/*
 * Khoros: $Id: tb_deps.c,v 1.2.2.2 1997/03/14 22:23:32 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: tb_deps.c,v 1.2.2.2 1997/03/14 22:23:32 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Routines to maintain toolbox dependencies
   >>>> 
   >>>>	Private: update_deps_in_file
   >>>>          kcms_find_tb_dependency
   >>>>	Static: 
   >>>>	Public: kcms_add_tb_dependency
   >>>>         kcms_delete_tb_dependency
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*-----------------------------------------------------------
| Routine Name: update_deps_in_file - update toolbox dependencies in file
|
| Purpose:      This routine is used to update toolbox dependencies
|               in the toolbox Imake configuration and include files
|
| Input:        routine - the calling routine
|               file_object - file object to update
|               file_description - text description of file
|               begin_key
|               end_key
|               format_string
|               convert_tb_to_upper
|               deplist
|
| Returns:      TRUE if successful, otherwise FALSE
|
| Written By:   Kevin Sullivan
| Date:         3-feb-97
----------------------------------------------------------*/
int
update_deps_in_file(
   kstring  routine,
   kobject  file_object,
   kstring  file_description,
   kstring  begin_key,
   kstring  end_key,
   kstring  format_string,
   int      convert_tb_to_upper,
   klist   *deplist)
{
   kfile    *file;
   kstring   path;
   int       status1, status2;
   kstring   begin, end, key1, key2;
   char      tbupper[KLENGTH];
   klist    *list;

   if ((file_object != NULL)
       && (kcms_get_attribute(file_object, KCMS_FOBJ_FULLNAME, &path)))
   {
      if ((file = kfopen(path, "r")) == NULL)
      {
         kerror(KCMS, routine,
                "Unable to open %s '%s'",
                file_description, path);
         errno = KCMS_METHODERR;
         return FALSE;
      }
      begin = kparse_file_scan(file, KPARSE_BOF, begin_key,
                               KLITERAL, NULL, &key1, &status1);
      end = kparse_file_scan(file, end_key, KPARSE_EOF,
                             KLITERAL, &key2, NULL, &status2);
      kfclose(file);
      if (status1 != KPARSE_OK ||
          status2 != KPARSE_OK)
      {
         kerror(KCMS, routine,
                "Unable to replace tag section in file '%s'", path);
         errno = KCMS_METHODERR;
         return FALSE;
      }
      if ((file = kfopen(path, "w")) == NULL)
      {
         kerror(KCMS, routine,
                "Unable to open '%s' to write %s",
                path, file_description);
         errno = KCMS_METHODERR;
         return FALSE;
      }
      kfputs(begin, file);
      kfputs(key1, file);
      kfputs("\n", file);
      for(list = deplist; list != NULL; list = klist_next(list))
      if (convert_tb_to_upper)
      {
         kstring_upper((kstring)list->client_data, tbupper);
         kfprintf(file, format_string, tbupper);
      }
      else
      {
         kfprintf(file, format_string, (kstring)list->client_data);
      }
      kfprintf(file, "\n%s", key2);
      kfputs(end, file);
      kfclose(file);
      kfree_and_NULL(key1);
      kfree_and_NULL(key2);
      kfree_and_NULL(begin);
      kfree_and_NULL(end);
      return TRUE;
   }
   else return FALSE;
}
/*-----------------------------------------------------------
| Routine Name: kcms_find_tb_dependency - find toolbox dependency
|
| Purpose:      This routine is used to find a toolbox dependency, 
|               i.e. to see if a toolbox depends on another toolbox
|
| Input:        tbobj   - The toolbox object
|               tbdep   - The name of the toolbox dependency to find
|
| Returns:      TRUE if found, otherwise FALSE
|
| Written By:   Kevin Sullivan
| Date:         13-mar-97
----------------------------------------------------------*/
int
kcms_find_tb_dependency(
   kobject     tbobj,
   kstring     tbdep
   )
{
   kstring   tbname;
   kstring   tbpath;
   klist    *list;
   int       found = FALSE;

   if (!kcms_get_attributes(tbobj, KCMS_NAME, &tbname,
                            KCMS_PATH, &tbpath,
                            KCMS_TB_DEPENDENCIES, &list,
                            NULL))
      return FALSE;

   for (; (list != NULL) && !found; list = klist_next(list))
   {
      if (kstrcmp((kstring)list->client_data, tbdep) == 0)
         found = TRUE;
      else
         if ((tbobj = kcms_open_toolbox((kstring)list->client_data)) == NULL)
            found = FALSE;
         else
            found = kcms_find_tb_dependency(tbobj, tbdep);
   }

   return found;
}

/************************************************************
* Routine Name: kcms_add_tb_dependency - add toolbox dependency
*
* Purpose:      This routine is used to add a toolbox dependency, 
*               i.e. to make a toolbox depend on another toolbox
*
* Input:        tbobj   - The toolbox object
*               tbdep   - The name of the toolbox dependency to add
*
* Returns:      TRUE if successful, otherwise FALSE
*
* Written By:   Kevin Sullivan
* Date:         23-jan-97
*************************************************************/
int
kcms_add_tb_dependency(
   kobject     tbobj,
   kstring     tbdep,
   int         force_flag
   )
{
   kstring   routine      = "kcms_add_tb_dependency";
   kstring   tbname;
   kstring   tbpath;
   char      command[KLENGTH];
   klist    *list;
   kobject   config_file, include_file;
   kstring   word;
   ktoken    token;

   if (!kcms_get_attributes(tbobj, KCMS_NAME, &tbname,
                            KCMS_PATH, &tbpath,
                            KCMS_TB_DEPENDENCIES, &list,
                            KCMS_TB_CONFIG_FILE, &config_file,
                            KCMS_TB_INCLUDE_FILE, &include_file,
                            NULL))
      return FALSE;

   if (kstrcmp(tbname, "BOOTSTRAP") == 0)
   {
      kerror(KCMS, routine,
             "The bootstrap toolbox may not depend on any other toolbox.");
      return FALSE;
   }

   word = kstrdup(tbdep);
   token = kstring_to_token(word);
   list = klist_add(list, (kaddr)token, (kaddr)word);
   if (!kcms_set_attribute(tbobj, KCMS_TB_DEPENDENCIES, list))
      return FALSE;

   /* write out toolbox imake config file, include file, and database */

   if (!update_deps_in_file(routine, config_file,
                            "toolbox Imake configuration file",
                            "/\\* -toolbox_include_toolboxes \\*/",
                            "/\\* -toolbox_include_toolboxes_end \\*/",
                            "\n#include %s_INCLUDE\n", TRUE, list))
      return FALSE;

   if (!update_deps_in_file(routine, include_file,
                            "toolbox include file",
                            "/\\* other toolbox includes go here \\*/",
                            "/\\* toplevel include for each library in this toolbox go below here \\*/",
                            "\n#include <%s.h>\n", FALSE, list))
      return FALSE;

   kcms_sync(tbobj);

   /* when creating a toolbox, objects dir is not yet created */
   ksprintf(command, "%s/objects", tbpath);
   if (kaccess(command, F_OK) == 0)
   {
      if (force_flag
       || kprompt(KSTANDARD, "Yes", "No", TRUE,
                  "After changing toolbox dependencies, \n"
                  "toolbox Makefiles must be regenerated. \n"
                  "Do you want me to do this for you? \n"))
      {
         if (!force_flag)
            kannounce(KCMS, routine, "Regenerating Makefiles for toolbox");
         ksprintf(command, "cd %s/objects;make Makefile;make Makefiles",
                  tbpath);
         ksystem(command);
      }
   }
   return TRUE;
}
/************************************************************
* Routine Name: kcms_delete_tb_dependency - delete toolbox dependency
*
* Purpose:      This routine is used to delete a toolbox dependency 
*
* Input:        tbobj   - The toolbox object
*               tbdep   - The name of the toolbox dependency to delete
*
* Returns:      TRUE if successful, otherwise FALSE
*
* Written By:   Kevin Sullivan
* Date:         23-jan-97
*************************************************************/
int
kcms_delete_tb_dependency(
   kobject     tbobj,
   kstring     tbdep,
   int         force_flag
   )
{
   kstring   routine      = "kcms_delete_tb_dependency";
   kstring   tbname;
   kstring   tbpath;
   char      command[KLENGTH];
   klist    *list;
   kobject   config_file, include_file;
   kstring   word         = tbdep;
   ktoken    token;

   if (!kcms_get_attributes(tbobj, KCMS_NAME, &tbname,
                            KCMS_PATH, &tbpath,
                            KCMS_TB_DEPENDENCIES, &list,
                            KCMS_TB_CONFIG_FILE, &config_file,
                            KCMS_TB_INCLUDE_FILE, &include_file,
                            NULL))
      return FALSE;

   if (strcmp(tbname, "bootstrap") == 0)
   {
      kerror(KCMS, routine,
             "The bootstrap toolbox may not depend on any other toolbox.");
      return FALSE;
   }

   token = kstring_to_token(word);
   list = klist_delete(list, (kaddr)token);

   /* all toolboxes but bootstrap must depend on at least bootstrap */
   if (list == NULL)
   {
      word = kstrdup("bootstrap");
      token = kstring_to_token(word);
      list = klist_add(list, (kaddr)token, (kaddr)word);
   }

   if (!kcms_set_attribute(tbobj, KCMS_TB_DEPENDENCIES, list))
      return FALSE;

   /* write out toolbox imake config file, include file, and database */

   if (!update_deps_in_file(routine, config_file,
                            "toolbox Imake configuration file",
                            "/\\* -toolbox_include_toolboxes \\*/",
                            "/\\* -toolbox_include_toolboxes_end \\*/",
                            "\n#include %s_INCLUDE\n", TRUE, list))
      return FALSE;

   if (!update_deps_in_file(routine, include_file,
                            "toolbox include file",
                            "/\\* other toolbox includes go here \\*/",
                            "/\\* toplevel include for each library in this toolbox go below here \\*/",
                            "\n#include <%s.h>\n", FALSE, list))
      return FALSE;

   kcms_sync(tbobj);

   if (force_flag || kprompt(KSTANDARD, "Yes", "No", TRUE,
               "After changing toolbox dependencies, \n"
               "toolbox Makefiles must be regenerated. \n"
               "Do you want me to do this for you? \n"))
   {
      if (!force_flag)
         kannounce(KCMS, routine, "Regenerating Makefiles for toolbox");
      ksprintf(command, "cd %s/objects;make Makefile;make Makefiles",
               tbpath);
      ksystem(command);
   }

   return TRUE;
}

