/*
 * Khoros: $Id: tb_public.c,v 1.14.2.1 1997/03/06 21:58:08 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: tb_public.c,v 1.14.2.1 1997/03/06 21:58:08 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Toolbox public functions
   >>>>   Public:
   >>>>                 kcms_create_toolbox()
   >>>>                 kcms_add_toolbox()
   >>>>                 kcms_get_objects()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"
#include "toolboxP.h"
#include "databaseP.h"


/************************************************************
* Routine Name:	kcms_create_toolbox - create a new toolbox object
*
* Purpose:	This function creates the directory structure for a new
*		toolbox, and the minimum number of configuration files
*		required for a toolbox.
*
* Input:	tbname	- The name to be given to the new toolbox.
*		tbpath	- The path to the top of the toolbox.
*
* Returns:	If the toolbox was successfully created,
*		then an object referenced is returned for the new toolbox.
*		If the toolbox could not be created,
*		then NULL is returned,
*		and \f(CW\s-1errno\s+1\fP is set to signify the error.
*
* Written By:	Neil Bowers
* Date:		26-mar-94
*************************************************************/
kobject 
kcms_create_toolbox(
   kstring  tbname,
   kstring  tbpath)
{
   kstring  routine  = "Create Toolbox";
   kobject  toolbox;
   char     fullpath[KLENGTH];


   if (tbname == NULL)
      kerror(KCMS, routine, "NULL string passed for toolbox name.");

   else if (tbpath == NULL)
      kerror(KCMS, routine, "NULL string passed for toolbox path.");

   else if (!kcms_legal_identifier(tbname, KOBJ_CMSTB))
      kerror(KCMS, routine, "Invalid toolbox name -- "
	     "\"%s\". \n"
             "A valid name must start with an alphabetic character \n"
             "and contain only alphanumeric characters or the \n"
             "underscore character, and may not be a reserved word.", tbname);

   else if (ktoolbox_path(tbname) != NULL)
      kerror(KCMS,routine,"There is already a `%s' toolbox.",tbname);

   else if (kfullpath(tbpath,NULL,fullpath) == NULL)
      kerror(KCMS, routine, "Could not expand path for Toolbox '%s'", tbname);

   else if (!ktoolbox_add(tbname, fullpath))
      kerror(KCMS, routine,
	     "Could not update a Toolbox file\n"
	     "Please make sure you have write\n"
	     "permission on your $HOME/.Toolboxes.\n");

   else
   {
      toolbox = kcms_tb_create(tbname,fullpath);
      if (toolbox != NULL)
	 return toolbox;
      ktoolbox_delete(tbname);
   }

   /*-- failed if we hit here --*/
   return NULL;
}

/************************************************************
* Routine Name:	kcms_add_toolbox - add an existing toolbox
*
* Purpose:	This routine adds an existing toolbox to your
*		khoros working environment, making that toolbox
*		visible to the user.
*
* Input:	tbname	- The name of an existing toolbox.
*		tbpath	- The path to the top of the toolbox.
* Output:	
* Returns:	A kobject pointer to the toolbox object,
*		or NULL if the toolbox could not be opened.
*		In the latter case,
*		\f(CW\s-1errno\s+1\fP is set to signify the error.
*
* Written By:	Neil Bowers
* Date:		24-mar-94
*************************************************************/
kobject 
kcms_add_toolbox(
   kstring  tbname,
   kstring  tbpath)
{
   kstring  routine            = "Add Toolbox";
   kobject  toolbox	       = NULL;
   char	    fullpath[KLENGTH];
   char     dbpath[KLENGTH];


   if (tbname == NULL || tbpath == NULL)
      kerror(KCMS,routine,"Name and/or Path are NULL\n");

   else if (ktoolbox_path(tbname) != NULL)
      kerror(KCMS,routine, "Toolbox '%s' is already installed", tbname);

   else if (kfullpath(tbpath,NULL,fullpath) == NULL)
      kerror(KCMS,routine, "Could not expand path for Toolbox '%s'", tbname);

   else
   {
      if (kstring_3cat(fullpath, "/", TOOLBOX_DB_PATH, dbpath) == NULL)
	 kerror(KCMS, routine,
	        "Couldn't determine path to toolbox database (toolbox: %s)\n",
	        tbname);

      if (kaccess(dbpath, R_OK) == -1)
      {
	 errno = KCMS_ENOTFOUND;
	 kerror(KCMS, routine, "Toolbox `%s' database does not exist", tbname);
      }

      else if (!ktoolbox_add(tbname, fullpath))
	 kerror(KCMS,routine,"Call to ktoolbox_add() failed\n");

      else if ((toolbox = kcms_open_toolbox(tbname)) == NULL)
      {
	 kerror(KCMS,routine,"Cannot open toolbox");
	 (void)ktoolbox_delete(tbname);
      }
   }

   return toolbox;
}

/************************************************************
* Routine Name:	kcms_get_objects - get a list of typed objects from toolbox
*
* Purpose:	This routine is used to query a toolbox for a list of all
*		software objects of a given type.
*
* Input:	toolbox - The toolbox object to query.
*		reqtype	- The type of object(s) we're interested in.
*			  This can be one or more of the supported software
*			  object types, which are:
*			  !\f(CW\s-2
*			  !   KCMS_KROUTINE
*			  !   KCMS_XVROUTINE
*			  !   KCMS_LIBRARY
*			  !   KCMS_PANE
*			  !   KCMS_SCRIPT
*			  !\s+2\fP
*			  The object type identifiers can be or'd together
*			  to request more than one object type:
*			  !\f(CW\s-2
*			  !   ( KCMS_KROUTINE | KCMS_XVROUTINE | KCMS_PANE )
*			  !\s+2\fP
*
* Returns:	A pointer to the first element of a klist of software
*		objects, or NULL if no software objects of the specified
*		type were found.
*
* Written By:	Neil Bowers
* Date:		23-jun-94
*************************************************************/
klist *
kcms_get_objects(
   kobject       toolbox,
   unsigned long reqtype)
{
   klist        *softobjects;
   klist        *result       = NULL;
   kobject       object;
   unsigned long otype;
   ktoken        token;


   if (!kcms_get_attribute(toolbox, KCMS_TB_SOFTWARE_OBJECTS, &softobjects))
      return NULL;

   for (; softobjects != NULL; softobjects = klist_next(softobjects))
   {
      object = (kobject)klist_clientdata(softobjects);
      if (object != NULL
	  && kcms_get_attribute(object, KCMS_CMOBJ_TYPE, &otype)
	  && kcms_get_attribute(object, KCMS_CMOBJ_ONAME_TKN, &token))
      {
	 if ((otype & reqtype) != KCMS_NONE)
	    result = klist_add(result, (kaddr)token, (kaddr)object);
      }
   }

   return result;
}
