/*
 * Khoros: $Id: tb_utils.c,v 1.12.2.1 1997/04/25 21:18:05 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: tb_utils.c,v 1.12.2.1 1997/04/25 21:18:05 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Toolbox creation routines
   >>>>		
   >>>>   Public:
   >>>>                 kcms_get_copyright()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"
#include "toolboxP.h"

typedef struct
{
   unsigned long language;
   kstring       opening;
   kstring       prefix;
   kstring       closing;
} CommentInfo;

static CommentInfo comments[] =
{
   {KCMS_LANG_NONE,       NULL,   NULL,      NULL},
   {KCMS_LANG_C,          "/*",   " * ",     " */"},
   {KCMS_LANG_FORTRAN,    NULL,   "C ",      NULL},
   {KCMS_LANG_PERL,       NULL,   "# ",      NULL},
   {KCMS_LANG_CSH,        NULL,   "# ",      NULL},
   {KCMS_LANG_SH,         NULL,   "# ",      NULL},
   {KCMS_LANG_KSH,        NULL,   "# ",      NULL},
   {KCMS_LANG_CPLUSPLUS,  NULL,   "// ",     NULL},
   {KCMS_LANG_ADA,        NULL,   "-- ",     NULL},

   {KCMS_LANG_ROFF,       NULL,   ".\\\" ",  NULL},
   {KCMS_LANG_TEX,        NULL,   "% ",      NULL},
   {KCMS_LANG_KHOROS_UIS, NULL,   "# ",      NULL},
   {KCMS_LANG_HTML,       "<!--", "  -- ",   "  -->"}
};

/************************************************************
* Routine Name: kcms_get_copyright - get copyright comment for toolbox
*
* Purpose:	This routine is used to determine the copyright statement
*		comment which should be used for a specified toolbox.
*
*		If the relevant copyright attribute is defined then this
*		will be used, otherwise the copyright statement from
*		the BOOTSTRAP toolbox will be used.
*
* Input:	toolbox   - The toolbox object we want a copyright for.
*	        attribute - The attribute identifier for the copyright to use;
*			    each toolbox can have two copyright statements
*			    associated with it, short and long.
*			    Legal values for this parameter are:
*			    \f(CW\s-2KCMS_TB_COPYRIGHT_SHORT\fP\s+2, or
*			    \f(CW\s-2KCMS_TB_COPYRIGHT_SHORT\fP\s+2.
*		language  - The language we want this for.
*			    The following languages are currently supported:
*			    !\f(CW\s-2
*			    !   KCMS_LANG_C
*			    !   KCMS_LANG_FORTRAN
*			    !   KCMS_LANG_PERL
*			    !   KCMS_LANG_CSH
*			    !   KCMS_LANG_SH
*			    !   KCMS_LANG_KSH
*			    !   KCMS_LANG_CPLUSPLUS
*			    !   KCMS_LANG_ADA
*			    !   KCMS_LANG_ROFF
*			    !   KCMS_LANG_TEX
*			    !   KCMS_LANG_HTML
*			    !   KCMS_LANG_KHOROS_UIS
*			    !\s+2\fP
*
* Returns:	If the appropriate copyright could be found, then a string
*		is returned which contains the copyright formatted within
*		a comment for the specified language.
*		If a copyright could not be found, or some other error
*		occurred, then NULL will be returned.
*
*		NOTE: the string returned is kmalloc'd for your use.
*		You should free the string when you are done.
* Written By:	Neil Bowers
* Date:		21-jun-94
*************************************************************/
kstring
kcms_get_copyright(
   kobject       toolbox,
   kcms_attr     attribute,
   unsigned long language)
{
   kstring       routine    = "kcms_get_copyright()";
   kstring       copyright;
   kstring       tbname;
   CommentInfo  *cinfo      = NULL;
   kstring       result     = NULL;
   kstring       tmp;
   size_t        i;
   char        **list;
   ssize_t       nlines;


   if (!kcms_legal_object(toolbox, routine))
      return NULL;

   /*-- get the raw copyright string from the toolbox -----------------*/
   if (!kcms_get_attribute(toolbox, attribute, &copyright)
       || !kcms_get_attribute(toolbox, KCMS_NAME, &tbname))
      return NULL;

   if (copyright == NULL)
   {
      kerror(KCMS, routine, "Could not determine a copyright string.");
      return NULL;
   }

   /*-- split the copyright into lines --------------------------------*/
   list = kparse_string_delimit(copyright, "\n", KDELIM_CLEAN, &nlines);

   for (i=0; i<knumber(comments); i++)
   {
      if (comments[i].language == language)
	 cinfo = &comments[i];
   }

   if (cinfo == NULL)
   {
      kerror(NULL, routine, "Could not find comment information for the "
	     "language specified (%d).\n", language);
      return NULL;
   }

   /*-- does the language have an opening comment construct? ----------*/
   if (cinfo->opening != NULL)
      result = kstring_cat(cinfo->opening, "\n", NULL);

   /*-- prepend the comment prefix to each line, add to comment -------*/
   for (i=0; i<nlines; i++)
   {
      tmp = kstring_3cat(result, cinfo->prefix, list[i], NULL);
      kfree_and_NULL(result);
      result = kstring_cat(tmp, "\n", NULL);
      kfree_and_NULL(tmp);
   }
   karray_free(list, KSTRING, (size_t) nlines, NULL);

   /*-- does the language have an closing comment construct? ----------*/
   if (cinfo->closing != NULL)
   {
      tmp = kstring_3cat(result, cinfo->closing, "\n", NULL);
      kfree_and_NULL(result);
      result = tmp;
   }

   return result;
}

/*-----------------------------------------------------------
| Routine Name:	kcms_tb_remove_cmobj - delete a partial object from toolbox
|
| Purpose:	This function is used to remove the traces of a partial
|		software object from a toolbox.
|
| Input:	toolbox     - toolbox object to delete from
|		object_name - name of software object to remove
|
| Returns:	TRUE (1) on success, FALSE (0) otherwise
|
| Written By:	Neil Bowers
| Date:		23-May-96
------------------------------------------------------------*/
int
kcms_tb_remove_cmobj(
   kobject  toolbox,
   kstring  object_name)
{
   kstring         routine        = "kcms_tb_remove_cmobj()";
   ktoken          oname_token;
   klist          *target;
   unsigned long   object_types[] = {KCMS_KROUTINE,
			 	     KCMS_XVROUTINE,
				     KCMS_LIBRARY,
				     KCMS_PANE,
				     KCMS_SCRIPT};
   int             i, j;
   kstring         typestring;
   char            object_path[KLENGTH];
   char            dir_path[KLENGTH];
   char            inc_path[KLENGTH];
   char            path[KLENGTH];
   kstring        *dirlist;
   size_t          dcount, count;


   /*-------------------------------------------------------------------
   | first we see if it's in the list of software objects in the toolbox
   -------------------------------------------------------------------*/
   if (toolbox->sw_objects != NULL)
   {
      oname_token = kstring_to_token(object_name);
      target = klist_locate(toolbox->sw_objects, oname_token);
      if (target != NULL)
      {
	 toolbox->sw_objects = klist_delete(toolbox->sw_objects, oname_token);
	 toolbox->update |= KCMS_UPDATE_SYNC;
	 kcms_sync(toolbox);
      }
   }

   /*-------------------------------------------------------------------
   | See if there is any of the object left in objects/<object-type>
   -------------------------------------------------------------------*/
   for (i = 0; i < knumber(object_types); i++)
   {
      typestring = kcms_attr_int2string(KCMS_CMOBJ_TYPE, object_types[i]);
      if (typestring == NULL)
	 continue;

      ksprintf(dir_path, "$%s/objects/%s", toolbox->name, typestring);

      ksprintf(object_path, "%s/%s", dir_path, object_name);

      if (kaccess(object_path, F_OK | X_OK) == 0)
      {
	 /*-------------------------------------------------------------
	 | If there are no other objects of this type, then delete the
	 | type directory as well.
	 -------------------------------------------------------------*/
	 dirlist = karray_dirlist(dir_path, NULL, NULL, KDIR, FALSE, &count);
	 for (j = 0, dcount = 0; j < count; j++)
	 {
	    if (kstrcasecmp(dirlist[j], object_name) == 0)
	       dcount++;
	    else if (kstrcmp(dirlist[j], "CVS") == 0)
	       dcount++;
	    else if (kstrcmp(dirlist[j], "..") == 0)
	       dcount++;
	 }
	 if (dcount != count)
	    kstrcpy(dir_path, object_path);
	 karray_free(dirlist, KSTRING, count, NULL);        
	 if (!kremove_dir(dir_path))
	    kerror(KCMS, routine, "\n"
		   "The object's directory structure could not "
		   "be completely deleted.\n"
		   "The most likely reason was that you were in "
		   "the directory you were trying to delete.\n");

         /*-- if a library, it's includes may be need to be removed --*/
         if (object_types[i] == KCMS_LIBRARY)
         {
            kcms_cmobj_destroy_checklib(toolbox->name, toolbox->path,
                                        object_name, object_name);
            ksprintf(inc_path, "$%s/include/%s", toolbox->name, object_name);

            if (!kremove_dir(inc_path))
               kerror(KCMS, routine, "\n"
                      "The library's include directory could not "
                      "be completely deleted.\n"
                      "The most likely reason was that you were in "
                      "the directory you were trying to delete.\n");
         }

	 /*-- remake imake and make files ----------------------------*/
	 if (kdirname(dir_path, path) != NULL)
	    kcms_dir_imakeandmake(path, toolbox->name, 1);

	 return TRUE;
      }
   }

   return FALSE;
}
