/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Routines to Generate .3-like html file for library
   >>>>
   >>>>  Private:
   >>>>                 kgen_clui_generate_htm3file()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*---------------------------------------------------------------
|
|  Routine Name: kgen_clui_generate_htm3file
|
|       Purpose: Creates the htm3 file for the library as a whole
|
|         Input: library   -  library object
|                prog_spec -  pointer to the internal PS struct
|                opath     -  path to the library object
|                tb_name   -  name of toolbox
|		 name      -  name of library object being generated
|
|        Output: Returns TRUE on success, FALSE on failure
|    Written By: Kevin Sullivan, after Danielle Argiro and Steve Jorgensen
|          Date: Oct 24, 1996 
| Modifications: 
|
--------------------------------------------------------------*/
int
kgen_clui_generate_htm3file(
   kobject library,
   char    **prog_spec,
   char    *opath,
   char    *tb_name,
   char    *oname)
{
   kstring      routine     = "kgen_clui_generate_htm3file()";
   kobject      toolbox     = NULL;
   kobject      file_object = NULL;
   int          force_flag  = FALSE;
   kfile       *file;
   kstring      filepath;
   kstring      copyright   = NULL;
   char         temp[KLENGTH], directory[KLENGTH];
   char        *pagename;
   header_info *header;
   klist       *headerlist;
   int          isnewfile   = FALSE;
   kstring      short_description = NULL;


   /*-- get file object representing htm3 page -----------------------*/
   if (!kcms_get_attributes(library,
			    KCMS_CMOBJ_GEN_HTM3,   &file_object,
			    KCMS_CMOBJ_SHORT_DESC, &short_description,
			    KCMS_END)
       || !kcms_query_bit(library, KCMS_CMOBJ_FLAGS, KCMS_BIT_CMOBJ_FORCE,
			  &force_flag))
      return FALSE;

   /*-------------------------------------------------------------------
   | 1st time htm3 page has been created - create file object,
   | so that kcms registers the htm3 page in the kcms database
   -------------------------------------------------------------------*/
   if (file_object == NULL)
   {
      ksprintf(temp, "%s/html/%s.htm", opath, oname);
      file_object = kcms_create_fileobj(library, temp, NULL,
					KCMS_FOBJ_TYPE_HTML,
					KCMS_FOBJ_SUBTYPE_HTM3,
					KCMS_FOBJ_GEN_HTM3,
					KCMS_FOBJ_ACCESS_RDWR);
      if (file_object == NULL)
	 return FALSE;
      isnewfile = TRUE;
      filepath = temp;
   }

   /*-------------------------------------------------------------------
   | The htm3 file has already been created & added to kcms database;
   | simply get the path to the htm3 page.
   -------------------------------------------------------------------*/
   else if (!kcms_get_attribute(file_object, KCMS_PATH, &filepath))
      return FALSE;

   /*-------------------------------------------------------------------
   | If somehow the database GHOST_HTM3 key got deleted from the kcms
   | database, but the htm3 file really is there, don't want to
   | clobber it!
   -------------------------------------------------------------------*/
   if (isnewfile && kaccess(filepath, R_OK) == 0)
   {
      kerror(KCODEGEN, routine,
	     "WARNING!  CMS database appears to have been corrupted;  "
	     "cms dbm key GHOST_HTM3 is missing, but the %s htm3 does "
	     "exist.  Regeneration of htm3 file will NOT take place. "
	     "Adding GHOST_HTM3 cms key to database; next run of "
	     "ghostwriter should be OK.", oname);
      return FALSE;
   }

   /*-------------------------------------------------------------------
   | get the directory to the htm3 file.  if the directory does
   | not exist, it will get created by kmake_dir();  if kmake_dir()
   | cannot create it, it will return the ENOENT error.
   -------------------------------------------------------------------*/
   (void) kdirname(filepath, directory);
   if (!kmake_dir(directory, (mode_t) 0777) && errno != ENOENT)
   {
      kerror(KCODEGEN, routine,
	     "Can't create directory %s in which to deposit htm3 file",
	     directory);
      return FALSE;
   }

   /*-------------------------------------------------------------------
   | see if htm3 file already exists.  If so, prompt to over-write,
   | return TRUE if the answer is NO
   -------------------------------------------------------------------*/
   if (!force_flag && koverwrite(kget_notify(), filepath) == FALSE)
      return TRUE;

   /*-- create & open htm3 file --------------------------------------*/
   if (kaccess(filepath, R_OK) != 0)
      isnewfile = TRUE;
   if ((file = kfopen(filepath, "w"))== NULL)
   {
      kerror(KCODEGEN, routine, "Could not create file '%s'", filepath);
      return FALSE;
   }
	
   /*-- print out first line of htm3 page with time -----------*/
   kfprintf(file, "<HTML>\n");
   kfprintf(file, "<HEAD>\n");
   kfprintf(file, "     <TITLE> %s Online Help Page </TITLE>\n", oname);
   kfprintf(file, "</HEAD>\n");
   kfprintf(file, "<BODY>\n");
	
   kfprintf(file, "<H2>%s (%s) library</H2><HR>\n",
	    kstring_upper(tb_name, NULL), oname);

   /*-- print out name & description ---------------------------------*/
   kfprintf(file, "<H3>LIBRARY NAME </H3>\n");
   if (short_description == NULL)
      short_description = "(no short description specified)";
   kfprintf(file, "%s - %s\n", oname, short_description);
	
   /*-- print out description ----------------------------------------*/
   kfprintf(file, "<H3>DESCRIPTION </H3>\n<P>\n");
   if (kstrlen(prog_spec[MAN3_LONGDESC]) > 0) 
      kgen_roff2html(file, prog_spec[MAN3_LONGDESC]);

   /*-- print out the list of library functions ----------------------*/
   kfprintf(file, "<H3>LIST OF LIBRARY FUNCTIONS </H3>\n<DL>\n");
   kcms_get_attribute(library, KCMS_CMOBJ_HEADERLIST, &headerlist);
   while (headerlist!= NULL)
   {
      header = (header_info *) headerlist->client_data;
      if (header->name != NULL)
      {
         pagename = kchar_replace(header->name, ':', '_', NULL);
         if (header->shortdesc != NULL)
	    kfprintf(file, "<DT> <A HREF=\"%s.htm\"> %s </A><DD> %s\n", 
                     pagename, header->name, header->shortdesc);
         else
	    kfprintf(file, "<DT> <A HREF=\"%s.htm\"> %s </A><DD>\n", 
                     pagename, header->name);
         kfree_and_NULL(pagename);
      }
      headerlist = klist_next(headerlist);
   }
   kfprintf(file, "</DL>\n");

   /*-- print out additional information -----------------------------*/
   kfprintf(file, "<H3>ADDITIONAL INFORMATION </H3>\n");
   if (kstrlen(prog_spec[MAN3_ADDINFO]) > 0) 
      kgen_roff2html(file, prog_spec[MAN3_ADDINFO]);

   /*-- print out files ----------------------------------------------*/
   kfprintf(file, "<H3>LOCATION OF SOURCE FILES </H3>\n");
   kfprintf(file, "<CODE>$%s/objects/library/%s/src</CODE>\n", 
	    kstring_upper(tb_name, NULL), oname);

   kfprintf(file, "<H3>LOCATION OF PUBLIC INCLUDE FILE </H3>\n");
   kfprintf(file, "<CODE>$%s/include/%s/%s.h</CODE>\n", 
	    kstring_upper(tb_name, NULL), oname, oname);

   kfprintf(file, "<H3>YOU MUST INCLUDE </H3>\n");
   kfprintf(file, "<CODE>#include &lt%s.h&gt</CODE>\n", 
            kstring_lower(tb_name, NULL));
	
   /*-- SEE ALSO section ---------------------------------------------*/
   kfprintf(file, "<H3>SEE ALSO </H3>\n");
   if (kstrlen(prog_spec[MAN3_SEEALSO]) > 0) 
      kgen_roff2html(file, prog_spec[MAN3_SEEALSO]);
	
   /*-- RESTRICTIONS section -----------------------------------------*/
   kfprintf(file, "<H3>SEE MANUAL </H3>\n");
   if (kstrlen(prog_spec[MAN3_MANUAL]) > 0) 
      kgen_roff2html(file, prog_spec[MAN3_MANUAL]);

   kfprintf (file, "<H3>COPYRIGHT </H3>\n");
	
   /*-- COPYRIGHT section --------------------------------------------*/
   if ((toolbox = kcms_open_toolbox(tb_name)) == NULL)
   {
      if (tb_name == NULL)
	 kerror(KCODEGEN, routine,
		"Cannot get copyright for lib html file, NULL input toolbox.  "
		"Copyright will be left blank in the file");
      else
	 kerror(KCODEGEN, routine,
		"Cannot open toolbox '%s' to get copyright for lib html file.  "
		"Copyright will be left blank in the file",
		tb_name);
   }
   else
   {
      copyright = kcms_get_copyright(toolbox, KCMS_TB_COPYRIGHT_SHORT,
				     KCMS_LANG_NONE);
      kcms_close(toolbox);
   }
   if (copyright != NULL)
   {
      kfprintf(file, "%s\n", copyright);
      kfree_and_NULL(copyright);
   }

   kfprintf(file, "</BODY>\n");
   kfprintf(file, "</HTML>\n");

   kfclose(file);

   return TRUE;
}


