/*
 * Khoros: $Id: Memory.c,v 1.2.2.1 1997/05/04 21:27:33 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: Memory.c,v 1.2.2.1 1997/05/04 21:27:33 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $CONVERT/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		UNIX Memory Transport Routines
   >>>>		
   >>>>   Static:
   >>>>                 memory_tempnam()
   >>>>                 memory_open()
   >>>>                 memory_close()
   >>>>                 memory_read()
   >>>>                 memory_write()
   >>>>                 memory_lseek()
   >>>>                 memory_access()
   >>>>                 memory_unlink()
   >>>>                 memory_lock()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"
#define KMEM_DEF
#if !defined(KMEM_DEF)
TransportInformation memory_transport[] = { {NULL} };
#else

static int memory_tempnam	 PROTO((const char *, const char *, char *));
static int memory_open	 PROTO((const char *, int, mode_t, kfile *));
static int memory_close	 PROTO((kfile *));
static ssize_t memory_read	 PROTO((kfile *, kaddr, size_t));
static ssize_t memory_write PROTO((kfile *, kaddr, size_t));
static off_t   memory_lseek PROTO((kfile *, off_t, int));

TransportInformation memory_transport[] =
{
    {
	"Internal Memory",
	"memory",
	FALSE,
	memory_tempnam,
	memory_open,
	memory_close,
	memory_read,
	memory_write,
	memory_lseek,
	NULL,
	NULL,
	NULL,
	NULL,
    }
};

/*
 *  Internal resource structures for "standard file" transport
 */
typedef struct
{
	int memory_id;
	size_t size;
	size_t position;
} ResourceStruct;

/* a hash of active memory ids */
static khash *memory_ids = NULL;


/*-------------------------------------------------------------------*
|  
|		  Private Memory Routines
|  
--------------------------------------------------------------------*/
/*ARGSUSED*/
static void memory_exit(kexit_status status, kaddr client_data)
{
   khash_free(memory_ids, NULL);
   memory_ids = NULL;
}

static void memory_initialize(void)
{
   static int initialized = FALSE;
   if (initialized == TRUE)
      return;
   initialized = TRUE;
   kexit_handler(memory_exit, NULL);
   memory_ids = khash_init(KINT, 10);
   khash_add(memory_ids, 0, NULL);
}

static int memory_availid(void)
{
   int id = 1;
   while (khash_check(memory_ids, id) == TRUE)
      id++;
   return(id);
}
static int memory_addid(int id)
{
   kaddr bogus = (kaddr) id;
   if (id <= 0)
      return FALSE;
   memory_initialize();
   if (khash_add(memory_ids, id, bogus) == FALSE)
      return FALSE;
   return TRUE;
}

static int memory_delid(int id)
{
   if (id <= 0) return FALSE;
   memory_initialize();
   if (!memory_ids)
      return TRUE;
   if (khash_delete(memory_ids, id) == FALSE)
      return FALSE;
   return TRUE;
}
/*-------------------------------------------------------------------*
|  
|		  Memory Routines
|  
--------------------------------------------------------------------*/

/*-------------------------------------------------------------
|  
|  Routine Name: memory_tempnam
|  
|       Purpose: This function initializes a tempnam for a "memory"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"memory=XXXXXXXXXXX"
|		  	     ...etc....
|  
|         Input:  identifier - the transport identifier
|		  template   - the filename template to be modeled
|        Output:  result     - the newly created transport name
|    Written By:  Mark Young and Steve Jorgensen
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int memory_tempnam(
   const char *identifier,
   const char *ktemplate,
   char *result)
{
	/*
	 *  Store the unique key so that we can use it later.
	 */
	memory_initialize();
	(void) sprintf(result,"memory=%d", memory_availid());

	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: 	memory_open
|  
|       Purpose: This function opens a "internal memory" connection.  It is the
|	         internal driver to open a file, which is called by
|	         the user when they call kopen().  kopen() calls when
|	         it gets an open request with a filename of the following
|	         syntax:
|  
|		  	"memory=XXXXXXXXX"
|         Input: 
|        Output:  returns 0 or -1 if an error occurs
|    Written By:  Mark Young and Steve Jorgensen
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int memory_open(
   const char *path,
   int    flags,
   mode_t mode,
   kfile  *file)
{
	int	   id;
	ResourceStruct *resources;

	/*
	 *  Get internal memory (shmat) protections from the open permission
	 *  flags...
	 */
	if (!kfile_iswrite(file) && !kfile_isread(file))
	{
	   errno = KMISSING_INPUT;
	   kerror("kutils", "memory_open", "Memory Transport is only useful "
		  "if opened KOPEN_RDWR");
	   return -1;
	}

        if (sscanf(path, "memory=%d", &id) != 1)
	{
           if (sscanf(path,"%d", &id) != 1)
           {
	      kinfo(KSYSLIB, "Unable to get memory id from path\n");
	      errno = KINVALID_FILE;
	      return -1;
           }
	}

        memory_initialize();
	if (khash_check(memory_ids, id) == TRUE)
	{
	   kinfo(KSYSLIB, "Trying to use id '%d', but it's already in use", id);
	   errno = EBADF;
	   return -1;
	}

	if (!memory_addid(id))
	{
	   kinfo(KSYSLIB, "Couldn't add '%d' to 'in use' hash", id);
	   return -1;
	}

	if ((resources = (ResourceStruct *)kcalloc(1, sizeof(ResourceStruct)))
	    == NULL)
	{
           kinfo(KSYSLIB, "memory_open: Not enough memory to kcalloc (%d) "
                          "bytes for the khoros socket structure.\n",
                          sizeof(ResourceStruct));
           return(-1);
	}
	resources->memory_id = id;
	file->resources = resources;
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: memory_close
|  
|       Purpose: This function closes a "memory".  It is the internal
|	         driver to close a data shm segment, which is called by the user
|	         when they call kclose().  kclose() calls the "internal memory"
|	         internal drivers by the nature of the transport, which is
|	         dictated by the initial kopen().
|  
|         Input:  file - the kfile structure.
|        Output:  returns whether we were able to close the file
|    Written By:  Mark Young and Steve Jorgensen
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int memory_close(
   kfile *file)
{
	int	  id;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call shmdt() to detach from internal memory segment
	 */
	id = resources->memory_id;
	if (memory_delid(id) == FALSE)
	{
	   errno = EBADF;
	   return -1;
	}
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: memory_read
|  
|       Purpose: This function reads a "internal memory".  It is the internal
|	         driver to read data from a shm, which is called by the user
|	         when they call kread().  kread() calls the "internal memory"
|	         internal drivers by the nature of the transport, which is
|	         dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes read from the shm
|    Written By:  Mark Young and Steve Jorgensen
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t memory_read(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	resources->position += nbytes;
	if (resources->position >= resources->size)
	   nbytes -= (resources->position - resources->size);

	resources->position = kmin(resources->size, resources->position);
	return((ssize_t) nbytes);
}

/*-------------------------------------------------------------
|  
|  Routine Name: memory_write
|  
|       Purpose: This function writes to a "internal memory".  It is the internal
|	         driver to write data from the supplied data array to a shm,
|	         which is called by the user when they call kwrite().  kwrite()
|	         calls the "internal memory" internal drivers by the nature of
|	         the transport, which is dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young and Steve Jorgensen
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t memory_write(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;
	resources->position += nbytes;
	resources->size = kmin(resources->size, resources->position);
	return((ssize_t) nbytes);
}

/*-------------------------------------------------------------
|  
|  Routine Name: memory_write
|  
|       Purpose: This function writes to a "internal memory".  It is the internal
|	         driver to write data from the supplied data array to a shm,
|	         which is called by the user when they call kwrite().  kwrite()
|	         calls the "internal memory" internal drivers by the nature of
|	         the transport, which is dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young and Steve Jorgensen
|  
-------------------------------------------------------------*/
static off_t   memory_lseek(
	kfile *file,
	off_t offset,
	int   whence)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call lseek() since "kfile" is just a regular shm
	 */
	if (whence == KSEEK_SET)
	   resources->position = offset;
	else if (whence == KSEEK_CUR)
	   resources->position += offset;
	else if (whence == KSEEK_END)
	   resources->position = resources->size + offset;
	else
	{
	   errno = EINVAL;
	   kinfo(KSYSLIB, "memory_lseek:  Invalid whence argument value for "
                          "file '%s'", file->path);
	   return(-1);
	}

	if (resources->position > resources->size)
	{
	   return(-1);
	}
	return(resources->position);
}

#endif /* KMEM_DEF */
