/*
 * Khoros: $Id: Pipe.c,v 1.24.2.2 1997/04/21 22:03:52 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: Pipe.c,v 1.24.2.2 1997/04/21 22:03:52 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		UNIX Pipe Transport Routines
   >>>>		
   >>>>   Static:
   >>>>                 pipe_tempnam()
   >>>>                 pipe_open()
   >>>>                 pipe_close()
   >>>>                 pipe_read()
   >>>>                 pipe_write()
   >>>>                 pipe_access()
   >>>>                 pipe_unlink()
   >>>>                 pipe_descriptors()
   >>>>   Public:
   >>>>                 kpinfo()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"	

static int pipe_tempnam	    PROTO((const char *, const char *, char *));
static int pipe_open	    PROTO((const char *, int, mode_t, kfile *));
static int pipe_close	    PROTO((kfile *));
static ssize_t pipe_read    PROTO((kfile *, kaddr, size_t));
static ssize_t pipe_write   PROTO((kfile *, kaddr, size_t));
static int pipe_access	    PROTO((const char *, const char *, int));
static int pipe_unlink	    PROTO((const char *, const char *));
static int pipe_descriptors PROTO((kfile *, int *, int *));

TransportInformation pipe_transport[] =
{
    {
	"Standard Unix Pipe",
	"pipe",
	FALSE,
	pipe_tempnam,
	pipe_open,
	pipe_close,
	pipe_read,
	pipe_write,
	NULL,
	pipe_access,
	pipe_unlink,
	NULL,
	pipe_descriptors,
    }
};

/*
 *  Internal resource structures for "standard pipe" transport
 */
typedef struct
{
	pid_t pid;
	int   fid[2];
} ResourceStruct;

/*-------------------------------------------------------------------*
|  
|                 Private Pipe Transport Routines
|  
--------------------------------------------------------------------*/

/*ARGSUSED*/
static pid_t pipe_child(
   kfile *file,
   const char *command,
   int   *fid)
{
	pid_t pid;
	char  *tmp, **args;
	int   to_child[2], from_child[2];

	if (kfile_isread(file) && pipe(from_child) == -1)
	   return(-1);
	if (kfile_iswrite(file) && pipe(to_child) == -1)
	   return(-1);

	tmp = kexpand_variable(command, NULL);
	args = kcommand_to_args(tmp);
	if ((pid = kfork()) == 0)
	{
	   if (kfile_iswrite(file))
	   {
	      dup2(to_child[0], STDIN_FILENO);
	      close(to_child[0]); close(to_child[1]);
	   }

	   if (kfile_isrdwr(file))
	      dup2(from_child[1], STDERR_FILENO);

	   /*
	    *  this is really bad, but we have no time.  phantomd really
	    *  needs to have stdout and stderr duped, but not stdin.  So
	    *  for now, put in this kludge to have phantomd always dup it's
	    *  stderr.
	    */
	   else if (kstrcmp(kprog_get_program(), "phantomd") == 0)
	      dup2(from_child[1], STDERR_FILENO);

	   if (kfile_isread(file))
	   {
	      dup2(from_child[1], STDOUT_FILENO);
	      close(from_child[1]); close(from_child[0]);
	   }
	   kexecvp(args[0], args);
	   fprintf(stderr,"\nUnable to execute command:\n\n '%s'\n", command);
	   fprintf(stderr,"args[0] = '%s':\n", args[0]);
	   fprintf(stderr,"errno = %d:\n", errno);
           _exit(KEXIT_FAILURE);
	}

        /*
         *  Check to see if we failed to create the process.
         */
        if (pid == -1)
        {
           kfprintf(kstderr, "\nUnable to create a process to execute the "
			    "command (errno=%d):\n\n  %s", errno, tmp);
	   kfree(tmp);
           return(-1);
        }
	kfree(tmp);
	if (kfile_iswrite(file)) close(to_child[0]);
	if (kfile_isread(file))  close(from_child[1]);
	fid[0] = kfile_isread(file)  ? from_child[0] : -1;
	fid[1] = kfile_iswrite(file) ? to_child[1]   : -1;
	return(pid);
}

/*-------------------------------------------------------------------*
|
|                 Pipe Transport Routines
|
--------------------------------------------------------------------*/

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_tempnam
|  
|       Purpose: This function initializes a tempnam for a "pipe"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"pipe=[in,out]"
|		  	"pipe=-"
|		  	"pipe=#"
|  
|	         This routine creates a pipe and returns the template
|	         in the following syntax:
|  
|		  	"pipe=[in,out]"
|  
|         Input:  identifier - the transport identifier
|		  ktemplate  - the filename template to be modeled
|        Output:  result     - the newly created transport name
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int pipe_tempnam(
   const char *identifier,
   const char *ktemplate,
   char *result)
{
	int	fid[2];

	/*
	 *  Create a pipe for the transport
	 */
	if (pipe(fid) == -1)
	{
	   kinfo(KSYSLIB, "pipe_tempnam: Unable to create temporary pipe");
	   return(-1);
	}
	else
	{
	   if (identifier == NULL)
	      (void) sprintf(result,"pipe=[%d,%d]", fid[0], fid[1]);
	   else
	      (void) sprintf(result,"%s=[%d,%d]", identifier, fid[0], fid[1]);
	}
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_open
|  
|       Purpose: This function opens a "pipe" connection.  It is the
|	         internal driver to open a pipe, which is called by
|	         the user when they call kopen().  kopen() calls when
|	         it gets an open request with a filename of the following
|	         syntax:
|  
|		  	"pipe=[in,out]"
|		  	"pipe=-"   (stdin or stdout)
|		  	"pipe=#"   (stderr)
|  
|         Input: 
|        Output:  returns 0 or -1 if an error occurs
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int pipe_open(
   const char   *path,
   int    flags,
   mode_t mode,
   kfile  *file)
{
	pid_t pid = -1;
	int   fid[2];
	ResourceStruct *resources;

	/*
	 *  Assign the string to the appropriate pipe.
	 */
	fid[0] = fid[1] = -1;
	if (sscanf(path," [ %d , %d ]", &fid[0], &fid[1]) != 2)
	{
	   if (kstrcmp(path,"#") == 0)
	   {
	      fid[1] = STDERR_FILENO;
	      ksetvbuf(file, NULL, _IONBF, (size_t) 0);
	   }
	   else if (kstrcmp(path,"-") == 0)
	   {
	      if (kfile_isrdwr(file))
	      {
                 kinfo(KSYSLIB, "pipe_open: Invalid open access for file '-'. "
			"The '-' can only be used for exclusively reading or "
			"writing.");
	         return(-1);
	      }
	      else if (kfile_isread(file))
	      {
	         fid[0] = STDIN_FILENO;
	         if (isatty(STDIN_FILENO))
		    ksetvbuf(file, NULL, _IOLBF, (size_t) 0);
	      }
	      else
	      {
	         fid[1] = STDOUT_FILENO;
	         if (/*isatty(STDOUT_FILENO)*/1)
		    ksetvbuf(file, NULL, _IONBF, (size_t) 0);
	      }
	   }
	   else
	   {
	      if ((pid = pipe_child(file, path, fid)) == -1)
		 return(-1);
	   }
	}

	/*
	 *  See if the user wants the KOPEN_NONBLOCK set
	 */
	if (flags & KOPEN_NONBLOCK)
	{
	   if (fid[0] != -1) fcntl(fid[0], F_SETFL, KOPEN_NONBLOCK);
	   if (fid[1] != -1) fcntl(fid[1], F_SETFL, KOPEN_NONBLOCK);
	}

	if ((resources = (ResourceStruct *) kcalloc((size_t)1,
			sizeof(ResourceStruct))) == NULL)
	{
	   kinfo(KSYSLIB, "pipe_open: Not enough memory to kcalloc (%d) bytes"
	      " for the khoros pipe structure.\n", sizeof(ResourceStruct));
	   return(-1);
	}
	file->resources = (kaddr) resources;

	/*
	 *  Save the read and write transport for the pipe and pid (if
	 *  a piped command).
	 */
	resources->fid[0] = fid[0];
	resources->fid[1] = fid[1];
	resources->pid = pid;
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_close
|  
|       Purpose: This function closes a "pipe".  It is the internal driver
|	         to close a data pipe, which is called by the user
|	         when they call kclose().  kclose() calls the "pipe" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the initial kopen().
|  
|         Input:  file - the kfile structure.
|        Output:  returns whether we were able to close the pipe
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int pipe_close(
   kfile *file)
{
	int fid, status = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call close() since "kfile" is just a regular file
	 *  descriptor.
	 */
	fid = resources->fid[0];
	if (fid != -1 && fid != STDIN_FILENO)
	   status |= close(fid);

	fid = resources->fid[1];
	if (fid != -1 && fid != STDOUT_FILENO && fid != STDERR_FILENO)
	   status |= close(fid);

	if (resources->pid != -1)
	   waitpid(resources->pid, &status, 0);

	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_read
|  
|       Purpose: This function reads a "pipe".  It is the internal driver
|	         to read data from a pipe, which is called by the user
|	         when they call kread().  kread() calls the "pipe" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the kopen().
|  
|         Input:  file - the kfile structure.
|		  ptr  - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes read from the file
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t pipe_read(
   kfile *file,
   kaddr ptr,
   size_t nbytes)
{
	size_t size;
	char *buf = (char *)ptr;
	ssize_t count, numread = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call read() since "kfile" is just a regular file
	 *  descriptor.
	 */
	do
	{
	   size = nbytes - numread;
	   count = read(resources->fid[0], (kaddr) buf, size);
	   numread += count; buf += count;
	}
	while (numread < nbytes && count > 0 && kfile_isstream(file));
	kinfo(KDEBUG,"pipe_read: read %d out of %d bytes '%s'\n", numread,
		     nbytes, file->path);
	return(numread);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_write
|  
|       Purpose: This function writes to a "pipe".  It is the internal driver
|	         to write data from the supplied data array to a pipe, which
|	         is called by the user when they call kwrite().  kwrite() calls
|	         the "pipe" internal drivers by the nature of the transport,
|	         which is dictated by the kopen().
|  
|         Input:  id   - the kfile structure.
|		  ptr  - the pointer to write the data from.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t pipe_write(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	size_t size;
	char *buf  = (char *)ptr;
	ssize_t count, numwrite = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Safety check..
	 */
	if (nbytes == 0)
	   return(0);

	/*
	 *  Simply call write() since "kfile" is just a regular file
	 *  descriptor.
	 */
	do
	{
	   size = nbytes - numwrite;
	   count = write(resources->fid[1], (kaddr) buf, size);
	   numwrite += count; buf += count;
	}
	while (numwrite < nbytes && count > 0);
	kinfo(KDEBUG, "pipe_write: wrote %d out of %d bytes '%s'\n", numwrite,
		      nbytes, file->path);
	return(numwrite);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_access
|  
|       Purpose: This function is used to do see if a pipe is accessable.
|	         This is called when the user calls kaccess().  kaccess()
|	         calls the "file" internal drivers by the nature of the
|	         transport, which is dictated by the ktransport_identifier().
|  
|         Input:  name - the original path name passed to kaccess
|		  path - the filename that they wish to access
|		  mode - the access mode information
|        Output:  returns the result
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int pipe_access(
   const char *name,
   const char *path,
   int  mode)
{
	int status = 0;  /* just for now. (MY) 4/21/96 */

	/*
	 *  Simply call access() since "path" is just a regular filename
	 */
	if (kstrcmp(path, "-") == 0 || kstrcmp(path, "#") == 0)
	   return(0);

	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_unlink
|  
|       Purpose: This function initializes a tempnam for a "file"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"pipe=[in,out]"
|  
|         Input:  name - the original path name passed to kunlink
|		  path - the filename that they wish to unlink
|        Output:  returns -1 or 0 depending whether we
|		  sucessfully created the template
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int pipe_unlink(
   const char *name,
   const char *path)
{
	int	fid[2], status1 = -1, status2 = -1;

	fid[0] = fid[1] = -1;
	(void) sscanf(path," [ %d , %d ]", &fid[0], &fid[1]);

	if (fid[0] != -1)
	{
	   if ((status1 = close(fid[0])) == -1)
	      if (errno == EBADF) status1 = 0;
	}

	if (fid[1] != -1)
	{
	   if ((status2 = close(fid[1])) == -1)
	      if (errno == EBADF) status2 = 0;
	}
	return(status1 | status2);
}

/*-------------------------------------------------------------
|  
|  Routine Name: pipe_descriptors
|  
|       Purpose: This function shouldn't exist, but there is occasion
|	         when the user wants to know the actual file descriptor
|	         associated with the file descriptor.
|  
|         Input: file structure
|        Output:  inum - the input file descriptor
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int pipe_descriptors(
   kfile *file,
   int   *inum,
   int   *onum)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;


	if (inum != NULL) *inum = resources->fid[0];
	if (onum != NULL) *onum = resources->fid[1];
	return(0);
}

/**************************************************************
*  
*  Routine Name: kpinfo - gets the associated process id
*
*  Purpose: Gets the pipe process id 
*
*    Input: file - tranport pointer to close
*  Returns: process id on success, -1 on failure
*    Written By: Mark Young  
*          Date: Jan 20, 1995
**************************************************************/
/*ARGSUSED*/
pid_t kpinfo(
   kfile *file)
{
        ResourceStruct *resources;

	/*
	 *  Make sure the entry is within our table.  If not then it
	 *  has probably been closed already.
	 */
	if ((file = kfile_checkfile(file)) == NULL ||
	    file->routines != pipe_transport)
	{
	   return(kreminfo(file));
	}
	resources = (ResourceStruct *) file->resources;
	return(resources->pid);
}
