/*
 * Khoros: $Id: Stream.c,v 1.14 1997/01/29 01:35:29 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: Stream.c,v 1.14 1997/01/29 01:35:29 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Stream (Named Pipe) Transport Routines
   >>>>		
   >>>>   Static:
   >>>>                 stream_tempnam()
   >>>>                 stream_open()
   >>>>                 stream_close()
   >>>>                 stream_read()
   >>>>                 stream_write()
   >>>>                 stream_access()
   >>>>                 stream_unlink()
   >>>>                 stream_descriptors()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


#if !defined(KSOCKET_DEF)
TransportInformation stream_transport[] = { {NULL} };
#else

static int stream_tempnam     PROTO((const char *, const char *, char *));
static int stream_open	      PROTO((const char *, int, mode_t, kfile *));
static int stream_close	      PROTO((kfile *));
static ssize_t stream_read    PROTO((kfile *, kaddr, size_t));
static ssize_t stream_write   PROTO((kfile *, kaddr, size_t));
static int stream_access      PROTO((const char *, const char *, int));
static int stream_unlink      PROTO((const char *, const char *));
static int stream_descriptors PROTO((kfile *, int *, int *));

TransportInformation stream_transport[] =
{
    {
	"Standard Stream (FIFO)",
	"stream",
	FALSE,
	stream_tempnam,
	stream_open,
	stream_close,
	stream_read,
	stream_write,
	NULL,
	stream_access,
	stream_unlink,
	NULL,
	stream_descriptors,
    }
};

/*
 *  Internal resource structures for "Standard Stream (FIFO)" transport
 */
typedef struct
{
	int	fid;
} ResourceStruct;



/*-------------------------------------------------------------------*
|  
|		  Stream Routines
|  
--------------------------------------------------------------------*/


/*-------------------------------------------------------------
|  
|  Routine Name: stream_tempnam
|  
|       Purpose: This function initializes a tempnam for a "stream"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"stream=XXXXXXXXXX"
|  
|         Input:  identifier - the transport identifier
|		  ktemplate  - the filename template to be modeled
|        Output:  result     - the newly created transport name
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int stream_tempnam(
   const char *identifier,
   const char *ktemplate,
   char *result)
{
	char	*temp, filename[KLENGTH];

	(void) kstrcpy(filename, ktemplate);
	if ((temp = ktmpnam(filename)) == NULL)
	   return(-1);

	/*
	 *  Create the FIFO stream if it doesn't already exist.
	 */
	if (mkfifo(temp, (mode_t) 0666) < 0)
	{
	   if (errno != EEXIST)
	   {
	      kinfo(KSYSLIB, "stream_tempnam:  Unable to create temporary "
			     "stream file '%s'", temp);
	      return(-1);
	   }
	}

	if (identifier == NULL)
	   (void) sprintf(result, "stream=%s", temp);
	else
	   (void) sprintf(result,"%s=%s", identifier, temp);

	kfree_and_NULL(temp);
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_open
|  
|       Purpose: This function opens a "stream" connection.  It is the
|	         internal driver to open a stream, which is called by
|	         the user when they call kopen().  kopen() calls when
|	         it gets an open request with a filename of the following
|	         syntax:
|  
|		  	"stream=filename"
|  
|         Input: 
|        Output:  returns 0 or -1 if an error occurs
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int stream_open(
   const char *path,
   int    flags,
   mode_t mode,
   kfile  *file)
{
	int    fid;
	mode_t stream_mode;
	ResourceStruct *resources;

	/*
	 *  Create the FIFO stream if it doesn't already exist.
	 */
	if (!(flags & KOPEN_CREAT))
	   stream_mode = 0666;
	else
	   stream_mode = mode & 0777;

	if (mkfifo(path, (mode_t) stream_mode) < 0)
	{
	   if (errno != EEXIST)
	   {
	      kinfo(KSYSLIB, "stream_open: Create Failure for stream file "
			     "'%s'", path);
	      return(-1);
	   }
	}

	if ((fid = open(path, flags, mode)) < 0)
	{
	   kinfo(KSYSLIB, "stream_open: Open failure for stream file '%s'",
			path);
	   return(-1);
	}

	if ((resources = (ResourceStruct *) kcalloc((size_t)1,
			sizeof(ResourceStruct))) == NULL)
	{
	   kinfo(KSYSLIB, "stream_open: Not enough memory to kcalloc (%d) "
			  "bytes for the khoros stream structure.\n",
			  sizeof(ResourceStruct));
	   return(-1);
	}

	resources->fid = fid;
	file->resources = (kaddr) resources;
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_close
|  
|       Purpose: This function closes a "stream".  It is the internal driver
|	         to close a data stream, which is called by the user
|	         when they call kclose().  kclose() calls the "stream" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the initial kopen().
|  
|         Input:  file - the kfile structure.
|        Output:  returns whether we were able to close the stream
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int stream_close(
   kfile *file)
{
	int	     fid, status;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call close() since "kfile" is just a regular file
	 *  descriptor.
	 */
	fid = resources->fid;
	status = close(fid);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_read
|  
|       Purpose: This function reads a "stream".  It is the internal driver
|	         to read data from a stream, which is called by the user
|	         when they call kread().  kread() calls the "stream" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes read from the file
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t stream_read(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	size_t size;
	char *buf = (char *)ptr;
	ssize_t count, numread = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call read() since "kfile" is just a regular file
	 *  descriptor.
	 */
	do
	{
	   size = nbytes - numread;
	   count = read(resources->fid, (kaddr) buf, size);
	   numread += count; buf += count;
	}
	while (numread < nbytes && count > 0 && kfile_isstream(file));
	kinfo(KDEBUG,"pipe_read: read %d out of %d bytes '%s'\n", numread,
		     nbytes, file->path);
	return(numread);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_write
|  
|       Purpose: This function writes to a "stream".  It is the internal driver
|	         to write data from the supplied data array to a stream, which
|	         is called by the user when they call kwrite().  kwrite() calls
|	         the "stream" internal drivers by the nature of the transport,
|	         which is dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to write the data from.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t stream_write(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	size_t size;
	char *buf  = (char *)ptr;
	ssize_t count, numwrite = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Safety check..
	 */
	if (nbytes == 0)
	   return(0);

	/*
	 *  Simply call write() since "kfile" is just a regular file
	 *  descriptor.
	 */
	do
	{
	   size = nbytes - numwrite;
	   count = write(resources->fid, (kaddr) buf, size);
	   numwrite += count; buf += count;
	}
	while (numwrite < nbytes && count > 0);
	kinfo(KDEBUG, "pipe_write: wrote %d out of %d bytes '%s'\n", numwrite,
		      nbytes, file->path);
	return(numwrite);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_access
|  
|       Purpose: This function is used to do see if a file is accessable.
|	         This is called when the user calls kaccess().  kaccess()
|	         calls the "stream" internal drivers by the nature of the
|	         transport, which is dictated by the ktransport_identifier().
|  
|         Input:  name - the original path name passed to kaccess
|		  path - the filename that they wish to access
|		  mode - the access mode information
|        Output:  returns the result
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int stream_access(
   const char *name,
   const char *path,
   int  mode)
{
	int	   status;

	/*
	 *  Simply call access() since "path" is just a stream file
	 */
	status = access(path, mode);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_unlink
|  
|       Purpose: This function initializes a tempnam for a "stream"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"stream=XXXXXXXXXX"
|  
|         Input:  name - the original path name passed to kunlink
|		  path - the filename that they wish to unlink
|        Output:  returns 0 or -1 depending whether we
|		  sucessfully created the template
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int stream_unlink(
   const char *name,
   const char *path)
{
	int	status;

	status = unlink(path);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: stream_descriptors
|  
|       Purpose: This function shouldn't exist, but there is occasion
|	         when the user wants to know the actual file descriptor
|	         associated with the file descriptor.
|  
|         Input: file structure
|        Output:  inum - the input file descriptor
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int stream_descriptors(
   kfile *file,
   int   *inum,
   int   *onum)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	if (inum != NULL) *inum = resources->fid;
	if (onum != NULL) *onum = resources->fid;
	return(0);
}

#endif  /* KSTREAM_DEF */
/* don't add after the endif */
