/*
 * Khoros: $Id: Tcpip.c,v 1.1.2.1 1997/05/24 17:50:07 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: Tcpip.c,v 1.1.2.1 1997/05/24 17:50:07 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Socket Transport Drivers
   >>>>		
   >>>>   Static:
   >>>>                 tcpip_connect()
   >>>>                 tcpip_tempnam()
   >>>>                 tcpip_open()
   >>>>                 tcpip_close()
   >>>>                 tcpip_read()
   >>>>                 tcpip_write()
   >>>>                 tcpip_access()
   >>>>                 tcpip_descriptors()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#define KPOSIX_OVERRIDE
#include "internals.h"


#if !defined(KSOCKET_DEF)
TransportInformation tcpip_transport[] = { {NULL} };
#else

/* Berkeley socket (inet) transport */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

static int tcpip_tempnam     PROTO((const char *, const char *, char *));
static int tcpip_open	     PROTO((const char *, int, mode_t, kfile *));
static int tcpip_close	     PROTO((kfile *));
static ssize_t tcpip_read    PROTO((kfile *, kaddr, size_t));
static ssize_t tcpip_write   PROTO((kfile *, kaddr, size_t));
static int tcpip_access      PROTO((const char *, const char *, int));
static int tcpip_descriptors PROTO((kfile *, int *, int *));

TransportInformation tcpip_transport[] =
{
    {
	"TCP Socket (AF_INET)",
	"tcpip",
	FALSE,
	tcpip_tempnam,
	tcpip_open,
	tcpip_close,
	tcpip_read,
	tcpip_write,
	NULL,
	tcpip_access,
	NULL,
	NULL,
	tcpip_descriptors,
    }
};

/*
 *  Internal resource structures for "Standard Socket (AF_UNIX)" transport
 */
typedef struct
{
	int	 socket;
	int      write_flag;
	unsigned long readnum;
} ResourceStruct;


/*-------------------------------------------------------------------*
|  
|		  Private Socket Routines
|  
--------------------------------------------------------------------*/

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_connect
|  
|       Purpose: This function opens a "socket" connection.  It is an
|	         internal driver to open a socket, which is called by
|	         the the internal socket routines such as tcpip_open().
|  
|		  	"tcpip=XXXXXXXXXX"
|  
|         Input:  file - the kfile structure describing the transport
|		  flags - the flags to open the connection
|		  path - the socket path
|        Output:  returns -1 or the socket id
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int tcpip_connect(
   kfile *file,
   int   flags,
   const char *path)
{
	char	 machine[KLENGTH];
	unsigned long inaddr;
	struct	 sockaddr_in server;
	struct   hostent *host_machine;
	int      port, status, sock, nsock, len, reuse;

	/*
	 *  Get the path machine to connect to, the socket name component of
	 *  the path.
	 */
	if (ksscanf(path,"tcpip=%[^:]:%d", machine, &port) != 2 &&
	    ksscanf(path,"%[^:]:%d", machine, &port) != 2)
	{
           kinfo(KSYSLIB,"tcpip_connect: Unknown machine or network address "
			 "'%s'\n", path);
	   return(-1);
	}
	kmemset((char *) &server, 0, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port   = htons(port);

	/*
	 *  Assign our server's address and then connect to the returned
	 *  socket "server_sock".
	 */
        if ((host_machine = gethostbyname(machine)) != NULL)
        {
           kmemcpy((kaddr) &server.sin_addr, (kaddr) host_machine->h_addr,
                        (size_t) host_machine->h_length);
        }
        else if ((inaddr=inet_network(machine)) != (unsigned long) -1)
        {
           inaddr = htonl(inaddr);
           kmemcpy((char *) &server.sin_addr, &inaddr, sizeof(inaddr));
        }
        else
        {
           /*
            *  failed to find the name..
            */
           errno = ENETUNREACH;
           kinfo(KSYSLIB,"tcpip_connect: Unknown machine or network address "
			 "'%s'\n", machine);
           return(-1);
        }

	/*
	 *  Create either a local UNIX connection based socket
	 */
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	   kinfo(KSYSLIB,"tcpip_connect:  Creation failure for socket '%s'",
			path);
	   return(-1);
	}

	reuse = TRUE;
	len = sizeof(reuse);
	status = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,(char *) &reuse,len);
	if (status < 0)
	{
	   kinfo(KSYSLIB,"tcpip_connect:  Cannot set option (REUSEADDR) for "
			 "socket '%s'", path);
	}

	if (flags & KOPEN_CREAT)
	{
	   len = sizeof(server);
	   server.sin_addr.s_addr = htonl(INADDR_ANY);
	   if (bind(sock, (struct sockaddr *) &server, len) == -1)
	   {
	      close(sock);
	      kinfo(KSYSLIB,"tcpip_connect:  Bind failure for socket '%s'",
			path);
	      return(-1);
	   }
	   listen(sock, 5);

	   if ((nsock = accept(sock, (struct sockaddr *) &server, &len)) == -1)
	   {
	      close(sock);
	      kinfo(KSYSLIB,"tcpip_connect:  Accept failure for socket '%s'",
			path);
	      return(-1);
	   }
	   close(sock);
	   sock = nsock;
	}
	else
	{
	   int i;

	   len = sizeof(server);
	   for (i = 0, status = -1; i < 5; i++)
	   {
	      status = connect(sock, (struct sockaddr *) &server, len);
	      if (status == 0)
		 break;
	   }

	   if (status == -1)
	   {
	      close(sock);
	      kinfo(KSYSLIB,"tcpip_connect:  Connect failure for socket '%s'",
			path);
	      return(-1);
	   }
	}
	return(sock);
}

/*-------------------------------------------------------------------*
|  
|		  Socket Routines
|  
--------------------------------------------------------------------*/

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_tempnam
|  
|       Purpose: This function initializes a tempnam for a "socket"
|	         type transport.  Given a request with a filename of
|	         the following syntax:
|  
|		  	"tcpip=XXXXXXXXXX"
|  
|         Input:  identifier - the transport identifier
|		  ktemplate  - the filename template to be modeled
|        Output:  result     - the newly created transport name
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int tcpip_tempnam(
   const char *identifier,
   const char *ktemplate,
   char *result)
{
	int	i, port ,sock;
	char	machine[KLENGTH];
	struct	sockaddr_in server;

	/*
	 *  Get local hostname
	 */
	if (kgethostname(machine, KLENGTH) != 0)
	   return(-1);

	/*
	 *  Create a local TCP/IP connection based socket
	 */
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	   kinfo(KSYSLIB, "tcpip_tempnam: Unable to create temporary socket "
			  "'%s'", machine);
	   return(-1);
	}

	/*
	 *  Assign our server's address and then connect to the returned
	 *  socket "server_sock".
	 */
	for (port = -1, i = 1024; i < 5000; i++)
	{
	   kmemset((char *) &server, 0, sizeof(server));
	   server.sin_port = htons(i);
	   server.sin_family = AF_INET;
	   server.sin_addr.s_addr = htonl(INADDR_ANY);

	   if (bind(sock, (struct sockaddr *) &server, sizeof(server)) >= 0)
	   {
	      port = i;
	      break;
	   }
	}
	close(sock);

	if (port == -1)
	{
	   kinfo(KSYSLIB, "tcpip_tempnam:  Bind failure for socket '%s'",
			   machine);
	   return(-1);
	}

	/*
	 *  Copy the result into the result array
	 */
	if (identifier == NULL)
	   (void) ksprintf(result,"%s=%s:%d", machine, port);
	else
	   (void) ksprintf(result,"%s=%s:%d", identifier, machine, port);

	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_open
|  
|       Purpose: This function opens a "socket" connection.  It is the
|	         internal driver to open a file, which is called by
|	         the user when they call kopen().  kopen() calls when
|	         it gets an open request with a filename of the following
|	         syntax:
|  
|		  	"filename"
|		  	"/path/filename"
|		  	"./path/filename"
|		  	"file=XXXXXXXXXX"
|		  	     ...etc....
|  
|         Input: 
|        Output:  returns 0 or -1 if an error occurs
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int tcpip_open(
   const char *path,
   int    flags,
   mode_t mode,
   kfile  *file)
{
	ResourceStruct *resources;
	int   sock;


	/*
	 *  Connect to the socket
	 */
	if ((sock = tcpip_connect(file, flags, path)) < 0)
	   return(-1);

	/*
	 *  Create the socket structure.
	 */
	if ((resources = (ResourceStruct *) kcalloc((size_t)1,
			sizeof(ResourceStruct))) == NULL)
	{
           kinfo(KSYSLIB, "tcpip_open: Not enough memory to kcalloc (%d) "
			  "bytes for the khoros socket structure.\n",
			  sizeof(ResourceStruct));
	   return(-1);
	}
	resources->socket = sock;
	file->resources = (kaddr) resources;
	kfile_setstate(file, KFILE_PERM);

	/* MY & SJ: Aren't able to support buffered read/write sockets */
	if (kfile_isrdwr(file))
	{
	   kfile_clrstate(file, KFILE_STREAM);
	   ksetvbuf(file, NULL, _IONBF, (size_t) 0);
	}
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_close
|  
|       Purpose: This function closes a "socket".  It is the internal driver
|	         to close a data socket, which is called by the user
|	         when they call kclose().  kclose() calls the "socket" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the initial kopen().
|  
|         Input:  file - the kfile structure.
|        Output:  returns whether we were able to close the socket
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int tcpip_close(
   kfile *file)
{
	int	     sock, status;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/*
	 *  Simply call close() since "kfile" is just a regular file
	 *  descriptor.
	 */
	sock = resources->socket;
	status = close(sock);
	return(status);
}

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_read
|  
|       Purpose: This function reads a "socket".  It is the internal driver
|	         to read data from a socket, which is called by the user
|	         when they call kread().  kread() calls the "socket" internal
|	         drivers by the nature of the transport, which is dictated
|	         by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to store the data into.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes read from the file
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t tcpip_read(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	long len;
        size_t size;
        char *buf = (char *) ptr;
        ssize_t count, numread = 0; 
	ResourceStruct *resources = (ResourceStruct *) file->resources;

        /*
         *  If write flag is TRUE, then write a -1 in order to break
         *  client read-deadlock.
         */
        if (resources->write_flag)
        {
           len = -1; len = htonl((unsigned long) len);
           if (write(resources->socket, (kaddr) &len, 4) < 0)
              return(-1);
           else resources->write_flag = FALSE;
        }
 
        if (resources->readnum == 0)
        {
           /* read in the number of bytes that are to be written */
           if (read(resources->socket, (kaddr) &len, 4) < 0)
               return(-1);

	   len = ntohl((unsigned long) len);
           if (len != -1)
               resources->readnum = len;
        }

	/*
	 *  Simply call read() since "kfile" is just a regular file
	 *  descriptor.
	 */
	do
	{
	   if (resources->readnum == 0)
	   {
	      /* read in the number of bytes that are to be written */
	      if (read(resources->socket, (kaddr) &len, 4) < 0)
	         return(numread);
	      else
	         resources->readnum = ntohl((unsigned long) len);
	   }

	   size = kmin(nbytes - numread, resources->readnum);
	   if ((count = read(resources->socket, (kaddr) buf, size)) > 0)
	   {
	      numread += count; buf += count; resources->readnum -= count;
	   }
	} while (numread < nbytes && count > 0);

	kinfo(KDEBUG,"tcpip_read: read %d out of %d bytes '%s'\n", numread,
		     nbytes, file->path);
	return(numread);
}

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_write
|  
|       Purpose: This function writes to a "socket".  It is the internal driver
|	         to write data from the supplied data array to a socket, which
|	         is called by the user when they call kwrite().  kwrite() calls
|	         the "socket" internal drivers by the nature of the transport,
|	         which is dictated by the kopen().
|  
|         Input:  file   - the kfile structure.
|		  ptr    - the pointer to write the data from.
|		  nbytes - the number of bytes to read.
|        Output:  returns the number of bytes written to the kfile
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static ssize_t tcpip_write(
   kfile  *file,
   kaddr  ptr,
   size_t nbytes)
{
	long len;
	size_t size;
	char *buf  = (char *)ptr;
	ssize_t count, numwrite = 0;
	ResourceStruct *resources = (ResourceStruct *) file->resources;

	/* write the number of bytes that are to be written */
	len = htonl((unsigned long) nbytes);
	if (write(resources->socket, (kaddr) &len, 4) < 0)
	    return(0);

	/*
	 *  Simply call write() since "socket" is just a regular socket
	 *  descriptor.
	 */
	do
	{
	   size = nbytes - numwrite;
	   count = write(resources->socket, (kaddr) buf, size);
	   numwrite += count; buf += count;
	} while (numwrite < nbytes && count > 0);

        /* set the write flag to TRUE in order to break client read-deadlock */
        resources->write_flag = TRUE;

	kinfo(KDEBUG, "pipe_write: wrote %d out of %d bytes '%s'\n", numwrite,
		      nbytes, file->path);
	return(numwrite);
}

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_access
|  
|       Purpose: This function is used to do see if a file is accessable.
|	         This is called when the user calls kaccess().  kaccess()
|	         calls the "socket" internal drivers by the nature of the
|	         transport, which is dictated by the ktransport_identifier().
|  
|         Input:  name - the original path name passed to kaccess
|		  path - the filename that they wish to access
|		  mode - the access mode information
|        Output:  returns the result
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int tcpip_access(
   const char *name,
   const char *path,
   int  mode)
{
	return(0);
}

/*-------------------------------------------------------------
|  
|  Routine Name: tcpip_descriptors
|  
|       Purpose: This function shouldn't exist, but there is occasion
|	         when the user wants to know the actual file descriptor
|	         associated with the file descriptor.
|  
|         Input: file structure
|        Output:  inum - the input file descriptor
|    Written By:  Mark Young
|  
-------------------------------------------------------------*/
/*ARGSUSED*/
static int tcpip_descriptors(
   kfile *file,
   int   *inum,
   int   *onum)
{
	ResourceStruct *resources = (ResourceStruct *) file->resources;


	if (inum != NULL) *inum = resources->socket;
	if (onum != NULL) *onum = resources->socket;
	return(0);
}

#endif  /* KSOCKET_DEF */
/* don't add after the endif */
