/*
 * Khoros: $Id: knegotiator.c,v 1.27.2.2 1997/03/10 17:22:58 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: knegotiator.c,v 1.27.2.2 1997/03/10 17:22:58 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Negotiator Support Utility
   >>>>		
   >>>>  Private:
   >>>>                 knegotiate_transport()
   >>>>                 knegotiate_initialize()
   >>>>
   >>>>                 knegotiate_continuous_run()
   >>>>                 knegotiate_check()
   >>>>                 knegotiate_open()
   >>>>                 knegotiate_close()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


/*
 *  The following are private routines that the user should not call but
 *  should be private to the kutils library that enables a negotiation
 *  method between processes thru some user defined negotiator.
 *
 *	routines:	knegotiate_transport
 *			knegotiate_open
 *			knegotiate_close
 */
static int enable_negotiation = TRUE;

static kaddr clui_form = NULL;
static void (*clui_get_args)(kaddr) = NULL;
static void (*clui_free_args)(kexit_status, kaddr) = NULL;
static int  (*clui_update_args)(kaddr, char *, int) = NULL;

/*-----------------------------------------------------------
|
|  Routine Name: knegotiate_transport - enable transport negotation for the
|				   local process.
|
|       Purpose: This function is used to enable the transport negotations
|		 for the default transport
|
|         Input: enable - whether local negotiations should occur or not 
|    Written By: Mark Young
|          Date: Apr 09, 1995
|
------------------------------------------------------------*/
int knegotiate_transport(
   int enable)
{
	int old_enable = enable_negotiation;

	kfile_initialize();
	enable_negotiation = enable;
	return(old_enable);
}

/*-----------------------------------------------------------
|
|  Routine Name: knegotiate_initialize - 
|
|       Purpose: This function is used to initialize callback routines
|		 so that the negotiator,
|		 if one exists.  The negotiator is informed
|		 that this process has closed this transport path.
|
|    Written By: Mark Young
|          Date: June 8, 1996
|
------------------------------------------------------------*/
void knegotiate_initialize(
    kaddr form,
    void (*get_args)    PROTO((kaddr)),
    void (*free_args)   PROTO((kexit_status, kaddr)),
    int  (*update_args) PROTO((kaddr, char *, int)))
{
	clui_form        = form;
	clui_get_args    = get_args;
	clui_free_args   = free_args;
	clui_update_args = update_args;
}

/*-----------------------------------------------------------
|
|  Routine Name: knegotiate_continuous_run - 
|
|       Purpose: This function is used to call the negotiator,
|		 if one exists.  The negotiator is informed
|		 that this process has closed this transport path.
|
|    Written By: Mark Young & Jeremy Worley
|          Date: Apr 8, 1996
|
------------------------------------------------------------*/
int knegotiate_continuous_run(void)
{
	kfile *ipc;
	int   result = -1;
	static char last_command[KLENGTH];
	char  buffer[KLENGTH], command[KLENGTH];

	static int initialized = FALSE;


	kinfo(KDEBUG, "Entering knegotiate_continuous_run");

	if ((ipc = kipc_getipc()) == NULL)
	{
	   static int tryonce = FALSE;
	   kstrcpy(last_command, "");
	   if (tryonce == FALSE)
	   {
	      tryonce = TRUE;
	      return(TRUE);
	   }
	   return(FALSE);
	}

	if (!initialized)
	{
	   initialized = TRUE;
	   ksprintf(buffer,"%s:%d: initialize contrun\n",KIPC_PROCESS,getpid());
	   kfputs(buffer, ipc);
	   return(TRUE);
	}

	ksprintf(buffer, "%s:%d: next contrun\n", KIPC_PROCESS, getpid());
	if (kfputs(buffer, ipc) != EOF)
	{
	   kinfo(KDEBUG, "knegotiate_continuous_run reading input");
	   if (kfgets(buffer, KLENGTH, kstdin) != NULL &&
	       ksscanf(buffer, "%[^\n]", command) == 1 &&
	       kstrlen(command) > 0)
	   {
	      kinfo(KDEBUG, "knegotiate_continuous_run '%s'", command);
	      if (kstrcasecmp(command, "exit") == 0)
	      {
	         kprintf("%d: got an exit request\n", getpid());
		 result = FALSE;
	      }
	      else if (kstrcmp(command, last_command) != 0)
	      {
 	         clui_update_args(clui_form, command, TRUE);
	         clui_free_args(KEXIT_SUCCESS, NULL);
	         clui_get_args(clui_form);
		 kstrcpy(last_command, command);
	         result = TRUE;
	      }
	      else result = TRUE;
	   }
	}
	kinfo(KDEBUG, "Leaving knegotiate_continuous_run '%d'", result);
	return(result);
}

/*-----------------------------------------------------------
|
|  Routine Name: knegotiate_check - check if a transport has been modified
|
|       Purpose: This function is used to call the negotiator,
|		 if one exists.  The negotiator is asked if
|		 the a transport has been modified.
|
|         Input: path  - the path that negotiator should be informed of
|       Returns: TRUE if allowed to modify, FALSE if not, and -1 unknown file 
|    Written By: Mark Young
|          Date: Oct 14, 1992
|
------------------------------------------------------------*/
int knegotiate_check(
   const char *path)
{
	kfile *ipc;
	int   result = -1;
	char  buffer[KLENGTH], type[KLENGTH];

	if (!path || !enable_negotiation || (ipc = kipc_getipc()) == NULL)
	   return(TRUE);

	kinfo(KDEBUG, "Entering knegotiate_check");

	kflags_to_type(KOPEN_RDONLY, type);
	ksprintf(buffer, "%s:%d: check '%s' '%s'\n", KIPC_TRANSPORT, getpid(),
			path, type);
	if (kfputs(buffer, ipc) != EOF)
	{
	   if (kfgets(buffer, KLENGTH, kstdin) != NULL)
	   {
	      kstring_lower(buffer, buffer);
	      if (kstrcmp(buffer, "yes\n") == 0)
	         result = TRUE;
	      else if (kstrcmp(buffer, "no\n") == 0)
	         result = FALSE;
	      else
	         result = FALSE;
	   }
	}
	return(result);
}

/*-----------------------------------------------------------
|
|  Routine Name: knegotiate_open - inform negotiator that we opening
|				   a file.
|
|       Purpose: This function is used to call the negotiator,
|		 if one exists.  The negotiator is informed
|		 that the this process will be opening a file
|		 by the name of "path".
|
|         Input: path  - the path that negotiator should be informed of
|		 flags - the flags in which we will be opening the file
|       Returns: TRUE if allowed to open, FALSE if not, and -1 on unknown file
|    Written By: Mark Young
|          Date: Oct 14, 1992
|
------------------------------------------------------------*/
int knegotiate_open(
   const char *path,
   int  flags)
{
	kfile *ipc;
	int   result = 0;
	char  buffer[KLENGTH], type[KLENGTH];

	if (!path || !enable_negotiation || (ipc = kipc_getipc()) == NULL)
	   return(FALSE);

	kinfo(KDEBUG, "Entering knegotiate_open");

        /*
         *  Currently cantata is the only applications monitoring open/close
         *  of transports.  And since cantata is only interested in output
         *  files if the file that we are about to open is KOPEN_RDONLY then
         *  don't bother.
         */
        if ((flags & KOPEN_ACCMODE) == KOPEN_RDONLY)
           return(TRUE);

	kflags_to_type(flags, type);
	ksprintf(buffer, "%s:%d: open '%s' '%s'\n", KIPC_TRANSPORT, getpid(),
			path, type);
	if (kfputs(buffer, ipc) != EOF)
	{
	   if (kfgets(buffer, KLENGTH, kstdin) != NULL)
	   {
	      kstring_lower(buffer, buffer);
	      if (kstrcmp(buffer, "yes\n") == 0)
	         result = TRUE;
	      else if (kstrcmp(buffer, "no\n") == 0)
	         result = FALSE;
	      else
	         result = -1;
	   }
	}
	return(result);
}

/*-----------------------------------------------------------
|
|  Routine Name: knegotiate_close - inform negotiator that we closing
|				    a file.
|
|       Purpose: This function is used to call the negotiator,
|		 if one exists.  The negotiator is informed
|		 that this process has closed this transport path.
|
|         Input: path - the path that the negotiator should be informed of
|		 flags - the flags in which we will have the file opened
|       Returns: TRUE if allowed to close, FALSE if not, and -1 on unknown file 
|    Written By: Mark Young
|          Date: Oct 14, 1992
|
------------------------------------------------------------*/
int knegotiate_close(
   const char *path,
   int  flags)
{
	kfile *ipc;
	int   result = 0;
	char  buffer[KLENGTH], type[KLENGTH];

	if (!path || !enable_negotiation || (ipc = kipc_getipc()) == NULL)
	   return(FALSE);

	kinfo(KDEBUG, "Entering knegotiate_close");

        /*
         *  Currently cantata is the only applications monitoring open/close
         *  of transports.  And since cantata is only interested in output
         *  files if the file that we are about to open is KOPEN_RDONLY then
         *  don't bother.
         */
        if ((flags & KOPEN_ACCMODE) == KOPEN_RDONLY)
           return(TRUE);

	kflags_to_type(flags, type);
	ksprintf(buffer, "%s:%d: close '%s' '%s'\n", KIPC_TRANSPORT, getpid(),
			path, type);

	if (kfputs(buffer, ipc) != EOF)
	{
	   if (kfgets(buffer, KLENGTH, kstdin) != NULL)
	   {
	      kstring_lower(buffer, buffer);
	      if (kstrcmp(buffer, "yes\n") == 0)
	         result = TRUE;
	      else if (kstrcmp(buffer, "no\n") == 0)
	         result = FALSE;
	      else
	         result = -1;
	   }
	}
	return(result);
}
