/*
 * Khoros: $Id: ktransport.c,v 1.29.2.4 1997/06/07 19:06:59 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: ktransport.c,v 1.29.2.4 1997/06/07 19:06:59 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		General Khoros Transport Routines
   >>>>		
   >>>>   Public:
   >>>>                 ktempnam()
   >>>>                 krename()
   >>>>                 kopen()
   >>>>                 kclose()
   >>>>                 kread()
   >>>>                 kwrite()
   >>>>                 kaccess()
   >>>>                 kunlink()
   >>>>                 kcreat()
   >>>>                 ktouch()
   >>>>                 kfcntl()
   >>>>                 kflock()
   >>>>                 klseek()
   >>>>                 ktell()
   >>>>                 kdup()
   >>>>                 kdup2()
   >>>>                 kpopen()
   >>>>                 kpclose()
   >>>>                 ksystem()
   >>>>                 kspawn()
   >>>>                 kexecvp()
   >>>>                 kfileno()
   >>>>                 kfidfile()
   >>>>                 ktmpfid()
   >>>>                 kexit_handler()
   >>>>                 kexit()
   >>>>                 kgetdescriptors()
   >>>>                 kgethostname()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


/**************************************************************
*  
*  Routine Name: ktempnam - create a name for a temporary khoros transport
*       Purpose: This module is used to create temporary files.
*		 The application program calls "ktempnam"
*	         with a template filename which is used to help
*	         identify the application routine.  The type of
*	         temporary created depends on the identifier.  If
*	         a unique temporary shared memory key is desired
*	         then the template would look like:
*  
*  		 !       "shm=XXXX"
*  
*	         If a file was desired then either
*  
*  		 !       "XXXXXXXXX"
*  		 !       "file=XXXX"
*  
*	         would work.  The convention should be as follows,
*	         an identifier followed by an "=" equals indicator,
*	         and finally followed by an optional template.
*         Input: identifier  - The transport identifier
*  		 ktemplate   - template of the file to be created.
*       Returns: a unique temporary filename on success, NULL otherwise
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*   Declaration: char *ktempnam(
*		 !   const char *identifier,
*		 !   const char *ktemplate)
**************************************************************/
/*ARGSUSED*/
char *ktempnam(
   const char *identifier,
   const char *ktemplate)
{
	TransportInformation *routines;
	char	*filename, itemp[5*KLENGTH],
                ftemp[5*KLENGTH], result[5*KLENGTH];

	/*
	 *  According to the transport routines
	 */
	if (identifier == NULL)
	   identifier = ktransport_identifier(ktemplate, itemp);
	if ((routines = ktransport_routines(identifier)) == NULL)
	   return(NULL);

	/*
	 *  Get the filename component
	 */
	if ((filename = ktransport_path(ktemplate, ftemp)) == NULL &&
	    ktemplate != NULL)
	   return(NULL);

	if (routines->tempnam == NULL)
	   return(NULL);

	if (routines->tempnam(identifier, filename, result) == -1)
	   return(NULL);
	else
	   return(kstrdup(result));
}

/**************************************************************
*  
*  Routine Name: krename - rename a khoros transport from path1 to path2
*       Purpose: krename() is used to move the contents from one
*		 file to another.  If the two files are standard
*		 UNIX files then krename() simply use the system
*		 rename(), but if this fails then the khoros transport
*		 mechanisms are used to copy the data from the
*		 first file to the second, and then the first
*		 is unlinked using kunlink().
*         Input: oldname - the old khoros transport name
*  		 newname - the new khoros transport name
*	Returns: 0 on success, -1 on failure and sets errno to
*		 indicate the error
*    Written By: Mark Young  
*          Date: Jan 20, 1993
**************************************************************/

int krename(
   const char *oldname,
   const char *newname)
{
	int   status;
	kfile *file1, *file2;
	char  name1[5*KLENGTH], name2[5*KLENGTH];

	/*
	 *  kfullpath() is used to expand the path before opening it...
	 */
	if (!kfullpath(oldname,NULL,name1) || !kfullpath(newname,NULL,name2))
	{
	   errno = ENOTDIR;
	   return(-1);
	}

	/*
	 *  First see if we can't rename() oldname to newname using
	 *  the system rename function.
	 */
	if (rename(name1, name2) == 0)
	   return(0);
	else
	   errno = 0;

	/*
	 *  Since that didn't work then try and copy the data from
	 *  the first file to the second.
	 */
	if ((file1 = kfile_open(name1, KOPEN_RDONLY, (mode_t) 0666)) == NULL)
	   return(-1);

	if ((file2 = kfile_open(name2, KOPEN_WRONLY|KOPEN_CREAT|KOPEN_TRUNC,
				(mode_t) 0666)) == NULL)
	{
	   kfile_close(file1);
	   return(-1);
	}
	status = kfile_copydata(kfileno(file1), kfileno(file2), (size_t) 0);
	(void) kfile_close(file1);
	(void) kfile_close(file2);

	/*
	 *  If we have successfully transfered the contents from file1
	 *  to file2, then go ahead and unlink file1.
	 */
	if (status != -1)
	{
	   kunlink(name1);
	   return(0);
	}
	else
	{
	   kunlink(name2);
	   return(-1);
	}
}


/**************************************************************
*  
*  Routine Name: kopen - open or create a file for reading and/or writing
*       Purpose: This function is a replacement for the system "open"
*	         call.  The only difference is that kopen() supports
*	         more than just files, it supports other data transports
*	         as well.  The path should be in the form of an identifier,
*	         followed by an "=" equal indicator, and then the transport
*	         token. ie) a Shared memory path would look like:
*  
*  		 !       "shm=1032"
*  
*	         If a file was desired then either
*  
*  		 !      "/usr/tmp/vadd1235"
*  		 !      "file=/usr/tmp/vadd1235"
*
*		 will work.
*         Input: path - is the string containing the path name to the
*		        desired file to be open.  The path name identifies
*		        the file to be opened, just like the regular UNIX
*		        open() command.
*	        flags - how the file is to be opened.
*  
*  		!	KOPEN_RDONLY - Open file for only reading
*  		!	KOPEN_WRONLY - Open file for only writing
*  		!	KOPEN_RDWR   - Open file for both reading & writing
*               !
*  		!       KOPEN_NONBLOCK - Whether to block when reading and
*  		!                        writing to the file.
*  		!       KOPEN_APPEND   - Append to the file (writing).
*  		!       KOPEN_TRUNC    - Truncate the file for writing.
*  		!       KOPEN_EXCL     - Error if file exists
*  		!       KOPEN_CREAT    - the file exists, this flag has no
*		!                         effect.
*		!       Otherwise, the file is created, and the
*		!       owner ID of the file is set to the effective
*		!       user ID of the process.
*  
*	         mode - the permissions to be used when creating a new
*		        file.  Note that this parameter is only needed
*			when KOPEN_CREAT is specified in the flags.
*       Returns: the kfile id on success, -1 otherwise
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*   Declaration: int kopen(
*		 !  const char *path,
*  		 !  int         flags,
*		 !  mode_t      mode)
**************************************************************/

int kopen(
   const char *path,
   int  flags,
   kvalist)
{
	kfile *file;
	kva_list args;
        mode_t mode = 0666;
        
        kva_start(args, flags);
        if ((flags & KOPEN_CREAT) == KOPEN_CREAT)
	{
	   if (sizeof(mode_t) == sizeof(unsigned short))
              mode = (mode_t) kva_arg(args, unsigned int);
	   else
              mode = (mode_t) kva_arg(args, mode_t);
	}

	if ((file = kfile_open(path, flags, mode)) == NULL)
	{
	   kva_end(args);
	   return(-1);
	}
	kva_end(args);
	return(file->id);

}

/**************************************************************
*  
*  Routine Name: kclose - close and delete a transport descriptor
*  
*  Purpose: This function is a replacement for the system "close"
*	   call.  The only difference is that kclose() supports
*	   more than just files, it supports other data transports
*	   as well, such as shared memory, pipes, files, etc.
*  
*  
*    Input: id - the file to be close which was opened earlier
*		 with kopen().
*   Output: none
*  Returns: -1 or 0 depending on whether the operation failed or succeeded.
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

int kclose(
   int id)
{
	kfile	*file;

	/*
	 *  Get the file entry associated with the id
	 */
	if ((file = kfile_checkid(id)) == NULL)
	   return(-1);

	return(kfile_close(file));
}

/**************************************************************
*  
*  Routine Name: kread - read input from a transport descriptor
*  
*  Purpose: This function is a replacement for the system "read"
*	   call.  The only difference is that kread() supports
*	   more than just files, it supports other data transports
*	   as well, such as shared memory, pipes, files, etc.
*  
*	   The routine will read nbytes into the character array
*	   "buf".  If not all nbytes can be read then the kread()
*	   routine returns the number of bytes actually read.
*  
*  
*   Input: id - the file id to be read which was opened earlier
*		 with kopen().
*  
*	    buf - the buffer to read the data into
*  
*	    nbytes - the number of bytes to be read
*  
*   Output: none
*
*  Returns: the number of bytes read, 0 when end
*	    of file is encountered, or -1 when an error
*	    is encountered.
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

ssize_t kread(
   int    id,
   kaddr  buf,
   size_t nbytes)
{
	ssize_t	num;
	kfile	*file;

	/*
	 *  Get the file entry associated with the id
	 */
	if ((file = kfile_checkid(id)) == NULL)
	   return(-1);

	/*
	 *  Call the "read()" routine to read from the transport
	 */
	num = kfile_read(file, (char *)buf, (size_t) nbytes);
	return(num);
}

/**************************************************************
*  
*  Routine Name: kwrite - write output to a transport descriptor
*  
*  Purpose: This function is a replacement for the system "write"
*	   call.  The only difference is that kwrite() supports
*	   more than just files, it supports other data transports
*	   as well, such as shared memory, pipes, files, etc.
*  
*	   The routine will write nbytes from the character array
*	   "buf" to the appropriate transport mechanism specified by
*	   id.  If not all nbytes can be read then the kread()
*	   routine returns the number of bytes actually read.
*  
*  
*   Input: id - the file id to be read which was opened earlier
*		 with kopen().
*  
*	    buf - the buffer to write the data from
*  
*	    nbytes - the number of bytes to be written
*  
*  Output:  none
*
*  Returns: The number of bytes written, or -1 when
*	    an error is encountered.
*  
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

ssize_t kwrite(
   int   id,
   kaddr buf,
   size_t   nbytes)
{
	ssize_t	num;
	kfile	*file;

	/*
	 *  Get the file entry associated with the id
	 */
	if ((file = kfile_checkid(id)) == NULL)
	   return(-1);

	/*
	 *  Call the "write()" routine to write to the transport
	 */
	num = kfile_write(file, (char *)buf, nbytes);
	return(num);
}

/**************************************************************
*  
*  Routine Name: kaccess - determine accessibility of file
*  
*  Purpose: This function is a replacement for the system "access"
*	   call.  The only difference is that kaccess() checks
*	   more than just files, it will also check other data
*	   transports as well, such as shared memory, pipes, files,
*	   etc.
*  
*	   The routine will check to see if the token specified by
*	   the transport is either readable (R_OK), writeable (W_OK),
*	   or executable (X_OK), or accessible (F_OK) any possible
*	   combination by or'ing the flags together.
*  
*  Input: path - is the string containing the path name to the
*		 desired file to be accessed.  The path name ident-
*		 ifies the file/token to be tested, just like the
*		 regular UNIX access() command.
*  
*	  mode - the mode in which kaccess() should test for the
*		 file.  The mode is comprised of an or'ing of the
*		 following flags:
*  
*  	   !	R_OK - the file readable
*  	   !	W_OK - the file writeable
*  	   !	X_OK - the file executable
*  	   !	F_OK - the file and directory leading to the
*  	   !	       file accessible.
*  
*   Output: none
*
*  Returns: return -1 for failure otherwise 0 for success
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

int kaccess(
   const char *path,
   int         mode)
{
	return(kfile_access(path, mode));
}

/**************************************************************
*  
*  Routine Name: kunlink - remove a filename from a directory entry
*  
*  Purpose: This function is used to unlink the kfile.  Depending
*	   on the different transport being used the behavior
*	   will be different.  For a file it unlinks the file, but
*	   for a shared memory segment it initializes it before
*	   using the shmctl() to destroy it.
*  
*  
*   Input:  path - the path to the object ot be unlinked.
*  Returns: -1 or 0 depending on whether the
*	    operation failed or succeeded.
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

int kunlink(
   const char *path)
{
	TransportInformation *routines;
	char	*identifier, *filename, name[5*KLENGTH], itemp[5*KLENGTH],
		ftemp[5*KLENGTH];

	/*
	 *  kfullpath() is used to expand the path before opening it...
	 */
	if (!kfullpath(path, NULL, name))
	   return(FALSE);

	/*
	 *  According to the transport routines
	 */
	identifier = ktransport_identifier(name, itemp);
	if ((routines = ktransport_routines(identifier)) == NULL)
	   return(-1);

	/*
	 *  Get the filename component
	 */
	if ((filename = ktransport_path(name, ftemp)) == NULL)
	   return(-1);

	if (routines->unlink == NULL)
	   return(unlink(filename));
	else
	   return(routines->unlink(name, filename));
}

/************************************************************
*
*  Routine Name: kcreat - routine for creating a khoros transport
*
*       Purpose: This function is used to create a khoros transport
*		 file.  Depending on the different transport being
*		 used the behavior will be different.  For a file the
*		 kcreat call simply create the file with the mode,
*		 
*         Input: path - path to file to be created
*                mode - the permissions or mode in which to create the file
*       Returns: open fid on success, -1 otherwise
*    Written By: Mark Young
*          Date: Oct 23, 1993
*
*************************************************************/

int kcreat(
   const char *path,
   mode_t mode)
{
	int  fid;

	if ((fid = kopen(path,KOPEN_WRONLY|KOPEN_CREAT|KOPEN_TRUNC,mode)) == -1)
	   return(-1);

	return(fid);
}

/************************************************************
*
*  Routine Name: ktouch - routine for touching a temporary transport
*
*       Purpose: This function is used to create a khoros transport
*		 file.  Depending on the different transport being
*		 used the behavior will be different.  For a file the
*		 ktouch call simply open the file and then close
*		 it.  This is analgous to the "touch" system call.
*		 
*         Input: path - path to file to be created
*                mode - the permissions or mode in which to touch the file
*       Returns: 0 on success, -1 otherwise
*    Written By: Mark Young
*          Date: Apr 27, 1994
*
*************************************************************/

int ktouch(
   const char *path,
   mode_t  mode)
{
	int  fid;

	if ((fid = kopen(path, KOPEN_WRONLY|KOPEN_CREAT, mode)) == -1)
	   return(-1);

	return(kclose(fid));
}

/**************************************************************
*  
*  Routine Name: kfcntl - manipulates an open transport descriptor
*  
*  Purpose: 
*  Input: id  - the id of the object to be fcntl'ed.
*  
*	  cmd - the command to run
* Returns: -1 or 0 depending on whether the
*	   operation failed or succeeded.
*    Written By: Mark Young  
*          Date: Jan 24, 1996
*  
**************************************************************/
/* ARGSUSED */
int kfcntl(
   int id,
   int cmd,
   kvalist)

{
	return(-1);
}

/**************************************************************
*  
*  Routine Name: kflock - apply or remove an advisory lock on an open transport
*			  descriptor
*  
*  Purpose: This function is used to lock the transport. Depending
*	   on the different transport being used the behavior
*	   will be different.  For a file it calls flock() to
*	   lock the file, but for a shared memory segment it
*	   it calls shmctl() to lock & unlock the shared memory
*	   segment.
*  
*  
*  Input: id	 - the id of the object to be flock.
*  
*	  operation - the mode in which kflock() should lock the
*		   file.  The operation is comprised of an or'ing of the
*		   following flags:
*  
*  	  !	KLOCK_SH - lock file shareable (when reading)
*  	  !	KLOCK_EX - lock file execlusive (when writing)
*  	  !	KLOCK_NB - lock file for no block (don't block)
*  	  !	KLOCK_UN - unlock the file (free previous lock)
* Returns: -1 or 0 depending on whether the
*	   operation failed or succeeded.
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

int kflock(
   int id,
   int operation)
{
	kfile	*file;
	TransportInformation *routines;

	/*
	 *  Get the file entry associated with the id
	 */
	if ((file = kfile_checkid(id)) == NULL)
	   return(-1);

	/*
	 *  Call the "lock()" routine to close the transport
	 */
	routines = file->routines;
	if (routines->lock != NULL)
	   return(routines->lock(file, operation));

	return(0);
}

/**************************************************************
*  
*  Routine Name: klseek - move read/write pointer of a transport descriptor
*  
*  Purpose: This function is used to seek the transport. Depending
*	   on the different transport being used the behavior
*	   will be different.  For a file it calls lseek() to
*	   lock the file, but for a shared memory segment it
*	   it simply resets the internal offset pointer.
*  
*  
*  Input:	id	 - the id of the object to be seeked.
*  		offset - the offset in which to seek
*  		whence - the control of how the offset will be applied
*  Returns: -1 or the new seeked position within the transport
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

off_t klseek(
   int id,
   off_t offset,
   int whence)
{
	kfile	*file;

	/*
	 *  Get the file entry associated with the id
	 */
	if ((file = kfile_checkid(id)) == NULL)
	   return(-1);

	/*
	 *  Call the "kfile_seek()" routine to seek on the transport
	 */
	return(kfile_seek(file, offset, whence));
}

/**************************************************************
*  
*  Routine Name: ktell - report the position of the read/write pointer
*  
*  Purpose: This function is used to tell the current position
*	   within the transport. Depending on the different
*	   transport being used the behavior will be different.
*	   For a file it calls tell() to locate the offset
*	   within the file, but for a shared memory segment it
*	   it simply indicates the internal offset.
*  
*  
*   Input:  id	 - the id of the object to be told.
*  Returns: -1 or the current offset depending on whether the
*	    operation failed or succeeded.
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*  
**************************************************************/

long int ktell(
   int id)
{
	kfile	*file;

	/*
	 *  Get the file entry associated with the id
	 */
	if ((file = kfile_checkid(id)) == NULL)
	   return(-1);

	/*
	 *  Call the "kfile_tell()" routine to tell on the transport
	 */
	return(kfile_tell(file));
}

/************************************************************
*
*  Routine Name: kdup - duplicate an existing khoros transport descriptor
*
*       Purpose: kdup() is used to duplicate an existing khoros transport
*		 descriptor.  The "id" is integer index in the process's
*		 transport descriptor table.  The new descriptor returned
*		 kdup() will be the lowest table entry id.  
*
*         Input: id - the existing khoros transport descriptor to be dup'ed
*       Returns: the newly dup'ed descriptor on success, or -1 upon failure
*    Written By: Mark Young
*          Date: Apr 28, 1993
*  Side Effects: If too many descriptors are active then errno will be
*		 set to EMFILE and -1 will be returned.  If id is not
*		 a valid or active descriptor then errno will be set to
*		 EBADF and -1 will be returned.
*
*************************************************************/

int kdup(
   int id)
{
	int newid;

	/*
	 *  Get the file entry associated with the id
	 */
	if (!kfile_checkid(id))
	   return(-1);

	newid = kfile_dup(id, -1);
	return(newid);
}

/************************************************************
*
*  Routine Name: kdup2 - specifically duplicate an existing khoros transport
*			 descriptor
*
*       Purpose: kdup2() is used to duplicate an existing khoros transport
*		 descriptor.  The "id1" is integer index in the process's
*		 transport descriptor table.  The new descriptor returned
*		 will the specific one specified by "id2".  If id2 is active
*		 then it will be closed (kclose) before being dup'ed.
*
*         Input: id1 - the existing khoros transport descriptor to be dup'ed
*		 id2 - the descriptor to be used for the newly dup'ed descriptor
*       Returns: the newly dup'ed descriptor on success, or -1 upon failure
*    Written By: Mark Young
*          Date: Apr 28, 1993
*  Side Effects: If too many descriptors are active then errno will be
*		 set to EMFILE and -1 will be returned.  If id is not
*		 a valid or active descriptor then errno will be set to
*		 EBADF and -1 will be returned.
*
*************************************************************/

int kdup2(
   int id1,
   int id2)
{
	int newid;

	/*
	 *  Get the file entry associated with the id
	 */
	if (!kfile_checkid(id1))
	   return(-1);

	(void) kclose(id2);
	newid = kfile_dup(id1, id2);
	return(newid);
}

/**************************************************************
*  
*  Routine Name: kpopen - open a pipe (for I/O) from or to a process
*  
*  Purpose: open a pipe (for I/O) from or to a process
*
*    Input: command - command to send pipe output to or from
*	    type    - open status either 'r' or 'w'
*  Returns: The open file transport pointer
*    Written By: Mark Young  
*          Date: Jul 08, 1992
**************************************************************/

kfile *kpopen(
   const char *command,
   const char *type)
{
	kfile *file;
	char  *tmp, temp[5*KLENGTH];

        /*
         *  Open transport as a Pipe since the Pipe transport is always
	 *  available and understands piped commands as transports.
         */
	if (kremote_location(command, temp) != NULL)
           tmp = kstring_cat("remotexec=", command, NULL);
	else
           tmp = kstring_cat("pipe=", command, NULL);

        file = kfopen(tmp, type); kfree(tmp);
	return(file);
}

/**************************************************************
*  
*  Routine Name: kpclose - close a pipe (for I/O) from or to a process
*
*  Purpose: Close a pipe (for I/O) from or to a process
*
*    Input: file - tranport pointer to close
*  Returns: 0 on success, -1 on failure
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*   Declaration: int kpclose(
*                !   kfile *file)
**************************************************************/
/*ARGSUSED*/
int kpclose(
   kfile *file)
{
	int status;

	/*
	 *  Make sure the entry is within our table.  If not then it
	 *  has probably been closed already.
	 */
	if ((file = kfile_checkfile(file)) == NULL)
	{
	   errno = EBADF;
	   return(-1);
	}
	status = kfile_close(file);
	return(status);
}

/**************************************************************
*  
*  Routine Name: ksystem - issue a shell command
*  
*  Purpose: This function is a replacement for the "system" library
*	   call.  The only difference is that ksystem() supports
*	   more than just files, it supports other data transports
*	   in order to allow remote execution of a job.
*  
*  Input:  command - the command to be systemed
*  Returns: The job status
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*
**************************************************************/

int ksystem(
   const char *command)
{
	int pid, status = 0;
	void (*old_handler)(int);

	old_handler = ksignal(SIGCHLD, SIG_DFL);
	pid = kspawn(command);
	waitpid(pid, &status, 0);
	ksignal(SIGCHLD, old_handler);
        return(status);
}

/************************************************************
*
*  Routine Name: kspawn - spawn a command
*
*       Purpose: This function is used to spawn a command.  Similar to
*		 "ksystem" call, the kspawn() supports   more than just
*		 files, it supports other data transports in order to
*		 allow remote execution of a job.
*
*         Input: command - the command to be spawn
*       Returns: the sub-process id on success, -1 on failure
*    Written By: Mark Young
*          Date: Apr 14, 1994
*
*************************************************************/

pid_t kspawn(
   const char *command)
{
	pid_t pid;
	char  **args, *tmp;

	/*
	 *  According to the transport routines
	 */
	tmp = kexpand_variable(command, NULL);
	args = kcommand_to_args(tmp);
	if ((pid = kfork()) == 0)
	{
	   kexecvp(args[0], args);
	   fprintf(stderr,"\nUnable to execute command:\n\n '%s'\n", tmp);
	   _exit(KEXIT_FAILURE);
	}
	kfree(tmp);
	return(pid);
}

/**************************************************************
*  
*  Routine Name: kexecvp - execute a command
*  
*  Purpose: This function is a replacement for the system "execvp"
*	   call.  The only difference is that kexecvp() supports
*	   executing processes on remote machines.
*  
*	   The routine will execute another process (specified by
*	   arg0) and replace the current core image with the
*	   specified core.  For example, the command:
*  
*  	   !	"vfileinfo -i1 file"
*  
*	   would be specified as:
*  
*  	   !		arg0 = "vfileinfo" 
*          !
*  	   !		args[0] = "vfileinfo" 
*  	   !		args[1] = "-i1" 
*  	   !		args[2] = "file" 
*  	   !		args[3] = NULL
*  
*	   If the command is to be executed on a different machine such
*	   as "borris" then the same command would look like:
*  
*  	   !		"vfileinfo@borris -i1 file"
*  
*	   would be specified as:
*  
*  	   !		arg0 = "vfileinfo@borris" 
*          !
*  	   !		args[0] = "vfileinfo@borris" 
*  	   !		args[1] = "-i1" 
*  	   !		args[2] = "file" 
*  	   !		args[3] = NULL
*  
*  
*	   If the command is to be executed on a different machine such
*	   as "borris", and the file to be read is on "natasha" then
*	   the command would look like this:
*  
*  	   !		"vfileinfo@borris -i1 file@natasha"
*  
*	   would be specified as:
*  
*  	   !		arg0 = "vfileinfo@borris" 
*          !
*  	   !		args[0] = "vfileinfo@borris" 
*  	   !		args[1] = "-i1" 
*  	   !		args[2] = "file@natasha" 
*  	   !		args[3] = NULL
*  
*  
*         Input: arg0 - the routine to be executed
*	         args - the arguments, include arg0, in which execute
*		        the command
*       Returns: Note: this call overlays the calling process with the named
*	         file, then transfers control to the new core image.  If the
*	         command is successfully executed then there will be no return
*	         from the new process.  If the core cannot be loaded or found
*  	         then we return with an error of -1.
*    Written By: Mark Young  
*          Date: Jul 08, 1992
**************************************************************/

int kexecvp(
   const char *arg0,
   char *const args[])
{
/*
 * This is so Irix 4.x will compile - JMAS.
 */
#if defined(__sgi) && defined(__SVR3)
	execvp(arg0, (const char **) args);
#else
	execvp(arg0, args);
#endif
	return(-1);
}


/**************************************************************
*  
*  Routine Name: kfileno - return the transport descriptor
*  
*  Purpose: This function is used to retrieve the corresponding
*	   kfile entry descriptor, which can be used with the
*	   (int fid) routines like: kread(), kwrite().
*  
*  Input: file - the kfile structure to be referenced.
*  Returns: The entry descriptor (or -1 if it doesn't exist)
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*
**************************************************************/
int kfileno(
   kfile *file)
{
	/*
	 *  Make sure the entry is within our table.  If not then it
	 *  has probably been closed already.
	 */
	if ((file = kfile_checkfile(file)) == NULL)
	{
	   errno = EBADF;
	   return(-1);
	}
	else
	   return(file->id);
}

/**************************************************************
*  
*  Routine Name: kfidfile - return the kfile structure associated with a fid
*  
*  Purpose: This function is used to retrieve the corresponding
*	   kfile structure given a khoros transport descriptor (fid).
*         Input: id   - the transport identifier opened earlier
*       Returns: The kfile pointer representing the open stream 
*                to the data transport on success, NULL on failure
*    Written By: Mark Young  
*          Date: Sept 4, 1995
* Modifications:
**************************************************************/

kfile *kfidfile(
   int  id)
{
	kfile *file;

	/*
	 *  Make sure the entry is within our table.  If not then it
	 *  has probably been closed already.
	 */
	if ((file = kfile_checkid(id)) == NULL)
	{
	   errno = EBADF;
	   return(NULL);
	}
	return(file);
}

/************************************************************
*
*  Routine Name: ktmpfid - create a temporary transport descriptor
*       Purpose: This function is similar to ktmpfile(), except that
*		 it returns a transport descriptor instead of a transport
*		 structure.
*                The data transport will automatically be deleted when closed.
*       Returns: The fid descriptor representing the temporary transport
*                on success, -1 on failure
*    Written By: Mark Young
*          Date: Sept 4, 1995
*************************************************************/

int ktmpfid(void)
{
	kfile *file;

	/*
	 *  Create temporary file and open it for reading and writing
	 */
	if ((file = ktmpfile()) == NULL)
	   return(-1);

	return(file->id);
}

/**************************************************************
*  
*  Routine Name: kexit_handler - adds an kexit handler
*  
*  Purpose: kexit_handler adds an handler so that upon call of kexit(),
*	    the handler is called.  kexit_handler() can be called
*	    several times, in which case the termination handlers
*	    will be called in reverse order of their instantiation.  The
*	    handlers should be declared as follows:
*	    !
*	    ! void handler(
*	    !    int status,
*	    !    kaddr client_data)
*	    !
*	    
*  
*  Input: handler - the handler to be called
*	  client_data - the data to passed when calling the routine
*  Returns: TRUE on success, FALSE on failure
*    Written By: Mark Young
*          Date: May 17, 1993 15:49
*   Declaration: int kexit_handler(
*		 !   kfunc_void handler,
*		 !   kaddr      client_data)
*
**************************************************************/

typedef struct _khandler
{
	void (*handler) PROTO((kexit_status, kaddr));
	kaddr      client_data;
} khandler;

static klist *kexit_list = NULL;


int kexit_handler(
   void (*handler) PROTO((kexit_status, kaddr)),
   kaddr client_data)
{
	khandler *temp;

	/*
	 *  Sanity check
	 */
	if (!handler)
	   return(FALSE);

	/*
	 *  Malloc handler
	 */
	if ((temp = (khandler *) kcalloc((size_t) 1, sizeof(khandler))) == NULL)
	   return(FALSE);

	temp->handler = handler;
	temp->client_data = client_data;

	kexit_list = klist_insert(kexit_list, (kaddr) temp, NULL,
				  KLIST_HEAD, FALSE);
	return(TRUE);
}

/**************************************************************
*  
*  Routine Name: kexit - terminate a process
*  
*  Purpose: closes any Khoros transports left open and
*           exits the system with the status passed in.
*  
*  Input: status - exit status of either KEXIT_SUCCESS or KEXIT_FAILURE
*  Returns:
*    Written By: Mark Young
*          Date: Aug 14, 1992
*
**************************************************************/

void kexit(
   kexit_status status)
{
	kfile    *ipc;
	khandler *tmp;
	kaddr    client_data;
	klist    *list = kexit_list;
	void     (*handler) PROTO((kexit_status, kaddr));
	char     temp[5*KLENGTH], command[5*KLENGTH], **args = NULL;

	while (list != NULL)
	{
	   tmp = (khandler *) klist_identifier(list);
	   handler = tmp->handler;
	   client_data = tmp->client_data;
	   kfree(tmp);

	   list = klist_next(list);
	   handler(status, client_data);
	}
	klist_free(kexit_list, NULL); kexit_list = NULL;

        if (status == KEXIT_SUCCESS && (ipc = kipc_getipc()) != NULL)
        {
           ksprintf(temp, "%s:%d: exit process\n", KIPC_PROCESS, getpid());
	   if (kfputs(temp, ipc) != EOF &&
	       kfgets(temp, KLENGTH, kstdin) != NULL &&
	       ksscanf(temp, "%[^\n]", command) == 1)
	   {
#if 0
	      if (kstrlen(command) > 0 && kstrcmp(command, "exit") != 0)
	         args = kcommand_to_args(command);
#endif
	   }
        }

        /*
         *  Do final cleanup by closing all remaining transports and
         *  calling explicit routines to cleanup any remaining memory
         *  in use.
         */
	kfile_closeall();
	kfile_cleanup();
	kenvp_cleanup();
	kerrno_cleanup();

	/*
	 *  If the current process is suppose to be re-execed then call
	 *  execvp to replace our current process.
	 */
	if (args && args[0])
	   execvp(args[0], args);

	/*  really exit with our exit status */
	exit((int) status);
}

/**************************************************************
*  
*  Routine Name: kgetdescriptors - get true UNIX file descriptors
*  
*  Purpose: This function is used to retrieve the input and
*	   output file descriptor associated with a transport.
*	   Note:  this routine should not be used without
*		   extreme caution, since certain transports
*		   don't have file descriptors (such as shared
*		   memory).
*  
*  Input: file - the kfile structure to be referenced.
*  Output: inum - the input descriptor if exists (or if non exists minus 1)
*	   onum - the output descriptor if exists (or if non exists minus 1)
*  Returns: TRUE on success, FALSE on failure
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*
**************************************************************/

int kgetdescriptors(
   kfile  *file,
   int    *inum,
   int    *onum)
{
	int	status;
	TransportInformation *routines;

	/*
	 *  Make sure the entry is within our table.  If not then it
	 *  has probably been closed already.
	 */
	if ((file = kfile_checkfile(file)) == NULL)
	{
	   errno = EBADF;
	   return(-1);
	}

	/*
	 *  Call the "filedescriptor()" routine to see if a input and output
	 *  file descriptor exist
	 */
	routines = file->routines;
	if (routines->descriptors != NULL)
	{
	   status = routines->descriptors(file, inum, onum);
	}
	else
	{
	   if (inum != NULL) *inum = -1;
	   if (onum != NULL) *onum = -1;
	   status = -1;
	}
	return(status);
}

/**************************************************************
*
*  Routine Name: kgethostname - get the current hostname
*  
*  Purpose:  This routine is the same as the system gethostname()
*	     call except that it will get the hostname depending on
*	     the machine name is specified.
*  
*  Input:   namelen - the size of the "hostname" buffer
*  Output:  hostname - the current hostname string
*  Returns: 0 on success, and -1 on failure
*    Written By: Mark Young  
*          Date: Jul 08, 1992
*
**************************************************************/

int kgethostname(
   char   *hostname,
   int     namelen)
{
	if (gethostname(hostname, namelen) == -1)
	   return(-1);

	return(0);
}
