/*
 * Khoros: $Id: token.c,v 1.9.2.4 1997/05/30 22:26:39 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: token.c,v 1.9.2.4 1997/05/30 22:26:39 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Token Utilities
   >>>>		
   >>>>  Private:
   >>>>   Public:
   >>>>                 kstring_to_token()
   >>>>                 ktoken_to_string()
   >>>>                 ktoken_check()
   >>>>                 ktoken_delete()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


/*
 *  This is the next available token, hash table, string table
 *
 *  NOTE:  token_num must be initially 1.
 */
static khash  *hash = NULL;

/*-----------------------------------------------------------
|
|  Routine Name: cleanup_handler - private routine for cleanup on exit
|
|       Purpose: The cleanup handler routine frees memory on termination
|		 of the program.
|
|         Input: status - the exit status of the program
|		 clientdata - not used
|    Written By: Mark Young  
|          Date: Mar 5, 1997
| Modifications:
|
------------------------------------------------------------*/
/* ARGSUSED */
static void cleanup_handler(
   kexit_status status,
   kaddr clientdata)
{
	khash_free(hash, NULL);
	hash = NULL;
}

/************************************************************
*
*  Routine Name: kstring_to_token - return the token that is associated with
*				 the specified string
*
*       Purpose: This function returns the token that is associated with
*		 the supplied string.  The idea is that a given string is
*		 tagged with a unique token, so that an application which is
*		 being slowed down by numerous string comparisons or searches
*		 through string attributes, can use the unique tokens for
*		 a value comparison instead using a traditional string
*		 compare call.  Thus, greatly reducing the comparison time.
*
*		 kstring_to_token() returns unique tokens for different
*		 strings, so that calling routines can perform string
*		 comparisons using "token1 == token2" instead of the
*		 time intensive kstrcmp(str1, str2) function call.
*
*		 If the string is NULL or if the space to copy the string
*		 cannot be allocated, then KTOKEN_NONE is returned.
*
*      Examples: For example to instatiate a string, using the function
*		 call kstring_to_token(), the programmer uses the following
*		 call:
*
*		 .begin code
*		 token1 = kstring_to_token("test string1");
*		 token2 = kstring_to_token("test string2");
*		 .end code
*
*		 This should yield two totally distinct tokens, which when
*		 compared (token1 == token2) will result that the two strings
*		 are not the same.
*
*	  Input: istr - The string to compute the token for.
*       Returns: The token computed or KTOKEN_NONE if an error occurs.
*    Written By: Mark Young
*          Date: Nov 16, 1992
*************************************************************/

ktoken kstring_to_token(
   const char *istr)
{
	long indx;
	char **table;


	/*
	 *  Sanity check
	 */
	if (istr == NULL)
	   return(KTOKEN_NONE);

	/*
	 *  First try to see if the token already exists...
	 */
	if (hash == NULL)
	{
	   hash = khash_init(KSTRING, 4013);
	   kexit_handler(cleanup_handler, NULL);
	}
	else if ((indx = khash_location(hash, istr)) != -1)
	{
	   table = (char **) hash->table;
	   return(table[indx]);
	}

	khash_add(hash, istr, NULL);
	if ((indx = khash_location(hash, istr)) == -1)
	   return(KTOKEN_NONE);

	table = (char **) hash->table;
	return((ktoken) table[indx]);
}

/************************************************************
*
*  Routine Name: ktoken_to_string - return the string associated with
*				 the specified token
*
*       Purpose: This function returns the string that is associated with
*		 the supplied token.  The idea is that a unique token
*		 is given for each string sent to the function
*		 kstring_to_token().  This routine is used to retrieve a
*		 pointer to the token's associated string.
*
*	  Input: token - The token to return the corresponding string for.
*       Returns: The token's corresponding string or NULL, if the token is
*		 not valid.
*    Written By: Mark Young
*          Date: Nov 16, 1992
*************************************************************/

char *ktoken_to_string(
   ktoken token)
{
	return((char *) token);
}

/************************************************************
*
*  Routine Name: ktoken_check - check to see if a string has been token'ized
*
*       Purpose: This function checks to see if the input string has 
*		 a token representation.
*
*	  Input: istr - The string to check for a token.
*       Returns: If the string is NULL or if the string has not been
*		 token'ized, KTOKEN_NONE is returned.  Otherwise the token
*		 corresponding to the string is returned.
*    Written By: Mark Young
*          Date: Nov 16, 1993
*************************************************************/

ktoken ktoken_check(
   const char *istr)
{
	long indx;
	char **table;

	/*
	 *  return whether a token already exists...
	 */
	if ((indx = khash_location(hash, istr)) == -1)
	   return(KTOKEN_NONE);

	table = (char **) hash->table;
	return((ktoken) table[indx]);
}

/************************************************************
*
*  Routine Name: ktoken_delete - delete the token'ized string from the list
*				 of tokens
*
*       Purpose: This function deletes the token'ized representation of a
*		 string from the token list.  If the string has a token
*		 representation then it is removed from the list.
*
*		 If the string is NULL, or the string or the token'ized
*		 string cannot be deleted, then FALSE is returned.
*
*	  Input: string - The string to delete.
*       Returns: TRUE (1) on success, FALSE (0) otherwise.
*    Written By: Mark Young
*          Date: Nov 16, 1993
*************************************************************/
int ktoken_delete(
   const char *string)
{
	return(khash_delete(hash, string));
}
