 /*
  * Khoros: $Id: lkcompare.c,v 1.16.2.3 1997/05/31 19:05:52 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lkcompare.c,v 1.16.2.3 1997/05/31 19:05:52 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for kcompare
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lkcompare
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/*
 * Operation structure declaration
 */
struct _operation
{
        char    *name;
	int     routine;
};

/*
 * Function declaration list
 */
static struct _operation operations[] =
{
#define KEQ     1
    { "eq",      KEQ },
#define KNE     2
    { "ne",      KNE  },
#define KGT     3
    { "gt",      KGT },
#define KGE     4
    { "ge",      KGE },
#define KLT     5
    { "lt",      KLT },
#define KLE     6
    { "le",      KLE },
};
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lkcompare - perform comparison operation on data object(s)
* 
*       Purpose: The following is the library routine that performs
*		 a comparison operation on two data objects.
*		 The routine takes two input data object (or a single data
*		 object and a constant), a function name, and an output
*		 data object in which to store the result of the
*		 function.
*
*		 The supported functions are:
*                .TS 
*                center tab(:) ;
*                l l .
*		 eq  :  o = i1 == (i2 or val)
*		 ne  :  o = i1 != (i2 or val)
*		 gt  :  o = i1 >  (i2 or val)
*		 ge  :  o = i1 >= (i2 or val)
*		 lt  :  o = i1 <  (i2 or val)
*		 le  :  o = i1 <= (i2 or val)
*                .TE
*
*		 Therefore if lkcompare() were called with two input data
*		 objects:
*
*		 !    lkcompare(i1, i2, 0.0, 1.0, 0.0, "eq", o);
*
*		 then the following would be performed:
*
*		 !    o = (i1 == i2) ? 1.0 : 0.0
*
*		 If lkcompare were called with a single data object and
*		 a constant:
*
*		 !    lkcompare(i1, NULL, 10.0, 1.0, 0.0, "eq", o);
*
*		 then the following would be performed:
*
*		 !    o = (i1 == 10.0) ? 1.0 : 0.0
*
*         Input: i1   - the input object to be processed
*		 i2   - the second input object (if NULL then uses the
*			 constant "val")
*		 val  - the constant to be instead of the second data object
*		 function - the function to be applied
*		 tval - the value to be assigned if comparison is TRUE
*		 fval - the value to be assigned if comparison is FALSE
*		 tol  - the plus/minus tolerence in which to do the comparison
*			(not functional yet)
*        Output: o    - the output object
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*  Restrictions: 
*    Written By: Donna Koechner & Mark Young
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lkcompare (
   kobject i1,
   kobject i2,
   double  val,
   char    *function,
   double  tval,
   double  fval,
   double  tol,
   kobject o)
/* -library_def_end */

/* -library_code */
{
        char *lib = "kdatamanip", *rtn = "lkcompare";
        klist *objlist=NULL;
	double	*idata1 = NULL;
        double  *idata2 = NULL;
        double  *odata  = NULL;
	int	routine, j, indx, num;
        int     num_lines,line_size;
        int	type1, type2, w1, w2, h1, h2,
		d1, d2, t1, t2, e1, e2;

        /* Reference the input objects to avoid side effects */
	KCALL((i1 = kpds_reference_object(i1)) != KOBJECT_INVALID);
        objlist = klist_add(objlist,i1,"KOBJECT");
        if (i2 != NULL)
          {
	    KCALL((i2 = kpds_reference_object(i2)) != KOBJECT_INVALID);
            objlist = klist_add(objlist,i2,"KOBJECT");
          }

        /* See what type of operation we are to perform */
        for (num = 0, indx = -1; num < knumber(operations); num++)
        {
           if (kstrcmp(function, operations[num].name) == 0)
           {
              indx = num;
              break;
           }
        }
        if (indx == -1)
        {
           kerror(lib, rtn, "Sorry, but the function "
                            "'%s' is not in the list of callable operations",
                            function);
           (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
           return(FALSE);
        }
        else if (operations[indx].routine == KNONE)
        {
           kerror(lib, rtn, "Sorry, but I am unable to run the operation "
                            "'%s' on data services objects at this time.", 
                            function);
           (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
           return(FALSE);
        }
        else
           routine = (int) operations[indx].routine;

        /* Come up with a reasonable output data type, based on the
           data types of the input data and the output values */
	KCALL(kpds_get_attribute(i1, KPDS_VALUE_DATA_TYPE, &type1));
        if (i2 != NULL)
        {
	   KCALL(kpds_get_attribute(i2, KPDS_VALUE_DATA_TYPE, &type2));
           type1 = kmax(type1, type2);
        }
        /* If we are handling an unsigned data type, but the output values
           are signed, then upgrade the data type to signed. If we have a
           small representable range (ubyte) then upgrade to SHORT as well.
           If the input is BIT and the T/F values are not 0 or 1 then
           just assume an output type of BYTE. None of these games guarantee 
           that the output values will fit properly within that data type, but 
           it's much better than trying to shove a signed output value into an 
           unsigned output data type! */ 
        if (type1 == KBIT && (fval < 0 || tval < 0 ||
                              fval > 1 || tval > 1)) type1 = KBYTE;
        else if (type1 == KUBYTE && (fval < 0 || tval < 0)) type1 = KSHORT;
        else if (type1 == KUSHORT && (fval < 0 || tval < 0)) type1 = KSHORT;
        else if (type1 == KUINT && (fval < 0 || tval < 0)) type1 = KINT;
        else if (type1 == KULONG && (fval < 0 || tval < 0)) type1 = KLONG;
        if (type1 > KDOUBLE)
          {
            kerror(lib, rtn, "Sorry, cannot handle KCOMPLEX or higher data");
            (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
            return(FALSE);
          }
        KCALL(kpds_set_attribute(o, KPDS_VALUE_DATA_TYPE, type1));

	/*
	 *  Get the maximum size of the value segment for i1 & i2 and make
	 *  sure that i1, i2, o are the same.
	 */
	KCALL(kpds_get_attribute(i1, KPDS_VALUE_SIZE, &w1, &h1, &d1, &t1, &e1));
        if (i2 != NULL)
        {
	   KCALL(kpds_get_attribute(i2, KPDS_VALUE_SIZE, 
                                        &w2, &h2, &d2, &t2, &e2));
	   w1 = kmax(w1, w2); h1 = kmax(h1, h2); d1 = kmax(d1, d2);
	   t1 = kmax(t1, t2); e1 = kmax(e1, e2);
	   KCALL(kpds_set_attribute(i2, KPDS_VALUE_SIZE, w1, h1, d1, t1, e1));
	   KCALL(kpds_set_attribute(i1, KPDS_VALUE_SIZE, w1, h1, d1, t1, e1));
        }
        KCALL(kpds_set_attribute(o, KPDS_VALUE_SIZE, w1, h1, d1, t1, e1));

        /* Now reference the output object - we'll muck with the presentation
           to simplify processing. */
	KCALL((o = kpds_reference_object(o)) != KOBJECT_INVALID);
        objlist = klist_add(objlist,o,"KOBJECT");
 
        /* Set the data type of the input and output presentation to DOUBLE
           for processing. */
	KCALL(kpds_set_attribute(i1, KPDS_VALUE_DATA_TYPE, KDOUBLE));
        if (i2 != NULL)
        {
	   KCALL(kpds_set_attribute(i2, KPDS_VALUE_DATA_TYPE, KDOUBLE));
        }
	KCALL(kpds_set_attribute(o, KPDS_VALUE_DATA_TYPE, KDOUBLE));

        /* Allocate work buffers */
        KCALL(kpds_get_attribute(i1, KPDS_VALUE_LINE_INFO, &line_size,
                                                           &num_lines));
        KCALL(!((odata = (double *)kmalloc((size_t)line_size*sizeof(double)))
                        == NULL));
        objlist = klist_add(objlist,odata,"KMALLOC");
        KCALL(!((idata1 = (double *)kmalloc((size_t)line_size*sizeof(double)))
                         == NULL));
        objlist = klist_add(objlist,idata1,"KMALLOC");
        if (i2 != NULL)
          {
            KCALL(!((idata2=(double *)kmalloc((size_t)line_size*sizeof(double)))
                            == NULL));
            objlist = klist_add(objlist,idata2,"KMALLOC");
          }

        /*
	 *  Do the processing by simply requesting the lines of data from
	 *  i1 & i2 and then evaluate the symbol list and then store the
	 *  result into the output. 
	 */
	for (num = 0; num < num_lines; num++)
	{
	   KCALL(kpds_get_data(i1, KPDS_VALUE_LINE, (kaddr)idata1));
      
           if (i2 != NULL)
           {
	      KCALL(kpds_get_data(i2, KPDS_VALUE_LINE, (kaddr)idata2));
           }

	   for (j = 0; j < line_size; j++)
	   {
	      switch (routine)
	      {
		  case KEQ:
		       odata[j] = (idata1[j] == ((!i2) ? val : idata2[j]) ?
					tval : fval);
		       break;

		  case KNE:
		       odata[j] = (idata1[j] != ((!i2) ? val : idata2[j]) ?
					tval : fval);
                       break;

		  case KGT:
		       odata[j] = (idata1[j] > ((!i2) ? val : idata2[j]) ?
					tval : fval);
                       break;

		  case KGE:
		       odata[j] = (idata1[j] >= ((!i2) ? val : idata2[j]) ?
					tval : fval);
                       break;

		  case KLT:
		       odata[j] = (idata1[j] < ((!i2) ? val : idata2[j]) ?
					tval : fval);
                       break;

		  case KLE:
		       odata[j] = (idata1[j] <= ((!i2) ? val : idata2[j]) ?
					tval : fval);
                       break;
	      }
	   }

	   KCALL(kpds_put_data(o, KPDS_VALUE_LINE, (kaddr) odata));
	}

        (void)klist_free(objlist, (void(*)(klist*))lkcall_free);

	return(TRUE);
}
/* -library_code_end */
