 /*
  * Khoros: $Id: lkelevation.c,v 1.10.2.2 1997/05/31 19:05:54 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lkelevation.c,v 1.10.2.2 1997/05/31 19:05:54 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for kelevation
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lkelevation
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
static int init_location PROTO((kobject, double, double, double,
				double, int, int, int));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lkelevation - use plane of value data to construct elevation
* 
*       Purpose: This routine will create explicit curvilinear location
*		 data containing a plane of elevation data.
*		 
*		 The generated location data will consist of a single
*		 width-height plane of location data, with a dimension
*		 size of three, for storing (x,y,z) values.  The (x,y)
*		 values are generated across a uniform range according
*		 to the specified (xbegin - xend) and (ybegin - yend)
*		 ranges.  The z-values will be filled out with a
*		 single plane of data extracted from the value segment
*		 of a provided input object.  The depth, time, and
*		 element position in the value segment of the plane to
*		 be used for z-values can also be specified.
*
*		 The width and height of the value segment in the
*		 provided input object will dictate the width and
*		 height of the resulting location segment.  The
*		 datatype of the resulting location segment can either
*		 be propagated from the input value segment, or can be
*		 specified explicitly.  If set explicitly, the data
*		 will be cast to this type as it is retrieved from the
*		 value segment.
*		 
*		 The (x,y) data is generated and stored as one block,
*		 so generating large elevation data sets is subject to
*		 the memory limitations of your machine.
*
*         Input: src_obj - the source object containing a value segment
*                xbegin  - the x range begin point
*                ybegin  - the y range begin point
*		 xend    - the x range end point
*		 yend    - the y range end point
*
*		 dpos    - depth position at which to extract value plane
*		 tpos    - time position at which to extract value plane
*		 epos    - elevation position at which to extract value plane
*
*		 datatype_string - datatype name to use when constructing
*				   the location data. Use "Propagate Value 
*			           Type" to propogate the input value 
*				   datatype to the output location.
*
*        Output: dst_obj - the destination object
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: Destination object must not already contain location data.
*    Written By: Steve Kubica
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lkelevation(
   kobject src_obj,
   double  xbegin,
   double  ybegin,
   double  xend,
   double  yend,
   int     dpos,
   int     tpos,
   int     epos,
   char   *datatype_string,
   kobject dst_obj)
/* -library_def_end */

/* -library_code */
{
   char *lib  = "kdatamanip";
   char *prog = "lkelevation";

   kaddr vdata = NULL;
   int   wid, hgt;
   int   datatype;
   int   vdatatype;
   
   /*-- work out the datatype we want --*/
   if (kstrcmp(datatype_string, "Propagate Value Type"))
   {
      datatype = kdatatype_to_define(datatype_string);

      if (datatype == 0)
      {
	 kerror(lib, prog, "Not a proper data type '%s'.", datatype_string);
	 return FALSE;
      }
   }

   /*-- verify that we have value data to work with --*/
   if (!kpds_query_value(src_obj))
   {
      kerror(lib, prog, "Source object must contain a value segment.");
      return FALSE;
   }

   /*-- get size and datatype information from the value segment --*/
   if (!kpds_get_attributes(src_obj, 
                            KPDS_VALUE_SIZE, &wid, &hgt, NULL, NULL, NULL, 
                            KPDS_VALUE_DATA_TYPE, &vdatatype, 
			    NULL))
   {
      kerror(lib, prog, 
             "Unable to retrieve size or data type of value segment");
      return FALSE;
   }

   /*-- propogate input value datatype, if desired --*/
   if (kstrcmp(datatype_string, "Propagate Value Type") == 0)
      datatype = vdatatype;

   /*-- initialize the (x,y) location data in the destination --*/
   if (!init_location(dst_obj, xbegin, ybegin, xend, yend, wid, hgt, datatype))
   {
      kerror(lib, prog, "Unable to initialize location segment.");
      return FALSE;
   }
   
   /*-- want plane of value data from desired place in desired data type --*/
   if (!kpds_set_attributes(src_obj, 
                            KPDS_VALUE_DATA_TYPE, datatype,
                            KPDS_VALUE_POSITION, 0, 0, dpos, tpos, epos, 
                            NULL))
   {
      kerror(lib, prog, "Unable to set attributes on value segment.");
      return FALSE;
   }

   /*-- want to put plane of value data in location segment z-plane --*/
   if (!kpds_set_attribute(dst_obj, KPDS_LOCATION_POSITION, 0, 0, 0, 2))
   {
      kerror(lib, prog, "Unable to set attributes on location segment.");
      return FALSE;
   }

   /*-- get the plane of value data... --*/
   if ((vdata = kpds_get_data(src_obj, KPDS_VALUE_PLANE, NULL)) == NULL)
   {
      kerror(lib, prog, "Unable to get data from the value segment.");
      return FALSE;
   }

   /*-- ... and put it in the location segment --*/
   if (!kpds_put_data(dst_obj, KPDS_LOCATION_PLANE, vdata))
   {
      kerror(lib, prog, "Unable to put data into the location segment.");
      return FALSE;
   }

   kfree(vdata);
   
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: init_location - construct curvilinear location data
|
|       Purpose: This function constructs a block of curvilinear 
|                data whose values range uniformly between the
|                specified x and y begin and end points.  The
|		 constructed location segment will have space
|		 for z values, but the z values will not be 
|		 initialized. 
|
|                The generated location data will be put with a 
|                location all primitive.
|
|         Input: obj      - object to create location segment in
|                xbegin   - x axis begin point
|                ybegin   - y axis begin point
|                xend     - x axis end point
|                yend     - y axis end point
|                wid      - width size for the constructed data
|                hgt      - height size for the constructed data
|                datatype - data type for the constructed data
|
|        Output: obj      - location segment is created and 
|		            initialized in obj
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: January 25, 1996
| Modifications:
|
------------------------------------------------------------*/
static int 
init_location(
   kobject obj,
   double  xbegin,
   double  ybegin,
   double  xend,
   double  yend,
   int     wid,
   int     hgt,
   int     datatype)
{
   kaddr  ldata = NULL;
   double del[3]  = {0.0, 0.0, 0.0};
   double ubeg[3] = {0.0, 0.0, 0.0};
   int    k[3];
   int    n;


   /*-- set the grid type in the destination object --*/
   if (!kpds_set_attribute(obj, KPDS_LOCATION_GRID, KCURVILINEAR))
      return FALSE;

   /*-- create the location segment in the destination --*/
   if (!kpds_create_location(obj))
      return FALSE;

   /*-- set the size and data type as specified --*/
   if (!kpds_set_attributes(obj,
                            KPDS_LOCATION_SIZE, wid, hgt, 1, 3,
                            KPDS_LOCATION_DATA_TYPE, datatype,
                            NULL))
      return FALSE;
   
   /*-- allocate sufficient data space --*/
   ldata = kmalloc((size_t)(wid*hgt*3 * kdata_size(datatype)));

   /*-- determine the deltas between adjacent points (no 0 deltas!) --*/
   del[0] = (xend - xbegin)/(double) ((wid-1 == 0) ? 1 : (wid-1));
   del[1] = (yend - ybegin)/(double) ((hgt-1 == 0) ? 1 : (hgt-1));

   /*-- this information will be much easier to manage in arrays --*/
   ubeg[0] = xbegin;
   ubeg[1] = ybegin;

   switch (datatype)
   {
      case KBYTE :
      {
         char *data = (char *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (char) ubeg[n] + del[n] * k[n];
      } break;

      case KUBYTE :
      {
         unsigned char *data = (unsigned char *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (unsigned char) ubeg[n] + del[n] * k[n];
      } break;

      case KSHORT :
      {
         short *data = (short *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (short) ubeg[n] + del[n] * k[n];
      } break;

      case KUSHORT :
      {
         unsigned short *data = (unsigned short *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (unsigned short) ubeg[n] + del[n] * k[n];
      } break;

      case KINT :
      {
         int *data = (int *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (int) ubeg[n] + del[n] * k[n];
      } break;

      case KUINT :
      {
         unsigned int *data = (unsigned int *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (unsigned int) ubeg[n] + del[n] * k[n];
      } break;

      case KLONG :
      {
         long *data = (long *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (long) ubeg[n] + del[n] * k[n];
      } break;

      case KULONG :
      {
         unsigned long *data = (unsigned long *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (unsigned long) ubeg[n] + del[n] * k[n];
      } break;

      case KFLOAT :
      {
         float *data = (float *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (float) ubeg[n] + del[n] * k[n];
      } break;

      case KDOUBLE :
      {
         double *data = (double *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
                  *data++ = (double) ubeg[n] + del[n] * k[n];
      } break;

      case KCOMPLEX :
      {
         kcomplex *data = (kcomplex *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
               {
                  (*data).r = (float) ubeg[n] + del[n] * k[n];
                  (*data).i = (float) 0.0;
                  data++;
               }
      } break;

      case KDCOMPLEX :
      {
         kdcomplex *data = (kdcomplex *) ldata;

         for (n = 0; n < 3; n++)
            for (k[1] = 0; k[1] < hgt; k[1]++)
               for (k[0] = 0; k[0] < wid; k[0]++)
               {
                  (*data).r = (double) ubeg[n] + del[n] * k[n];
                  (*data).i = (double) 0.0;
                  data++;
               }
      } break;

      default:
         return FALSE;
   }

   /*-- set the region size to be the planes of (x,y) data ...  --*/
   if (!kpds_set_attribute(obj, KPDS_LOCATION_REGION_SIZE, wid, hgt, 1, 2))
      return FALSE;
   
   /*-- ... and put the data --*/
   if (!kpds_put_data(obj, KPDS_LOCATION_REGION, ldata))
      return FALSE;

   kfree(ldata);

   return TRUE;
}
/* -library_code_end */
