 /*
  * Khoros: $Id: lktranspose.c,v 1.1.2.3 1997/05/31 19:07:15 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lktranspose.c,v 1.1.2.3 1997/05/31 19:07:15 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for ktranspose
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lktranspose
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
#include "internals.h"


/*-- this is a private function in kdataccess, but it does what we want --*/
kaddr kdas_transpose_data  PROTO((kaddr, int, int, int *, int *,
				  int *, int *, kaddr));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lktranspose - transpose data dimensions
* 
*       Purpose: The following is the library routine that performs
*	 	 data transposition with respect to the 5 polymorphic
*		 dimensions, width, height, depth, time, and elements.
*                The routine takes an input data object and a newn
*		 index order defined by the norder array, which specifies how 
*		 each dimension in the old object will be mapped to the 
*		 new object. The data is transposed, with the old
*		 dimensional sizes reassigned to the desired new 
*		 dimensions.  
*
*		 The norder_map array is a set of five integers that 
*		 can have one of the following values: 
*		 !	KWIDTH
*		 !	KHEIGHT
*		 !	KDEPTH
*		 !	KTIME
*		 !	KELEMENTS
*
*		 Each dimension can only be mapped to once.  
*		 
*		 norder[0] 
*		 !	defines where the WIDTH dimension of the source 
*		 !	object will be mapped to in the destination object.
*		 norder[1] 
*		 !	defines where the HEIGHT dimension of the source 
*		 !	object will be mapped to in the destination object.
*		 norder[2] 
*		 !	defines where the DEPTH dimension of the source 
*		 !	object will be mapped to in the destination object.
*		 norder[3] 
*		 !	defines where the TIME dimension of the source 
*		 !	object will be mapped to in the destination object.
*		 norder[4] 
*		 !	defines where the ELEMENTS dimension of the source 
*		 !	object will be mapped to in the destination object.
*
*		 This routine is similar to the lkorient routine,
*		 except that dimensions can not be combined in the
*		 mapping.  It is also significantly faster than the
*		 lkorient routine.
*
*                If the source object has map data, the map data cannot
*                have depth, time, or elements dimensions of greater than
*                one.  The map width and height dimensions can be any size.
*                
*                If the source object has rectilinear or curvilinear 
*                location or time data, lktranspose will fail.  If the 
*                source object has uniform location data, lktranspose will
*                continue, but the location data will not be modified
*                to reflect the reorientation.
*
*                If the object has mask data, it will be transposed along
*                with the value data.
*
*                lktranspose will fail if the source object does not contain
*                value data.
*         Input: src_obj   - the input object to be processed
*                norder - new ordering for the dimensions
*        Output: dst_obj   - the output object
*
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: 
*    Written By: Steve Kubica
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lktranspose (
   kobject src_obj,
   int     *norder,
   kobject dst_obj)
/* -library_def_end */

/* -library_code */
{
   char	 *rtn="lktranspose", *lib="kdatamanip";
   klist *objlist=NULL;
   int	  has_mask=FALSE;
   int    i;
   
   if (!kpds_query_value(src_obj))
   {
      kerror(lib, rtn, "source object must contain value data");
      return(FALSE);
   }
   
   if (kpds_query_map(src_obj))
   {
      int mw, mh, md, mt, me;

      KCALL(kpds_get_attribute(src_obj,KPDS_MAP_SIZE,&mw,&mh,&md,&mt,&me));
      if ((md > 1) || (mt > 1) || (me > 1))
      {
	 kerror(lib, rtn, 
		"Unable operate on object with map depth, time, or elements "
		"dimension greater than 1 (md=%d, mt=%d, me=%d)", md, mt, me);
	 return(FALSE);
      }
   }
   
   /* Object may not have rectilinear or curvilinear location data */
   if (kpds_query_location(src_obj))
   {
      int grid;
      
      KCALL(kpds_get_attribute(src_obj, KPDS_LOCATION_GRID, &grid));
      if (grid != KUNIFORM && (grid == KRECTILINEAR ||
			       grid == KCURVILINEAR))
      {
	 kerror(lib, rtn,
		"Source object cannot contain explicit "
		"rectilinear or curvilinear location data.\n"
		"Use the kaxis program for reorienting data along "
		"the width, height, and depth dimensions");
              return(FALSE);
      }
   }
   
   /* Object may not have time data */
   if (kpds_query_time(src_obj))
   {
      kerror(lib, rtn,
	     "Source object cannot contain explicit time data. ");
      return(FALSE);
   }
   
   if (kpds_query_mask(src_obj))
      has_mask = TRUE;
   
   /* reference the source object */
   KCALL((src_obj = kpds_reference_object(src_obj)) != KOBJECT_INVALID);
   objlist = klist_add(objlist,src_obj,"KOBJECT");

   {
      kaddr oblock = NULL, omblock = NULL;
      kaddr nblock = NULL, nmblock = NULL;
      
      int osize [5], nsize [5];
      int ounit [5], nunit [5];
      int oorder[5] = {KWIDTH, KHEIGHT, KDEPTH, KTIME, KELEMENTS};
      int type, mtype;
      int num, num_regions;

      /*-- determine original size --*/
      KCALL(kpds_get_attributes(src_obj, 
				KPDS_VALUE_SIZE,
				&osize[0], 
				&osize[1], 
				&osize[2], 
				&osize[3], 
				&osize[4],
				KPDS_VALUE_DATA_TYPE, &type, 
				NULL));
      if (has_mask)
	 KCALL(kpds_get_attribute(src_obj, KPDS_MASK_DATA_TYPE, &mtype));
      
      /*-- determine source unit size --*/
      KCALL(kpds_get_attribute(src_obj, 
			       KPDS_VALUE_OPTIMAL_REGION_SIZE, 
			       &ounit[0], 
			       &ounit[1], 
			       &ounit[2], 
			       &ounit[3], 
			       &ounit[4], &num_regions));

      /*-- set up the destination ordered size and unit --*/
      kmemcpy(nsize, osize, sizeof(int) *  5);
      kdas_array_reorder(nsize, 5, norder, oorder);
      kmemcpy(nunit, ounit, sizeof(int) *  5);
      kdas_array_reorder(nunit, 5, norder, oorder);
      
      /*-- set up the output object --*/
      KCALL(kpds_set_attributes(dst_obj,
				KPDS_VALUE_SIZE,
				nsize[0], 
				nsize[1], 
				nsize[2], 
				nsize[3], 
				nsize[4],
				KPDS_VALUE_DATA_TYPE, type, 
				NULL));

      /*-- set attributes transposed unit size --*/
      KCALL(kpds_set_attribute(src_obj, 
			       KPDS_VALUE_REGION_SIZE, 
			       ounit[0], 
			       ounit[1], 
			       ounit[2], 
			       ounit[3], 
			       ounit[4]));
      
      KCALL(kpds_set_attribute(dst_obj, 
			       KPDS_VALUE_REGION_SIZE, 
			       nunit[0], 
			       nunit[1], 
			       nunit[2], 
			       nunit[3], 
			       nunit[4]));

      if (has_mask)
      {
	 KCALL(kpds_set_attributes(dst_obj,
				   KPDS_MASK_SIZE,
				   nsize[0], 
				   nsize[1], 
				   nsize[2], 
				   nsize[3], 
				   nsize[4],
				   KPDS_MASK_DATA_TYPE, mtype, 
				   NULL));

	 KCALL(kpds_set_attribute(src_obj, 
				  KPDS_MASK_SIZE, 
				  ounit[0], 
				  ounit[1], 
				  ounit[2], 
				  ounit[3], 
				  ounit[4]));
      
	 KCALL(kpds_set_attribute(dst_obj, 
				  KPDS_MASK_SIZE, 
				  nunit[0], 
				  nunit[1], 
				  nunit[2], 
				  nunit[3], 
				  nunit[4]));

	 KCALL(kpds_set_attribute(src_obj, 
				  KPDS_MASK_REGION_SIZE, 
				  ounit[0], 
				  ounit[1], 
				  ounit[2], 
				  ounit[3], 
				  ounit[4]));
	 
	 KCALL(kpds_set_attribute(dst_obj, 
				  KPDS_MASK_REGION_SIZE, 
				  nunit[0], 
				  nunit[1], 
				  nunit[2], 
				  nunit[3], 
				  nunit[4]));
      }
      
      /*-- compute the number of elements in the the block --*/
      for (i = 0, num = 1; i < 5; i++)
	 num *= ounit[i];
      
      /*-- allocate data blocks for reordering --*/
      oblock = kmalloc((size_t)(num*kdata_size(type)));
      nblock = kmalloc((size_t)(num*kdata_size(type)));

      if (has_mask)
      {
	 omblock =  kmalloc((size_t)(num*kdata_size(mtype)));
	 nmblock =  kmalloc((size_t)(num*kdata_size(mtype)));
      }
      
      for (i = 0; i < num_regions; i++)
      {
	 /*-- read a unit from the src object --*/
	 oblock = kpds_get_data(src_obj, KPDS_VALUE_REGION, oblock);
	 
	 /*-- reorder the data if necessary and store --*/
	 kpds_put_data(dst_obj, KPDS_VALUE_REGION, 
		       kdas_transpose_data(oblock, 5, type, ounit, nunit, 
				           oorder, norder, nblock));
	 if (has_mask)
	 {
	    omblock = kpds_get_data(src_obj, KPDS_MASK_REGION, omblock);
	 
	    kpds_put_data(dst_obj, KPDS_MASK_REGION, 
			  kdas_transpose_data(omblock, 5, mtype, ounit, nunit, 
					      oorder, norder, nmblock));
	 }
      }
      
      /*-- cleanup --*/
      kfree(oblock);  kfree(nblock);
      kfree(omblock); kfree(nmblock);
   }

   (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
   return(TRUE);
}
/* -library_code_end */
