 /*
  * Khoros: $Id: Avs.c,v 1.8 1997/02/07 21:42:57 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: Avs.c,v 1.8 1997/02/07 21:42:57 mirage Exp $";

#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Khoros AVS Data Service Routines
   >>>>		
   >>>>  Private:
   >>>>                 init_avs()
   >>>>                 init_avs()
   >>>>   Static:
   >>>>                 avs_check()
   >>>>                 avs_input()
   >>>>                 avs_output()
   >>>>                 avs_destroy()
   >>>>                 avs_prep()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


#if !defined(KAVS_DEF)

void
init_avs(void) /*-- NULL avs initialization --*/
{
   return;
}

#else 

#include "kdatafmt/avs.h"

static int avs_check   PROTO((int));
static int avs_input   PROTO((kobject, int, int));
static int avs_output  PROTO((kobject, int, int));
static int avs_destroy PROTO((kobject));
static int avs_prep    PROTO((kobject));

/*=====================================================================*/

/*-----------------------------------------------------------
|  Routine Name: init_avs
|       Purpose: initialize the avs file format
------------------------------------------------------------*/
void
init_avs(void)
{
   kdms_add_file_format("avs", 
			"Advanced Visualization Software Image format (AVS)", 
			avs_check, 
			avs_input, 
			avs_output, 
			avs_destroy, 
			avs_prep,
			NULL, NULL, NULL, NULL, NULL);

   /*-- need some colorspace attributes --*/
   kcolor_init();

   return;
}

/*=====================================================================*/

/*
 *  Internal Resource Structure 
 *  avs  -  Advanced Visualization Software Image format (AVS)
 */
typedef struct
{
   int  fid;
   avs *image;
}
ResourceStruct;

/****    =============================================================
 ****    Format Methods
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: avs_check - Check if a file is an AVS file
|
|       Purpose: This function is used to check the supposed size
|                of the AVS image according to the first few bytes 
|                of the file and compares it to the overall size of
|                the file to see if it really is an AVS file.
|
|         Input: object - the avs object to be checked
|
|        Output: none
|
|       Returns: TRUE if the given file is an AVS file,
|                FALSE otherwise
|
|    Written By: Mark Young
|          Date: July 13, 1992
| Modifications:
|
------------------------------------------------------------*/
static int
avs_check(int fid)
{
   int size;
   int width;
   int height;

   /*---------------------------------------------------------------------
   |  Data is always stored in Sun format so we need to set the
   |  descriptor machine type to be of type KMACH_BIG_ENDIAN_IEEE.  This
   |  will allow us to read the 4 byte width and height fields
   |  properly.
   \---------------------------------------------------------------------*/
   kfile_setmachtype(fid, KMACH_BIG_ENDIAN_IEEE);

   if (kread_int(fid, &width,  (size_t) 1) != 1 || 
       kread_int(fid, &height, (size_t) 1) != 1)
      return FALSE;

   /*-- verify that the image has the right number of bytes that --*/
   size = klseek(fid, (off_t) 0, SEEK_END);

   if ((size - 8) != width * height * 4)
      return FALSE;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: avs_input - Read an AVS image
|
|       Purpose: This function is used to read in the image if the
|                supplied data is an AVS image.  
|
|                This routine will convert from the AVS index order
|                into the more standard index order of width by
|                height by elements.  It will also shift the 
|                data ordering from alphaRGB to RGBalpha.
|                
|         Input: object - the avs object to be initialized
|        Output: none
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Mark Young and Steve Kubica
|          Date: July 13, 1992
| Modifications:
|
------------------------------------------------------------*/
/*ARGSUSED */
static int
avs_input(kobject object, int fid, int flags)
{
   ResourceStruct *resources;
   avs *image;

   int size[KDMS_MAX_DIM];
   int i;
   int total;
   kubyte *r, *g, *b, *a;
   unsigned char *rdata = NULL;

   /*-- get the resource structure --*/ 
   if (!kdms_initialized(object))
   {
      if ((resources = (ResourceStruct *)kcalloc((size_t)1,
					sizeof(ResourceStruct))) == NULL)
         return(FALSE);

      kdms_glue_set_resources(object, resources);
   }
   else
      resources = (ResourceStruct *) kdms_glue_get_resources(object);

   resources->fid = fid;

   if ((image = resources->image = avs_fdread(fid)) == NULL)
   {
      kdas_set_error(KDMS_EFMT_FAILURE);
      return FALSE;
   }

   /*-----------------------------------------------------------------------
   |  value data
   \-----------------------------------------------------------------------*/
   if (image->width != 0 && image->height != 0)
   {
      /*-- create the value segment --*/
      kpds_create_value(object);

      size[0] = image->width;
      size[1] = image->height;
      size[2] = size[3] = 1;
      size[4] = 4;              /* alpha R G B */


      total = image->width * image->height;

      rdata = (unsigned char *)kmalloc(sizeof(unsigned char) * total * 4);

      r = rdata;
      g = r + total;
      b = g + total;
      a = b + total;

      /*----------------------------------------------------------------
      |   This loops mainly exists to reorder the data from aRGB to
      |   RGBa.  The second purpose is to change the index order from
      |   elements first to elements last (might as well do this here
      |   since we are looping through every data point anyway).
      \-----------------------------------------------------------------*/
      for (i = 0; i < total; i++)
      {
         *r++ = image->data[1 + i * 4];
         *g++ = image->data[2 + i * 4];
         *b++ = image->data[3 + i * 4];

         /*-- note AVS uses this for opacity, we use it for alpha --*/
         *a++ = 255 - image->data[i * 4];
      }

      kdms_set_segment(object, KDMS_SEGMENT_VALUE, 5, size, NULL, 
		       KUBYTE, rdata);

      kcolor_set_attribute(object, _INTERNAL_COLORSPACE, KRGB);
      kcolor_set_attribute(object, KCOLOR_HAS_ALPHA, TRUE);
   }

   return TRUE;
}


/*-----------------------------------------------------------
|
|  Routine Name: avs_output - Closes an AVS image
|
|       Purpose: This function is used to close the AVS image.  
|                The data will be reordered by this function
|                from our default index order to the proper 
|                AVS index order.  It will also shift the 
|                data ordering from RGBalpha to alphaRGB.
|
|         Input: object - the avs object to be closed
|
|        Output: Returns: TRUE on success, FALSE otherwise
|
|    Written By: Mark Young and Steve Kubica
|          Date: Jul 13, 1992 15:50
| Modifications:
|
------------------------------------------------------------*/
/*ARGSUSED */
static int
avs_output(kobject object, int fid, int flags)
{
   ResourceStruct *resources;

   avs *image = NULL;
   int type;
   int w, h;
   int i;
   int total;
   kubyte *r, *g, *b, *a;
   kubyte *data;
   kubyte *reordered_data = NULL;

   /*-- get the resource structure --*/ 
   if (!kdms_initialized(object))
   {
      if ((resources = (ResourceStruct *) kcalloc((size_t)1,
					sizeof(ResourceStruct))) == NULL)
         return(FALSE);

      kdms_glue_set_resources(object, resources);
   }
   else
      resources = (ResourceStruct *) kdms_glue_get_resources(object);

   if (resources->image == NULL)
   {
      if ((image = resources->image = avs_create()) == NULL)
      {
         kdas_set_error(KDMS_EFMT_FAILURE);
         return FALSE;
      }
   }
   else
      image = resources->image;
 
   /*
    * Initialize the image's data arrays according to the data found in the
    * data segment.  
    *
    * The avs structure contains only value data.  (image->value)
    */

   /*-----------------------------------------------------------------------
   |  value data -
   \-----------------------------------------------------------------------*/
   if (kpds_query_value(object))
   {
      kpds_get_attribute(object, KPDS_VALUE_DATA_TYPE, &type);

      if (type != KUBYTE)
	 type = KUBYTE;

      kpds_set_attributes(object, 
			  KPDS_COUPLING,        KUNCOUPLED,
			  KPDS_VALUE_DATA_TYPE, type,
			  NULL);

      kpds_get_attribute(object, KPDS_VALUE_SIZE, &w, &h, NULL, NULL, NULL);

      /*-- we're guaranteed to have 4 planes courtesy of the prep routine --*/
      image->width  = w;
      image->height = h;

      /*-- retrieve all value data --*/
      if ((data = (kubyte *)kpds_get_data(object,KPDS_VALUE_ALL,NULL)) == NULL)
      {
         kdas_set_error(KDMS_EFMT_FAILURE);
         return FALSE;
      }

      /*-- precompute array offsets for faster indexing --*/
      total = image->width * image->height;
      r = data;
      g = r + total;
      b = g + total;
      a = b + total;

      /* 
       *  This next loop serves two purposes.  The first purpose is to
       *  reorder the data from [R G B alpha] to [alpha R G B]. 
       *  The second purpose is to change the index order from 
       *  elements last to elements first (might as well do this
       *  here since we are looping through every data point anyway).
       */
      reordered_data = (kubyte *) kmalloc(sizeof(kubyte) * total * 4);

      for (i = 0; i < total; i++)
      {
	 /* note that AVS uses this for opacity, we use it for alpha */
	 reordered_data[i * 4] = 255 - *a++;
	    
	 reordered_data[1 + i * 4] = *r++;
	 reordered_data[2 + i * 4] = *g++;
	 reordered_data[3 + i * 4] = *b++;
      }

      /* free the image and temporary data */
      kfree_and_NULL(data);
      kfree_and_NULL(image->data);

      /* Assign the reordered image data */
      image->data = reordered_data;
   }

   /*-- write the avs structure out to the specified filename --*/
   return avs_fdwrite(fid, image);
}

/*-----------------------------------------------------------
|
|  Routine Name: avs_destroy - Frees an avs image
|
|       Purpose: This function is used to destroy an AVS image. 
|
|         Input: data - initial sample of data
|                num  - number of bytes in the sample
|
|        Output: Returns: TRUE on success, FALSE otherwise
|
|    Written By: Mark Young
|          Date: Jul 13, 1992 15:50
| Modifications:
|
------------------------------------------------------------*/
static int
avs_destroy(kobject object)
{
   ResourceStruct *resources;

   if ((resources = (ResourceStruct *) kdms_glue_get_resources(object))==NULL)
      return TRUE;

   /*-- free the avs structure --*/
   avs_free(resources->image);

   /*-- free the resources --*/
   kfree(resources);

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: avs_prep - prepare an object for output to avs
|
|       Purpose: This function is used to prepare an object for
|		 output to avs.  It will print out information 
|		 about what is not getting stored out to the avs
|		 file due to the severe limitations of the avs format.	
|
|         Input: object - the object to be prepared
|        Output: none
|	Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 18, 1996
|
------------------------------------------------------------*/
/*ARGSUSED2*/
static int
avs_prep(kobject object)
{
   char *warn = NULL;
   int   d, t, e;
   int   type;
   int   had_map = FALSE;
   

   /*-----------------------------------------------------------------------
   |  mask, location, and time data
   \-----------------------------------------------------------------------*/
   if (kpds_query_mask(object))
      warn = kstring_append(warn, "- no support for storing mask "
			    ": mask segment ignored.\n");

   if (kpds_query_location(object))
      warn = kstring_append(warn, "- no support for storing location "
			    ": location segment ignored.\n");

   if (kpds_query_time(object))
      warn = kstring_append(warn, "- no support for storing time "
			    ": time segment ignored.\n");

   if (kpds_query_map(object))
   {
      had_map = TRUE;
      
      kpds_set_attribute(object, KPDS_MAPPING_MODE, KMAPPED);
      
      warn = kstring_append(warn, "- no support for map segment "
			    ": mapping data through the map.\n");
   }
   
   /*-- support for checking extraneous attributes coming soon! --*/

   /*-----------------------------------------------------------------------
   |  value data
   \-----------------------------------------------------------------------*/
   if (kpds_query_value(object))
   {
      /*-- this gets hairy : uncouple so we don't thrash the disk --*/
      kpds_set_attribute(object, KPDS_COUPLING, KUNCOUPLED);


      /*-- size issue : can only store 2D data plane --*/
      kpds_get_attribute(object, KPDS_VALUE_SIZE, NULL, NULL, &d, &t, &e);

      if (d > 1 || t > 1)
      {
	 kpds_set_attribute(object, KPDS_VALUE_SIZE, -1, -1, 1, 1, -1);
	 warn = kstring_append(warn, "- ignoring depth and time planes of "
			       "value segment.\n");
      }

      /*-- special case : if we have only one plane of color... --*/
      if (e == 1)
      {
	 if (had_map)
	    warn = kstring_append(warn, "- only one plane of value "
				  "information : image may appear strange.\n");

	 else /*-- ...create a greyscale map --*/
	 {
	    kcolor_set_attribute(object, KCOLOR_MAP_AUTOCOLOR, KGREYSCALE);
	    kpds_set_attribute(object, KPDS_MAPPING_MODE, KMAPPED);
	 
	    warn = kstring_append(warn, "- only one plane of value "
				  "information : creating greyscale "
				  "colormap\n");
	    warn = kstring_append(warn, "  check initial image value range if "
				  "results are poor...\n");
	 }
      }
      
      /*-- check our element size again --*/
      kpds_get_attribute(object, KPDS_VALUE_SIZE, NULL, NULL, NULL, NULL, &e);

      /*-- pad out so it looks like we have four planes --*/ 
      kpds_set_attributes(object, 
			  KPDS_VALUE_INTERPOLATE, KPAD,
			  KPDS_VALUE_PAD_VALUE,   255.0, 0.0,
			  KPDS_VALUE_SIZE,        -1, -1, 1, 1, 4,
			  NULL);
      

      /*-- type issue : can only store bit & unsigned byte images --*/
      kpds_get_attribute(object, KPDS_VALUE_DATA_TYPE, &type);

      if (type != KUBYTE)
	 warn = kstring_append(warn, "- casting value data to unsigned byte "
			       "for storage.\n");
   }

   /*-- if we have any news to report, report it now --*/
   if (warn)
   {
      kinfo(KSTANDARD, "AVS format output :\n%s", warn);
      kfree_and_NULL(warn);
   }

   return TRUE;
}

#endif /* KAVS_DEF */
/* don`t add after the endif */
