 /*
  * Khoros: $Id: copy.c,v 1.12.2.2 1997/03/11 17:15:19 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: copy.c,v 1.12.2.2 1997/03/11 17:15:19 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Copying Utilites
   >>>>		
   >>>>   Public:
   >>>>                 kaps_copy_through_pres()
   >>>>                 kaps_copy_beneath_pres()
   >>>>
   >>>>                 kaps_copy_object()
   >>>>                 kaps_copy_global_attributes()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*-----------------------------------------------------------
|
|  Routine Name: kaps_copy_through_pres
|
|       Purpose: Copy a segment from one object to another 
|		 through the presentations of both objects.
|
|		 The datatype and size of the destination
|		 copy is overwritten with the datatype and 
|		 size of the source.  The index order of 
|		 the destination is set to the polymorphic 
|		 default index order for that segment.
|
|		 During the copy data will be read through
|		 the presentation of the first object and 
|		 put through the presentation of the second
|		 object.   
| 
|		 Some presentation attributes are considered
|		 only from the source. The position and offset 
|		 attributes are considered for the get data,
|		 but not considered for the put.  By setting
|		 the position attribute on the source before
|		 the copy, the copied data is shifted by the
|		 position.
|
|		 Mapping mode is also considered only for the
|		 source. If the source object is mapped, 
|		 the destination object will be considered 
|                unmapped.
|
|		 If the segment does not yet exist in destination
|		 object, it will be created there.
|
|		 Non-transferable segments can not be copied with
|		 this function.
|
|         Input: src 	 - the source object to copy from
|		 dst 	 - the destination object to copy to
|	 	 segment - the name of the segmen to copy
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: April 25, 1996
|
------------------------------------------------------------*/
int
kaps_copy_through_pres(
   kobject  src,
   kobject  dst,
   char    *segment)
{
   kobject rsrc, rdst;
   kaddr   data = NULL;
   ktoken  seg = kstring_to_token(segment);

   int  unit[5];
   int  wbeg[5];
   int  wend[5];
   int  beg [5];
   int  pof [5];
   int *pos, *off, *size, *order;
   int  dim, datatype, i;
   int 	vorder[5] = {KWIDTH, KHEIGHT, KDEPTH, KTIME, KELEMENTS};
   int 	lorder[4] = {KWIDTH, KHEIGHT, KDEPTH, KDIMENSION};
   int 	torder[1] = {KTIME};
   int 	morder[5] = {KMAP_WIDTH, KMAP_HEIGHT, KMAP_DEPTH, KMAP_TIME, 
		     KMAP_ELEMENTS};


   if (src == NULL || dst == NULL)
      return FALSE;   /*-- set invalid object errorno here --*/

   if (!kdms_query_segment(src, segment))
      kdas_fail_error(KDMS_ESEG_NONEXIST);

   /*-- get physical characteristics --*/
   dim  = kpds_get_dimension(src, seg);
   kpds_get_size(src, seg, &size);
   kdms_get_attribute(src, segment, KDMS_DATA_TYPE, &datatype);

   /*-- deal with all segment specific information --*/ 
   if (kstrcmp(segment, KPDS_SEGMENT_LOCATION) == 0)
      order = lorder;
   else if (kstrcmp(segment, KPDS_SEGMENT_MAP) == 0)
      order = morder;
   else if (kstrcmp(segment, KPDS_SEGMENT_TIME) == 0)
      order = torder;
   else  if (kstrcmp(segment, KPDS_SEGMENT_MASK) == 0)
      order = vorder;
   else  if (kstrcmp(segment, KPDS_SEGMENT_VALUE) == 0)
   {
      order = vorder;

      /*-- value datatype changes with mapping mode --*/
      kpds_get_attribute(src, KPDS_VALUE_DATA_TYPE, &datatype);
   }
   else /*-- something else --*/
      order = vorder;


   /*-- make sure the segment exists in the destination --*/
   if (!kdms_query_segment(dst, segment))
      if (!kdms_create_segment(dst, segment))
	 return FALSE;

   /*-- the destination should not be mapped --*/
   if (!kpds_set_attribute(dst, KPDS_MAPPING_MODE, KUNMAPPED))
      return FALSE;

   /*-- dest will have the default order, copy src size to physical as well--*/
   kdms_set_attributes(dst, segment,
		       KDMS_INDEX_ORDER, order,
		       KDMS_SIZE,        size,
		       KDMS_DATA_TYPE,   datatype,
		       NULL);

   /*-- get position and offset from source before referencing --*/
   kpds_get_access_info(src, seg, NULL, &pos, &off, NULL, NULL);

   /*-- create a reference to avoid side effects on the source --*/
   rsrc = kdms_reference(src);
   rdst = kdms_reference(dst);

   /*-- find optimal region size --*/ 
   kdas_optimal_region_size(dim, size, datatype, unit);
   
   /*-- determine initial position --*/
   for (i = 0; i < dim; i++)
   {
      beg[i] = 0;
      pof[i] = pos[i] + off[i];
   }

   /*-- copy the data from the source to the destination --*/
   do
   {
      /*-- shift get by position --*/
      for (i = 0; i < dim; i++)
	 wbeg[i] = beg[i] + pof[i], wend[i] = wbeg[i] + unit[i] - 1;

      data = kdms_get_data(rsrc, segment, wbeg, wend, data);

      /*-- don't shift put by position --*/
      for (i = 0; i < dim; i++)
	 wbeg[i] = beg[i], wend[i] = wbeg[i] + unit[i] - 1;

      kdms_put_data(rdst, segment, wbeg, wend, data);
   }
   while (kdas_advance_position(dim, size, unit, beg));


   /*-- close the references --*/
   kdms_close(rsrc);
   kdms_close(rdst);

   kfree_and_NULL(data);

   return TRUE;
}


/*-----------------------------------------------------------
|
|  Routine Name: kaps_copy_beneath_pres
|
|       Purpose: This routine will copy all the data in a specified
|		 segment from one object to another.  This copy
|		 copies "underneath" the presentation functionality
|		 of data services.  No presentation considerations are 
|		 taken into account with this copy. This copy duplicates 
|		 the segment exactly.  
|
|		 The attribute dimension, size, dataytype, and 
|		 index order will be copied from the source to 
|		 the destination since they describe the physical 
|		 state of the data.
|
|		 The copy is done by creating references of both
|		 the source and destination, reseting the presentation
|		 attributes to match the physical, and zeroing out
|		 all presentation attributes which affect the data
|		 pipeline, such as size and datatype.
|
|		 If the segment does not yet exist in destination
|		 object, it will be created there.
|
|		 Non-transferable segments can not be copied with
|		 this function.
|
|         Input: src 	 - the source object to copy from
|		 dst 	 - the destination object to copy to
|	 	 segment - the name of the segmen to copy
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: December 26, 1996
|
------------------------------------------------------------*/
int
kaps_copy_beneath_pres(
   kobject object1,
   kobject object2,
   char   *segment)
{
   kobject src;
   kobject dest;

   int   begin[KDMS_MAX_DIM] = {0, 0, 0, 0, 0};
   int   end[KDMS_MAX_DIM]   = {0, 0, 0, 0, 0};
   int  *unit;
   int  *size;
   int   dim;
   int   i;
   int   transferable;
   kaddr data = NULL;

   if (object1 == NULL || object2 == NULL)
      return FALSE;   /*-- set invalid object errorno here --*/
   
   /*-- make sure the segment exists in the source -- */
   if (!kdms_query_segment(object1, segment))
      kdas_fail_error(KDMS_ESEG_NONEXIST);

   /*-- if the segment is not transferable, then bail out -- */
   if (!kdms_get_attribute(object1, segment, KDMS_TRANSFERABLE, &transferable))
      return FALSE;

   if (!transferable)
      return TRUE;
   
   /*-- make sure the segment exists in the destination -- */
   if (!kdms_query_segment(object2, segment))
      if (!kdms_create_segment(object2, segment))
	 return FALSE;


   /*-- create reference we can muck with the presentation at will -- */
   src  = kdms_reference(object1);
   dest = kdms_reference(object2);

   /*-- sync source presentation to reflect physical source -- */
   if (!kdms_set_attribute(src, segment, KDMS_COUPLING, KDEMAND) &&
       !kdms_sync(src, segment, KPHYS2PRES))
      return FALSE;

   /*-- want to affect physical layer on the destination -- */
   if (!kdms_set_attribute(dest, segment, KDMS_COUPLING, KCOUPLED))
      return FALSE;

   /*-- migrate over the dimension, size, data type, and index order -- */
   kdms_copy_attributes(src, dest, segment,
			KDMS_DIMENSION, 
			KDMS_SIZE, 
			KDMS_DATA_TYPE,
			KDMS_INDEX_ORDER, 
                        NULL);

   /*-- none out all other presentation attribute before the copy -- */
   if (!kdms_set_attribute(src, segment, KDMS_SCALING, KNONE))
      return FALSE;
   
   if (!kdms_set_attribute(dest, segment, KDMS_SCALING, KNONE))
      return FALSE;


   /*-- determine the optimal region size with which to copy -- */
   if (!kdms_get_attributes(src, segment, 
			    KDMS_DIMENSION,           &dim,
			    KDMS_SIZE,                &size,
			    KDMS_OPTIMAL_REGION_SIZE, &unit, NULL))
      return FALSE;

   /*-- perform the copy using iterative gets-n-puts --*/
   do
   {
      for (i = 0; i < dim; i++)
	 end[i] = unit[i] - 1;

      data = kdms_get_data(src, segment, begin, end, data);
      kdms_put_data(dest, segment, begin, end, data);
   }
   while (kdas_advance_position(dim, size, unit, begin));

   kfree_and_NULL(data);

   /*-- close down the references -- */
   kdms_close(src);
   kdms_close(dest);

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kaps_copy_object
|
|       Purpose: This routine will copy all of the 
|		 segments from one object to another.
|
|		 
|
|         Input: src_obj         - the object to copy from
|		 dest_obj        - the object to copy to
|		 copy_attributes - copy the segment attributes 
|		                   and object level attributes
|		 copy_data       - copy the segment data
|
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: Dec 01, 1993 09:00
| Modifications:
|
------------------------------------------------------------*/
int
kaps_copy_object(
   kobject src_obj,
   kobject dest_obj,
   int copy_attributes,
   int copy_data)
{
   char **segments = NULL;
   int i;
   int nsegments = 0;


   segments = kdms_get_segment_names(src_obj, &nsegments);

   for (i = 0; i < nsegments; i++)
   {

      /*  if the segment appears to exists in the src, then copy it */
      if (kdms_query_segment(src_obj, segments[i]))
      {
	 if (kstrcmp(segments[i], KPDS_SEGMENT_VALUE)==0)
	 {
	    if (copy_attributes)
	       if (!kpds_copy_value_attr(src_obj, dest_obj))
		  return FALSE;
	    
	    if (copy_data)
	       if (!kpds_copy_value_data(src_obj, dest_obj, TRUE))
		  return FALSE;
	 }
	 else if (kstrcmp(segments[i], KPDS_SEGMENT_MASK)==0)
	 {
	    if (copy_attributes)
	       if (!kpds_copy_mask_attr(src_obj, dest_obj))
		  return FALSE;
	    
	    if (copy_data)
	       if (!kpds_copy_mask_data(src_obj, dest_obj, TRUE))
		  return FALSE;
	 }
	 else if (kstrcmp(segments[i], KPDS_SEGMENT_MAP)==0)
	 {
	    if (!kpds_is_mapped(src_obj))
	    {
	       if (copy_attributes)
		  if (!kpds_copy_map_attr(src_obj, dest_obj))
		     return FALSE;
	    
	       if (copy_data)
		  if (!kpds_copy_map_data(src_obj, dest_obj, TRUE))
		     return FALSE;
	    }
	 }
	 else if (kstrcmp(segments[i], KPDS_SEGMENT_LOCATION)== 0 ||
		  kstrcmp(segments[i], KPDS_SEGMENT_UNIFORM) == 0 ||
		  kstrcmp(segments[i], KPDS_SEGMENT_WIDTH)   == 0 ||
		  kstrcmp(segments[i], KPDS_SEGMENT_HEIGHT)  == 0 ||
		  kstrcmp(segments[i], KPDS_SEGMENT_DEPTH)   == 0)
	 {
	    if (copy_attributes)
	       if (!kpds_copy_location_attr(src_obj, dest_obj))
		  return FALSE;
	    
	    if (copy_data)
	       if (!kpds_copy_location_data(src_obj, dest_obj, TRUE))
		  return FALSE;
	 }
	 else if (kstrcmp(segments[i], KPDS_SEGMENT_TIME)==0)
	 {
	    if (copy_attributes)
	       if (!kpds_copy_time_attr(src_obj, dest_obj))
		  return FALSE;
	    
	    if (copy_data)
	       if (!kpds_copy_time_data(src_obj, dest_obj, TRUE))
		  return FALSE;
	 }
	 else
	 {
	    if (copy_attributes)
	       if (!kdms_copy_segment_attributes(src_obj,  segments[i],
						 dest_obj, segments[i]))
		  return FALSE;

	    if (copy_data)
	       if (!kaps_copy_beneath_pres(src_obj, dest_obj, segments[i]))
		  return FALSE;
	 }
      }
   }

   /*-- tidy up segment names --*/
   karray_free(segments, KSTRING, (size_t) nsegments, NULL);

   /* copy the object level attributes if desired */
   if (copy_attributes)
      if (!kdms_copy_segment_attributes(src_obj, NULL, dest_obj, NULL))
	 return FALSE;



   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kaps_copy_global_attributes
|
|       Purpose: This routine will copy all the global
|		 object-level attributes from one object 
|		 to another.
|
|         Input: src_obj         - the object to copy from
|		 dest_obj        - the object to copy to
|
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Aug 19, 1994 10:26
| Modifications:
|
------------------------------------------------------------*/
int
kaps_copy_global_attributes(
   kobject src_obj,
   kobject dest_obj)
{
   if (!kdms_copy_segment_attributes(src_obj, NULL, dest_obj, NULL))
      return FALSE;

   return TRUE;
}
