/*
 * Khoros: $Id: utilities.c,v 1.11.2.2 1997/03/11 17:15:49 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: utilities.c,v 1.11.2.2 1997/03/11 17:15:49 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Library Routine for #object-name#
   >>>>		
   >>>>  Private:
   >>>>                 kapu_multi_input_loc()
   >>>>                 kapu_multi_input_time()
   >>>>   Public:
   >>>>                 kcall_analyze()
   >>>>                 lkcall_free()
   >>>>                 kapu_mask_ops()
   >>>>                 kapu_upgrade_objects()
   >>>>                 kapu_scale_data()
   >>>>                 kapu_minmax()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


/****************************************************************
*
*  Routine Name: kcall_analyze() - analyze the source code text and
*                print appropriate error messages.
*
*       Purpose: analyze the source code text and print appropriate
*                error messages (i.e. related to the text). This code
*                is used internally by the KCALL macro, but may find
*                other uses as well. Note that it is assumed that an
*                error has occured while executing the source code.
*
*         Input: char *lib - library name
*                char *rtn - routine name
*                char *code - code text
*
*        Output: None.
*
*       Returns: TRUE if successful, FALSE if a failure occurs.
*
*  Restrictions: Restrictions on data or input as applicable
*    Written By: Scott Wilson
*          Date: Apr 29, 1996
*      Verified:
*  Side Effects:
* Modifications:
****************************************************************/
void kcall_analyze(char *lib, char *rtn, char *code)
  {
    if (kstrstr(code,"_set_attribute") != NULL)
      kerror(lib,rtn,"Unable to set object attributes");
    else if (kstrstr(code,"_get_attribute") != NULL)
      kerror(lib,rtn,"Unable to get object attributes");
    else if (kstrstr(code,"_get_data") != NULL)
      kerror(lib,rtn,"Unable to get data from object");
    else if (kstrstr(code,"_put_data") != NULL)
      kerror(lib,rtn,"Unable to put data to object");
    else if (kstrstr(code,"_reference_object") != NULL)
      kerror(lib,rtn,"Unable to reference_object");
    else if (kstrstr(code,"_create_") != NULL)
      kerror(lib,rtn,"Unable to create segment or attribute");
    else if (kstrstr(code,"_destroy_") != NULL)
      kerror(lib,rtn,"Unable to destroy segment or attribute");
    else if (kstrstr(code,"_open_") != NULL)
      kerror(lib,rtn,"Unable to open object");
    else if (kstrstr(code,"_close_") != NULL)
      kerror(lib,rtn,"Unable to close object");
    else if (kstrstr(code,"_copy_") != NULL)
      kerror(lib,rtn,"Unable to copy segment, attribute, data, or object");
    else
      kerror(lib,rtn,"Got an error from subroutine call:\n  %s\n",code);
    kinfo(KVERBOSE,"%s:%s: Offending source code:\n  %s\n",lib,rtn,code);
}

/****************************************************************
*
*  Routine Name: lkcall_free() - go through a list of objects
*                and free them using the necesssary routines,
*                typically kfree_and_NULL() or kpds_close_object().
*
*       Purpose: Given an list of type *klist, step through
*                each entry and free memory resources
*                according the the following scheme: 1) if the
*                client_data for the list entry is "KMALLOC",
*                then call kfree_and_NULL() on the list identifier. If the client_data
*                for the list object is "KOBJECT", then call kpds_close_object()
*                on the list identifier.
*
*                By immediately adding items that allocate dynamic memory
*                resources to the klist, a single call to lkcall_free()
*                will return all of those resources in a systematic
*                and reliable manner. This routines is called, for example,
*                by the KCALL() macro, which relieves the programmer of the
*                need to check the return value of each kpds, kdms, or other
*                Data Services call and to clean up if a failure occurs.
*
*		 Items can be added to or removed from the object list by
*		 using the routines klist_add() and klist_delete() from the
*		 kutils(3) library.
*
*         Input: klist *objlist - klist of things to be free'd
*
*        Output: None.
*
*       Returns: TRUE if successful, FALSE if a failure occurs.
*
*  Restrictions: Restrictions on data or input as applicable
*    Written By: Scott Wilson
*          Date: Nov 15, 1994
*      Verified:
*  Side Effects:
* Modifications:
****************************************************************/
int lkcall_free(klist *p)
  {
    char *lib="kapputils";
    char *rtn="lkcall_free";
 
    if (kstrcmp((char *) p->client_data,"KOBJECT")==0)
      {
        if (!kpds_close_object((kobject)(p->identifier)))
          kerror(lib,rtn,"Unable to close object");
      }
    else if (kstrcmp((char *) p->client_data,"KMALLOC")==0)
      kfree_and_NULL(p->identifier);
    else
        (void)kinfo(KSYSLIB,"lkcall_free: Unrecognized item in list: %s\n",
                             p->client_data);
    return(TRUE);
  }


/************************************************************
*
*  Routine Name: kapu_mask_ops - perform AND, OR, or XOR on data object masks
*
*       Purpose: This function combines the masks of the source
*		 objects according to the operation flag supplied.
*		 Valid operations are:
*		 !  KAPU_MASK_AND
*		 !  KAPU_MASK_OR
*		 !  KAPU_MASK_XOR
*
*		 If no source object has a mask, then kapu_mask_ops
*		 returns without modifying dest_obj.
*
*		 If only one source object has a mask, then its mask
*		 is copied to the destination object.
*
*		 If more than one source object has a mask, then the
*		 specified operation is performed on the source mask
*		 data, and stored in dest_obj.  dest_obj mask data,
*		 if it exists, is not included in the operation.
*
*		 Invalid objects (KOBJECT_INVALID) may be passed in.
*
*         Input: operation - defines operation to perform
*                num_src_objs - number of source objects provided in kvalist
*                kvalist - list of source objects (kobject)
*
*        Output: dest_obj - destination object where resulting mask is stored
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Donna Koechner
*          Date: Jun  8, 1994
*      Verified:
*  Side Effects: Data may be written to mask of dest_object (mask may 
*		 be created if it does not already exist).
* Modifications:
*
*************************************************************/

int kapu_mask_ops(
   kobject  dest_obj,
   int	    operation,
   int	    num_src_objs,
   kvalist)
{
        kva_list        valist;
        kobject         *objlist;
	int		*mask_exists; 
	int		i, j, k, mask_count=0, error=FALSE;
	int		dw, dh, dd, dt, de;
	int		rgn_w, rgn_h, rgn_d, rgn_t, rgn_e, num_rgns, num_pts;
	unsigned char	*data=NULL, *result=NULL;
	char		*prog="kapu_mask_ops";
	char		*lib="kdatamanip";

	/*
	 * Destination object must be valid.
	 */
	if (dest_obj == KOBJECT_INVALID)
	{
	   kerror(lib,prog,"destination object must be a valid object");
	   return(FALSE);
	}

        if ((objlist = (kobject *) kmalloc((size_t)num_src_objs * 
					   sizeof(kobject))) 
		== NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for objlist");
	   return(FALSE);
	}

        if ((mask_exists = 
		(int *) kmalloc((size_t) num_src_objs * sizeof(int))) == NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for mask_exists array");
	   kfree_and_NULL(objlist);
	   return(FALSE);
	}

        kva_start(valist, num_src_objs);

        for (i=0; i<num_src_objs; i++)
	{
	   objlist[i] = kva_arg(valist, kobject);
	   if (objlist[i] == KOBJECT_INVALID)
	      mask_exists[i] = FALSE;
	   else 
	   {
	      if ((mask_exists[i] = kpds_query_mask(objlist[i])) == TRUE)
	         mask_count ++;
	   }
	}

        kva_end(valist);

	/* 
	 *  If none of the objects contain a mask, do nothing and return.
	 */
	if (mask_count == 0)
	{
	   kfree_and_NULL(objlist);
	   kfree_and_NULL(mask_exists);
	   return(TRUE);
	}
		
	/* 
	 *  Get and set attributes of destination object.
	 */
	if (!kpds_query_mask(dest_obj))
	{
	   if (!kpds_create_mask(dest_obj))
	   {
	      kerror(lib,prog,"Unable to create destination object mask.");
	      kfree_and_NULL(objlist);
	      kfree_and_NULL(mask_exists);
	      return(FALSE);
	   }
	}
	if (!kpds_get_attribute(dest_obj, KPDS_MASK_SIZE, &dw,&dh,&dd,&dt,&de))
	{
	   kerror(lib,prog,"Unable to get destination object attributes.");
           kfree_and_NULL(objlist);
           kfree_and_NULL(mask_exists);
           return(FALSE);
	}

	/* 
	 *  If only one object contains a mask, simply copy it to the 
	 *  destination object and return.
	 */
	if (mask_count == 1)
	{
	   for (i=0; i<num_src_objs; i++)
	   {
	      if (mask_exists[i])
	      {
		 if ((objlist[i] = kpds_reference_object(objlist[i])) 
			== KOBJECT_INVALID)
		 {
	   	    kerror(lib,prog,"Unable to reference object %d", i);
	   	    error = TRUE;
		 }
	         if (!kpds_set_attribute(objlist[i], KPDS_MASK_SIZE, 
				dw,dh,dd,dt,de))
	         {
	            kerror(lib,prog,"Unable to set mask size of object %d.",i);
                    kfree_and_NULL(objlist);
                    kfree_and_NULL(mask_exists);
                    return(FALSE);
	         }
	         if (!kpds_copy_mask((objlist[i]), dest_obj, TRUE))
		    error = TRUE;

		 kpds_close_object(objlist[i]);
	         kfree_and_NULL(objlist);
	         kfree_and_NULL(mask_exists);
		 if (error)
		    return(FALSE);
		 else
	            return(TRUE);
	      }
	   }
	}

	if ((dest_obj = kpds_reference_object(dest_obj)) == KOBJECT_INVALID)
	{
	   kerror(lib,prog,"Unable to reference destination object");
	   error = TRUE;
	}

	if (!kpds_set_attribute(dest_obj, KPDS_MASK_DATA_TYPE, KUBYTE))
	{
	   kerror(lib,prog,"Unable to set destination object mask data type.");
	   error = TRUE;
	}
	if (!kpds_get_attributes(dest_obj, 
			KPDS_MASK_OPTIMAL_REGION_SIZE, 
			&rgn_w, &rgn_h, &rgn_d, &rgn_t, &rgn_e, &num_rgns, 
			NULL))
	{
	   kerror(lib,prog,"Unable to get destination object attributes.");
	   error = TRUE;
	}
	if (!kpds_set_attributes(dest_obj, 
		KPDS_MASK_REGION_SIZE, rgn_w, rgn_h, rgn_d, rgn_t, rgn_e, 
		KPDS_MASK_POSITION, 0,0,0,0,0,
		NULL))
	{
	   kerror(lib,prog,"Unable to set destination object mask attributes.");
	   error = TRUE;
	}

	if (error)
	{
	   kpds_close_object(dest_obj);
	   kfree_and_NULL(objlist);
	   kfree_and_NULL(mask_exists);
	   return(FALSE);
	}

	/* 
	 *  If more than one mask exists, we will need set attributes and
	 *  get and put data.  Therefore, reference all objects that contain 
	 *  masks so that their attributes are not permanently affected.  
	 */
        for (i=0; i<num_src_objs; i++)
	{
	   if (mask_exists[i])
	   {
	      if ((objlist[i] = kpds_reference_object(objlist[i])) 
			== KOBJECT_INVALID)
	      {
	         kerror(lib,prog,"Unable to reference object %d.",i);
		 error = TRUE;
	      }
	      if (!kpds_set_attributes(objlist[i], 
			KPDS_MASK_SIZE, dw, dh, dd, dt, de,
			KPDS_MASK_REGION_SIZE, rgn_w,rgn_h,rgn_d,rgn_t,rgn_e, 
			KPDS_MASK_POSITION, 0,0,0,0,0,
			KPDS_MASK_DATA_TYPE, KUBYTE,
			NULL))
	      {
	         kerror(lib,prog,"Unable to set attributes for object %d.",i);
		 error = TRUE;
	      }

	      if (error)
	      {
	         kpds_close_object(dest_obj);
	         for (j=0; j<i; j++)
		 {
		    if (mask_exists[j])
		       kpds_close_object(objlist[j]);
	         }
		 kfree_and_NULL(mask_exists);
		 kfree_and_NULL(objlist);
	         return(FALSE);
	      }
	   }
	}
	
	num_pts = rgn_w * rgn_h * rgn_d * rgn_t * rgn_e;

	/*
	 *  Initialize a temporary results array, initial values are
	 *  based on operation being performed (AND - initialize to 1, 
	 *  OR - initialize to 0, XOR - initialize to 0).
	 */

        if ((result=(unsigned char *)kmalloc((size_t) num_pts * 
					     sizeof(unsigned char))) 
		== NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for result array.");
	   kpds_close_object(dest_obj);
	   for (j=0; j<i; j++)
	   {
	      if (mask_exists[j])
	         kpds_close_object(objlist[j]);
	   }
	   kfree_and_NULL(mask_exists);
	   kfree_and_NULL(objlist);
	   return(FALSE);
	}
	if (operation == KAPU_MASK_AND)
	{
	   for (i=0; i<num_pts; i++)
	      result[i] = 1;
	}
	else
	{
	   for (i=0; i<num_pts; i++)
	      result[i] = 0;
	}
	
	for (i=0; i<num_rgns; i++)
	{
	   for (j=0; j<num_src_objs; j++)
	   {
	      if (mask_exists[j])
	      {
	         if ((data = (unsigned char *) kpds_get_data(objlist[j],
				KPDS_MASK_REGION, data)) == NULL)
		 {
	            kerror(lib,prog,"Unable to get region from object %d.",j);
		    error = TRUE;
		    break;
	 	 }
	         if (error)  break;
		 if (operation == KAPU_MASK_AND)
		 {
		    for (k=0; k<num_pts; k++)
		       result[k] = result[k] & data[k]; 
		 }
		 else if (operation == KAPU_MASK_OR)
		 {
		    for (k=0; k<num_pts; k++)
		       result[k] = result[k] | data[k]; 
		 }
		 else if (operation == KAPU_MASK_XOR)
		 {
		    for (k=0; k<num_pts; k++)
		       result[k] = result[k] ^ data[k]; 
		 }
		 else 
		 {
	            kerror(lib,prog,"Unrecognized operation (%d).",operation);
		    error = TRUE;
		    break;
		 }
	      }
	      if (error)  break;
	   }

	   if (!kpds_put_data(dest_obj, KPDS_MASK_REGION, result))
	   {
	      kerror(lib,prog,"Unable to get region from object %d.",j);
	      error = TRUE;
	      break;
	   }
	   if (error)  break;
	}
	
	/* clean up */
	kpds_close_object(dest_obj);
	for (j=0; j<i; j++)
	{
	   if (mask_exists[j])
	      kpds_close_object(objlist[j]);
	}
	kfree_and_NULL(objlist);
	kfree_and_NULL(mask_exists);
	kfree_and_NULL(result);
	kfree_and_NULL(data);
	if (error)
	   return(FALSE);
	else
	   return(TRUE);
	
}


/*-----------------------------------------------------------
|
|  Routine Name: kapu_multi_input_loc - check object list for location data
|	
|       Purpose: This function checks whether location data exists in
|		 the source objects.  Currently, the function returns 
|		 TRUE if the first object has location, and FALSE if it 
|		 does not.
|
|		 The grid_type parameter is set to the grid type of the 
|		 first input object if it has location.
|
|		 The function will print a warning (KSYSLIB) if any
|		 of the other source objects have location, and it is
|		 being lost.
|
|		 kapu_multi_input_loc is under construction.  In the future, 
|		 as we start supporting location more fully, it will do more 
|		 than just check the first object and set the grid type to 
|		 that object's type.  For example, 
|		 the grid_type parameter will be set to the lowest common
|		 grid type of the input objects, there will be the option
|		 to set up location on all of the objects so that they are
|		 presented the same, and more information on which source 
|		 objects have location will be returned.  This will happen
|		 later. 
|
|         Input: num_src_objs - number of source objects provided in kvalist
|                kvalist - list of source objects (kobject)
|
|        Output: grid_type - processing grid type: set to KNONE, KUNIFORM, 
|			     KRECTILINEAR, or KCURVILINEAR
|
|       Returns: TRUE (1) if location exists in first source, FALSE (0) 
|		 otherwise
|
|  Restrictions: 
|    Written By: Donna Koechner
|          Date: Jan  3, 1995
|      Verified:
|  Side Effects: None
| Modifications:
|
------------------------------------------------------------*/

int kapu_multi_input_loc(
   int	    *grid_type,
   int	    num_src_objs,
   kvalist)
{
        kva_list        valist;
        kobject         *objlist;
	int		*grid; 
	int		*loc_exists; 
	int		i, loc_count=0;
	char		*prog="kapu_multi_input_loc";
	char		*lib="kdatamanip";

        if ((objlist = (kobject *) kmalloc((size_t) num_src_objs * 
					   sizeof(kobject))) 
		== NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for objlist");
	   return(FALSE);
	}

        if ((grid = (int *) kmalloc((size_t) num_src_objs * 
				    sizeof(int))) == NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for grid array");
	   kfree_and_NULL(objlist);
	   return(FALSE);
	}

        if ((loc_exists = (int *) kmalloc((size_t) num_src_objs * 
					  sizeof(int))) == NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for loc_exists array");
	   kfree_and_NULL(objlist);
	   kfree_and_NULL(grid);
	   return(FALSE);
	}

        kva_start(valist, num_src_objs);

        for (i=0; i<num_src_objs; i++)
	{
	   objlist[i] = kva_arg(valist, kobject);
	   grid[i] = KNONE;
	   if (objlist[i] != KOBJECT_INVALID)
	   {
	      if ((loc_exists[i] = kpds_query_location(objlist[i])) == TRUE)
	      {
	         loc_count ++;
		 if (!kpds_get_attribute(objlist[i], 
				KPDS_LOCATION_GRID, &grid[i]))
		 {
                    kerror(lib,prog,
			   "Unable to get location grid type for object %d",i);
                    kfree_and_NULL(objlist); kfree_and_NULL(grid); kfree_and_NULL(loc_exists);
                    return(FALSE);
		 }
	      }
	   }
	}
        kva_end(valist);

        kfree_and_NULL(objlist);

        *grid_type = KNONE;

	if (loc_count > 0)
	{
	   if (loc_exists[0] && (loc_count == 1))
	      *grid_type = grid[0];
	   else
	   {
	      kinfo(KSYSLIB, "Only Location data from the first object will "
			      "be propagated. \nLocation that exists in other "
			      "source objects will be lost. \n"
		     	      "(%d out of %d have location)", 
				loc_count, num_src_objs);
	      *grid_type = grid[0];
	   }
	}
	else
	{
           kfree_and_NULL(loc_exists);
           kfree_and_NULL(grid);
	   return(FALSE);
	}
	
        kfree_and_NULL(loc_exists);
        kfree_and_NULL(grid);
	return(TRUE);
}


/*-----------------------------------------------------------
|
|  Routine Name: kapu_multi_input_time - check object list for time data
|	
|       Purpose: This function checks whether time data exists in
|		 the source objects.  Currently, the function returns 
|		 TRUE if the first object has time, and FALSE if it 
|		 does not.
|
|		 The function will print a warning (KSYSLIB) if any
|		 of the other source objects have time, and it is
|		 will potentially be lost.
|
|		 kapu_multi_input_time is under construction.  In the future, 
|		 as we start supporting time more fully, it will do more 
|		 than just check the first object.  For example a grid type 
|		 will be added when data services supports it.
|		 The grid_type parameter will be set to the lowest common
|		 grid type of the input objects, there will be the option
|		 to set up time on all of the objects so that they are
|		 presented the same, and more information on which source 
|		 objects have time will be returned.  This will happen
|		 later. 
|
|         Input: num_src_objs - number of source objects provided in kvalist
|                kvalist - list of source objects (kobject)
|
|       Returns: TRUE (1) if time exists in first source, FALSE (0) otherwise
|  Restrictions: 
|    Written By: Donna Koechner
|          Date: Jan  3, 1995
|      Verified:
|  Side Effects: None
| Modifications:
|
------------------------------------------------------------*/

int kapu_multi_input_time(
   int	    num_src_objs,
   kvalist)
{
        kva_list        valist;
        kobject         *objlist;
	int		*time_exists; 
	int		i, count=0;
	int		result;
	char		*prog="kapu_multi_input_time";
	char		*lib="kdatamanip";

        if ((objlist = (kobject *) kmalloc((size_t) num_src_objs * 
					   sizeof(kobject))) 
		== NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for objlist");
	   return(FALSE);
	}

        if ((time_exists = (int *) kmalloc((size_t) num_src_objs * 
					   sizeof(int))) == NULL)
	{
	   kerror(lib,prog,"Unable to allocate memory for time_exists array");
	   kfree_and_NULL(objlist);
	   return(FALSE);
	}

        kva_start(valist, num_src_objs);

        for (i=0; i<num_src_objs; i++)
	{
	   objlist[i] = kva_arg(valist, kobject);
	   if (objlist[i] != KOBJECT_INVALID)
	   {
	      if ((time_exists[i] = kpds_query_time(objlist[i])) == TRUE)
	         count ++;
	   }
	}
        kva_end(valist);

        kfree_and_NULL(objlist);

	result = FALSE;
	if (count > 0)
	{
	   if (time_exists[0] && (count == 1))
	      result = TRUE;
	   else
	   {
	      kinfo(KSYSLIB, "Only Time data from the first object will "
			      "be propagated. \nTime that exists in other "
			      "source objects will be lost. \n"
		     	      "(%d out of %d have time)", count, num_src_objs);
	      result = time_exists[0];
	   }
	}
	
        kfree_and_NULL(time_exists);
	return(result);
}


/************************************************************
*
*  Routine Name: kapu_upgrade_objects - check objects for value, map, and mask data, and set up for polymorphic processing
*	
*       Purpose: This function checks the source objects for map, 
*		 value, and mask data, and it sets the processing 
*		 flags and type parameters passed in to indicate how the
*		 objects would best be processed as polymorphic objects.  
*		 If the upgrade flags are TRUE, this function will also 
*		 set up the objects for the recommended processing. 
*
*		 .IP "Upgrade Parameters:" 5
*		 The upgrade parameters are applied in this order - mapping
*		 first, then type, then size.
*		 .RS
*		 .IP "upgrade_mapping"
*		 If the upgrade_mapping parameter is true, and the 
*		 recommended processing is not process_map, then any 
*		 object that has map data will be modified so that the
*		 KPDS_MAPPING_MODE is KMAPPED.
*		 .IP "upgrade_type" 
*		 If the upgrade_type flag is TRUE, then all segments 
*		 that exist in the data objects will be upgraded to 
*		 the highest common data type for that segment.  For
*		 example, if the value data in source 1 is unsigned 
*		 byte, and the data type in source 2 is float, source 
*		 1's value data will be upgraded to float.  If mapping 
*		 has already been upgraded, the data type upgrade will be 
*		 applied to the value data (since setting KPDS_MAPPING_MODE
*	         to KMAPPED maps the data through the value segment, 
*		 removing the map).
*		 .IP "upgrade_size"
*		 If the upgrade_size flag is TRUE, then all segments that
*		 exist in the data objects will be upgraded to the highest
*		 size of each dimension, for each segment.  Again, if
*		 mapping has already been upgraded, size upgrading will 
*		 not be applied to map data.
*		 .RE
*
*		 .IP "Subobject Position Flag" 5
*		 If the use_subobj_pos flag is TRUE, the subobject position
*		 for each source object that has value or mask data is 
*		 recorded.  The minimum subobject position that is common 
*		 to all source objects is determined.  This minimum position, 
*		 along with the individual subobject positions, is used when 
*		 calculating the destination object size.  Size is calculated 
*		 so that subobject positions are aligned.  
*
*		 If the upgrade_size flag is true, the object sizes, the 
*		 value and mask segment offsets, and the subobject position 
*		 attributes will be upgraded.
*		 
*		 If invalid objects are passed in, this function will not
*		 fail, but will pass over those objects.  Invalid objects
*		 will not be taken into consideration.
*
*         Input: 
*		 upgrade_mapping - if TRUE, upgrade mapping mode is necessary
*		 upgrade_type - if TRUE, upgrade map, value, mask data types
*		 upgrade_size - if TRUE, upgrade map, value, mask data size
*		 use_subobj_pos - if TRUE, use subobject position in size 
*				  calculation
*		 num_src_objs - number of source objects provided in kvalist
*                kvalist - list of source objects (kobject)
*
*        Output: 
*		 process_map - set to TRUE when map processing is recommended
*		 process_value - set to TRUE when value processing recommended
*		 process_mask - set to TRUE when mask processing recommended
*		 map_type - recommended processing type for map data
*		 value_type - recommended processing type for value data
*		 mask_type - recommended processing type for mask data
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Donna Koechner
*          Date: Feb 10, 1995
*      Verified:
*  Side Effects: 
* Modifications:
*
*************************************************************/

int kapu_upgrade_objects(
   int	    *process_map,
   int	    *process_value,
   int	    *process_mask,
   int	    *map_type,
   int	    *value_type,
   int	    *mask_type,
   int	     upgrade_mapping,
   int	     upgrade_type,
   int	     upgrade_size,
   int	     use_subobj_pos,
   int	     num_src_objs,
   kvalist)
{
   klist    *objlist = NULL;
   char	    *rtn = "kapu_upgrade_objects";
   char	    *lib = "kapputils";
   int	    *has_value, *has_mask, *has_map;
   int	    *mw, *mh, *md, *mt, *me;
   int	    *vw, *vh, *vd, *vt, *ve;
   int	    *woff=NULL, *hoff=NULL, *doff=NULL, *toff=NULL, *eoff=NULL;
   int	     new_woff=0, new_hoff=0, new_doff=0, new_toff=0, new_eoff=0;
   int	     map_w=0, map_h=0, map_d=0, map_t=0, map_e=0;
   int	     value_w=0, value_h=0, value_d=0, value_t=0, value_e=0;
   kobject  *src_list;
   kva_list  valist;
   int	     valid_objects, i;
   int	     first_value;
   int	     same_sub_pos;
   int	     map_count = 0, mask_count = 0, value_count = 0;
   int	     maptype = KBIT, masktype = KBIT, valuetype = KBIT, type = KBIT;


   /*-------------------------------------------------------------------
   | Allocate arrays
   -------------------------------------------------------------------*/
   KCALL((src_list = (kobject *)
	  kmalloc((size_t) num_src_objs * sizeof(kobject))) != NULL);
   objlist = klist_add(objlist, src_list, "KMALLOC");

   KCALL((has_map = (int *) 
	  kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, has_map, "KMALLOC");

   KCALL((has_mask = (int *) 
	  kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, has_mask, "KMALLOC");

   KCALL((has_value = (int *) 
	  kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, has_value, "KMALLOC");

   KCALL((mw = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, mw, "KMALLOC");

   KCALL((mh = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, mh, "KMALLOC");

   KCALL((md = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, md, "KMALLOC");

   KCALL((mt = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, mt, "KMALLOC");

   KCALL((me = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, me, "KMALLOC");

   KCALL((vw = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, vw, "KMALLOC");

   KCALL((vh = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, vh, "KMALLOC");

   KCALL((vd = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, vd, "KMALLOC");

   KCALL((vt = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, vt, "KMALLOC");

   KCALL((ve = (int *) kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
   objlist = klist_add(objlist, ve, "KMALLOC");

   if (use_subobj_pos)
   {
      KCALL((woff = (int *) 
	     kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
      objlist = klist_add(objlist, woff, "KMALLOC");

      KCALL((hoff = (int *) 
	     kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
      objlist = klist_add(objlist, hoff, "KMALLOC");

      KCALL((doff = (int *) 
	     kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
      objlist = klist_add(objlist, doff, "KMALLOC");

      KCALL((toff = (int *) 
	     kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
      objlist = klist_add(objlist, toff, "KMALLOC");

      KCALL((eoff = (int *) 
	     kmalloc((size_t) num_src_objs * sizeof(int))) != NULL);
      objlist = klist_add(objlist, eoff, "KMALLOC");
   }

   /*-------------------------------------------------------------------
   | Get map, mask, and value information about each source object.
   | At the end of this, "valid_objects" will be the count of objects
   | passed in that are valid.  Segment information for all invalid 
   | objects will be set to FALSE.
   -------------------------------------------------------------------*/
   kva_start(valist, num_src_objs);

   same_sub_pos = TRUE;
   first_value = TRUE;
   valid_objects = num_src_objs;
   for (i = 0; i < num_src_objs; i++)
   {
      src_list[i] = kva_arg(valist, kobject);
      if (src_list[i] == KOBJECT_INVALID)
      {
         has_map[i] = FALSE;
	 has_mask[i] = FALSE;
	 has_value[i] = FALSE;
	 valid_objects --;
      }
      else 
      {
         if ((has_map[i] = kpds_query_map(src_list[i])) == TRUE)
	 {
	    KCALL(kpds_get_attribute(src_list[i], KPDS_MAP_DATA_TYPE, &type));
	    maptype = kdatatype_cast_output(maptype, type);
	    KCALL(kpds_get_attribute(src_list[i],
			KPDS_MAP_SIZE, &mw[i],&mh[i],&md[i],&mt[i],&me[i]));
	    map_w = kmax(map_w, mw[i]);
	    map_h = kmax(map_h, mh[i]);
	    map_d = kmax(map_d, md[i]);
	    map_t = kmax(map_t, mt[i]);
	    map_e = kmax(map_e, me[i]);
	    map_count ++;
         }
         if ((has_mask[i] = kpds_query_mask(src_list[i])) == TRUE)
	 {
	    KCALL(kpds_get_attribute(src_list[i], KPDS_MASK_DATA_TYPE, &type));
	    masktype = kdatatype_cast_output(masktype, type);
	    mask_count ++;
	 }
	 if ((has_value[i] = kpds_query_value(src_list[i])) == TRUE)
         {
	    KCALL(kpds_get_attribute(src_list[i], KPDS_VALUE_DATA_TYPE, &type));
	    valuetype = kdatatype_cast_output(valuetype, type);
	    KCALL(kpds_get_attribute(src_list[i],
			KPDS_VALUE_SIZE, &vw[i],&vh[i],&vd[i],&vt[i],&ve[i]));
	    value_count ++;
         }
	 if (has_mask[i] && !has_value[i])
         {
	    KCALL(kpds_get_attribute(src_list[i],
			KPDS_MASK_SIZE, &vw[i],&vh[i],&vd[i],&vt[i],&ve[i]));
	 }
	 if (use_subobj_pos && (has_value[i] || has_mask[i]))
         {
            KCALL(kpds_get_attribute(src_list[i], KPDS_SUBOBJECT_POSITION, 
				&woff[i],&hoff[i],&doff[i],&toff[i],&eoff[i]));
            if (first_value == TRUE)
	    {
	       new_woff = woff[i];
	       new_hoff = hoff[i];
	       new_doff = doff[i];
	       new_toff = toff[i];
	       new_eoff = eoff[i];
	       first_value = FALSE;
            }
	    else
	    {
	       new_woff = kmin(new_woff, woff[i]);
	       new_hoff = kmin(new_hoff, hoff[i]);
	       new_doff = kmin(new_doff, doff[i]);
	       new_toff = kmin(new_toff, toff[i]);
	       new_eoff = kmin(new_eoff, eoff[i]);
	    }

            /*----------------------------------------------------------
            | If all subobject positions are the same, don't need to 
	    | worry about them.
            ----------------------------------------------------------*/
	    if ((new_woff != woff[i]) || (new_hoff != hoff[i]) ||
	        (new_doff != doff[i]) || (new_toff != toff[i]) ||
		(new_eoff != eoff[i]) )
	    {
	       same_sub_pos = FALSE;
	    }
	 } 
      }
   }
   kva_end(valist);

   /*-------------------------------------------------------------------
   | If all of the subobject positions were the same, it isn't necessary
   | to take it into consideration when calculating size and offset.
   -------------------------------------------------------------------*/
   if ((use_subobj_pos == TRUE) && (same_sub_pos == TRUE))
      use_subobj_pos = FALSE;

   if (map_count == 0)
      *process_map = FALSE;
   if (mask_count == 0)
      *process_mask = FALSE;
   if (value_count == 0)
   {
      *process_value = FALSE;
      *process_mask  = FALSE;
   }

   /*-------------------------------------------------------------------
   | If both map and value exist, then any object that has map must
   | also have value.  
   -------------------------------------------------------------------*/
   if ((valid_objects == 1) && (map_count > 0))
   {
      *process_map = TRUE;
      *process_value = FALSE;
      if (mask_count > 0)
         *process_mask = TRUE;
   }
   else if ((value_count > 0) && (map_count > 0))
   {
      for (i=0; i<num_src_objs; i++)
      {
         if (has_map[i])
	 {
	    if (!has_value[i])
	    {
	       kerror(lib, rtn, "Object %d must have value data.",i);
	       (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
            }
         }
      }
      *process_map = FALSE;
      *process_value = TRUE;
      if (mask_count > 0)
         *process_mask = TRUE;
   }
   else if ((value_count > 0) && (map_count == 0))
   {
      *process_map = FALSE;
      *process_value = TRUE;
      if (mask_count > 0)
         *process_mask = TRUE;
   }
   else if ((value_count == 0) && (map_count > 0))
   {
      *process_map = TRUE;
      *process_value = FALSE;
      *process_mask = FALSE;
   }

   /*-------------------------------------------------------------------
   | If the flag is true, set the objects up for processing.  First 
   | modify type, then mapping (may need to redo size indices), then 
   | size
   -------------------------------------------------------------------*/
   if (upgrade_mapping && (map_count > 0) && (value_count > 0))
   {
      kva_start(valist, num_src_objs);

      for (i =0 ; i < num_src_objs; i++)
      {
         src_list[i] = kva_arg(valist, kobject);
	 if (src_list[i] != KOBJECT_INVALID)
	 {
            /*----------------------------------------------------------
	    | Set mapping mode on all objects with maps, set map 
	    | count to zero.  determine greater of map/value type.
            ----------------------------------------------------------*/
	    if (has_map[i])
	    {
	       KCALL(kpds_set_attribute(src_list[i],
					KPDS_MAPPING_MODE, KMAPPED));
	       map_count --;
	       has_map[i] = FALSE;
	       KCALL(kpds_get_attribute(src_list[i], KPDS_VALUE_SIZE, 
				&vw[i], &vh[i], &vd[i], &vt[i], &ve[i]));
	       valuetype = kdatatype_cast_output(valuetype, maptype);
	    }
         }
      }
      kva_end(valist);
   }

   /*-------------------------------------------------------------------
   | If there is value or mask data, need to calculate value size and 
   | offsets.  (the max map size has already been determined)
   -------------------------------------------------------------------*/

   if (upgrade_size && (process_value || process_mask))
   for (i = 0; i < num_src_objs; i++)
   {
      if (has_value[i] || has_mask[i])
      {
         if (use_subobj_pos)
	 {
            /*----------------------------------------------------------
            | If using subobject position, calculate new size wrt to 
            | minimum subobject position found. 
            ----------------------------------------------------------*/
	    vw[i] = vw[i] + woff[i] - new_woff;
	    vh[i] = vh[i] + hoff[i] - new_hoff;
	    vd[i] = vd[i] + doff[i] - new_doff;
	    vt[i] = vt[i] + toff[i] - new_toff;
	    ve[i] = ve[i] + eoff[i] - new_eoff;
         }
         /*-------------------------------------------------------------
	 | Determine the maximum of all sizes, 
         -------------------------------------------------------------*/
	 value_w = kmax(value_w, vw[i]);
	 value_h = kmax(value_h, vh[i]);
         value_d = kmax(value_d, vd[i]);
	 value_t = kmax(value_t, vt[i]);
	 value_e = kmax(value_e, ve[i]);
      }
   }

   if (upgrade_type || upgrade_size)
   {
      kva_start(valist, num_src_objs);

      for (i = 0; i < num_src_objs; i++)
      {
         src_list[i] = kva_arg(valist, kobject);
	 if (src_list[i] != KOBJECT_INVALID)
	 {
	    if (upgrade_type)
	    {
               /*-------------------------------------------------------
	       | If count of segment > 0, set type on all objects 
               -------------------------------------------------------*/
	       if (map_count > 0)
	       {
	          if (has_map[i])
		  {
		     KCALL(kpds_set_attribute(src_list[i],
					KPDS_MAP_DATA_TYPE, maptype));
		  }
               }
               if (value_count > 0)
	       {
	          if (has_value[i])
		  {
		     KCALL(kpds_set_attribute(src_list[i],
					KPDS_VALUE_DATA_TYPE, valuetype));
		  }
               }
	       if (mask_count > 0)
	       {
	          if (has_mask[i])
		  {
		     KCALL(kpds_set_attribute(src_list[i],
					KPDS_MASK_DATA_TYPE, masktype));
		  }
               }
            }

            if (upgrade_size)
	    {
               /*-------------------------------------------------------
	       | If upgrade_size, match size of value and or map 
	       | segments if their count is greater than 1.  don't 
	       | operate on mask 
               -------------------------------------------------------*/
	       if (map_count > 0)
	       {
	          if (has_map[i])
		  {
		     KCALL(kpds_set_attribute(src_list[i], KPDS_MAP_SIZE, 
				map_w, map_h, map_d, map_t, map_e));
		  }
               }
               /*-------------------------------------------------------
	       | Set size on all objects.
	       | If subobject position is being used, set offset on value 
	       | and or mask to be offset[i] - new_offset, and set the 
	       | subobject position of all objects to be the new_offset.  
	       | Make sure to set offsets in both value and mask data so 
	       | that padding is correct when subobject positions are used.
               -------------------------------------------------------*/
	       if (value_count > 0)
	       {
	          if (has_value[i])
		  {
		     KCALL(kpds_set_attribute(src_list[i], KPDS_VALUE_SIZE, 
				value_w, value_h, value_d, value_t, value_e));
		     if (use_subobj_pos) 
		     {
		        KCALL(kpds_set_attribute(src_list[i], KPDS_VALUE_OFFSET,
				(new_woff - woff[i]), (new_hoff - hoff[i]), 
				(new_doff - doff[i]), (new_toff - toff[i]), 
				(new_eoff - eoff[i])));
		        KCALL(kpds_set_attribute(src_list[i],
				KPDS_SUBOBJECT_POSITION, new_woff, new_hoff, 
				new_doff, new_toff, new_eoff));
			if (has_mask[i])
			{
		           KCALL(kpds_set_attribute(src_list[i],
				KPDS_MASK_OFFSET, (new_woff - woff[i]), 
				(new_hoff - hoff[i]), (new_doff - doff[i]), 
				(new_toff - toff[i]), (new_eoff - eoff[i])));
			}
		     }
		  }
		  else if (has_mask[i])
		  {
		     KCALL(kpds_set_attribute(src_list[i], KPDS_MASK_SIZE, 
				value_w, value_h, value_d, value_t, value_e));
		     if (use_subobj_pos) 
		     {
		        KCALL(kpds_set_attribute(src_list[i], KPDS_MASK_OFFSET,
				(new_woff - woff[i]), (new_hoff - hoff[i]), 
				(new_doff - doff[i]), (new_toff - toff[i]), 
				(new_eoff - eoff[i])));
		        KCALL(kpds_set_attribute(src_list[i],
				KPDS_SUBOBJECT_POSITION, new_woff, new_hoff, 
				new_doff, new_toff, new_eoff));
		     }
		  }
               }
            }
         }
      }
      kva_end(valist);
   }

   *map_type = maptype;
   *value_type = valuetype;
   *mask_type = masktype;

   (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
   return TRUE;
}

/****************************************************************
* 
*  Routine Name: kapu_scale_data - scales the data
* 
*       Purpose: For unmapped data, the value segment is scaled
*                using KPDS_VALUE_SCALING.  For mapped data, the
*                map segment is scaled using KPDS_MAP_SCALING.
*                Attributes are set so that any subsequent calls
*                to kpds_get_data will provide scaled data.  Note
*                that the actual data is not transformed by this
*                function.  If KPDS_DATA_RANGE is set, then these
*                values are used for the input minimum and maximum; 
*                otherwise, the data is scanned to determine the
*                input minimum and maximum.
*
*         Input: src_obj - object to be scaled.
*                min - minimum of scaled data.
*                max - maximum of scale data.
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Thomas H. Robey
*          Date: Nov 15, 1994
*      Verified: 
*  Side Effects: 
* Modifications:
****************************************************************/
int 
kapu_scale_data(kobject *src_obj, double min, double max)
{
   int i;
   double datamin, datamax;
   double factor, offset;
   char *lib = "licomposite";
   char *rtn = "kapu_scale_data";
   klist *objlist = NULL;

   /* Check input */

   i = kpds_get_attribute(*src_obj, KPDS_DATA_RANGE, &datamin, &datamax);

   if (kpds_query_map(*src_obj))
   {
      if (i == 0)
         KCALL(kapu_minmax(*src_obj, KPDS_MAP_ALL, &datamin, &datamax));

      factor = (max - min) / (datamax - datamin);
      offset = min - datamin * factor;
      KCALL(kpds_set_attributes(*src_obj,KPDS_MAP_SCALE_FACTOR,factor,
            KPDS_MAP_SCALE_OFFSET,offset,0.0,
            KPDS_MAP_SCALING,KSCALE,NULL));
   }
   else
   {
      if (i == 0)
         KCALL(kapu_minmax(*src_obj, KPDS_VALUE_ALL, &datamin, &datamax));

      factor = (max - min) / (datamax - datamin);
      offset = min - datamin * factor;
      KCALL(kpds_set_attributes(*src_obj,KPDS_VALUE_SCALE_FACTOR,factor,
            KPDS_VALUE_SCALE_OFFSET,offset,0.0,
            KPDS_VALUE_SCALING,KSCALE,NULL));
   }

   return TRUE;
}

/****************************************************************
* 
*  Routine Name: kapu_minmax - Obtains minimum and maximum of
*                segment
* 
*       Purpose: Obtains the minimum and maximum for the primitive 
*                of the specified object.  KPDS_VALUE_ALL reads
*                optimally-sized regions; otherwise, the entire 
*                primitive is read in one chunk. 
*
*         Input: obj - object
*                segment - segment 
*
*        Output: min - minimum of segment
*                max - maximum of segment
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: The only location primitive supported is 
*                KPDS_LOCATION_ALL
*    Written By: Thomas H. Robey
*          Date: Nov 15, 1994
*      Verified: 
*  Side Effects: 
* Modifications:
****************************************************************/
int
kapu_minmax(kobject obj, char *primitive, double *min, double *max)
{
   double  nmin;
   double  nmax;
   kaddr   data = NULL;
   kobject tmp = NULL;
   int datatype;
   int w;
   int h;
   int d;
   int t;
   int e;
   int n;
   int j;
   int size = 0;
   char *segment = NULL;
   char *string;
   char *lib = "licomposite";
   char *rtn = "kapu_minmax";
   klist *objlist = NULL;

   *min=0.0;
   *max=0.0;

   /* Determine the segment */

   string=kstrstr(primitive,"Value");
   if (string!=NULL)
      segment = "value";
   string=kstrstr(primitive,"Mask");
   if (string!=NULL)
      segment = "mask";
   string=kstrstr(primitive,"Map");
   if (string!=NULL)
      segment = "map";
   string=kstrstr(primitive,"Location");
   if (string!=NULL)
      segment = "location";
   string=kstrstr(primitive,"Time");
   if (string!=NULL)
      segment = "time";
   if (segment == NULL)
   {
      kerror(lib, rtn, "Invalid segment.");
      return FALSE;
   }

   /* verify that the segment is present */
   if ((kstrcmp(segment,"value") == 0 && !kpds_query_value(obj)) || 
      !kdms_query_segment(obj, segment))
   {
      kerror(lib, rtn, "Segment %s doesn't exist.", segment);
      return FALSE;
   }

   tmp = kdms_reference(obj);
   objlist = klist_add(objlist, tmp, "KOBJECT");

   /* get the current presentation data type */
   if (kstrcmp(segment, "value") == 0)
      kpds_get_attribute(tmp, KPDS_VALUE_DATA_TYPE, &datatype);
   else
      kdms_get_attribute(tmp, segment, KDMS_DATA_TYPE, &datatype);
   


   /* can't deal with complex and double complex, so convert to double */
   if (datatype == KCOMPLEX || datatype == KDCOMPLEX)
   {
      datatype = KDOUBLE;
      KCALL(kdms_set_attribute(tmp, primitive, KDMS_DATA_TYPE, KDOUBLE));
   }

   if (datatype == KBIT)
   {
      datatype = KUBYTE;
      KCALL(kdms_set_attribute(tmp, primitive, KDMS_DATA_TYPE, KUBYTE));
   }

   if (kstrcmp(segment,"value") == 0)
   {
      if (kstrcmp(primitive, KPDS_VALUE_ALL) == 0)
      {

         KCALL(kpds_get_attribute(tmp,KPDS_VALUE_OPTIMAL_REGION_SIZE,
               &w,&h,&d,&t,&e,&n));

         KCALL(kpds_set_attributes(tmp,KPDS_VALUE_REGION_SIZE,w,h,d,t,e,
               KPDS_VALUE_POSITION,0,0,0,0,0,NULL));

         for (j = 0; j < n; j++)
         {
            KCALL((data=kpds_get_data(tmp,KPDS_VALUE_REGION,data)) != NULL);
            KCALL(kdata_minmax(data, w*h*d*t*e, datatype, &nmin, &nmax));
            if (j > 0)
            {
               *min = kmin(nmin,*min);
               *max = kmax(nmax,*max);
            }
            else
            {
               *min = nmin;
               *max = nmax;
            }
         }

         kfree_and_NULL(data);
      }
      else
      {
         if (kstrcmp(primitive, KPDS_VALUE_POINT) == 0)
            size = 1;
         else if (kstrcmp(primitive, KPDS_VALUE_VECTOR) == 0)
         {
            KCALL(kpds_get_attribute(tmp, KPDS_VALUE_VECTOR_INFO, &size, &n));
         }
         else if (kstrcmp(primitive, KPDS_VALUE_LINE) == 0)
         {
            KCALL(kpds_get_attribute(tmp, KPDS_VALUE_LINE_INFO, &size, &n));
         }
         else if (kstrcmp(primitive, KPDS_VALUE_PLANE) == 0)
         {
            KCALL(kpds_get_attribute(tmp, KPDS_VALUE_PLANE_INFO, &w, &h, &n));
            size = w * h;
         }
         else if (kstrcmp(primitive, KPDS_VALUE_VOLUME) == 0)
         {
            KCALL(kpds_get_attribute(tmp,KPDS_VALUE_VOLUME_INFO,&w,&h,&d,&n));
            size = w * h * d;
         }
         else if (kstrcmp(primitive, KPDS_VALUE_REGION) == 0)
         {
            KCALL(kpds_get_attribute(tmp,KPDS_VALUE_REGION_SIZE,&w,&h,&d,&t,
                  &e));
            size = w * h * d * t * e;
         }
         else
         {
	    (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
            kerror(lib, rtn, "Invalid primitive %s.", primitive);
            return FALSE;
         } 

         KCALL((data = kpds_get_data(tmp, primitive, data)) != NULL)
         KCALL(kdata_minmax(data, size, datatype, min, max));

         kfree_and_NULL(data);
      }
   }
   else if (kstrcmp(segment,"map") == 0)
   {
      if (kstrcmp(primitive, KPDS_MAP_POINT) == 0)
         size = 1;
      else if (kstrcmp(primitive, KPDS_MAP_VECTOR) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MAP_VECTOR_INFO, &size, &n));
      }
      else if (kstrcmp(primitive, KPDS_MAP_LINE) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MAP_LINE_INFO, &size, &n));
      }
      else if (kstrcmp(primitive, KPDS_MAP_PLANE) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MAP_PLANE_INFO, &w, &h, &n));
         size = w * h;
      }
      else if (kstrcmp(primitive, KPDS_MAP_VOLUME) == 0)
      {
         KCALL(kpds_get_attribute(tmp,KPDS_MAP_VOLUME_INFO,&w,&h,&d,&n));
         size = w * h * d;
      }
      else if (kstrcmp(primitive, KPDS_MAP_REGION) == 0)
      {
         KCALL(kpds_get_attribute(tmp,KPDS_MAP_REGION_SIZE,&w,&h,&d,&t,&e));
         size = w * h * d * t * e;
      }
      else if (kstrcmp(primitive, KPDS_MAP_ALL) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MAP_SIZE, &w, &h, &d, &t, &e));
         size = w * h * d * t * e;
      }
      else
      {
	 (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
         kerror(lib, rtn, "Invalid primitive %s.", primitive);
         return FALSE;
      } 

      KCALL((data = kpds_get_data(tmp, primitive, data)) != NULL)
  
      KCALL(kdata_minmax(data, size, datatype, min, max));

      kfree_and_NULL(data);
   }
   else if (kstrcmp(segment,"mask") == 0)
   {
      if (kstrcmp(primitive, KPDS_MASK_POINT) == 0)
         size = 1;
      else if (kstrcmp(primitive, KPDS_MASK_VECTOR) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MASK_VECTOR_INFO, &size, &n));
      }
      else if (kstrcmp(primitive, KPDS_MASK_LINE) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MASK_LINE_INFO, &size, &n));
      }
      else if (kstrcmp(primitive, KPDS_MASK_PLANE) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MASK_PLANE_INFO, &w, &h, &n));
         size = w * h;
      }
      else if (kstrcmp(primitive, KPDS_MASK_VOLUME) == 0)
      {
         KCALL(kpds_get_attribute(tmp,KPDS_MASK_VOLUME_INFO,&w,&h,&d,&n));
         size = w * h * d;
      }
      else if (kstrcmp(primitive, KPDS_MASK_REGION) == 0)
      {
         KCALL(kpds_get_attribute(tmp,KPDS_MASK_REGION_SIZE,&w,&h,&d,&t,&e));
         size = w * h * d * t * e;
      }
      else if (kstrcmp(primitive, KPDS_MASK_ALL) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_MASK_SIZE, &w, &h, &d, &t, &e));
         size = w * h * d * t * e;
      }
      else
      {
	 (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
         kerror(lib, rtn, "Invalid primitive %s.", primitive);
         return FALSE;
      } 

      KCALL((data = kpds_get_data(tmp, primitive, data)) != NULL)
  
      KCALL(kdata_minmax(data, size, datatype, min, max));

      kfree_and_NULL(data);
   }
   else if (kstrcmp(segment,"location") == 0)
   {
      int grid;

      if (kstrcmp(primitive, KPDS_LOCATION_ALL) != 0)
      {
	 (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
         kerror(lib, rtn, "%s currently unsupported.", primitive);
         return FALSE;
      }

      /* get the current grid type -- KRECTILINEAR or KCURVILINEAR */
      KCALL(kpds_get_attribute(obj, KPDS_LOCATION_GRID, &grid));

      /*
       *  if the grid type is rectlinear, then we need to get the
       *  location points from the width, height, and depth segments.
       */
      if (grid == KRECTILINEAR)
      {
         kaddr data_w = NULL;
         kaddr data_h = NULL;
         kaddr data_d = NULL;


         data_w = kpds_get_data(tmp, KPDS_LOCATION_WIDTH_REGION, NULL);
         data_h = kpds_get_data(tmp, KPDS_LOCATION_HEIGHT_REGION, NULL);
         data_d = kpds_get_data(tmp, KPDS_LOCATION_DEPTH_REGION, NULL);

         if (data_w == NULL || data_h == NULL || data_d)
         {
	    (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
            kerror(lib, rtn, "Data channel error.");
            return FALSE;
         }

         KCALL(kpds_get_attribute(tmp,KPDS_LOCATION_REGION_SIZE,&w,&h,&d,&e));
         KCALL(kdata_minmax(data_w, w, datatype, &nmin, &nmax));
         *min = kmin(nmin,*min);
         *max = kmin(nmax,*max);
         KCALL(kdata_minmax(data_h, h, datatype, &nmin, &nmax));
         *min = kmin(nmin,*min);
         *max = kmin(nmax,*max);
         KCALL(kdata_minmax(data_d, d, datatype, &nmin, &nmax));
         *min = kmin(nmin,*min);
         *max = kmin(nmax,*max);


         kfree_and_NULL(data_w);
         kfree_and_NULL(data_h);
         kfree_and_NULL(data_d);
      }
      else if (grid == KCURVILINEAR)
      {
         KCALL((data = kpds_get_data(tmp, KPDS_LOCATION_REGION, NULL)) != NULL)

         KCALL(kpds_get_attribute(tmp,KPDS_LOCATION_REGION_SIZE,&w,&h,&d,&e));

         KCALL(kdata_minmax(data, w*h*d*e, datatype, min, max));
      }

      kfree_and_NULL(data);
   }
   else if (kstrcmp(segment,"time") == 0)
   {
      if (kstrcmp(primitive, KPDS_TIME_POINT) == 0)
         size = 1;
      else if (kstrcmp(primitive, KPDS_TIME_REGION) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_TIME_REGION_SIZE, &size));
      }
      else if (kstrcmp(primitive, KPDS_TIME_ALL) == 0)
      {
         KCALL(kpds_get_attribute(tmp, KPDS_TIME_SIZE, &size));
      }
      else
      {
	 (void)klist_free(objlist, (void(*)(klist*))lkcall_free);
         kerror(lib, rtn, "Invalid primitive %s.", primitive);
         return FALSE;
      } 

      KCALL((data = kpds_get_data(tmp, primitive, data)) != NULL)
  
      KCALL(kdata_minmax(data, size, datatype, min, max));

      kfree_and_NULL(data);
   }

   (void)klist_free(objlist, (void(*)(klist*))lkcall_free);

   return TRUE;
}

