/*
  * Khoros: $Id: format.c,v 1.6.2.3 1997/03/11 17:15:54 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: format.c,v 1.6.2.3 1997/03/11 17:15:54 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Management Services Format Glue Code
   >>>>		
   >>>>  Private:
   >>>>                 kdas_format_construct()
   >>>>                 kdas_format_dup()
   >>>>                 kdas_format_destruct()
   >>>>                 kdas_to_kfile_flags()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/****    =============================================================
 ****    Format Manifestation Methods
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_format_construct - construct a data services format
|
|       Purpose: This routine will construct a format structure
|		 using the given format name and glue functions.
|
|         Input: name        - the name of the format to construct
|		 description - the ascii description of the format to construct
|		 check          - check function for format
|		 input          - input function for format
|		 output         - output function for format
|		 destroy        - destroy function for format
|		 prepare_output - prepare output function for format
|		 locate_segment - locate segment function for format
|		 reader         - optional reader function for format
|		 writer         - optional writer function for format
|		 seeker         - optional seeker function for format
|		 teller         - optional teller function for format
|
|        Output: none
|
|       Returns: the constructed format or NULL if it can't be created
|
|    Written By: Steve Kubica
|          Date: June 1, 1994
| Modifications:
|
------------------------------------------------------------*/
kdas_format *
kdas_format_construct(
   char    *name,
   char    *description,
   int      (*check)   	     PROTO((int)), 
   int      (*input)   	     PROTO((kobject, int, int)),
   int      (*output)  	     PROTO((kobject, int, int)),
   int      (*destroy) 	     PROTO((kobject)),
   int      (*prepare_output) PROTO((kobject)),
   int      (*locate_segment) PROTO((kobject, char *)),
   ssize_t  (*reader)         PROTO((int, kaddr, size_t, int)),
   ssize_t  (*writer)         PROTO((int, kaddr, size_t, int )),
   off_t    (*seeker)         PROTO((int, off_t, int)),
   long int (*teller)         PROTO((int)))
{
   kdas_format *format;
   

   /*-- sanity check to make sure that name or description are not NULL --*/
   if (!name || !description) return NULL;
   
   /*-- allocate the new format and initialze its fields --*/
   format = (kdas_format *) kmalloc(sizeof(kdas_format));

   /*-- we must have either an input or output function --*/
   if (input == NULL && output == NULL )
   {
      kfree(format);
      return NULL; /*-- kerror, gotta have either input or output --*/
   }

   format->check   	  = check;
   format->input   	  = input;
   format->output   	  = output;
   format->destroy 	  = destroy;
   format->prepare_output = prepare_output;

   /*-- large data set support functons are optional --*/
   if (locate_segment)
   {
      format->locate_segment = locate_segment;

      /*-- use the reader and writer if provided, else use generic ones --*/
      format->reader = (reader == NULL) ? kread_generic  : reader;
      format->writer = (writer == NULL) ? kwrite_generic : writer;
      format->seeker = (seeker == NULL) ? klseek         : seeker;
      format->teller = (teller == NULL) ? ktell          : teller;
   }
   
   /*-- no large data set support, so bag all these functions --*/
   else 
   {
      format->locate_segment = NULL;
      format->reader         = NULL;
      format->writer         = NULL;
      format->seeker         = NULL;
      format->teller         = NULL;
   }

   format->name        = kstrdup(name);
   format->description = kstrdup(description);

   /*-- these are used on a per-object basis --*/
   format->resources = NULL;
   format->flags     = 0;

   return format;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_format_dup - duplicate a data services format
|
|       Purpose: This routine will duplicate a given format structure.
|
|         Input: format - the format to duplicate
|
|        Output: none
|
|       Returns: the duplicated format structuer
|
|    Written By: Steve Kubica
|          Date: March 7, 1996
| Modifications:
|
------------------------------------------------------------*/
kdas_format *kdas_format_dup(kdas_format *format)
{
   kdas_format *nformat;
   
   if (format == NULL)
      return NULL;

   nformat = (kdas_format *) kdupalloc(format, sizeof(kdas_format));
   nformat->name        = kstrdup(format->name);
   nformat->description = kstrdup(format->name);

   return nformat;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_format_destruct - destruct a data services format
|
|       Purpose: This routine will destruct a given format structure.
|
|         Input: format - the format to destruct
|
|        Output: none
|
|       Returns: nothing
|
|    Written By: Steve Kubica
|          Date: March 11, 1994
| Modifications:
|
------------------------------------------------------------*/
void kdas_format_destruct(kdas_format *format)
{
   /*-- trivial free --*/
   if (format == NULL)
      return;

   kfree(format->name);
   kfree(format->description);
   kfree(format);
   return;
}

/****    =============================================================
 ****    Format Utility Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_to_kfile_flags - convert data service flags
|
|       Purpose: This routine will convert the given data services
|		 flags to the corresponding transport flags.
|
|         Input: flags - the data service flags to convert
|
|        Output: none
|
|       Returns: the converted tranport flags
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: June 2, 1994
| Modifications:
|
------------------------------------------------------------*/
int kdas_to_kfile_flags(int flags)
{
   /*-- read-write case -- */
   if ((KOBJ_WRITE & flags) && (KOBJ_READ & flags))
      return O_RDWR;

   /*-- read-only case --*/
   if (!(KOBJ_WRITE & flags) && (KOBJ_READ & flags))
      return O_RDONLY;

   /*-- write-only case (this breaks gdbm) --*/
   if ((KOBJ_WRITE & flags) && !(KOBJ_READ & flags)) 
      return O_WRONLY | O_CREAT | O_TRUNC;

   return 0;
}
