 /*
  * Khoros: $Id: seg.c,v 1.11.2.3 1997/03/11 17:16:01 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: seg.c,v 1.11.2.3 1997/03/11 17:16:01 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Segment Code
   >>>>		
   >>>>  Private:
   >>>>                 kdas_segment_construct()
   >>>>                 kdas_segment_destruct()
   >>>>                 kdas_segment_set_all()
   >>>>                 kdas_segment_get_all()
   >>>>                 kdas_segment_set_dimension()
   >>>>                 kdas_segment_set_size()
   >>>>                 kdas_segment_set_order()
   >>>>                 kdas_segment_set_datatype()
   >>>>                 kdas_segment_move_memory_to_temp()
   >>>>                 kdas_segment_move_temp_to_memory()
   >>>>                 kdas_segment_get_data()
   >>>>                 kdas_segment_put_data()
   >>>>                 kdas_segment_write()
   >>>>   Static:
   >>>>                 _kdas_segment_access()
   >>>>                 _kdas_find_data()
   >>>>                 _kdas_read_all_data()
   >>>>                 _kdas_find_big_data()
   >>>>                 _wrap_region()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

typedef void (*kdas_memory_func)    PROTO((kaddr, kaddr, int, int, 
					   int *, int *, int *, int *, int *, 
					   int *, int));
typedef void (*kdas_transport_func) PROTO((kdas_transport *, kaddr, int, int, 
					   int *, int *, int *, int *, int *, 
					   int *, int));

static kdas_memory_func    
       memory_access[2]    = {kdas_memory_clip, kdas_memory_interp};

static kdas_transport_func 
       transport_access[2] = {kdas_transport_clip, kdas_transport_interp};


static int  _kdas_segment_access      PROTO((kdas_segment *, int *, int *, 
					     int *, int *, int *, int, 
					     int, kaddr));
static int   _kdas_find_data          PROTO((kdas_segment *, int, kaddr *, 
					     kdas_transport **));
static kaddr _kdas_read_all_data      PROTO((kdas_segment *));

static kaddr _kdas_find_big_data      PROTO((kdas_segment *, int));

static void  _wrap_region             PROTO((kaddr, int *, int *, int *, int *,
					     int *, int *, int *, kaddr));

/****    =============================================================
 ****    Segment Manifestation Methods
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_construct
|
|       Purpose: This routine will construct a segment using the 
|		 given segment name. All fields of the segment 
|		 structure will be initialized to be 0 or NULL.
|
|         Input: segment_name - the name of the segment
|
|        Output: none
|
|       Returns: the constructed segment or NULL if it can't create
|
|    Written By: Steve Kubica
|          Date: March 11, 1994
| Modifications:
|
------------------------------------------------------------*/
kdas_segment *
kdas_segment_construct(char *segment_name)
{
   kdas_segment *segment;
   

   /*-- sanity check to make sure that string is not NULL --*/
   if (!segment_name) return NULL;

   /*-- allocate the new segment and initialze its fields --*/
   segment = (kdas_segment *) kmalloc(sizeof(kdas_segment));

   segment->id = kstring_to_token(segment_name);

   segment->attributes = NULL;
   segment->callbacks  = NULL;

   segment->data = NULL;
   segment->temp = NULL;
   segment->dimension = -1;
   segment->datatype  = -1;
   segment->size      = NULL;
   segment->order     = NULL;

   segment->rval = 0.0;
   segment->ival = 0.0;

   segment->format = NULL;
   segment->object = NULL;

   segment->references = NULL;
		     
   return segment;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_destruct
|
|       Purpose: This routine will destruct a given segment.
|
|         Input: segment  - the segment to destruct
|
|        Output: none
|
|       Returns: nothing
|
|    Written By: Steve Kubica
|          Date: March 11, 1994
| Modifications:
|
------------------------------------------------------------*/
void kdas_segment_destruct(kdas_segment *segment)
{
   /*-- trivial free --*/
   if (segment == NULL)
      return;

   kfree(segment->data);
   kdas_temp_destruct(segment->temp);

   /*-- if there are any attributes, then delete those --*/
   if (segment->attributes)
      klist_free(segment->attributes, kdas_attribute_list_free);

   /*-- if there are any callbacks, then delete those --*/
   if (segment->callbacks)
      klist_free(segment->callbacks, NULL);

   kfree(segment->size);
   kfree(segment->order);
   kfree(segment);
   return;
}

/****    =============================================================
 ****    Segment Characteristics Methods
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_set_all - assign everything for a segment
|
|       Purpose: This routine will set the dimensionality, size, 
|		 index order and datatype of a segment as well as 
|		 optionally allowing you to assign the actual segment 
|		 data directly.
|
|		 Anything that was previously contained in the 
|		 segment will be discarded.
|
|		 The segment data should only be assigned directly
|		 if it is all present.  The act of assigning this
|		 data with this routine will cause the large
|		 data set management to be bypassed.  If you
|		 have the capability of handling large data sets
|		 then NULL should be passed for the data as the
|		 data will be clipped directly from the file.
|
|         Input: segment   - segment on which we will set everything
|		 dimension - new dimensionality to set
|		 size      - new size to set
|		 order     - new order to set
|		 datatype  - new datatype to set
|		 data      - segment data if all is present
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 11, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_set_all(
   kdas_segment *segment,
   int           dimension,
   int          *size,
   int          *order,
   int           datatype,
   kaddr         data)
{
   if (segment == NULL)
      return FALSE;

   /*-- tidy up if we have to --*/
   if (size  && segment->size)  kfree(segment->size);
   if (order && segment->order) kfree(segment->order);
   if (segment->data)  kfree(segment->data);
   if (segment->temp != NULL) kdas_temp_destruct(segment->temp);
   
   /*-- assign what we've got --*/
   if (dimension != -1) segment->dimension = dimension;
   if (datatype  != -1) segment->datatype  = datatype;
   if (size)     segment->size  = (int *) kmalloc(sizeof(int) * dimension);
   if (order)    segment->order = (int *) kmalloc(sizeof(int) * dimension);

   if (size)
      kmemcpy(segment->size,  size,  sizeof(int) * dimension);

   if (order)
      kmemcpy(segment->order, order, sizeof(int) * dimension);

   /*-- if they gave us data, then we've got it all --*/
   if (data != NULL) 
      segment->data = data;
   else
      data = NULL;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_get_all - retrieve everything from a segment
|
|       Purpose: This routine will get the dimensionality, size, 
|		 index order and datatype of a segment as well as 
|		 optionally allowing you to retreive the actual segment 
|		 data directly.
|
|		 The data pointer returned may be NULL if the data
|		 is either still in the original file, or if the
|		 data is in a temporary segment, as it would
|		 be if it qualifies as a "large" data set.
|
|         Input: segment   - segment on which we will set everything
|
|        Output: dimension - segment dimensionality
|		 size      - segment size
|		 order     - segment order
|		 datatype  - segment datatype
|		 data      - segment data if it is present
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 11, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_get_all(
   kdas_segment *segment,
   int          *dimension,
   int         **size,
   int         **order,
   int          *datatype,
   kaddr        *data)
{
   if (segment == NULL)
      return FALSE;
   
   if (dimension) *dimension = segment->dimension;
   if (size)      *size      = segment->size;
   if (order)     *order     = segment->order;
   if (datatype)  *datatype  = segment->datatype;
   if (data)      *data      = segment->data;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_set_dimension - set the segment dimensionality
|
|       Purpose: Set the dimensionality of a segment.  
|
|		 Setting the dimensionality is what manages the
|		 size of the segment's size array.  The size of
|		 the segment's size array should only be changed
|		 by this routine.
|
|         Input: segment   - segment on which we will set the dimensionality
|		 dimension - new dimensionality to set
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 11, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_set_dimension(
   kdas_segment *segment,
   int           dimension)
{
   int i;
   
   if (dimension == -1) /*-- nullop --*/
      return TRUE;
   
   if (segment == NULL)
      return FALSE;

   /*----------------------------------------------------------------
   |   modify the segment size dimensionality
   \-----------------------------------------------------------------*/
   if (segment->size == NULL)
   {
      /*-- set the size array up to be of new dimension and initialize --*/
      segment->size = (int *) kmalloc(sizeof(int) * dimension);
      for (i = 0; i < dimension; i++)
	 segment->size[i] = 0; /* -1; */
   }
   else /*-- dimensionality changing --*/
   {
      if (dimension < segment->dimension)  /*-- smaller --*/
      {
	 /*-- roll old sizes into the last dimension --*/
	 for (i = dimension; i < segment->dimension; i++)
	    segment->size[dimension-1] *= segment->size[i];

	 segment->size = (int *) krealloc(segment->size, sizeof(int) * dimension);
      }
      else if (dimension > segment->dimension)  /*-- larger --*/
      {
	 segment->size  = (int *) krealloc(segment->size, sizeof(int) * dimension);
	 
	 /*-- pad out new dimension sizes with 1 --*/ 
	 for (i = segment->dimension; i < dimension; i++)
	    segment->size[i] = 1;
      }
   }

   /*----------------------------------------------------------------
   |   modify the segment index order dimensionality
   \-----------------------------------------------------------------*/
   if (segment->order == NULL)
   {
      /*-- set the order array up to be of new dimension and initialize --*/
      segment->order = (int *) kmalloc(sizeof(int) * dimension);
      for (i = 0; i < dimension; i++)
	 segment->order[i] = -1;
   }
   else /*-- dimensionality changing --*/
   {
      segment->order = (int *) krealloc(segment->order, sizeof(int) * dimension);

      /*-- needed for kimportasc : null out the index order --*/
      for (i = 0; i < dimension; i++)
	 segment->order[i] = -1;
   }

   /*-- assign the new dimension --*/
   segment->dimension = dimension;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_set_size - set the segment size
|
|       Purpose: Set the size of a segment.  
|
|		 The dimensionality of the segment must be 
|		 set before the size can be set.
|
|		 This routine assumes that the segment's size
|		 array is already initialized to the proper size.
|		 Note that setting the dimensionality is what 
|		 allocates the segment's size array.  
|
|         Input: segment - segment on which we will set the size
|		 size    - new size to set
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 11, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_set_size(
   kdas_segment *segment,
   int          *size)
{
   int nsize[KDAS_MAX_DIM] = KDAS_ONE_ARRAY;
   int uninitialized = FALSE;
   int big_resize;
   int onum;
   int nnum;
   int i;


   if (segment == NULL)
      return FALSE;

   if (segment->dimension == -1) /*-- dimension has not been set before --*/
   {
      kset_errno(KDAS_ESEG_UNKNOWN_DIMENSIONALITY);
      return FALSE;
   }

   /*----------------------------------------------------------------
   |   construct a new size array, ignoring any new sizes that are
   |   set to -1 (which means to use the old size) -- at the same
   |   time we'll figure out how many total numbers of elements are
   |   used by both the old and new size.
   \-----------------------------------------------------------------*/
   for (i = 0, onum = 1, nnum = 1; i < segment->dimension; i++)
   {
      /*-- ignore new sizes of -1 --*/
      nsize[i] = (size[i] == -1) ? segment->size[i] : size[i];

      /*-- figure out the old and new overall sizes --*/
      if (segment->size[i] > 0)
	 onum *= segment->size[i];

      /*-- keep track if any of the new size is still not initialized --*/
      if (nsize[i] != -1)
	 nnum *= nsize[i];
      else
	 uninitialized = TRUE; 
   }

   /*----------------------------------------------------------------
   |   resize only if *all* of the following conditions are true :
   |   [1] the overall size has changed 
   |   [2] the new size is completely initialized
   |   [3] our old presentation was completely initialized
   \-----------------------------------------------------------------*/
   if ((onum != nnum) && !uninitialized && 
       (segment->datatype != -1) &&
       kdas_size_initialized(segment->size, segment->dimension))
   {
      /*----------------------------------------------------------------
      |   if we are an input object, and we don't have data yet, use the
      |   find data routine and tell it we want to write to force us to 
      |   dup the data to a place where we can work on it
      \-----------------------------------------------------------------*/
      if (segment->format && (segment->format->flags & KOBJ_READ))
      {
	 if (segment->data == NULL && segment->temp == NULL)
	 {
	    kaddr           memory;
	    kdas_transport *transport;
	    
	    _kdas_find_data(segment, FALSE, &memory, &transport);
	 }
      }

      /*----------------------------------------------------------------
      |   if we actually have data in this segment, contained either
      |   in a file or in memory, then we should resize it
      \-----------------------------------------------------------------*/
      if (segment->data || segment->temp)
      {
	 /*-- will this segment get big or small after resize? --*/
	 big_resize = kdas_is_big(segment->dimension,nsize,segment->datatype);
      
	 /*-- getting big? --*/
	 if (big_resize && segment->data && segment->format->locate_segment)
	    kdas_segment_move_memory_to_temp(segment);
	 
	 /*-- segment data is all in memory --*/
	 if (segment->data)
	    segment->data = kdas_memory_resize(segment->data, 
					       segment->dimension,
					       segment->datatype, 
					       segment->size, nsize,
					       segment->rval, segment->ival);
      
	 /*-- segment data is in a tempfile --*/
	 else if (segment->temp != NULL)
	    segment->temp = kdas_temp_resize(segment->temp, segment->dimension,
					     segment->datatype, segment->size, 
					     nsize, segment->rval, 
					     segment->ival);
	 
	 /*-- getting small? --*/
	 if (!big_resize && segment->temp)
	    kdas_segment_move_temp_to_memory(segment);
      }
   }

   /*-- assign the new size to the segment's size array --*/
   kmemcpy(segment->size, nsize, sizeof(int) * segment->dimension);
   
   return TRUE;
}


/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_set_order - set the segment order
|
|       Purpose: Set the index order of a segment.  
|
|		 The dimensionality of the segment must be 
|		 set before the order can be set.
|
|		 This routine assumes that the segment's order
|		 array is already initialized to the proper size.
|		 Note that setting the dimensionality is what 
|		 allocates the segment's order array.  
|
|         Input: segment - segment on which we will set the order
|		 order   - new order to set
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 11, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_set_order(
   kdas_segment *segment,
   int          *order)
{
   int  norder[KDAS_MAX_DIM] = KDAS_ONE_ARRAY;
   int  uninitialized = FALSE;
   int  order_changing = FALSE;
   int *nsize = NULL;
   int  i;


   if (segment == NULL)
      return FALSE;

   if (segment->dimension == -1) /*-- dimension has not been set before --*/
   {
      kset_errno(KDAS_ESEG_UNKNOWN_DIMENSIONALITY);
      return FALSE;
   }

   /*----------------------------------------------------------------
   |   construct a new order array, ignoring any new orders that are
   |   set to -1 (which means to use the old order) 
   \-----------------------------------------------------------------*/
   for (i = 0; i < segment->dimension; i++)
   {
      /*-- ignore new orders of -1 --*/
      norder[i] = (order[i] == -1) ? segment->order[i] : order[i];
      
      /*-- see if any of the new order is still not initialized --*/
      if (norder[i] == -1)
	 uninitialized = TRUE; 

      else  /*-- is the index order being changed? --*/
	 if ((order[i] != -1) && 
	     (segment->order[i] != -1) &&
	     (segment->order[i] != norder[i]))
	    order_changing = TRUE;
   }
   

   /*----------------------------------------------------------------
   |   reorder the size array that so it is in the correct new order.
   \-----------------------------------------------------------------*/
   if (order_changing && !uninitialized)
   {
      nsize = (int *) kdupalloc(segment->size, sizeof(int)*segment->dimension);
      kdas_array_reorder(nsize, segment->dimension, segment->order, norder);
   }

   /*----------------------------------------------------------------
   |   reorder only if *all* of the following conditions are true :
   |   [1] the index order has changed 
   |   [2] the new order has been completely initialized
   |   [3] our old presentation was completely initialized
   \-----------------------------------------------------------------*/
   if (order_changing && !uninitialized &&
       (segment->datatype != -1) &&
       kdas_size_initialized(segment->size, segment->dimension) && 
       kdas_initialized(segment->order, segment->dimension))
   {
      /*----------------------------------------------------------------
      |   if we are an input object, and we don't have data yet, use the
      |   find data routine and tell it we want to write to force us to 
      |   dup the data to a place where we can work on it
      \-----------------------------------------------------------------*/
      if (segment->format && (segment->format->flags & KOBJ_READ))
      {
	 if (segment->data == NULL && segment->temp == NULL)
	 {
	    kaddr           memory;
	    kdas_transport *transport;
	    
	    _kdas_find_data(segment, FALSE, &memory, &transport);
	 }
      }

      /*----------------------------------------------------------------
      |   if we actually have data in this segment, contained either
      |   in a file or in memory, then we should consider reordering it
      \-----------------------------------------------------------------*/
      if (segment->data || segment->temp)
      {
	 /*-- segment data is all in memory --*/
	 if (segment->data)
	    segment->data = kdas_memory_reorder(segment->data, 
						segment->dimension,
						segment->datatype,
						segment->size, nsize,
						segment->order, norder);
	 
	 /*-- segment data is in a tempfile --*/
	 else if (segment->temp)    
	    segment->temp = kdas_temp_reorder(segment->temp, 
					      segment->dimension,
					      segment->datatype,
					      segment->size, nsize, 
					      segment->order, norder);
      }
   }

   /*-- assign the reordered size --*/
   if (order_changing && !uninitialized)
   {
      kfree(segment->size);
      segment->size = nsize;
   }

   /*-- assign the new order to the segment's order array --*/
   kmemcpy(segment->order, norder, sizeof(int) * segment->dimension);

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_set_datatype - set the segment datatype
|
|       Purpose: Set the datatype of a segment.  
|
|         Input: segment  - segment on which we will set the datatype
|		 datatype - new datatype to set
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 11, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_set_datatype(
   kdas_segment *segment,
   int           datatype)
{
   int big_recast = FALSE;
   

   if (segment == NULL)
      return FALSE;

   /*----------------------------------------------------------------
   |   recast only if *all* of the following conditions are true :
   |   [1] the datatype has changed
   |   [2] our old presentation was completely initialized
   \-----------------------------------------------------------------*/
   if ((datatype != -1) && (segment->datatype != datatype) &&
       (segment->datatype != -1) &&
       kdas_size_initialized(segment->size, segment->dimension))
   {
      /*----------------------------------------------------------------
      |   if we are an input object, and we don't have data yet, use the
      |   find data routine and tell it we want to write to force us to 
      |   dup the data to a place where we can work on it
      \-----------------------------------------------------------------*/
      if (segment->format && (segment->format->flags & KOBJ_READ))
      {
	 if (segment->data == NULL && segment->temp == NULL)
	 {
	    kaddr           memory;
	    kdas_transport *transport;

	    _kdas_find_data(segment, FALSE, &memory, &transport);
	 }
      }

      /*----------------------------------------------------------------
      |   if we actually have data in this segment, contained either
      |   in a file or in memory, then we should recast it
      \-----------------------------------------------------------------*/
      if (segment->data || segment->temp)
      {
	 /*-- will this segment get big or small after recast? --*/
	 big_recast = kdas_is_big(segment->dimension, segment->size, datatype);
	 
	 /*-- getting big? --*/
	 if (big_recast && segment->data && segment->format->locate_segment)
	    kdas_segment_move_memory_to_temp(segment);

	 /*-- segment data is all in memory --*/
	 if (segment->data)
	    segment->data = kdas_memory_recast(segment->data, 
					       segment->dimension,
					       segment->size, 
					       segment->datatype, datatype);
	 
	 /*-- segment data is in a tempfile --*/
	 else if (segment->temp)    
	    segment->temp = kdas_temp_recast(segment->temp, 
					     segment->dimension,
					     segment->size, 
					     segment->datatype, datatype);

	 /*-- getting small? --*/
	 if (!big_recast && segment->temp)
	    kdas_segment_move_temp_to_memory(segment);
      }
   }

   /*--  assign the new datatype to the segment's datatype --*/
   segment->datatype = datatype;
   
   return TRUE;
}


/****    =============================================================
 ****    Segment Data Utilities
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_move_memory_to_temp - move memory to a tempfile
|
|       Purpose: This routine will migrate the data contained in 
|		 segment->data to a tempfile.
|		
|		 The data will be freed.
|
|         Input: segment - the segment to operate on
|		 
|        Output: none
|
|       Returns: nothing
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
void
kdas_segment_move_memory_to_temp(kdas_segment *segment)
{
   kdas_transport *ntemp;
   int             position[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;


   /*-- create a new tempfile --*/
   ntemp = kdas_temp_construct(segment->dimension, segment->size, 
			       segment->datatype, segment->rval, 
			       segment->ival);

   /*-- put all the data in the tempfile --*/
   kdas_transport_clip(ntemp, segment->data, segment->dimension,
		       segment->datatype, segment->size, segment->size, NULL,
		       position, position, segment->size, KDAS_WRITE);

   kfree(segment->data);
   segment->data = NULL;
   segment->temp = ntemp;

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_move_temp_to_memory - move a tempfile to memory
|
|       Purpose: This routine will move the data contained in the
|		 segment->temp file and migrate it to a block of memory.
|
|		 The tempfile will be closed.
|
|         Input: segment - the segment to operate on
|
|        Output: none
|
|       Returns: nothing
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
void
kdas_segment_move_temp_to_memory(kdas_segment *segment)
{
   kaddr ndata;
   int   position[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;


   /*-- create a new memory block --*/
   ndata = kdas_memory_construct(segment->dimension, segment->size, 
				 segment->datatype, 0.0, 0.0);

   /*-- get all the data from the tempfile --*/
   kdas_transport_clip(segment->temp, ndata, segment->dimension, 
		       segment->datatype, segment->size, segment->size, NULL,
		       position, position, segment->size, KDAS_READ);

   kdas_temp_destruct(segment->temp);  
   segment->temp = NULL;
   segment->data = ndata;

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_get_data - retreive data block from data space
|
|       Purpose: This routine will clip a data block from a data space.
|		 The location of the data block is specified by a begin
|		 point and end point which are relative to the zero 
|		 origin of the data space.  The data block must be
|		 inside the bounds of the data space.  The data space 
|		 can have an "apparent size" different from its actual 
|		 size, in which case, the data space will either be
|		 padded, wrapped, or interpolated to reconcile the
|		 size difference.
|                         
|	                   DATA BLOCK 
|	                       |            __end
|	           (0,0)__     |           /
|	                  \    |          /      DATA SPACE
|	                   +-+-|-+-+-+---/-----+   of size "apparent_size"
|	                   | | V | | | |/| | | |
|	                   +-+-----+-+-/-+-+-+-+
|                          |/|*|*|*| |/| | | | |
|                   begin__/-|-+-+-|-/-+-+-+-+-+
|                          | |*|*|*|/| | | | | |
|                          +-+-----+/\-+-+-/\+-+
|                          | /\| | /  \/\|/  \ |
|                          |/  \/\/  :   V    \|
|                                    :
|                                    :
|
|		 The data block can be preallocated or can be allocated
|		 by this routine.  A pointer to the retrieved data block 
|		 will be returned by this routine.
|
|		 The begin and end points *should* be within the apparent
|		 size of the data.  This routine will not deal with the
|		 presentation of out-of-bounds data.
|
|         Input: segment - the segment to get data from
|                begin   - the first corner bounding the region to retrieve
|                end     - the end point bounding the region to retrieve
|	         apparent_size - the size the data space appears to be;
|			         pass NULL if it should appear to be the
|		 	         size it actually is.
|		 interpolate   - indicates how to resolve the size 
|			         difference between the segment size
|		 		 and the apparent size : KNONE, KPAD,
|				 KWRAP, KZERO_ORDER.
|		 rpad    - if padding, the real value to use
|		 ipad    - if padding, the imaginary value to use
|		 data    - the data block to use when returning the retrieved
|			   data; if NULL, then space will be allocated.
|        Output: none
|
|       Returns: a pointer to the data block being retrieved is returned
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
kaddr
kdas_segment_get_data(
   kdas_segment *segment, 
   int   	*begin, 
   int   	*end, 
   int   	*apparent_size, 
   int           interpolate,
   double        rpad,
   double        ipad,
   kaddr 	 data)  
{
   
   kaddr block;
   int   block_size[KDAS_MAX_DIM];
   int   clip_unit[KDAS_MAX_DIM];
   int   position_in_block[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int   position_in_data[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int   tmp[KDAS_MAX_DIM];
   int  *asize;
   int   clip;
   int   outside;
   int   i;

   /*-- sanity check --*/
   if (!begin || !end)
      return NULL;  /*-- kerror : invalid data access --*/

   /*-- figure out the overall size of the data being retrieved --*/
   for (i = 0; i < segment->dimension; i++)
      block_size[i] = end[i] - begin[i] + 1;

   /*-- if the destination data is NULL, then allocate space for it --*/
   if (data == NULL)
      block = kdas_memory_construct(segment->dimension, block_size, 
				    segment->datatype, 0.0, 0.0);
   else
      block = data;

   /*----------------------------------------------------------------
   |   determine the position of the unit we need to clip : 
   |      the unit's position will be given relative to both 
   |      the data space and data block.
   |
   |      we will clip this to the data's actual size, unless we are
   |      interpolating, then we will clip to the apparent size.
   \-----------------------------------------------------------------*/
   asize = (interpolate == KZERO_ORDER) ? apparent_size: segment->size;
   kdas_clip_corners(segment->dimension, asize, begin, end, &clip, &outside,
		     position_in_data, tmp, position_in_block, clip_unit);

   /*----------------------------------------------------------------
   |   if we've gone outside the data's actual size, we must deal with
   |   the out of bounds data ... we must either fail, pad, or wrap
   \-----------------------------------------------------------------*/
   if (clip)
   {
      if (interpolate == KNONE || interpolate == KZERO_ORDER)
      {
	 kfree(block);
	 return NULL;  /*-- not an error --*/
      }
      else if (interpolate == KWRAP)
      {
	 kdas_wrap_data(segment, segment->dimension, segment->size, 
			block_size, begin, _wrap_region, block);
	 return block;
      }
      else if (interpolate == KPAD)
	 kdas_preset_data(block, segment->dimension, block_size, 
			  segment->datatype, rpad, ipad);
   }

   /*-- get the in-bounds data --*/
   if (!outside)
      _kdas_segment_access(segment, apparent_size, block_size, clip_unit,
			   position_in_data, position_in_block, 
			   interpolate, KDAS_READ, block);
   return block;
}


/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_put_data - store data block in data space
|
|       Purpose: This routine will clip a data block into a data space.
|		 The location of the data block is specified by a begin
|		 point and end point which are relative to the zero 
|		 origin of the data space.  The data block must be
|		 inside the bounds of the data space.  The data space 
|		 can have an "apparent size" different from its actual 
|		 size, in which case, the data will either be clipped
|		 or interpolated to reconcile the size difference.
|                         
|	                   DATA BLOCK 
|	                       |            __end
|	           (0,0)__     |           /
|	                  \    |          /      DATA SPACE
|	                   +-+-|-+-+-+---/-----+   of size "apparent_size"
|	                   | | V | | | |/| | | |
|	                   +-+-----+-+-/-+-+-+-+
|                          |/|*|*|*| |/| | | | |
|                   begin__/-|-+-+-|-/-+-+-+-+-+
|                          | |*|*|*|/| | | | | |
|                          +-+-----+/\-+-+-/\+-+
|                          | /\| | /  \/\|/  \ |
|                          |/  \/\/  :   V    \|
|                                    :
|                                    :
|
|		 If the region being put is out of bounds of the
|		 apparent size of the data, it will be clipped.
|
|         Input: segment - the segment to put data into
|                begin   - the first corner bounding the region being stored
|                end     - the end point bounding the region begin stored
|	         apparent_size - the size the data space appears to be;
|			         pass NULL if it should appear to be the
|		 	         size it actually is.
|		 interpolate   - indicates how to resolve the size 
|			         difference between the segment size
|		 		 and the apparent size : KNONE, KPAD,
|				 KWRAP, KZERO_ORDER.
|		 rpad    - if padding, the real value to use
|		 ipad    - if padding, the imaginary value to use
|		 data    - the data block being put
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: May 23, 1994
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_put_data(
   kdas_segment *segment, 
   int   	*begin, 
   int   	*end, 
   int   	*apparent_size, 
   int           interpolate,
   kaddr 	 data)  
{
   int   block_size[KDAS_MAX_DIM];
   int   clip_unit[KDAS_MAX_DIM];
   int   position_in_block[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int   position_in_data[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int   tmp[KDAS_MAX_DIM];
   int  *asize;
   int   clip;
   int   outside;
   int   i;


   /*-- sanity check --*/
   if (!begin || !end)
      return FALSE;  /*-- kerror : invalid data access --*/

   /*-- figure out the overall size of the data being retrieved --*/
   for (i = 0; i < segment->dimension; i++)
      block_size[i] = end[i] - begin[i] + 1;

   /*-- if we have no apparent size, then we appear to be the size we are --*/
   if (apparent_size == NULL)
      apparent_size = segment->size;

   /*----------------------------------------------------------------
   |   determine the position of the unit we need to clip : 
   |      the unit's position will be given relative to both 
   |      the data space and data block.
   |
   |      we will clip to the data's actual size, unless we are
   |      interpolating, then we will clip to the apparent size.
   \-----------------------------------------------------------------*/
   asize = (interpolate == KZERO_ORDER) ? apparent_size: segment->size;
   kdas_clip_corners(segment->dimension, asize, begin, end, &clip, &outside,
		     position_in_data, tmp, position_in_block, clip_unit);

   /*-- clip the block into the segment --*/
   if (!outside)
      return _kdas_segment_access(segment, 
				  apparent_size, block_size, clip_unit, 
				  position_in_data, position_in_block, 
				  interpolate, KDAS_WRITE, data);
   else
      return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_segment_write
|
|       Purpose: This routine will write a data segment
|		 to a file at the location indicated by the
|		 format segment locator.  
|
|		 This routine should only be called when a 
|		 data object with format support for large
|		 data sets is being closed.
|
|         Input: segment - the segment to write out
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: November 21, 1995
| Modifications:
|
------------------------------------------------------------*/
int
kdas_segment_write(kdas_segment *segment)
{
   kdas_transport orig;

   kaddr block = NULL;

   int unit    [KDAS_MAX_DIM];
   int position[KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;
   int zero    [KDAS_MAX_DIM] = KDAS_ZERO_ARRAY;


   /*-- make sure we have everything we need --*/
   if (!segment || !segment->format || !segment->format->locate_segment)
   {
      /*-- error - trying to write segment with no support --*/
      return FALSE;
   }

   /*-- construct transport for original file --*/
   orig.reader = segment->format->reader;
   orig.writer = segment->format->writer;
   orig.seeker = segment->format->seeker;
   orig.teller = segment->format->teller;
   orig.id     = segment->format->locate_segment(segment->object, 
						 (char *)segment->id);

   /*-- if data is all in a temporary file then migrate it out --*/
   if (segment->temp) 
   {
      /*-- get the optimal region size to move data with --*/
      kdas_optimal_region_size(segment->dimension, segment->size, 
			       segment->datatype, unit);

      block = kdas_memory_construct(segment->dimension, unit,
				    segment->datatype, 0.0, 0.0);
      do
      {
	 /*-- read from the old transport --*/
	 kdas_transport_clip(segment->temp, block, segment->dimension, 
			     segment->datatype, segment->size, unit, NULL,
			     position, zero, unit, KDAS_READ);

	 /*-- write into the original transport --*/
	 kdas_transport_clip(&orig, block, segment->dimension, 
			     segment->datatype, segment->size, unit, NULL,
			     position, zero, unit, KDAS_WRITE);
      }
      while (kdas_advance_position(segment->dimension, segment->size, 
				   unit, position));
      kfree(block);
   }

   /*-- if data is all in memory, then just write it out --*/
   else if (segment->data)
      kdas_transport_clip(&orig, segment->data, segment->dimension, 
			  segment->datatype, segment->size, segment->size, 
			  NULL, zero, zero, segment->size, KDAS_WRITE);
   else
   {
      /*-- error  where's the data? --*/
      return FALSE;
   }
   
   return TRUE;
}

/****    =============================================================
 ****    Static Data Access Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdas_segment_access
|
|       Purpose: This routine will read or write a clip unit
|		 in a data segment transfering it either to
|		 or from a block of data.  The clip unit can
|		 be an arbitrarily positioned subspace in both 
|		 the segment, as well as in the block of data.  
|
|		 The segment may also have an apparent size which 
|		 is different from its actual size for use 
|		 with interpolated data clipping.  If interpolated
|		 then the data will appear to be clipped from the
|		 bigger interpolated data space.  If interpolation
|		 is not desired, then the apparent size will be 
|		 ignored.
|
|                         
|	                   DATA BLOCK 
|	        block is of :  |            
|	          "block size" |            
|	                       |               SEGMENT DATA SPACE
|	                   +-+-|-+-+-+-+-+-+-+-+   segment is of : 
|	                   | | V | | | | | | | |     "apparent_size"
|	                   +-+---------+-+-+-+-+
|                          | |*|*|*|*|*| | | | |
|                          +-/- - - - -|-+-+-+-+
|                          |/|*|*|*|*|*| | | | |
|                          /-+------/\-+-+-/\+-+
|      position in space__/| /\| | /  \/\|/  \ |
|                  (1,1)   |/  \/\/  :   V    \|
|                                    :
|                                    :
|
|         Input: segment       - the segment to access data from
|		 apparent_size - the apparent size of the segment
|		 block_size    - the overall size of the data block
|		 clip_unit     - the size of the data to clip
|		 position_in_data  - the position of clip unit in segment space
|		 position_in_block - the position of clip unit in block space
|		 interpolate   - at this point, indicates if we should
|				 do zero order interpolation or not.
|		 reading       - TRUE if data should be read from the segment
|		  		 into the block, FALSE if the data should be
|			 	 written from the block into the segment.
|		 block         - preallocated block of memory to clip either
|			         into or out of.
|
|        Output: data is clipped either into or out of block
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: June 5, 1995
| Modifications:
|
------------------------------------------------------------*/
static int
_kdas_segment_access(
   kdas_segment *segment,
   int  	*apparent_size,
   int  	*block_size,
   int          *clip_unit,
   int  	*position_in_data,
   int  	*position_in_block,
   int           interpolate,
   int  	 reading,
   kaddr	 block)
{
   kaddr           memory;
   kdas_transport *transport;
   int             access_type;
   

   /*----------------------------------------------------------------
   |   we will either clip or interpolate the data from memory or
   |   transport depending on our interpolation type.   try to 
   |   avoid interpolating even if it says we are supposed to 
   |   if our actual size is the same as our apparent size.
   \-----------------------------------------------------------------*/
   if (interpolate == KZERO_ORDER)
   {
      if (kmemcmp(apparent_size,segment->size,sizeof(int)*segment->dimension))
	 access_type = 1;
      else
	 access_type = 0;
   }
   else
      access_type = 0;
   
   /*----------------------------------------------------------------
   |   find the data : it will be either in memory, or in a transport
   \-----------------------------------------------------------------*/
   _kdas_find_data(segment, reading, &memory, &transport);

   /*-- segment data is all in memory --*/
   if (memory)
      memory_access[access_type](memory, block, segment->dimension, 
				 segment->datatype, segment->size,
				 block_size, apparent_size,
				 position_in_data, position_in_block,
				 clip_unit, reading);
   
   /*-- segment data is all in a temporary transport --*/
   else if (transport)
      transport_access[access_type](transport, block, segment->dimension, 
				    segment->datatype, segment->size,
				    block_size, apparent_size,
				    position_in_data, position_in_block, 
				    clip_unit, reading);
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdas_find_data
|
|       Purpose: This routine finds where the data is located
|		 in a segment.  
|
|		 The data can be in one of three locations :
|	
|		 [1] all in memory
|		 [2] all in a temp file
|		 [3] all in the original file
|	
|		 If the data is small enough to fit in memory,
|		 and isn't in memory yet, then this routine will 
|		 read the data into memory.
|
|		 If the data is large, and we are reading, then 
|		 this routine will return a transport which
|		 is set up to the beginning of the segment in 
|		 the original file.
|
|		 If the data is large, and we are writing, then
|		 this routine will set up a temporary transport
|		 for us to write to.  If we are writing to an 
|		 input object, then the temporary transport will
|		 be a copy of the original data segment.
|
|		 This routine maintains the temp and data fields
|		 of the data segment, so that later accesses
|		 will know where the data is located.
|
|		 This routine will return either a memory 
|		 location or a transport in which the data can
|		 be found.  The one where the data is not will
|		 be returned as NULL.
|
|         Input: segment       - the segment to access data from
|		 reading       - TRUE if we are reading data from the segment,
|		  		 FALSE if we are writing data to the segment.
|        Output: memory        - the memory with the data or NULL
|		 transport     - the transport with the data or NULL
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: October 31, 1995
| Modifications:
|
------------------------------------------------------------*/
static int
_kdas_find_data(
   kdas_segment    *segment,
   int              reading,
   kaddr           *memory,
   kdas_transport **transport)
{
   kdas_transport *trans = segment->temp;

   /*----------------------------------------------------------------
   |   if we don't have the data already in memory or in a temp file,
   |   then we've need to perform our initialization
   \-----------------------------------------------------------------*/
   if (!segment->data && !segment->temp)
   {
      /*----------------------------------------------------------------
      |   initialize MEMORY if either the segment will fit entirely 
      |   into memory, or if we have no large data set support.
      |
      |   if there is no locate segment function, then this segment
      |   has probably been created, and we need to allocate space
      |	  whether we are reading or not.
      \-----------------------------------------------------------------*/
      if (!kdas_is_big(segment->dimension, segment->size, segment->datatype)
	  || !segment->format->locate_segment)
      {
	 /*-- input object -- */
	 if ((segment->format->flags & KOBJ_READ) &&
	      segment->format->locate_segment)  
	    segment->data = _kdas_read_all_data(segment);
	 
	 /*-- output object --*/
	 else 
	    segment->data = kdas_memory_construct(segment->dimension,
						  segment->size, 
						  segment->datatype,
						  segment->rval,
						  segment->ival);
      }

      /*----------------------------------------------------------------
      |   initialize TRANSPORT if we have a large data set and the format
      |   actually supports the reading of large data sets
      \-----------------------------------------------------------------*/
      else 
      {
	 /*-- input object : may be orig, so don't assign to segment->temp --*/
	 if ((segment->format->flags & KOBJ_READ) &&
	      segment->format->locate_segment)  
	    trans = (kdas_transport *) _kdas_find_big_data(segment, reading); 

	 /*-- output object --*/
	 else
	    trans = segment->temp = kdas_temp_construct(segment->dimension,
							segment->size,
							segment->datatype,
							segment->rval,
							segment->ival);
      }
   }

   /*-- return where the data is : one of these will be NULL --*/
   *memory    = segment->data;
   *transport = trans;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdas_read_all_data
|
|       Purpose: This routine will read all the data for a
|		 a segment from an original file using the 
|		 transport clipper.  It is assumed that the
|		 segment's format has support for locating
|		 the segment in the original file.
|
|         Input: segment - the segment to access data from
|
|        Output: none
|
|       Returns: the allocated memory on success, NULL on failure
|
|    Written By: Steve Kubica
|          Date: October 31, 1995
| Modifications:
|
------------------------------------------------------------*/
static kaddr
_kdas_read_all_data(kdas_segment *segment)
{
   kdas_transport orig;
   kaddr data;
   int   zero[KDAS_MAX_DIM] = {0, 0, 0, 0, 0};
   int   fid;


   /*-- could have been created : make sure it really is in original file --*/
   fid = segment->format->locate_segment(segment->object, (char *)segment->id);

   /*-- construct memory for new segment --*/
   data = kdas_memory_construct(segment->dimension, segment->size,
				segment->datatype, segment->rval,
				segment->ival);
   
   /*-- if segment was just created, just return new memory --*/
   if (fid == -1)
      return data;
   
   /*-- construct transport for original file --*/
   orig.reader = segment->format->reader;
   orig.writer = segment->format->writer;
   orig.seeker = segment->format->seeker;
   orig.teller = segment->format->teller;
   orig.id     = fid;

   /*-- read from the original file --*/
   kdas_transport_clip(&orig, data, segment->dimension, segment->datatype, 
		       segment->size, segment->size, NULL, zero, zero,
		       segment->size, KDAS_READ);
   return data;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdas_find_big_data
|
|       Purpose: This routine will return the transport
|		 needed to read data for the segment using the
|		 transport clipper.  It is assumed that the
|		 segment's format has support for locating
|		 the segment in the original file.  If the
|		 data is still in the original file and we
|		 are reading, temporarily return the original
|		 transport.  Otherwise, construct a tranport
|		 and return that.  In the event that we
|		 contsruct a transport, assign it internally
|		 to the segment->temp pointer so we have
|		 access to it later.
|
|         Input: segment - the segment to access data from
|		 reading - TRUE if we are reading
|
|        Output: none
|
|       Returns: the transport on success, NULL on failure
|
|    Written By: Steve Kubica
|          Date: October 31, 1995
| Modifications:
|
------------------------------------------------------------*/
static kaddr
_kdas_find_big_data(kdas_segment *segment, int reading)
{
   kdas_transport       *trans;
   static kdas_transport original;
   int                   fid;

   /*-- could have been created : make sure it really is in original file --*/
   fid = segment->format->locate_segment(segment->object, (char *)segment->id);

   /*-- if segment was just created, create new memory --*/
   if (fid != -1)
   {
      /*-- construct transport for original file --*/
      original.reader = segment->format->reader;
      original.writer = segment->format->writer;
      original.seeker = segment->format->seeker;
      original.teller = segment->format->teller;
      original.id     = fid;
      
      /*-- if writing, make copy of input to a temporary file --*/
      if (!reading)
	 trans = segment->temp = kdas_temp_copy(&original, 
						segment->dimension, 
						segment->datatype, 
						segment->size);
      else
	 trans = &original; /*-- static, so return it temporarily --*/
   }
   else
      trans = segment->temp = kdas_temp_construct(segment->dimension, 
						  segment->size, 
						  segment->datatype,
						  segment->rval, 
						  segment->ival);
   return trans;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _wrap_region
|
|       Purpose: This routine is used with kdas_wrap_data to 
|		 extract a region from the segment and insert
|		 it into a data block.
|
|    Written By: Steve Kubica
|          Date: January 6, 1996
| Modifications:
|
------------------------------------------------------------*/
static void
_wrap_region(
   kaddr source,
   int  *data_size,
   int  *block_size,
   int  *sb,
   int  *se,
   int  *db,
   int  *de,
   int  *unit,
   kaddr destination)
{
   kdas_segment *segment = (kdas_segment *) source;

   /*-- fill out this piece of the block --*/
   _kdas_segment_access(segment, data_size, block_size, unit, sb, db, 
			KNONE, KDAS_READ, destination);
   return;
}
