/*
  * Khoros: $Id: wrap.c,v 1.4 1997/01/20 04:56:40 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: wrap.c,v 1.4 1997/01/20 04:56:40 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Pipeline Code
   >>>>		
   >>>>  Private:
   >>>>                 kdas_construct_partitions()
   >>>>                 kdas_wrap_data()
   >>>>   Static:
   >>>>                 _kdas_add_partition()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

static void _kdas_add_partition PROTO((int **, int *, int));


/****    =============================================================
 ****    Partition Code
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdas_construct_partitions
|
|       Purpose: This routine will construct partition arrays which
|		 provide the partition points dividing an N dimensional
|		 region up according to some size data set.  The data
|		 block is specified in space relative to the data set.
|
|		 This example illustrates a partitioning in 2D :
|
|                  -5     0       5      10      15 17 ------> 1st dimension
|		 -3 +-----+-------+---------------+-+
|                   |     |       |       |       | | 
|                 0 + - - +-------+ - - - + - - - +-+
|                   |     |\\\\\\\|       |       | |    \\\\ actual data
|                   |     |\\\\\\\|                 |
|                   |     |\\\\\\\|       |       | |
|                   |     |\\\\\\\|                 |
|                 6 + - - +-------+ - - - + - - - +-+
|                   |     |       |\      |       | |
|		  9 +---------------\-------------+-+
|                                    \__size = (5,6)
|		  |
|		  V 2nd dimension
|		 
|
|                The partition arrays will be allocated for you,
|		 you must free them later.
|
|         Input: dimension  - the dimensionality we are partitioning
|		 size       - the size of the actual data
|		 begin      - the begin point of the block being acquired
|		 block_size - the size of the block being acquired
|
|        Output: number     - the number of partitions
|
|       Returns: the constructed partition array 
|
|    Written By: Steve Kubica and Jeremy Worley
|          Date: October 31, 1994
| Modifications:
|
------------------------------------------------------------*/
int **
kdas_construct_partitions(
   int    dimension,
   int   *size,
   int   *begin,
   int   *block_size,
   int  **number)
{
   int **partitions;
   int   done;
   int   position;
   int   i;

   
   /*-- sanity checking --*/
   if (number == NULL)
      return FALSE;

   /*----------------------------------------------------------------
   |   construct the partition arrays : 
   |     for each dimension, we will partition it into sections which
   |     are the size of the overall data set size in that dimension.  
   |   
   |   the 'n' array will contain the number of partition points
   |   the 'partitions' array will contain the partition bounds
   |
   |   example partitioning for a dimension size 5 with a region {-10, 15} :
   |       paritions[i] = { -10 5 0 5 10 15 }   
   |       number[i] = 6
   \-----------------------------------------------------------------*/
   *number = (int *) kmalloc(sizeof(int) * dimension);
   partitions = (int **) kmalloc(sizeof(int *) * dimension);
   
   /*-- initialize the partition arrays --*/
   for (i = 0; i < dimension; i++)
   {
      int end = begin[i] + block_size[i] - 1;

      (*number)[i] = 0;
      partitions[i] = NULL;

      /*-- initial partition --*/
      _kdas_add_partition(&(partitions)[i], &(*number)[i], begin[i]);

      /*-- find the first aligned position --*/
      position = size[i] * (int)(begin[i]/size[i] + ((begin[i]>0) ? 1 : 0));

      if (position <= end)
	 _kdas_add_partition(&(partitions)[i], &(*number)[i], position);

      done = FALSE;
      do
      {
	 if (position + size[i] > end)
	 {
	    position = end + 1;
	    done = TRUE;
	 }
	 else
	    position += size[i];

	 _kdas_add_partition(&(partitions)[i], &(*number)[i], position);

      } while (!done);
   }
   
   return partitions;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdas_add_partition
|
|       Purpose: This utility routine will add a number to a 
|		 partition array, resizing the array and 
|		 incrementing the number of array elements.
|
|         Input: partition    - pointer to a array to add partition
|	         number       - the number of partition points
|		 p            - the partition value to add
|
|        Output: partition - the partition array is grown in place
|		 number    - reflects the eventual size of the partition array
|
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: October 27, 1994
| Modifications:
|
------------------------------------------------------------*/
static void
_kdas_add_partition(
   int **partition, 
   int  *number, 
   int   p)
{
   *number = *number + 1;
   *partition = (int *) krealloc(*partition, (size_t)(*number*sizeof(int)));
   (*partition)[*number-1] = p;

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdas_wrap_data
|
|       Purpose: This routine will fill in both out-of-bounds data
|		 and in-bounds data in a data block wrapping the
|		 data appropriately.
|
|         Input: source      - kaddr client data with source information
|		 dimension   - the dimensionality of the data space
|		 data_size   - the size of the data space
|		 block_size  - the size of the block to fill out
|		 begin       - the begin point of the in-bounds data access
|		 fill_region - function to use for transfering region from
|			       the source to the destination
|
|        Output: destination - kaddr client data with destination information 
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: January 7, 1996
| Modifications:
|
------------------------------------------------------------*/
void
kdas_wrap_data(
   kaddr  source,
   int    dimension,
   int   *data_size,
   int   *block_size,
   int   *begin,
   void (*fill_region) PROTO((kaddr, int *, int *, int *, int *, int *, 
			      int *, int *, kaddr)),
   kaddr  destination)
{
   int **partitions = NULL;
   int  *n = NULL;
   int  *index;
   int  *one;
   int  *unit;
   int  *nsize;
   
   int *db, *de;  /*-- destination begin and end --*/
   int *sb, *se;  /*-- source begin and end --*/
   int i;


   /*-- allocate working arrays --*/
   index = (int *) kmalloc(sizeof(int) * dimension);
   one   = (int *) kmalloc(sizeof(int) * dimension);
   db    = (int *) kmalloc(sizeof(int) * dimension);
   de    = (int *) kmalloc(sizeof(int) * dimension);
   sb    = (int *) kmalloc(sizeof(int) * dimension);
   se    = (int *) kmalloc(sizeof(int) * dimension);
   unit  = (int *) kmalloc(sizeof(int) * dimension);
   nsize = (int *) kmalloc(sizeof(int) * dimension);
   
   /*----------------------------------------------------------------
   |   construct the partition arrays
   \-----------------------------------------------------------------*/
   partitions = kdas_construct_partitions(dimension, data_size, begin, 
					  block_size, &n);

   for (i = 0; i < dimension; i++)
   {
      nsize[i] = n[i]-1;
      index[i] = 0;
      one[i]   = 1;
   }


   /*----------------------------------------------------------------
   |   traverse the partition arrays :
   |      we will visit each region bounded in n dimensions by the 
   |      points specified along each dimensionension in the partition 
   |      arrays.   
   |
   |   for each region : 
   |      determine corners in the destination space
   |      determine corresponding corners in the source space
   |      shift the destination corners to the data space of the destination
   |   - and finally - 
   |      get data from the source, and put in the destination
   |	  using the handy user provided function
   \-----------------------------------------------------------------*/
   do
   {
      for (i = 0; i < dimension; i++)
      {
	 /*-- compute region in destination --*/
	 db[i] = partitions[i][ index[i] ];
	 de[i] = partitions[i][ (index[i]+1) % n[i] ] - 1;

	 /*-- determine corresponding region in source --*/
	 if (db[i] < 0)
	 {
	    sb[i] = (db[i] + ((kabs(db[i])/data_size[i])+1) * data_size[i]) 
	                                                    % data_size[i];
	    se[i] = sb[i] + (de[i] - db[i]);
	 }
	 else 
	 {
	    se[i] = de[i] % data_size[i];
	    sb[i] = se[i] - (de[i] - db[i]);
	 }

	 /*-- shift destination region relative to destination space --*/
	 db[i] -= begin[i];
	 de[i] -= begin[i];

	 unit[i] = de[i] - db[i] + 1;
      }

      /*-- fill out this piece using the user provided function --*/
      fill_region(source, data_size, block_size, sb, se, db, de, unit, 
		  destination);
      
   } while (kdas_advance_position(dimension, nsize, one, index));

   /*-- clean up the partitions array --*/
   for (i = 0; i < dimension; i++)
      kfree(partitions[i]);
   kfree(partitions);

   kfree(n);
   kfree(nsize);
   kfree(index);
   kfree(one);
   kfree(unit);

   kfree(db);   kfree(de);
   kfree(sb);   kfree(se);

   return;
}
