 /*
  * Khoros: $Id: callbacks.c,v 1.15.2.2 1997/03/11 17:16:16 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: callbacks.c,v 1.15.2.2 1997/03/11 17:16:16 mirage Exp $";

#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Callback Routines
   >>>>		
   >>>>  Private:
   >>>>                 kdms_disable_callbacks()
   >>>>                 kdms_enable_callbacks()
   >>>>                 kdms_process_data_callbacks()
   >>>>                 kdms_process_object_callbacks()
   >>>>                 kdms_delete_callbacks_for_object()
   >>>>   Static:
   >>>>                 _kdms_callback_scale_coord()
   >>>>   Public:
   >>>>                 kdms_add_callback()
   >>>>                 kdms_remove_callback()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

static void _kdms_callback_scale_coord PROTO ((int *, kposition *, int *,
					       int *, int *, int *, int, int));

static int _kdms_disable_callbacks = FALSE;

/****    =============================================================
 ****    Public Callback Routines 
 ****    =============================================================
 ****/

/************************************************************
*
*  Routine Name: kdms_add_callback - add a callback associated
*                with an object's data or attribute.
*
*       Purpose: This function is used to add a callback on a 
*                segment in a data object so that operations
*                can be performed and monitored in an event
*                driven environment.
*
*                A callback is a mechanism for operating on a data
*                object whenever an asynchronous event occurs rather
*                than in a strictly sequential manner.  This function
*                is particularly useful in conjunction with
*                kdms_reference.  The callbacks operate on the data,
*                which is shared between referenced objects.  Thus
*                object A can set a callback such that when data get
*                changed via object B, operations can be performed on
*                object A.
*
*                The object and segment arguments are used to specify
*                which segment and object will have the callback added
*                to.
*
*                The type argument specifies what kind of callback
*                should be added.  This argument may take on the
*                following values:
*
*       !       KDMS_CALLBACK_CHANGE  - generate a callback whenever data 
*       !                            	on the specified segment is changed 
*       !                            	(via kdms_put_data or kdms_copy_data).
*       !       KDMS_CALLBACK_ACCESS  - generate a callback whenever data
*       !                            	on the specified segment is accessed
*       !                            	(via kdms_get_data or kdms_copy_data).
*       !       KDMS_CALLBACK_DELETE  - generate a callback whenever the
*       !                            	segment specified is about to be
*       !                            	deleted, either through a 
*	!			     	kdms_destroy_segment() call, or 
*	!			     	via a kdms_close().
*       !       KDMS_CALLBACK_DESTROY - generate a callback whenever the
*       !                            	last instance of an object or 
*	!				segment is about to be closed.
*       !       KDMS_CALLBACK_SAVE    - generate a callback whenever the
*       !                            	specified segment is about to be
*       !                            	changed.  This is similar to the
*       !                            	KDMS_CALLBACK_CHANGE, except that
*       !                            	the callback is generated before
*       !                            	the data is changed rather than
*       !                            	afterward.
*
*		 If the segment specified is NULL, then the callback
*		 is placed on the entire object.  Currently, only
*		 DELETE and DESTROY callbacks can be placed on an
*		 object.
*
*                The callback mechanims in Data Services are analogous
*                to the callback mechanisms that are available in the
*                Xvwidget library or in Xt.  The motivation for
*                callbacks in Data Services is that they facilitate
*                functionality in the Xvisual library, where they allow
*		 different visual objects tied to a single data object
*		 to communicate with each other.
*
*         Input: object     - The object to add callback to.
*                segment    - The segment in object to add callback to.
*                             If NULL, then it is considered an object
*			      callback. 
*                type       - The type of callback to add.
*                callfunc   - The function to call when the callback
*                             occurs.
*                clientData - Any clientdata to be passed to the callfunc.
*
*        Output:
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: Only KDMS_CALLBACK_DELETE and KDMS_CALLBACK_DESTROY 
*		 callbacks work for objects.
*    Written By: Jeremy Worley
*          Date: April 21, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_add_callback(
   kobject     object,
   char       *segment,
   char       *type,
   kfunc_void  callfunc,
   kaddr       clientData)
{
   kcallback *cb;
   
   KDMS_INIT;

   /*-- sanity check --*/
   if (object == NULL || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   /*-- only DELETE and DESTROY callbacks can be used at the object level --*/
   if ((kstrcmp(type, KDMS_CALLBACK_DELETE ) == 0 && segment != NULL) || 
       (kstrcmp(type, KDMS_CALLBACK_DESTROY) == 0 && segment != NULL))
      return FALSE;

   /*-- manifest new callback --*/ 
   if ((cb = (kcallback *) kcalloc((size_t) 1, sizeof(kcallback))) == NULL)
      return FALSE;
   
   cb->object  	  = object;
   cb->segment 	  = kstring_to_token(segment);
   cb->clientData = clientData;
   cb->type   	  = kstring_to_token(type);
   
   if (segment == NULL)
      cb->callback.object = (void(*)(kobject, char*, void*))callfunc;  /*-- object callback --*/
   else
      cb->callback.data   = (void(*)(kobject, char*, void*, kdms_callback*))callfunc;  /*-- segment callback --*/


   /*-- add new cb to the object --*/
   if (segment != NULL)
   {
      kdas_presentation *pres;
      
      if ((pres = kdms_get_segment_info(object, segment)) == NULL)
      {
	 kdas_set_error(KDMS_ESEG_NONEXIST);
	 return FALSE;
      }

      pres->segment->callbacks = klist_add(pres->segment->callbacks, cb, NULL);
#if 0
      kprintf("\nAdded Callback with the following info\n");
      kprintf("    Segment Name    '%s'\n", segment);
      kprintf("    Segment Address '%x'\n", pres);
      kprintf("    Type of callback '%s'\n", type);
      kprintf("    Routine '%x'\n", callfunc);
      kprintf("\n");
#endif
   }
   else
   {
      object->phys->callbacks = klist_add(object->phys->callbacks, cb, NULL);
#if 0
      kprintf("\nAdded Object with the following info\n");
      kprintf("    Object File '%s'\n", object->phys->filename);
      kprintf("    Object Address '%x'\n", object);
      kprintf("    Type of callback '%s'\n", cb->type);
      kprintf("    Routine '%x'\n", callfunc);
      kprintf("\n");
#endif
   }
   

   return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_remove_callback - remove a callback associated
*                with an object's data or attribute.
*
*       Purpose: This function is used to remove a callback that was
*                previously added on a segment in a data object so that 
*                operations can be performed and monitored in an event
*                driven environment.
*
*                This function will not remove a callback unless all
*                of the arguments that are passed to it are the same
*                as those passed to the kdms_create_callback
*                previously.  This allows for multiple callbacks that
*                are similar in nature to be removed without
*                confusion.
*
*                A callback is a mechanism for operating on a data
*                object whenever an asynchronous event occurs rather
*                than in a strictly sequential manner.  This function
*                is particularly useful in conjunction with
*                kdms_reference.  The callbacks operate on the data,
*                which is shared between referenced objects.  Thus
*                object A can set a callback such that when data get
*                changed via object B, operations can be performed on
*                object A.
*
*                The object and segment arguments are used to specify
*                which segment and object will have the callback added
*                to.
*
*                The type argument specifies what kind of callback
*                should be added.  This argument may take on the
*                following values:
*
*       !       KDMS_CALLBACK_CHANGE  - generate a callback whenever data 
*       !                            	on the specified segment is changed 
*       !                            	(via kdms_put_data or kdms_copy_data).
*       !       KDMS_CALLBACK_ACCESS  - generate a callback whenever data
*       !                            	on the specified segment is accessed
*       !                            	(via kdms_get_data or kdms_copy_data).
*       !       KDMS_CALLBACK_DELETE  - generate a callback whenever the
*       !                            	segment specified is about to be
*       !                            	deleted, either through a 
*	!			     	kdms_destroy_segment() call, or 
*	!			     	via a kdms_close().
*       !       KDMS_CALLBACK_DESTROY - generate a callback whenever the
*       !                            	last instance of an object or
*	!				segment is about to be closed.
*       !       KDMS_CALLBACK_SAVE    - generate a callback whenever the
*       !                            	specified segment is about to be
*       !                            	changed.  This is similar to the
*       !                            	KDMS_CALLBACK_CHANGE, except that
*       !                            	the callback is generated before
*       !                            	the data is changed rather than
*       !                            	afterward.
*
*		 If the segment specified is NULL, then the callback
*		 is placed on the entire object.  Currently, only
*		 DELETE and DESTROY callbacks can be placed on an
*		 object.
*
*                The callback mechanims in Data Services are analogous
*                to the callback mechanisms that are available in the
*                Xvwidget library or in Xt.  The motivation for
*                callbacks in Data Services is that they facilitate
*                functionality in the Xvisual library, where they
*                allow different visual objects tied to a single
*                data object to communicate with each other.
*
*         Input: object     - object to add callback to.
*                segment    - segment in object to add callback
*                             to.
*                type       - type of callback to add.
*                callfunc   - function to call when callback
*                             occurs.
*                clientData - data to be passed to callfunc.
*
*        Output:
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: Only KDMS_CALLBACK_DELETE and KDMS_CALLBACK_DESTROY 
*		 callbacks work for objects.
*    Written By: Jeremy Worley
*          Date: April 21, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_remove_callback(
   kobject     object,
   char       *segment,
   char       *type,
   kfunc_void  callfunc,
   kaddr       clientData)
{
   kdas_presentation *lpres = NULL;
   kcallback *cb, *rcallback;
   klist     *l;
   ktoken     stoken;
   
   KDMS_INIT;

   /*-- sanity check --*/
   if (object == NULL || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   stoken = kstring_to_token(segment);

   /*-- get the callback list that it will be on --*/
   if (segment != NULL)
   {
      if ((lpres = kdms_get_segment_info(object, segment)) == NULL)
      {
	 kdas_set_error(KDMS_ESEG_NONEXIST);
	 return FALSE;
      }

      l = lpres->segment->callbacks;
   }
   else
      l = object->phys->callbacks;

   /*-- find the callback --*/
   for (l = klist_head(l), rcallback = NULL; l != NULL; l = klist_next(l))
   {
      cb = (kcallback *) klist_identifier(l);
      
      /*-- everything must match --*/
      if (cb->object  	    == object 	&&
	  cb->segment 	    == stoken 	&&
	  cb->callback.data == (void(*)(kobject, char*, void*, kdms_callback*))callfunc &&  /* union catches both */
	  cb->clientData    == clientData)
	 rcallback = cb;
   }

   /*-- remove this callback from the object --*/
   if (rcallback != NULL)
   {
      if (segment != NULL)
	 lpres->segment->callbacks = klist_delete(lpres->segment->callbacks,
						  rcallback);
      else
	 object->phys->callbacks = klist_delete(object->phys->callbacks,
						rcallback);
   }
   
   /*-- and finally free the callback  --*/
   kfree_and_NULL(rcallback);

   return TRUE;
}


/****    =============================================================
 ****    Callback Disable/Enable Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|  Routine Name: kdms_disable_callbacks
|       Purpose: External callback control to disbale callbacks
|		 from occuring.
|    Written By: Jeremy Worley
|          Date: October 27, 1993
------------------------------------------------------------*/
void
kdms_disable_callbacks(void)
{
   _kdms_disable_callbacks = TRUE;
   return;
}

/*-----------------------------------------------------------
|  Routine Name: kdms_enable_callbacks
|       Purpose: External callback control to enable callbacks
|		 to occur.
|    Written By: Jeremy Worley
|          Date: October 27, 1993
------------------------------------------------------------*/
void
kdms_enable_callbacks(void)
{
   _kdms_disable_callbacks = FALSE;
   return;
}

/****    =============================================================
 ****    Process Callbacks
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdms_process_data_callbacks
|
|       Purpose: Process all the callbacks of a given type on the
|		 provided data object.  For all callbacks, scale 
|		 the begin and end points into the space of the object 
|		 being called back on.
|
|         Input: object - object generating callback
|		 pres   - presentation generating callback
|		 begin  - begin point for data access relative to object
|		 end    - end point for data access relative to objecte
|		 type   - callback type
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: April 21, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_process_data_callbacks(
   kobject            object,
   kdas_presentation *pres,
   int               *begin,
   int               *end,
   char              *type)
{
   static int disable = FALSE;

   kdms_callback      callstruct;
   kdas_presentation *tpres;
   kcallback         *cb;
   klist             *l;
   ktoken             ttoken;
   int                cbegin[KDAS_MAX_DIM];
   int                cend  [KDAS_MAX_DIM];
   

   if (pres->segment->callbacks == NULL)
      return TRUE;  /*-- trivial case --*/

   ttoken = kstring_to_token(type);

   /*-- blindly disable callbacks while callbacks are being serviced --*/
   if (disable == TRUE || _kdms_disable_callbacks == TRUE)
      return (TRUE);

   disable = TRUE;

#if 0
   kprintf("\nIn process callback with the following info\n");
   kprintf("    Segment Name    '%s'\n", ktoken_to_string(pres->id));
   kprintf("    Segment Address '%x'\n", pres->segment);
   kprintf("    Number of callbacks '%d'\n", 
	   klist_size(pres->segment->callbacks));
   kprintf("\n");
#endif

   callstruct.segment = ktoken_to_string(pres->id);
   
   /*---------------------------------------------------------------
   |  callbacks may be reentrant, so keep our own copy of these
   |  so that we don't lose track of our begin and end points as
   |  we cycle through all the callbacks for this presentation. 
   \---------------------------------------------------------------*/
   kmemcpy(cbegin, begin, sizeof(int) * pres->dimension);
   kmemcpy(cend,   end,   sizeof(int) * pres->dimension);
   
   /*---------------------------------------------------------------
   |  run through all the callbacks for this segment...
   \---------------------------------------------------------------*/
   for (l = klist_head(pres->segment->callbacks); l != NULL; l = klist_next(l))
   {
      cb = (kcallback *) klist_identifier(l);
 
      if (cb->type == ttoken)
      {
	 callstruct.type = ktoken_to_string(cb->type);

	 /*-- get the segment presentation from the callback's object --*/
         tpres = kdms_get_segment_info(cb->object,
				       ktoken_to_string(cb->segment));
	 if (tpres == NULL)
	    return FALSE;

         _kdms_callback_scale_coord(cbegin, &callstruct.begin,
				    pres->size, tpres->size,
				    pres->order, tpres->order, 
				    tpres->dimension, 1);
         _kdms_callback_scale_coord(cend, &callstruct.end, 
				    pres->size, tpres->size,
				    pres->order, tpres->order, 
				    tpres->dimension, 1);
         callstruct.object = cb->object;
         cb->callback.data(object, ktoken_to_string(cb->segment),
			   cb->clientData, &callstruct);
      }
   }

   disable = FALSE;

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_process_object_callbacks
|
|       Purpose: Process all the object callbacks of a given type 
|		 on the provided data object.
|
|         Input: object - object generating callback
|		 type   - callback type
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: May 26, 1994 
| Modifications:
|
------------------------------------------------------------*/
int
kdms_process_object_callbacks(kobject object, char *type)
{
   static int disable = FALSE;

   kcallback *cb;
   klist     *l;
   ktoken     ttoken;

   if (object->phys->callbacks == NULL)
      return TRUE;  /*-- trivial case --*/

   ttoken = kstring_to_token(type);

   /*-- blindly disable callbacks while callbacks are being serviced --*/
   if (disable == TRUE || _kdms_disable_callbacks == TRUE)
      return (TRUE);

   disable = TRUE;

#if 0
      kprintf("\nIn callback with the following info\n");
      kprintf("    Object File '%s'\n", object->phys->filename);
      kprintf("    Object Address '%x'\n", object);
      kprintf("    Number of callbacks '%d'\n", 
	      klist_size(object->phys->callbacks));
#endif

   for (l = klist_head(object->phys->callbacks); l != NULL; l = klist_next(l))
   {
      cb = (kcallback *) klist_identifier(l);

      if (cb->type == ttoken)
      {
	 cb->callback.object(object, type, cb->clientData);
#if 0
	 kprintf("\n    Type of callback '%s'\n", cb->type);
	 kprintf("    Routine '%x'\n", cb->callback.object);
	 kprintf("\n");
#endif
      }
   }

   disable = FALSE;

   return TRUE;
}


/****    =============================================================
 ****    Delete Callbacks
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdms_delete_callbacks_for_object
|
|       Purpose: Deletes all callbacks for the given object and 
|		 segment.  This is used when objects
|                are closed.  It deletes all callbacks which
|		 refer to the object and segment, so they
|		 are not called anymore.  If the pres 
|		 argument is NULL, then it will delete object
|		 callbacks.
|
|         Input: object - object which is going away
|		 pres   - segment presentation which is going away.
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: December 20, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_delete_callbacks_for_object(
   kobject            object, 
   kdas_presentation *pres)
{
   kdms_callback *cb;
   klist         *l;

   if (object == NULL || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   /*-- delete all object callbacks for this object --*/
   if (pres == NULL)
   {
      if (object->phys->callbacks)
      {
	 /*-- work backwards, since we are deleting --*/
	 l = klist_tail(object->phys->callbacks);
	 do 
	 {
	    cb = (kdms_callback *) klist_identifier(l);

	    /*-- go to the prev entry, -then- worry about the current one --*/ 
	    l = klist_prev(l);
	    
	    /*---------------------------------------------------------------
	    |  if object matches, then remove current entry
	    |  unless it's a destroy callback, in which case we shouldn't
	    |  remove it because it applies to the low level physical object
	    \---------------------------------------------------------------*/
	    if ((cb->object == object) && 
		(kstrcmp((char *)cb->type, KDMS_CALLBACK_DESTROY) != 0))
	    {
	       object->phys->callbacks = klist_delete(object->phys->callbacks, 
						      cb);
	       kfree(cb);
	    }
	    
	 } while (l != NULL);
      }
   }
   
   /*-- delete all segment callbacks for this object --*/
   else
   {
      if (pres->segment->callbacks)
      {
	 /*-- work backwards, since we are deleting --*/
	 l = klist_tail(pres->segment->callbacks);
	 do
	 {
	    cb = (kdms_callback *) klist_identifier(l);
	 
	    /*-- go to the prev entry, -then- worry about the current one --*/
	    l = klist_prev(l);

	    /*-- if object and segment match, then remove current entry --*/
	    if (cb->object == object && pres->id == cb->segment)
	    {
	      pres->segment->callbacks = klist_delete(pres->segment->callbacks,
						      cb);
	      kfree(cb);
	    }

	 } while (l != NULL);
      }
   }
   
   return TRUE;
}

/****    =============================================================
 ****    Static Utility Routine
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdms_callback_scale_coord
|
|       Purpose: Scale one point in a space specified by a source
|		 size and index order into a space specified by
|		 another source and index order.   The point
|		 is returned via the classic kposition structure,
|		 and as such is organized into the polymorphic
|		 value segment size and order.  This may have strange 
|		 results for other segments...
|
|         Input: spoint  - source point to scale
|		 ssize   - source size
|		 sorder  - source order 
|		 dsize   - destination size
|		 dorder  - destination order 
|		 dim     - dimensionality
|		 def     - default value to use when order is not 
|		  	   one of the polymorphic order values of
|			   KWIDTH, KHEIGHT, KDEPTH, or KTIME.
|		 
|        Output: dpoint  - destination point in the form of a kposition
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: March 29, 1994
| Modifications:
|
------------------------------------------------------------*/
static void
_kdms_callback_scale_coord(
   int       *spoint, 
   kposition *dpoint,
   int 	     *size,
   int 	     *dsize,
   int 	     *sorder,
   int 	     *dorder,
   int 	      dim,
   int 	      def)
{
   double stride;
   int    i, index = 0, tmp;
   int   *d;

   /*-- manually reorder for optimal speed --*/
   for (i = 0, d = dorder; i < KDMS_MAX_DIM; i++, d++)
   {
      if      (sorder[0] == *d) index = 0;
      else if (sorder[1] == *d) index = 1;
      else if (sorder[2] == *d) index = 2;
      else if (sorder[3] == *d) index = 3;
      else if (sorder[4] == *d) index = 4;

      stride = (double) (dsize[i]-1) / kmax((size[index]-1), 1);
      tmp    = (int) ((double) spoint[index] * stride);

      /*-- finally fill out the position in "standard" polymorphic order --*/
      switch (*d)
      {
	 case KWIDTH  	: dpoint->w = tmp; break;
	 case KHEIGHT 	: dpoint->h = tmp; break;
	 case KDEPTH  	: dpoint->d = tmp; break;
	 case KTIME   	: dpoint->t = tmp; break;
	 case KELEMENTS : dpoint->e = tmp; break;
      }
   }

   return;
}
