/*
  * Khoros: $Id: glue.c,v 1.7.2.2 1997/03/11 17:16:20 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: glue.c,v 1.7.2.2 1997/03/11 17:16:20 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Data Services Glue Management
   >>>>		
   >>>>  Private:
   >>>>                 kdms_initialized()
   >>>>                 kdms_identify_format()
   >>>>                 kdms_locate_format()
   >>>>                 kdms_free_formats()
   >>>>   Static:
   >>>>                 _kdms_get_default_format()
   >>>>   Public:
   >>>>                 kdms_add_file_format()
   >>>>                 kdms_support()
   >>>>                 kdms_glue_set_resources()
   >>>>                 kdms_glue_get_resources()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*- if there is not an user-defined default, then this is the default --*/
#define DEFAULT_FORMAT "kdf"

/*-- if we want a raw format, we'll get it! --*/
#define RAW_FORMAT     "raw"

/*-- list of defined formats --*/
static kdas_format **defined_formats     = NULL;
static int           num_defined_formats = 0;

/*-- default data format --*/
static kdas_format  *default_format      = NULL;
static int           default_format_init = FALSE;

static kdas_format *_kdms_get_default_format PROTO((void));

/****    =============================================================
 ****    Public Routines
 ****    =============================================================
 ****/

/************************************************************
*
*  Routine Name: kdms_add_file_format
*
*       Purpose: This function is used to add support for a new
*		 file format to data services.  
*
*         Input: name          	- A short name for the file format.  This
*		             	  is the name that will be used when setting 
*		 	     	  the KDMS_FORMAT attribute.
*
*		 description   	- A short description of the file format.
*
*		 check       	- This function that takes an open fid, 
*			     	  examines the file to determine if it 
*			     	  is stored in this format.
*
*		 input       	- This function called when check returns
*			     	  TRUE.  This function opens the file,
*			     	  reads the header, and sets attributes 
*			     	  of a data object based on the information
*			     	  in the header.  This function is optional
*				  provided there is an output function given.
*
*		 output      	- This function is called when an output 
*			          object is about to be closed.  This 
*				  function will open a permanent output 
*				  file and write a header based on object 
*				  attributes. This function is optional
*				  provided there is an input function given.
*
*		 destroy     	- This function is called when any object is 
*				  about to be closed.  This routine frees up 
*				  memory stored in object resources used 
*				  by the file format support.
*
*		 prepare_output - Optional function for preparing an
*				  output object to be written to the disk.  
*			 	  This format should consider any 
*				  limitations of the format and modify
*				  the object to best accomodate the output
*				  format.  For example, this routine may
*				  turn on presentation mapping for formats
*				  with no support for map storage.  If
*				  NULL, this function will simply be ignored.
*
*		 locate_segment - Optional function for large data set
*				  support.  This will move return an
*				  fid which has been placed at the first
*				  byte of a named segment.
*
*		 reader         - Optional function for large data set
*				  support.  This should be capable of
*				  reading a given number of data values
*				  in the datatypes supported by the
*			  	  file format.  If NULL, this will default
*				  to kread_generic().  
*
*		 writer         - Optional function for large data set
*				  support.  This should be capable of
*				  writing a given number of data values
*				  in the datatypes supported by the
*			  	  file format.  If NULL, this will default
*				  to kwrite_generic().  
*
*		 seeker         - Optional function for large data set
*				  support.  This should be capable of seeking
*				  a file id some given number of bytes in the 
*				  data file.  If NULL, this will default to
*				  klseek().
*
*		 teller         - Optional function for large data set
*				  support.  This should be capable of telling
*				  the position of a file id in terms of bytes
*				  in the data file.  If NULL, this will 
*				  default to ktell().
*        Output: None
*
*       Returns: TRUE on success, FALSE otherwise
*  Restrictions:
*    Written By: Jeremy Worley
*          Date: December 7, 1994
*      Verified:
*  Side Effects:
* Modifications:
*   Declaration:
*
*************************************************************/
int
kdms_add_file_format(
   char    *name,
   char    *description,
   int      (*check)          PROTO((int)),
   int      (*input)          PROTO((kobject, int, int)),
   int      (*output)  	     PROTO((kobject, int, int)),
   int      (*destroy) 	     PROTO((kobject)),
   int      (*prepare_output) PROTO((kobject)),
   int      (*locate_segment) PROTO((kobject, char *)),
   ssize_t  (*reader)         PROTO((int, kaddr, size_t, int)),
   ssize_t  (*writer)         PROTO((int, kaddr, size_t, int )),
   off_t    (*seeker)         PROTO((int, off_t, int)),
   long int (*teller)         PROTO((int)))
{
   kdas_format *tmp;

   KDMS_INIT;

   /*-- construct the new format based on provided info --*/
   tmp = kdas_format_construct(name, description, check, input, output, 
			       destroy, prepare_output, locate_segment, 
			       reader, writer, seeker, teller);
   if (tmp == NULL)
      kdas_fail_error(KINTERNAL);

   /*-- add the new format to end of the definition list --*/
   defined_formats = (kdas_format **) karray_insert(defined_formats, KSTRUCT,
				   tmp, (size_t) num_defined_formats, 
				   KLIST_TAIL, FALSE);
   num_defined_formats++;
   
   if (defined_formats == NULL)
      kdas_fail_error(KINTERNAL);

   return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_support - obtain a list of file formats
*                supported by data services.
*
*       Purpose: This function is used to obtain a list of the file
*                formats supported by the data abstraction.
*
*         Input: list_inputs  - TRUE if formats with input support
*			     	should be included in the list
*                list_outputs - TRUE if formats with output support
*			        should be included in the list
*
*        Output: num - The number of formats in the returned 
*		       string array.
*
*       Returns: An array of strings (char **) containing
*                a list of all formats that are currently
*                defined in Data Services.
*
*  Restrictions:
*    Written By: Jeremy Worley
*          Date: October 18, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
char **
kdms_support(
   int *num,
   int list_inputs,
   int list_outputs)
{
   char **names;
   int    i, number = 0;

   KDMS_INIT;

   names = (char **) kcalloc((size_t)num_defined_formats, sizeof(char *));

   if (names == NULL)
      return NULL;

   for (i = 0; i < num_defined_formats; i++)
   {
      if (list_inputs && defined_formats[i]->input)
      {
	 names[number] = kstring_copy(defined_formats[i]->name, NULL);
	 number++;
      }
      

      if (list_outputs && defined_formats[i]->output)
      {
	 names[number] = kstring_copy(defined_formats[i]->name, NULL);
	 number++;
      }
   }
   
   if (num)
      *num = number;

   return names;
}

/************************************************************
*
*  Routine Name: kdms_glue_set_resources - set resource pointer
*
*       Purpose: This function is used to set a resource pointer
*	         within a data object.  This should be used
*		 by format glue routines to associate any format
*		 specific information with a data object for
*		 later use.
*
*         Input: object   - Object to set resources on.
*		 resource - Resource to set.	
*
*        Output: none
*
*       Returns: TRUE on success, FALSE if object is invalid
*
*  Restrictions:
*    Written By: Jeremy Worley
*          Date: November 8, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int 
kdms_glue_set_resources(
   kobject object, 
   kaddr   resource)
{
   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   object->phys->resources = resource;

   return TRUE;
}

/************************************************************
*
*  Routine Name: kdms_glue_get_resources - get resource pointer
*
*       Purpose: This function is used to retreive a resource pointer
*	         within a data object.  This should be used
*		 by format glue routines to retrieve any format
*		 specific information which was associated with
*		 the data object via the kdms_glue_set_resources()
*		 call.
*
*         Input: object   - Object to get resources from.
*
*        Output: none
*
*       Returns: resources if available, FALSE otherwise
*
*  Restrictions:
*    Written By: Jeremy Worley
*          Date: November 8, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kaddr 
kdms_glue_get_resources(kobject object)
{
   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   return object->phys->resources;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_initialized
|
|       Purpose: This routine returns true if the resource
|                structure has been initialized, and false
|                otherwise.
|
|         Input: object - object to check 
|
|        Output:
|
|       Returns: TRUE if format resources for this object are present, 
|		 FALSE otherwise
|
|    Written By: Jeremy Worley
|          Date: May 10, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_initialized(kobject object)
{
   return (object->phys->resources != NULL);
}

/****    =============================================================
 ****    Format Routines
 ****    =============================================================
 ****/

/*-----------------------------------------------------------
|
|  Routine Name: kdms_identify_format - find out what format an open fid is in 
|
|       Purpose: This function takes an open fid, and, if it is an input
|		 object, will check the fid against all known file formats,
|		 returning a copy of the identified file format if one is
|		 found.
|		
|		 If the fid is for an output object, this routine will 
|		 return the default output file format.  
|
|		 If the object is specified to be a raw data file, then it 
|		 will be opened as a raw data file, regardless.
|
|         Input: fid   - open fid containing the file to check
|                flags - flags with which the data object was opened,
|			 these are data service flags, not file flags	
|
|        Output: none
|
|       Returns: identified format on success, NULL otherwise
|
|    Written By: Jeremy Worley and Steve Kubica
|          Date: January 4, 1995
| Modifications:
|
------------------------------------------------------------*/
kdas_format *
kdms_identify_format(
   int fid, 
   int flags)
{
   kdas_format *format = NULL;

 
   /*-- force raw if that's what we requested --*/
   if (flags & KOBJ_RAW)
   {
      format = kdms_locate_format(RAW_FORMAT);
      return kdas_format_dup(format);
   }
   
   /*-- if input object, check against all existing types --*/
   if (flags & KOBJ_READ) 
   {
      int i, check;
      
      for (i = 0; i < num_defined_formats; i++)
      {
	 /*------------------------------------------------------------
	 |  Check to see if this is the format which corresponds to 
	 |  the open fid.  If it is, then return it, but first rewind
	 |  it so we can read it properly later.
	 \------------------------------------------------------------*/
	 if ((defined_formats[i] != NULL) && 
	     (defined_formats[i]->check != NULL))
	 {
	    check = defined_formats[i]->check(fid);

	    /*-- rewind --*/
	    klseek(fid, (off_t) 0, KSEEK_SET);

	    if (check == TRUE)
	       return kdas_format_dup(defined_formats[i]);
	 }
      }
   
      /*--------------------------------------------------------------
      |  Couldn't identify it, maybe we can still glean what it is
      |  to give the use a clue about what's wrong.	
      |
      |  ... the file command and /etc/magic is the answer...
      \--------------------------------------------------------------*/
      kdas_set_error(KDMS_EFMT_CLUELESS);

      return NULL;
   }

   /*-- if output object, return default format --*/
   if (flags & KOBJ_WRITE)
   {
      format = _kdms_get_default_format();

      /*-- make sure we have an output method... --*/
      if (format->output == NULL)
      {
	 /*-- ...if not then we will resort to the hardwired default --*/
	 kerror("kdataman", "kdms_identify_format", 
		"The format %s specified by KHOROS_FORMAT has no output "
		"support : defaulting output format to %s",
		format->name, DEFAULT_FORMAT);
 
	 format = kdms_locate_format(DEFAULT_FORMAT);
      }
   }

   return kdas_format_dup(format);
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_locate_format - given a name, find the format
|
|       Purpose: This function will take a given name, search the
|		 list of defined formats, and return the one which
|		 matches.  If none match, return NULL.
|
|         Input: name - name to search for
|
|        Output: none
|
|       Returns: format matching name if defined, NULL otherwise
|
|    Written By: Steve Kubica
|          Date: February 29, 1996
| Modifications:
|
------------------------------------------------------------*/
kdas_format *
kdms_locate_format(char *name)
{
   int i;

   /*-- mindless search of the list to find the name --*/
   for (i = 0; i < num_defined_formats; i++)
   {
      if (kstrcmp(defined_formats[i]->name, name) == 0)
	 return defined_formats[i];
   }

   return NULL;
   
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kdms_get_default_format - retrieve default format
|
|       Purpose: This function will, if the default format has not been
|		 determined yet, will initialize it by checking the 
|		 environment variable KHOROS_FORMAT, and, if that as
|		 not been defined, by retrieving the hardwired default 
|		 format from the format list.
|
|		 If the format inidicated by the environment variable
|		 is not defined, then an error message will be printed,
|		 and the default will be assigned to the hardwared 
|		 default. 
|		
|		 The environment variable KHOROS_KDMS_FORMAT is supported
|		 for backwards compatibility.
|
|         Input: none
|        Output: none
|       Returns: default format
|
|    Written By: Steve Kubica
|          Date: February 29, 1996
| Modifications:
|
------------------------------------------------------------*/
static kdas_format *
_kdms_get_default_format(void)
{
   char *name;
   
   /*-- have we initialized the default yet? --*/
   if (!default_format_init)
   {
      /*-----------------------------------------------------------
      |  all these checks are done separately, so we can identify 
      |  the source of any bogus format defaults.
      \-----------------------------------------------------------*/

      /*-- check environment variable first --*/
      if ((name = kgetenv("KHOROS_FORMAT")) != NULL)
      {
	 if ((default_format = kdms_locate_format(name)) == NULL)
	 {
	    kerror("kdataman", "_kdms_get_default_format",
		   "The format %s defined by KHOROS_KDMS_FORMAT is unknown\n"
		   "Defaulting to use %s.", name, DEFAULT_FORMAT);
	    default_format = kdms_locate_format(DEFAULT_FORMAT);
	 }
      }
      
      /*-- check old environment variable next : backwards compatibility --*/
      else if ((name = kgetenv("KHOROS_KDMS_FORMAT")) != NULL)
      {
	 if ((default_format = kdms_locate_format(name)) == NULL)
	 {
	    kerror("kdataman", "_kdms_get_default_format",
		   "The format %s defined by KHOROS_KDMS_FORMAT is unknown\n"
		   "Defaulting to use %s.", name, DEFAULT_FORMAT);
	    default_format = kdms_locate_format(DEFAULT_FORMAT);
	 }
      }
	    
      /*-- hardwired default --*/
      else 
	 default_format = kdms_locate_format(DEFAULT_FORMAT);

      /*-- sanity check -- make sure hardwired default is valid! --*/
      if (default_format == NULL)
      {
	 kerror("kdataman", "_kdms_get_default_format",
		"Not good : the hardwared default format %s is not known.\n"
		"Someone has switched data services on you!\n"
		"Try setting another format using the KHOROS_FORMAT\n"
		"environment variable to get around this.\n", DEFAULT_FORMAT);
	 return NULL;
      }

      default_format_init = TRUE;
   }

   return default_format;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_free_formats
|
|       Purpose: Free all defined formats.
|
|         Input: none
|
|        Output: none
|
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: March 8, 1996
| Modifications:
|
------------------------------------------------------------*/
void
kdms_free_formats(void)
{
   int i;
   
   if (num_defined_formats == 0)
      return;

   for (i = 0; i < num_defined_formats; i++)
      kdas_format_destruct(defined_formats[i]);
   
   kfree_and_NULL(defined_formats);

   return;
}
