 /*
  * Khoros: $Id: public.c,v 1.17.2.2 1997/03/11 17:16:21 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: public.c,v 1.17.2.2 1997/03/11 17:16:21 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		
   >>>>  Private:
   >>>>                 kdms_locate()
   >>>>                 kdms_get_segment_list()
   >>>>                 kdms_copy_object()
   >>>>   Public:
   >>>>                 kdms_create()
   >>>>                 kdms_open()
   >>>>                 kdms_close()
   >>>>                 kdms_reference()
   >>>>                 kdms_reopen()
   >>>>                 kdms_sync()
   >>>>                 kdms_update_references()
   >>>>                 kdms_reference_list()
   >>>>                 kdms_get_segment_names()
   >>>>                 kdms_close_hook()
   >>>>                 kdms_copy_segment_data()
   >>>>                 kdms_copy_remaining_data()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/************************************************************
*
*  Routine Name: kdms_create - create a temporary data object.
*
*       Purpose: kdms_create is used to instantiate a data object
*                (kobject) when it will only be used for temporary
*                storage of information.  If you are intending to
*                process an object that already exists as a file or
*                transport (input), or you are planning on saving the
*                kobject to a file or transport (output), then the
*                appropriate routines to use are kdms_input,
*                kdms_output, or kdms_open.
*
*                This function creates an instance of a data object
*                that will have associated with it a temporary
*                transport that will be used for buffering large
*                amounts of data.  This temporary transport will be
*                automatically removed when the process terminates.
*                There is no way to rename the temporary file or
*                replace the temporary file with a permanent one.
*
*                The kdms_create function call creates what is
*                essentially a "blank" object.  That is, the object
*                will initially have no segments, and almost all
*                global attributes will be initialized to default
*                values.  If a default is not appropriate, then the
*                attribute will be uninitialized.  The default values
*                for attributes are described in Chapter 5 of Program
*		 Services Volume 2. of the Khoros 2.0 Manual.
*
*                An object that is created with this function call
*                behaves similarly to an output object that is created
*                with the kdms_output function call.  Thus, it is
*                necessary to create each segment that is desired and
*                initialize attributes such as size and datatype prior
*                to using the object.
*
*         Input: none
*
*        Output: none
*
*       Returns: kobject on success, KOBJECT_INVALID upon failure
*
*  Restrictions: 
*    Written By: Mark Young and Jeremy Worley
*          Date: July 13, 1992
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kobject
kdms_create(void)
{
   kobject object;
   
   KDMS_INIT;

   object = kdms_object_construct(TRUE);

   /*-- this will return the default format for us --*/
   object->phys->format = kdms_identify_format(-1, KOBJ_WRITE);
   object->phys->temporary = TRUE;

   return object;
}

/************************************************************
*
*  Routine Name: kdms_open - create an object associated with an input
*                or output transport.
*
*       Purpose: kdms_open is used to instantiate a data object
*                (kobject) that is assocated with a permanent file or
*                transport.  If a permanent file is not desired
*                (i.e. the object is going to be used as temporary
*                storage, and will not be used by any other process)
*                then the kdms_create function call should be used
*                instead.
*
*                The first argument to this function is the transport
*                or file name.  This argument indicates the name of
*                the transport that is associated with the object.
*                The transport name can be any legal khoros transport
*                description.  While this will usually be a regular
*                UNIX file name, it is also possible to specify such
*                things as shared memory pointers, memory map
*                pointers, sockets, streams, and even transports on
*                other machines.  For more information on the syntax
*                of a Khoros transport name, refer to the online man
*                page for the Khoros function kopen.
*
*                The second argument to the kdms_open function call is
*                used to provide data services with specific
*                information about how the object is going to be
*                manipulated.  The flags argument is analogous to
*                kopen's flags argument.  The flags argument is
*                constructed by bitwise OR'ing predefined values from
*                the following list:
*
*		.RS
*       	.IP "KOBJ_READ" 10
*				Open an existing file or transport
*                               for reading (input).  By using this flag,
*                               you are indicating that the file or transport
*                               exists and that it contains valid data.
*                               If it does not exist, or the data is not
*                               recognized, then an error message will
*                               be generated and this function will
*                               return KOBJECT_INVALID.
*        
*        	.IP "KOBJ_WRITE" 10
*				Open a file or transport for writing
*                               (output).  By using this flag, you are
*                               indicating that any data that you write
*                               to the object will be stored in the
*                               file or transport specified.
*        
*        	.IP "KOBJ_STREAM" 10
*				If the transport is a stream then polymorphic
*	                        services does not attempt to buffer the
*	                        data.  By setting this value, you are
*	                        indicating that that process reads and/or
*	                        writes the data strictly sequentially from the
*                               object in its native index order, and
*                               it is not necessary to buffer the data
*                               in a temporary transport that can be
*                               accessed randomly.
*        
*        	.IP "KOBJ_RAW" 10
*				When an object is opened, data
*	 			services usually attempts to recognize 
*			        the file format by examining the first 
*		  		part of the file.  By setting this value,
*	 			you will bypass this operation, forcing
*			        the file to be read as raw unformatted
*				data.
*		.RE
*
*                These flags can be combined arbitrarily (with the exceptions
*                given above) to alter the interpretation of the file 
*                or transport.  For example, specifying both KOBJ_READ
*                and KOBJ_WRITE will result in a read/write file object.
*                This implies that the file already exists and will be
*                read from using kdms_get_data and written to using
*                kdms_put_data.  When kdms_close is called, the changes
*                that are a result of calls to kdms_put_data will be
*                stored to the file.
*
*                However, if you intend to open an output object, but
*                you need to occationally read data from it that you
*                have already written, it is not necessary to specify 
*                KOBJ_READ (in fact, doing so may result result in an
*                error if the file or transport does not already exist).
*
*                Likewise, it is possible to call kdms_put_data on
*                an input object (one which was opened without the 
*                KOBJ_WRITE flag).  If this is done, then subsequent 
*                calls to kdms_get_data on a region that has been
*                written to will contain the new data.  However,
*                the file that is associated with this input object
*                will not be changed.  Thus, the KOBJ_READ and KOBJ_WRITE
*                flags only indicate what operations are allowed on the
*                permanent file that is associated with the object, not
*                what operations are allowable on the object itself.
*
*                If KOBJ_READ is specified, then the Data Services
*                will attempt to recognize the file format automatically.
*                If it fails, then this function will return KOBJECT_INVALID,
*                indicating that it was unable to open the object,
*                unless the KOBJ_RAW flag was also specified,
*                in which case, it will assume that the input file
*                is simply headerless data.  The structured file formats
*                that are currently recognized are Kdf (The Khoros
*                2.0 standard file format), Viff (The Khoros 1.0
*                standard file format, Pnm (Portable Any Map, which includes
*                PBM, PGM, and PNM), and Sun Raster.
*                
*         Input: name - a string that contains the path name
*                       of a file or transport that will be 
*                       associated with the object.
*
*                flags - how the object is to be opened.
*                        a combination of KOBJ_READ, KOBJ_WRITE,
*                        KOBJ_STREAM and KOBJ_RAW as 
*                        described above.
*
*        Output: none
*       Returns: kobject on success, KOBJECT_INVALID upon failure
*
*  Restrictions: The KOBJ_RAW flag will have unpredictable 
*                results if it is combined with the KOBJ_WRITE flag.
*                This limitation will be removed in a later release of 
*                the Khoros 2.0 system.
*
*                The Kdbm file format has a bug in it that prevents
*                it from being used on a stream base output.  This
*                will be fixed in a future release of the Khoros 2.0
*                system.
*    Written By: Jeremy Worley
*          Date: July 13, 1992
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kobject
kdms_open(
    char *name,
    int   flags)
{
   KDMS_INIT;
   
   return kdms_object_open(name, flags);
}

/************************************************************
*
*  Routine Name: kdms_close - close an open data object.
*
*       Purpose: This function is called on an object when all
*                interaction with the object is completed.
*                In addition to freeing resources that were
*                used to manage the object, this function also,
*                performs any "last minute" manipulation on the
*                file or transport that is associated with the
*                object.
*
*                If the object was created with the kdms_reference
*                function call, or if another object was created as
*                a reference of the one being closed, then the object 
*                might be sharing some of its resources with other 
*                objects.  If this is the case, then those shared 
*                resources will not be freed, but rather they will 
*                be disassociated from the object being closed.  Thus, 
*                closing an object does not affect any other object.
*
*         Input: object - the object to be closed.
*
*        Output: none
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions:
*    Written By: Mark Young and Jeremy Worley
*          Date: July 13, 1992
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_close(kobject object)
{
   KDMS_INIT;
   
   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);
   
   return kdms_object_close(object);
}

/************************************************************
*
*  Routine Name: kdms_reference - create a reference of a
*                data object.
*
*       Purpose: This function is used to create a reference
*                of a data object that can be treated as a second
*                independant data object under most circumstances.
*                A referenced object is similar conceptually to a 
*                symbolic link in a UNIX file system in most 
*                respects.  For example, getting data from an
*                input object and a reference of the object will
*                result in the same data.  Data that is put on an
*                output object can then be gotten from one of its 
*                references.  
*
*                The similarity ends there.  Once an object is
*                referenced, the two resulting objects are 
*                equivelant--there is no way to distinguish the
*                original from the reference.  In fact, closing
*                the original does not in any way affect the 
*                reference, and visa-versa.
*
*                kdms_reference creates a new object that has
*                presentation attributes that are independant
*                of the original object's presentation attributes.
*                The presentation attributes are UNCOUPLED from
*                the physical attributes, see the description
*                found in Chapter 6 of the the Khoros Programmer's
*                Manual on the KDMS_COUPLING attribute for more
*                information.  The two objects (or more if there are
*                future calls to kdms_reference) share all physical
*                resources.
*
*         Input: object - the abstract data object to be
*                         reference.
*
*        Output: none
*       Returns: a kobject that is a reference of the input
*                object on success, KOBJECT_INVALID upon failure
*
*  Restrictions: 
*
*    Written By: Mark Young and Jeremy Worley
*          Date: August 20, 1992
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kobject
kdms_reference(kobject object)
{
   KDMS_INIT;
   
   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);
   
   return kdms_object_reference(object);
}

/************************************************************
*
*  Routine Name: kdms_reopen - associate new data with an existing object
*
*       Purpose: This function is used to associate new stored data
*		 with an already open data object.  This operation can
*		 only be performed on an input object.  This function
*		 is typically used in an interactive environment when
*		 many references to replace the data set being operated
*		 on or visualized with a new data set, without having
*		 to close all of the references and replace them with
*		 new ones.
*
*         Input: object   - data object re-open
*		 filename - transport name of new data set to associate with
*		       	    the object.
*		 flags    - flags to use when opening the data set specified
*			    by the filename argument.
*
*        Output: 
*
*       Returns: object (the input argument) on success, NULL on failure.
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: November 15, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kobject
kdms_reopen(
   kobject object,
   char   *name,
   int     flags)
{
   KDMS_INIT;

   return kdms_object_reopen(object, name, flags);
}

/************************************************************
*
*  Routine Name: kdms_sync - synchronize physical and presentation
*                layers of a data object.
*
*       Purpose: This function is used to update physical
*                attributes of a data object to match those
*                of the presentation layer, or visa-versa.
*                When an attribute is set via kdms_set_attribute(s)
*                or kdms_copy_attribute(s) calls, the presentation
*                version of the attribute is the only thing
*                that is directly manipulated.  The KDMS_COUPLING
*                attribute is used at that time to determine if
*                the physical attribute should be updated to
*                correspond to its value at the presentation
*                level.  The KDMS_COUPLING attribute can take
*                on one of three values: KUNCOUPLED, KCOUPLED, or
*                KDEMAND.  If it is set to KUNCOUPLED or KDEMAND,
*                then Data Services will not update the physical
*                layer.  If the attribute is set to KCOUPLED, then
*                data services immediately updates the physical
*                layer.  If the attribute is set to KDEMAND,
*                then this updating will only occur when kdms_sync
*                is called.  If the KDMS_COUPLING attribute is
*                set to KUNCOUPLED, then this routine will simply
*                return, without issuing an error message.
*
*         Input: object  - data object to be resynchronized.
*                segment - segment re-synchronize.  If this is
*                          set to KDMS_OBJECT, then all segments
*                          are re-synchronized.
*                direction - the desired direction of the synchronization.
*                            the legal values are KPRES2PHYS, which
*                            indicates that the physical layer will be
*                            updated to correspond to the presentation
*                            layer; and KPHYS2PRES, which indicates
*                            that the presentation layer will be updated
*                            to correspond to the physical layer.
*
*        Output:
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: September 1, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kdms_sync(
   kobject object,
   char   *segment,
   int     direction)
{
   kdas_presentation *pres;
   
   KDMS_INIT;
   
   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   /*-- null segment means to sync everything --*/
   if (segment == NULL)
   {
      klist *l;
      
      for (l = object->presentations ; l != NULL; l = klist_next(l))
	 kdms_segment_sync((kdas_presentation *)l->client_data, direction);
   }
   else
   {
      if ((pres = kdms_get_segment_info(object, segment)) == NULL)
      {
	 kdas_set_error(KDMS_ESEG_NONEXIST);
	 return FALSE;
      }
      kdms_segment_sync(pres, direction);
   }
   
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_locate
|
|       Purpose:
|
|         Input:
|
|        Output:
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Jeremy Worley
|          Date: June 11, 1993
| Modifications:
|
------------------------------------------------------------*/
kobject
kdms_locate(char *filename)
{
   KDMS_INIT;

   return kdms_object_locate(filename);
}

/************************************************************
*
*  Routine Name: kdms_update_references - update segment presentation
*  				          of all reference objects.
*
*       Purpose: This function propagates the values of the
*                presentation attributes from the specified 
*                object and segment to the corresponding segments
*                in all of the object's references.
*
*         Input: object - object containing segment to propagate.
*                segment - name of segment to be propagated.
*
*        Output: 
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: October 26, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int 
kdms_update_references(
   kobject object,
   char *segment)
{
   KDMS_INIT;

   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   return kdms_object_update_references(object, segment);
}

/************************************************************
*
*  Routine Name: kdms_reference_list - return a klist of references.
*
*       Purpose: This function returns a klist of objects that are
*                references of the object passed in as the argument
*                to this function.  The object passed in will also
*                be in this list.  NOTE:  This list is the one
*                actually used by KDMS to manage references.  Destroying
*                this list will cause grief like you've never seen
*                before.
*
*         Input: object - the object to get references of.
*
*        Output: 
*
*       Returns: a klist * on success, NULL on failure.
*
*  Restrictions: 
*    Written By: Jeremy Worley
*          Date: October 27, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
klist *
kdms_reference_list(kobject object)
{
   KDMS_INIT;

   if (!object || object->type != KOBJ_DATA)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   return (klist_head(object->phys->references));
}

/************************************************************
*
*  Routine Name: kdms_get_segment_names - get an array of 
*		 segment names for the object specified.
*
*       Purpose: Given an object, obtain an array of strings
*		 which are the names of all segments which
*		 exist in the object.
*
*         Input: object - the object to get the segment names
*		          from.
*

*        Output: number - the number of segments in the object
*			  (and thus, the number strings in the
*			  array that is returned.
*
*       Returns: an array of strings containing the names of the
*		 segments present in the object on success, NULL
*		 on failure.
*
*  Restrictions: 
*    Written By: Steve Kubica
*          Date: May 2, 1994
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
char **
kdms_get_segment_names(kobject object, int *number)
{
   klist *seg;
   char  *s = NULL;
   char **segments = NULL;
   int    nsegments = 0;

   KDMS_INIT;

   for (seg = klist_head(object->presentations);
        seg != NULL; seg = klist_next(seg))
   {
      s = ktoken_to_string(klist_identifier(seg));
      segments = (char **) karray_add(segments, KSTRING, kstrdup(s),
			(size_t) nsegments);
      nsegments++;
   }

   if (number)
      *number = nsegments;

   return segments;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_get_segment_list
|
|       Purpose: I don't really like this function.  Especially,
|		 given the fact that another function (kdms_get_segment_names)
|		 exists which performs a similar function without
|		 jeopardizing the integrity of the object.
|         Input:
|        Output:
|       Returns:
|    Written By: Jeremy Worley
|          Date: August 18, 1994
| Modifications:
|
------------------------------------------------------------*/
klist *
kdms_get_segment_list(kobject object)
{
   KDMS_INIT;

   return object->presentations;
}

/************************************************************
*
*  Routine Name: kdms_close_hook - insert a service to be called
*		 when an object is closed.
*
*       Purpose: This function is used to insert a special function that is
*                called immediately before an object is closed (after
*                a call to kdms_close) that can perform any cleanup
*                that may be required before the object is written if
*                it is an output object.
*
*         Input: object - object to have close hook function added
*		          to.
*		 func   - function to set as close hook.  This function 
*			  has the following prototype:  func(kobject).
*
*        Output: 
*
*       Returns: The close_hook that was assigned prior to this
*		 new assignment. 
*
*  Restrictions: 
*    Written By: Jeremy Worley & John Salas
*          Date: November 15, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kfunc_int 
kdms_close_hook(
   kobject   object, 
   kfunc_int func)
{
   kfunc_int tmp;

   KDMS_INIT;

   if (!object || object->type != KOBJ_DATA)
      return (NULL);

   tmp = (kfunc_int) object->phys->close_hook;

   object->phys->close_hook = (int(*)(kobject)) func;

   return tmp;
}

/************************************************************
*
*  Routine Name: kdms_copy_segment_data - copy data from one data
*                                         object to another.
*
*       Purpose: This function copies data associated with one segment 
*                or all segments from a source object to a destination 
*                output object.  
*
*                The first argument to this function is the source
*                object.  The third argument is the destination
*                object.  The second argument to this function
*                determines which data segment will be copied.  If
*                this argument is NULL, then this instructs the
*                kdms_copy_data function to copy the data for all
*                segments to the destination object.
*
*                If a segment that is being copied already exists in
*                the destination object, then the segment data is
*                removed and replaced by the source's segment data.
*                In this event, the physical attributes of the
*                original data in the destination object will remain
*                unchanged.  The data will be automatically translated
*                to fit the physical attributes that are already
*                present.
*
*                If the destination object does not contain a segment
*                that exists on the source object and that is being
*                copied with this function, then the said segment will
*                be created.  In this event, the physical attributes
*                that describe the new segment will be used to
*                initialize the presentation attributes for the
*                destination object.
*
*                Data Services attempts to minimize the number of
*                attributes that are copied with the data, as there
*                another function for that purpose
*                (kdms_copy_attributes).  However, as mentioned above,
*                there are cases when it is necessary to copy certain
*                attributes that are critical to the interpretation of
*                the data.  However, this should only occur when the
*                destination does not have the segment or segments
*                being copied.
*
*         Input: sobject  - the object that serves as a source
*                           for the segment data.
*                segment  - the data segment to be copied.
*
*        Output: dobject  - the object that will serve as
*                           a destination in the copy
*                           operation.
*		 dsegment - optional destination segment, if 
*			    different from the segment argument.
*
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions:
*    Written By: Mark Young, Jeremy Worley, and Steve Kubica
*          Date: August 20, 1992
*      Verified:
*  Side Effects: 
* Modifications:
*
*************************************************************/
int
kdms_copy_segment_data(
   kobject sobject,
   char   *segment,
   kobject dobject,
   char   *dsegment)
{
   kdas_presentation *spres = NULL;
   kdas_presentation *dpres = NULL;
   klist *l;
   int status = TRUE;

   /*-- if optional destination is NULL, then use source name --*/
   if (dsegment == NULL)
      dsegment = segment;
   
   /*-- sanity check --*/
   if (sobject == NULL || dobject == NULL)
      kdas_fail_error(KDMS_ENULL_KOBJECT);

   /*-- copying to ourself is trivial case --*/
   if (sobject == dobject && (kstrcmp(segment, dsegment) == 0))
      return TRUE;

   /*-- copy data for all segments --*/
   if (segment == NULL)
   {
      /*-- copy the data for each segment --*/
      for (l = klist_head(sobject->presentations); l != NULL; 
	   l = klist_next(l))
      {
	 spres = (kdas_presentation *) l->client_data;

	 /*-- this is the segment's current object for copying --*/
	 spres->segment->object = sobject;

	 /*-- instantiate destination segment, if we haven't yet--*/
	 if ((dpres = kdms_get_segment_info(dobject, dsegment)) == NULL)
	    dpres = kdms_segment_create(dobject, ktoken_to_string(spres->id));
	 
	 kdms_segment_copy_physical_attributes(spres, dpres, FALSE);
	 status &= kdms_segment_copy_data(spres, dpres);
      }
   }
   else
   {
      /*-- get the presentations from both --*/
      if ((spres = kdms_get_segment_info(sobject, segment)) == NULL)
      {
	 kdas_set_error(KDMS_ESEG_NONEXIST);
	 return FALSE;
      }

      /*-- instantiate destination if we haven't yet --*/
      if ((dpres = kdms_get_segment_info(dobject, dsegment)) == NULL)
	 dpres = kdms_segment_create(dobject, dsegment);

      kdms_segment_copy_physical_attributes(spres, dpres, FALSE);

      status = kdms_segment_copy_data(spres, dpres);
   }

   return status;
}

/************************************************************
*
*  Routine Name: kdms_copy_remaining_data - copy remaining data
*
*       Purpose: kdms_copy_remaining_data copies any data 
*		 from segments which are instantiated in 
*		 both the source and destination objects
*		 and have not yet had their data modified
*		 in the destination.  
*
*         Input: sobject - object to get data from
*		 dobject - object to copy data to
*
*        Output:
*
*       Returns:
*
*  Restrictions:
*    Written By: Jeremy Worley and Steve Kubica
*          Date: June 24, 1994
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int 
kdms_copy_remaining_data(
   kobject sobject,
   kobject dobject)
{
   kdas_presentation *pres;
   klist *l;

   KDMS_INIT;

   /*-- traverse all destination segments --*/
   for (l = klist_head(dobject->presentations); l != NULL; l = klist_next(l))
   {
      pres = (kdas_presentation *) l->client_data;

      /*-- any data at all in this destination segment? --*/
      if ((pres->segment->data == NULL) && (pres->segment->temp == NULL))
      {
	 char *segment = ktoken_to_string(pres->id);
	 
	 /*-- if this segment exists in the source, then copy it --*/
	 if (kdms_query_segment(sobject, segment))
	 {
	    if (!kdms_copy_segment_data(sobject, segment, dobject, NULL))
	       return FALSE;
	 }
      }
   }

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kdms_copy_object
|
|       Purpose: 
|
|         Input: 
|
|        Output: 
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: December 1, 1993
| Modifications:
|
------------------------------------------------------------*/
int
kdms_copy_object(
   kobject src_obj,
   kobject dest_obj,
   int     copy_attributes,
   int     copy_data)
{
   char **segments = NULL;
   int i;
   int j;
   int nsegments = 0;

   segments = kdms_get_segment_names(src_obj, &nsegments);

   for (i = 0; i < nsegments; i++)
   {
      /*--  if the segment appears to exists in the src, then copy it --*/
      if (kdms_query_segment(src_obj, segments[i]))
      {

         /*-- copy the attributes if so desired --*/
         if (copy_attributes)
            if (!kdms_copy_segment_attributes(src_obj, segments[i],
                                              dest_obj, segments[i])) 
	    {
	       for (j = 0; j < nsegments; j++)
		  kfree_and_NULL(segments[j]);
	       kfree_and_NULL(segments);
	       
	       return (FALSE);
	    }
	 
         /*-- copy the data if so desired --*/
         if (copy_data)
            if (!kdms_copy_segment_data(src_obj, segments[i],
                                        dest_obj, segments[i]))
	    {
	       for (j = 0; j < nsegments; j++)
		  kfree_and_NULL(segments[j]);
	       kfree_and_NULL(segments);
	       
               return (FALSE);
	    }
	 
      }
   }

   /*-- copy the object level attributes if desired --*/
   if (copy_attributes)
      if (!kdms_copy_segment_attributes(src_obj, NULL, dest_obj, NULL))
      {
	 for (j = 0; j < nsegments; j++)
	    kfree_and_NULL(segments[j]);
	 kfree_and_NULL(segments);
	 return (FALSE);
      }
   
   for (j = 0; j < nsegments; j++)
      kfree_and_NULL(segments[j]);
   kfree_and_NULL(segments);

   return (TRUE);
}

