 /*
  * Khoros: $Id: cast.c,v 1.1.2.1 1997/03/18 23:18:47 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: cast.c,v 1.1.2.1 1997/03/18 23:18:47 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		   Geometry Services Data Code
   >>>>
   >>>>   Public:	
   >>>>			kgeom_new_location_dim()
   >>>>			kgeom_new_layout()
   >>>>			kgeom_new_has_alpha()
   >>>>  Private:
   >>>>			kgeom_new_primitive_location_dim()
   >>>>			kgeom_new_primitive_layout()
   >>>>			kgeom_new_primitive_has_alpha()
   >>>>   Static:
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/************************************************************
*
*  Routine Name: kgeom_new_location_dim - cast a geometry object to one with a new location dimensionality
*
*       Purpose: This function will "cast" an existing geometry object
*		 to have a new location dimensionality.  If the requested
*		 location dimensionality is different than the location 
*		 dimensionality of the provided object, then a new object 
*	 	 is returned which matches the original object, but has
*		 the requested location dimensionality.
*
*		 Specifically, this routine will adjust the size of
*		 the location vectors in each primitive of the given
*		 object to the requested location dimensionailty.
*
*		 This function is intended to simplify processing.
*		 For example, a routine which only wants 2D location
*		 data can use this function to guarantee that the
*		 geometry it will process has only 2D data.
*		 
*		 The cast geometry object is a distinct from the
*		 original geometry object; it is constructed by first
*		 making a copy of the original geometry object and
*		 then resizing all the location vectors.  The original
*		 object will be destroyed after the copy.  If there
*		 are no changes necessary, then the original object will
*		 be returned.
*
*         Input: object - geometry object to cast
*		 new_location_dim - new location dimensionality to cast to
*        Output: none
*       Returns: a pointer to the cast object on success, 
*                NULL on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_new_location_dim(
   kgeom_object *object, 
   int           new_location_dim)
{
   kgeom_object    *new_object;
   kgeom_primitive *primitive;
   int              number, i;
   
   if (object == NULL)
      return NULL;

   if (new_location_dim == object->location_dim)
      return object;

   new_object = kgeom_copy_object(object, NULL);
   new_object->location_dim = new_location_dim;

   number = kgeom_number_primitives(new_object);

   for (i = 0; i < number; i++)
   {
      primitive = kgeom_get_primitive(new_object, i);
      kgeom_new_primitive_location_dim(object, primitive, new_location_dim);
   }
   
   kgeom_blast_object(object);

   return new_object;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_new_primitive_location_dim
|
|       Purpose: This routine will resize the location data in the
|		 the given primitive to the requested location
|		 dimensionailty.
|
|         Input: object    - object to which primitive is associated
|		 primitive - primitive to modify
|		 new_location_dim - new location dimensionality to use
|
|        Output: none
|
|       Returns: TRUE on success, FALSE on failure
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
| Modifications:
|
------------------------------------------------------------*/
void
kgeom_new_primitive_location_dim(
   kgeom_object    *object,
   kgeom_primitive *primitive,
   int              new_location_dim)
{
   int nlocs, nnorms;
   int o = object->location_dim;
   int n = new_location_dim;

   /*-- locations and normals are affected by the location_dim --*/
   nlocs  = kgeom_number(primitive, KPER_VERTEX);
   nnorms = kgeom_number(primitive, object->layout);

   switch (primitive->type)
   {
      case KGEOM_POLYLINE_DISJOINT :
      case KGEOM_POLYLINE_CONNECTED :
	 {
	    kgeom_polyline *p = (kgeom_polyline *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_TRIANGLES_DISJOINT :
      case KGEOM_TRIANGLES_CONNECTED :
	 {
	    kgeom_triangles *p = (kgeom_triangles *) primitive;
	    p->locs  = kgeom_resize_vectors(p->locs,  nlocs, o, n);
	    p->norms = kgeom_resize_vectors(p->norms, nnorms, o, n);
	 } break;
      case KGEOM_SPHERES :
	 {
	    kgeom_spheres *p = (kgeom_spheres *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_TEXT :
	 {
	    kgeom_text *p = (kgeom_text *) primitive;
	    p->locs    = kgeom_resize_vectors(p->locs,    nlocs, o, n);
	    p->offsets = kgeom_resize_vectors(p->offsets, nlocs, o, n);
	    p->ups     = kgeom_resize_vectors(p->ups,     nlocs, o, n);
	    p->flows   = kgeom_resize_vectors(p->flows,   nlocs, o, n);
	    } break;
      case KGEOM_CYLINDERS :
	 {
	    kgeom_cylinders *p = (kgeom_cylinders *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_DIRECTED_POINTS :
	 {
	    kgeom_directed_points *p = (kgeom_directed_points *) primitive;
	    p->locs  = kgeom_resize_vectors(p->locs,  nlocs,  o, n);
	    p->norms = kgeom_resize_vectors(p->norms, nnorms, o, n);
	 } break;
      case KGEOM_MARKERS :
	 {
	    kgeom_markers *p = (kgeom_markers *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_POLYGON :
	 {
	    kgeom_polygon *p = (kgeom_polygon *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_RECTANGLES :
	 {
	    kgeom_rectangles *p = (kgeom_rectangles *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_ELLIPSES :
	 {
	    kgeom_ellipses *p = (kgeom_ellipses *) primitive;
	    p->locs = kgeom_resize_vectors(p->locs, nlocs, o, n);
	 } break;
      case KGEOM_QUADMESH :
	 {
	    kgeom_quadmesh *p = (kgeom_quadmesh *) primitive;
	    p->locs  = kgeom_resize_vectors(p->locs,  nlocs,  o, n);
	    p->norms = kgeom_resize_vectors(p->norms, nnorms, o, n);
	 } break;
      case KGEOM_OCTMESH :
	 {
	    kgeom_octmesh *p = (kgeom_octmesh *) primitive;
	    p->locs  = kgeom_resize_vectors(p->locs,  nlocs,  o, n);
	    p->norms = kgeom_resize_vectors(p->norms, nnorms, o, n);
	 } break;
   }
   return;
}

/************************************************************
*
*  Routine Name: kgeom_new_layout - cast a geometry object to one with a new layout
*
*       Purpose: This function will "cast" an existing geometry object
*		 to have a new layout.  If the requested layout is 
*		 different than the layout of the provided object, then 
*		 a new object is returned which matches the original 
*		 object, but has the requested layout.
*
*		 Specifically, this routine will adjust the number of
*		 normal and color vectors contained in each primitive
*		 of the given object to match the requested layout.
*
*		 This function is intended to simplify processing.
*		 For example, a routine which only wants KPER_VERTEX
*		 data can use this function to guarantee that the
*		 geometry it will process has only KPER_VERTEX data.
*		 
*		 The cast geometry object is a distinct from the
*		 original geometry object; it is constructed by first
*		 making a copy of the original geometry object and
*		 then adjusting the number of color and normal
*		 vectors. The original object will be destroyed after
*		 the copy.  If there are no changes necessary, then
*		 the original object will be returned.
*
*         Input: object - geometry object to cast
*		 new_location_dim - new location dimensionality to cast to
*        Output: none
*       Returns: a pointer to the cast object on success, 
*                NULL on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_new_layout(
   kgeom_object *object, 
   int           new_layout)
{
   kgeom_object    *new_object;
   kgeom_primitive *primitive;
   int              number, i;
   
   if (object == NULL)
      return NULL;

   if (new_layout == object->layout)
      return object;

   new_object = kgeom_copy_object(object, NULL);
   new_object->layout = new_layout;

   number = kgeom_number_primitives(new_object);

   for (i = 0; i < number; i++)
   {
      primitive = kgeom_get_primitive(new_object, i);
      kgeom_new_primitive_layout(object, primitive, new_layout);
   }
   
   kgeom_blast_object(object);

   return new_object;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_new_primitive_layout
|
|       Purpose: This routine will adjust the number of normal and
|		 color vectors contained in the given primitive 
|		 to match the requested layout.
|
|         Input: object    - object to which primitive is associated
|		 primitive - primitive to modify
|		 new_layout - new layout to use
|
|        Output: none
|
|       Returns: TRUE on success, FALSE on failure
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
| Modifications:
|
------------------------------------------------------------*/
void
kgeom_new_primitive_layout(
   kgeom_object    *object,
   kgeom_primitive *primitive,
   int              new_layout)
{
   int csize = (object->has_alpha) ? 4 : 3;
   int nsize = object->location_dim;
   int o, n;

   /*-- colors and normals are affected by the layout --*/
   o = kgeom_number(primitive, object->layout);
   n = kgeom_number(primitive, new_layout);

   switch (primitive->type)
   {
      case KGEOM_POLYLINE_DISJOINT :
      case KGEOM_POLYLINE_CONNECTED :
	 {
	    kgeom_polyline *p = (kgeom_polyline *) primitive;
	    p->cols = kgeom_resample_vectors(p->cols, csize, o, n);
	 } break;
      case KGEOM_TRIANGLES_DISJOINT :
      case KGEOM_TRIANGLES_CONNECTED :
	 {
	    kgeom_triangles *p = (kgeom_triangles *) primitive;
	    p->cols  = kgeom_resample_vectors(p->cols, csize, o, n);
	    p->norms = kgeom_resample_vectors(p->norms, nsize, o, n);
	 } break;
      case KGEOM_CYLINDERS :
	 {
	    kgeom_cylinders *p = (kgeom_cylinders *) primitive;
	    p->cols = kgeom_resample_vectors(p->cols, csize, o, n);
	 } break;
      case KGEOM_QUADMESH :
	 {
	    kgeom_quadmesh *p = (kgeom_quadmesh *) primitive;
	    p->cols  = kgeom_resample_vectors(p->cols, csize, o, n);
	    p->norms = kgeom_resample_vectors(p->norms, nsize, o, n);
	 } break;
      case KGEOM_OCTMESH :
	 {
	    kgeom_octmesh *p = (kgeom_octmesh *) primitive;
	    /* !$#@! unsigned byte color data for octmeshes! */
	    /* p->cols  = kgeom_resample_vectors(p->cols, csize, o, n); */
	    p->norms = kgeom_resample_vectors(p->tcoords, nsize, o, n);
	 } break;
   }
   return;
}

/************************************************************
*
*  Routine Name: kgeom_new_has_alpha - cast a geometry object to one with a new has_alpha
*
*       Purpose: This function will "cast" an existing geometry object
*		 to have a new value of has_alpha.  If the requested 
*		 has_alpha value is different than the has_alpha value 
*		 of the provided object, then a new object is returned 
*		 which matches the original object, but has the requested 
*		 has_alpha value.
*
*		 Specifically, this routine will adjust the size of
*		 the color vectors contained in each primitive of the
*		 given object to match the requested has_alpha value.
*
*		 This function is intended to simplify processing.
*		 For example, a routine which only wants color data
*		 with alpha values can use this function to guarantee
*		 that the geometry it will process has only
*		 color data with alpha values.
*		 
*		 The cast geometry object is a distinct from the
*		 original geometry object; it is constructed by first
*		 making a copy of the original geometry object and
*		 then adjusting the size of color vectors so that they
*		 include an alpha component.  The original object will
*		 be destroyed after the copy.  If there are no changes
*		 necessary, then the original object will be returned.
*
*         Input: object - geometry object to cast
*		 new_has_alpha - new has_alpha value to cast to
*        Output: none
*       Returns: a pointer to the cast object on success, 
*                NULL on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_new_has_alpha(
   kgeom_object *object, 
   int           new_has_alpha)
{
   kgeom_object    *new_object;
   kgeom_primitive *primitive;
   int              number, i;
   
   if (object == NULL)
      return NULL;

   if (new_has_alpha == object->has_alpha)
      return object;

   new_object = kgeom_copy_object(object, NULL);
   new_object->has_alpha = new_has_alpha;

   number = kgeom_number_primitives(new_object);

   for (i = 0; i < number; i++)
   {
      primitive = kgeom_get_primitive(new_object, i);
      kgeom_new_primitive_has_alpha(object, primitive, new_has_alpha);
   }
   
   kgeom_blast_object(object);

   return new_object;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_new_primitive_has_alpha
|
|       Purpose: This routine will adjust the size of the
|		 color vectors contained in the given primitive 
|		 to match the requested has_alpha value.
|
|         Input: object    - object to which primitive is associated
|		 primitive - primitive to modify
|		 new_has_alpha - new has_alpha value to use
|
|        Output: none
|
|       Returns: TRUE on success, FALSE on failure
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
| Modifications:
|
------------------------------------------------------------*/
void
kgeom_new_primitive_has_alpha(
   kgeom_object    *object,
   kgeom_primitive *primitive,
   int              new_has_alpha)
{
   int ncols;
   int o = (object->has_alpha) ? 4 : 3;
   int n = (new_has_alpha)     ? 4 : 3;

   /*-- locations and normals are affected by the location_dim --*/
   ncols  = kgeom_number(primitive, object->layout);

   switch (primitive->type)
   {
      case KGEOM_POLYLINE_DISJOINT :
      case KGEOM_POLYLINE_CONNECTED :
	 {
	    kgeom_polyline *p = (kgeom_polyline *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_TRIANGLES_DISJOINT :
      case KGEOM_TRIANGLES_CONNECTED :
	 {
	    kgeom_triangles *p = (kgeom_triangles *) primitive;
	    p->cols  = kgeom_resize_vectors(p->cols,  ncols, o, n);
	 } break;
      case KGEOM_SPHERES :
	 {
	    kgeom_spheres *p = (kgeom_spheres *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_TEXT :
	 {
	    kgeom_text *p = (kgeom_text *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	    } break;
      case KGEOM_CYLINDERS :
	 {
	    kgeom_cylinders *p = (kgeom_cylinders *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_DIRECTED_POINTS :
	 {
	    kgeom_directed_points *p = (kgeom_directed_points *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols,  o, n);
	 } break;
      case KGEOM_MARKERS :
	 {
	    kgeom_markers *p = (kgeom_markers *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_POLYGON :
	 {
	    kgeom_polygon *p = (kgeom_polygon *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_RECTANGLES :
	 {
	    kgeom_rectangles *p = (kgeom_rectangles *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_ELLIPSES :
	 {
	    kgeom_ellipses *p = (kgeom_ellipses *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols, o, n);
	 } break;
      case KGEOM_QUADMESH :
	 {
	    kgeom_quadmesh *p = (kgeom_quadmesh *) primitive;
	    p->cols = kgeom_resize_vectors(p->cols, ncols,  o, n);
	 } break;
      case KGEOM_OCTMESH :
	 {
	    kgeom_octmesh *p = (kgeom_octmesh *) primitive;
	    /* !$#@! unsigned byte color data for octmeshes! */
	    /* p->cols  = kgeom_resize_vectors(p->cols, ncols,  o, n); */
	 } break;
   }
   return;
}
