 /*
  * Khoros: $Id: data.c,v 1.1.2.5 1997/04/17 22:12:37 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: data.c,v 1.1.2.5 1997/04/17 22:12:37 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		   Geometry Services Data Code
   >>>>
   >>>>   Public:	
   >>>>  Private:
   >>>>			kgeom_number()
   >>>>			kgeom_alloc()
   >>>>			kgeom_dup()
   >>>>			kgeom_resize_vectors()
   >>>>			kgeom_resample_vectors()
   >>>>   Static:
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_number
|
|       Purpose: This routine, for a given primitive and layout, 
|		 will determine how many vectors are present.
|
|		 The number of vertices is returned.
|		
|    Written By: Steve Kubica
|          Date: December 27, 1996
|
------------------------------------------------------------*/
int
kgeom_number(
   kgeom_primitive  *primitive,
   int               layout)
{
    int number;
    int width, height, depth;

    /*------------------------------------------------------
    |  the mesh and texture primitives don't use nverts to
    |  track the number of vertices.  we have to treat them
    |  in terms of width, height, and possibly depth.
    \------------------------------------------------------*/
    if (primitive->type == KGEOM_QUADMESH || 
	primitive->type == KGEOM_TEXTURE2D)
    {
       /*-- use quadmesh to get at the width and height fields --*/
       kgeom_quadmesh *p = (kgeom_quadmesh *) primitive;

       width  = p->width;
       height = p->height;
    }
    else if (primitive->type == KGEOM_OCTMESH || 
	     primitive->type == KGEOM_TEXTURE2D)
    {
       /*-- use octmesh to get at the width, height, and depth fields --*/
       kgeom_octmesh *p = (kgeom_octmesh *) primitive;

       width  = p->width;
       height = p->height;
       depth  = p->depth;
    }
    else /*-- all other primitives use nverts --*/ 
    {
       /* use polyline to get at the nverts field */
       kgeom_polyline *p = (kgeom_polyline *) primitive;
       
       number = p->nverts;
    }

    /* check for each primitive which is affected by layout */
    switch (primitive->type)
    {
       case KGEOM_POLYLINE_DISJOINT : 
	  if (layout == KPER_LINE) number /= 2;
	  break;
       case KGEOM_POLYLINE_CONNECTED : 
	  if (layout == KPER_LINE) number -= 1;
	  break;
       case KGEOM_TRIANGLES_DISJOINT :
	  if (layout == KPER_FACE) number /= 3;
	  break;
       case KGEOM_TRIANGLES_CONNECTED :
	  if (layout == KPER_LINE)  number -= 1;
	  if (layout == KPER_FACE)  number -= 2;
	  break;
       case KGEOM_CYLINDERS :
	  if (layout == KPER_LINE)  number /= 2;
	  break;
       case KGEOM_QUADMESH :
	  if (layout == KPER_VERTEX) 
	     number = width * height;
	  else
	     number = (width - 1) * (height - 1);
	  break;
       case KGEOM_OCTMESH :
	  if (layout == KPER_VERTEX)
	     number = width * height * depth;
	  else
	     number = (width - 1) * (height - 1) * (depth - 1);
	  break;
       case KGEOM_TEXTURE2D :
	  number = width * height;
	  break;
       case KGEOM_TEXTURE3D :
	  number = width * height * depth;
	  break;
    }
    
    return number;
}

/* 
 *    ================================================================== 
 *    Geometry Custom Data Handlers 
 *    ==================================================================
 */

/*-----------------------------------------------------------
|  Routine Name: kgeom_alloc - allocate a block of data
|
|         Input: number - number of points to allocate
|
|        Output: none
|
|       Returns: pointer to the allocated data
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
------------------------------------------------------------*/
float *
kgeom_alloc(int number_of_points)
{
   float *data = NULL;

   if (number_of_points == 0)
      return NULL;  /* well, it has 0 points in it, doesn't it? */

   if ((data = (float *) kmalloc(number_of_points * sizeof(float))) == NULL)
   {
      kset_errno(KMEMORY_ALLOCATION);
      return NULL;
   }
   
   return data;
}

/*-----------------------------------------------------------
|  Routine Name: kgeom_dup - duplicate a block of data
|
|         Input: number - number of points to duplicate
|		 data   - pointer to the data to be duplicated
|
|        Output: none
|
|       Returns: pointer to the duplicated data
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
------------------------------------------------------------*/
float *
kgeom_dup(float *data, int number_of_points)
{
   float *rdata;
   
   if (data == NULL) /*-- want to duplicate whatever we get --*/
      return NULL;

   if ((rdata = (float *) kdupalloc((const kaddr) data, 
				    sizeof(float)*number_of_points)) == NULL)
   {
      kset_errno(KMEMORY_ALLOCATION);
      return NULL;
   }
   return rdata;
}

/* 
 *    ================================================================== 
 *    Data Presenters
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_resize_vectors - resize the vectors in a list
|
|         Input: old - old block of floating point data
|		 npoints - number of vectors in data
|		 old_size - old vector size
|		 new_size - new vector size 
|
|        Output: none
|
|       Returns: pointer to newly allocated block, old is freed
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
------------------------------------------------------------*/
float * 
kgeom_resize_vectors(
   float *old,
   int    npoints,
   int    old_size,
   int    new_size)
{
   float *tmp;
   float *n, *o;
   int    growing;
   int    min_size;
   int    i, j;

   if (old == NULL) 
      return NULL;
   
   tmp = (float *) kcalloc(sizeof(float), (size_t) npoints * new_size);

   growing = (new_size > old_size);
   min_size = (growing) ? old_size : new_size;
   
   for (i = 0, o = old, n = tmp; i < npoints; i++)
   {
      for (j = 0; j < min_size; j++)
	 *n++ = *o++;

      if (growing)
	 for (j = old_size; j < new_size; j++) n++;
      else
	 for (j = new_size; j < old_size; j++) o++;
   }

   kfree(old);

   return tmp;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_resample_vectors - change the number of vectors in list
|
|         Input: old - old block of floating point vector data
|		 size - size of each vector
|		 old_number - old number of vectors
|		 new_number - new number of vectors
|
|        Output: none
|
|       Returns: pointer to newly allocated block, old is freed
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
------------------------------------------------------------*/
float * 
kgeom_resample_vectors(
   float *old,
   int    size,
   int    old_number,
   int    new_number)
{
   float index, increment, *new1;
   int i;
   
   if (old == NULL) 
      return NULL;

   new1 = (float *) kmalloc((size_t) new_number * size * sizeof(float));

   increment = old_number / new_number;

   /*-- blind resampling of vectors --*/
   for (i = 0, index = 0.0; i < new_number; i++, index += increment) 
   {
      new1[i]   = old[((int)index)];
      new1[i+1] = old[((int)index)+1];
      new1[i+2] = old[((int)index)+2];
   }

   kfree(old);

   return new1;
}
