 /*
  * Khoros: $Id: object.c,v 1.1.2.5 1997/05/27 21:25:04 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: object.c,v 1.1.2.5 1997/05/27 21:25:04 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		   Geometry Services Object Code
   >>>>
   >>>>   Public:	
   >>>>			kgeom_new_object()
   >>>>			kgeom_blast_object()
   >>>>
   >>>>			kgeom_write_header()
   >>>>			kgeom_read_header()
   >>>>			kgeom_done_with_file()
   >>>>
   >>>>			kgeom_write_object()
   >>>>			kgeom_start_writing_object()
   >>>>			kgeom_write_object_and_primitives()
   >>>>			kgeom_finish_writing_object()
   >>>>
   >>>>			kgeom_read_object()
   >>>>			kgeom_start_reading_object()
   >>>>			kgeom_read_object_and_primitives()
   >>>>			kgeom_finish_reading_object()
   >>>>
   >>>>			kgeom_number_primitives()
   >>>>			kgeom_add_primitive()
   >>>>			kgeom_get_primitive()
   >>>>			kgeom_remove_primitive()
   >>>>
   >>>>			kgeom_get_root()
   >>>>  Private:
   >>>>   Static:
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

static unsigned char magic[] = { 'k', 'g', 'e', 'o', 'm' };
static unsigned char version = 1;

/*
 *    ================================================================== 
 *    kgeom_object constructor and destructor
 *    ==================================================================
 */

/************************************************************
*
*  Routine Name: kgeom_new_object - construct a new geometry object
*
*       Purpose: This function will construct a new geometry object
*                structure and initialize all of its fields with their
*                default values.  
*
*		 Once constructed, the fields of this geometry object
*		 can be assigned specific values.  For a complete
*		 explanation of these fields, see the
*		 \fIkgeom_object\fP man page.
*
*		 \fIGeometry Primitives and Primitive Lists\fP
*
*		 A geometry primitive contains geometric data which
*		 describes a shape in space. Lines, triangles, and
*		 spheres, are all considered to be geometric
*		 primitives. For a complete description of all the
*		 geometry primitives, see the \fIkgeom_primitive\fP
*		 man page.
*
*		 Each geometry object contains a list of such geometry
*		 primitives.  New primitives can be created using the
*		 function \f(CWkgeom_new_primitive()\fP.  
*		 Once a new primitive has been created and its data
*		 assigned, it can be added to the primitive list of
*		 the object using the function
*		 \f(CWkgeom_add_primitive()\fP.  A corresponding
*		 function \f(CWkgeom_remove_primitive()\fP allows you
*		 to remove a primitive from the primitive list of an
*		 object. 
*
*		 The number of primitives which have been added to the
*		 primitive list of an object is returned by the
*		 function \f(CWkgeom_number_primitives()\fP.  A
*		 primitive at a specific position in the primitive
*		 list can be retrieved using the function
*		 \f(CWkgeom_get_primitive()\fP.
*
*		 The primitive functions should be the only means used
*		 for accessing the primitives on a primitive list.
*		 Even though the primitive list of the geometry object
*		 could be acccessed directly, it should be considered
*		 to be private.  By using the primitive list function
*		 calls, you can be certain to maintain the integrity
*		 of the primitive list.
*		 
*		 \fIReading and Writing Geometry Objects\fP
*
*		 A geometry object can be written out to a file, or
*		 other transport using the function
*		 \f(CWkgeom_write_object()\fP.  This function will
*		 write out all the object and primitive information to
*		 a specified file.
*
*		 Geometry that has been stored in a file can be read
*		 back with the function \f(CWkgeom_read_object()\fP.
*		 This function will construct a new object into which
*		 it will read the geometry; it is not necessary to
*		 pre-construct an object with the
*		 \f(CWkgeom_new_object()\fP function prior to reading.
*
*		 \fIOther Geometry Object Functions\fP
*
*		 A geometry object can be copied using the function
*		 \f(CWkgeom_copy_object()\fP.  
*
*		 A geometry object can be destroyed with a call to the
*		 function \f(CWkgeom_geom_blast_object()\fP.  This
*		 will destroy the object and all of its primitives,
*		 freeing all associated memory.  Be careful not to
*		 access any of the geometry object's data after
*		 blasting it.
*
*      Examples: This simple code illustrates how a
*                new geometry object would be constructed :
*       
*		 .begin code
*  		  kgeom_object *obj;
*  
*  		  obj = kgeom_new_object();
*		 .end code
*		 
*         Input: none
*        Output: none
*       Returns: a pointer to the constructed object on success, 
*                NULL on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_new_object(void)
{
   kgeom_object *object;

   float color [3]  = KGEOM_COLOR_DEFAULT;
   float acolor[3]  = KGEOM_AMBIENT_COLOR_DEFAULT;
   float dcolor[3]  = KGEOM_DIFFUSE_COLOR_DEFAULT;
   float scolor[3]  = KGEOM_SPECULAR_COLOR_DEFAULT;
   float matrix[16] = KGEOM_MATRIX_DEFAULT;
   

   if ((object = (kgeom_object *) kcalloc((size_t) 1,
			sizeof(kgeom_object))) == NULL)
   {
      kset_errno(KMEMORY_ALLOCATION);
      return NULL;
   }

   /*-- initialize all default values --*/
   if ((object->primitives = (kgeom_primitive **)
	kmalloc(sizeof(kgeom_primitive *))) == NULL)
   {
      kset_errno(KMEMORY_ALLOCATION);
      return NULL;
   }
   object->primitives[0] = NULL;

   object->name 	     = NULL;
   object->has_bounding_box  = FALSE;
   object->has_center        = FALSE;
   object->has_color         = FALSE;
   object->has_alpha         = FALSE;

   kmemcpy(object->ambient_color,  acolor, sizeof(float) * 3);
   kmemcpy(object->diffuse_color,  dcolor, sizeof(float) * 3);
   kmemcpy(object->specular_color, scolor, sizeof(float) * 3);
   kmemcpy(object->color,          color,  sizeof(float) * 3);
   kmemcpy(object->matrix,         matrix, sizeof(float) * 16);

   object->specular_exponent = KGEOM_SPECULAR_EXPONENT_DEFAULT;
   object->opacity           = 1.0;
   object->visible           = TRUE;
   object->modeling_space    = KGEOM_MODELING_SPACE_DEFAULT;
   object->location_dim      = 3;
   object->texture_coord_dim = 2;

   object->fid = -1;
   object->parent = NULL;

   return object;
}

/************************************************************
*
*  Routine Name: kgeom_blast_object - free a geometry object
*
*       Purpose: This function will destroy the given geometry 
*                object, first freeing all the primitives on the
*		 object's primitive list, and then freeing the 
*		 object itself.
*
*		 Note that geometry services will free \fIall\fP the
*		 primitives and data pointers within the geometry
*		 object.  Since geometry services will try to free all
*		 memory pointed to by the geometry object, be careful
*		 not to place any static primitives or point to static
*		 arrays of data from within the geometry object.  If
*		 geometry services tries to free a statically
*		 allocated pieces of memory, it will result in a fatal
*		 error.
*
*		 If you wish to keep any component of the geometry
*		 object in memory, simply remove it from the geometry
*		 object prior to calling this function.  For instance,
*		 to keep a primitive from an object in memory after
*		 the object has been destroyed, that primitive could
*		 first be removed with the
*		 \f(CWkgeom_remove_primitive()\fP function.
*		 Similarly, to keep a particular array of data from a
*		 primitive in memory after the associated primitive
*		 has been destroyed, the pointer to that array could
*		 be assigned to NULL.  This routine will only free the
*		 primitives and data pointers which are seen within
*		 the object.
*
*         Input: object - geometry object to destroy
*        Output: none
*       Returns: 
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
void
kgeom_blast_object(kgeom_object *object)
{
   int i = 0;
   
   if (object == NULL)
      return;

   if (object->fid != -1)
      kgeom_done_with_file(object);

   /*-- destroy any primitives that we have --*/
   while (object->primitives[i] != NULL)
      kgeom_blast_primitive((kgeom_primitive *) object->primitives[i]),  i++;

   kfree(object->primitives);
   kfree(object->name);
   kfree(object);
   
   return;
}

/*
 *    ================================================================== 
 *    kgeom_object copy function
 *    ==================================================================
 */

/************************************************************
*
*  Routine Name: kgeom_copy_object - copy a geometry object
*
*       Purpose: This function will copy the given geometry 
*                object, creating a duplicate of the object as well as
*                of all the object's primitives.  All the data contained in 
*		 each of the object's primitives will be copied as well.
*
*		 If desired, a preallocated object can be provided as
*		 the destination for the copy.  Any data or primitives
*		 which may have existed in the destination prior to
*		 calling this routine will be destroyed.
*
*		 If no destination object is provided, one will be
*		 allocated and returned.
*		 
*         Input: object - geometry object to copy
*        Output: new_object - a pointer to geometry object that will
*			      serve as a destination for the copy. If
*			      NULL, then a new destination object will
*			      be allocated.
*       Returns: copy of object
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_copy_object(kgeom_object *object, kgeom_object *new_object)
{
   kgeom_primitive *primitive;
   kgeom_primitive *new_primitive;
   int number, i;
   
   if (object == NULL)
      return NULL;

   /*-- make sure the provided new object is empty --*/
   if (new_object != NULL)
   {
      kfree_and_NULL(new_object->name);

      /*-- destroy any primitives that we have --*/
      number = kgeom_number_primitives(new_object);
      for (i = 0; i < number; i++) 
      {
	 kgeom_blast_primitive(new_object->primitives[i]);
	 new_object->primitives[i] = NULL;
      }
   }
   else
      new_object = kgeom_new_object();

   /*-- copy all object fields --*/
   new_object->name = kstrdup(object->name);

   new_object->has_bounding_box = object->has_bounding_box;
   new_object->has_center       = object->has_center;
   new_object->has_color        = object->has_color;
   new_object->has_alpha        = object->has_alpha;

   kmemcpy(new_object->ambient_color, object->ambient_color, sizeof(float)*3);
   kmemcpy(new_object->diffuse_color, object->diffuse_color, sizeof(float)*3);
   kmemcpy(new_object->specular_color,object->specular_color,sizeof(float)*3);
   kmemcpy(new_object->color,         object->color,         sizeof(float)*3);
   kmemcpy(new_object->matrix,        object->matrix,        sizeof(float)*16);

   new_object->specular_exponent = object->specular_exponent;
   new_object->opacity           = object->opacity;
   new_object->visible           = object->visible;
   new_object->modeling_space    = object->modeling_space;
   new_object->location_dim      = object->location_dim;
   new_object->texture_coord_dim = object->texture_coord_dim;

   /*-- copy over all primitives --*/
   number = kgeom_number_primitives(object);
   for (i = 0; i < number; i++)
   {
      primitive     = kgeom_get_primitive(object, i);
      new_primitive = kgeom_copy_primitive(object, primitive, NULL);
      kgeom_add_primitive(new_object, new_primitive);
   }

   return new_object;
}

/*
 *    ================================================================== 
 *    kgeom_object header writing and reading
 *    ==================================================================
 */

/************************************************************
*
*  Routine Name: kgeom_write_header - write a geometry object header
*
*       Purpose: This function will write a header for a given 
*                geometry object and keep track of the open fid in the
*                geometry object structure.  This fid will be used
*		 for subsequent write calls for this geometry
*		 object.
*
*		 In general, the \f(CWkgeom_write_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for writing a geometry object.  It should
*		 only be used when you wish to write a geometry object
*		 primitive-by-primitive.  More information on this is
*		 available in the \f(CWkgeom_write_primitive()\fP man
*		 page.
*
*		 A geometry object, when written to a file, will
*		 basically consist of a geometry header, geometry
*		 object data, and geometry primitive data.  This
*		 routine allows you to write only the geometry header.
*
*		 Functionally, the \f(CWkgeom_write_object()\fP routine
*		 is equivalent to a call to
*		 \f(CWkgeom_write_header\fP, followed by calls to
*		 \f(CWkgeom_write_object_and_primitives()\fP and 
*		 \f(CWkgeom_done_writing()\fP.
*
*         Input: object - geometry object to write
*		 filename - filename to write header to
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_write_header(
   kgeom_object *object,
   char         *filename)
{
   kubyte mach = (unsigned char) kmach_local();

   if (object == NULL)
   {
      kset_errno(KNULL_PARAMETER);
      return FALSE;
   }

   if (filename == NULL)
   {
      kset_errno(KNULL_PARAMETER);
      return FALSE;
   }
   
   if ((object->fid = kopen(filename, 
			    KOPEN_WRONLY | KOPEN_CREAT | KOPEN_TRUNC, 
			    0666)) == -1)
   {
      kset_errno(KCANT_OPEN_OUTFILE);
      return FALSE;
   }

   /*-- at this point, we have a valid fid. write the header --*/
   kwrite_ubyte(object->fid, magic, (size_t) 5);
   kwrite_ubyte(object->fid, &mach, (size_t) 1);

   /*-- header is now written and object->fid has open file id --*/
   return TRUE;
}

/************************************************************
*
*  Routine Name: kgeom_read_header - read a geometry object header
*
*       Purpose: This routine will open a file and try to read a 
*		 geometry header.  If the file does not contain a 
*		 valid geometry header, then it will be closed and
*		 nothing will be returned.  If it does contain a valid
*		 header, then a new object will be returned 
*		 containing the valid open fid.
*
*		 In general, the \f(CWkgeom_read_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for reading a geometry object.  It should
*		 only be used when you wish to read a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_read_primitive()\fP man
*		 page.
*
*		 A geometry object, when written to a file basically
*		 consists of a geometry header, geometry object data,
*		 and geometry primitive data.  This routine allows you
*		 to read only the geometry header.
*
*		 Functionally, the \f(CWkgeom_read_object()\fP routine
*		 is equivalent to a call to
*		 \f(CWkgeom_read_header\fP, followed by calls to
*		 \f(CWkgeom_read_object_and_primitives()\fP and 
*		 \f(CWkgeom_done_reading()\fP.
*
*         Input: filename - filename to read header from.
*        Output: 
*       Returns: new geometry object structure on success, NULL otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_read_header(char *filename)
{
   kgeom_object *object;
   kubyte m[5], mach;
   int    type;
   

   if (filename == NULL)
   {
      kset_errno(KNULL_PARAMETER);
      return NULL;
   }

   object = kgeom_new_object();

   if ((object->fid = kopen(filename, KOPEN_RDONLY, 0666)) == -1)
   {
      kset_errno(KCANT_OPEN_INFILE);
      kgeom_blast_object(object);
      return NULL;
   }

   /*-- get the magic number -- if it matches, we're in luck --*/
   kread_ubyte(object->fid, &m, (size_t) 5);

   if (kmemcmp(magic, &m, (size_t) 5) != 0)
   {
      kset_errno(KINVALID_INPUT);
      kgeom_blast_object(object);
      return NULL;
   }

   /*-- okay, it's likely a geometry file... get the mach type --*/
   kread_ubyte(object->fid, &mach, (size_t) 1);
   kfile_setmachtype(object->fid, (int) mach);

   /*-- we need to read the object type here (trust me) --*/
   kread_int(object->fid, &type, (size_t) 1); 
   /*-- must read this here ... 
        read primitive must read type before read_object ... --*/

   return object;
}


/*-----------------------------------------------------------
|
|  Routine Name: kgeom_done_with_file - close a file for a given object
|
|       Purpose: This routine will close the file associated
|		 with the provided geometry object.  
|
|		 For clarity, the macro functions kgeom_done_reading() 
|		 and kgeom_done_writing() should be used instead of 
|		 this function.
|	
|         Input: object - object with fid that we're done with
|
|        Output: none
|
|       Returns: TRUE on success, FALSE on failure
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
| Modifications:
|
------------------------------------------------------------*/
int
kgeom_done_with_file(kgeom_object *object)
{
   if (object == NULL || object->fid == -1)
      return TRUE;  /*-- that was easy! --*/

   if (kclose(object->fid) == -1)
   {
      kset_errno(KINTERNAL);
      return FALSE;
   }
   object->fid = -1;

   return TRUE;
}

/*
 *    ================================================================== 
 *    kgeom_object writing
 *    ==================================================================
 */

/************************************************************
*
*  Routine Name: kgeom_write_object - write a geometry object
*
*       Purpose: This function will write the given geometry 
*                object to a specified file.   The object can 
*		 later be read from that file using the 
*		 \f(CWkgeom_read_object()\fP function. 
*
*		 This function will write everything contained in the
*		 geometry object to the file.  All the object
*		 information will be written followed by the
*		 information and data contained in each primitive in
*		 the object's primitive list.
*		
*		 After the object has been written, you may free it
*		 using the \f(CWkgeom_blast_object()\fP function.
*
*      Examples: The following code would be used to write a geometry
*		 object out to the file "data.kgm":
*
*		 .begin code
*		 kgeom_object *geom;
*
*		 // presumably some primitives would be added here 
*
*		 kgeom_write_object(geom, "data.kgm");
*		 .end code  
*
*         Input: object - geometry object to write
*		 filename - filename to write geometry object to.
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 24, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_write_object(kgeom_object *object, char *filename)
{

   if (!kgeom_write_header(object, filename))
      return FALSE;
   
   if (!kgeom_write_object_and_primitives(object))
      return FALSE;

   if (!kgeom_done_writing(object))
      return FALSE;

   return TRUE;
}

/************************************************************
*
*  Routine Name: kgeom_start_writing_object - write the first part of a geometry object
*
*       Purpose: This routine will write the first part of a
*		 geometry object to the open fid.  It should be
*		 followed by a write of all the primitives, and
*		 then by a finish of the object.
*
*		 In general, the \f(CWkgeom_write_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for writing a geometry object.  It should
*		 only be used when you wish to write a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_write_primitive()\fP man
*		 page.
*
*		 The first part of a geometry object contains
*		 information which will be needed to later read the
*		 geometry primitive data.  Specifically, the layout,
*		 location_dim, texture_coord_dim, and has_alpha fields
*		 are written as the first part of the object.
*
*  	         All other information contained within the geometry
*  	         object is written after the primitives by the 
*  	         \f(CWkgeom_finish_writing_object()\fP call.
*
*		 The fid internal to the geometry object specifies the
*		 file to which the the object will be written.  In
*		 general, this fid is set when the geometry header is
*		 written by the \f(CWkgeom_write_header()\fP function.
*
*         Input: object - geometry object to write
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 24, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_start_writing_object(kgeom_object *object)
{
   int s = TRUE;
   int type = KGEOM_OBJECT;

   if (object == NULL)
   {
      kset_errno(KNULL_PARAMETER);
      return FALSE;
   }

   if (object->fid == -1)
   {
      kset_errno(KINVALID_FILE);
      return FALSE;
   }

   /*-- write object type identifier --*/
   CK(kwrite_int(object->fid, &type, (size_t) 1));

   /*-- write the object version number (unused for now) --*/
   CK(kwrite_ubyte(object->fid, &version, (size_t) 1));

   /*-- we'll need this much of the header before we can read anything --*/
   CK(kwrite_int  (object->fid, &object->layout, (size_t) 1));
   CK(kwrite_int  (object->fid, &object->location_dim, (size_t) 1));
   CK(kwrite_int  (object->fid, &object->texture_coord_dim, (size_t) 1));
   CK(kwrite_ubyte(object->fid, &object->has_alpha, (size_t) 1));

   return s;
}

/************************************************************
*
*  Routine Name: kgeom_finish_writing_object - write the last part of a geometry object 
*
*       Purpose: This routine will terminate the primitive list
*		 by writing a -1 to an open fid.  It will then write 
*		 out all the object specific information.
*
*		 In general, the \f(CWkgeom_write_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for writing a geometry object.  It should
*		 only be used when you wish to write a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_write_primitive()\fP man
*		 page.
*	
*		 The first part of a geometry object should have
*		 already been written by a call to
*		 \f(CWkgeom_start_writing_object()\fP.  That call
*		 would have written any information which would be
*		 needed to later read the geometry primitive data.
*		 Specifically, the layout, location_dim,
*		 texture_coord_dim, and has_alpha fields of the object
*		 would already have been written out as the first part
*		 of the object.
*
*  	         All other information contained within the geometry
*  	         object is written after the primitives by this call.
*
*		 The fid internal to the geometry object specifies the
*		 file to which the the object will be written.  In
*		 general, this fid is set when the geometry header is
*		 written by the \f(CWkgeom_write_header()\fP function.
*
*         Input: object - geometry object to write
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 24, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_finish_writing_object(kgeom_object *object)
{
   int s = TRUE;
   int type = -1;
   
   /*-- write the terminator for the primitive list --*/
   kwrite_int(object->fid, &type, (size_t) 1);


   /*-- layout, location_dim, texture_coord_dim were written previously --*/

   /*-- write the object version number (unused for now) --*/
   CK(kwrite_ubyte(object->fid, &version, (size_t) 1));

   /*-- write the object specific information --*/
   CK(kwrite_string(object->fid, &object->name, (size_t) 1));
   CK(kwrite_float (object->fid,  object->ambient_color, (size_t) 3));
   CK(kwrite_float (object->fid,  object->diffuse_color, (size_t) 3));
   CK(kwrite_float (object->fid,  object->specular_color, (size_t) 3));
   CK(kwrite_float (object->fid, &object->specular_exponent, (size_t) 1));
   CK(kwrite_int   (object->fid, &object->modeling_space, (size_t) 1));
   CK(kwrite_int   (object->fid, &object->visible, (size_t) 1));

   CK(kwrite_ubyte (object->fid, &object->has_bounding_box, (size_t) 1));
   CK(kwrite_ubyte (object->fid, &object->has_center, (size_t) 1));
   CK(kwrite_ubyte (object->fid, &object->has_matrix, (size_t) 1));
   CK(kwrite_ubyte (object->fid, &object->has_color, (size_t) 1));
   CK(kwrite_float (object->fid,  object->min, (size_t) 3));
   CK(kwrite_float (object->fid,  object->max, (size_t) 3));
   CK(kwrite_float (object->fid,  object->center, (size_t) 3));
   CK(kwrite_float (object->fid,  object->matrix, (size_t) 16));
   CK(kwrite_float (object->fid,  object->color, (size_t) 3));
   CK(kwrite_float (object->fid, &object->opacity, (size_t) 1));

   return s;
}

/************************************************************
*
*  Routine Name: kgeom_write_object_and_primitives - write a geometry object and all of its primitives
*
*       Purpose: This routine will write an object and all of its
*		 primitives to the open fid specified within the object.
*
*		 In general, the \f(CWkgeom_write_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for writing a geometry object.  It should
*		 only be used when you wish to write a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_write_primitive()\fP man
*		 page.
*		 
*		 This function is equivalent to calling the function
*		 \f(CWkgeom_start_writing_object()\fP, followed by a
*		 call to \f(CWkgeom_write_primitive()\fP for each
*		 primitive on the object's primitive list, followed by
*		 a call to \f(CWkgeom_finish_writing_object()\fP to
*		 terminate the primitive list and write all the
*		 object-specific information.
*
*		 The fid internal to the geometry object specifies the
*		 file to which the the object will be written.  In
*		 general, this fid is set when the geometry header is
*		 written by the \f(CWkgeom_write_header()\fP function.
*
*         Input: object - geometry object to write
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_write_object_and_primitives(kgeom_object *object)
{
   int j = 0;
   
   if (!kgeom_start_writing_object(object))
      return FALSE;
   
   while (object->primitives[j] != NULL)
      if (!kgeom_write_primitive(object, object->primitives[j++]))
	 return FALSE;

   if (!kgeom_finish_writing_object(object))
      return FALSE;

   return TRUE;
}

/*
 *    ================================================================== 
 *    kgeom_object reading
 *    ==================================================================
 */

/************************************************************
*
*  Routine Name: kgeom_read_object - read a geometry object
*
*       Purpose: This function will read an entire geometry object
*                object from the specified file.   
*
*		 If the file does not contain a valid geometry object,
*		 then it will be closed and nothing will be returned.
*		 If it does contain a valid header, then a new object
*		 containing the geometry data from the file will be
*		 returned.
*
*      Examples: The following code would be used to read in a geometry
*		 object from the file "data.kgm":
*
*		 .begin code
*		 kgeom_object *geom;
*
*		 geom = kgeom_read_object("data.kgm");
*		 .end code  
*
*         Input: filename - filename to read geometry from
*        Output: 
*       Returns: pointer to a new geometry object containing the geometry
*		 data contained in the file, NULL otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 24, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_object *
kgeom_read_object(char *filename)
{
   kgeom_object *object = NULL;
   
   if ((object = kgeom_read_header(filename)) == NULL)
      return NULL;
   
   if (!kgeom_read_object_and_primitives(object))
   {
      kgeom_blast_object(object);
      return NULL;
   }

   if (!kgeom_done_reading(object))
   {
      kgeom_blast_object(object);
      return NULL;
   }

   return object;
}


/************************************************************
*
*  Routine Name: kgeom_start_reading_object - read the first part of a geometry object 
*
*       Purpose: This routine will read the first part of a
*		 geometry object from an open fid.  It should be
*		 followed by a read of all the primitives, and
*		 then by a finish of the object.
*
*		 In general, the \f(CWkgeom_read_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for reading a geometry object.  It should
*		 only be used when you wish to read a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_read_primitive()\fP man
*		 page.
*
*		 The first part of a geometry object contains
*		 information which is needed to read the geometry
*		 primitive data.  Specifically, the layout,
*		 location_dim, texture_coord_dim, and has_alpha fields
*		 are read in as the first part of the object.
*
*  	         All other information contained within the geometry
*  	         object is read after the primitives by the 
*  	         \f(CWkgeom_finish_reading_object()\fP call.
*
*		 The fid internal to the geometry object specifies the
*		 file from which the the object will be read.  In
*		 general, this fid is set when the geometry header is
*		 read by the \f(CWkgeom_read_header()\fP function.
*
*         Input: object - geometry object to read
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 24, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_start_reading_object(kgeom_object *object)
{
   int s = TRUE;
   unsigned char v;

   if (object == NULL)
   {
      kset_errno(KNULL_PARAMETER);
      return FALSE;
   }

   if (object->fid == -1)
   {
      kset_errno(KINVALID_FILE);
      return FALSE;
   }

   /*-- read the version number (version is unused for now) --*/
   CK(kread_ubyte (object->fid, &v, (size_t) 1));

   /*-- we'll need this much of the header before we can read anything --*/
   CK(kread_int	 (object->fid, &object->layout, (size_t) 1));
   CK(kread_int	 (object->fid, &object->location_dim, (size_t) 1));
   CK(kread_int	 (object->fid, &object->texture_coord_dim, (size_t) 1));
   CK(kread_ubyte(object->fid, &object->has_alpha, (size_t) 1));

   return TRUE;
}

/************************************************************
*
*  Routine Name: kgeom_finish_reading_object - read the last part of a geometry object 
*
*       Purpose: This routine will read in all the object 
*		 specific information.
*
*		 In general, the \f(CWkgeom_read_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for reading a geometry object.  It should
*		 only be used when you wish to read in a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_read_primitive()\fP man
*		 page.
*	
*		 The first part of a geometry object should have
*		 already been read by a call to
*		 \f(CWkgeom_start_reading_object()\fP.  That call
*		 would have read any information which was 
*		 needed to read the geometry primitive data.
*		 Specifically, the layout, location_dim,
*		 texture_coord_dim, and has_alpha fields of the object
*		 would already have been read in as the first part
*		 of the object.
*
*  	         All other information contained within the geometry
*  	         object is read after the primitives by this call.
*
*		 The fid internal to the geometry object specifies the
*		 file from which the object will be read.  In
*		 general, this fid is set when the geometry header is
*		 read by the \f(CWkgeom_read_header()\fP function.
*	
*         Input: object - geometry object to read
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 24, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_finish_reading_object(kgeom_object *object)
{
   int s = TRUE;
   unsigned char v;

   /*-- type identifier was read in the read_header or read_primitive... --*/

   /*-- ...primitive terminator of -1 was also read by read_primitive --*/


   /*-- layout, location_dim, texture_coord_dim were read previously --*/

   /*-- read the version number (version is unused for now) --*/
   CK(kread_ubyte (object->fid, &v, (size_t) 1));

   CK(kread_string(object->fid, &object->name, (size_t) 1));
   CK(kread_float (object->fid,  object->ambient_color, (size_t) 3));
   CK(kread_float (object->fid,  object->diffuse_color, (size_t) 3));
   CK(kread_float (object->fid,  object->specular_color, (size_t) 3));
   CK(kread_float (object->fid, &object->specular_exponent, (size_t) 1));
   CK(kread_int   (object->fid, &object->modeling_space, (size_t) 1));
   CK(kread_int   (object->fid, &object->visible, (size_t) 1));
   CK(kread_ubyte (object->fid, &object->has_bounding_box, (size_t) 1));
   CK(kread_ubyte (object->fid, &object->has_center, (size_t) 1));
   CK(kread_ubyte (object->fid, &object->has_matrix, (size_t) 1));
   CK(kread_ubyte (object->fid, &object->has_color, (size_t) 1));
   CK(kread_float (object->fid,  object->min, (size_t) 3));
   CK(kread_float (object->fid,  object->max, (size_t) 3));
   CK(kread_float (object->fid,  object->center, (size_t) 3));
   CK(kread_float (object->fid,  object->matrix, (size_t) 16));
   CK(kread_float (object->fid,  object->color, (size_t) 3));
   CK(kread_float (object->fid, &object->opacity, (size_t) 1));

   return s;
}

/************************************************************
*
*  Routine Name: kgeom_read_object_and_primitives - read a geometry object and all of its primitives
*
*       Purpose: This routine will read an object and all of its
*		 primitives from the open fid specified within the object.
*
*		 In general, the \f(CWkgeom_read_object()\fP function
*		 should be used instead of this function. This
*		 function has been made public to provide complete
*		 flexibility for reading a geometry object.  It should
*		 only be used when you wish to read a geometry object
*		 primitive-by-primitive. More information on this is
*		 available in the \f(CWkgeom_read_primitive()\fP man
*		 page.
*		 
*		 This function is equivalent to calling the function
*		 \f(CWkgeom_start_reading_object()\fP, followed by a
*		 call to \f(CWkgeom_read_primitive()\fP for each
*		 primitive on the object's primitive list, followed by
*		 a call to \f(CWkgeom_finish_reading_object()\fP read
*		 all the object-specific information.
*
*		 The fid internal to the geometry object specifies the
*		 file from which the the object will be read.  In
*		 general, this fid is set when the geometry header is
*		 read by the \f(CWkgeom_read_header()\fP function.
*
*         Input: object - geometry object to be read
*        Output: 
*       Returns: TRUE on success, FALSE otherwise
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_read_object_and_primitives(kgeom_object *object)
{
   kgeom_primitive *primitive = NULL;
   
   if (!kgeom_start_reading_object(object))
      return FALSE;
   
   while ((primitive = kgeom_read_primitive(object)) != NULL)
      kgeom_add_primitive(object, primitive);

   if (!kgeom_finish_reading_object(object))
      return FALSE;

   return TRUE;
}


/*
 *    ================================================================== 
 *    kgeom_object primitive list management
 *    ==================================================================
 */

/************************************************************
*
*  Routine Name: kgeom_number_primitives - count the number of primitives in the given object
*
*       Purpose: This function will return the number of primitives
*		 contained in the primitive list of the given object.
*
*		 \fIGeometry Primitives and Primitive Lists\fP
*
*		 A geometry primitive contains geometric data which
*		 describes a shape in space. Lines, triangles, and
*		 spheres, are all considered to be geometric
*		 primitives. For a complete description of all the
*		 geometry primitives, see the \fIkgeom_primitive\fP
*		 man page.
*
*		 Each geometry object contains a list of such geometry
*		 primitives.  New primitives can be created using the
*		 function \f(CWkgeom_new_primitive()\fP.  The function
*		 \f(CWkgeom_remove_primitive()\fP allows you to remove
*		 a primitive from the primitive list of an object
*		 after it has been added.  A primitive
*		 at any given position can be retrieved using the
*		 function \f(CWkgeom_get_primitive()\fP.
*
*		 The primitive functions should be the only means used
*		 for accessing the primitives on a primitive list.
*		 Even though the primitive list of the geometry object
*		 could be acccessed directly, it should be considered
*		 to be private.  By using the primitive list function
*		 calls, you can be certain to maintain the integrity
*		 of the primitive list.
*
*         Input: object - object to count primitives in
*        Output: 
*       Returns: the number of primitives contained in the object
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_number_primitives(kgeom_object *object)
{
   int number = 0;
   
   if (object == NULL)
      return 0;
   
   while (object->primitives[number] != NULL) 
      number++;
   
   return number;
}

/************************************************************
*
*  Routine Name: kgeom_add_primitive - add a primitive to a geometry object
*
*       Purpose: This function will add the given primitive to the
*		 end of the primitive list of the given object.
*	
*		 \fIGeometry Primitives and Primitive Lists\fP
*
*		 A geometry primitive contains geometric data which
*		 describes a shape in space. Lines, triangles, and
*		 spheres, are all considered to be geometric
*		 primitives. For a complete description of all the
*		 geometry primitives, see the \fIkgeom_primitive\fP
*		 man page.
*
*		 Each geometry object contains a list of such geometry
*		 primitives.  New primitives can be created using the
*		 function \f(CWkgeom_new_primitive()\fP.  The function
*		 \f(CWkgeom_remove_primitive()\fP allows you to remove
*		 a primitive from the primitive list of an object
*		 after it has been added.  At any time, the number of
*		 primitives which have been added to the primitive
*		 list of an object can be determined by the function
*		 \f(CWkgeom_number_primitives()\fP.  A primitive
*		 at any given position can be retrieved using the
*		 function \f(CWkgeom_get_primitive()\fP.
*
*		 The primitive functions should be the only means used
*		 for accessing the primitives on a primitive list.
*		 Even though the primitive list of the geometry object
*		 could be acccessed directly, it should be considered
*		 to be private.  By using the primitive list function
*		 calls, you can be certain to maintain the integrity
*		 of the primitive list.
*
*         Input: object - object to add primitive to
*		 primitive - primitive to add
*        Output: 
*       Returns: TRUE on success, FALSE on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
int
kgeom_add_primitive(
   kgeom_object *object,
   kgeom_primitive *primitive)
{
   int length = 0;

   if (object == NULL || primitive == NULL)
      return FALSE;

   length = kgeom_number_primitives(object);  /*-- not including NULL --*/
   length++;
   
   object->primitives = (kgeom_primitive **) 
      krealloc(object->primitives, sizeof(kgeom_primitive *) * (length+1));

   object->primitives[length] = NULL;

   /*-- if primitive is an object, we need to encase it in a candy shell --*/
   if (primitive->type == KGEOM_OBJECT)
   {
      kgeom_object    *o = (kgeom_object *) primitive;
      kgeom_primitive *p = kgeom_new_primitive(KGEOM_OBJECT);

      /*-- assign the parent --*/
      o->parent = object;

      /*-- shell will point to the object .. we add shell to the list --*/
      ((kgeom_subobject *)p)->object = (kaddr) o;
      object->primitives[length-1] = primitive;
   }
   else /*-- just add it --*/
      object->primitives[length-1] = primitive;

   return TRUE;
}

/************************************************************
*
*  Routine Name: kgeom_get_primitive - get a primitive from a geometry object
*
*       Purpose: This routine will get the primitive from
*		 the specified position in the primitive list. If the
*		 position is outside the bounds of the primitive list
*		 then a NULL pointer will be returned.
*
*		 \fIGeometry Primitives and Primitive Lists\fP
*
*		 A geometry primitive contains geometric data which
*		 describes a shape in space. Lines, triangles, and
*		 spheres, are all considered to be geometric
*		 primitives. For a complete description of all the
*		 geometry primitives, see the \fIkgeom_primitive\fP
*		 man page.
*
*		 Each geometry object contains a list of such geometry
*		 primitives.  New primitives can be created using the
*		 function \f(CWkgeom_new_primitive()\fP.  Once a new
*		 primitive has been created and its data assigned, it
*		 can be added to the primitive list of the object
*		 using the function \f(CWkgeom_add_primitive()\fP.  A
*		 corresponding function
*		 \f(CWkgeom_remove_primitive()\fP allows you to remove
*		 a primitive from the primitive list of an object.
*		 The number of primitives which have been added to the
*		 primitive list of an object is returned by the
*		 function \f(CWkgeom_number_primitives()\fP.
*
*		 The primitive functions should be the only means used
*		 for accessing the primitives on a primitive list.
*		 Even though the primitive list of the geometry object
*		 could be acccessed directly, it should be considered
*		 to be private.  By using the primitive list function
*		 calls, you can be certain to maintain the integrity
*		 of the primitive list.
*	
*         Input: object - object to retrieve primitive from
*		 position - position in list from which the primitive should be retrieved
*        Output: 
*       Returns: TRUE on success, FALSE on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_primitive *
kgeom_get_primitive(
   kgeom_object *object,
   int           position)
{
   int length = 0;

   if (object == NULL)
      return FALSE;

   length = kgeom_number_primitives(object);  /*-- not including NULL --*/

   if (position >= length)
      return NULL;

   /*-- if object is a subobject, extract it from its candy shell --*/
   if (object->primitives[position]->type == KGEOM_OBJECT)
   {
      kgeom_subobject *p = (kgeom_subobject *) object->primitives[position];

      return (kgeom_primitive *) p->object;
   }
   else /*-- just return it --*/
      return object->primitives[position];
}

/************************************************************
*
*  Routine Name: kgeom_remove_primitive - remove a primitive from a geometry object
*
*       Purpose: This routine will remove the given primitive 
*		 from the primitive list of the specified object.
*		 If the given primitive is not on the primitive list
*		 of that object, then no action is taken.
*
*		 \fIGeometry Primitives and Primitive Lists\fP
*
*		 A geometry primitive contains geometric data which
*		 describes a shape in space. Lines, triangles, and
*		 spheres, are all considered to be geometric
*		 primitives. For a complete description of all the
*		 geometry primitives, see the \fIkgeom_primitive\fP
*		 man page.
*
*		 Each geometry object contains a list of such geometry
*		 primitives.  New primitives can be created using the
*		 function \f(CWkgeom_new_primitive()\fP.  Once a new
*		 primitive has been created and its data assigned, it
*		 can be added to the primitive list of the object
*		 using the function \f(CWkgeom_add_primitive()\fP.
*
*		 The number of primitives which have been added to the
*		 primitive list of an object is returned by the
*		 function \f(CWkgeom_number_primitives()\fP.  A
*		 primitive at a specific position in the primitive
*		 list can be retrieved using the function
*		 \f(CWkgeom_get_primitive()\fP.
*
*		 The primitive functions should be the only means used
*		 for accessing the primitives on a primitive list.
*		 Even though the primitive list of the geometry object
*		 could be acccessed directly, it should be considered
*		 to be private.  By using the primitive list function
*		 calls, you can be certain to maintain the integrity
*		 of the primitive list.
*		 
*         Input: object - object to remove primitive from
*		 primitive - primitive to remove
*        Output: 
*       Returns: TRUE on success, FALSE on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 18, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*
*************************************************************/
int
kgeom_remove_primitive(
   kgeom_object *object,
   kgeom_primitive *primitive)
{
   int done = FALSE;
   int i = 0, number, j;
   
   number = kgeom_number_primitives(object);
   
   while (object->primitives[i] != NULL && !done)
   {
      if (object->primitives[i] == primitive)
	 object->primitives[i] = NULL, done = TRUE;
      else
	 i++;
   }

   /*-- shrink the list --*/
   for (j = i; j < number-1; j++)
      object->primitives[j] = object->primitives[j+1];

   /*-- resize down to number (not number-1 since we have NULL terminator --*/
   object->primitives = (kgeom_primitive **) 
      krealloc(object->primitives, number * sizeof(kgeom_primitive *));

   return TRUE;
}

/*
 *    ================================================================== 
 *    kgeom_object misc
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_get_root - for a given object, get it's root object 
|
|       Purpose: This routine will traverse the parent list and 
|		 find the root parent object.  It is assumed
|		 that only the root knows the correct file to write to.
|
|		 If this object is the root, it will return itself.
|	
|         Input: object - object to find root for
|
|        Output: none
|
|       Returns: root object
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
| Modifications:
|
------------------------------------------------------------*/
kgeom_object *
kgeom_get_root(kgeom_object *object)
{
   if (object == NULL)
      return NULL;
   
   if (object->parent != NULL)
      return kgeom_get_root(object->parent);
   else
      return object;
}
