 /*
  * Khoros: $Id: primitive.c,v 1.1.2.3 1997/03/18 23:19:17 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: primitive.c,v 1.1.2.3 1997/03/18 23:19:17 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Geometry Services Primitive Code
   >>>>
   >>>>   Public:	
   >>>>			kgeom_new_primitive()
   >>>>  Private:
   >>>>			primitive_function_index()
   >>>>   Static:
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/************************************************************
*
*  Routine Name: kgeom_new_primitive - construct a new geometry primitive
*
*       Purpose: This function will construct a new geometry primitive
*                structure of the given type and initialize all default
*		 values.
*
*		 A geometry primitive contains geometric data which
*		 describes a shape in space. Lines, triangles, and
*		 spheres, are all considered to be geometric
*		 primitives.  Each "type" of primitive is actually
*		 represented by a distinct structure; these structures
*		 have been unioned into a common
*		 \f(CWkgeom_primitive\fP structure.  
*
*		 The type of primitive is declared on instantiation,
*		 as in this example :
*
*  !		\f(CWkgeom_primitive *prim;\fP
*  !
*  !		\f(CWprim = kgeom_new_primitive(KGEOM_POLYLINE_CONNECTED);\fP
*
*		 The type of primitive is tracked in a type field
*		 which is common to all the primitive structures. It
*		 can always be accessed as \f(CWprim->type\fP from any
*		 primitive structure.
*
*		 For a complete description of all the types of
*		 geometry primitives, see the \fIkgeom_primitive\fP
*		 man page, or consult the Data Services Manual :
*		 Programming Services Volume II
*
*		 Note that a geometry primitive is not limited to a
*		 single instance of a shape.  For example, a single
*		 spheres primitive can contain many spheres and a
*		 single polyline primitive can contain many lines.
*
*		 The data within a primitive consists of several
*		 components; the predominate component which actually
*		 defines the size and position of the geometry is the
*		 location data.  The location data consists of a list
*		 of vertices.  The number of these vertices is a
*		 specified by a field within the primitive and is a
*		 defining characteristic for most geometry primitives.
*
*		 Additional components which may be present in a
*		 geometry primitive are color, normals, or texture
*		 coordinates.  All components except for the location
*		 component are optional in any given geometry
*		 primitive.
*
*		 Most the fields of a geometry primitive are pointers
*		 to allocated arrays which contain these components.
*		 The size of these arrays is implied from the number
*		 of vertices present in the geometry primitive, and
*		 the presentation characteristics (layout,
*		 location_dim, texture_coord_dim, and has_alpha) set
*		 in the associated geometry object.  Most data within
*		 a geometry primitive will be floating point.
*
*		 Geometric data is always stored "elements first",
*		 with the vectors forming the leading dimension of the
*		 array.  For example, 3D location vertices will be
*		 stored in the order XYZXYZXYZ...
*
*		 The number of location vertices is by definition the
*		 number of vertices in the primitive. The size of each
*		 vertex will be determined by the location_dim field
*		 of the associated object.
*
*		 The number of color and normal vectors will be a
*		 function of the number of vertices, but will vary
*		 depending on the primitive and the layout field of
*		 associated geometry object.  For example, a disjoint
*		 polyline primitive will have 1 color per vertex for a
*		 KPER_VERTEX layout, but will have 1 color per line
*		 for a KPER_LINE primitive.
*
*		 The size of each color vector will either be 3 (for
*		 RGB), of 4 (for RGBa) if the has_alpha field of the
*		 associated object is TRUE.  The size of each normal
*		 vector will match the size of location_dim field in
*		 the associated object.  
*
*		 If texture coordinate vectors are present, the number
*		 of texture coordinates must match the number of
*		 vertices in the primitive.  The size of each texture
*		 coordinate will be determined by the texture_coord_dim 
*		 field of the associated object.
*
*		 If the modeling_space field of the associated object
*		 is set to KNDC_SPACE, the location data will be
*		 interpreted as normalized device coordinates and thus
*		 should range between 0.0 and 1.0.
*
*		 Color data is represented by RGB vectors with three
*		 floats determining the red, green, and blue
*		 intensity.  These numbers should range from 0.0 to
*		 1.0, with 1.0 implying maximum intensity. If an alpha
*		 component is present it will determine the
*		 transparency of the primitive.  It also should range
*		 from 0.0 to 1.0, with 0.0 implying that the primitive
*		 is totally transparent, and 1.0 implying that the
*		 primitive is totally opaque.
*
*		 Since the fields of a geometry primitive vary from
*		 type to type, we must specify the type of primitive
*		 we are working with in order to access the
*		 components of the primitive structure.  For example,
*		 if we had a spheres primitive, we would have to do
*		 the following to access the specific sphere fields :
*
*  !		\f(CWkgeom_primitive *prim;\fP
*  !
*  !		\f(CWprim = kgeom_new_primitive(KGEOM_SPHERES);\fP
*  !
*  !		\f(CW// specify that we want 45 spheres\fP 
*  !
*  !		\f(CWprim->spheres.nverts = 45;\fP 
*
*		 This can alternatively be done by casting the
*		 structure to its specific primitive type as follows : 

*  !		\f(CWkgeom_spheres *s;\fP
*  !
*  !		s = (kgeom_spheres *) kgeom_new_primitive(KGEOM_SPHERES);
*  !
*  !		\f(CW// specify that we want 45 spheres\fP 
*  !
*  !		\f(CWs->nverts = 45;\fP 
*
*		 Primitives can be cast to any type at any time, but
*		 you should be sure that you cast back to the general
*		 \f(CWkgeom_primitive *\fP type before passing a
*		 specific primitive into any functions which take a
*		 primitive argument.
*
*         Input: type - type of primitive to construct
*        Output: none
*       Returns: a pointer to the constructed primitive on success, 
*                NULL on failure
*
*  Restrictions: 
*
*    Written By: Steve Kubica
*          Date: December 1, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
kgeom_primitive *
kgeom_new_primitive(int type)
{
   kgeom_primitive *primitive = NULL;

   /*-- calloc will init everything to 0 or NULL --*/
   primitive = (kgeom_primitive *) kcalloc((size_t)1, sizeof(kgeom_primitive));
      
   if (primitive == NULL)
   {
      kerror("kgeom", "kgeom_new_primitive", "unable to allocate space");
      return NULL;
   }

   primitive->type = type;

   /*-- do any primitive specific, non-NULL initialization --*/
   switch (type)
   {
      case KGEOM_POLYLINE_DISJOINT : 
      case KGEOM_POLYLINE_CONNECTED :
	 {
	    kgeom_polyline *p = (kgeom_polyline *) primitive;
	    p->linetype  = 1; /* KLINE_SOLID; */
	    p->linewidth = 1;
	 }
	 break;
      case KGEOM_MARKERS :
	 {
	    kgeom_markers *p = (kgeom_markers *) primitive;
	    p->markertype  = 1; /* KMARKER_SQUARE; */
	 }
	 break;
      case KGEOM_POLYGON :
	 {
	    kgeom_polygon *p = (kgeom_polygon *) primitive;
	    p->linetype  = 1; /* KLINE_SOLID; */
	    p->linewidth = 1;
	 }
	 break;
      case KGEOM_RECTANGLES :
	 {
	    kgeom_rectangles *p = (kgeom_rectangles *) primitive;
	    p->linetype  = 1; /* KLINE_SOLID; */
	    p->linewidth = 1;
	 }
	 break;
      case KGEOM_ELLIPSES :
	 {
	    kgeom_ellipses *p = (kgeom_ellipses *) primitive;
	    p->linetype  = 1; /* KLINE_SOLID; */
	    p->linewidth = 1;
	 }
	 break;
      case KGEOM_VIEW_XFORM :
	 {
	    kgeom_view_xform *p = (kgeom_view_xform *) primitive;
	    float matrix[16] = KGEOM_VIEW_XFORM_MATRIX_DEFAULT;
	    
	    kmemcpy(p->matrix, matrix, sizeof(float) * 16);
	 }
	 break;
      case KGEOM_OBJECT_XFORM :
	 {
	    kgeom_object_xform *p = (kgeom_object_xform *) primitive;
	    float matrix[16] = KGEOM_OBJECT_XFORM_MATRIX_DEFAULT;
	    
	    kmemcpy(p->matrix, matrix, sizeof(float) * 16);
	    p->relative = KGEOM_OBJECT_XFORM_RELATIVE_DEFAULT;

	    p->translate[0] = 0.0;
	    p->translate[1] = 0.0;
	    p->translate[2] = 0.0;
	 }
	 break;
      case KGEOM_CAMERA_ORIENTATION :
	 {
	    kgeom_camera_orientation *p = 
	       (kgeom_camera_orientation *) primitive;

	    float eye[3] = KGEOM_CAMERA_EYE_DEFAULT;
	    float at [3] = KGEOM_CAMERA_AT_DEFAULT;
	    float up [3] = KGEOM_CAMERA_UP_VECTOR_DEFAULT;
	    
	    kmemcpy(p->eye, eye, sizeof(float) * 3);
	    kmemcpy(p->at,  at,  sizeof(float) * 3);
	    kmemcpy(p->up,  up,  sizeof(float) * 3);

	    p->field_of_view = KGEOM_CAMERA_FIELD_OF_VIEW_DEFAULT;
	    p->hither        = KGEOM_CAMERA_HITHER_DEFAULT;
	    p->yon           = KGEOM_CAMERA_YON_DEFAULT;

	    p->is_stereo     = KGEOM_CAMERA_IS_STEREO_DEFAULT;
	    p->eye_distance  = KGEOM_CAMERA_EYE_DISTANCE_DEFAULT;
	    p->focal_length  = KGEOM_CAMERA_FOCAL_LENGTH_DEFAULT;
	 }
	 break;
   }
   
   return primitive;
}

/*
 *    ================================================================== 
 *    primitive type to array index conversion
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: primitive_function_index 
|
|       Purpose: This routine will return the index into a table
|		 of primitive functions which corresponds to the
|		 given primitive.
|
|         Input: primitive - primitive type to return index for
|        Output: none
|       Returns: index into list
|
|    Written By: Steve Kubica
|          Date: December 18, 1996
| Modifications:
|
------------------------------------------------------------*/
int primitive_function_index(int type)
{
   switch (type)
   {
      case KGEOM_POLYLINE_DISJOINT : 
      case KGEOM_POLYLINE_CONNECTED :
	 return 1;
      case KGEOM_TRIANGLES_DISJOINT : 
      case KGEOM_TRIANGLES_CONNECTED :
	 return 2;
      case KGEOM_SPHERES : 
	 return 3;
      case KGEOM_TEXT : 
	 return 4;
      case KGEOM_CYLINDERS : 
	 return 5;
      case KGEOM_DIRECTED_POINTS : 
	 return 6;
      case KGEOM_MARKERS : 
	 return 7;
      case KGEOM_POLYGON : 
	 return 8;
      case KGEOM_RECTANGLES : 
	 return 9;
      case KGEOM_ELLIPSES : 
	 return 10;
      case KGEOM_QUADMESH : 
      case KGEOM_OCTMESH : 
	 return 11;
      case KGEOM_TEXTURE2D : 
      case KGEOM_TEXTURE3D : 
	 return 12;
      case KGEOM_VIEW_XFORM : 
	 return 13;
      case KGEOM_OBJECT_XFORM : 
	 return 14;
      case KGEOM_CAMERA_ORIENTATION : 
	 return 15;
      case KGEOM_VR_EVENT : 
	 return 16;
      case KGEOM_OBJECT : 
	 return 17;
      default : 
	 return 0;
   }
}

