 /*
  * Khoros: $Id: geom_fest.c,v 1.1 1997/02/28 21:41:15 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: geom_fest.c,v 1.1 1997/02/28 21:41:15 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Geometry Application Services
   >>>>		Primitive List Routines 
   >>>>		
   >>>>  Private:
   >>>>                 kgeom_construct_component()
   >>>>                 kgeom_establish_manifest_primitive()
   >>>>                 kgeom_verify_manifest_primitive()
   >>>>                 kgeom_get_manifest_primitive_list()
   >>>>                 kgeom_set_manifest_primitive_list()
   >>>>                 kgeom_get_manifest_number_primitives()
   >>>>                 kgeom_set_manifest_number_primitives()
   >>>>                 kgeom_write_transcript()
   >>>>                 kgeom_read_transcript()
   >>>>                 kgeom_get_manifest()
   >>>>                 kgeom_destroy_manifest()
   >>>>                 kgeom_free_manifest()
   >>>>   Static:
   >>>>                 _kgeom_free_component()
   >>>>                 _kgeom_construct_primitive()
   >>>>                 _kgeom_free_primitive()
   >>>>                 _kgeom_construct_manifest()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/* 
 *    ================================================================== 
 *    Component Methods
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_construct_component
|
|       Purpose: This simple routine will construct a
|                component.  
|
|         Input: present     - TRUE if the component is present
|		 begin_point - beginning index into segment
|		 size[2]     - array indicating size of component
|
|        Output: none
|
|       Returns: the constructed component structure
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1995 15:29
| Modifications:
|
------------------------------------------------------------*/
kgeom_component *
kgeom_construct_component(
   int   present,
   int   begin_point,
   int  *size)
{
   kgeom_component *knew = NULL;
   
   knew = (kgeom_component *) kcalloc(sizeof(kgeom_component), (size_t)1);
   
   knew->present     = present;
   knew->begin_point = begin_point;
   knew->size[0]     = size[0];
   knew->size[1]     = size[1];
   
   return knew;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kgeom_free_component
|
|       Purpose: This simple routine will free a component
|                structure.
|
|         Input: comp - pointer to the component definition 
|			to be freed
|
|        Output: none
|
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: Feb 16, 1994 16:24
| Modifications:
|
------------------------------------------------------------*/
static void 
_kgeom_free_component(kgeom_component *comp)
{
   /*-- trivial free --*/
   if (comp == NULL) return;
   
   kfree_and_NULL(comp);
   return;
}

/* 
 *    ================================================================== 
 *    Primitive Methods
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kgeom_construct_primitive
|
|       Purpose: This simple routine will construct a primitive
|                structure.  The components will be initialized
|		 to be NULL.
|
|         Input: primitive      - primitive identifier
|		 num_components - number of empty components to 
|				  allocate
|
|        Output: none
|
|       Returns: the constructed primitive.
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
static kgeom_primitive_la *
_kgeom_construct_primitive(
   int   primitive,
   int   num_components)
{
   kgeom_primitive_la *knew = NULL;
   
   knew = (kgeom_primitive_la *) kmalloc(sizeof(kgeom_primitive_la));
   
   /*-- assign the primitive type --*/
   knew->primitive = primitive;

   /*-- initially has no vertices --*/
   knew->number  = 0;

   /*-- initially all per-primitive attributes --*/
   knew->texture = NULL;
   knew->type    = KGEOM_LINE_TYPE_DEFAULT;
   knew->width   = KGEOM_LINE_WIDTH_DEFAULT;

   /*-- initialize null matrix --*/
   kmemset(knew->matrix, 0, (size_t) 16*sizeof(float));
   kmemset(knew->translate, 0, (size_t) 3*sizeof(float));
   knew->relative = KRELATIVE;
   knew->name     = NULL;

   /*-- allocate space for the given num_components --*/
   knew->num_components = num_components;
   knew->components = (kgeom_component **) 
      kcalloc(sizeof(kgeom_component *), (size_t) num_components);

   return knew;
}

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kgeom_free_primitive
|
|       Purpose: This simple routine will free a primitive 
|                structure along with all of its component structures.
|
|         Input: prim - pointer to the primitive structure to be freed
|
|        Output: none
|
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: Feb 16, 1994 16:24
| Modifications:
|
------------------------------------------------------------*/
static void 
_kgeom_free_primitive(kgeom_primitive_la *prim)
{
   int i;
   
   /*-- trivial free --*/
   if (prim == NULL) return;
   
   for( i = 0; i < prim->num_components; i++)
      _kgeom_free_component(prim->components[i]);
   
   kfree(prim->components);
   kfree(prim->texture);
   kfree(prim->name);

   return;
}

/* 
 *    ================================================================== 
 *    Manifest Methods
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: (static) _kgeom_construct_manifest
|
|       Purpose: This simple routine will construct a primitive
|                manifest structure.  The initial number of primitives
|		 will be zero.
|
|         Input: none
|
|        Output: none
|
|       Returns: the constructed manifest.
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
static kgeom_manifest *
_kgeom_construct_manifest(void)
{
   kgeom_manifest *knew = NULL;
   
   knew = (kgeom_manifest *) kmalloc(sizeof(kgeom_manifest));
   
   knew->num_primitives = 0;
   knew->primitives     = NULL;
   
   return knew;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_free_manifest
|
|       Purpose: This simple routine will free a primitive manifest
|                structure along with all of its primitive and
|		 component structures.
|
|         Input: manifest - pointer to the manifest structure to be freed
|
|        Output: none
|
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: Feb 16, 1994 16:24
| Modifications:
|
------------------------------------------------------------*/
void 
kgeom_free_manifest(kgeom_manifest *manifest)
{
   int i;
   
   /*-- trivial free --*/
   if (manifest == NULL) return;
   
   for (i = 0; i < manifest->num_primitives; i++)
   {
      _kgeom_free_primitive(manifest->primitives[i]);
      kfree(manifest->primitives[i]);
   }

   kfree(manifest->primitives);
   kfree(manifest);
   
   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_establish_manifest_primitive
|
|       Purpose: This routine will establish a primitive within
| 		 the primitive manifest at the given primitive 
|		 position according to the given primitive definition.
|
|         Input: manifest   - the manifest from which to establish
|			      the primitive
|		 pos        - the position at which to establish 
|			      the primitive
|		 definition - the definition of the primitive to 
|		 	      be established
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_establish_manifest_primitive(
   kgeom_manifest 	 *manifest,
   int            	  pos,
   kgeom_primitive_defin *defin)
{
   /*-- sanity check --*/
   if (manifest == NULL || defin == NULL)
      return FALSE;

   if (pos > manifest->num_primitives)
      return FALSE;

   /*-- if there's something there already, get rid of it --*/
   if (manifest->primitives[pos] != NULL)
   {
      _kgeom_free_primitive(manifest->primitives[pos]);
      kfree(manifest->primitives[pos]);
   }

   /*-- construct new primitive according to given defin --*/
   manifest->primitives[pos] = _kgeom_construct_primitive(defin->primitive, 
							defin->num_components);
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_verify_manifest_primitive
|
|       Purpose: This routine will verify that the primitive at 
|		 a given position matches the specified primitive.
|
|         Input: manifest   - the manifest from which to verify
|			      the primitive
|		 primitive  - the primitive to verify	
|		 pos        - the position at which to verify
|			      the primitive
|		 definition - the definition of the primitive to 
|		 	      be established.  if null, then the
|			      definition will be found within this routine
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_verify_manifest_primitive(
   kgeom_manifest 	  *manifest,
   int            	   primitive,
   int            	   pos,   
   kgeom_primitive_defin *defin)
{
   /*-- sanity check --*/
   if (manifest == NULL)
      return FALSE;

   if (pos > manifest->num_primitives)
      return FALSE;

   if (defin == NULL)
      defin = kgeom_locate_primitive_defin(primitive);
   
   /*-- if they match, then all is well --*/
   if (manifest->primitives[pos]->primitive == primitive)
      return TRUE;

   /*-- if no primitive has yet been set, then establish this one --*/
   if (manifest->primitives[pos]->primitive == KNULL_PRIMITIVE)
      return kgeom_establish_manifest_primitive(manifest, pos, defin);
   
   /*-- primitive doesn't match --*/
   kdas_set_error(KGEOM_EPRIMITIVE_NO_MATCH_POSITION);

   return FALSE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_get_manifest_primitive_list
|
|       Purpose: This routine will return a static array of 
|                containing the primitive identifiers contained
|		 within the specified geometry manifest.
|
|         Input: manifest - the manifest from which to get 
|			    the primitive list
|        Output: none
|
|       Returns: the primitive list
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int *
kgeom_get_manifest_primitive_list(
   kgeom_manifest *manifest)
{
   static int *list = NULL;
   static int  list_size;
   int         i;
   int         inc;
   

   /*-- sanity check --*/
   if (manifest == NULL)
      return NULL;

   /*-- incrment by at least KLENGTH, or more if necessary --*/
   if (manifest->num_primitives > KLENGTH)
      inc = KLENGTH;
   else
      inc = manifest->num_primitives;

   /*-- make sure static primitive list can hack the number of prims --*/
   if (manifest->num_primitives > list_size)
   {
      list_size += inc;
      list = (int *) krealloc(list, manifest->num_primitives * sizeof(int));
   }
   

   /*-- traverse the primitives in the manifest and fill out the list --*/
   for (i = 0; i < manifest->num_primitives; i++)
      list[i] = manifest->primitives[i]->primitive;

   return list;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_set_manifest_primitive_list
|
|       Purpose: This routine will take a new primitive
|		 list, traverse the primitives in the 
|		 manifest, destroy any primitives in the
|		 manifest that do not match the new list,
|		 and establish any primitives in the
|		 new list which have not yet been established.
|
|		 After this routine, the primitives in the
|		 manifest will correctly represent the 
|		 primitive identifiers in the list that
|		 was provided.
|
|         Input: manifest - the manifest in which to set 
|			    the new primitive list
|		 list     - the new list of primitives
|
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_set_manifest_primitive_list(
   kgeom_manifest *manifest,
   int            *list)
{
   int i;

   /*-- sanity checks --*/
   if (manifest == NULL)
      return FALSE;

   if ((manifest->num_primitives != 0) && list == NULL)
      return FALSE;
   
   /*-- traverse the primitives in the manifest and fill out the list --*/
   for (i = 0; i < manifest->num_primitives; i++)
   {
      /*-- check if new primitive matches what is on old manifest --*/
      if (manifest->primitives[i]->primitive != list[i])  
      {
	 /*-- destroy the old one --*/
	 _kgeom_free_primitive(manifest->primitives[i]);
	 kfree(manifest->primitives[i]);

	 /*-- construct the new one --*/
	 if (list[i] == KNULL_PRIMITIVE)
	    manifest->primitives[i] = _kgeom_construct_primitive(list[i], 0);
	 else
	    kgeom_establish_manifest_primitive(manifest, i, 
				        kgeom_locate_primitive_defin(list[i]));
      }
   }
   
   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_get_manifest_number_primitives
|
|       Purpose: This routine will return the number of 
|		 primitives containined within a manifest.
|
|         Input: manifest - the manifest from which to get 
|			    the number of primitives
|        Output: none
|
|       Returns: the number of primitives
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_get_manifest_number_primitives(
   kgeom_manifest *manifest)
{
   /*-- sanity check --*/
   if (manifest == NULL)
      return 0;

   return manifest->num_primitives;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_set_manifest_number_primitives
|
|       Purpose: This routine will change the number of
|		 primitives in a given manifest to match
|		 the new specified number.  If the new
|		 number necessitates the list to grow,
|		 the list will be filled out with 
|	     	 null primitives.
|
|         Input: manifest - the manifest in which to set 
|			    the new number of primitives
|		 num      - the new number of primitives
|
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_set_manifest_number_primitives(
   kgeom_manifest *manifest,
   int             num)
{
   int i;

   /*-- sanity checks --*/
   if (manifest == NULL)
      return FALSE;

   /*-- make sure this number isn't something outrageous --*/
   if (num < 0) 
      kdas_fail_error(KGEOM_EINVALID_NUMBER_PRIMITIVES);

   /*-- that was sure easy --*/
   if (manifest->num_primitives == num)
      return TRUE;

   /*-- if shrinking, free the excess primitives the list --*/
   if (manifest->num_primitives > num)
   {
      for (i = num; i < manifest->num_primitives; i++)
      {
	 _kgeom_free_primitive(manifest->primitives[i]);
	 kfree(manifest->primitives[i]);
      }

      manifest->primitives = (kgeom_primitive_la **) krealloc(
					manifest->primitives, 
				         num * sizeof(kgeom_primitive_la *));
   }
   

   /*-- if growing, fill out the new primitives with null primitives --*/
   else 
   {
      manifest->primitives = (kgeom_primitive_la **) krealloc(
					manifest->primitives, 
				         num * sizeof(kgeom_primitive_la *));

      for (i = manifest->num_primitives; i < num; i++)
	 manifest->primitives[i] = _kgeom_construct_primitive(KNULL_PRIMITIVE,
							      0);
   }

   /*-- make the list equal to the new size --*/
   manifest->num_primitives = num;


   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_write_transcript
|
|       Purpose: This routine will store a given manifest
|		 into the permanent attribute 
|		 KGEOM_MANIFEST_TRANSCRIPT.
|
|		 The manifest structure is basically unrolled
|		 in a deterministic manner into a huge array
|		 of integers and stored in a single attribute.		
|
|	         Additionally, any texture strings which 
|	         may be stored per primitive are stored 
|                here in a separate string attribute.
|
|         Input: obj      - the object in which to store
|			    the manifest
|		 manifest - the manifest to be stored
|
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) on failure
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_write_transcript(
   kobject         obj,
   kgeom_manifest *manifest)
{
   char **textures = NULL;
   char **xnames = NULL;
   float *xmatrices;
   float *xtrans;
   int   *xrels;
   int    arg_size = 1;
   int   *transcript;
   int   *ltypes;
   int   *lwidths;
   int   *p;
   int    number;
   int    i;
   int    j;

   /*-- sanity check --*/
   if (manifest == NULL)
      return FALSE;
   
   /*-- if any transcript attributes exist already, then blast it --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TYPE, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_TYPE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH);


   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME, 
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME);
   
   /* 
    *    create transcript of proper size and data type :
    *      there will be 1 int for the number of primitives 
    * 	   followed by 3 ints/primitive and 4 ints/component
    *
    *    additional transcript information will be a list of 
    *    strings with one per number of primitives as well as
    *    two integer arrays.  we have to add one to the array
    *    to cope with the single primitive case, as data 
    *    services does not know how to handle arrays of size 1.
    *
    */
   for (i = 0; i < manifest->num_primitives; i++, arg_size += 3)
      arg_size += manifest->primitives[i]->num_components * 4;

   /*-- allocate the transcript --*/
   transcript = (int *)kmalloc((size_t) arg_size * sizeof(int));
   textures   = (char **) kcalloc((size_t)(manifest->num_primitives+1),
				  sizeof(char *));
   ltypes     = (int *)kcalloc((size_t)(manifest->num_primitives+1),
			       sizeof(int));
   lwidths    = (int *)kcalloc((size_t)(manifest->num_primitives+1),
			       sizeof(int));

   xmatrices  = (float *)kcalloc((size_t)(manifest->num_primitives+1), 
			(size_t) 16 * sizeof(float));
   xtrans     = (float *)kcalloc((size_t)(manifest->num_primitives+1), 
			(size_t) 3 * sizeof(float));
   xrels      = (int *) kcalloc((size_t)(manifest->num_primitives+1),
			sizeof(int));
   xnames     = (char **) kcalloc((size_t)(manifest->num_primitives+1),
				  sizeof(char *));
   

   /*-- fill out transcript information --*/
   transcript[0] = manifest->num_primitives;
   for (i = 0, p = &(transcript[1]); i < manifest->num_primitives; i++)
   {
      *p++ = manifest->primitives[i]->primitive;
      *p++ = manifest->primitives[i]->num_components;
      *p++ = manifest->primitives[i]->number;

      if (manifest->primitives[i]->components != NULL)
	 for (j = 0; j < manifest->primitives[i]->num_components; j++)
	 {
	    *p++ = manifest->primitives[i]->components[j]->present;
	    *p++ = manifest->primitives[i]->components[j]->begin_point;
	    *p++ = manifest->primitives[i]->components[j]->size[0];
	    *p++ = manifest->primitives[i]->components[j]->size[1];
	 }

      /*-- per-primitive attributes --*/
      textures[i] = manifest->primitives[i]->texture;
      ltypes[i]   = manifest->primitives[i]->type;
      lwidths[i]  = manifest->primitives[i]->width;

      /*-- xform attributes --*/
      kmemcpy((kaddr) &xmatrices[i*16], manifest->primitives[i]->matrix,
	      (size_t) 16 * sizeof(float));
      kmemcpy((kaddr) &xtrans[i*3], manifest->primitives[i]->translate,
	      (size_t) 3 * sizeof(float));

      xrels[i]  = manifest->primitives[i]->relative;
      xnames[i] = manifest->primitives[i]->name;
   }
   
   /*-- store the transcript --*/
   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT, 1,
			      arg_size, KINT, TRUE, TRUE))
      return FALSE;

   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT, transcript))
      return FALSE;

   /*-- store the per-primitive attributes --*/
   number = manifest->num_primitives+1;

   /*-- per-primitive textures --*/
   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE, 1,
			      number, KSTRING, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE, textures))
      return FALSE;

   /*-- per-primitive types --*/
   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_TYPE, 1,
			      number, KINT, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_TYPE, ltypes))
      return FALSE;

   /*-- per-primitive width --*/
   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH, 1,
			      number, KINT, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH, lwidths))
      return FALSE;


   /*-- per-primitive transformation attributes --*/
   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX, 1,
			      number*16, KFLOAT, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX, xmatrices))
      return FALSE;

   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE, 1,
			      number, KINT, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE, xrels))
      return FALSE;

   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE, 1,
			      number * 3 , KFLOAT, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE, xtrans))
      return FALSE;

   if (!kdms_create_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME, 1,
			      number, KSTRING, TRUE, TRUE))
      return FALSE;
   if (!kdms_set_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME, xnames))
      return FALSE;

   kfree_and_NULL(transcript);
   kfree_and_NULL(textures);
   kfree_and_NULL(ltypes);
   kfree_and_NULL(lwidths);

   kfree_and_NULL(xmatrices);
   kfree_and_NULL(xrels);
   kfree_and_NULL(xtrans);
   kfree_and_NULL(xnames);

   return TRUE;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_read_transcript
|
|       Purpose: This routine will retrieve a given manifest
|		 from the permanent attribute 
|		 KGEOM_MANIFEST_TRANSCRIPT.
|
|         Input: obj - the object from which to get the manifest
|			    
|        Output: none
|
|       Returns: the manifest
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
kgeom_manifest *
kgeom_read_transcript(kobject obj)
{
   kgeom_manifest  *manifest;
   kgeom_primitive_la *primitive;
   char **textures = NULL;
   char **xnames = NULL;
   float *xmatrices = NULL;
   float *xtrans = NULL;
   int   *xrels = NULL;
   int 	 *transcript = NULL;
   int   *ltypes   = NULL;
   int   *lwidths  = NULL;
   int 	 *p;
   int 	  present;
   int 	  begin_point;
   int    primitive_id;
   int    num_components;
   int 	  size[2];
   int 	  i;
   int 	  j;
   
   /*-- get the attribute which is the transcript of the manifest  --*/
   if (!kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT, 
			     NULL, NULL, NULL, NULL))
      return NULL;

   if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT, &transcript))
      return NULL;

   /*-- get the per-primitive attributes --*/

   /*-- per-primitive textures --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE, &textures))
	 return NULL;

   /*-- per-primitive types --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TYPE,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_TYPE, &ltypes))
	 return NULL;
   
   /*-- per-primitive widths --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH, &lwidths))
	 return NULL;

   /*-- per-primitive transformation attributes  --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX, 
			      &xmatrices))
	 return NULL;
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE, 
			      &xrels))
	 return NULL;
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE, 
			      &xtrans))
	 return NULL;
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME,
			    NULL, NULL, NULL, NULL))
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME, 
			      &xnames))
	 return NULL;


   p = transcript;

   /*-- start with the manifest --*/
   manifest = _kgeom_construct_manifest();

   /*-- allocate pointers to the number of primitives needed --*/
   manifest->num_primitives = *p++;
   manifest->primitives = (kgeom_primitive_la **)kmalloc((size_t)
				  manifest->num_primitives *
				  sizeof(kgeom_primitive_la *));

   /*-- get each primitive --*/
   for (i = 0; i < manifest->num_primitives; i++)
   {
      primitive_id   = *p++;
      num_components = *p++;
      
      primitive = _kgeom_construct_primitive(primitive_id, num_components);

      primitive->number  = *p++;

      /*-- per primitive attributes --*/
      primitive->texture = (textures == NULL) ? NULL : textures[i];
      primitive->width   = (lwidths  == NULL) ? 1 : lwidths[i];
      primitive->type    = (ltypes   == NULL) ? 1 : ltypes[i];

      if (xmatrices != NULL)
	 kmemcpy(primitive->matrix, (kaddr) &xmatrices[i*16], 
		 (size_t) 16*sizeof(float));
      if (xtrans != NULL)
	 kmemcpy(primitive->translate, (kaddr) &xtrans[i*3], 
		 (size_t) 3*sizeof(float));

      primitive->name     = (xnames == NULL) ? NULL      : xnames[i];
      primitive->relative = (xrels == NULL)  ? KRELATIVE : xrels[i];

      /*-- get each component within that primitive --*/
      for (j = 0; j < primitive->num_components; j++)
      {
	 present     = *p++;
	 begin_point = *p++;
	 size[0]     = *p++;
	 size[1]     = *p++;
	 
	 primitive->components[j] = kgeom_construct_component(present, 
							      begin_point, 
							      size);
      }
      manifest->primitives[i] = primitive;
   }

   return manifest;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_get_manifest
|
|       Purpose: This routine will get the primitive manifest
|		 from the specified object.  If the object does
|		 not have a manifest, one will be retored from
|		 the manifest transcript.  If there is no manifest
|		 transcript, then there is no manifest to return.
|
|         Input: obj - the object from which to get the manifest
|			    
|        Output: none
|
|       Returns: the manifest
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
kgeom_manifest *kgeom_get_manifest(kobject obj)
{
   kgeom_manifest *manifest;
   
   /*-- do we have the manifest already figured out? --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST, 
			    NULL, NULL, NULL, NULL))
   {
      if (!kdms_get_attribute(obj, NULL, KGEOM_MANIFEST, &manifest))
	 return NULL;
      else
	 return manifest;
   }

   /*-- see if there's a transcript of earlier manifest --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT,
			    NULL, NULL, NULL, NULL))
   {
      manifest = kgeom_read_transcript(obj);
      
      /*-- we want to keep this manifest around in memory... --*/
      kdms_create_attribute(obj, NULL, KGEOM_MANIFEST, 1, 1, KSTRUCT, 
			    FALSE, TRUE);

      /*-- ... so we associate it with the object --*/
      kdms_set_attribute(obj, NULL, KGEOM_MANIFEST, manifest);

      return manifest;
   }

   /*-- no manifest, or transcript of there ever having been one --*/
   kdas_set_error(KGEOM_ENO_PRIMITIVE_LIST);
   
   return NULL;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_destroy_manifest
|
|       Purpose: This routine will destroy the primitive manifest
|		 and the manifest transcipt from the specified object.
|
|         Input: obj - the object from which to destroy
|			    
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Feb 04, 1994 15:29
| Modifications:
|
------------------------------------------------------------*/
int kgeom_destroy_manifest(kobject obj)
{
   kgeom_manifest *manifest;
   
   /*-- free the manifest if we already have it in memory --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST, 
			    NULL, NULL, NULL, NULL))
   {
      kdms_get_attribute(obj, NULL, KGEOM_MANIFEST, &manifest);
      kgeom_free_manifest(manifest);

      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST);
   }

   /*-- blast the manifest transcript if we have one --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_TRANSCRIPT);

   /*-- blast all per-primitive attributes if we have them --*/
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_TEXTURE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_TYPE,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_TYPE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_WIDTH);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_MATRIX);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_RELATIVE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_TRANSLATE);
   if (kdms_query_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME,
			    NULL, NULL, NULL, NULL))
      kdms_destroy_attribute(obj, NULL, KGEOM_MANIFEST_XFORM_OBJECT_NAME);
   
   return TRUE;
}
