/*
  * Khoros: $Id: geom_utils.c,v 1.1 1997/02/28 21:41:27 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: geom_utils.c,v 1.1 1997/02/28 21:41:27 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Geometry Application Services
   >>>>		Utility Routines
   >>>>		
   >>>>  Private:
   >>>>                 kgeom_init()
   >>>>                 kgeom_layout()
   >>>>                 kgeom_color_size()
   >>>>                 kgeom_attribute_name()
   >>>>   Static:
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

int _geom_initialized = FALSE;

/* 
 *    ================================================================== 
 *    Initialization Routines
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_init
|
|       Purpose: Initialize all of the primitives and
|                attributes needed by geometry services 
|
|         Input: none
|
|        Output: none
|
|       Returns: TRUE on success, FALSE otherwise
|
|    Written By: Steve Kubica
|          Date: March 6, 1994
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_init(void)
{
   if ( _geom_initialized == TRUE )
      return TRUE;

   /*-- prevent this routine from being called again --*/
   _geom_initialized = TRUE;

   /*-- initialize all geometry services error messages --*/
   kgeom_init_errors();

   /*-- initialize all geometry services attributes --*/ 
   kgeom_define_attributes();

   /*-- initialize all geometry services primitives --*/
   kgeom_define_primitives();

   /*-- initialize all geometry services meshes --*/
   kgeom_define_meshes();

   /*-- initialize all geometry services segments --*/
   kgeom_define_segments();

   /*-- initialize all polymorphic services segments --*/
   kpds_init();

   /*--------------------------------------------------------------
   |  it is possible to generate errnos with duplicate attribute
   |  definitions.  To prevent the user from seeing this, we'll
   |  re-zero the errno.
   \--------------------------------------------------------------*/
   errno = 0;

   return TRUE;
}

/* ################################################################## */

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_layout
|
|       Purpose: This routine returns the current setting
|		 of the KGEOM_LAYOUT attribute
|
|         Input: obj - the object to retrieve the attribute from
|
|        Output: none
|
|       Returns: the current setting of KGEOM_LAYOUT
|
|    Written By: Steve Kubica
|          Date: Aug 20, 1994 11:56
| Modifications:
|
------------------------------------------------------------*/
int
kgeom_layout(kobject obj)
{
   int layout;
   
   if (!kgeom_get_attribute(obj, KGEOM_OBJECT, KGEOM_LAYOUT, &layout))
      return -1;
   else
      return layout;
}

/*-----------------------------------------------------------
|
|  Routine Name: kgeom_color_size
|
|       Purpose: This routine checks the current colorspace and
|		 the has alpha attributes to determine the current 
|		 size of a color vector for the given segment.
|	
|		 In general, the color size is determined by the
|		 colorspace model and the presence or absence of
|		 the alpha channel.  
|
|		 If the segment being checked is not the map, then
|		 the colorsize will be 1 if there is a map present,
|		 (meaning that the color data will be mapped) --
|		 otherwise, the colorsize will be determined by
|		 the colorspace.   If there is a map segment, it
|		 is assumed that the colorsize will represent
|		 the width of the map.
|
|         Input: object  - the object to determine the colorsize for
|		 segment - the segment being checked
|
|        Output: None
|
|       Returns: the color vector size on success, -1 on failure
|
|    Written By: Steven Kubica 
|          Date: May 11, 1994 15:35
| Modifications:
|
------------------------------------------------------------*/
int 
kgeom_color_size(
   kobject  obj,
   char    *segment)
{
   int cspace;
   int has_alpha;
   int csize;
   int has_map;


   has_map = kdms_query_segment(obj,KDMS_SEGMENT_MAP);

   /* 
    *   if this is not the map segment and we have a map, then our 
    *   color is a map index and our colorsize is going to be one.
    */
   if (segment != NULL)
   {
      if (has_map && (kstrcmp(segment,KDMS_SEGMENT_MAP) != 0))
         return 1;

      /* if this the the map segment, and we have no map, then 	
	 return -1 to indicate that there is no map width */
      if (!has_map && (kstrcmp(segment,KDMS_SEGMENT_MAP) == 0))
	 return -1;
      
   }
   

   /* othewise, the color is determined by the colorspace and has alpha */
   if (!kgeom_get_attributes(obj,KGEOM_OBJECT,
	  		         KGEOM_COLORSPACE,&cspace,
			         KGEOM_HAS_ALPHA, &has_alpha,
			         NULL))
      return -1;

   switch(cspace)
   {
        case KNONE:  
	case KGREY:		csize = 1;  break;
	case KRGB: 		
	case KCMY: 		
	case KYIQ: 		
	case KHSV: 		
	case KHLS: 		
	case KIHS: 		
	case KXYZ: 		
	case KUVW: 		
	case KUCSUVW:		
	case KUCSSOW:		
	case KUCSLab:		
	case KUCSLuv:		csize = 3;  break;
	case KUSERDEFINED: 	csize = 3;  break;

	default:
	   /* errno = INVALID COLORSPACE! */
     	   return -1;
   }

   /* The alpha channel adds one to the color size */
   if (has_alpha) csize++;

   return csize;
}


/* ################################################################## */

/*
 *    ==================================================================
 *    Naming Routines
 *    ==================================================================
 */

/*-----------------------------------------------------------
|  Routine Name: kgeom_attribute_name
|       Purpose: generate unique string for given position
|
|  		"position" + "attribute"
|
|    Written By: Steve Kubica
|          Date: Apr 14, 1993 16:05
-----------------------------------------------------------*/
char *kgeom_attribute_name(
   char *attribute,
   int   position)
{
   char temp[KLENGTH];

   ksprintf(temp, "%d.%s\0", position, attribute);
   return kstrdup(temp);
}
