/*
  * Khoros: $Id: model_data.c,v 1.1.2.4 1997/03/12 23:53:25 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: model_data.c,v 1.1.2.4 1997/03/12 23:53:25 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Polymophic Model Data Handlers
   >>>>		
   >>>>  Private:
   >>>>                 kaps_offset_corners()
   >>>>                 kaps_get_primitive()
   >>>>                 kaps_put_primitive()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*
 *    ================================================================== 
 *    Polymorphic Data Access Routines
 *    ==================================================================
 */

/*-----------------------------------------------------------
|
|  Routine Name: kaps_offset_corners
|
|       Purpose: Given a zero-based begin and end point
|		 defining the absolute size of a region,
|		 offset that region into space according
|		 to the position and offsets for a given
|		 segment.   A mask can be provided to determine
|		 which positional components are to be used.
|		 This routine uses zeros in the mask to 
|		 mark which positional components are 
|		 going to be used (this lets me use the 
|		 usage mask directly).
|
|         Input: obj     - the object to retrieve the position 
|			   and offset from
|		 segment - the segment to retrieve the position 
|			   and offset from
|		 mask    - an array indicating which positional
|		           components are to be used.  Zeros mark
|			   which positional components are to be used.
|			   If this is not present, then it defaults to use
|			   all.
|
|	         begin   - the begin point to be shifted
|	         end     - the end point to be shifted
|
|        Output: begin   - the begin point is shifted by the position
|			   and offset in place
|		 end     - the end point is shifted by the position
|			   and offset in place
|       Returns: none
|
|    Written By: Steve Kubica
|          Date: Apr 29, 1994 11:17
| Modifications:
|
------------------------------------------------------------*/
void
kaps_offset_corners(
   kobject obj,
   char   *segment,
   int    *mask,
   int    *begin,
   int    *end)
{
   int *pos = NULL;
   int *off = NULL;
   int  dim;
   int  i;
   int  total_offset;

   dim = kaps_get_segment_dimension(obj, segment);
   pos = kaps_get_segment_position(obj, segment);
   off = kaps_get_segment_offset(obj, segment);

   if (pos == NULL || off == NULL || mask == NULL)
      return;

   for (i = 0; i < dim; i++)
   {
      if (mask[i] == 0)
      {				/* zero's indicate what to use */
	 total_offset = 0;
         total_offset += (pos != NULL) ? pos[i] : 0;
	 total_offset += (off != NULL) ? off[i] : 0;

	 begin[i] += total_offset;
	 end[i]   += total_offset;
      }
   }

   return;
}

/*-----------------------------------------------------------
|
|  Routine Name: kaps_get_primitive
|
|       Purpose: This routine gets a primitive.  The autoincrementing
|		 functionality is invoked here.  This routine
|		 can handle all regularly defined primitives.  
|		 Region-based primitives have their own routine.
|
|         Input: obj  - the object to retrieve the primitive from
|		 prim - the primitive to retrieve
|		 data - pointer to allocated data with which to
|		        return the primitive, or NULL if we should
|		        perform the allocation.
|
|        Output: none
|
|       Returns: pointer to the primitive data on success,
|		 NULL otherwise
|
|    Written By: Steve Kubica
|          Date: Apr 29, 1994 11:17
| Modifications:
|
------------------------------------------------------------*/
kaddr
kaps_get_primitive(
   kobject         obj,
   kaps_primitive *prim,
   kaddr           data)
{
   static int begin[KDMS_MAX_DIM];
   static int end[KDMS_MAX_DIM];
   int *size;
   int  i;


   /* verify that the segment is present and get its size */
   if ((size = kaps_get_segment_size(obj, prim->segment)) == NULL)
      return NULL;
   
   /* figure out the begin and end points according to the primitive */
   for (i = 0; i < KDMS_MAX_DIM; i++)
   {
      begin[i] = 0;
      end[i]   = (prim->usage[i] == 1) ? (size[i] - 1) : 0;
   }

   /* increment the position for this segment */
   kaps_advance_segment_position(obj, prim->segment, begin, end, size, TRUE);

   /* move the begin and end to the appropriate position and offset */
   kaps_offset_corners(obj, prim->segment, prim->usage, begin, end);

   return (kdms_get_data(obj, prim->segment, begin, end, data));
}

/*-----------------------------------------------------------
|
|  Routine Name: kaps_put_primitive
|
|       Purpose: This routine gets a primitive.  The autoincrementing
|		 functionality is invoked here.  This routine
|		 can handle all regularly defined primitives.  
|		 Region-based primitives have their own routine.
|
|         Input: obj  - the object to put the primitive into
|		 prim - the primitive to store
|		 data - pointer to the primitive data being stored
|
|        Output: none
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Steve Kubica
|          Date: Apr 29, 1994 11:17
| Modifications:
|
------------------------------------------------------------*/
int
kaps_put_primitive(
   kobject         obj,
   kaps_primitive *prim,
   kaddr           data)
{
   static int begin[KDMS_MAX_DIM];
   static int end[KDMS_MAX_DIM];
   int       *size;
   int        i;


   /* verify that the segment is present and get its size */
   if ((size = kaps_get_segment_size(obj, prim->segment)) == NULL)
      return FALSE;

   /* figure out the begin and end points according to the primitive */
   for (i = 0; i < KDMS_MAX_DIM; i++)
   {
      begin[i] = 0;
      end[i]   = (prim->usage[i] == 1) ? (size[i] - 1) : 0;
   }

   /* increment the position for this segment */
   kaps_advance_segment_position(obj, prim->segment, begin, end, size, FALSE);

   /* move the begin and end to the appropriate position and offset */
   kaps_offset_corners(obj, prim->segment, prim->usage, begin, end);

   return (kdms_put_data(obj, prim->segment, begin, end, data));
}
