/*
 * Khoros: $Id: roi_util.c,v 1.16.2.1 1997/04/29 22:30:43 mirage Exp $
 */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: roi_util.c,v 1.16.2.1 1997/04/29 22:30:43 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		ROI Extraction Utility Routines
   >>>>		
   >>>>  Private:
   >>>>                 xvw_getroi_line()
   >>>>                 xvw_getroi_rectangle()
   >>>>                 xvw_getroi_circle()
   >>>>                 xvw_getroi_ellipse()
   >>>>                 xvw_getroi_polyline()
   >>>>                 xvw_extract_roi()
   >>>>   Static:
   >>>>                 xvw_extract_region_mask()
   >>>>                 xvw_flood_fill()
   >>>>                 xvw_scanline_fill()
   >>>>                 xvw_draw_line()
   >>>>                 xvw_draw_rectangle()
   >>>>                 xvw_draw_circle()
   >>>>                 xvw_draw_ellipse()
   >>>>                 xvw_compute_extents()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

/*-----------------------------------------------------------
|
|  Routine Name: xvw_extract_region_mask 
|
|       Purpose: This routine takes a mask that was created earlier
|		 with xvw_draw_xxx() and puts its mask data into the
|		 roi kobject, so that that ROI has the correct shape 
|		 and location.
|
|         Input: object   - the object from which the ROI is being extracted
|		 roi      - the roi kobject in which to insert the mask
|                x,y      - the roi's x & y position
|		 w,h      - the roi's width & height
|		 d        - depth of roi 
|		 t        - time of roi 
|		 e        - elements of roi
|		 mask     - the mask defining the shape
|		 multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|		 policy   - the ROI Presentation policy, one of:
|			    KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE.
|			    Note: if object passed in is not an image object,
|			    just use KROI_INSIDE.
|			    
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
static int xvw_extract_region_mask(
   xvobject object,
   kobject  roi,
   int      x,
   int      y,
   int      w,
   int      h,
   int      d,
   int      t,
   int      e,
   char     *mask,
   int      multiple,
   int      policy)
{
	int      i, j, k, xpos, ypos, xoffset = 0, yoffset = 0;

	/*
	 *  With multiple ROI's, the mask will have already been created.
	 *  with single ROI's, we'll need to create the mask.
	 */
	if (!kpds_query_mask(roi))
	{
	   kpds_create_mask(roi);
	   kpds_set_attribute(roi, KPDS_MASK_DATA_TYPE, KUBYTE);
	   kpds_set_attribute(roi, KPDS_MASK_SIZE, w, h, d, t, e);
	   kpds_initialize_mask(roi, 0.0);
	}
	
	/*
	 *  For single ROI's using a policy of "inside shape", 
	 *  put mask at (0,0) and set the subobject position
	 */
	if ((!multiple) && (policy != KIMAGE_ROI_OUTSIDE))
	{
	    xpos = 0;
	    ypos = 0;
	    kpds_set_attribute(roi, KPDS_SUBOBJECT_POSITION, x, y, d, t, e);
	}

	/*
	 *  For a policy of "outside shape" and for multiple ROI's, need to 
	 *  put the mask wherever it was originally drawn by the user in the 
	 *  image;  setting subobject position doesn't really make sense.
	 */
	else
	{
  	    xpos = x;
	    ypos = y;
	}

	if (multiple)
        {
	    char *oldmask;
	    char *tmpnew, *tmpold;

	    kpds_set_attributes(roi, 
			KPDS_MASK_REGION_SIZE, w, h, 1, 1, 1,
			KPDS_MASK_POSITION, xpos+xoffset, ypos+yoffset, 0, 0, 0,
			NULL);
	    oldmask = (char *)kpds_get_data(roi, KPDS_MASK_REGION, NULL);
	    tmpnew = mask;
	    tmpold = oldmask;
	    
	    if (policy != KIMAGE_ROI_OUTSIDE)
	        for (i = 0; i < w*h; i++, tmpold++)
	    	    *tmpnew++ = (*tmpnew > 0) ? *tmpnew : *tmpold;
	    else
                for (i = 0; i < w*h; i++, tmpold++)
                    *tmpnew++ = (*tmpnew == 0) ? 0 : *tmpold;
	    kfree_and_NULL(oldmask);
	}

	/*
	 *  Put the mask into the ROI kobject
	 */
	kpds_set_attributes(roi,
		KPDS_MASK_POSITION, xpos+xoffset, ypos+yoffset, 0, 0, 0,
		KPDS_MASK_REGION_SIZE, w, h, 1, 1, 1,
		NULL);

	for (i = 0; i < e; i++) 
	for (j = 0; j < t; j++) 
	for (k = 0; k < d; k++) 
	{
	     kpds_set_attribute(roi, KPDS_MASK_POSITION, 
				xpos+xoffset, ypos+yoffset, k, j, i);
	     kpds_put_data(roi, KPDS_MASK_REGION, mask);
	}

	return(TRUE);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_flood_fill 
|
|       Purpose: This is a flood fill routine that fills a shape in a mask
|		 according to the shape outline which is already in the 
|		 mask array.  The shape outline must be defined by the "val"
|		 values, while the shape background must be defined by the "set"
|		 values. For a single inside ROI extraction, the "val" value 
|		 should be a 1, and the "set" value should be a 0.  For example,
|		 a mask array passed in might look like:
|
|		 0 0 0 1 0 0 0 
|		 0 0 1 0 1 0 0 
|		 0 1 0 0 0 1 0 
|		 1 0 0 0 0 0 1 
|		 0 1 1 1 1 1 0 
|		 0 0 0 0 0 0 0 
|
|		 Given this, xvw_flood_fill() would fill the shape with the
|		 "val" value (in this example, 1) so that it looked like:
|
|                0 0 0 1 0 0 0 
|                0 0 1 1 1 0 0
|                0 1 1 1 1 1 0
|                1 1 1 1 1 1 1
|                0 1 1 1 1 1 0
|                0 0 0 0 0 0 0
|
|		 This mask would result in the original image data appearing
|		 within the shape, and the default background color to appear
|		 around the edges.
|
|		 This situation is reversed with an outside ROI extraction,
|		 where the mask array passed in has the shape defined by
|		 the "val" value of 0, and the background defined by the "set"
|		 value of 1; the resulting mask would look like:
|
|                1 1 1 0 1 1 1
|                1 1 0 0 0 1 1
|                1 0 0 0 0 0 1
|                0 0 0 0 0 0 0
|                1 0 0 0 0 0 1
|                1 1 1 1 1 1 1 
|
|		 This would cause the original image data to appear outside the
|		 shape, and the default background color to appear within the 
| 		 shape.  This works similarly for multiple ROI's, except that 
|		 the "val" value should be incremented for each successive ROI.
|
|         Input: mask - the mask on which to perform the flood fill
|                w    - the width of the mask array
|                h    - the height of the mask array
|		 x    - the initial x position to start the fill
|		 y    - the initial y position to start the fill
|		 set  - the value to expect in the background area of the mask 
|                val  - the value to expect as defining the shape outline 
|			within the mask, which is also the value that will
|			be used to fill the shape
|
|    Written By: Mark Young
|          Date: Jan 02, 1995
|
------------------------------------------------------------*/
/*ARGSUSED*/
static void xvw_flood_fill(
  char *mask,
  unsigned int  w,
  unsigned int  h,
  int  x,
  int  y,
  int  set,
  int  val)
{
        int j, k, indx = y*w;

        j = x-1; k = x+1;
        mask[indx+x] = val;

        while (j >= 0 && mask[indx+j] == set) mask[indx + j--] = val;
        while (k < w  && mask[indx+k] == set) mask[indx + k++] = val;

        for (x = j+1, indx = j+1 + y*w; x < k; x++, indx++)
        {
           if (y > 0 && mask[indx-w] == set)
              xvw_flood_fill(mask, w, h, x, y-1, set, val);
           if (y < h-1 && mask[indx+w] == set)
              xvw_flood_fill(mask, w, h, x, y+1, set, val);
        }
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_scanline_fill
|				  
|
|       Purpose: This modified scanline fills a polygonal shape in a 
|		 mask according to the shape outline which is already stored
|		 in the mask.  It uses the xvw_flood_fill() routine to
|		 do most of the work (see above).

|         Input: mask   - the mask on which to perform the scanline fill
|		 width  - the width of the mask array
|		 height - the height of the mask array
|                set    - the value to expect in the background area of the mask
|                val    - the value to expect as defining the shape outline 
|                         within the mask, which is also the value that will
|                         be used to fill the shape
|    Written By: Mark Young
|          Date: Dec 18, 1994
|
------------------------------------------------------------*/
/*ARGSUSED*/
static void xvw_scanline_fill(
  char         *mask,
  unsigned int w,
  unsigned int h,
  int          set,
  int          val)
{
	char *tmp;
	int  i, j, put;

	for (i = 0, j = w*(h-1); i < w; i++, j++)
	{
	   if (mask[i] == set) 	
		xvw_flood_fill(mask, w, h, i, 0, set, -1);
	   if (mask[j] == set) 	
		xvw_flood_fill(mask, w, h, i, (int) h-1, set, -1);
	}
	for (i = 0, j = w-1; i < h; i++)
	{
	   if (mask[i*w] == set) 
		xvw_flood_fill(mask, w, h, 0, i, set, -1);
	   if (mask[i*w+j] == set) 
		xvw_flood_fill(mask, w, h, j, i, set, -1);
	}

	for (j = 0, tmp = mask; j < h; j++)
	{
	   for (i = 0, put = TRUE; i < w; i++)
	   {
	      if (*tmp++ == set)
		 xvw_flood_fill(mask, w, h, i, j, set, !put ? -1 : val);
	   }
	}

	for (i = 0; i < w*h; i++)
	   mask[i] = (mask[i] == -1) ? set : mask[i];
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_draw_line 
|
|       Purpose: This modified bresenham algorithm draws a line into 
|		 a mask, given a pre-allocated and initialized mask array, 
|		 and the value to use for the line.
|
|         Input: mask     - the mask in which to draw the line
|                width    - the width of the mask array
|                height   - the height of the mask array
|                x_1, y_1 - the begin point of the line
|                x_2, y_2 - the end point of the line
|		 val      - the value to use for the line outline
|
|    Written By: Mark Young
|          Date: Dec 18, 1994
|
------------------------------------------------------------*/
/*ARGSUSED*/
static void xvw_draw_line(
  char *mask,
  unsigned int width,
  unsigned int height,
  int  x_1,
  int  y_1,
  int  x_2,
  int  y_2,
  int  val)
{
	int    i, position, dx, dy, xincr, yincr;

	dx = kabs(x_2 - x_1);
	dy = kabs(y_2 - y_1);
	xincr = (x_1 < x_2) ? 1 : -1;
	yincr = (y_1 < y_2) ? 1 : -1;

	if (dx > dy)
	{
	   position = 2*dy - dx;
	   for (i = 0; i <= dx; i++)
	   {
	      mask[y_1 * (int) width + x_1] = val; x_1 += xincr;
	      if (position >= 0)
	      {
		 y_1 += yincr;
		 position += 2 * (dy - dx);
	      }
	      else position += 2 * dy;
	   }
	}
	else
	{
	   position = 2*dx - dy;
	   for (i = 0; i <= dy; i++)
	   {
	      mask[y_1 * (int) width + x_1] = val; y_1 += yincr;
	      if (position >= 0)
	      {
		 x_1 += xincr;
		 position += 2 * (dx - dy);
	      }
	      else position += 2 * dx;
	   }
	}
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_draw_rectangle
|
|       Purpose: This modified bresenham algorithm draws a rectangle into 
|                a mask, given a pre-allocated and initialized mask array, 
|                and the value to use for the rectangle.
|
|         Input: mask     - the mask in which to draw the rectangle
|                width    - the width of the mask array
|                height   - the height of the mask array
|		 x        - the x position of the rectangle
|		 y        - the y position of the rectangle
|                w        - the width of the rectangle
|                h        - the height of the rectangle
|                val      - the value to use for the rectangle outline
|
|    Written By: Mark Young
|          Date: Dec 18, 1994
|
------------------------------------------------------------*/
/*ARGSUSED*/
static void xvw_draw_rectangle(
  char *mask,
  unsigned int width,
  unsigned int height,
  int  x,
  int  y,
  int  w,
  int  h,
  int  val)
{
	int x_1 = x, y_1 = y, x_2 = x+w-1, y_2 = y+h-1;

	xvw_draw_line(mask, width, height, x_1, y_1, x_2, y_1, val);
	xvw_draw_line(mask, width, height, x_2, y_1, x_2, y_2, val);
	xvw_draw_line(mask, width, height, x_2, y_2, x_1, y_2, val);
	xvw_draw_line(mask, width, height, x_1, y_2, x_1, y_1, val);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_draw_circle
|
|       Purpose: This modified bresenham algorithm draws a circle into 
|                a mask, given a pre-allocated and initialized mask array, 
|                and the value to use for the circle.
|
|         Input: mask     - the mask in which to draw the circle
|                width    - the width of the mask array
|                height   - the height of the mask array
|		 xc,yc    - the position of the circle's center
|                radius   - the radius of the circle
|		 val      - the value to use for the circle outline 
|
|    Written By: Mark Young
|          Date: Dec 18, 1994
|
------------------------------------------------------------*/
/*ARGSUSED*/
static void xvw_draw_circle(
  char *mask,
  unsigned int width,
  unsigned int height,
  int  xc,
  int  yc,
  int  radius,
  int  val)
{
	int x, y, position;

	x = 0; y = radius; position = 3 - 2 * radius;
	while (x <= y)
	{
	   mask[xc + x  +  (yc + y) * (int) width] = val;
	   mask[xc - x  +  (yc + y) * (int) width] = val;
	   mask[xc + x  +  (yc - y) * (int) width] = val;
	   mask[xc - x  +  (yc - y) * (int) width] = val;
	   mask[xc + y  +  (yc + x) * (int) width] = val;
	   mask[xc - y  +  (yc + x) * (int) width] = val;
	   mask[xc + y  +  (yc - x) * (int) width] = val;
	   mask[xc - y  +  (yc - x) * (int) width] = val;
	   if (x == y)
	      break;

	   if (position < 0)
	      position += 4*x + 6;
	   else
	      position += 4*(x - y--) + 10;

	   x++;
	}
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_draw_ellipse 
|
|       Purpose: This modified bresenham algorithm draws an ellipse into
|                a mask, given a pre-allocated and initialized mask array,
|                and the value to use for the ellipse.
|
|         Input: mask   - the mask in which to draw the ellipse
|                width  - the width of the mask array
|                height - the height of the mask array
|                xc, yc - the center point of the ellipse
|                a, b   - the width & height (major / minor axis)
|		 val    - the value to use for the ellipse outline
|
|    Written By: Mark Young
|          Date: Dec 18, 1994
|
------------------------------------------------------------*/
/*ARGSUSED*/
static void xvw_draw_ellipse(
  char *mask,
  unsigned int width,
  unsigned int height,
  int  xc,
  int  yc,
  int  a,
  int  b,
  int  val)
{
	double position;
	int    x, y, asqr, bsqr;

	x = 0; y = b;
	asqr = ksqr(a);
	bsqr = ksqr(b);

	position = bsqr - asqr*b + asqr/4;
	mask[xc + x  +  (yc + y) * (int) width] = val;
	mask[xc - x  +  (yc + y) * (int) width] = val;
	mask[xc + x  +  (yc - y) * (int) width] = val;
	mask[xc - x  +  (yc - y) * (int) width] = val;
	while (asqr*(y - 0.5) > bsqr*(x + 1.0))
	{
	   if (position < 0)
	      position += bsqr*(2*x + 3);
	   else
	      position += bsqr*(2*x + 3) + asqr*(-2*y-- + 2);

	   x++;
	   mask[xc + x  +  (yc + y) * (int) width] = val;
	   mask[xc - x  +  (yc + y) * (int) width] = val;
	   mask[xc + x  +  (yc - y) * (int) width] = val;
	   mask[xc - x  +  (yc - y) * (int) width] = val;
	}

	position = bsqr*ksqr(x + 0.5) + asqr*ksqr(y - 1.0) - asqr*bsqr;
	while (y > 0)
	{
	   if (position < 0)
	      position += bsqr*(2*x++ + 2) + asqr*(-2*y + 3);
	   else
	      position += asqr*(-2*y + 3);

	   y--;
	   mask[xc + x  +  (yc + y) * (int) width] = val;
	   mask[xc - x  +  (yc + y) * (int) width] = val;
	   mask[xc + x  +  (yc - y) * (int) width] = val;
	   mask[xc - x  +  (yc - y) * (int) width] = val;
	}
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_compute_extents
|
|       Purpose: This routine computes the extents for a given set of points.
|		 It returns the begin and end points, as well as the overall
|		 width and height in pixels.
|
|         Input: pts - the points
|		 num - the number of points
|        Output: x,y - the begin point
|		 w,h - the overall width & height in pixels
|
|    Written By: Mark Young
|          Date: Dec 18, 1994
|
------------------------------------------------------------*/
/* ARGSUSED */
static void xvw_compute_extents(
   XPoint *pts,
   int	  num,
   int	  *x,
   int	  *y,
   unsigned int	*w,
   unsigned int	*h)
{
	int i, xmax, ymax;

	*x = xmax = pts[0].x;
	*y = ymax = pts[0].y;
	for (i = 1; i < num; i++)
	{
	   if (*x > pts[i].x)
	      *x = pts[i].x;
	   else if (xmax < pts[i].x)
	      xmax = pts[i].x;

	   if (*y > pts[i].y)
	      *y = pts[i].y;
	   else if (ymax < pts[i].y)
	      ymax = pts[i].y;
	}
	*w = (xmax - *x) + 1;
	*h = (ymax - *y) + 1;
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_getroi_line 
|
|       Purpose: This routine extracts a line region from an area object.
|
|         Input: object   - the area object to extract from
|                x_1,y_1  - the starting x&y point
|                x_2,y_2  - the ending x&y point
|		 val	  - the value
|                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|                policy   - one of KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE
|		 d        - depth of mask to be created
|		 t        - time of mask to be created
|		 e        - elements of mask to be created
|
|       Returns: the newly created region on success, NULL otherwise
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
kobject xvw_getroi_line(
   xvobject object,
   kobject  roi,
   int	    x_1,
   int	    y_1,
   int	    x_2,
   int	    y_2,
   int	    val,
   int      multiple,
   int      policy,
   int      d,
   int      t,
   int      e)
{
	char    *mask = NULL;
	kobject  tmp = NULL;
	int      x, y;
	unsigned int w, h;

	/*
	 *  Create the initial roi object
	 */
	if (roi == NULL)
	   tmp = roi = kpds_create_object();

	/* Get the extent */
	w = kabs(x_2-x_1)+1; h = kabs(y_2-y_1)+1;
	x = kmin(x_1,x_2);   y = kmin(y_1,y_2);

	/* Draw the line and extract the region */
	mask = (char *)kcalloc((size_t) 1, (size_t) w*h);
	xvw_draw_line(mask, w, h, x_1-x, y_1-y, x_2-x, y_2-y, val);

	if (!xvw_extract_region_mask(object, roi, x, y, (int) w, (int) h, 
				     d, t, e, mask, multiple, policy))
	{
	   if (tmp) kpds_close_object(tmp);
	   kfree_and_NULL(mask);
	   return(NULL);
	}
	kfree_and_NULL(mask);
	return(roi);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_getroi_rectangle
|
|       Purpose: This routine extracts a rectangular region from an
|		 area object.
|
|         Input: object   - the area object to extract from
|                x, y     - the x & y point
|                w, h     - the width & height
|		 val      - the value to set in the mask for drawing
|		 fill     - whether to fill the rectangle or not
|                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|                policy   - one of KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE
|		 d        - depth of mask to be created
|		 t        - time of mask to be created
|		 e        - elements of mask to be created
|
|       Returns: the newly created region on success, NULL otherwise
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
kobject xvw_getroi_rectangle(
   xvobject     object,
   kobject      roi,
   int	        x,
   int	        y,
   unsigned int w,
   unsigned int h,
   int	        val,
   int	        fill,
   int          multiple,
   int          policy,
   int          d,
   int          t,
   int          e)
{
	char    *mask = NULL;
	int      set = 0;
	kobject  tmp = NULL;

	/*
	 *  Create the initial roi object
	 */
	if (roi == NULL)
	   roi = kpds_create_object();

	/* Draw the rectangle and create rectangular mask */
	mask = (char *)kcalloc((size_t) 1, (size_t) w*h);

	if (policy == KROI_OUTSIDE)
        {
           val = 0;
           set = 1;
           kmemset(mask, set, (size_t) w*h);
        }
	xvw_draw_rectangle(mask, w, h, 0, 0, (int) w, (int) h, val);

	if (fill == TRUE)
	   xvw_flood_fill(mask, w, h, (int) w/2, (int) h/2, set, val);

	if (!xvw_extract_region_mask(object, roi, x, y, (int) w, (int) h, 
				     d, t, e, mask, multiple, policy))
	{
	   if (tmp) kpds_close_object(tmp);
	   kfree_and_NULL(mask);
	   return(NULL);
	}
	kfree_and_NULL(mask);
	return(roi);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_getroi_circle
|
|       Purpose: This routine extracts a circle region from an area object.
|
|         Input: object   - the area object to extract from
|                x,y      - the center x&y point
|                r        - the radius of the circle
|		 val      - the value to set in the mask for drawing
|		 fill     - whether to fill the circle or not
|                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|                policy   - one of KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE
|		 d        - depth of mask to be created
|		 t        - time of mask to be created
|		 e        - elements of mask to be created
|
|       Returns: the newly created region on success, NULL otherwise
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
kobject xvw_getroi_circle(
   xvobject     object,
   kobject      roi,
   int	        x,
   int	        y,
   unsigned int r,
   int	        val,
   int	        fill,
   int          multiple,
   int          policy,
   int          d,
   int          t,
   int          e)
{
	char    *mask  = NULL;
	int      set = 0;
	kobject  tmp = NULL;
	int      x_1, y_1; 
	unsigned int w, h;

	/*
	 *  Create the initial roi object
	 */
	if (roi == NULL)
	   roi = kpds_create_object();

	/* Get the extent */
	w = h = r*2+1;

	/* Draw the circle and extract the region */
	mask = (char *)kcalloc((size_t) 1, (size_t) w*h);

	if (policy == KROI_OUTSIDE)
	{
	   val = 0;
	   set = 1;
	   kmemset(mask, set, (size_t) w*h);
	}
	xvw_draw_circle(mask, w, h, (int) r, (int) r, (int) r, val);

        if (fill == TRUE)
           xvw_flood_fill(mask, w, h, (int) r, (int) r, set, val);

	x_1 = x - r;  y_1 = y - r;
	if (!xvw_extract_region_mask(object, roi, x_1, y_1, (int) w, (int) h, 
				     d, t, e, mask, multiple, policy))
	{
	   if (tmp) kpds_close_object(tmp);
	   kfree_and_NULL(mask);
	   return(NULL);
	}
	kfree_and_NULL(mask);
	return(roi);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_getroi_ellipse 
|
|       Purpose: This routine extracts a ellipse region from an area object.
|
|         Input: object   - the area object to extract from
|                x,y      - the center x&y point
|                a,b      - the major/minor axis
|		 val      - the value to set in the mask for drawing
|		 fill     - whether to fill the ellipse or not
|                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|                policy   - one of KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE
|		 d        - depth of mask to be created
|		 t        - time of mask to be created
|		 e        - elements of mask to be created
|
|       Returns: the newly created region on success, NULL otherwise
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
kobject xvw_getroi_ellipse(
   xvobject     object,
   kobject      roi,
   int	        x,
   int	        y,
   unsigned int	a,
   unsigned int	b,
   int		val,
   int		fill,
   int          multiple,
   int          policy,
   int          d,
   int          t,
   int          e)
{
	char    *mask = NULL;
	int      set = 0;
	kobject  tmp = NULL;
	int      x_1, y_1; 
	unsigned int w, h;

	/*
	 *  Create the initial roi object
	 */
	if (roi == NULL)
	   tmp = roi = kpds_create_object();

	/* Get the extent */
	w = a*2+1; h = b*2+1;

	/* Draw the ellipse and extract the region */
	mask = (char *)kcalloc((size_t) 1, (size_t) w*h);

	if (policy == KIMAGE_ROI_OUTSIDE)
        {
           val = 0;
           set = 1;
           kmemset(mask, set, (size_t) w*h);
        }
	xvw_draw_ellipse(mask, w, h, (int) a, (int) b, (int) a, (int) b, val);

	if (fill == TRUE)
	   xvw_flood_fill(mask, w, h, (int) a, (int) b, set, val);

	x_1 = x - a;  y_1 = y - b;
	if (!xvw_extract_region_mask(object, roi, x_1, y_1, (int) w, (int) h, 
				     d, t, e, mask, multiple, policy))
	{
	   if (tmp) kpds_close_object(tmp);
	   kfree_and_NULL(mask);
	   return(NULL);
	}
	kfree_and_NULL(mask);
	return(roi);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_getroi_polyline 
|
|       Purpose: This routine extracts a polyline region from an area object.
|
|         Input: object   - the area object to extract from
|                pts	  - the polyline points to be extracted
|                num      - the number of points
|		 val      - the value to set in the mask for drawing
|		 fill     - whether to fill the polygon or not
|                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|                policy   - one of KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE
|		 d        - depth of mask to be created
|		 t        - time of mask to be created
|		 e        - elements of mask to be created
|
|       Returns: the newly created region on success, NULL otherwise
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
kobject xvw_getroi_polyline(
   xvobject object,
   kobject  roi,
   XPoint   *pts,
   int	    num,
   int	    val,
   int	    fill,
   int      multiple,
   int      policy,
   int      d,
   int      t,
   int      e)
{
	char    *mask = NULL;
	int      set = 0;
	kobject  tmp = NULL;
	int      i, x, y; 
	unsigned int w, h;

	/*
	 *  Create the initial roi object
	 */
	if (roi == NULL)
	   tmp = roi = kpds_create_object();

	/* Get the extent */
	xvw_compute_extents(pts, num, &x, &y, &w, &h);

	/* Draw the polyline and extract the region */
	mask = (char *)kcalloc((size_t) 1, (size_t) w*h);

	if (policy == KROI_OUTSIDE)
        {
           val = 0;
           set = 1;
           kmemset(mask, set, (size_t) w*h);
        }
	for (i = 1; i < num; i++)
	{
	   xvw_draw_line(mask, w, h, pts[i-1].x - x, pts[i-1].y - y,
                        pts[i].x - x, pts[i].y - y, val);
	}
	xvw_draw_line(mask, w, h, pts[0].x - x, pts[0].y - y,
                      pts[num-1].x - x, pts[num-1].y - y, val);

	if (fill == TRUE)
	   xvw_scanline_fill(mask, w, h, set, val);

	if (!xvw_extract_region_mask(object, roi, x, y, (int) w, (int) h, 
				     d, t, e, mask, multiple, policy))
	{
	   if (tmp) kpds_close_object(tmp);
	   kfree_and_NULL(mask);
	   return(NULL);
	}
	kfree_and_NULL(mask);
	return(roi);
}

/*-----------------------------------------------------------
|
|  Routine Name: xvw_extract_roi 
|
|       Purpose: This routine will get the roi extents which will then
|		 can be used to get 
|
|         Input: object   - the object from which to get the ROI 
|		 roi      - the kobject to store the roi in
|		 val      - the value to set in the mask for drawing
|		 fill     - whether to fill the roi shape or not
|	         shape    - the ROI shape to extract
|	         erase    - TRUE to erase ROI shape afterwards, FALSE not to
|                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
|                policy   - the ROI Presentation policy, one of:
|                           KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE.
|                           Note: if object passed in is not an image object,
|                           just use KROI_INSIDE.
|
|       Returns: TRUE (1) on success, FALSE (0) otherwise
|
|    Written By: Mark Young & Danielle Argiro
|          Date: August 22, 1995
|
------------------------------------------------------------*/
/* ARGSUSED */
int xvw_extract_roi(
   xvobject object,
   kobject  roi,
   int      val,
   int      fill,
   int      shape,
   int      erase,
   int      multiple,
   int      policy, 
   int      d, 
   int      t,
   int      e)
{
	XPoint *pts  = NULL;
	unsigned int w, h, a, b, r;
	int x, y, x_1, y_1, x_2, y_2, ptnum;

	switch (shape)
	{
	   case KROI_POLYLINE:
	   case KROI_FREEHAND:
	        if (shape == KROI_POLYLINE)
		   pts = xvw_getextent_polyline(object, erase, &ptnum);
		else
		   pts = xvw_getextent_freehand(object, erase, &ptnum);

		if (pts == NULL)
	           return(FALSE);

		xvw_getroi_polyline(object, roi, pts, ptnum, val, 
				    fill, multiple, policy, d, t, e);
		break;

	   case KROI_RECTANGLE:
	        if (!xvw_getextent_rectangle(object, erase, &x, &y, &w, &h))
	           return(FALSE);
		
	        if (w == 0 || h == 0)
	        {
		    kerror(XVISUAL, "xvw_get_specified_roi",
		           "Cannot create ROI of zero width and/or height");
		    return(FALSE);
	        }
	        xvw_getroi_rectangle(object, roi, x, y, w, h, val, 
				     fill, multiple, policy, d, t, e);
		break;

	   case KROI_CIRCLE:
	        if (!xvw_getextent_circle(object, erase, &x, &y, &r))
	           return(FALSE);

	        if (r == 0)
                {
                    kerror(XVISUAL, "xvw_get_specified_roi",
                           "Cannot create ROI of zero radius");
                    return(FALSE);
                }
		xvw_getroi_circle(object, roi, x, y, r, val, 
				  fill, multiple, policy, d, t, e);
		break;

	   case KROI_ELLIPSE:
	        if (!xvw_getextent_ellipse(object, erase, &x, &y, &a, &b))
	           return(FALSE);

                if (a == 0 || b == 0)
                {
                    kerror(XVISUAL, "xvw_get_specified_roi",
                           "Cannot create ROI of zero width and/or height");
                    return(FALSE);
                }
		xvw_getroi_ellipse(object, roi, x, y, a, b, val, 
				   fill, multiple, policy, d, t, e);
		break;

	   case KROI_LINE:
	        if (!xvw_getextent_line(object, erase, &x_1, &y_1, &x_2, &y_2))
	           return(FALSE);

		xvw_getroi_line(object, roi, x_1, y_1, x_2, y_2, 
				val, multiple, policy, d, t, e);
		break;
	}
	return(TRUE);
}
/************************************************************
*
*  Routine Name: xvw_get_specified_roi - extracts the specified ROI from 
*					 an image object.
*
*       Purpose: Obtains a Region of Interest from the image object
*		 passed in, where the shape, location, and dimensions of the
*		 ROI are specified as parameters rather than interactively
*		 by the user;  meant for use with a GUI allowing keyboard
*		 extraction.  Provide values for ROI shape, the value to
*		 set in the mask for drawing, whether to fill the ROI shape
*		 or not, and ROI presentation policy.  For the remaining
*		 positional and sizing arguments, specify desired values for 
*		 those relevant to to the ROI shape specified; the others
*		 are ignored.  This routine does not support multiple ROI
*		 extraction.
*
*         Input: image    - image visual object displaying image from which
*		            ROI is to be extracted
*                shape    - the ROI shape to extract; one of:
*    	                    KROI_RECTANGLE, KROI_CIRCLE, KROI_ELLIPSE, or 
*		            KROI_LINE.  Note that neither KROI_POLYLINE or 
*                           KROI_FREEHAND are supported.
*                val      - the value to set in the mask for drawing
*                multiple - TRUE if extracting multiple ROI's, FALSE otherwise
*                policy   - the ROI Presentation policy, one of:
*                           KROI_INSIDE, KROI_OUTLINE, or KROI_OUTSIDE.
*                           Note: if object passed in is not an image object,
*                           just use KROI_INSIDE.
*		 x, y    -  For a shape of KIMAGE_ROI_RECTANGLE, x and y
*			    specify the upper left hand corner of the
*			    rectangle.  For a shape of KIMAGE_ROI_CIRCLE or,
*                           KIMAGE_ROI_ELLIPSE, x and y specify the center of 
*	                    the circle or the ellipse. For a shape of 
*		            KIMAGE_ROI_LINE, x and y specify one endpoint of
*		            the line.
*		w, h     - For a shape of KIMAGE_ROI_RECTANGLE, w and h
*		           specify the width and height of the rectangle.
*               radius   - For a shape of KIMAGE_ROI_CIRCLE, specifies the
*                          radius of the circle.
*		a, b     - For a shape of KIMAGE_ROI_ELLIPSE, a and b
*		           specify the width and height of the ellipse.
*		x2, y2   - For a shape of KIMAGE_ROI_LINE, x2 and y2
*		           specify the second endpoint of the line.
*
*        Output: 
*
*	Returns: The ROI kobject on success, NULL on failure
*
*  Restrictions:
*    Written By: Danielle Argiro
*          Date: March 23, 1996
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/

kobject xvw_get_specified_roi(
   xvobject image,
   int      shape,
   int      val,
   int      policy,
   int      x,
   int      y,
   unsigned int w,
   unsigned int h,
   unsigned int radius,
   unsigned int a,
   unsigned int b,
   int      x2,
   int      y2)
{
	kobject iobject; 
	kobject roi = NULL;
        int     d, t, e;
	int     fill, multiple = FALSE;
        int     xoffset, yoffset, type, reload, update;
        kaddr   data;
	
	fill = (policy != KIMAGE_ROI_OUTLINE);

        xvw_get_attributes(image,
                           XVW_IMAGE_IMAGEOBJ, &iobject,
                           XVW_IMAGE_XOFFSET,  &xoffset,
                           XVW_IMAGE_YOFFSET,  &yoffset,
                           NULL);

        kpds_get_attribute(iobject, KPDS_VALUE_SIZE, NULL, NULL, &d, &t, &e);

        if (shape == KIMAGE_ROI_RECTANGLE)
        {
	    if (w == 0 || h == 0)
	    {
		kerror(XVISUAL, "xvw_get_specified_roi",
		       "Cannot create ROI of zero width and/or height");
		return(NULL);
	    }
	    roi = xvw_getroi_rectangle(image, NULL, x, y, w, h, val, 
                                            fill, multiple, policy, d, t, e);
        }

        else if ((shape == KIMAGE_ROI_POLYLINE) ||
                 (shape == KIMAGE_ROI_FREEHAND))
        {
            kerror(XVISUAL, "xvw_get_specified_roi",
                   "Sorry, polyline and freehand ROIs are not supported");
            return(NULL);
        }

        else if (shape == KIMAGE_ROI_LINE)
        {
	    roi = xvw_getroi_line(image, NULL, x, y, x2, y2, val,
                                  multiple, policy, d, t, e);
        }

        else if (shape == KIMAGE_ROI_CIRCLE)
        {
	    if (radius == 0)
            {
                kerror(XVISUAL, "xvw_get_specified_roi",
                       "Cannot create ROI of zero radius");
                return(NULL);
            }
            roi = xvw_getroi_circle(image, NULL, x, y, radius, val, fill, 
				    multiple, policy, d, t, e);
        }
        else if (shape == KIMAGE_ROI_ELLIPSE)
        {
            if (a == 0 || b == 0)
            {
                kerror(XVISUAL, "xvw_get_specified_roi",
                       "Cannot create ROI of zero width and/or height");
                return(NULL);
            }
            roi = xvw_getroi_ellipse(image, NULL, x, y, a, b, val, fill,
                                     multiple, policy, d, t, e);
        }
        kpds_get_attribute(iobject, KPDS_VALUE_DATA_TYPE, &type);
        kpds_create_value(roi);

        kpds_get_attribute(roi, KPDS_MASK_SIZE, &w, &h, &d, &t, &e);

        kpds_get_attribute(roi, KPDS_SUBOBJECT_POSITION,
                               &x, &y, NULL, NULL, NULL);

        /* copy over all global attributes */
        kpds_copy_global_attr(iobject, roi);

	/*
         * want the resulting ROI to have width & height set from mask,
         * but need to preserve the depth, time, and elements dimensions
         * from the value segment of the original kobject
         */

        kpds_get_attribute(roi, KPDS_MASK_SIZE, &w, &h, NULL, NULL, NULL);
        kpds_get_attribute(iobject, KPDS_VALUE_SIZE, NULL, NULL, &d, &t, &e);

        kpds_set_attributes(roi,
                    KPDS_VALUE_DATA_TYPE, type,
                    KPDS_VALUE_SIZE, w, h, d, t, e,
                    NULL);

        kpds_set_attributes(iobject,
                    KPDS_VALUE_POSITION, x+xoffset, y+yoffset, 0, 0, 0,
                    KPDS_VALUE_REGION_SIZE, w, h, d, t, e,
                    NULL);

        /*
         *  Get the data from the original image object, and then put it
         *  into the newly create image object.
         */
        xvw_get_type(iobject, &reload, &update);
        xvw_set_type(iobject, reload, IMAGE_ROI_UPDATE);
        data = kpds_get_data(iobject, KPDS_VALUE_REGION, NULL);
        if (data == NULL)
        {
               kpds_close_object(roi); roi = NULL;
               return(NULL);
        }
        xvw_set_type(iobject, reload, update);
        kpds_put_data(roi, KPDS_VALUE_ALL, data);

	if (kpds_query_map(iobject))
 	    kpds_copy_map(iobject, roi, FALSE);

	/* copy over all untouched remaining data sets */
        kpds_copy_remaining_data(iobject, roi);

        kfree_and_NULL(data);
	return(roi);
}
