 /*
  * Khoros: $Id: callback.c,v 1.14.2.2 1997/04/29 22:37:34 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: callback.c,v 1.14.2.2 1997/04/29 22:37:34 mirage Exp $";
#endif



/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		Generalized Routines
   >>>>		
   >>>>  Private:
   >>>>                 xvw_callback_intermediary()
   >>>>   Public:
   >>>>                 xvw_insert_callback()
   >>>>                 xvw_add_callback()
   >>>>                 xvw_remove_callback()
   >>>>                 xvw_call_callback()
   >>>>                 xvw_destroy_callbacks()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"


/*
 *  intermediary callback structure; one structure 
 *  per object that has a callback added in the global list
 *  of intermediary structures.  
 */
typedef struct _xvw_intermediary
{
   ktoken      type;         /* our callback type, eg. XVW_BUTTON_SELECT  */
   int	       busy;         /* to busy the window upon invoking callback */
   xvobject    object;       /* object to which callback is added         */
   kaddr       client_data;      /* client data for callback              */
   void  (*callback_routine)(xvobject, kaddr, kaddr); /* callback routine */
   Time        time;		 /* list action time entry		  */
   int	       indx;		 /* list action indx entry		  */
}  xvw_intermediary;


/*-----------------------------------------------------------
|
|  Routine Name: xvw_callback_intermediary()
|
|       Purpose: In order to support multiple object sets, this routine
|                must be called to mediate when button is clicked on,
|                as opposed to the callback that was installed
|                on the scroll bar being called directly. 
|         Input: object      - scroll bar on which callback was installed.
|                client_data - a general client_data structure imposed by
|                              xvw_insert_callback, which includes the type and
|                              the name of the callback, as well as the pointer
|                              to the original client_data.
|                call_data   - the structure passed in as the call_data will
|                              vary, according to the object set used;  in all
|                              cases, it will indicate the end user's movement
|                              of the scroll bar.
|        Output: none
|       Returns: none
|    Written By: Danielle Argiro & Mark Young
|          Date: Feb 26, 1993
| Modifications:
|
------------------------------------------------------------------*/

void xvw_callback_intermediary(
   Widget widget,
   kaddr  client_data,
   kaddr  call_data)
{
	void (*callback_routine)(xvobject, kaddr, kaddr);

	int      busy;
	xvobject object;
        xvw_intermediary *intermediary;

        /*
         * the intermediary client_data holds the callback to call
         * as well as the original client_data
         */
        intermediary = (xvw_intermediary *) client_data;
	object = intermediary->object;

        /*
         *  call the callback that was originally specified
         *  for execution on the click of the button.
         */
	busy = intermediary->busy;
	callback_routine = intermediary->callback_routine;
	if (callback_routine)
	{
	   if (busy == TRUE)
	      xvw_busy(NULL, TRUE);

	   (*callback_routine) (object, intermediary->client_data, call_data);
	   if (busy == TRUE)
	      xvw_busy(NULL, FALSE);
	}
}

/************************************************************
*
*  Routine Name: xvw_insert_callback - insert a callback to a GUI object
*
*       Purpose: Installs a callback on a GUI object.  The only difference
*		 between this routine and \fIxvw_add_callback()\fP is that it
*		 allows you to specify whether the object should be busied
*		 upon invocation of the callback.  This is typically used in 
*		 callbacks that won't take a lot of time, and thus don't want 
*		 to incur the unnecessary overhead involved with making them
*		 "busy".
*
*         Input: object      - object on which to add the raw callback.
*                type        - event type on which to add callback, one of:\f(CW
*                !             XVW_BUTTON_SELECT
*                !             XVW_DESTROY
*                !             XVW_GENERAL
*                !             XVW_LIST_HIGHLT_ELEM
*                !             XVW_LIST_ITEM_ACTION
*                !             XVW_LIST_ITEM_SELECT
*                !             XVW_LIST_UNHIGHLT_ELEM
*                !             XVW_SCROLL_CONT_MOTION
*                !             XVW_SCROLL_INCR_MOTION
*                !
*                !             XVW_DOUBLE_CALLBACK
*                !             XVW_ERROR_CALLBACK
*                !             XVW_FLOAT_CALLBACK
*                !             XVW_GEOMETRY_CALLBACK
*                !             XVW_HELP_CALLBACK
*                !             XVW_INFO_CALLBACK
*                !             XVW_INPUTFILE_CALLBACK
*                !             XVW_INTEGER_CALLBACK
*                !             XVW_LAYOUT_CALLBACK
*                !             XVW_OUTPUTFILE_CALLBACK
*                !             XVW_SELECT_CALLBACK
*                !             XVW_TEXTDISPLAY_CALLBACK
*                !             XVW_TEXTINPUT_CALLBACK
*                !             XVW_WARN_CALLBACK\fP
*
*		 busy	     - whether to busy the windows when invoking the
*			       callback
*
*                callback_routine - callback routine to install 
*                client_data - pointer to private application data that
*                              will be passed to callback routine
*        Output: 
*	Returns: 
*  Restrictions:
*    Written By: Mark Young & Danielle Argiro
*          Date: Aug 03, 1994
*      Verified: 
*  Side Effects:
* Modifications:
*
*******************************************************************/

void xvw_insert_callback(
   xvobject   object,
   char       *type,
   int	      busy,
   void       (*callback_routine)(xvobject, kaddr, kaddr),
   kaddr      client_data)
{
	Widget widget = xvw_widget(object);

	ktoken		 token;
	xvw_intermediary *intermediary;

	/*
	 *  allocate intermediary structure, return on failure
	 */
	if ((intermediary = (xvw_intermediary *) kcalloc((size_t) 1, 
				sizeof(xvw_intermediary))) == NULL)
	{
	    kerror("xvwidgets", "xvw_insert_callback",
		   "Unable to allocate intermediary structure for list "
		   "callback, cannot add callback");
	    return;
	}
	token = kstring_to_token(type);
	intermediary->type        = token;
	intermediary->busy        = busy;
	intermediary->object      = object;
        intermediary->client_data = client_data;
        intermediary->callback_routine = callback_routine;

	/*
	 *  If the list of widget class callback lists is NULL, add the
         *  callback directly using the specified type.
	 */
	if (kstrcmp(type, XVW_DESTROY) != 0)
	{
	   XtAddCallback(widget, type, xvw_callback_intermediary, (kaddr)
			intermediary);
	}
	object->callbacks = klist_insert(object->callbacks,
		(kaddr) object, (kaddr) intermediary, KLIST_HEAD, TRUE);
}

/************************************************************
*
*  Routine Name: xvw_add_callback - add a callback to a GUI object
*
*       Purpose: Installs a callback on a GUI object.  When the GUI object
*                is used as specified by the \fItype\fP argument, the 
*                callback will be called. 
*
*		 A callback \fImust\fP be defined as follows:\f(CW
*		 !void callback(
*		 !    xvobject object,
*		 !    kaddr  client_data,
*		 !    kaddr  call_data)\fP
*
*                !\fIobject -\fP
*    		 !     The GUI object for which the callback was invoked.
*
*                !\fIclient_data -\fP
*                !     The pointer to the private client data, used to pass 
*                !     parameters from the application to the callback routine.
*                !\fIcall_data -\fP
*                !     The \fIcall_data\fP is the mechanism through which the 
*                !     GUI object itself can pass parameters to a callback when *                !     applicable.  The structure type of the call_data is 
*                !     defined by the GUI object; it must be cast to its correct
*                !     structure type before being used in the callback.  
*                !     Please see the documentation on the particular GUI object
*                !     of interest for the definition and use of the call_data 
*                !     pointer by a particular GUI object.
*
*         Input: object      - object on which to add callback.
*                type        - event type on which to add callback, one of:\f(CW
*		
*                !             XVW_BUTTON_SELECT
*                !             XVW_DESTROY
*                !             XVW_GENERAL
*                !             XVW_LIST_HIGHLT_ELEM
*                !             XVW_LIST_ITEM_ACTION
*                !             XVW_LIST_ITEM_SELECT
*                !             XVW_LIST_UNHIGHLT_ELEM
*                !             XVW_SCROLL_CONT_MOTION
*                !             XVW_SCROLL_INCR_MOTION
*		 !
*		 !             XVW_DOUBLE_CALLBACK
*		 !             XVW_ERROR_CALLBACK
*		 !             XVW_FLOAT_CALLBACK
*		 !             XVW_GEOMETRY_CALLBACK
*		 !             XVW_HELP_CALLBACK
*		 !             XVW_INFO_CALLBACK
*		 !             XVW_INPUTFILE_CALLBACK
*		 !             XVW_INTEGER_CALLBACK
*		 !             XVW_LAYOUT_CALLBACK
*		 !             XVW_OUTPUTFILE_CALLBACK
*		 !             XVW_SELECT_CALLBACK
*		 !             XVW_TEXTDISPLAY_CALLBACK
*		 !             XVW_TEXTINPUT_CALLBACK
*		 !             XVW_WARN_CALLBACK\fP
*
*                callback_routine - callback routine to install 
*                client_data - pointer to private application data that
*                              will be passed to callback routine
*        Output: 
*	Returns: 
*  Restrictions:
*    Written By: Danielle Argiro & Mark Young
*          Date: Apr 9, 1993
*      Verified: 
*  Side Effects:
* Modifications:
*
*******************************************************************/

void xvw_add_callback(
   xvobject   object,
   char       *type,
   void       (*callback_routine)(xvobject, kaddr, kaddr),
   kaddr      client_data)
{
	xvw_insert_callback(object, type, TRUE, callback_routine, client_data);
}

/************************************************************
*
*  Routine Name: xvw_remove_callback - remove a callback from a GUI object
*
*       Purpose: Removes a callback from a GUI object.
*
*         Input: object      - GUI object from which to remove callback.
*                type        - event type from which to remove callback, one of:
*                                                   \f(CW
*                !             XVW_BUTTON_SELECT
*                !             XVW_DESTROY
*                !             XVW_GENERAL
*                !             XVW_LIST_HIGHLT_ELEM
*                !             XVW_LIST_ITEM_ACTION
*                !             XVW_LIST_ITEM_SELECT
*                !             XVW_LIST_UNHIGHLT_ELEM
*                !             XVW_SCROLL_CONT_MOTION
*                !             XVW_SCROLL_INCR_MOTION
*                !
*                !             XVW_DOUBLE_CALLBACK
*                !             XVW_ERROR_CALLBACK
*                !             XVW_FLOAT_CALLBACK
*                !             XVW_GEOMETRY_CALLBACK
*                !             XVW_HELP_CALLBACK
*                !             XVW_INFO_CALLBACK
*                !             XVW_INPUTFILE_CALLBACK
*                !             XVW_INTEGER_CALLBACK
*                !             XVW_LAYOUT_CALLBACK
*                !             XVW_OUTPUTFILE_CALLBACK
*                !             XVW_SELECT_CALLBACK
*                !             XVW_TEXTDISPLAY_CALLBACK
*                !             XVW_TEXTINPUT_CALLBACK
*                !             XVW_WARN_CALLBACK\fP
*
*                client_data - pointer to private application data that
*                              was being passed to callback routine
*        Output:
*	Returns:
*  Restrictions:
*    Written By: Danielle Argiro & Mark Young
*          Date: Apr 9, 1993
*      Verified: 
*  Side Effects:
* Modifications:
*
*******************************************************************/
/* ARGSUSED */
void xvw_remove_callback(
   xvobject   object,
   char       *type,
   void       (*callback_routine)(xvobject, kaddr, kaddr),
   kaddr      client_data)
{
	ktoken token;
	Widget widget = xvw_widget(object);
	klist  *list, *entry = NULL;
	xvw_intermediary *intermediary = NULL;

	/*
	 *  find the intermediary to be deleted
	 */
	list = object->callbacks;
	token = kstring_to_token(type);
	while ((list = klist_locate(list, (kaddr) object)) != NULL)
	{
	   intermediary = (xvw_intermediary *) klist_clientdata(list);
	   if (intermediary->type == token)
	   {
	      if (intermediary->client_data == client_data)
	      {
	         entry = list;
		 break;
	      }
	      else entry = list;
	   }
	   list = klist_next(list);
	}

	intermediary = (xvw_intermediary *) klist_clientdata(entry);
	if (!intermediary || intermediary->type != token)
	{
	    kinfo(KXVLIB,"Can't find object from which to remove callback\n");
	    return;
	}

        /*
         *  If the widget is not an XVW_DESTROY callback remove it.
         */
	if (kstrcmp(type, XVW_DESTROY) != 0)
	{
           XtRemoveCallback(widget, type, xvw_callback_intermediary, (kaddr)
                            intermediary);
        }

	/*
	 *  Remove the intermidary
	 */
	entry = klist_delete(entry, (kaddr) object);
	object->callbacks = klist_head(entry);
	kfree_and_NULL(intermediary);
}

/************************************************************
*
*  Routine Name: xvw_call_callback - call a callback for a GUI object
*
*       Purpose: Calls a callback routine for a GUI object.  Currently, this
*		 routine does not support the passing of call_data, so
*		 if the callback routine expects to get calldata,
*		 \fIdo not\fP use this routine.
*
*         Input: object      - GUI object from which to remove callback.
*                type        - event type from which to call callback, one of:
*                !             XVW_BUTTON_SELECT
*                !             XVW_DESTROY
*                !             XVW_GENERAL
*                !             XVW_LIST_HIGHLT_ELEM
*                !             XVW_LIST_ITEM_ACTION
*                !             XVW_LIST_ITEM_SELECT
*                !             XVW_LIST_UNHIGHLT_ELEM
*                !             XVW_SCROLL_CONT_MOTION
*                !             XVW_SCROLL_INCR_MOTION
*                !
*                !             XVW_DOUBLE_CALLBACK
*                !             XVW_ERROR_CALLBACK
*                !             XVW_FLOAT_CALLBACK
*                !             XVW_GEOMETRY_CALLBACK
*                !             XVW_HELP_CALLBACK
*                !             XVW_INFO_CALLBACK
*                !             XVW_INPUTFILE_CALLBACK
*                !             XVW_INTEGER_CALLBACK
*                !             XVW_LAYOUT_CALLBACK
*                !             XVW_OUTPUTFILE_CALLBACK
*                !             XVW_SELECT_CALLBACK
*                !             XVW_TEXTDISPLAY_CALLBACK
*                !             XVW_TEXTINPUT_CALLBACK
*                !             XVW_WARN_CALLBACK
*        Output:
*	Returns:
*  Restrictions:
*    Written By: Mark Young
*          Date: May 22, 1994
*      Verified: 
*  Side Effects:
* Modifications:
*
*******************************************************************/
/* ARGSUSED */
void xvw_call_callback(
   xvobject   object,
   char       *type)
{
	ktoken     token;
	klist	   *callbacks;
	int	   busy = FALSE;
	void       (*callback_routine)(xvobject, kaddr, kaddr);
	xvw_intermediary *intermediary;

	token = kstring_to_token(type);
	callbacks = object->callbacks;
	while (callbacks != NULL)
	{
	    intermediary = (xvw_intermediary *) klist_clientdata(callbacks);
	    callback_routine = intermediary->callback_routine;
	    if (intermediary->type == token && callback_routine != NULL)
	    {
	       if (intermediary->busy && !busy)
	       {
                  xvw_busy(NULL, TRUE);
		  busy = TRUE;
	       }
               (*callback_routine) (object, intermediary->client_data, NULL);
	    }
	    callbacks = klist_next(callbacks);
	}

	if (busy == TRUE) xvw_busy(NULL, FALSE);
}

/************************************************************
*
*  Routine Name: xvw_destroy_callbacks - destroy all callbacks associated with
*					 a GUI object
*
*       Purpose: Destroys all callbacks installed on a GUI object.
*
*		 This function is automatically called when an object
*		 is destroyed, but is publicly provided since it is
*		 sometimes necessary to explicitly remove all callbacks 
*		 from an object.
*
*         Input: object - object/gadget to destroy all callbacks for
*
*    Written By: Mark Young
*          Date: May 19, 1994
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/

void xvw_destroy_callbacks(
   xvobject   object)
{
	klist  *list;
	xvw_intermediary *intermediary;

	/*
	 *  Now that we deleted the real event handler we need to cleanup
	 *  our handler event list.
	 */
	list = object->callbacks;
	while (list != NULL)
	{
	   intermediary = (xvw_intermediary *) klist_clientdata(list);
	   kfree_and_NULL(intermediary);
	   list = klist_next(list);
	}
	klist_free(object->callbacks, NULL);
	object->callbacks = NULL;
}
