 /*
  * Khoros: $Id: file.c,v 1.19.2.1 1997/04/29 22:37:46 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: file.c,v 1.19.2.1 1997/04/29 22:37:46 mirage Exp $";
#endif



/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		File Change Detect Routines
   >>>>		
   >>>>		These routines are used to detect whether a file
   >>>>		has been changed.  The file detect routine will
   >>>>		call the programmer's callback routine when the
   >>>>		file has been detected to be changed.  The program-
   >>>>		mer also specifies how often we are supposed
   >>>>		to check the file for being modified.
   >>>>		
   >>>>   Static:
   >>>>                 check_detectfile()
   >>>>                 free_detectfile()
   >>>>   Public:
   >>>>                 xvw_add_detectfile()
   >>>>                 xvw_call_detectfile()
   >>>>                 xvw_remove_detectfile()
   >>>>                 xvw_destroy_detectfiles()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"	


typedef struct _xvw_file
{
	xvobject  object;
	XtIntervalId id;
	time_t    time;
	char	  *filename;
	unsigned long interval;
	int      (*routine)(xvobject, char *, kaddr);
	kaddr	  client_data;
} xvw_file;

static klist *file_list = NULL;

/*-----------------------------------------------------------
|
|  Routine Name: check_detectfile - checks to see if a file has changed
|
|       Purpose: Checks to see if a file has changed.  If so
|                we call the user's callback routine with the
|                id and client data.  If the user returns TRUE
|		 then we update the time otherwise we will check
|		 the file again with the same time.
|
|         Input: data - our xvw_file structure
|                id   - the current timeout id
|    Written By: Mark Young
|          Date: Apr 26, 1996
|
------------------------------------------------------------*/
/* ARGSUSED */
static void check_detectfile(
   kaddr        data,
   XtIntervalId *id)
{
	struct stat buf, tmp_buf;
	int (*routine)(xvobject, char *, kaddr);
	xvw_file  *handler = (xvw_file *) data;


	/*
	 *  Get the stat's for the handler filename, unless we have a force
	 *  callback.  If a force callback that simply do the callback no
	 *  matter what state we are in.
	 */
	if (stat(handler->filename, &buf) != -1)
	{
	   /*
	    *  check to see if the file has been modified since the last
	    *  time we read it.
	    */
	   if (handler->time != buf.st_mtime)
	   {
	      /*
	       *  since the file has been modified call the user's callback
	       *  routine and update the read time.
	       */
	      routine    = handler->routine;

	      tmp_buf.st_mtime = handler->time;
	      handler->time = buf.st_mtime;
	      if (!routine(handler->object, handler->filename,
		  	   handler->client_data))
	         handler->time = tmp_buf.st_mtime;
	      else if (stat(handler->filename, &buf) != -1)
		 handler->time = buf.st_mtime;
	   }
	}

	/*
	 *  Re-add check_detectfile to the time out handler.
	 */
	if (handler->interval > 0)
	{
	   handler->id = XtAppAddTimeOut(xvw_appcontext(NULL),
			handler->interval, check_detectfile, data);
	}
}

/*-----------------------------------------------------------
|
|  Routine Name: free_detectfile - frees single detect handler entry from 
|			           inpout handler list
|
|       Purpose: This utility frees a single action handler entry from
|		 the action list of an object, called by xvw_destroy_detectfile()
|		 to delete individual handlers from the object detect handler 
|		 list.
|
|         Input: listptr - pointer to detect handler entry to be freed
|    Written By: Danielle Argiro
|          Date: August 18, 1995
|
------------------------------------------------------------*/

static void free_detectfile(
    klist *listptr)
{
	xvw_file *handler;

	handler = (xvw_file *) klist_clientdata(listptr);
	XtRemoveTimeOut(handler->id);
	kfree_and_NULL(handler->filename);
	kfree_and_NULL(handler);
}


/************************************************************
*
*  Routine Name: xvw_add_detectfile - add a (file) detect handler to an object
*
*       Purpose: Causes a detection mechanism to be installed on the
*                specified file.  After the specified interval of time
*                has elapsed, the file will be checked for any modification.
*		 If a modification to the file is detected, then the specified
*                detect handler is called.
*
*		 The detect handler can be associated with an object, so that
*		 file detection is automatically discontinued when the object is
*		 destroyed.  If NULL is passed for the object, then the detect
*                handler is added to the global file detection list.
*
*	         !The detect file callback must be declared in the following
*		 !form:\f(CW
*	         !int detect_handler(
*	         !     xvobject object,
*	         !     char *filename,
*	         !     kaddr client_data)\fP
*
*		 !\fIobject -\fP 
*		 !    If \fIxvw_add_detectfile()\fP is called with a particular 
*                !    xvobject, that object will be passed into the detect handler.
*
*		 !\fIfilename -\fP
*                !    This is the name of the file being monitored for change.
*
*
*                !\fIclient_data -\fP
*                !    The pointer to the client data, used to pass parameters
*                !    from the application to the detect handler.
*
*         Input: object      - object on which to place the detect handler.  
*                              Pass NULL if the detect handler is to be invoked 
*                              in general, and not associated with any 
*                              particular object.
*		 filename    - the name of the file to be monitored.
*                argtime     - the interval (in seconds) at which the file 
*                              should be checked for changes.
*                routine     - the detect handler routine to be called when 
*                              change to the file is detected.
*                client_data - pointer to client data that
*                              will be passed to detect handler
*        Output:
*	Returns: 
*  Restrictions: 
*    Written By: Mark Young
*          Date: Jul 15, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
void xvw_add_detectfile(
   xvobject  object,
   char      *filename,
   double    argtime,
   int       (*routine)(xvobject, char *, kaddr),
   kaddr     client_data)
{
	char     *fullpath;
	xvw_file *handler;
	struct stat buf;
	klist    *filelist_ptr;

	/*
	 *  Only allow a particular file handler to be added to the
         *  same file once.
	 */
	filelist_ptr = file_list;
	while (filelist_ptr != NULL)
	{
	     handler = (xvw_file *) klist_clientdata(filelist_ptr);
	     if ((handler->routine == routine) &&
		 (kstring_to_token(filename) == 
		  kstring_to_token(handler->filename)))
		 return;
	     filelist_ptr = klist_next(filelist_ptr);
	}

	/*
	 *  Get the last modified time of the file in question.
	 */
	if ((fullpath = kfullpath(filename, NULL, NULL)) == NULL)
	{
	   errno = KINVALID_FILE;
	   kerror("xvutils", "xvw_add_detectfile", 
		  "Invalid filename '%s' found.", !filename ? "(NULL)" :
		 filename);
	   return;
	}

	if ((handler = (xvw_file *) kcalloc((size_t) 1, 
					sizeof(xvw_file))) == NULL)
	{
	   kerror("xvutils", "xvw_add_detectfile", 
		  "Cannot allocate memory for xvw_file structure");
	   return;
	}

	if (stat(fullpath, &buf) != -1)
	   handler->time = buf.st_mtime;
	else
	   handler->time = time(NULL);

	/*
	 *  update the rest of detect structure info.  We also need
	 *  compute the update interval in terms of millaseconds instead
	 *  of seconds.
	 */
	handler->object = object;
	handler->interval = argtime * 1000;
	handler->filename = fullpath;
	handler->routine  = routine;
	handler->client_data  = client_data;
	if (handler->interval > 0 && (!kipc_getipc() || object != NULL))
	{
	   handler->id = XtAppAddTimeOut(xvw_appcontext(NULL),
				handler->interval, check_detectfile,
				(kaddr) handler);
	}

	/*
	 *  Add the handler to the detect file handler list.  If the object is
	 *  NULL then add it to the global file list.  
	 */
	if (!object)
	{
	   file_list = klist_insert(file_list, (kaddr)
			kstring_to_token(filename), handler, KLIST_HEAD, TRUE);
	}
	else
	{
	   object->detectfile = klist_insert(object->detectfile, (kaddr)
			kstring_to_token(filename), handler, KLIST_HEAD, TRUE);
	}
}

/************************************************************
*
*  Routine Name: xvw_call_detectfile - call a (file) detect handler for
*				       an object
*
*       Purpose: Causes the detection mechanism previously installed with
*                \fIxvw_add_detectfile()\fP to be called for the specified
*                file.
*
*         Input: object      - object in which to call the detect file
*			       handler, if NULL then call from the global
*			       list.
*		 filename    - the name of the file that is being monitored.
*        Output:
*	Returns:
*  Restrictions: 
*    Written By: Mark Young
*          Date: Jul 15, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
void xvw_call_detectfile(
   xvobject  object,
   char      *filename)
{
	ktoken   name;
	xvw_file *handler;
	struct stat buf, tmp_buf;
	int (*routine)(xvobject, char *, kaddr);
	klist	 *temp, **list = !object ? &file_list : &object->detectfile;

	temp = *list; name = kstring_to_token(filename);
#if 0
	kfprintf(kstderr, "Inside call handler.\n");
#endif

	while (temp != NULL)
	{
	   handler = (xvw_file *) klist_clientdata(temp);
	   temp = klist_next(temp);

	   /*
	    *  Get the stat's for the handler filename, unless we have a force
	    *  callback.  If a force callback that simply do the callback no
	    *  matter what state we are in.
	    */
	   if (/*handler->time != 0 &&*/ !knegotiate_check(handler->filename))
	   {
#if 0
	      kfprintf(kstderr,"  No change for file '%s'\n",handler->filename);
#endif
	      continue;
	   }

	   /*
	    *  since the file has been modified call the user's callback
	    *  routine and update the read time.
	    */
	   routine = handler->routine;

	   tmp_buf.st_mtime = handler->time;
	   handler->time = buf.st_mtime;
	   if (!routine(handler->object,handler->filename,handler->client_data))
	      handler->time = tmp_buf.st_mtime;
	   else if (stat(handler->filename, &buf) != -1)
	      handler->time = buf.st_mtime;
#if 0
	   kfprintf(kstderr, "  Calling handler for %s\n", handler->filename);
#endif
	}
#if 0
	kfprintf(kstderr, "Leaving call handler\n");
#endif
}

/************************************************************
*
*  Routine Name: xvw_remove_detectfile - remove a (file) detect handler 
*					 from an object
*
*       Purpose: Causes the detection mechanism previously installed with
*                \fIxvw_add_detectfile()\fP to be removed from the specified
*                file descriptor.
*
*                If the detect handler was associated with an object, the
*                file detection will be automatically removed when the object
*                is destroyed. Alternatively, it can be removed before the
*		 object is destroyed by using this routine.  If NULL was 
*                passed to \fIxvw_add_detectfile()\fP for the object, then 
*                you must call this routine if you need to remove the installed 
*                detect handler.
*
*         Input: object      - object in which to remove the detect file
*			       handler, if NULL then removed from the global
*			       list.
*		 filename    - the name of the file that was being monitored.
*                routine     - the detect handler routine which was being
*			       called when change to the file was detected.
*                client_data - pointer to private application data that
*                              was being passed to action handler
*        Output:
*	Returns:
*  Restrictions: 
*    Written By: Mark Young
*          Date: Jul 15, 1993
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
void xvw_remove_detectfile(
   xvobject  object,
   char      *filename,
   int       (*routine)(xvobject, char *, kaddr),
   kaddr     client_data)
{
	ktoken   name;
	xvw_file *handler;
	klist	 *temp, **list = !object ? &file_list : &object->detectfile;

	temp = *list; name = kstring_to_token(filename);
	while ((temp = klist_locate(temp, (kaddr) name)) != NULL)
	{
	   handler = (xvw_file *) klist_clientdata(temp);
	   if (routine == handler->routine &&
	       client_data == handler->client_data)
	   {
	      XtRemoveTimeOut(handler->id);
	      kfree_and_NULL(handler->filename); kfree_and_NULL(handler);
	      temp = klist_delete(temp, temp->identifier);
	      *list = klist_head(temp);
	      return;
	   }
	   temp = klist_next(temp);
	}
}

/************************************************************
*
*  Routine Name: xvw_destroy_detectfiles - destroy all file detect handlers 
*					  currently installed on an object
*
*       Purpose: Destroys all file detect handlers currently installed on a
*                GUI or visual object.
*
*                This function is automatically called when an object
*                is destroyed, but is publicly provided since it is
*                sometimes necessary to explicitly remove all action handlers
*                from an object.
*
*         Input: object - object for which to destroy all action handlers
*
*    Written By: Danielle Argiro
*          Date: August 18, 1995
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/

void xvw_destroy_detectfiles(
   xvobject   object)
{
        if (object == NULL)
	    return;

	klist_free(object->detectfile, free_detectfile);
}
