
/*
 * Copyright (c) 1996,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */
 
#include <geometry.h>
#include "gextents.h"

/**
  *
  * the vertices of the box are in the following order, where index
  * 0 (zero) corresponds to the "min" location, and index 7 corresponds
  * to the "max" location.
  *
  * the extents box is dispatched off to geom services as a series of
  * disjoint line segments, a total of 12 in all.  the x-lines are
  * first, followed by the y-lines, followed by the z-lines.
  *
  * if the user chooses the have a "colored" box, the x-lines will be
  * red, the y-lines will be green, and the z-lines will be cyan (blue
  * is too hard to see).
  *
  *                   6                 7
  *                   -----------------
  *                  /|              /|
  *              2  /---------------/3|
  *                 | |             | |
  *                 | |4            | |
  *                 | +-------------|-+ 5
  *                 |/              |/
  *               0 +---------------+ 1
  *
  * the box type flag will affect the way in which the box is computed.
  * a type of "min/max" will always produce a rectangular box, the corners
  * of which represent the absolute min/max of the location data.
  *
  * a box type of "bounding hull" will cause a different type of box to
  * be produced.  this box more accurately reflects that shape of the grid
  * defined by the location data, and is especially useful for curvilinear
  * grids.  the hull produced by this box type will cause a number of
  * polyline segments to be generated, one for each of the 12 edges of
  * logical grid.  the number of segments is a function of the amount of
  * data - this box takes up much more space than the min/max box.
  *
  * for nonsensical situations, such as trying to make a hull from
  * scatter (1-d, 3 space) data, the default recovery procedure will be
  * to create a min/max box without issuing an error message.
  *
  * NOTE: if no location data is present, or if the grid is a rectilinear
  * one, the results of the two box types will be identical.
**/
  

static rgb_vertex box_colors[12]=
{
    {{1.,0.,0.}},
    {{1.,0.,0.}},
    {{1.,0.,0.}},
    {{1.,0.,0.}},
    {{0.,1.,0.}},
    {{0.,1.,0.}},
    {{0.,1.,0.}},
    {{0.,1.,0.}},
    {{0.,1.,1.}},
    {{0.,1.,1.}},
    {{0.,1.,1.}},
    {{0.,1.,1.}}
};

static rgb_vertex uplane_color = {1.,0.,0.};
static rgb_vertex vplane_color = {0.,1.,0.};
static rgb_vertex wplane_color = {0.,1.,1.};

#define UAXIS 0
#define VAXIS 1
#define WAXIS 2

#define RMIN 0
#define RMAX 1

static void make_hull PROTO((kobject,kgeom_object*,int,int,int,int,int,int,int,int,int,int,int));
static void make_minmax_box PROTO((kobject,kgeom_object *,int,int,int,int));
static void construct_2d_wire_mesh PROTO((kobject,kgeom_object *,int,int,int,int,int,rgb_vertex *,int *));
static void do_hull PROTO((kgeom_object *,vertex_3d *,vertex_3d *,int,int,int *,rgb_vertex *,int,int,int));
int varray_transpose PROTO((vertex_3d *,vertex_3d **,int,int));

/*-----------------------------------------------------------
| 
|  Routine Name: make_extents_box
| 
|       Purpose: compute the extents, create the geometry
b| 
|         Input: kobject - containing the data from which to make the box.
| 
|        Output: kobject - containing the geometry representing the box.
|    Written By: wes
|          Date: May 19, 1993
| Modifications: 
|                *** 14 June 1993, wes
|                make use of lget_coords() library call, remove
|                _all_ data services references to location data.
|
|                *** 6 July 1993 wes
|                add box type parameter and supporting code.
| 
------------------------------------------------------------*/
int make_extents_box(kobject input,   /* input data */
		     kgeom_object *output,  /* output geometry */
		     int     style,   /* colored vs. white flag */
		     char   *name,    /* name of geom object */
		     int     type,/* hull vs. min/max flag */
		     int show_bounds,
		     int umin,
		     int umax,
		     int vmin,
		     int vmax,
		     int wmin,
		     int wmax)
{
    int has_colors;
    int val_avail,loc_avail;
    int dw,dh,dd,dt,de;  /* dimensions of value data. */
    int do_hull;

    /**
      * if style is 1, then the user hasn't specified anything
      * for the style, so they get a "white" box (white by the
      * fact that when no colors are specified, the default
      * object color is white).  if the style is 2 (not equal
      * to 1 at this point in time), then the box
      * gets colors.  the colors are such that x-lines are red,
      * y lines are green and z lines are cyan.
    **/
    
    has_colors = (style == 1) ? 0 : 1;

    /* check the availability of value data */
    val_avail = kpds_query_value(input);
    loc_avail = kpds_query_location(input);

    /* get the dimensions of the value data */
    if (val_avail != FALSE)
    {
	kpds_get_attribute(input,KPDS_VALUE_SIZE,&dw,&dh,&dd,&dt,&de);
#if 0
	fprintf(kstderr," data dims: w,h,d,t,e = %d,%d,%d,%d,%d \n",dw,dh,dd,dt,de);
#endif
	/**
	  * figure out if there's 3D of data, and if the user has asked
	  * for a hull.  if so, set the internal flag so that gets
	  * computed.  otherwise, set the flag so that a min/max box
	  * gets computed instead.  no warning msg gets issues.
	**/
	if ((dw != 1) && (dh != 1) && (dd != 1) && (type==2))
	    do_hull = 1;
	else
	    do_hull = 0;
    }
    else
    {
	/* no value data -- see if there's location data */
	/* loc_avail = kpds_query_location(input);  << do this earlier */
	if (loc_avail != FALSE)
	{
	    kpds_get_attribute(input,KPDS_LOCATION_SIZE,&dw,&dh,&dd,&dt);
#if 0
	    fprintf(kstderr," location dims: w,h,d,t,e = %d,%d,%d,%d,%d \n",dw,dh,dd,dt,de);
#endif
	    /**
	     * figure out if there's 3D of data, and if the user has asked
	     * for a hull.  if so, set the internal flag so that gets
	     * computed.  otherwise, set the flag so that a min/max box
	     * gets computed instead.  no warning msg gets issues.
	     **/
	    if ((dw != 1) && (dh != 1) && (dd != 1) && (type==2))
		do_hull = 1;
	    else
		do_hull = 0;

	}
	else
	{
/* KERROR CAT: "geom_annotate" */
	    kerror("GEOMETRY","gextents","there is no value or location data in the input file.");
	    return(WHACKED);
	}
    }

    /* -- if we have value but no location, create location -- */
    if (val_avail && !loc_avail)  
    {
       /* -- create uniform location -> this can be accessed as
          -- curvilienar location later in this routine 
	  -- */
       kpds_set_attribute(input, KPDS_LOCATION_GRID, KUNIFORM);
       kpds_create_location(input);
       
       /* -- create uniform location from (0,0,0)<->(dw,dh,dd) -- */
       kpds_set_attributes(input, 
			   KPDS_LOCATION_BEGIN, 0., 0., 0.,
			   KPDS_LOCATION_END, (double)dw, 
			   (double)dh, (double)dd, NULL);
    }

    /* -- assert that we want floating point location (x,y,z) data -- */
    kpds_set_attributes(input, 
			KPDS_LOCATION_DATA_TYPE, KFLOAT,
			KPDS_LOCATION_SIZE, -1, -1, -1, 3, NULL);


    /* set the layout of color information to per-segment */
    output->layout = KPER_LINE;

    /* this is erroneous - hulls use multiple polyline primitives,
       while min/max boxes use only a ssingle primitive consisting
       of a bunch of disjoint line segments. */

    /* do the name */
    output->name = kstrdup(name);

    if (do_hull == 1)
    {
      make_hull(input,output,dw,dh,dd,has_colors,show_bounds,
		umin,umax,vmin,vmax,wmin,wmax);
    }
    else
    {
      if (show_bounds)
      {
	make_minmax_box(input,output,dw,dh,dd,has_colors);
      }
    }

    return(CHILL);
}

static void
make_minmax_box(kobject input,
		kgeom_object *output,
		int dw,
		int dh,
		int dd,
		int has_colors)
{
    vertex_3d min,max,center;
    vertex_3d box[24];    /* holds the vertices for the 12 box segments */
    float *color_info;
    float *plane;
    int i,j,k,l;
    int num_prims;
    kgeom_polyline *lines;

    
    for (i=0;i<3;i++)
    {
	min.v[i] = KGEOM_BIGNUM;
	max.v[i] = KGEOM_SMALLNUM;
    }

    /* -- get planes of location -- */
    plane = NULL;
    for (k=0;k<dd;k++)
    {
       for (l=0;l<3;l++)
       {
	  /* -- look for min/max points for plane k -- */
	  kpds_set_attribute(input, KPDS_LOCATION_POSITION, 0, 0, k, l);
	  plane = (float *)kpds_get_data(input, KPDS_LOCATION_PLANE, plane);
	  for (j=0;j<dh*dw;j++)
	  {
	     if (plane[j] > max.v[l])
		max.v[l] = plane[j];
	     if (plane[j] < min.v[l])
		min.v[l] = plane[j];
	  }
       }
    }
    kfree(plane);

#if 0
    for (k=0;k<dd;k++)
	for (j=0;j<dh;j++)
	    for (i=0;i<dw;i++)
	    {
		lget_coords(input,i,j,k,&t);

		for (l=0;l<3;l++)
		{
		    if (t.v[l] > max.v[l])
			max.v[l] = t.v[l];
		    if (t.v[l] < min.v[l])
			min.v[l] = t.v[l];
		}
	    }
#endif

    for (i=0;i<3;i++)
	center.v[i] = (max.v[i] + min.v[i]) * 0.5;

#if 0
    fprintf(kstderr," min point is %f,%f,%f\n",min.v[0],min.v[1],min.v[2]);
    fprintf(kstderr," max point is %f,%f,%f\n",max.v[0],max.v[1],max.v[2]);
#endif
    /* do line from u=0,v=0,w=0 to u=1,v=0,w=0 */
    memcpy((char *)(box+0),(char *)&min,sizeof(vertex_3d));
    memcpy((char *)(box+1),(char *)&min,sizeof(vertex_3d));
    box[1].v[0] = max.v[0];
    
    /* do line from u=0,v=1,w=0 to u=1,v=1,w=0 */
    memcpy((char *)(box+2),(char *)&min,sizeof(vertex_3d));
    box[2].v[1] = max.v[1];
    memcpy((char *)(box+3),(char *)&max,sizeof(vertex_3d));
    box[3].v[2] = min.v[2];
    
    /* do line from u=0,v=0,w=1 to u=1,v=0,w=1 */
    memcpy((char *)(box+4),(char *)&min,sizeof(vertex_3d));
    box[4].v[2] = max.v[2];
    memcpy((char *)(box+5),(char *)&max,sizeof(vertex_3d));
    box[5].v[1] = min.v[1];
    
    /* do line from u=0,v=1,w=1 to u=1,v=1,w=1 */
    memcpy((char *)(box+6),(char *)&max,sizeof(vertex_3d));
    box[6].v[0] = min.v[0];
    memcpy((char *)(box+7),(char *)&max,sizeof(vertex_3d));

    

    /* ok, now do the y lines */
    
    /* do line from u=0,v=0,w=0 to u=0,v=1,w=0 */
    memcpy((char *)(box+8),(char *)&min,sizeof(vertex_3d));
    memcpy((char *)(box+9),(char *)&min,sizeof(vertex_3d));
    box[9].v[1] = max.v[1];
    
    /* do line from u=1,v=0,w=0 to u=1,v=1,w=0 */
    memcpy((char *)(box+10),(char *)&min,sizeof(vertex_3d));
    box[10].v[0] = max.v[0];
    memcpy((char *)(box+11),(char *)&max,sizeof(vertex_3d));
    box[11].v[2] = min.v[2];
    
    /* do line from u=0,v=0,w=1 to u=0,v=1,w=1 */
    memcpy((char *)(box+12),(char *)&min,sizeof(vertex_3d));
    box[12].v[2] = max.v[2];
    memcpy((char *)(box+13),(char *)&max,sizeof(vertex_3d));
    box[13].v[0] = min.v[0];
    
    /* do line from u=1,v=0,w=1 to u=1,v=1,w=1 */
    memcpy((char *)(box+14),(char *)&max,sizeof(vertex_3d));
    box[14].v[1] = min.v[1];
    memcpy((char *)(box+15),(char *)&max,sizeof(vertex_3d));

    
    /* now do all the Z lines */
    
    /* do line from u=0,v=0,w=0 to u=0,v=0,w=1 */
    memcpy((char *)(box+16),(char *)&min,sizeof(vertex_3d));
    memcpy((char *)(box+17),(char *)&min,sizeof(vertex_3d));
    box[17].v[2] = max.v[2];
    
    /* do line from u=1,v=0,w=0 to u=1,v=0,w=1 */
    memcpy((char *)(box+18),(char *)&min,sizeof(vertex_3d));
    box[18].v[0] = max.v[0];
    memcpy((char *)(box+19),(char *)&max,sizeof(vertex_3d));
    box[19].v[1] = min.v[1];
    
    /* do line from u=0,v=1,w=0 to u=0,v=1,w=1 */
    memcpy((char *)(box+20),(char *)&min,sizeof(vertex_3d));
    box[20].v[1] = max.v[1];
    memcpy((char *)(box+21),(char *)&max,sizeof(vertex_3d));
    box[21].v[0] = min.v[0];
    
    /* do line from u=1,v=1,w=0 to u=1,v=1,w=1 */
    memcpy((char *)(box+22),(char *)&max,sizeof(vertex_3d));
    box[22].v[2] = min.v[2];
    memcpy((char *)(box+23),(char *)&max,sizeof(vertex_3d));

    /**
      * we need to tell geometry services about this object.
    **/

    /* ok, now put the data into the primitive */

    if (has_colors)
	color_info = (float *)box_colors;
    else
	color_info = NULL;
    
    lines = (kgeom_polyline *) kgeom_new_primitive(KGEOM_POLYLINE_DISJOINT);
    lines->nverts = 24;
    lines->locs = (float *) kgeom_dup_locs((float *)box, lines, output);
    lines->cols = (float *) kgeom_dup_cols((float *)color_info, lines, output);
    kgeom_add_primitive(output, (kgeom_primitive *)lines);

    kmemcpy((kaddr)output->min, (kaddr) &min, sizeof(float) * 3);
    kmemcpy((kaddr)output->max, (kaddr) &max, sizeof(float) * 3);
    kmemcpy((kaddr)output->center, (kaddr) &center, sizeof(float) * 3);
    output->has_bounding_box = TRUE;
    output->has_center       = TRUE;
}

static void
make_hull(kobject input,
	  kgeom_object *output,
	  int dw,
	  int dh,
	  int dd,
	  int has_colors,
	  int show_bounds,
	  int umin,
	  int umax,
	  int vmin,
	  int vmax,
	  int wmin,
	  int wmax)

{
    /*
     * the algorithm used for this function is as follows:
     * 1. for each of the six planes of the volume, use data services
     *    to grab the location data corresponding to the plane (we're
     *    assuming that data services is reasonably efficient at this.
     *    if performance sucks, then we should rewrite the code to do
     *    the re-ordering ourselves).
     * 2. construct the box using the six arrays of location data.
     */
    int lod=0;			/* placeholder */
    vertex_3d *uzero,*uone,*vzero,*vone,*wzero,*wone; /* the six planes */
    int num_prims=1;

    /* grab the coordinates at the w=0 plane */

    kpds_set_attribute(input,KPDS_ELEMENTS_FIRST,TRUE);
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,0,0,0,0);
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,
		       dw,dh,1,3);
    wzero = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,NULL);

    /* grab the coordinates at the v=0 plane */
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,0,0,0,0);
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,
		       dw,1,dd,3);
    vzero = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,NULL);

    /* grab the coordinates at the v=1 plane */
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,0,dh-1,0,0);
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,
		       dw,1,dd,3);
    vone = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,NULL);

    /* grab the coordinates at the u=0 plane */
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,0,0,0,0);
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,
		       1,dh,dd,3);
    uzero = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,NULL);

    /* grab the coordinates at the u=1 plane */
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,dw-1,0,0,0);
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,
		       1,dh,dd,3);
    uone = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,NULL);

    /* grab the coordinates at the w=1 plane */

    kpds_set_attribute(input,KPDS_ELEMENTS_FIRST,TRUE);
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,0,0,dd-1,0);
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,
		       dw,dh,1,3);
    wone = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,NULL);

    /* now we have all the location data we need...proceed to
     construct the geometry..*/
    {
	vertex_3d *wzero_t,*wone_t,*vzero_t,*vone_t,*uzero_t,*uone_t;

	varray_transpose(wzero,&wzero_t,dw,dh);
	do_hull(output,wzero,wzero_t,dw,dh,&num_prims,
		has_colors?&wplane_color:NULL,show_bounds,wmin,lod);
	varray_transpose(wone,&wone_t,dw,dh);
	do_hull(output,wone,wone_t,dw,dh,&num_prims,
		has_colors?&wplane_color:NULL,show_bounds,wmax,lod);

	varray_transpose(vzero,&vzero_t,dw,dd);
	do_hull(output,vzero_t,vzero,dd,dw,&num_prims,
		has_colors?&vplane_color:NULL,show_bounds,vmin,lod);
	varray_transpose(vone,&vone_t,dw,dd);
	do_hull(output,vone_t,vone,dd,dw,&num_prims,
		has_colors?&vplane_color:NULL,show_bounds,vmax,lod);

	varray_transpose(uzero,&uzero_t,dh,dd);
	do_hull(output,uzero,uzero_t,dh,dd,&num_prims,
		has_colors?&uplane_color:NULL,show_bounds,umin,lod);
	varray_transpose(uone,&uone_t,dh,dd);
	do_hull(output,uone,uone_t,dh,dd,&num_prims,
		has_colors?&uplane_color:NULL,show_bounds,umax,lod);

    }
}


static void
do_hull(kgeom_object *output,		/* the output geom obejct */
	vertex_3d *plane,	/* the input 2d plane of 3space coords */
	vertex_3d *plane_transpose,
	int usize,		/* size of I axis of plane */
	int vsize,		/* size of J axis of plane */
	int *nprims,		/* number of primitives in output geom */
	rgb_vertex *color,	/* the color to use for this plane */
	int show_bounds,        /* flag to control if exterior is built */
	int do_interior,        /* flag to control if interior is built */
	int lod)		/* level-of-detail flag (not implemented) */
{
    kgeom_polyline *lines;
    int i,j,n,nloops;
    rgb_vertex *c;
    vertex_3d *v,*tv,*source;
    int l_usize,l_vsize;
    int do_colors = 1;

    n = MAX(usize,vsize);

    do_colors = color ? 1 : 0;
    
    /* malloc space for vertices for a single polyline prim */
    v = (vertex_3d *)kmalloc(sizeof(vertex_3d)*n);
    /* malloc and load up the colors - we only need to do this once */
    if (do_colors)
    {
      c = (rgb_vertex *)kmalloc(sizeof(rgb_vertex)*n);
      for (i=0;i<n;i++)
      {
	kmemcpy(c+i,color,sizeof(rgb_vertex));
      }
    }

    nloops = (do_interior == 0) ? 1 : 2; 

    for (j=0;j<nloops;j++)
    {
	if (j==0)
	{
	    source = tv = plane;
	    l_usize = usize;
	    l_vsize = vsize;
	}
	else
	{
	    source = tv = plane_transpose;
	    l_usize = vsize;
	    l_vsize = usize;
	}

	if (show_bounds || do_interior)
	{
	  lines = (kgeom_polyline *) 
	  kgeom_new_primitive(KGEOM_POLYLINE_CONNECTED);
	  lines->nverts = l_usize;
	  lines->locs = kgeom_dup_locs((float *)tv, lines, output);
	  if (do_colors)
	  {
	    lines->cols = kgeom_dup_cols((float *)c, lines, output);
	  }
	  kgeom_add_primitive(output, (kgeom_primitive *) lines);
	}
	
	/* do the interior, if requested */
	if (do_interior)
	{
	    for (i=1;i<l_vsize-1;i++)
	    {
		tv += l_usize;

		lines = (kgeom_polyline *) 
		   kgeom_new_primitive(KGEOM_POLYLINE_CONNECTED);
		lines->nverts = l_usize;
		lines->locs = kgeom_dup_locs((float *)tv, lines, output);
		if (do_colors)
		{
		  lines->cols = kgeom_dup_cols((float *)c, lines, output);
		}
		kgeom_add_primitive(output, (kgeom_primitive *) lines);
	    }
	}

	if (show_bounds || do_interior)
	{
	  /* do the segment at v=1 */
	  tv = source + l_usize*(l_vsize-1);
	  
	  lines = (kgeom_polyline *) 
	  kgeom_new_primitive(KGEOM_POLYLINE_CONNECTED);
	  lines->nverts = l_usize;
	  lines->locs = kgeom_dup_locs((float *)tv, lines, output);
	  if (do_colors)
	  {
	    lines->cols = kgeom_dup_cols((float *)c, lines, output);
	  }
	  kgeom_add_primitive(output, (kgeom_primitive *) lines);
	}
    }

    if (do_colors)
    {
      kfree(c);
    }
    kfree(v);
}

int
varray_transpose(vertex_3d *s,
		 vertex_3d **d,
		 int usize,
		 int vsize)
{
    vertex_3d *t;
    
    t = (vertex_3d *)kmalloc(sizeof(vertex_3d)*usize*vsize);

    /* build an array of vertices in which the u & v columns are transposed */
    {
	int u,v;
	int o_offset=0,i_offset;
	
	for (u=0;u<usize;u++)
	{
	    i_offset = u;
	    for (v=0;v<vsize;v++,i_offset+=usize)
		t[o_offset++] = s[i_offset];
	}
    }
    *d = t;
    return(CHILL);
}
