/*
 * Khoros: #RCSID#
 */
 
#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: #RCSID#";
#endif

/*
 * Copyright (C) 1996, Regents of the University of California.
 * See $GEOMETRY/misc/license/copyright.txt for full copyright verbiage.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>            File Title
   >>>>
   >>>>  Static:
   >>>>             _static_routines()
   >>>>  Public:
   >>>>             public_routines()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "ghogs.h"

typedef struct plane
{
    vertex_3d **v;		/* vertex info */
    vertex_3d **d;		/* data info */
    double **m;			/* vector magnitude (computed) */
    int usize,vsize;		/* size of plane */
} plane;

static int deinit_output_buffers PROTO((void));
static int init_output_buffers PROTO((clui_info_struct *));
static void load_vector PROTO((vertex_3d *,int,kgeom_object *));
static void load_triangles PROTO((vertex_3d *,vertex_3d *,int n,kgeom_object *));
static void flush_vectors PROTO((int, kgeom_object *));
static void flush_triangles PROTO((int, kgeom_object *));
static void final_flush PROTO((
   void (*hogsfunc) PROTO((plane *, int, clui_info_struct *, kgeom_object *)),
   kgeom_object *));

static int set_up_output_object PROTO((kgeom_object *));
static int malloc_plane PROTO((plane **,int,int));
static int dealloc_plane PROTO((plane *));
static int load_plane PROTO((kobject,plane *,int));
static int massage_plane PROTO((plane *,int, clui_info_struct *));


static void footballs PROTO((plane *,int, clui_info_struct *,kgeom_object *));
static void cones PROTO((plane *,int, clui_info_struct *,kgeom_object *));
static void barbs PROTO((plane *,int, clui_info_struct *,kgeom_object *));
static void arrows PROTO((plane *, int,clui_info_struct *,kgeom_object *));

static void build_unit_circle PROTO((int s));
static int orient_and_translate_circle PROTO((vertex_3d *,vertex_3d *,vertex_3d *,int,float *,float *,float *));
static int build_orientation_matrix PROTO((vertex_3d *tp,double *m11,double *m12,double *m13,double *m31,double *m32,double *m33));

#define MAXSIDES 64

/*-----------------------------------------------------------
|
|  Routine Name: lghogs() - create geometric icons from vector data.
|
|       Purpose: To create geometric icons at each grid location of
|                a vector field.  
|
|         Input: input kobject
|                clui_info - contains various parms.
|
|        Output: output kobject
|                argument5 - explanation
|       Returns: CHILL or WHACKED.
|
|    Written By: Wes Bethel
|          Date: Fri Aug 16 13:15:21 PDT 1996
| Modifications:
|
------------------------------------------------------------*/

int
lghogs(kobject input,
       kgeom_object *output,
       clui_info_struct *ci)
{
    plane *p;
    int w,h,d,t,e,n;
    int iw;
    static void (*hogsfunc) PROTO((plane *, int, clui_info_struct *, 
			   	   kgeom_object *));
    
    /* sanity-check the input data. */
    kpds_get_attribute(input,KPDS_VALUE_SIZE,&w,&h,&d,&t,&e);
    kpds_set_attribute(input,KPDS_VALUE_DATA_TYPE,KFLOAT);
    kpds_set_attribute(input,KPDS_LOCATION_DATA_TYPE,KFLOAT);

    if ((e != 2) && (e != 3))
    {
	kerror("I AM","CONFUSED"," This kroutine wants either 2- or 3-element input data.  You gave me data with e=%d.",e);
	return(WHACKED);
    }

    /* coerce presentation into a known form.. */
    kpds_set_attribute(input,KPDS_VALUE_SIZE,w,h,d,t,3); /* set to 3-vector */
    kpds_set_attribute(input,KPDS_LOCATION_SIZE,-1,-1,-1,3); /* and 3 space */
    kpds_set_attribute(input,KPDS_ELEMENTS_FIRST,TRUE);
    
    set_up_output_object(output);
    init_output_buffers(ci);

    malloc_plane(&p,w,h);

    /* set up a function pointer to do the right thing. */
    switch (ci->icon_choice_list)
    {
    case 1:
	hogsfunc = barbs;
	break;
    case 2:
	hogsfunc = arrows;
	break;
    case 3:
	hogsfunc = cones;
	break;
    case 4:
	hogsfunc = footballs;
	break;
    }

    n = w*h;
    
    /* loop through each plane of data. */
    for (iw=0;iw<d;iw++)
    {
	load_plane(input,p,iw);
	/* do a step where the vectors are normalized, magnitudes
	 recomputed based upon linear/log, etc. */
	massage_plane(p,n,ci);

	/* call the routine that builds all the icons for this slice */
	(*hogsfunc)(p,n,ci,output);
    }

    final_flush(hogsfunc,output);
    
    dealloc_plane(p);
    deinit_output_buffers();

    return(CHILL);
}


static int
set_up_output_object(kgeom_object *o)
{
    /* set up the output geometry object */
   o->name = kstrdup("hogs");
   return(CHILL);
}


static int
malloc_plane(plane **p,
	     int usize,
	     int vsize)
{
    plane *t;
    vertex_3d *tv,*t2;
    double *d,*d2;
    int i;

    t = (plane *)kmalloc(sizeof(plane));

    t2 = tv = (vertex_3d *)kmalloc(sizeof(vertex_3d)*usize*vsize);
    t->v = (vertex_3d **)kmalloc(sizeof(vertex_3d *)*vsize);
    for (i=0;i<vsize;i++,t2+=usize)
	t->v[i] = t2;
    
    t2 = tv = (vertex_3d *)kmalloc(sizeof(vertex_3d)*usize*vsize);
    t->d = (vertex_3d **)kmalloc(sizeof(vertex_3d *)*vsize);
    for (i=0;i<vsize;i++,t2+=usize)
	t->d[i] = t2;

    d2 = d = (double *)kmalloc(sizeof(double)*usize*vsize);
    t->m = (double **)kmalloc(sizeof(double *)*vsize);
    for (i=0;i<vsize;i++,d2+=usize)
	t->m[i] = d2;

    t->usize = usize;
    t->vsize = vsize;

    *p = t;
    return(CHILL);
}

static int
dealloc_plane(plane *p)
{
    kfree(p->v[0]);
    kfree(p->v);
    kfree(p->d[0]);
    kfree(p->d);
    kfree(p);
    return(CHILL);
}


static int
load_plane(kobject input,
	   plane *p,
	   int iw)
{
    /* this code is flawed because there's no notion of time. */

    /* grab the location data. */
    
    kpds_set_attribute(input,KPDS_LOCATION_POSITION,0,0,iw,0);

    /* we've already asked for 3-element data, so hard-coding here is safe. */
    kpds_set_attribute(input,KPDS_LOCATION_REGION_SIZE,p->usize,
		       p->vsize,1,3);
    p->v[0] = (vertex_3d *)kpds_get_data(input,KPDS_LOCATION_REGION,p->v[0]);
    
    /* grab the value data. */
    
    kpds_set_attribute(input,KPDS_VALUE_POSITION,0,0,iw,0,0);
    /* we've already asked for 3-element data, so hard-coding here is safe. */
    kpds_set_attribute(input,KPDS_VALUE_REGION_SIZE,p->usize,
		       p->vsize,1,1,3);

    p->d[0] = (vertex_3d *)kpds_get_data(input,KPDS_VALUE_REGION,p->d[0]);
    return(CHILL);

}

static void
barbs(plane *p,
      int n,
      clui_info_struct *ci,
      kgeom_object *o)
{
    register double *rmag;
    register vertex_3d *rv,*rd;
    vertex_3d barb[2];
    int i;

    rv = p->v[0];
    rd = p->d[0];
    rmag = p->m[0];
    
    for (i=0;i<n;i++,rv++,rd++,rmag++)
    {
	if (*rmag != 0.)
	{
	    /* load one endpoint */
	    kmemcpy(barb,rv,sizeof(vertex_3d));

	    /* the second enpoint is computed */
	    barb[1].v[0] = barb[0].v[0] + rd->v[0] * *rmag;
	    barb[1].v[1] = barb[0].v[1] + rd->v[1] * *rmag;
	    barb[1].v[2] = barb[0].v[2] + rd->v[2] * *rmag;

	    load_vector(barb,2,o);
	}
    }
}

#define ARROW_NSIDES 4
#define ARROW_NSIDES2 8

static void
arrows(plane *p,
       int n,
       clui_info_struct *ci,
       kgeom_object *o)
{
    int nsides;
    double penum_half_width = 45.0;
    double projection_length = 0.125; /* tweak this to make the arrowheads
				       proportionally larger or smaller. */
    float fval,fval2,rscale;

    vertex_3d work[ARROW_NSIDES];
    vertex_3d work2[ARROW_NSIDES2];
    
    register double *rmag;
    register vertex_3d *rv,*rd;
    vertex_3d barb[2];
    int i,j;

    penum_half_width = DEGREES_TO_RADIANS(penum_half_width);
    penum_half_width = ksin(penum_half_width) * projection_length;
    nsides = ARROW_NSIDES;
    
    rv = p->v[0];
    rd = p->d[0];
    rmag = p->m[0];
    
    build_unit_circle(ARROW_NSIDES);

    for (i=0;i<n;i++,rv++,rd++,rmag++)
    {
	if (*rmag != 0.)
	{
	    fval2 = projection_length;
	    fval = *rmag;
	    rscale = fval * penum_half_width;
	    
	    orient_and_translate_circle(rv,rd,work,ARROW_NSIDES,&fval,&fval2,&rscale);

	    /* same as the barb code...load up the arrow shaft */
	    kmemcpy(barb,rv,sizeof(vertex_3d));

	    /* the second enpoint is computed */
	    barb[1].v[0] = barb[0].v[0] + rd->v[0] * *rmag;
	    barb[1].v[1] = barb[0].v[1] + rd->v[1] * *rmag;
	    barb[1].v[2] = barb[0].v[2] + rd->v[2] * *rmag;
	    load_vector(barb,2,o);

	    /* ok, now assemble the arrowhead and dispatch off.. */
	    for (j=0;j<ARROW_NSIDES2;j++) 
	    {
		if (j & 1)
		    kmemcpy(work2+j,barb+1,sizeof(vertex_3d));
		else
		    kmemcpy(work2+j,work+(j>>1),sizeof(vertex_3d));
	    }
	    load_vector(work2,ARROW_NSIDES2,o); 
	}
    }
}

static void
cones(plane *p,
      int n,			/* how many grid points */
      clui_info_struct *ci,
      kgeom_object *o)
{
    register double *rmag;
    register vertex_3d *rv,*rd;
    
    double penum_half_width = 15.0; /* for cones w/ 30 degree heads. */
    double projection_length = 1.0; /* tweak this to make the arrowheads
				       proportionally larger or smaller. */
    int cone_nsides;
    float fval,fval2,rscale;

    vertex_3d *work;
    vertex_3d *vlist,*nlist;
    
    vertex_3d conehead;
    int i,j,offset;

    cone_nsides = 2 << (ci->icon_smooth_int); /* will produce cones with
						 4, 8, 16, etc. sides*/

    /* malloc space for the computed base of the cone. */
    work = (vertex_3d *)kmalloc(sizeof(vertex_3d)*cone_nsides);
    /* malloc space for disjoint triangle vertices and normals */
    vlist = (vertex_3d *)kmalloc(sizeof(vertex_3d)*cone_nsides*3);
    nlist = (vertex_3d *)kmalloc(sizeof(vertex_3d)*cone_nsides*3);

    penum_half_width = DEGREES_TO_RADIANS(penum_half_width);
    penum_half_width = ksin(penum_half_width) * projection_length;
    
    rv = p->v[0];
    rd = p->d[0];
    rmag = p->m[0];
    
    build_unit_circle(cone_nsides);

    for (i=0;i<n;i++,rv++,rd++,rmag++)
    {
	if (*rmag != 0.)
	{
	    fval2 = projection_length;
	    fval = *rmag;
	    rscale = fval * penum_half_width;
	    
	    orient_and_translate_circle(rv,rd,work,cone_nsides,&fval,&fval2,&rscale);

	    conehead.v[0] = rv->v[0] + rd->v[0] * *rmag;
	    conehead.v[1] = rv->v[1] + rd->v[1] * *rmag;
	    conehead.v[2] = rv->v[2] + rd->v[2] * *rmag;

	    offset=0;
	    for (j=0;j<cone_nsides;j++)
	    {
		kmemcpy(vlist+offset,&conehead,sizeof(vertex_3d));
		offset++;
		kmemcpy(vlist+offset,work+((j+1) % cone_nsides),sizeof(vertex_3d));
		offset++;
		kmemcpy(vlist+offset,work+j,sizeof(vertex_3d));
		offset++;

		/* the normal for the vertices at the base of the
		 cone is just the difference between the computed
		 unit&oriented circle and the original grid point. */
		npoint_diff((float *)(vlist+(j*3+1)),
			    (float *)(rv),
			    (float *)(nlist+(j*3+1)),1);
		
		npoint_diff((float *)(vlist+(j*3+2)),
			    (float *)(rv),
			    (float *)(nlist+(j*3+2)),1);

		/* the normal at the conehead is the average of the
		 previous 2 normals. */
		nlist[j*3].v[0] = 0.5 * (nlist[j*3+1].v[0] +
					   nlist[j*3+2].v[0]);
		nlist[j*3].v[1] = 0.5 * (nlist[j*3+1].v[1] +
					   nlist[j*3+2].v[1]);
		nlist[j*3].v[2] = 0.5 * (nlist[j*3+1].v[2] +
					   nlist[j*3+2].v[2]);
	    }

	    /* unitize the normals */
	    nnormal_unit(nlist,nlist,cone_nsides*3);
	    
	    load_triangles(vlist,nlist,cone_nsides*3,o);
	}
    }
    kfree(work);
}

static void
footballs(plane *p,
	  int n,
	  clui_info_struct *ci,
	  kgeom_object *o)
{
}

static double
linear_scale_func(double d,float s)
{
    return(d*s);
}

static double
log_scale_func(double d,float s)
{
    d += 1.;			/* so that log10 is positive */
    d = klog10(d);
    return(d*s);
}

static int
massage_plane(plane *p,
	      int n,
	      clui_info_struct *ci)
{
    register float acc,v0,v1,v2;
    register double mag;
    register vertex_3d *rd;
    register double d,di,*rm;
    register float scale;
    int i,j;
    static double (*scalefunc)(double,float);

    scale = ci->icon_scale_float;
    if (ci->linear_log_list == 1) /* linear */
	scalefunc = linear_scale_func;
    else
	scalefunc = log_scale_func;

    rd = p->d[0];
    rm = p->m[0];
    
    for (i=0;i<n;i++,rd++,rm++)
    {
	v0 = rd->v[0];
	v1 = rd->v[1];
	v2 = rd->v[2];

	d = v0*v0 + v1*v1 + v2*v2;
	if (d != 0.)
	{
	    /* compute length */
	    d = sqrt(d);
	    di = 1./d;

	    /* normalize the vector */
	    v0 *= di;
	    v1 *= di;
	    v2 *= di;

	    rd->v[0] = v0;
	    rd->v[1] = v1;
	    rd->v[2] = v2;

	    /* scale d  - right now we'll just assume linear */

/*	    d *= scale; */
	    d = (*scalefunc)(d,scale);
	}
	*rm = d;
    }
    return(CHILL);
}
	      
#define NVERTS_PER_BUFFER 8192
static vertex_3d *vbuffer;
static vertex_3d *nbuffer;
static int nverts=0;
static int nprims=1;

static int
init_output_buffers(clui_info_struct *ci)
{
    vbuffer = (vertex_3d *)kmalloc(sizeof(vertex_3d)*NVERTS_PER_BUFFER);
    nbuffer = (vertex_3d *)kmalloc(sizeof(vertex_3d)*NVERTS_PER_BUFFER);
    /* could do some error checking here.. */
    return(CHILL);
}

static int
deinit_output_buffers()
{
    kfree(vbuffer);
    kfree(nbuffer);
    return(CHILL);
}

static void
load_triangles(vertex_3d *b,
	       vertex_3d *normals,
	       int n,		/* total # of vertices */
	       kgeom_object *o)
{
    if (nverts+n >= NVERTS_PER_BUFFER)
    {
	flush_triangles(nverts,o);
	nverts=0;
    }
    kmemcpy(vbuffer+nverts,b,n*sizeof(vertex_3d));
    kmemcpy(nbuffer+nverts,normals,n*sizeof(vertex_3d));
    nverts += n;
}

static void
load_vector(vertex_3d *b,
	    int n,
	    kgeom_object *o)
{
    if (nverts+n >= NVERTS_PER_BUFFER)
    {
	flush_vectors(nverts,o);
	nverts = 0;
    }
    kmemcpy(vbuffer+nverts,b,n*sizeof(vertex_3d));
    nverts += n;
}

static void
flush_vectors(int n,
	      kgeom_object *o)
{
    kgeom_polyline p;
   
    p.type     = KGEOM_POLYLINE_DISJOINT;
    p.nverts   = n;
    p.locs     = (float *)vbuffer;
    p.cols     = NULL;
    kgeom_write_primitive(o, (kgeom_primitive *) &p);
}

static void
flush_triangles(int n,
		kgeom_object *o)
{
    kgeom_triangles p;
   
    p.type = KGEOM_TRIANGLES_CONNECTED;
    p.nverts = n;
    p.locs  = (float *) vbuffer;
    p.norms = (float *) nbuffer;
    p.cols     = NULL;
    p.tcoords  = NULL;
    p.texture  = NULL;
    kgeom_write_primitive(o, (kgeom_primitive *)&p);
}

static void
final_flush(void (*hogsfunc) PROTO((plane *, int, clui_info_struct *, 
				    kgeom_object *)), kgeom_object *output)
{
    if (nverts == 0)
	return;
    
    if ((hogsfunc == barbs) || (hogsfunc == arrows))
	flush_vectors(nverts,output);
    else
	flush_triangles(nverts,output);
}

static float ctable[128],stable[128];
#define PI 3.1415927

static void
build_unit_circle(int s)
{
    int i,n;
    double t,dt,t2;
    double dscale;

/*    n = 2<<(s-1);*/
    n=s;

/*    dscale = *scale; */
    dscale = 1.;

    dt = (2. * PI) / n;
    t = 0.;
    for (i=0;i<n;i++,t+=dt)
    {
	t2 = dscale * kcos(t);
	ctable[i] = t2;
	t2 = dscale * ksin(t);
	stable[i] = t2;
    }
}

static int
orient_and_translate_circle(vertex_3d *base, /* grid point */
			    vertex_3d *vect, /* vector data */
			    vertex_3d *dest, /* resultant array */
			    int nsides,	/* of size 'nsides' */
			    float *lscale, /* vector magnitude */
			    float *proj_val,
			    float *rscale) /* radius scale */
{
    register float rrscale;
    vertex_3d translate_vect;
    double m11,m12,m32,m13,m33,m31,scale;
    int i;

    scale = *lscale * (1.- *proj_val);

    rrscale = *rscale;

    translate_vect.v[0] = base->v[0] + vect->v[0] * scale;
    translate_vect.v[1] = base->v[1] + vect->v[1] * scale;
    translate_vect.v[2] = base->v[2] + vect->v[2] * scale;
    
    /* build the transformation coefficients.  see speech in tubeness module.*/

    build_orientation_matrix(vect,&m11,&m12,&m13,&m31,&m32,&m33);

    m11 *= rrscale;
    m12 *= rrscale;
    m13 *= rrscale;
    
    m31 *= rrscale;
    m32 *= rrscale;
    m33 *= rrscale;
    
    /* transform all the points on the scaled unit circle to their
       new location. */
    
    for (i=0;i<nsides;i++,dest++)
    {
	dest->v[0] = (ctable[i] * m11) + (stable[i] * m31) +
	    translate_vect.v[0];
	dest->v[1] = (ctable[i] * m12) + (stable[i] * m32) +
	    translate_vect.v[1];
	dest->v[2] = (ctable[i] * m13) + (stable[i] * m33) +
	    translate_vect.v[2];
    }
    return(CHILL);
}


static int
build_orientation_matrix(vertex_3d *tp,
			 double *m11,
			 double *m12,
			 double *m13,
			 double *m31,
			 double *m32,
			 double *m33)
{
    double a2c2;
    
    a2c2 = tp->v[0]*tp->v[0] + tp->v[2]*tp->v[2];
    if (a2c2 == 0.)  /* a degenerate condition; make an identity matrix */
    {
	*m11 = 1.;
	*m12 = 0.;
	*m13 = 0.;

	*m31 = 0.;
	*m32 = 0.;
	*m33 = 1.;
	
    }
    else
    {
	a2c2 = 1./a2c2;
	
	*m11 = tp->v[0]*tp->v[0]*tp->v[1] + tp->v[2]*tp->v[2];
	*m11 *= a2c2;

	*m12 = -1. * tp->v[0];
	
	*m13 = tp->v[0] * tp->v[1] * tp->v[2] - tp->v[0] * tp->v[2];
	*m13 *= a2c2;

	*m31 = *m13;

	*m32 = -1. * tp->v[2];

	*m33 = tp->v[1] * tp->v[2] * tp->v[2] + tp->v[0] * tp->v[0];
	*m33 *= a2c2;
    }
    return(CHILL);
}

