
/*
 * Copyright (c) 1996,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */
 
/*
 * Khoros: $Id: otree.c,v 1.2.2.2 1997/05/02 15:37:09 mirage Exp $
 */
 
#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: otree.c,v 1.2.2.2 1997/05/02 15:37:09 mirage Exp $";
#endif

#include "gstreamlines.h"
#include "otree.h"

static void compute_col_extents PROTO((col_extents *p,int i,int j,int ksize,int xsize,float *xptr,float *yptr,float *zptr));

static void union_min_extents PROTO((vertex *d,vertex *s));
static void union_max_extents PROTO((vertex *d,vertex *s));
static void init_hit_stack PROTO(());
static void get_cell_coord_min PROTO((int ival,int jval,int kval,float *ptr,float *val));

static void get_cell_coord_max PROTO((int ival,int jval,int kval,float *ptr, float *zval));

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/**
  * this file provides support for solving the problem of having
  * an (x,y,z) point, and needing to know where in a grid you are.
  *
  * we are using basically an octree type structure, based upon the
  * work of Globus, et. al, Nasa Ames 1990
  *
  * this work makes a fundamental assumption for the sake of some
  * optimization.  we assume a compuational grid which is 3 dimensional,
  * that is, has (i,j,k) indices.  we further assume that only the
  * z-coordinates along the k-dimension vary at any given (i,j).  in
  * other words, for any (i,j,k=1,n), the x,y coordinates remain
  * constant.
  *
  * based upon this assumption, we can construct a quad-tree which represents
  * the x,y extents of each column of cells.  at most, we'll have to search
  * 4 different columns of cells.
  *
  * note that we will have to construct a tree which is log (base 2) n
  * deep, where n is the maximum of (i,j).
  *
  * init_otree is called to build the x-y quad tree, and to compute
  * the extents of each column of cells.  in addition, all non-leaf nodes
  * of the tree will contain min/max extents values for all children nodes.
**/

static grid_info *xg,*yg,*zg;
static int table_size;
static FILE *fdump;
static col_extents *extents_array=NULL;
static int r_xsize=0,r_xysize=0;
static qtree *malloc_qtree PROTO((void));
static float *s_xptr,*s_yptr,*s_zptr;

/* a "hit stack" is used to accumulate viable "hits" when performing
 a search.  the number of maximum hits that can occur is a function of
 how much the cell-columns overlap in the x-y plane. this number may
 need to be adjusted for a particular data set. */

#define MAX_HITS 32
static hit_struct hit_stack[MAX_HITS];
static int hit_stack_ptr=0;

static int build_qtree PROTO((qtree **q,int imin,int imax,int jmin,int jmax,int kmin,int kmax,col_extents *ex_array,int ex_xsize));

int
init_otree(streamlines_struct *ss,
	   qtree *tree,
	   float **coords_arrays)
{
    /* build the original oct tree. */
    vertex bmin,bmax;
    int npts,xysize;
    int istart,iend,jstart,jend,kstart,kend;
    int i,j;
    qtree *t;
    float *xptr,*yptr,*zptr;

    r_xsize = ss->ijkdims[0];
    r_xysize = xysize = ss->ijkdims[0] * ss->ijkdims[1];
    
    npts = xysize * ss->ijkdims[2];

    if (extents_array != NULL)
	free((char *)extents_array);
    extents_array  = (col_extents *)kmalloc(sizeof(col_extents)*xysize);
    if (extents_array == NULL)
    {
	kerror("GEOMETRY","gstreamlines", "couldn't malloc space for spacetree.");
	return(0);
    }

    /* grab ALL of the location data from the vector data.  unfortunately,
     this is necessary. */

    coords_arrays[0] = s_xptr = xptr = (float *)kmalloc(sizeof(float)*npts*3);
    coords_arrays[1] = s_yptr = yptr = xptr + npts;
    coords_arrays[2] = s_zptr = zptr = yptr + npts;

    kpds_set_attribute(ss->vectordata,KPDS_LOCATION_POSITION,0,0,0,0);
    kpds_get_data(ss->vectordata,KPDS_LOCATION_VOLUME,xptr);

    kpds_set_attribute(ss->vectordata,KPDS_LOCATION_POSITION,0,0,0,1);
    kpds_get_data(ss->vectordata,KPDS_LOCATION_VOLUME,yptr);
    
    kpds_set_attribute(ss->vectordata,KPDS_LOCATION_POSITION,0,0,0,2);
    kpds_get_data(ss->vectordata,KPDS_LOCATION_VOLUME,zptr);

    for (j=0;j<(ss->ijkdims[1])-1;j++)
    {
	for (i=0;i<(ss->ijkdims[0])-1;i++)
	{
	    /* compute extents of this column of cells. */
	    compute_col_extents(extents_array,i,j,ss->ijkdims[2],
				r_xsize,xptr,yptr,zptr);
	}
    }
#ifdef DEBUG
    fprintf(stderr," finished building col_extents array.\n");
#endif

    tree->imin = 0;
    tree->imax = ss->ijkdims[0]-1;
/*    tree->imax = v_fld->dimensions[0]-1; */
    tree->jmin = 0;
/*    tree->jmax = v_fld->dimensions[1]-1; */
    tree->jmax = ss->ijkdims[1]-1;

    
    tree->kmin = 0;
/*    tree->kmax = v_fld->dimensions[2]-1; */
    tree->kmax = ss->ijkdims[2]-1;


    tree->q00 = tree->q01 = tree->q10 = tree->q11 = NULL;
    tree->leaf_node = 0;

#ifdef DEBUG
    fdump = fopen("/tmp/fdump","w");
#endif

    {
	int i0,imid,i1,j0,jmid,j1;

	/* NOTE that we are making the assumption here that the grid size
	   at this point is at least >= 2 in each of I and J. */
	
	i0 = 0;
/*	i1 = v_fld->dimensions[0]-2; */
	i1 = ss->ijkdims[0] - 2;
	imid = (i1 >> 1);
	j0 = 0;
/*	j1 = v_fld->dimensions[1]-2; */
	j1 = ss->ijkdims[1] - 2;
	jmid = (j1 >> 1);

	/* build upper left subtree */
	build_qtree(&(tree->q00),i0,imid,j0,jmid,tree->kmin,
		    tree->kmax,extents_array,ss->ijkdims[0]-1);
		    /* v_fld->dmensions[0]-1); */
	
	union_min_extents(&(tree->min),&(tree->q00->min));
	union_max_extents(&(tree->max),&(tree->q00->max));

	/* build upper right subtree */
	build_qtree(&(tree->q10),imid+1,i1,j0,jmid,tree->kmin,
		    tree->kmax,extents_array,ss->ijkdims[0]-1);
		    /* v_fld->dimensions[0]-1); */
	
	union_min_extents(&(tree->min),&(tree->q10->min));
	union_max_extents(&(tree->max),&(tree->q10->max));

	/* build lower left subtree */
	build_qtree(&(tree->q01),i0,imid,jmid+1,j1,tree->kmin,
		    tree->kmax,extents_array,ss->ijkdims[0]-1);
		    /* v_fld->dimensions[0]-1); */
	
	union_min_extents(&(tree->min),&(tree->q01->min));
	union_max_extents(&(tree->max),&(tree->q01->max));

	/* build lower right subtree */
	build_qtree(&(tree->q11),imid+1,i1,jmid+1,j1,tree->kmin,
		    tree->kmax,extents_array,ss->ijkdims[0]-1);
		    /* v_fld->dimensions[0]-1); */
	
	union_min_extents(&(tree->min),&(tree->q11->min));
	union_max_extents(&(tree->max),&(tree->q11->max)); 
    }
#ifdef DEBUG
    fprintf(stderr," finished building qtree. \n");
    print_qtree(tree,fdump,0);
    fclose(fdump);
#endif

    
}

static int
build_qtree(qtree **q,
	    int imin,
	    int imax,
	    int jmin,
	    int jmax,
	    int kmin,
	    int kmax,
	    col_extents *ex_array,
	    int ex_xsize)
{
    int i0,i1,j0,j1,k0,k1;
    int imid,jmid;
    int no_i_subdivisions,no_j_subdivisions;
    qtree *t;
    vertex tmin,tmax;
    col_extents *p;

    no_i_subdivisions = (imin == imax) ? 1 : 0;
    no_j_subdivisions = (jmin == jmax) ? 1 : 0;

    t = malloc_qtree();
    
    *q = t;
    t->imin = imin;
    t->imax = imax;
    t->jmin = jmin;
    t->jmax = jmax;
    t->kmin = kmin;
    t->kmax = kmax;
    
    if (imin > imax)
	fprintf(stderr,"fatal error: imin > imax.\n");
    if (jmin > jmax)
	fprintf(stderr,"fatal error: jmin > jmax. \n");

    if (no_i_subdivisions && no_j_subdivisions) /* termination condition */
    {
	/* just copy over extents info to the leaf node.  no combining
	 required at this point. */
	
	p = ex_array + jmin*ex_xsize + imin;
	memcpy((char *)&(t->min),(char *)&(p->cmin),sizeof(vertex));
	memcpy((char *)&(t->max),(char *)&(p->cmax),sizeof(vertex));
	t->leaf_node = 1;
	return(CHILL);
    }

    t->leaf_node = 0;
    imid = ((imax - imin) >> 1);
    jmid = ((jmax - jmin) >> 1);

    if ((no_i_subdivisions == 0) && (no_j_subdivisions == 1))
    {
	/* do left half */
	i0 = imin;
	i1 = imin + imid;
	build_qtree(&(t->q00),i0,i1,jmin,jmax,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q00->min));
	union_max_extents(&(t->max),&(t->q00->max));
	
	/* do right half */
	i0 = imin + imid + 1;
	i0 = (i0 <= imax) ? i0 : imax;
	i1 = imax;
	build_qtree(&(t->q10),i0,i1,jmin,jmax,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q10->min));
	union_max_extents(&(t->max),&(t->q10->max));
	
	t->q11 = t->q01 = NULL;
    }
    else if ((no_i_subdivisions == 1) && (no_j_subdivisions == 0))
    {
	/* do upper half */
	j0 = jmin;
	j1 = jmin+jmid;
	build_qtree(&(t->q01),imin,imax,j0,j1,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q01->min));
	union_max_extents(&(t->max),&(t->q01->max));
	
	
	/* do lower half */
	j0 = jmin+jmid+1;
	j0 = (j0 <= jmax) ? j0 : jmax;
	j1 = jmax;
	build_qtree(&(t->q11),imin,imax,j0,j1,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q11->min));
	union_max_extents(&(t->max),&(t->q11->max));
	t->q00 = t->q10 = NULL;
    }
    else
    {
	/* upper left */
	i0 = imin;
	i1 = imin + imid;
	j0 = jmin;
	j1 = jmin+jmid;
	build_qtree(&(t->q00),i0,i1,j0,j1,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q00->min));
	union_max_extents(&(t->max),&(t->q00->max));

	/* upper right */
	i0 = imin+imid+1;
	i0 = (i0 <= imax) ? i0 : imax;
	i1 = imax;
	build_qtree(&(t->q10),i0,i1,j0,j1,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q10->min));
	union_max_extents(&(t->max),&(t->q10->max));

	/* lower left */
	j0 = jmin+jmid+1;
	j0 = (j0 <= jmax) ? j0 : jmax;
	j1 = jmax;
	i0 = imin;
	i1 = imin+imid;
	build_qtree(&(t->q01),i0,i1,j0,j1,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q01->min));
	union_max_extents(&(t->max),&(t->q01->max));
	
	/* lower right */
	i0 = imin+imid+1;
	i0 = (i0 <= imax) ? i0 : imax;
	i1 = imax;
	build_qtree(&(t->q11),i0,i1,j0,j1,kmin,kmax,ex_array,ex_xsize);
	union_min_extents(&(t->min),&(t->q11->min));
	union_max_extents(&(t->max),&(t->q11->max));
    }
}

#define MAXMATCHES 128

int
find_grid_ijk(vertex *v, int *ijk, qtree *tree)
{
    /* the purpose of this routine is to find the i,j,k index in the
     pile o' cells that contains the point *vertex. */
    int status;

    init_hit_stack();

    /* the first search is a "rough-cut" which finds colums of cells which
     contain the point in question.  later, we will examine these columns
     of cells more closely to determine exactly which cell contains the
     point. */
    status = qtree_find_point(v,tree);
    if (status == 0)
    {
	ijk[0] = ijk[1] = -1;	/* don't care about ijk[2] */
	return(0);
    }

#ifdef DEBUG
    fprintf(stderr,"\t found %d columns which contain this point.\n",hit_stack_ptr);
#endif
    
    /* need to look at each column of cells to find the cell which
     contains the point. */

    status = look_at_columns(v,s_xptr,s_yptr,s_zptr,ijk);
    if (status == 0)
    {
	ijk[2] = -1;
	return(0);
    }
    return(1);
}


static void
compute_col_extents(col_extents *p,
		    int i,
		    int j,
		    int ksize,
		    int xsize,
		    float *xptr,
		    float *yptr,
		    float *zptr)
{
    /* in this khoros implementation, we are computing the space
     bounded by adjacent lattice points. thus, there is some redundancy
     built into the loops below.  this redundancy will not exist if we're
     working with strictly-cell based data (which we're not, here). */
    
    int k;
    int offset;
    float *t;

    /* xsize-1 because we're computing a pointer into an array that
     is one less in size than the AVS field array in each dimension - this
     won't be the case if we're looking directly at a cell's points array,
     rather than computing cells based upon a lattice. */
    
    p = p + j*(xsize-1) + i;

    p->cmin.x = p->cmin.y = p->cmin.z = BIGNUM ;
    p->cmax.x = p->cmax.y = p->cmax.z = SMALLNUM ;

    /* use xsize because we're indexing into the lattice */
    t = xptr + i + j*xsize;
    for (k=0;k<ksize;k++,t += r_xysize)
    {
	/* current x point less than xmin? */
	p->cmin.x = MIN(*t,p->cmin.x);
	p->cmax.x = MAX(*t,p->cmax.x);
	
	/* xpoint+1 > xmax? */
	p->cmin.x = MIN(*(t+1),p->cmin.x);
	p->cmax.x = MAX(*(t+1),p->cmax.x);

	/* check xpoint+xsize */
	p->cmin.x = MIN(*(t+xsize),p->cmin.x);
	p->cmax.x = MAX(*(t+xsize),p->cmax.x);
	
	/* check xpoint+xsize+1 */
	p->cmin.x = MIN(*(t+xsize+1),p->cmin.x);
	p->cmax.x = MAX(*(t+xsize+1),p->cmax.x);
    }
    
    t = yptr + i + j*xsize;
    for (k=0;k<ksize;k++,t += r_xysize)
    {
	/* current y point less than ymin? */
	p->cmin.y = MIN(*t,p->cmin.y);
	p->cmax.y = MIN(*t,p->cmax.y);

	/* ypoint+1 > ymax? */
	p->cmax.y = MAX(*(t+1),p->cmax.y);
	p->cmin.y = MIN(*(t+1),p->cmin.y);

	/* ypoint+xsize */
	p->cmax.y = MAX(*(t+xsize),p->cmax.y);
	p->cmin.y = MIN(*(t+xsize),p->cmin.y);

	/* ypoint+xsize+1 */
	p->cmax.y = MAX(*(t+xsize+1),p->cmax.y);
	p->cmin.y = MIN(*(t+xsize+1),p->cmin.y);
    }
    
    t = zptr + i + j*xsize;
    for (k=0;k<ksize;k++,t += r_xysize)
    {
	/* z point < zmin? */
	p->cmin.z = MIN(*t,p->cmin.z);
	p->cmax.z = MAX(*t,p->cmax.z);

	/* check zpoint+1  */
	p->cmin.z = MIN(*(t+1),p->cmin.z);
	p->cmax.z = MAX(*(t+1),p->cmax.z);
	
	/* check zpoint+xsize  */
	p->cmin.z = MIN(*(t+xsize),p->cmin.z);
	p->cmax.z = MAX(*(t+xsize),p->cmax.z);
	
	/* check zpoint+xsize+1  */
	p->cmin.z = MIN(*(t+xsize+1),p->cmin.z);
	p->cmax.z = MAX(*(t+xsize+1),p->cmax.z);
    }
}


#ifdef DEBUG
print_qtree(t,f,level)
qtree *t;
FILE *f;
int level;
{
    int i;
    int leaf_node;

#if 0
    leaf_node = ((t->q00 == NULL) && (t->q01 == NULL) &&
		 (t->q10 == NULL) && (t->q11 == NULL));
#endif
    leaf_node = t->leaf_node;

    if (leaf_node)
	fprintf(f,"*");
    else
	fprintf(f," ");

    for (i=0;i<level;i++)	/* indent based upon tree level */
	fprintf(f,"\t");

    fprintf(f,"\t%d\t%d\t%d\t%d\n",t->imin,t->imax,t->jmin,t->jmax);
    
    for (i=0;i<level;i++)	/* indent based upon tree level */
	fprintf(f,"\t");
    fprintf(f,"\t%g %g %g\t%g %g %g\n",t->min.x,t->min.y,t->min.z,t->max.x,t->max.y,t->max.z);


    /* print subtrees */
    if (t->q00 != NULL)
	print_qtree(t->q00,f,level+1);
    
    if (t->q10 != NULL)
	print_qtree(t->q10,f,level+1);
    
    if (t->q01 != NULL)
	print_qtree(t->q01,f,level+1);
    
    if (t->q11 != NULL)
	print_qtree(t->q11,f,level+1);
    return(CHILL);
}
#endif

static void
union_min_extents(vertex *d,
		  vertex *s)
{
    d->x = MIN(d->x,s->x);
    d->y = MIN(d->y,s->y);
    d->z = MIN(d->z,s->z);
}

static void
union_max_extents(vertex *d,
		  vertex *s)
{
    d->x = MAX(d->x,s->x);
    d->y = MAX(d->y,s->y);
    d->z = MAX(d->z,s->z);
}


static qtree
*malloc_qtree(void)
{

    qtree *t;
    
    t = (qtree *)malloc(sizeof(qtree));
    t->min.x = t->min.y = t->min.z = BIGNUM;
    t->max.x = t->max.y = t->max.z = SMALLNUM;
    t->q00 = t->q10 = t->q01 = t->q11 = NULL;

    return(t);
}

int
qtree_find_point(vertex *v, qtree *tree)
{
    /* for this node (tree), 1st check to see if the point lies
     within the min/max. */

    int status;
    int status00,status01,status10,status11;

    status = is_contained(v,&(tree->min),&(tree->max));
    
    if ((tree->leaf_node) && (status)) /* this cell contains the point */
    {
	add_to_hit_stack(tree);
	return(status);
    }    
    else if (status)		/* not a leaf node, but contains the point */
    {
	status00 = status01 = status10 = status11 = 0;
	
	/* continue searching - check each of children..*/
	if (tree->q00)
	    status00 = qtree_find_point(v,tree->q00);
	if (tree->q01)
	    status01 = qtree_find_point(v,tree->q01);
	if (tree->q10)
	    status10 = qtree_find_point(v,tree->q10);
	if (tree->q11)
	    status11 = qtree_find_point(v,tree->q11);
	return(status00 || status01 || status10 || status11);
    }
    else
	return(status);
}
    
int
is_contained(vertex *v,vertex *min,vertex *max)
{
    if ((v->x >= min->x) && (v->x <= max->x) &&
	(v->y >= min->y) && (v->y <= max->y) && 
	(v->z >= min->z) && (v->z <= max->z))
	return(1);
    else
	return(0);
}

static void
init_hit_stack()
{
    hit_stack_ptr = 0;
}

int
add_to_hit_stack(qtree *t)
{
    if (hit_stack_ptr < MAX_HITS)
    {
	hit_stack[hit_stack_ptr].column_info = t;
	hit_stack_ptr++;
	return(CHILL);
    }
    else
    {
	fprintf(stderr,"\tError: Hit Stack Overflow....\n");
	return(WHACKED);
    }
}

int
look_at_columns(vertex *v,float *xptr,float *yptr,float *zptr,int *ijk)
{
    /* at this point, we're being given a vertex, V, and several
     columns (in the hit stack) which contain the point.  the task
     now is to determine which of the cells contains the point. */

    int l,kstart,kend,ival,jval,k;
    qtree *p;
    float zmin,zmax;
    float xmin,xmax;
    float ymin,ymax;

    /* loop over each column */
    for (l=0;l<hit_stack_ptr;l++)
    {
	p = hit_stack[l].column_info;
	
	kstart = p->kmin;
	kend = p->kmax;
	ival = p->imin;
	jval = p->jmin;
	
	for (k=kstart;k<kend;k++) /* examine each cell in the column */
	{
	    /* first best thing to look at is depth.  for now, we'll
	     just proceed linearly down the column of cells.  this could
	     be improved by using a binary search. */

	    get_cell_coord_min(ival,jval,k,s_zptr,&zmin);
	    get_cell_coord_max(ival,jval,k,s_zptr,&zmax);

	    if ((v->z >= zmin) && (v->z <= zmax))
	    {
		/* bracketed in Z.  now check X & Y. */
		get_cell_coord_min(ival,jval,k,s_xptr,&xmin);
		get_cell_coord_max(ival,jval,k,s_xptr,&xmax);

		if ((v->x >= xmin) && (v->x <= xmax))
		{
		    /* bracketed in X.  now check Y. */
		    get_cell_coord_min(ival,jval,k,s_yptr,&ymin);
		    get_cell_coord_max(ival,jval,k,s_yptr,&ymax);

		    if ((v->y >= ymin) && (v->y <= ymax))
		    {
			ijk[0] = ival;
			ijk[1] = jval;
			ijk[2] = k;
			return(1);
		    }
		}
	    }
	}
    }
    return(0);
}

static void
get_cell_coord_min(int ival,
		   int jval,
		   int kval,
		   float *ptr,
		   float *val)
{
    float tz,*t;
    
    tz = BIGNUM;
    t = ptr + ival + jval*r_xsize + kval*r_xysize;

    /* look at each of "cell's" points (whether they be X's, Y's or
       Z's) to find the minimum value.
     this process could be expidited at the expense of more set-up work
     and space. */
    
    tz = *t;
    tz = MIN(*(t+1),tz);
    tz = MIN(*(t + r_xsize),tz);
    tz = MIN(*(t + r_xsize+1),tz);

    t += r_xysize;
    tz = MIN(*t,tz);
    tz = MIN(*(t+1),tz);
    tz = MIN(*(t + r_xsize),tz);
    tz = MIN(*(t + r_xsize+1),tz);
    
    *val = tz;
}

static void
get_cell_coord_max(int ival,
		   int jval,
		   int kval,
		   float *ptr,
		   float *zval)
{
    float tz,*t;
    
    tz = SMALLNUM;
    t = ptr + ival + jval*r_xsize + kval*r_xysize;

    /* look at each of "cell's" z-points to find the maximum value.
     this process could be expidited at the expense of more set-up work
     and space. */
    
    tz = *t;
    tz = MAX(*(t+1),tz);
    tz = MAX(*(t + r_xsize),tz);
    tz = MAX(*(t + r_xsize+1),tz);

    t += r_xysize;
    tz = MAX(*t,tz);
    tz = MAX(*(t+1),tz);
    tz = MAX(*(t + r_xsize),tz);
    tz = MAX(*(t + r_xsize+1),tz);
    *zval = tz;
    
}

