 /*
  * Khoros: $Id: lgbindtexture.c,v 1.9.2.2 1997/05/31 18:58:34 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lgbindtexture.c,v 1.9.2.2 1997/05/31 18:58:34 mirage Exp $";
#endif

/*
 * Copyright (C) 1996,1997 , The Regents of the University of California.
 * All rights reserved.  See $GEOMETRY/repos/license/LBL_Copyright.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for gbindtexture
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lgbindtexture
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
static int check_for_texture_coords PROTO((kgeom_object *));
static int set_texture_transport    PROTO((kgeom_object *, char *));
static kgeom_primitive *find_first_texture PROTO((kgeom_object *geom));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lgbindtexture - *
* 
*       Purpose: This should be a complete description that anyone
*                could understand;  it should have acceptable grammar
*                and correct spelling.
*
*         Input: in_geom - the geometry object to which the texture
*                          will be applied
*                in_texture - (presumably an image or volume) the texture
*                          object which will provide color information in
*                          the texture mapping renderer
*                in_texture_tname - name of the transport containing the
*                          texture
*
*        Output: out_geom - copy of input object with a texture attribute
*                          set
*
*       Returns: CHILL on success, WHACKED otherwise
*
*  Restrictions: dimensionality of texture coords and input texture must agree
*    Written By: wes
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Sep 1995 - renamed to lgbindtexture for jargon-consistency
*                with GL, etc.
****************************************************************/
/* -library_def */
int
lgbindtexture(kgeom_object *geom,
	      kgeom_object *texture,
	      char *texture_name)
/* -library_def_end */

/* -library_code */
{
    /**
      * validity checking on input consists of checking the dimensionality
      * of the input texture and comparing it to the dimensionality of
      * the texture coordinates on the input object.
      *
      * for the sake of simplicity, we will assume
      *
      * if they jibe, the texture attribute will be set in the geom
      * object.  if not, an error message will be issued notifying the
      * user of the inconsistency between the dimensionality of the
      * texture and texture coordinates.
    **/

    kgeom_primitive *texture_prim;
    int tdim, any_texture_coords;

    /* this becomes more complicated now that textures can co-exist with
       other geometry..  take the first texture as "the one" */
    texture_prim = find_first_texture(texture);

    if (texture_prim == NULL)
    {
       kerror("GEOMETRY","lkapplytexture",
	      "Texture input does not contain a texture!");
       return WHACKED;
    }

    /**
      * assume that octmeshes have 3D texture coords, all other texture
      * coords are 2D.   assume that the input texture has a reasonable
      * width, height, (and depth) as enforced by the texture creation 
      * routine.  
    **/
    tdim = (texture_prim->type == KGEOM_TEXTURE2D) ? 2 : 3;

    /**
      * try to grab a texture coord to see if there's anything really there. 
     **/
    any_texture_coords = check_for_texture_coords(geom);
    
    if (!any_texture_coords)
    {
	kerror("GEOMETRY","lgbindtexture","The input geometry object contains NO texture coordinates.  I would like geometry which HAS texture coordinates. Aborting.\n");
	return WHACKED;
    }

    if (tdim != geom->texture_coord_dim)
    {
	/* bark at user. */
	kerror("GEOMETRY","lgbindtexture","There is a dimension mismatch between the input geometry and input texture.\n Input texture dims = %d, texture coordinate dims = %d.\nThese values MUST be identical.  Aborting.\n", 
	       tdim, geom->texture_coord_dim);
	return WHACKED;
    }

    /**
      * we've found an acceptable match.
      **/
    set_texture_transport(geom,texture_name); 

    return CHILL;
}

static int
check_for_texture_coords(kgeom_object *geom)
{
    kgeom_primitive *prim;
    int found_some=FALSE;
    int i,nprims;
    float *tc=NULL;

    nprims = kgeom_number_primitives(geom);
    
    for (i=0;i<nprims;i++)
    {
        prim = kgeom_get_primitive(geom, i);
	switch(prim->type)
	{
	   /* check only those object types which can have texture coords.*/
	   case KGEOM_TRIANGLES_DISJOINT:
	   case KGEOM_TRIANGLES_CONNECTED:
	      if (((kgeom_triangles *)prim)->tcoords)
		 found_some = TRUE;
	      break;
	   case KGEOM_QUADMESH:
	      if (((kgeom_quadmesh *)prim)->tcoords)
		 found_some = TRUE;
	      break;
	   case KGEOM_OCTMESH:
	      if (((kgeom_octmesh *)prim)->tcoords)
		 found_some = TRUE;
	      break;
	}
    }
    return(found_some);
}

static int
set_texture_transport(kgeom_object *geom,
		      char *texture_name)
{
    kgeom_primitive *prim;
    int i,nprims;

    nprims = kgeom_number_primitives(geom);

    for (i=0;i<nprims;i++)
    {
       prim = kgeom_get_primitive(geom, i);
       switch(prim->type)
       {
	  /* assign only to primitives which have texture coords */
	  case KGEOM_TRIANGLES_DISJOINT:
	  case KGEOM_TRIANGLES_CONNECTED:
	     if (((kgeom_triangles *)prim)->tcoords)
		((kgeom_triangles *)prim)->texture = kstrdup(texture_name);
	      break;
	  case KGEOM_QUADMESH:
	     if (((kgeom_quadmesh *)prim)->tcoords)
		((kgeom_quadmesh *)prim)->texture = kstrdup(texture_name);
	     break;
	  case KGEOM_OCTMESH:
	     if (((kgeom_octmesh *)prim)->tcoords)
		((kgeom_octmesh *)prim)->texture = kstrdup(texture_name);
	     break;
       }
    }
    return(CHILL);
}

static 
kgeom_primitive *find_first_texture(kgeom_object *geom)
{
    kgeom_primitive *prim;
    int i,nprims;

    nprims = kgeom_number_primitives(geom);

    for (i=0;i<nprims;i++)
    {
       prim = kgeom_get_primitive(geom, i);

       if (prim->type == KGEOM_TEXTURE2D ||
	   prim->type == KGEOM_TEXTURE3D)
	  return prim;
    }
    return NULL;
}
/* -library_code_end */
