 /*
  * Khoros: $Id: lgquadmesh.c,v 1.14.2.5 1997/05/31 18:58:56 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lgquadmesh.c,v 1.14.2.5 1997/05/31 18:58:56 mirage Exp $";
#endif

/*
 * Copyright (C) 1996,1997 , The Regents of the University of California.
 * All rights reserved.  See $GEOMETRY/repos/license/LBL_Copyright.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for gquadmesh
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lgquadmesh
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
#include "internals.h"


static int ok_mesh_input PROTO(( kobject ));
static int make_quad_mesh PROTO(( kobject, kgeom_object *, int *, vis_cmap *, int,int,float *,int));
static int make_tstrip_mesh PROTO(( kobject, kgeom_object *, int *, vis_cmap *, int,int,float *,int));
static void viffloc_to_geomloc PROTO((float *dst,float *iloc,int usize,int vsize));
void meshrow_to_tstrip PROTO((char *dst, char *src, int n, int size));
static int assimilate_location_data PROTO((kobject kin,float *a,int row_num,int usize));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lgquadmesh - *
* 
*       Purpose: to create a quad mesh geom object from 2D data.
*                the geom may be optionally colorized with
*                a colormap, whose presence is optional.
*
*         Input: inobj - an opened kobject containing the source data
*                cmapobj - an opened kobject containing a visualization cmap
*                name - a char string containing the name for the geom object
*
*                have_mapfile - boolean indicating if we have a colormap
*                have_tcoords - boolean indicating if we have are to
*                    generate texture coords for the quadmesh.
*
*                NOTE: the use of a colormap and generation of texture coords
*                    are mutually exclusive.
*
*                tc - array of 4 floats specifing umin,vmin,umax,vmax for
*                    texture coords.  this routine will compute texture coords
*                    such that (umin,vmin) will be at the index (0,0) in
*                    the quadmesh, and (umax,vmax) will be at index
*                    (usize-1,vsize-1) of the quadmesh.
*
*                optimization - int value used to determine what form of
*                    mesh optimization is used.  0=none, a plain old
*                    quadmesh is output; 1=t-strips, a mesh is broken up
*                    into t-strips.  more will come in the future.
*
*                do_normals - int value indicating whether or not vertex
*                    normals will be computed.
*
*        Output: outobj - an opened kobject containing the geom object
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: Restrictions on data or input as applicable
*    Written By: Yo Momma
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: do not drive heavy machinery while operating this software
*      Examples: 
* Modifications: *
*                Sep 1995
*                   new tstrip code generates a bunch of triangle strips from
*                   2d input data (w/ 3d location data).  an object is created
*                   which has v-1 triangle strips, where v is the number of
*                   rows of data.
****************************************************************/
/* -library_def */
int lgquadmesh(kobject inobj,
	       int have_mapfile,
	       kobject cmapobj,
	       kgeom_object *outobj,
	       char *name,
	       int have_tcoords,
	       float *tc,
	       int optimization,
	       int do_normals)
/* -library_def_end */

/* -library_code */
{
    vis_cmap cmap;
    int i,num_dims,dims[5];

    /** first, verify that the input object is processable **/

    /** at present, we want 2 dimensional data and 3d coords and
      will barf if we get anything else **/

    if (ok_mesh_input(inobj) == WHACKED)
	return(WHACKED);

    /* next, convert the colormap if it's present to a vis_cmap */
    if (have_mapfile)
    {
	if (lkobj_to_vismap(cmapobj,&cmap) == WHACKED)
        {
	    errno = KINVALID_DATATYPE;
	    kerror("GEOMETRY","lquadmesh","The object supplied for a vis colormap is invalid. \n");
	    return(WHACKED);
	}
    }

    /* set up the output object */
    outobj->name      = kstrdup(name);
    outobj->layout    = KPER_VERTEX;
    outobj->has_alpha = TRUE;
    
    /* build the output object */

    /* figure out if we are working with 2 or 3d data. */
    kpds_get_attribute(inobj,KPDS_VALUE_SIZE,dims,dims+1,dims+2,
		       dims+3,dims+4);
	
    /* in the following loop, we`ll scan through the sizes associated
       with each value segment dimension (EXCEPT for elements) and
       count if the value is greater than 1 */

    num_dims=0;
    for (i=0;i<4;i++)
	if (dims[i] > 1)
	    num_dims++;

    /* if num_dims==2, then we'll (try to) make a quadmesh, else if
       num_dims==3, then we'll try to make an oct mesh */

    if (num_dims == 2)
    {
	switch (optimization)
	{
	case 0:
	    make_quad_mesh(inobj,outobj,dims,&cmap,have_mapfile,
			   have_tcoords,tc,do_normals);
	    break;
	case 1:
	    make_tstrip_mesh(inobj,outobj,dims,&cmap,have_mapfile,
			     have_tcoords,tc,do_normals);
	    break;
	default:
	    kerror("GEOMETRY","lquadmesh","Optimization value of %d is not supported.\n",optimization);
	    break;
	}
    }
    else
    {
	errno = KINVALID_DATATYPE;
	kerror("GEOMETRY","lquadmesh","There must be 2 dimmensions of value data.  You have given me %d\n",num_dims);
	return(WHACKED);
    }
    return(CHILL);
}

static int
ok_mesh_input(kobject inobj)
{
    int lw,lh,ld,ldims;
    int vw,vh,vd,ve,vt;

    if (kpds_query_location(inobj))
    {
       kpds_get_attribute(inobj,KPDS_LOCATION_SIZE,&lw,&lh,&ld,&ldims);
       if (ldims != 3)
       {
	  errno = KINVALID_DATATYPE;
	  kerror("GEOMETRY","lquadmesh","The location data must have dimensions=3.  The input file has dimensions=%d.\n",ldims);
	  return(WHACKED);
       }
    }
    else
    {
      int wid,hgt,dep;
 
      /* Here's where we create a location segment for the location
          challenged dataset.  we are going to set the uniform
          location min and max to be from (0,0,0) <-> (wid,hgt,dep).
          The rest of the routine will process as if we had regular
          curvilinear location data present */
 
      kpds_get_attribute(inobj,KPDS_VALUE_SIZE,&wid,&hgt,&dep,NULL,NULL);
 
      kpds_set_attribute(inobj, KPDS_LOCATION_GRID, KUNIFORM);
      kpds_create_location(inobj);
      kpds_set_attributes(inobj,
                          KPDS_LOCATION_BEGIN, 0.,0.,0.,
                          KPDS_LOCATION_END,(double)wid,
                          (double)hgt, (double)dep, NULL);

    }

    /* set up a happy presentation for the location data */
    kpds_set_attribute(inobj, KPDS_LOCATION_DATA_TYPE, KFLOAT);

    kpds_get_attribute(inobj,KPDS_VALUE_SIZE,&vw,&vh,&vd,&vt,&ve);
    if (ve != 1)  /* it's not scalar data .. */
    {
	if (ve != 4) /* can't possibly be anything but RGBA data */
	{
	    errno = KINVALID_DATATYPE;
	    kerror("GEOMETRY","lquadmesh","This datafile has elements = %d.  Only elements=1 or elements=4 (interpreted as RGBA) are supported.");
	    return(WHACKED);
	}
    }

    /* set up a happy presentation for the value data */
    kpds_set_attribute(inobj, KPDS_VALUE_DATA_TYPE, KFLOAT);

    return(CHILL);
}

static int
make_quad_mesh(kobject kin,
	       kgeom_object *kout,
	       int *dims,
	       vis_cmap *cmap,
	       int use_map,
	       int use_tc,
	       float *tc,
	       int do_normals)
{
    int i;
    int usize,vsize;
    float *iloc=NULL,*ival=NULL,*oloc=NULL,*temp=NULL,*ocolors=NULL;
    float *tcoords=NULL;
    vertex_3d *normals=NULL;
    int vw,vh,vd,vt,ve;
    int index;
    vertex_3d bmin,bmax,center;
    kgeom_quadmesh *mesh;

    /* sep 95 - this routine could be improved by looping over rows
     of U (or something similar) and putting out data row or two at
     a time. this way, the entire quadmesh dataset wouldn't have to
     be in memory at one time.  the problem areas in the code below
     which will need some attention are: 1. computing of vertex normals,
     if they're not present; 2. computation of texture coords, if requested.
     everything else should work pretty much ok or with little modification. */

    /* sep 95 - do normals not yet implemented.  will always generate normals */

    /* assign values to usize and vsize.  from the calling routine,
       dims[] holds the size of the value data.  i'm assuming that
       there are at two and only two values in this array that are > 1.
       such an assumption eliminates a lot of boring error checking
       code, but the below code will seg fault if this is not the case */
    
    for (i=0;i<4;i++)
	if (dims[i] != 1)
	    break;
    
    usize = dims[i];
    for (i=i+1;i<4;i++)
	if (dims[i] != 1)
	    break;
    vsize = dims[i];

    /* grab the location data from the input. */

    /* at some point, we'll want to do this in a more intelligent way
       than to just malloc space for the entire shootin' match.
       maybe loop over columns or something */

    iloc = (float *)kcalloc((size_t) usize*vsize*3, sizeof(float)); 
    oloc = (float *)kcalloc((size_t) usize*vsize*3, sizeof(float));

    kpds_set_attribute(kin,KPDS_LOCATION_POSITION,0,0,0,0);
    kpds_get_data(kin,KPDS_LOCATION_ALL,iloc);

    
    /* unscramble location data from XXXYYYZZZ to XYZXYZXYZ */

    viffloc_to_geomloc(oloc,iloc,usize,vsize);

    /**
      * compute the normals for the quadmesh and put them, too.
    **/

    lcompute_quadmesh_vertex_normals((vertex_3d *)oloc,
				     &normals,
				     usize,vsize);

    /**
      *  start setting up the output mesh 
     **/
    mesh = (kgeom_quadmesh *) kgeom_new_primitive(KGEOM_QUADMESH);
    mesh->width  = usize;
    mesh->height = vsize;
    mesh->locs   = (float *)oloc;
    mesh->norms  = (float *)normals;

    /**
      * compute the bounding box and center point.
    **/

    lcompute_bbox((vertex_3d *)oloc,usize*vsize,&bmin,&bmax);
    for (i=0;i<3;i++)
	kout->center[i] = bmin.v[i] + (bmax.v[i]-bmin.v[i])*0.5;

    kmemcpy(kout->min, &(bmin.v[0]), sizeof(float)*3);
    kmemcpy(kout->max, &(bmax.v[0]), sizeof(float)*3);
    kout->has_bounding_box = TRUE;
    kout->has_center       = TRUE;
    
    /**
     * the possibilities for colorizing the output quadmesh are as
     * follows:
     * 1. use_map == TRUE, && scalar data,
     * 2. use_map == FALSE, && e==4 value data (interpreted as
     *    RGBA in [0..1] )
     * 3. use_map == FALSE, scalar data, (or no data), AND presence
     *     of texture coords.
    **/
    
    kpds_get_attribute(kin,KPDS_VALUE_SIZE,&vw,&vh,&vd,&vt,&ve);
    if (use_map)
    {
	if (ve != 1)
	{
	    errno = KINVALID_DATATYPE;
	    kerror("GEOMETRY","lquadmesh","When using a colormap with 2D data, the elements size of the value segment MUST be equal to 1.  The file you supplied has something different.");
	    return(WHACKED);
	}
	ival = (float *)kmalloc((unsigned)(sizeof(float)*usize*vsize));
	temp = ocolors = (float *)kmalloc((unsigned)(sizeof(float)*usize*vsize*4));
	kpds_set_attribute(kin,KPDS_VALUE_POSITION,0,0,0,0,0);
	kpds_get_data(kin,KPDS_VALUE_ALL,ival);

	for (i=0;i<usize*vsize;i++,temp+=4)
	{
	    index = lget_map_index(cmap,ival+i);
	    lget_map_rgba(cmap,index,temp,temp+1,temp+2,temp+3);
	}

	mesh->cols = (float *) ocolors;
    }
    else
    {
	/* don't care what kind of value data we have */
	if (use_tc == 2) /* make 2d texture coords */
	{
	    double U,dU,V,dV;
	    int iu,iv,offset=0;
	    
	    tcoords = (float *)kmalloc((unsigned)(usize*vsize*2*sizeof(float)));

	    U = tc[0];
	    dU = (tc[1]-tc[0])/(usize-1);
	    V = tc[2];
	    dV = (tc[3]-tc[2])/(vsize-1);
	    
	    for (iv = 0;iv<vsize;iv++)
	    {
		U = tc[0];
		for (iu=0;iu<usize;iu++,U+=dU)
		{
		    tcoords[offset++] = U;
		    tcoords[offset++] = V;
		}
		V += dV;
	    }
	    
	    kout->texture_coord_dim = 2;
	    mesh->tcoords = tcoords;
	}
	else /* assume use_tc == 3 */
	{
	    double U,dU,V,dV,W,dW;
	    int iu,iv,offset=0;
	    
	    tcoords = (float *)kmalloc((unsigned)(usize*vsize*3*sizeof(float)));
	    
	    U = tc[0];
	    dU = (tc[1]-tc[0])/(usize-1);
	    V = tc[2];
	    dV = (tc[3]-tc[2])/(vsize-1);
	    W = tc[4];
	    dW = (tc[5]-tc[4])/(vsize-1);

	    /**
	      * in the absence of anything better, W will be interpolated
	      * across the V dimension.
	    **/
	    for (iv = 0;iv<vsize;iv++)
	    {
		U = tc[0];
		for (iu=0;iu<usize;iu++,U+=dU)
		{
		    tcoords[offset++] = U;
		    tcoords[offset++] = V;
		    tcoords[offset++] = W;
		}
		V += dV;
		W += dW;
	    }
	    kout->texture_coord_dim = 3;
	    mesh->tcoords = tcoords;

	}
    }

    kgeom_add_primitive(kout, (kgeom_primitive *)mesh);

    return(CHILL);
}

int
lcompute_quadmesh_vertex_normals(vertex_3d *oloc,
				 vertex_3d **normals,
				 int usize,
				 int vsize)
{
    int uv,vv;
    int i,j,k;
    vertex_3d *n;
    vertex_3d s1,s2;
    vertex_3d *v;
    double mag;

    /**
      * here, we compute normals on a per-vertex basis.  if per-facet
      * normals are desired, they could be derived from those
      * computed here, or from the raw vertices directly
    **/

    uv = usize;
    vv = vsize;

    /* alloc space for one normal per vertex */
    if (*normals == NULL)
	n = (vertex_3d *)kmalloc((unsigned)(sizeof(vertex_3d)*uv*vv));
    else
        n = *normals;
    
    v = oloc;

    /* do general case first. */
    for (j=0;j<vv;j++)
    {
	for (i=0;i<uv;i++)
	{
	    if ((i != 0) && (i != uv-1) && (j != 0) && (j != vv-1))
	    {
		for (k=0;k<3;k++)
		{
		    s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
		    s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
		}
		vertex_unit(&s1,&mag);
		vertex_unit(&s2,&mag);
		vertex_cross(&s1,&s2,(n+i+j*uv));
	    }
	}
    }
    
    /* do the j=0 row. */
    j=0;
    for (i=0;i<uv;i++)
    {
	if ((i != 0) && (i != uv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+1)->v[k] - (v+i-1)->v[k];
		s2.v[k] = (v+i+uv)->v[k] - (v+i)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i));
	}
    }
    
    /* do the j=vv-1 row. */
    j=vv-1;
    for (i=0;i<uv;i++)
    {
	if ((i != 0) && (i != uv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
		s2.v[k] = (v+i+j*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i+j*uv));
	}
    }
    
    /* do the i=0 column */
    i=0;
    for (j=0;j<vv;j++)
    {
	if ((j != 0) && (j != vv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i+j*uv)->v[k];
		s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i+j*uv));
	}
    }
    
    /* do the i=uv column */
    i=uv-1;
    for (j=0;j<vv;j++)
    {
	if ((j != 0) && (j != vv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
		s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i+j*uv));
	}
    }
    /* now, do the four corners. these normals are computed as the
      cross product of the segments adjoining the 2 nearest points. */
    
    /* i==0,j==0 */
    i=0;
    j=0;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i+j*uv)->v[k];
	s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+j*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    

    /* i=uv-1,j=0 */
    i=uv-1;
    j=0;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
	s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+j*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    
    /* i=0,j=vv-1 */
    i=0;
    j=vv-1;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i+j*uv)->v[k];
	s2.v[k] = (v+i+j*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    
    /* i=uv-1,j=vv-1 */
    i=uv-1;
    j=vv-1;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
	s2.v[k] = (v+i+j*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    
/*    kgeom_add_normals(kobj,n,uv*vv); */

#if 0
    free((char *)v);
    free((char *)n);
#endif
    *normals = n;
    return(CHILL);
}


static int
make_tstrip_mesh(kobject kin,
		 kgeom_object *kout,
		 int *dims,
		 vis_cmap *cmap,
		 int use_map,
		 int use_tc,
		 float *tc,
		 int do_normals)
{
    int i;
    int usize,vsize;
    float *oloc=NULL,*temp=NULL,*ocolors=NULL;
    float *tcoords=NULL;
    vertex_3d *verts = NULL;
    int vw,vh,vd,vt,ve;
    int index;
    vertex_3d bmin,bmax,center;
    vertex_3d wbmin,wbmax;
    int u_index,v_index;
    int rows;
    float *ival,*colors;
    float *iloc,*iloc_row_prev,*iloc_row0,*iloc_row1,*iloc_row_next;
    float *otc=NULL;
    vertex_3d *norms,*w_norms,*w2_norms;

    kgeom_triangles *tri;
    

    /* assign values to usize and vsize.  from the calling routine,
       dims[] holds the size of the value data.  i'm assuming that
       there are at two and only two values in this array that are > 1.
       such an assumption eliminates a lot of boring error checking
       code, but the below code will seg fault if this is not the case */
    
    for (i=0;i<4;i++)
	if (dims[i] != 1)
	    break;

    u_index = i;
    usize = dims[u_index];
    
    for (i=i+1;i<4;i++)
	if (dims[i] != 1)
	    break;
    v_index = i;
    vsize = dims[v_index];

    
    /* tweak pds attributes of input so that the first index corresponds to
     the u_index above, and the second index corresponds to the v_index
     above.

     NOTE:  sep 1995 - this hasn't been written.  we just assume that this
     is the case and proceed.  if this isn't the case, the output data will
     probably look like  a bunch of zeroes.  */
 
    if ((u_index != 0) && (v_index != 1))
    {
	kerror("GEOMETRY","lgquadmesh","lgquadmesh wants the data to be in the width and height axes.  Your data is organized in some other way.  Bye.\n");
	return(WHACKED);
    }
    
    kpds_get_attribute(kin,KPDS_VALUE_SIZE,&vw,&vh,&vd,&vt,&ve);
    if ((ve == 1) && (use_map))
    {
	ocolors = (float *)kmalloc((unsigned)(sizeof(float)*usize*2*4));
	colors = (float *)kmalloc((unsigned)(sizeof(float)*usize*2*4));
    }
    else
    {
	ocolors = NULL;
	colors = NULL;
    }

    if (!use_map)
    {
	if (use_tc)
	{
	    tcoords = (float *)kmalloc((unsigned)sizeof(float)*usize*2*2);

	    /* temp hack till we get the tcoord code working here.. */
	    for (i=0;i<usize*2*2;i++)
	       tcoords[i] = 0.;
#if 0	
	    kerror("GEOMETRY","lquadmesh","The implementation of tcoords for tstrips is incomplete in lgquadmesh.");

#endif	
	
	    kout->texture_coord_dim = 2;
	}
	else
	    tcoords = NULL;
    }
    else
    {
	tcoords = NULL;
    }

    if (do_normals)
    {
	w_norms = (vertex_3d *)kmalloc((unsigned)sizeof(vertex_3d)*usize*4);
	w2_norms = (vertex_3d *)kmalloc((unsigned)sizeof(vertex_3d)*usize*2);
	norms = (vertex_3d *)kmalloc((unsigned)sizeof(vertex_3d)*usize*2);
    }
    else
    {
	w_norms = NULL;
	w2_norms = NULL;
	norms = NULL;
    }
	
    
    /* malloc temporary space for two rows worth of data: locs, colors,
       normals, tc's, etc. */
    
    iloc = (float *)kmalloc((unsigned)(sizeof(float)*usize*4*3));
    iloc_row_prev = iloc;
    iloc_row0 = iloc + usize*3;
    iloc_row1 = iloc_row0 + usize*3;
    iloc_row_next = iloc_row1 + usize*3;
    
    oloc = (float *)kmalloc((unsigned)(sizeof(float)*usize*2*3));
    verts = (vertex_3d *)kmalloc((unsigned)(sizeof(vertex_3d)*usize*2));
    ival = (float *)kmalloc((unsigned)(sizeof(float)*usize*2));
    if (tcoords)
	otc = (float *)kmalloc((unsigned)(sizeof(float)*usize*2*2));
    else
	otc = NULL;



    /*
     * main loop:
     *
     * we grab a row's worth of information from the input, and spit out
     * a new tstrip.
     *
     * we keep 4 rows worth of location data in memory at once.  this much
     * is needed so that we can compute 2nd order-smooth normals for each
     * tstrip.  a lot of the code deals with buffer copying and that sort of
     * thing.  this is done to set up data correctly for the function that
     * computes the normals.  perhaps at some pooint in the future, that
     * routine could be rewritten to deal with something less general
     * (than the way it's presently set up).
     */

    /* grab row #0's location data. */
    assimilate_location_data(kin,iloc_row0,0,usize);
    lcompute_bbox((vertex_3d *)iloc_row0,usize,&bmin,&bmax);
    
    /* copy it to the "previous row" pointer */
    kmemcpy(iloc_row_prev,iloc_row0,sizeof(float)*usize*3);
    assimilate_location_data(kin,iloc_row1,1,usize);
    lcompute_bbox((vertex_3d *)iloc_row1,usize,&wbmin,&wbmax);

    lunion_boxes(&bmin,&bmax,&wbmin,&wbmax,&bmin,&bmax);

    for (rows=1;rows < vsize;rows++)
    {
	/* grab location data from input */
	if (rows+1 != vsize)
	{
	    assimilate_location_data(kin,iloc_row_next,rows+1,usize);
	    lcompute_bbox((vertex_3d *)iloc_row_next,usize,&wbmin,&wbmax);
	    lunion_boxes(&bmin,&bmax,&wbmin,&wbmax,&bmin,&bmax);
	}

	/* compute normals if requested */
	if (do_normals)
	{
	    lcompute_quadmesh_vertex_normals((vertex_3d *)iloc,&w_norms,usize,4);
	    if (rows == vsize-1)
		kmemcpy(w2_norms,w_norms,sizeof(vertex_3d)*2*usize);
	    else
		kmemcpy(w2_norms,w_norms+usize,sizeof(vertex_3d)*2*usize);
		
	    meshrow_to_tstrip((char *)norms,(char *)w2_norms,usize,sizeof(vertex_3d));
	}

	meshrow_to_tstrip((char *)verts,(char *)iloc_row0,usize,sizeof(vertex_3d));

	/* finally, copy the next row's worth of data to the starting
	   location in the input work array */

	/* -----  deal with colors or texture coords ----- */
	if (use_map)
	{
	    temp = ocolors;

	    /* grab the value data from the input object */
	    if (rows == 1)
	    {
		kpds_set_attribute(kin,KPDS_VALUE_POSITION,0,0,0,0,0);
		kpds_get_data(kin,KPDS_VALUE_LINE,ival);
		
		for (i=0;i<usize;i++,temp+=4)
		{
		    index = lget_map_index(cmap,ival+i);
		    lget_map_rgba(cmap,index,temp,temp+1,temp+2,temp+3);
		}
	    }
	    else
	    {
		kmemcpy(ocolors,ocolors+usize*4,sizeof(float)*usize*4);
		temp = ocolors + usize*4;
	    }
	    	    
	    kpds_set_attribute(kin,KPDS_VALUE_POSITION,0,rows,0,0,0);
	    kpds_get_data(kin,KPDS_VALUE_LINE,ival);

	    /* convert the value data to colors.. */
	    for (i=0;i<usize;i++,temp+=4)
	    {
		index = lget_map_index(cmap,ival+i);
		lget_map_rgba(cmap,index,temp,temp+1,temp+2,temp+3);
	    }
	    meshrow_to_tstrip((char *)colors,(char *)ocolors,usize,sizeof(float)*4);
	}
	else if (otc != NULL)
	{
	    /* texture coords -
	     tc[0] = umin, tc[1] = umax, tc[2] = vmin,tc[3] = vmax...*/
	    
	    float U,dU,V,dV;
	    int iu,oindex;
	    
	    dV = (tc[3]-tc[2])/(vsize-1);
	    V = tc[2] + (rows-1)*dV;

	    U = tc[0];
	    dU = (tc[1]-tc[0])/(usize-1);

	    oindex = 0;
	    for (iu=0;iu<usize;iu++,U+=dU)
	    {
		tcoords[oindex++] = U;
		tcoords[oindex++] = V;
	    }

	    U = tc[0];
	    V += dV;
	    for (iu=0;iu<usize;iu++,U+=dU)
	    {
		tcoords[oindex++] = U;
		tcoords[oindex++] = V;
	    }
	    
	    meshrow_to_tstrip((char *)otc,(char *)tcoords,usize,
			      sizeof(float)*2);
	    
	}
	

	tri = (kgeom_triangles *) 
	   kgeom_new_primitive(KGEOM_TRIANGLES_CONNECTED);
	tri->nverts = 2*usize;
	tri->locs    = kgeom_dup_locs((float *)verts, tri, kout);
	tri->cols    = kgeom_dup_cols((float *)colors, tri, kout);
	tri->norms   = kgeom_dup_norms((float *)norms, tri, kout);
	tri->tcoords = kgeom_dup_tcoords((float *)otc, tri, kout);
	kgeom_add_primitive(kout, (kgeom_primitive *) tri);

	/* now, update the vertex info. */
	kmemcpy(iloc,iloc_row0,sizeof(float)*3*usize);
	kmemcpy(iloc_row0,iloc_row1,sizeof(float)*3*usize);
	kmemcpy(iloc_row1,iloc_row_next,sizeof(float)*3*usize);

    }

    for (i=0;i<3;i++)
	kout->center[i] = bmin.v[i] + (bmax.v[i]-bmin.v[i])*0.5;

    kmemcpy(kout->min, (float *) &(bmin.v[0]), sizeof(float) * 3);
    kmemcpy(kout->max, (float *) &(bmax.v[0]), sizeof(float) * 3);
    kout->has_bounding_box = TRUE;
    kout->has_center       = TRUE;
    
    kfree(iloc);
    kfree(oloc);
    kfree(verts);
    kfree(otc);
    if (ival != NULL)
	kfree(ival);
    if (ocolors != NULL)
	kfree(ocolors);
    if (colors != NULL)
	kfree(colors);
    if (tcoords != NULL)
	kfree(tcoords);

    return(CHILL);
}

    /* unscramble location data from XXXYYYZZZ to XYZXYZXYZ */

static void
viffloc_to_geomloc(float *dst,
		   float *iloc,
		   int usize,
		   int vsize)
{
    register float *x,*y,*z,*d;
    int j;
	
    x = iloc;
    y = x + usize*vsize;
    z = y + usize*vsize;

    d = dst;
	
    for (j=0;j<usize*vsize;j++)
    {
	*d++ = *x++;
	*d++ = *y++;
	*d++ = *z++;

	if ((double) *x == 0.0) *x = 0.0;
	if ((double) *y == 0.0) *y = 0.0;
	if ((double) *z == 0.0) *z = 0.0;
    }
}

void
meshrow_to_tstrip(char *dst,
		  char *src,
		  int n,
		  int size)
{
    /*
     * utility routine to copy from basically a 2D array of values
     * comprising adjacent rows in a quadmesh to a linear array of
     * values which forms a t-strip.
     *
     * src array - contains 2*n items if size "size" each
     * dst array - must have space for 2*n items of size "size"
     */
    int i;
    int row_stride;
    char *base_src,*base_dst;

    base_src = src;
    base_dst = dst;

    row_stride = size*n;

    for (i=0;i<n;i++,base_src+=size)
    {
	kmemcpy(base_dst,base_src,(size_t)size);
	base_dst += size;
	kmemcpy(base_dst,base_src+row_stride,(size_t)size);
	base_dst += size;
    }
}

static int
assimilate_location_data(kobject kin,
			 float *a,
			 int row_num,
			 int usize)
{
    int j,index;
    float *tmp;

    tmp = (float *)kmalloc((unsigned)(sizeof(float)*usize*3));
    
    for (j=0,index=0;j<3;j++,index+=usize)
    {
	kpds_set_attribute(kin,KPDS_LOCATION_POSITION,0,row_num,0,j);
	kpds_get_data(kin,KPDS_LOCATION_LINE,a+index);
    }
    /* unscramble from viff xyz ordering to geom svcs xyz ordering. */
    viffloc_to_geomloc(tmp,a,usize,1);
    kmemcpy(a,tmp,sizeof(float)*usize*3);
    kfree(tmp);

    return TRUE;
}
/* -library_code_end */
