 /*
  * Khoros: $Id: lgtransform.c,v 1.9.2.4 1997/05/31 18:59:00 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lgtransform.c,v 1.9.2.4 1997/05/31 18:59:00 mirage Exp $";
#endif

/*
 * Copyright (C) 1996,1997 , The Regents of the University of California.
 * All rights reserved.  See $GEOMETRY/repos/license/LBL_Copyright.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for gtransform
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lgtransform
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lgtransform - perform geometric transformation upon the vertex (and possibly normals) data in a geom object.
* 
*       Purpose: To facilitate arbitrary transformations of geometry.
*
*         Input: input - the input geometry object
*                scale - float[3] used for scaling factors for x,y,z
*                    axes, respectively
*                trans - float[3] used for translating along x,y,z
*                    axes, respectively
*                do_normals - permit or inhibit transformation of surface
*                    normals, if present.
*
*        Output: output - the transformed geometry object
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: Restrictions on data or input as applicable
*                radii of spheres are not altered in this version of
*                code - only the vertex data.
*    Written By: wes
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lgtransform(kgeom_object *geom,
		float *scale,
		float *trans,
		int do_normals)
/* -library_def_end */

/* -library_code */
{
    int status, num_prims;
    int has_alpha;
    int i;
    vertex_3d bmin,bmax,lbmin,lbmax;

    num_prims = kgeom_number_primitives(geom);
    
    for (i=0;i<3;i++)
    {
	bmin.v[i] = KGEOM_BIGNUM;
	bmax.v[i] = -1. * KGEOM_BIGNUM;
    }

    for (i=0;i<num_prims;i++)
    {
       kgeom_primitive *prim = kgeom_get_primitive(geom, i);
       
       switch (prim->type)
       {
	 case KGEOM_POLYLINE_DISJOINT:
	 case KGEOM_POLYLINE_CONNECTED:
	 {
	    kgeom_polyline *lines = (kgeom_polyline *) prim;
	    float *verts=NULL;
	    vertex_3d *outverts;
	    int nverts;
	    matrix4x4 m;

	    nverts = lines->nverts;
	    verts  = lines->locs;
	    
	    if (nverts == 0)
		break;
	    
	    outverts = (vertex_3d *)kmalloc(sizeof(vertex_3d)*nverts);
	    
	    identity_4x4(&m);
	    m.m[0][0] = scale[0];
	    m.m[1][1] = scale[1];
	    m.m[2][2] = scale[2];

	    m.m[3][0] = trans[0];
	    m.m[3][1] = trans[1];
	    m.m[3][2] = trans[2];

	    npoint_xfrm(verts,&m,(float *)outverts,nverts);

	    lines->locs = (float *) outverts;
	    kfree(verts);
	    
	    lcompute_bbox(outverts,nverts,&lbmin,&lbmax);
	    lunion_boxes(&lbmin,&lbmax,&bmin,&bmax,&bmin,&bmax);
	}
	break;
	
        case KGEOM_TRIANGLES_CONNECTED:
        case KGEOM_TRIANGLES_DISJOINT:
	{
	    /* Changed from line below, Terry Ligocki:
	     * kgeom_triangles *tri = (kgeom_triangles *) tri;
	    */
	    kgeom_triangles *tri = (kgeom_triangles *) prim;
	    int nverts;
	    float *verts;
	    vertex_3d *outverts;
	    matrix4x4 m;
	    
	    nverts = tri->nverts;
	    verts  = tri->locs;

	    if (nverts == 0)
		break;
	    
	    outverts = (vertex_3d *)kmalloc(sizeof(vertex_3d)*nverts);

	    identity_4x4(&m);
	    m.m[0][0] = scale[0];
	    m.m[1][1] = scale[1];
	    m.m[2][2] = scale[2];

	    m.m[3][0] = trans[0];
	    m.m[3][1] = trans[1];
	    m.m[3][2] = trans[2];

	    npoint_xfrm(verts,&m,(float *)outverts,nverts);

	    /* Changed to check "do_normals" also, Terry Ligocki */
	    if (tri->norms && do_normals == TRUE)
	       kerror("GEOMETRY","lgtransform",
		      "transformation of normals is not ready yet.");

	    tri->locs = (float *) outverts;
	    kfree(verts);
	    
	    lcompute_bbox(outverts,nverts,&lbmin,&lbmax);
	    lunion_boxes(&lbmin,&lbmax,&bmin,&bmax,&bmin,&bmax);
	}
	break;
	default:
	    kerror("GEOMETRY","lgtransform",
		"only line and triangles can be transformed currently.");
	    break;
	}
	
    }

    kmemcpy(geom->min, &(bmin.v[0]), sizeof(float)*3);
    kmemcpy(geom->max, &(bmax.v[0]), sizeof(float)*3);
    for (i=0;i<3;i++)
	geom->center[i] = bmin.v[i] + 0.5 * (bmax.v[i] - bmin.v[i]);

    geom->has_bounding_box = TRUE;
    geom->has_center       = TRUE;
    
    
    return TRUE;
}
/* -library_code_end */
