
/*
 * Copyright (c) 1996,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */
 

#include <design.h>
#include <geometry.h>
#include "render.h"
#include "camera.h"
#include "imaging.h"

#define MAX_RENDERERS 4

static renderer_struct *avail_renderers[MAX_RENDERERS];
static int number_avail_renderers=0;

static renderer_struct software_renderer;
static renderer_struct native_renderer; /* the native hardware renderer */
static int active_renderer=0;
static int requested_renderer=0;

int
set_active_renderer(int i)
{
    if ((i >= 0) && (i < number_avail_renderers))
    {
	active_renderer=i;
	return(CHILL);
    }
    else
	return(WHACKED);
}

int
get_active_renderer(void)
{
    return(active_renderer);
}

static
renderer_struct *get_active_render_struct(void)
{
    int active;
    active=get_active_renderer();
    return(avail_renderers[active]);
}

int
is_active_renderer_stereo_capable()
{
    renderer_struct *r;
    r = get_active_render_struct();
    return(r->stereo_capable);
}


int
have_graphics_hardware(renderer_struct *renderer,screen_stuff *s)
{
    char *xdisp_name;
    Display *xdisp;
    int nexts;
    char **exts;
    int status;
    char *server_vendor;
    
    /**
      * determine if there is support for hardware rendering on this
      * platform.  if so, return TRUE, otherwise, return FALSE.
    **/

    if (s->display == NULL)  /* shouldn't ever happen...*/
    {
	fprintf(stderr," Hey bagel-brain, set $DISPLAY to something, ok?\n");
	kexit(0);
    }

    xdisp = XOpenDisplay(getenv("DISPLAY"));
    server_vendor = XServerVendor(xdisp);

    exts = XListExtensions(xdisp,&nexts);

    /**
    * 2 May 1996 - in previous incarnations, like for GL and
    * XGL, we would look for extensions in the X-server that
    * indicated support for some kind of hardware-assisted
    * rendering.  presently, we're only doing OpenGL, which is
    * technically an X extension.  rather than query the server
    * for this extension, we're just going to make the OpenGL
    * setup call.  if it fails, well there's no OpenGl.  in other
    * words, this code is not really useful and should be removed...wes
    **/
    status = find_gfx_hardware(exts,nexts,server_vendor,
			       &(renderer->renderer_name[0]),
			       s->display,s->screen);  /* in machine.c */

    XFreeExtensionList(exts);
    return(status);
}


void
identify_native_renderer(void)
{
    char *xdisp_name;
    screen_stuff s;
    int depth;
    
    /**
      * the purpose of this routine is to determine some information
      * about $DISPLAY, like whether or not is supports graphics
      * acceleration, etc.  .
    **/
    
    xdisp_name = getenv("DISPLAY");
    s.display = XOpenDisplay(xdisp_name);
    s.screen = DefaultScreen(s.display);
    
#ifdef _USE_GFX_HARDWARE
    if (have_graphics_hardware(&native_renderer,&s) == TRUE)
    {
	avail_renderers[number_avail_renderers] = &native_renderer;
	native_renderer.renderer_active = 1;
	number_avail_renderers++;

	/** the routine init_native_renderer() is present in ogl.c,.
	using this way of organizing things, it is impossible to
	have more than one flavor of hardware assisted rendering
	per version of render. **/
	init_native_renderer(&native_renderer);

    }
#endif
}


void
identify_software_renderer(void)
{
    /**
      * the purpose of this routine is to determine some information
      * about $DISPLAY, like whether or not is supports graphics
      * acceleration, etc.
      *
      * this routine must be called prior to establishing any other
      * renderers.
    **/
    software_renderer.ndisplay_options = 5;
    software_renderer.selected_display_option = 0;  /* default is 3-3-2  */

    software_renderer.display_options_txt[0] = s332_STRING;
    software_renderer.display_options_txt[1] = s24BIT_STRING;
    software_renderer.display_options_txt[2] = sQUANT_STRING;
    software_renderer.display_options_txt[3] = sDITHER_STRING;
    software_renderer.display_options_txt[4] = sNODISPLAY_STRING;

    software_renderer.is_hardware_renderer = 0;
    
    kstrcpy(software_renderer.renderer_name,"Software Renderer");
    
    software_renderer.renderer_active = 1;

    number_avail_renderers=1;
    avail_renderers[0] = &software_renderer;

    software_renderer.init_output_func = init_render_output;
    software_renderer.render_func = software_render_image;
    software_renderer.shutdown_output_func = shutdown_render_output;
    software_renderer.set_camera_func = software_set_camera;
    software_renderer.window_resize_func = software_window_resize;
    software_renderer.refresh_func = NULL; /* backing store works ok in
					      software..no need for this. */
    software_renderer.stereo_capable = FALSE;
}

int
get_all_renderers_info(int *n,
		       renderer_struct ***r)
{
    *n = number_avail_renderers;
    *r = avail_renderers;
    return(CHILL);
}

int
select_hardware_renderer()
{
    renderer_struct **r;
    int nrenderers,i;

    get_all_renderers_info(&nrenderers,&r);
    for (i=0;i<nrenderers;i++)
    {
	if (r[i]->is_hardware_renderer)
	{
	    /* choose the first hardwar */
	    set_active_renderer(i);
	    break;
	}
    }
    return(CHILL);
}

void
renderer_init_func (void)
{
    int w,h,img_display_choice,do_named_flag;
    int xpos,ypos;
    char transport_name[128];
    renderer_struct *r;
    int (*fptr)(int,int,int,int,int,int,kstring);

    get_render_image_position(&xpos,&ypos);
    get_image_dims(&w,&h);
    get_image_display_choice(&img_display_choice);
    get_output_transport_name(transport_name);
    get_do_named_output(&do_named_flag);
    r = get_active_render_struct();

    fptr = r->init_output_func;
    (*fptr)(w,h,xpos,ypos,img_display_choice,do_named_flag,transport_name);
}

void
renderer_shutdown_func(void)
{
    renderer_struct *r;
    int (*fptr)(void);
    
    r = get_active_render_struct();

    fptr = r->shutdown_output_func;
    (*fptr)();
}

void
renderer_invoke_func(void)
{
    renderer_struct *r;
    int (*fptr)(void);
    
    r = get_active_render_struct();

    fptr = r->render_func;
    (*fptr)();
}

void
renderer_resize_func(int new_width,
		    int new_height)
{
    renderer_struct *r;
    int (*fptr)(int,int);
    
    r = get_active_render_struct();

    fptr = r->window_resize_func;
    (*fptr)(new_width,new_height);
    compute_view_xform();
}

void
renderer_refresh_func(void)
{
    renderer_struct *r;
    int (*fptr)(void);
    
    r = get_active_render_struct();

    fptr = r->refresh_func;
    if (fptr != NULL)
	(*fptr)();
}

int
set_requested_renderer (int i)
{
    requested_renderer = i;
    return(CHILL);
}

int
get_requested_renderer()
{
    return(requested_renderer);
}
