
/*
 * Copyright (c) 1996,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 */
 

/**********************************************************************

  Log:

  
1/1/95, wes. changed the way that arrays returned by kgeom_get_attribute()
    are handled.  previously, this call returned an array.  presently, it
    returns a pointer to an area of memory under control of the system.
    hence, it can't be free'd, nor otherwise used.  it is read-only and
    volatile.  the changes made were to cull out free's and to do 
    kmemcpy's where previously these arrays were copied directly into 
    render_obj structs.

10 jan 95, wes.    cleaned up file detect code.
30 jan 95, wes.    connected triangle strips without normals will no
    longer be converted to disjoint triangles.  rather, the normals
    will be computed on a per-face basis, resulting in images in which
    the objects appear faceted.  users now bear the responsiblility
    of generating appropriate normals if this behavior is not desireable.

7 july 95,  wes.    optimization of geom_xfrm_points, stuff for threads.

12 jan 96, wes. add support for text.

Thu Mar  7 15:39:45 PST 1996 wes
    - more text support
    - more input objects (now 7 in number)

Wed May 22 14:33:29 MDT 1996
    - support for vr xforms.

Tue Mar 18 10:50:31 PST 1997
    - preliminary support for opengl display lists for disjoint 
	triangle primitives ONLY.

*********************************************************************/

#ifdef _USE_THREADS
#include "geometry/gthreads.h"
#endif

#include <design.h>

#include "render.h"

#include "render_objs.h"
#include "shader.h"
#include "clip.h"
#include "vrxfrm.h"
#include "texture.h"

#include "geometry/rmatrix.h"

int reread_input_object PROTO((xvobject, char *, kaddr));

static int squelch_the_squawk=0;
static renderobj root;
static int num_objects_in_tree=0;
static renderobj *current_object = &root;
static matrix4x4 current_parent_xfrm,current_parent_xfrm_inverse;

static int whats_in_this_object PROTO((char *));
static void free_renderobj PROTO((renderobj *r));
static void free_render_prim PROTO((render_prim *p));
static int build_disjoint_from_connected_perface PROTO((render_prim *,int,float *,float *,float *,float *,int, int));
static int build_disjoint_from_connected_nonorms PROTO((render_prim *p,int nverts,float *verts,float *colors,int layout,int has_alpha));
static int tstrip_to_disjoint_v PROTO((float *s,float *d,int n,int size_s));
static int tstrip_to_disjoint_f PROTO((float *s,float *d,int n,int size_s));
static int compute_octmesh_radii PROTO((int *mesh_dims,float *r,float *v));

static float fdetect_wait_time=2.0;  /* time in seconds between select() calls
					on input file descriptors. */
static float vr_fdetect_wait_time= 0.; /* we want 20 vr events/sec */

int i_zero=0,i_one=1,i_two=2,i_three=3,i_four=4,i_five=5,i_six=6;
static renderobj *input_objects[NUMBER_RENDER_INPUT_OBJECT_PORTS]=
{
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};


/**
  * the following utility routines are responsible for
  * converting the various khoros specific primitives into
  * render "native" format.
  *
  * in the future, there will be a "low memory" version of
  * this code which won't make a copy of all the vertex, etc.
  * information.
  *
  * NOTE: the order of the following routines is the same as in
  * $DATASERV/include/kappserv/kgeometry.h!!!!
**/

static void rm_nullprim PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kpolyline_dis PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kpolyline_con PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_ktriangles_dis PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_ktriangles_con PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kspheres PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kcylinders PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kquadmesh PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_koctmesh PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kdirpoints PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kpolyhedra_face PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_kpolyhedra_verts PROTO((render_prim *, kgeom_object *, kgeom_primitive *,int, int));
static void rm_ktexture2d PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_ktexture3d PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));
static void rm_ktext PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int));


static int primitives[] = 
{
	KNULL_PRIMITIVE,
	KGEOM_POLYLINE_DISJOINT,
	KGEOM_POLYLINE_CONNECTED,
	KGEOM_TRIANGLES_DISJOINT,
	KGEOM_TRIANGLES_CONNECTED,
	KGEOM_SPHERES, 
	KGEOM_CYLINDERS, 
	KGEOM_DIRECTED_POINTS, 
	KGEOM_QUADMESH, 
	KGEOM_OCTMESH,  
	KGEOM_TEXT
#if 0
	KGEOM_TEXTURE2D, 
	KGEOM_TEXTURE3D, 
#endif
};

static void (*convert_prim_funcs[]) PROTO((render_prim *, kgeom_object *, kgeom_primitive *, int, int)) =
{
    rm_nullprim,
    rm_kpolyline_dis,
    rm_kpolyline_con,
    rm_ktriangles_dis,
    rm_ktriangles_con,
    rm_kspheres,
    rm_kcylinders,
    rm_kdirpoints,
    rm_kquadmesh,
    rm_koctmesh,
    rm_ktext
#if 0
    rm_ktexture2d,
    rm_ktexture3d
#endif
};

/* determine the number of primitive types from the above table */
static int num_prim_types = sizeof(primitives) / sizeof(primitives[0]);

#define NUM_PRIM_TYPES num_prim_types


static void copy_kattribs_to_renderobj PROTO((kgeom_object *,renderobj *,int));
static void compute_render_bbox PROTO((renderobj *));
static void determine_transparency PROTO((renderobj *));
static void determine_default_shader PROTO((renderobj *));
static int search_for_named_object PROTO((char *, renderobj **, renderobj *));
static void union_bounding_boxes PROTO((renderobj *a,renderobj *b,vertex_3d *bmin, vertex_3d *bmax));

int load_child_names PROTO((char **, int *, renderobj *, int));

/**
  * this file contains routines which are repsonsible for
  * converting between khoros geom objects and render
  * objects.  this routine knows about the clui_info_struct
  * as well as the render-specific objects.
*/



int
init_object_list(clui_info_struct *rmc)
{
    int status;
    renderobj *t;
    int root_num_kids = 0;
    

    /* initialize the root object */
    init_render_obj(&root,ROOT_OBJECT_NAME);
    current_object = &root;
    num_objects_in_tree = 1;

    /* now, read in each of the objects.it would be nice if
     the i1_files in the command structure were in an array....
     also, this is where the scene services interface will go
     to grab kgeom_objects whenever that finally gets built */

    /*
     * Wed May 22 14:33:29 MDT 1996
     *
     * with the addition of transformation events, we no longer create
     * an object for every input file.  KGEOM_OBJECT_XFORM events will be
     * applied to the transformation of a particular object;
     * KGEOM_CAMERA_XFORM updates the camera view matrix (which is ill
     * coded right now), KGEOM_CAMERA_ORIENTATION changes various viewing
     * parameters and is the only guaranteed way to change render's
     * camera - KGEOM_CAMERA_XFORM is intended to be output by render
     * and used by other modules, such as gvr_xform_obj.
     *
     * a change that is needed is to first process all the geometry,
     * then process all the vr transformation commands, otherwise a named
     * object might not be found cuz it hasn't been added to the object
     * tree yet.
     */

    if (rmc->i1_flag)
    {
        status = whats_in_this_object(rmc->i1_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i1_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[0] = t;
		xvw_add_detectfile(NULL,rmc->i1_file,fdetect_wait_time,
				   reread_input_object,&i_zero);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	    vr_process_object(rmc->i1_file);
	    xvw_add_detectfile(NULL,rmc->i1_file,vr_fdetect_wait_time,
			       reread_input_object,&i_zero);
	}
	/* else we'll just continue.. */
    }

    /* check the additional flags.  this is where having an
       array would be nice.  c'est la vie. */

    if (rmc->i2_flag)
    {
        status = whats_in_this_object(rmc->i2_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i2_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[1] = t;
		xvw_add_detectfile(NULL,rmc->i2_file,fdetect_wait_time,
				   reread_input_object,&i_one);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	    vr_process_object(rmc->i2_file);
	    xvw_add_detectfile(NULL,rmc->i2_file,vr_fdetect_wait_time,
			       reread_input_object,&i_one);
	}
	/* else we'll just continue.. */
    }
    
    if (rmc->i3_flag)
    {
        status = whats_in_this_object(rmc->i3_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i3_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[2] = t;
		xvw_add_detectfile(NULL,rmc->i3_file,fdetect_wait_time,
				   reread_input_object,&i_two);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	    vr_process_object(rmc->i3_file);
	    xvw_add_detectfile(NULL,rmc->i3_file,vr_fdetect_wait_time,
			       reread_input_object,&i_two);
	}
	/* else we'll just continue.. */
    }
    
    if (rmc->i4_flag)
    {
        status = whats_in_this_object(rmc->i4_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i4_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[3] = t;
		xvw_add_detectfile(NULL,rmc->i4_file,fdetect_wait_time,
				   reread_input_object,&i_three);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	    vr_process_object(rmc->i4_file);
	    xvw_add_detectfile(NULL,rmc->i4_file,vr_fdetect_wait_time,
			       reread_input_object,&i_three);
	}
	/* else we'll just continue.. */
    }

    if (rmc->i5_flag)
    {
        status = whats_in_this_object(rmc->i5_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i5_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[4] = t;
		xvw_add_detectfile(NULL,rmc->i5_file,fdetect_wait_time,
				   reread_input_object,&i_four);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	    vr_process_object(rmc->i5_file);
	    xvw_add_detectfile(NULL,rmc->i5_file,vr_fdetect_wait_time,
			       reread_input_object,&i_four);
	}
	/* else we'll just continue.. */
    }

    if (rmc->i6_flag)
    {
        status = whats_in_this_object(rmc->i6_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i6_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[5] = t;
		xvw_add_detectfile(NULL,rmc->i6_file,fdetect_wait_time,
				   reread_input_object,&i_five);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	    vr_process_object(rmc->i6_file);
	    xvw_add_detectfile(NULL,rmc->i6_file,vr_fdetect_wait_time,
			       reread_input_object,&i_five);
	}
	/* else we'll just continue.. */
    }

    if (rmc->i7_flag)
    {
        status = whats_in_this_object(rmc->i7_file);

	if (status == RM_GEOMETRY)
	{
	    t = create_renderobj();
	    status = convert_kobj_to_renderobj(rmc->i7_file,t);
	    if (status != CHILL)
/* KERROR CAT: "init_object_list" */
	        kerror("RENDER","render","failed to convert kgeom_object (-i) to internal format");
	    else
	    {
	        add_child(&root,t);
		num_objects_in_tree++;
		/*	    replace_input_object(0,t); */
		input_objects[6] = t;
		xvw_add_detectfile(NULL,rmc->i7_file,fdetect_wait_time,
				   reread_input_object,&i_six);
	    }
	}
	else if (status == RM_VRSTUFF)
	{
	  float one=1.0;
	    vr_process_object(rmc->i7_file);
	    xvw_add_detectfile(NULL,rmc->i7_file,one,
			       reread_input_object,&i_six);
	}
	/* else we'll just continue.. */
    }

    set_current_object(&root);
    
    return(CHILL);
}

void
init_render_obj(renderobj *r,char *name)
{
    memset((char *)r,0,sizeof(renderobj));
    kstrncpy(r->name,name,KGEOM_MAX_NAME_LENGTH);

    identity_4x4(&(r->xfrm));

    /* default surface attributes */
    r->attribs.amb = KGEOM_AMBIENT_DEFAULT;
    r->attribs.diff = KGEOM_DIFFUSE_DEFAULT;
    r->attribs.spec = KGEOM_SPECULAR_COEFFICIENT_DEFAULT;
    r->attribs.spec_exp = KGEOM_SPECULAR_EXPONENT_DEFAULT;
    r->attribs.spec_metal = KGEOM_METAL_DEFAULT;

    /* set center to origin */
    r->center.v[0] = r->center.v[1] = r->center.v[2] = 0.;

    /* default color */
    r->colors.r = r->colors.g = r->colors.b = r->colors.a = 1.;

    /* default display status */
    r->do_display = TRUE;

    /* default shader */
    r->shader = KSHADER_INHERIT;
}


renderobj *
create_renderobj()
{
    renderobj *t;
    t = (renderobj *)kmalloc(sizeof(renderobj));
    init_render_obj(t,NULL);
    return(t);
}

int
add_child(renderobj *parent, renderobj *child)
{
    int i;
    vertex_3d union_box_min,union_box_max;
    
    if ((parent->nchildren == MAXOFFSPRING) ||
	(child->nparents == MAXOFFSPRING))
	return(WHACKED);
    
    parent->children[parent->nchildren++] = child;
    child->parents[child->nparents++] = parent;

    /**
      * update parent's bounding box to reflect extents of the
      * child's bounding box.  this is union type operation.
    **/

    if (!(child->flags & IGNOREBOX))
    {
	union_bounding_boxes(parent,child,&union_box_min,&union_box_max);

	VCOPY(&(parent->bmin),&union_box_min);
	VCOPY(&(parent->bmax),&union_box_max);
    }

    for (i=0;i<3;i++)
	parent->center.v[i] = parent->bmin.v[i] + 0.5 * (parent->bmax.v[i] - parent->bmin.v[i]);

    return(CHILL);
}

int
convert_kobj_to_renderobj(char *file,
			    renderobj *r)
{
    kgeom_object *k;
    kgeom_primitive *kprim;
    char *name=NULL;
    int nprims,i, j;
    render_prim *p;
    int *primlist=NULL;
    
    k = kgeom_read_object(file);
    if (k == NULL)
    {
	kerror("RENDER","convert_kobj_to_renderobj","The input object <%s> is problematic, and can't be opened.",file);
	return(WHACKED);
    }

    /*
     * we will be taking over data from the geometry object.  by setting
     * pointers to NULL after we've snagged them, that frees them from
     * the control of the object
     */

    /* get the name */
    name = k->name;   k->name = NULL;
    
    /* assign name if there is one   -- added 7/11/93  // Steve K */
    if (name)
      strncpy((char *)r->name,(char *)name,KGEOM_MAX_NAME_LENGTH);

    /* get the number of primitives in this object */
    nprims = kgeom_number_primitives(k);
    r->nprims = nprims;

    /* grab some memory for the primitives. */
    r->prims = (render_prim *)kmalloc(sizeof(render_prim)*nprims);
    memset((char *)(r->prims),0,sizeof(render_prim)*nprims);

    r->layout       = k->layout;
    r->has_alpha    = k->has_alpha;
    r->verts_nspace = k->location_dim;
    
    /* if 2-space vertices, ignore this object on transformations. */
    if (r->verts_nspace == 2)
       r->flags = IGNOREBOX;
    else
       r->flags = 0;

    /* actually, we want all data to appear to be in 3 space */
    k = kgeom_new_location_dim(k, 3);

    for (i=0;i<nprims;i++)
    {
       kprim = kgeom_get_primitive(k, i);
       p = r->prims+i;
       p->type = kprim->type;

       for (j=0;j<NUM_PRIM_TYPES;j++)
	  if (p->type == primitives[j])
	  {
	     (*convert_prim_funcs[j])(p,k,kprim,r->layout,r->has_alpha);
	     break;
	  }
       
       copy_kattribs_to_renderobj(k,r,r->has_alpha);
       determine_default_shader(r);
    }
    
/*    kfree(name);  1/1/95 don't free stuff returned by get_attrib(). wes*/
    kgeom_blast_object(k);
    return(CHILL);
}

int
return_object_names(char ***list,int *n)
{
    char *t;
    int i;
    
    t = (char *)kmalloc(sizeof(char)*KGEOM_MAX_NAME_LENGTH);
    *list = (char **)kmalloc(sizeof(char *)*num_objects_in_tree);
    
    kstrncpy((char *)t,(char *)(root.name),KGEOM_MAX_NAME_LENGTH);

    *list[0] = t;  /* set entry 0 to the root object. */
    *n = 1;

    for (i=0;i<root.nchildren;i++)
	load_child_names(*list,n,root.children[i],1);

    return(CHILL);
}

int
free_object_names(char **list,int n)
{
    int i;
    for (i=0;i<n;i++)
	kfree(list[i]);
    kfree(list);
    return(CHILL);
}

int
load_child_names(char **list,int *n,renderobj *r,int level)
{
    int i;
    char *t,*t2;

    t = t2 = (char *)kmalloc(sizeof(char)*KGEOM_MAX_NAME_LENGTH);
    
    for (i=0;i<level;i++)
	*t2++ = ' ';
    
    kstrncpy((char *)t2,(char *)(r->name),KGEOM_MAX_NAME_LENGTH-level);
    list[*n] = t;
    *n += 1;

    if (r->nchildren > 0)
    {
	/* recursively descend into tree */
	for (i=0;i<r->nchildren;i++)
	    load_child_names(list,n,r->children[i],level+1);
    }

    return(CHILL);
}

static void
rm_nullprim (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
}

static void
rm_ktext (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int nstrings,i;

    char **strings=NULL;
    char **fonts=NULL;
    float *locs=NULL,*offset=NULL,*ups=NULL,*flows=NULL;
    float *colors=NULL, *back_colors=NULL;
    render_text_struct *rmt;

    kgeom_text *text = (kgeom_text *) k;
    

    nstrings = text->nverts;
    strings  = text->strings;  text->strings = NULL;
    fonts    = text->fonts;    text->fonts = NULL;
    locs     = text->locs;     text->locs = NULL;
    offset   = text->offsets;  text->offsets = NULL;
    ups      = text->ups;      text->ups = NULL;
    flows    = text->ups;      text->flows = NULL;
    colors   = text->cols;     text->cols = NULL;
    back_colors = text->bgcols; text->bgcols = NULL;
    

    p->nverts = nstrings;
    p->type = KGEOM_TEXT;

    rmt = p->prim_specifics.text = (render_text_struct *)kmalloc(sizeof(render_text_struct)*nstrings);

    /*
     * loop which preprocesses text strings for the renderer.
     * for the software renderer, we'll make calls to X at this point.
     * the results of these calls will be arrays of pixels which will be
     * subsequently composited into the image.  the hardware renderers
     * can work with text directly.
     *
     * since render provides no facilities for manipulating the
     * characteristics of text strings (10 jan 96), other than overall color,
     * any pre-rendering rasterization can occur here.  font, etc. changes
     * must happen to the text object outside of render.  the file
     * detect will make sure the object gets reread, and this routine
     * will get called again....
     */
    
    for (i=0;i<nstrings;i++)
    {
	/* grab the string... */
	rmt[i].source_string = (char *)kmalloc(sizeof(char)*strlen(strings[i]));
	strncpy((char *)(rmt[i].source_string),strings[i],strlen(strings[i]));

	/* grab the font... */
	if (fonts == NULL)
	    rmt[i].source_font = NULL;
	else
	{
	    rmt[i].source_font = (char *)kmalloc(sizeof(char)*strlen(fonts[i]));
	    strncpy((char *)(rmt[i].source_font),fonts[i],strlen(fonts[i]));
	}

	/* grab the colors.. */
	if (colors)
	{
	    rmt[i].fg_color = (surface_colors_rgb *)kmalloc(sizeof(surface_colors_rgb));
	    kmemcpy(&(rmt[i].fg_color),colors+i,sizeof(surface_colors_rgb));
	}
	else
	    rmt[i].fg_color = NULL;

	/* call the routine that builds the ROP mask.. the mask is
	 independent of color. */
	rmt[i].mask = build_text_mask(strings[i],
				      (fonts != NULL) ? fonts[i] : NULL,
				      rmt+i);
    }

    p->verts = (vertex_3d *)locs;

    
}

static void
rm_kpolyline_dis (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int nverts;
    float *verts=NULL,*colors=NULL;

    kgeom_polyline *lines = (kgeom_polyline *) k;

    nverts = lines->nverts;
    
#ifdef DEBUG
    kprintf("\tthere are %d verts in this kpolyline_dis prim.\n",nverts);
#endif

    verts  = lines->locs;  lines->locs = NULL;
    colors = lines->cols;  lines->cols = NULL;

    if (verts == NULL) /* strange object */
	p->nverts = 0;
    else
	p->nverts = nverts;
    p->verts = (vertex_3d *)verts;

    if (colors == NULL)
	p->ncolors = 0;
    else if (layout == KPER_VERTEX)
	p->ncolors = nverts;
    else /*  (layout == KPER_VECTOR) */
	p->ncolors = nverts/2;

    if (has_alpha)
	p->rgba_color_list = (surface_colors_rgba *)colors;
    else
	p->rgb_color_list = (surface_colors_rgb *)colors;
    
    p->type = KGEOM_POLYLINE_DISJOINT;

    p->clipfunc = (int(*)(void))clip_polyline_dis;

}

static void
rm_kpolyline_con (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int nverts;
    float *verts=NULL,*colors=NULL;
    kgeom_polyline *lines = (kgeom_polyline *) k;

    nverts = lines->nverts;
    
#ifdef DEBUG
    kprintf("\tthere are %d verts in this kpolyline_dis prim.\n",nverts);
#endif

    verts  = lines->locs;  lines->locs = NULL;
    colors = lines->cols;  lines->cols = NULL;

    if (verts == NULL) /* strange object */
	p->nverts = 0;
    else
	p->nverts = nverts;
    p->verts = (vertex_3d *)verts;

    if (colors == NULL)
	p->ncolors = 0;
    else if (layout == KPER_VERTEX)
	p->ncolors = nverts;
    else /*  (layout == KPER_VECTOR) */
	p->ncolors = nverts-1;

    if (has_alpha)
	p->rgba_color_list = (surface_colors_rgba *)colors;
    else
	p->rgb_color_list = (surface_colors_rgb *)colors;
    
    p->type = KGEOM_POLYLINE_CONNECTED;

    p->clipfunc = (int(*)(void))clip_polyline_con;
}

static void
rm_ktriangles_dis (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int nverts;
    vertex_3d p1,p2;
    vertex_3d *vnormals,*vverts;
    int nnormals,i;
    int index,j;
    double dmag;
    float *verts=NULL,*colors=NULL,*normals=NULL,*tcoords=NULL;
    kgeom_triangles *tri = (kgeom_triangles *) k;

    nverts = tri->nverts;
    
#ifdef DEBUG
    kprintf("\tthere are %d verts in this ktriangles_dis prim.\n",nverts);
#endif

    verts   = tri->locs;      tri->locs = NULL;
    colors  = tri->cols;      tri->cols = NULL;
    normals = tri->norms;     tri->norms = NULL;
    tcoords = tri->tcoords;   tri->tcoords = NULL;
    
    if (verts == NULL) /* strange object */
	p->nverts = 0;
    else
	p->nverts = nverts;
    p->verts = (vertex_3d *)verts;

    if (colors == NULL)
	p->ncolors = 0;
    else if (layout == KPER_VERTEX)
	p->ncolors = nverts;
    else /*  (layout == KPER_FACE) */
	p->ncolors = nverts/3;

    if (has_alpha)
	p->rgba_color_list = (surface_colors_rgba *)colors;
    else
	p->rgb_color_list = (surface_colors_rgb *)colors;

    if (normals == NULL)
    {
	/**
	  * this section of code has not been tested.  8/17/94
	**/
	/**
	 * if there were no normals present in the object, compute them
	 * here.
	 **/
	if (layout == KPER_VERTEX)
	    nnormals = nverts;
	else
	    nnormals = nverts/3;
	vnormals = (vertex_3d *)kmalloc(sizeof(vertex_3d)*nnormals);
	vverts = (vertex_3d *)verts;
	
	for (i=0;i<nnormals;i++)
	{
	    /**
	      * compute index so that we grab the correct vertices for
	      * each triangle.  
	    **/

	    if (layout == KPER_VERTEX)
		index = i/3;
	    else
		index = i*3;
	    
	    for (j=0;j<3;j++)
	    {
		p1.v[j] = vverts[index+1].v[j] - vverts[index].v[j];
		p2.v[j] = vverts[index+2].v[j] - vverts[index+1].v[j];
	    }
	    
	    vertex_cross(&p1,&p2,vnormals+i);
	    vertex_unit(vnormals+i,&dmag);
	    
	    if (i & 1) /* flip every other normal */
	    {
		for (j=0;j<3;j++)
		    vnormals[i].v[j] *= -1.;
	    }
	}
	p->nthings = nnormals;
/*	normals = (float *)vnormals; */
	p->prim_specifics.normals = vnormals;
    }
    else			/* have normals.. */
    {
	/* first set the number of normals based upon the layout flag */
	if (layout == KPER_VERTEX)
	    p->nthings = nverts;
	else /* layout == KPER_FACE */
	    p->nthings = nverts/3;
	p->prim_specifics.normals = (vertex_3d *)normals;

	/* now, possibly unitize the normals */
	{
	    double mag;
	    /* look at one of the normals.  if it's not a unit normal,
	       normalize them.. */
	    mag = normals[0]*normals[0] + normals[1]*normals[1] +
		normals[2]*normals[2];
	    mag = fabs(mag-1.0);
	    if (mag > 1e-5)		/* allow for some roundoff... */
	    {
		nnormal_unit((vertex_3d*)normals,(vertex_3d *)normals,
			     p->nthings);
	    }
	}
    }

    p->type = KGEOM_TRIANGLES_DISJOINT;

    p->clipfunc = (int(*)(void))clip_polygon_cheap;

    /* blow off texture coordinates until things get straightened out */
}


static void
rm_ktriangles_con (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int nverts;
    int i;
    int index,j;
    double dmag;
    float *verts=NULL,*colors=NULL,*normals=NULL,*tcoords=NULL;
    char *tname=NULL;
    int has_tc,tc_size;
    kgeom_triangles *tri = (kgeom_triangles *) k;
    /**
      * 1/30/95
      * tstrips which don't have normals will have them computed here
      * in this routine.  we have two choices, either to compute the
      * normals per-face (cross product of triangle edges), or to compute
      * the normals so as to approximate some kind of smooth surface.
      * we choose to compute the normals in the former way.  the images
      * resulting from computing normals in this way will appear faceted.
      * if the user wishes to have the t-strip approximate a smooth
      * surface, they must compute the normals themselves outside of
      * render.
    **/

    nverts = tri->nverts;
    
#ifdef DEBUG
    kprintf("\tthere are %d verts in this ktriangles_con prim.\n",nverts);
#endif

    verts   = tri->locs;      tri->locs = NULL;
    colors  = tri->cols;      tri->cols = NULL;
    normals = tri->norms;     tri->norms = NULL;
    tcoords = tri->tcoords;   tri->tcoords = NULL;

    tname = tri->texture;  tri->texture = NULL;
    tc_size = o->texture_coord_dim;


    /**
      * check for whether or not we're going to build a disjoint or
      * connected primitive.
    **/

    /* the layout attribute in the kgeom object refers to BOTH colors
       and normals.  that is, in the khoros data services model, if both
       colors and normals are defined, there are the same number of each,
       and the layout is either KPER_FACE or KPER_VERTEX. */
    
    if (layout == KPER_FACE) 
    {
	/**
	  * build a disjoint triangle primitive.
	**/

	build_disjoint_from_connected_perface(p,nverts,verts,colors,normals,
					      tcoords,layout,has_alpha);
    }
    else if (normals == NULL)	/* no normals */
    {
	build_disjoint_from_connected_nonorms(p,nverts,verts,colors,
					      layout,has_alpha);
    }
    else			/* layout is per-vertex */
    {
	/**
	  * build a connected triangle primitive.  we assume per-vertex
	  * everything.
	**/
	p->nverts = nverts;
	p->verts = (vertex_3d *)verts;

	if (colors)
	    p->ncolors = nverts;
	else
	    p->ncolors = 0;
	if (has_alpha)
	    p->rgba_color_list = (surface_colors_rgba *)colors;
	else
	    p->rgb_color_list = (surface_colors_rgb *)colors;


	/* this may be broken.. */
	p->nthings = nverts;
	p->prim_specifics.normals = (vertex_3d *)normals;
	
	p->type = KGEOM_TRIANGLES_CONNECTED;

    }

    p->clipfunc = (int(*)(void))clip_polygon_cheap;


    /* deal with texture coordinates */
    if ((tname == NULL) || (tname[0] == '\0'))
	p->has_texture = 0;
    else
    {
	p->has_texture = 1;
	strcpy(p->texture_transport_name,tname);

	/* do some sanity checking. */
	/* first, check to see if there are texture coords to go with the texture.. */
	if (tcoords == NULL)
	{
	    kerror("RENDER","render","the input connected triangles doo-hickey has a texture but no texture coordinates.  ignoring the texture and continuing...");
	    p->has_texture = 0;
	}

	/* next, check to see if the dimensionality of the texture agrees
	 with the dimensionality of the texture coords. */
	{
	    kgeom_primitive *tprim;
	    int tsize;
	    int tex_has_alpha;

	    tprim = get_first_khoros_texture(tname, &tex_has_alpha);

	    /* this is extra-cautious to make sure we have not only a
               texture, but a good texture with an appropriate size */
	    if (tprim)
	       switch (tprim->type)
	       {
		  case KGEOM_TEXTURE2D :
		     {
			kgeom_texture2D *t = (kgeom_texture2D *) tprim;
			if (t->width <= 1 || t->height <= 1)
			   tsize = 0, kerror("RENDER","render",
					     "2d texture must have width and height > 1");
			else
			   tsize = 2;
		     } break;
		  case KGEOM_TEXTURE3D :
		     {
			kgeom_texture3D *t = (kgeom_texture3D *) tprim;
			if (t->width <= 1 || t->height <= 1 || t->depth <= 1)
			   tsize = 0, kerror("RENDER","render",
					     "3d texture must have width, height, and depth > 1");
			else
			   tsize = 3;
		     } break;
		  default:
		     tsize = 0;
	       }
	    else
	       tsize = 0;

	    if (tprim)
	    {
	       if (tsize != tc_size)
	       {
		  kerror("RENDER","render","there is a mismatch between the dimensionality of the texture coordinates and the dimensionality of the input texture.  ignoring the texture and continuing....");
		  p->has_texture = 0;
	       }
	       else
	       {
		  p->tcoords_dimensions = tc_size;
		  p->tcoords = tcoords;

		  /* what we want is some way to avoid "loading" a texutre
		     multiple times.  */
		  if (p->the_texture == NULL) 
		     load_render_texture(tprim,&(p->the_texture),
					   tex_has_alpha,tname);
	       }
	    }
	    else /* no texture available : user already warned */
	       p->has_texture = 0;

	    kgeom_blast_primitive(tprim);
	}
    }
}

    

static void
rm_kspheres (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int nspheres;
    float *v=NULL,*c=NULL,*r=NULL;
    kgeom_spheres *spheres = (kgeom_spheres *) k;

    nspheres = spheres->nverts;
    v = spheres->locs;   spheres->locs = NULL;
    c = spheres->cols;   spheres->cols = NULL;
    r = spheres->radii;  spheres->radii = NULL;
    
    if (v != NULL)
    {
	p->nverts = nspheres;
	p->verts = (vertex_3d *)v;
    }
    else
    {
	p->nverts = 0;
	p->verts = NULL;
    }

    if (r != NULL)
    {
	p->prim_specifics.radii = r;
	p->nthings = nspheres;
    }
    else
    {
	p->prim_specifics.radii = NULL;
	p->nthings = 0;
    }

    if (c != NULL)
    {
	p->ncolors = nspheres;
	if (has_alpha)
	{
	    p->rgba_color_list = (surface_colors_rgba *)c;
	    p->rgb_color_list = NULL;
	}
	else
	{
	    p->rgba_color_list = NULL;
	    p->rgb_color_list = (surface_colors_rgb *)c;
	}
    }
    else
    {
	p->ncolors = 0;
    }

    p->clipfunc = (int(*)(void))clip_polygon_cheap;

    
}

static void
rm_kcylinders (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
}

static void
rm_kquadmesh (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    int usize,vsize,nverts,has_loc,has_color,has_normals,has_tc,tc_size;
    float *verts=NULL,*normals=NULL,*colors=NULL,*tc=NULL;
    vertex_3d *tn=NULL;
    char *tname=NULL;
    kgeom_quadmesh *qmesh = (kgeom_quadmesh *) k;

    usize = qmesh->width;
    vsize = qmesh->height;
    nverts = usize*vsize;

    verts   = qmesh->locs;    qmesh->locs    = NULL; 
    colors  = qmesh->cols;    qmesh->cols    = NULL; 
    normals = qmesh->norms;   qmesh->norms   = NULL; 
    tc      = qmesh->tcoords; qmesh->tcoords = NULL; 

    has_loc     = (verts != NULL);
    has_color   = (colors != NULL);
    has_normals = (normals != NULL);
    has_tc      = (tc != NULL);

    tc_size = o->texture_coord_dim;
    tname = qmesh->texture;

    if ((tname == NULL) || (tname[0] == '\0'))
	p->has_texture = 0;
    else
    {
	p->has_texture = 1;
	strcpy(p->texture_transport_name,tname);

	/* do some sanity checking. */
	/* first, check to see if there are texture coords to go with the texture.. */
	if (tc == NULL)
	{
	    kerror("RENDER","render","the input quadmesh has a texture but no texture coordinates.  ignoring the texture and continuing...");
	    p->has_texture = 0;
	}
	/* next, check to see if the dimensionality of the texture agrees
	 with the dimensionality of the texture coords. */
	{
	    kgeom_primitive *tprim;
	    int tsize, tex_has_alpha;

	    tprim = get_first_khoros_texture(tname, &tex_has_alpha);

	    /* this is extra-cautious to make sure we have not only a
	       texture, but a good texture with an appropriate size */
	    if (tprim)
	       switch (tprim->type)
	       {
		  case KGEOM_TEXTURE2D :
		     {
			kgeom_texture2D *t = (kgeom_texture2D *) tprim;
			if (t->width <= 1 || t->height <= 1)
			   tsize = 0, kerror("RENDER","render",
					     "2d texture must have width and height > 1");
			else
			   tsize = 2;
		     } break;
		  case KGEOM_TEXTURE3D :
		     {
			kgeom_texture3D *t = (kgeom_texture3D *) tprim;
			if (t->width <= 1 || t->height <= 1 || t->depth <= 1)
			   tsize = 0, kerror("RENDER","render",
					     "3d texture must have width, height, and depth > 1");
			else
			   tsize = 3;
		     } break;
		  default:
		     tsize = 0;
	       }
	    else
	       tsize = 0;

	    if (tprim)
	    {
	       if (tsize != tc_size)
	       {
		  kerror("RENDER","render","there is a mismatch between the dimensionality of the texture coordinates and the dimensionality of the input texture.  ignoring the texture and continuing....");
		  p->has_texture = 0;
	       }
	       else
	       {
		  p->tcoords_dimensions = tc_size;
		  p->tcoords = tc;
		  load_render_texture(tprim,&(p->the_texture),
					tex_has_alpha,tname);
	       }
	    }
	    else /* no texture available : user already warned */
	       p->has_texture = 0;
	    
	    kgeom_blast_primitive(tprim);
	}
    }

    p->type = KGEOM_QUADMESH;
    if (verts == NULL) /* strange object */
    {
	p->nverts = 0;
	p->mesh_dims[0] = 0;
	p->mesh_dims[1] = 0;
	p->mesh_dims[2] = 0;
    }
    else
    {
	p->nverts = usize*vsize;
	p->mesh_dims[0] = usize;
	p->mesh_dims[1] = vsize;
	p->mesh_dims[2] = 1; /* for sanity reasons */
	p->nverts = nverts;
    }

    if (!has_normals)  /* then compute them */
    {
	/**
	  * make a call to compute a bunch of per-vertex
	  * normals for the vertices.  we'll compute them once
	  * here and keep them around rather than recomputing
	  * them each time we rerender.
        **/
/*	lcompute_normals(verts,&normals,usize,vsize); */
	lcompute_quadmesh_vertex_normals((vertex_3d *)verts,&tn,
					 usize,vsize);
	normals = (float *)tn;
    }
    
    p->verts = (vertex_3d *)verts;

    if (colors == NULL)
	p->ncolors = 0;
    else if (layout == KPER_VERTEX)
	p->ncolors = nverts;
    else /*  (layout == KPER_FACE) */
	p->ncolors = (usize-1)*(vsize-1);

    if (has_alpha)
	p->rgba_color_list = (surface_colors_rgba *)colors;
    else
	p->rgb_color_list = (surface_colors_rgb *)colors;

    if (normals == NULL)
	p->nthings = 0;
    else if (layout == KPER_VERTEX)
	p->nthings = nverts;
    else /* layout == KPER_FACE */
	p->nthings = (usize-1)*(vsize-1);
    p->prim_specifics.normals = (vertex_3d *)normals;

    p->clipfunc = (int(*)(void))clip_polygon_cheap;
    /* blow off texture coordinates until things get straightened out */

}

static void
rm_koctmesh (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    /**
      * right now, we are handling ONLY curvilinear meshes.  this needs
      * to be changed to handle UNIFORM and RECTILINEAR meshes.
    **/
    
    int usize,vsize,wsize,nverts,has_loc,has_color,has_normals;
    float *verts=NULL,*normals=NULL,*texture=NULL;
    int tc_size;
    char *tname=NULL;
    unsigned char *colors=NULL;
    kgeom_octmesh *omesh = (kgeom_octmesh *) k;
    

    /* we're going to assume that the color info in the octmesh is
     unsigned bytes.  wes, 29 oct 96 */

    usize = omesh->width;
    vsize = omesh->height;
    wsize = omesh->depth;
    nverts = usize*vsize*wsize;

    verts   = omesh->locs;    omesh->locs    = NULL; 
    colors  = omesh->cols;    omesh->cols    = NULL;
    normals = omesh->norms;   omesh->norms   = NULL; 

    has_loc     = (verts != NULL);
    has_color   = (colors != NULL);
    has_normals = (normals != NULL);

    tc_size = o->texture_coord_dim;
    tname = omesh->texture; omesh->texture = NULL;

    p->type = KGEOM_OCTMESH;
    if (verts == NULL) /* strange object */
    {
	p->nverts = 0;
	p->mesh_dims[0] = 0;
	p->mesh_dims[1] = 0;
	p->mesh_dims[2] = 0;
    }
    else
    {
	p->nverts = usize*vsize*wsize;
	p->mesh_dims[0] = usize;
	p->mesh_dims[1] = vsize;
	p->mesh_dims[2] = wsize;
	p->nverts = nverts;
    }

    /* compute the octmesh radii.  each "radius" is an approximation of
     the radius of a circle formed by the projection of the voxel vertices
     into some plane.  the radius is used later on in the splat-rendering
     code to compute the size of the voxel splat footprint. */

    if (verts != NULL)
    {
	p->oct_mesh_radii = (float *)kmalloc(sizeof(float)*nverts);
	compute_octmesh_radii(p->mesh_dims,p->oct_mesh_radii,verts);
    }
    else
	p->oct_mesh_radii = NULL;

    
    if (!has_normals)  /* then compute them */
    {
#if 0
	/**
	  * make a call to compute a bunch of per-vertex
	  * normals for the vertices.  we'll compute them once
	  * here and keep them around rather than recomputing
	  * them each time we rerender.
        **/
/*	lcompute_normals(verts,&normals,usize,vsize); */
	lcompute_quadmesh_vertex_normals((vertex_3d *)verts,&normals,
					 usize,vsize);
#endif
    }
    
    p->verts = (vertex_3d *)verts;

    if (colors == NULL)
	p->ncolors = 0;
    else if (layout == KPER_VERTEX)
	p->ncolors = nverts;
    else /*  (layout == KPER_FACE) */
	p->ncolors = (usize-1)*(vsize-1)*(wsize-1);

    p->byte_rgba_color_list = (byte_rgba_colors *)colors;
#if 0
    if (has_alpha)
	p->rgba_color_list = (surface_colors_rgba *)colors;
    else
	p->rgb_color_list = (surface_colors_rgb *)colors;
#endif

    if (normals == NULL)
	p->nthings = 0;
    else if (layout == KPER_VERTEX)
	p->nthings = nverts;
    else /* layout == KPER_FACE */
	p->nthings = (usize-1)*(vsize-1);
    p->prim_specifics.normals = (vertex_3d *)normals;

    p->clipfunc = (int(*)(void))clip_nullfunc;
    /* blow off texture coordinates until things get straightened out */
}

static void
rm_kdirpoints (render_prim *p, kgeom_object *o, kgeom_primitive *k,int layout, int has_alpha)
{
    kfprintf(kstderr," directed point primitives not yet supported.\n");
}


static void
rm_kpolyhedra_face (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
}

static void
rm_kpolyhedra_verts (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
}


static void
rm_ktexture2d (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpha)
{
    kfprintf(kstderr," texture 2d primitives not yet supported.\n");
}

static void
rm_ktexture3d (render_prim *p, kgeom_object *o, kgeom_primitive *k, int layout, int has_alpa)
{
    kfprintf(kstderr," texture 3d primitives not yet supported.\n");
}

static void
copy_kattribs_to_renderobj(kgeom_object *k,renderobj *p, int has_alpha)
{
    double d;
    int status,i,j,offset;
    float *f,work[4];
    float *bmin,*bmax;

    f = work;
    /**
      * for any object attribute which isn't defined, we use a
      * default or otherwise try and do something sensible.
    **/

    /* first, do the surface attributes */

    /* ignore until we have support for ambient_color 
    p->attribs.amb = k->ambient_color;
    */
    p->attribs.amb = KGEOM_AMBIENT_DEFAULT;

    /* ignore until we have support for diffuse_color 
    p->attribs.diff = k->diffuse_color;
    */
    p->attribs.diff = KGEOM_DIFFUSE_DEFAULT;

    /* ignore until we have support for specular_color 
    p->specular.amb = k->specular_color;
    */
    p->attribs.spec = KGEOM_SPECULAR_COEFFICIENT_DEFAULT;

    p->attribs.spec_exp   = k->specular_exponent;

    /* metalness is going away */
    p->attribs.spec_metal = KGEOM_METAL_DEFAULT;

    /* now, do the surface color */
    if (k->has_color)
    {
	p->colors.r = k->color[0];
	p->colors.g = k->color[1];
	p->colors.b = k->color[2];

	if (has_alpha)
	    p->colors.a = k->opacity;
	else
	    p->colors.a = 1.0;
    }

    /* do the bounding box */
    bmin = bmax = NULL;
    if (k->has_bounding_box)
       bmin = k->min, bmax = k->max;

    if (!k->has_bounding_box)
    {
	/* no box defined, compute one. */
	if (!(p->flags & IGNOREBOX))
	    compute_render_bbox(p);
    }
    else
    {
	for (i=0;i<3;i++)
	{
	    p->bmin.v[i] = (float)bmin[i];
	    p->bmax.v[i] = (float)bmax[i];
	}
    }

    /* do the center point */
    if (!k->has_center) 
    {
	/* no center point, compute one. */
	for (i=0;i<3;i++)
	    p->center.v[i] = (p->bmax.v[i] + p->bmin.v[i]) * 0.5;
    }
    else
    {
       for (i=0;i<3;i++)
	  p->center.v[i] = k->center[i];
    }


    /* get the transformation matrix. */
    if (!k->has_matrix)
	identity_4x4(&(p->xfrm));
    else
    {
        bmin = k->matrix;
	offset=0;
	for (j=0;j<4;j++)
	    for (i=0;i<4;i++)
		p->xfrm.m[i][j] = (KGEOM_VERTEX_TYPE)bmin[offset++];
    }

    /* get the display status */
    p->do_display = k->visible;

    /* determine if any part of this object is transparent */
    if (k->has_alpha)
    {
	/* need to scan through the list of colors */
	determine_transparency(p);
    }
    else
	p->is_transparent = FALSE;
       
}

static void
compute_render_bbox(renderobj *p)
{
    float *mins,*maxs;
    float *f;
    int i;

    mins = (float *)kmalloc(sizeof(float)*p->nprims*3);
    maxs = (float *)kmalloc(sizeof(float)*p->nprims*3);

    for (i=0;i<p->nprims;i++)
    {
	f = (float *)&(p->prims[i].verts[0]);
	vect_min(f,p->prims[i].nverts*3,0,3,mins+i);
	vect_min(f+1,p->prims[i].nverts*3,0,3,mins+i+p->nprims);
	vect_min(f+2,p->prims[i].nverts*3,0,3,mins+i+(p->nprims*2));
	
	vect_max(f,p->prims[i].nverts*3,0,3,maxs+i);
	vect_max(f+1,p->prims[i].nverts*3,0,3,maxs+i+p->nprims);
	vect_max(f+2,p->prims[i].nverts*3,0,3,maxs+i+(p->nprims*2));
    }

    for (i=0;i<3;i++)
	vect_min(mins,p->nprims,i*p->nprims,1,&(p->bmin.v[i]));
    
    for (i=0;i<3;i++)
	vect_max(maxs,p->nprims,i*p->nprims,1,&(p->bmax.v[i]));

    kfree(mins);
    kfree(maxs);
}

static void
determine_transparency(renderobj *p)
{
    int has_prim_color,i,j;
    int has_some_transparency;
    surface_colors_rgba *c;
    byte_rgba_colors *b;


    /* first, look at the global color information for non-zero alpha. */
    if (p->colors.a != 1.)
    {
	p->is_transparent = TRUE;
	return;
    }
    else
	p->is_transparent = FALSE; /* so far, anyway. */

    
    /* if there is no color information at the primitive level,
       then just look at the global object color vector. */
    for (i=0;i<p->nprims;i++)
    {
	if (p->prims[i].ncolors != 0)
	    break;
    }
    if (i < p->nprims)
	has_prim_color = TRUE;
    else
	has_prim_color = FALSE;

    if (has_prim_color == FALSE)
    {
	if (p->colors.a != 1.)
	    p->is_transparent = TRUE;
	else
	    p->is_transparent = FALSE;
    }
    else
    {
	has_some_transparency = FALSE;
	for (i=0;i<p->nprims;i++)
	{
	    if ((c = p->prims[i].rgba_color_list) != NULL)
/*	    if (c != null)*/
	    {
		for (j=0;j<p->prims[i].ncolors;j++,c++)
		    if (c->a != 1.)
		    {
			has_some_transparency = TRUE;
			break;
		    }
	    }
	    else if ((b = p->prims[i].byte_rgba_color_list) != NULL)
	    {
		for (j=0;j<p->prims[i].ncolors;j++,b++)
		    if (b->a != 255)
		    {
			has_some_transparency = TRUE;
			break;
		    }
	    }
	    if (has_some_transparency == TRUE)
		break;
	}
	if (has_some_transparency == TRUE)
	    p->is_transparent = TRUE;
	else
	    p->is_transparent = FALSE;
    }
}

static void
determine_default_shader(renderobj *r)
{
    /* taken from render_objs.h: */
    /**
      * since a geom object is comprised of several primitives,
      * the notion of an object level shader applying to heterogeneous
      * primitives is slippery.  for now, we are assuming that
      * all the primitive types are homogeneous (ie, all volumetric,
      * all surface based, or all wireframe) and that all of the
      * primitives will use the object-level shader.
    **/

    if (r->nprims == 0)
	r->shader = KSHADER_INHERIT;
    else
    {

	/* all objects who's verts_nspace parameter is set to 2 get
	 a no-light shader by default.  usually, the only things that
	 will have verts_nspace == 2 are text and other annotation things
	 like color legends. */

	if (r->verts_nspace == 2)
	    r->shader = KSHADER_NO_LIGHT;
	else
	{
	    /* just look at the first primitive - assume they're
	       all homogeneous */
	
	    switch (r->prims[0].type)
	    {
	    case KGEOM_POLYLINE_DISJOINT:
	    case KGEOM_POLYLINE_CONNECTED:
		r->shader = KSHADER_NO_LIGHT;
		break;
            
	    case KGEOM_TRIANGLES_DISJOINT:
	    case KGEOM_TRIANGLES_CONNECTED:
		if (r->prims[0].prim_specifics.normals != NULL)
		    r->shader = KSHADER_GOUROUD;
		else
		    r->shader = KSHADER_FLAT;
		break;
	    case KGEOM_QUADMESH:
		    r->shader = KSHADER_GOUROUD;
		break;
		
	    case KGEOM_CYLINDERS:
	    case KGEOM_DIRECTED_POINTS:
		r->shader = KSHADER_FLAT;
		break;

	    case KGEOM_SPHERES:
		r->shader = KSHADER_FLAT;
		/*  r->shader = KSHADER_GOUROUD;  slow.. :(  // Steve K */
		break;

	    case KGEOM_OCTMESH:
		/*          r->shader = KSHADER_SPLAT_HOMOGENEOUS; */
		r->shader = KSHADER_VOXEL_DOT; /* this one is fastest (in software) */
		/* r->shader = KSHADER_VOXEL_3D_TEXTURE;  */
		break;

#if 0
	    case KGEOM_TEXTURE2D:
	    case KGEOM_TEXTURE3D:
#endif
	    default:
		r->shader = KSHADER_INHERIT;
		break;
	    }
	}
    }
}

int
find_renderobj(char *search_for_name, renderobj **r)
{
    int status;

    status = search_for_named_object(search_for_name,r,&root);
    return(status);
}

static int
search_for_named_object(char *name, renderobj **rval, renderobj *r)
{
    int i,status;

    if (kstrcmp(name,r->name) == 0)
    {
	*rval = r;
	return(CHILL);
    }
    else
    {
	for (i=0;i<r->nchildren;i++)
	{
	    status = search_for_named_object(name,rval,r->children[i]);
	    if (status == CHILL)
		return(CHILL);
	}
    }
    return(WHACKED);
}

renderobj *get_current_object()
{
    return(current_object);
}

void
get_parent_xfrm(matrix4x4 *m)
{
    matrix_4x4copy(m,&current_parent_xfrm);
}

void
get_parent_xfrm_inverse(matrix4x4 *m)
{
    matrix_4x4copy(m,&current_parent_xfrm_inverse);
}


static void
build_parent_xfrm(renderobj *r,
		  matrix4x4 *m)
{
    renderobj *p;
    matrix4x4 t1,comp;
    
    if (r->nparents > 1)  /* deep shit */
    {
	kerror("RENDER","render","this object has more than one parent. you cannot transform this object. ");
	return;
    }
    else if (r->nparents == 0) /* top of tree */
    {
	return;
    }
    else
    {
	p = r->parents[0];
	/* validity check on p ?? */

	/* construct forward transformation for p. this consists of
	   the following matrix concatenations:
	       -Center * Scale * Rotate * Center * Translate
	   this composite matrix will be postmultiplied to the one
	   passed in to this routine.
	 */
	identity_4x4(&t1);
	identity_4x4(&comp);

	t1.m[3][0] = -1.*p->center.v[0];
	t1.m[3][1] = -1.*p->center.v[1];
	t1.m[3][2] = -1.*p->center.v[2];

	/* for now, scaling and rotation are collapsed into a single
	   matrix in the object.  */
	
	mmul_4x4(&t1,&(p->xfrm),&comp);

	t1.m[3][0] = p->center.v[0] + p->translate_vect.v[0];
	t1.m[3][1] = p->center.v[1] + p->translate_vect.v[1];
	t1.m[3][2] = p->center.v[2] + p->translate_vect.v[2];

	mmul_4x4(&comp,&t1,&comp);

	
	build_parent_xfrm(p,&comp);
	mmul_4x4(m,&comp,m);
    }
}

int
set_current_object_lite(renderobj *r)
{
  /* like set_current_object, but we only update the pointer and don't
   update the GUI. */
    current_object = r;
    return(CHILL);
}



int
set_current_object(renderobj *r)
{
    matrix4x4 m;
    
    /* check to make sure it exists?? */
    current_object = r;

    /**
      * also, set the current transformation matrix and center point.
      * this is the spot where we'll catch impossible conditions, such
      * as attempting to set the cumulative transformation (bottom-up)
      * for objects with more than one parent.
      *
      * we store, along with a pointer to the current object, a single
      * transformation matrix which represents that transformation
      * of the parent.  in other words, to compute the position in view
      * space, we need the transformation of the current object, plus
      * the transformation of the parent.  since the reference object may
      * be at an arbitrary place in the object tree, we need to backwards-
      * compute a transformation, representing the transformation for
      * the parent.  the viewspace vector would computed by transforming
      * and object's vertex value through its local transformation matrix,
      * then through that of it's parent (postmultiplication).
      *
      * NOTE: this backwards transformation should not be confused with
      * the inverse of the parent's transform.
      *
      * the static variable current_parent_xfrm holds the above described
      * transformation matrix, and is computed by the routine
      * build_parent_xfrm().
    **/
       
    identity_4x4(&m);

    build_parent_xfrm(r,&m);
    matrix_4x4copy(&current_parent_xfrm,&m);
    matrix_inverse(&current_parent_xfrm,&current_parent_xfrm_inverse);

    /* update the transformation UI */
    update_transformation_pane(r);
    update_object_pane(r);

    return(CHILL);
}

int
set_object_shader (renderobj *r, int shader)
{
    /* range check ?? */
    r->shader = shader;
    return(CHILL);
}

int
get_object_shader (renderobj *r, int *shader)
{
    *shader = r->shader;
    return(CHILL);
}

int
set_object_display_status (renderobj *r, int value)
{
    /** range check ?? */
    r->do_display = value;
    return(CHILL);
}

int
get_object_display_status (renderobj *r, int *value)
{
    *value = r->do_display;
    return(CHILL);
}

int
get_object_rgba (renderobj *o, float *r, float *g, float *b, float *a)
{
    *r = o->colors.r;
    *g = o->colors.g;
    *b = o->colors.b;
    *a = o->colors.a;
    return(CHILL);
}

int
set_object_rgba (renderobj *o, float r, float g, float b, float a)
{
    /* range check on input ?? */
    o->colors.r = r;
    o->colors.g = g;
    o->colors.b = b;
    o->colors.a = a;
    determine_transparency(o);
    return(CHILL);
}

int
set_object_amb (renderobj *r, float f)
{
    /* range check input?? */
    r->attribs.amb = f;
    return(CHILL);
}

int
get_object_amb (renderobj *r, float *f)
{
    *f = r->attribs.amb;
    return(CHILL);
}

int
set_object_diff (renderobj *r, float f)
{
    /* check input ?? */
    r->attribs.diff = f;
    return(CHILL);
}

int
get_object_diff (renderobj *r, float *f)
{
    *f = r->attribs.diff;
    return(CHILL);
}

int
set_object_spec (renderobj *r, float f)
{
    /* check input ?? */
    r->attribs.spec = f;
    return(CHILL);
}

int
get_object_spec (renderobj *r , float *f)
{
    /* check input ?? */
    *f = r->attribs.spec;
    return(CHILL);
}

int
set_object_spec_exp (renderobj *r, float f)
{
    /* check input ?? */
    r->attribs.spec_exp = f;
    return(CHILL);
}

int
get_object_spec_exp (renderobj *r, float *f)
{
    *f = r->attribs.spec_exp;
    return(CHILL);
}

int
set_object_metal (renderobj *r, float f)
{
    /* check input ? */
    r->attribs.spec_metal = f;
    return(CHILL);
}

int
get_object_metal (renderobj *r, float *f)
{
    *f = r->attribs.spec_metal;
    return(CHILL);
}

renderobj *
get_root_object()
{
    return(&root);
}

static void
union_bounding_boxes(renderobj *a,
		     renderobj *b,
		     vertex_3d *bmin,
		     vertex_3d *bmax)
{
    vertex_3d l_amin,l_amax,l_bmin,l_bmax;
    
    /**
      * if the boxes of either object are {0.,0.,0},{0.,0.,0.},
      * we will assume that the box hasn't been defined.
      *
      * the box in object "a" will not be transformed (matrix or translate
      * vector).  the box in object "b" will be transformed.
    **/

    int a_undefined=0,b_undefined=0,i;
    
    VCOPY(&l_amin,&(a->bmin));
    VCOPY(&l_amax,&(a->bmax));
    VCOPY(&l_bmin,&(b->bmin));
    VCOPY(&l_bmax,&(b->bmax));

    if ((l_amin.v[0] == 0.) &&
	(l_amin.v[1] == 0.) &&
	(l_amin.v[2] == 0.) &&
	(l_amax.v[0] == 0.) &&
	(l_amax.v[1] == 0.) &&
	(l_amax.v[2] == 0.))
	a_undefined = 1;

    if ((l_bmin.v[0] == 0.) &&
	(l_bmin.v[1] == 0.) &&
	(l_bmin.v[2] == 0.) &&
	(l_bmax.v[0] == 0.) &&
	(l_bmax.v[1] == 0.) &&
	(l_bmax.v[2] == 0.))
	b_undefined = 1;
    
    /* transform "b" points back to the center of the object. */

    npoint_diff((KGEOM_VERTEX_TYPE *)&(l_bmin.v[0]),
		(KGEOM_VERTEX_TYPE *)&(b->center.v[0]),
		(KGEOM_VERTEX_TYPE *)&(l_bmin.v[0]),1);

    npoint_diff((KGEOM_VERTEX_TYPE *)&(l_bmax.v[0]),
		(KGEOM_VERTEX_TYPE *)&(b->center.v[0]),
		(KGEOM_VERTEX_TYPE *)&(l_bmax.v[0]),1);


    /* apply these points through the transformation matrix */
    npoint_xfrm(&(l_bmin.v[0]),&(b->xfrm),&(l_bmin.v[0]),1);
    npoint_xfrm(&(l_bmax.v[0]),&(b->xfrm),&(l_bmax.v[0]),1);

    /* add back the center, plus the translation vector */

    for (i=0;i<3;i++)
    {
	l_bmin.v[i] = l_bmin.v[i] + b->translate_vect.v[i] + b->center.v[i];
	l_bmax.v[i] = l_bmax.v[i] + b->translate_vect.v[i] + b->center.v[i];
/***
	l_bmin.v[i] += b->translate_vect.v[i] + b->center.v[i];
	l_bmax.v[i] += b->translate_vect.v[i] + b->center.v[i];
****/
    }

    if (a_undefined && b_undefined) /* compute the boxes. */
    {
	/* 2 May 1997 - the code needs to be written which will compute
	 an object bounding box.  at present, there is some ambiguity
	 in how to address the problem of vertices from heterogeneous
	 dimensionality...ie, both 2d and 3d verts.  for now, what i'm
	 going to do is add a switch that will shut off the following
	 message after it's displayed once.
	*/
	if (squelch_the_squawk == 0)
	{
	    kerror("GEOMETRY","render",
		   " null bounding boxes encountered in the input objects.  it's going to be hard to automatically compute a viewpoint. this message will not be repeated.  weird things may happen. the solution is to make sure that all of your input geometry objects have a bounding box.");
	    squelch_the_squawk = 1;
	}
    }
    else if (a_undefined && !b_undefined) 
    {
	VCOPY(bmin,&(l_bmin));
	VCOPY(bmax,&(l_bmax));
    }
    else if (b_undefined && !a_undefined)
    {
	VCOPY(bmin,&(l_amin));
	VCOPY(bmax,&(l_amax));
    }
    else
    {
	for (i=0;i<3;i++)
        {
	    bmin->v[i] = MIN(l_amin.v[i],l_bmin.v[i]);
	    bmin->v[i] = MIN(bmin->v[i],l_bmax.v[i]);
	    bmin->v[i] = MIN(bmin->v[i],l_amax.v[i]);
	    
	    bmax->v[i] = MAX(l_amin.v[i],l_bmin.v[i]);
	    bmax->v[i] = MAX(bmax->v[i],l_amax.v[i]);
	    bmax->v[i] = MAX(bmax->v[i],l_bmax.v[i]);
	}
    }
}

static void
local_sync_root_box(renderobj *r)
{
    int i;
    renderobj *c;

    /* this routine is incomplete.  it only gets the first generation of
       root descendants */

    for (i=0;i<3;i++)
    {
	r->bmin.v[i] = r->bmax.v[i] = 0.;
    }
    
    for (i=0;i<r->nchildren;i++)
    {
	c = r->children[i];
	union_bounding_boxes(r,c,&(r->bmin),&(r->bmax));
    }

#if 0
    /**
      * recompute the center point so that rotations happen
      * "nicely".  this might not be the best way, and is a hack,
      * oh well.
   **/

    for (i=0;i<3;i++)
	r->center.v[i] = r->bmin.v[i] + (r->bmax.v[i] - r->bmin.v[i]) * 0.5;
#endif
}

/**
  * here, we're going to tranverse downward through the object
  * tree with the intent of recomputing the roots min/max box.
  * this is necessary to do since the user may move objects around,
  * then expect the root object to encompass them all in subsequent
  * transformations.
**/
void
sync_root_box()
{

    local_sync_root_box(&root);
    
}

void
get_bounding_box(renderobj *r,
		 vertex_3d *bmin,
		 vertex_3d *bmax)
{
    /**
      * this routine will return the bounding box for the object "r",
      * subject to whatever local transformations are applied to "r",
      * such as scaling, rotations and translations.
    **/

    vertex_3d box[2];
    int i,j;

    VCOPY(box,&(r->bmin));
    VCOPY(box+1,&(r->bmax));

    /* translate the box back to the center point of the object. */

    for (i=0;i<2;i++)
    {
	for (j=0;j<3;j++)
	    box[i].v[j] -= r->center.v[j];
    }

    /* apply transformation matrix */
    npoint_xfrm((KGEOM_VERTEX_TYPE *)&(box[0].v[0]),
		&(r->xfrm),
		(KGEOM_VERTEX_TYPE *)&(box[0].v[0]), 2);

    /* now, add back in the center point and the object translational comp. */
    
    for (i=0;i<2;i++)
    {
	for (j=0;j<3;j++)
	    box[i].v[j] += r->center.v[j] + r->translate_vect.v[j];
    }

    VCOPY(bmin,box);
    VCOPY(bmax,box+1);
}

void
geom_xform_points(matrix4x4 *child_xfrm,
		  vertex_3d *child_tv,
		  vertex_3d *child_center,
		  vertex_3d *in,
		  vertex_3d *out,
		  matrix4x4 *parent_xfrm,
		  vertex_3d *parent_tv,
		  int n)
/**
  * this routine is used to apply geometric transformations to lists of points.
  * it is called from the rendering routines in render.c, as well as
  * the xor bounding box code in render_mouse.c.
  *
  * the order of operations is as follows:
  * 1. translate the points back to the local origin
  * 2. apply scaling then rotational transformations
  * 3. translate by -1 * local_origin + object translation
  * 4. apply the transformation specified by the parent.
  *    this will be represented by the composite matrix:
  *        [Scale*Rotate + Translate]
  *
  * 7 july 95, wes.  collapsed all operations into a single matrix op.
**/
{
    matrix4x4 t,composite;

    identity_4x4(&composite);
    
    /* translate to local origin */
    identity_4x4(&t);
    t.m[3][0] = -1. * child_center->v[0];
    t.m[3][1] = -1. * child_center->v[1];
    t.m[3][2] = -1. * child_center->v[2];

    /* apply child's transformation matrix */
    mmul_4x4(&t,child_xfrm,&composite);
    
    /* translate the origin back to the childs center point, and
       add child's translation vector */

    identity_4x4(&t);
    t.m[3][0] = child_center->v[0];
    t.m[3][1] = child_center->v[1];
    t.m[3][2] = child_center->v[2];

    if (child_tv != NULL)
    {
	t.m[3][0] += child_tv->v[0];
	t.m[3][1] += child_tv->v[1];
	t.m[3][2] += child_tv->v[2];
    }

    mmul_4x4(&composite,&t,&composite);

    /* premultiply the child's matrix with the parent's matrix.  */

    if (parent_xfrm != NULL)
	mmul_4x4(&composite,parent_xfrm,&composite); 

    /* add parent's translation vector after applying parent's xformation
     matrix. */
    
    identity_4x4(&t);
    if (parent_tv != NULL)
    {
	t.m[3][0] += parent_tv->v[0];
	t.m[3][1] += parent_tv->v[1];
	t.m[3][2] += parent_tv->v[2];
    }
    mmul_4x4(&composite,&t,&composite);

    /* blast all the points thru the single composite matrix. */
    npoint_xfrm(&(in->v[0]),&composite,&(out->v[0]),n); 
}
void
geom_xform_normals(matrix4x4 *child_xfrm,
		   vertex_3d *child_tv,
		   vertex_3d *child_center,
		   vertex_3d *in,
		   vertex_3d *out,
		   matrix4x4 *parent_xfrm,
		   vertex_3d *parent_tv,
		   int n)
/**
  * this routine is used to apply geometric transformations to lists of norms.
  * it is called from the rendering routines in render.c, as well as
  * the xor bounding box code in render_mouse.c.
  *
  * the order of operations is as follows:
  * 1. translate the points back to the local origin (NOT)
  * 2. apply scaling then rotational transformations
  * 3. translate by -1 * local_origin + object translation (NOT)
  * 4. apply the transformation specified by the parent.
  *    this will be represented by the composite matrix:
  *        [Scale*Rotate + Translate(NOT)]
  *
  * 7 july 95, wes.  collapsed all operations into a single matrix op.
**/
{
    matrix4x4 t,composite;

    identity_4x4(&composite);
    
    /* translate to local origin */
    identity_4x4(&t);

    t.m[3][0] = -1. * child_center->v[0];
    t.m[3][1] = -1. * child_center->v[1];
    t.m[3][2] = -1. * child_center->v[2];

    /* apply child's transformation matrix */

    mmul_4x4(&t,child_xfrm,&composite);
    
    /* translate the origin back to the childs center point, and
       add child's translation vector */

    identity_4x4(&t);

    t.m[3][0] = child_center->v[0];
    t.m[3][1] = child_center->v[1];
    t.m[3][2] = child_center->v[2];

    if (child_tv != NULL)
    {
	t.m[3][0] += child_tv->v[0];
	t.m[3][1] += child_tv->v[1];
	t.m[3][2] += child_tv->v[2];
    }

    mmul_4x4(&composite,&t,&composite);

    /* premultiply the child's matrix with the parent's matrix.  */

    if (parent_xfrm != NULL)
	mmul_4x4(&composite,parent_xfrm,&composite); 

    /* add parent's translation vector after applying parent's xformation
     matrix. */
    
    identity_4x4(&t);

    if (parent_tv != NULL)
    {
	t.m[3][0] += parent_tv->v[0];
	t.m[3][1] += parent_tv->v[1];
	t.m[3][2] += parent_tv->v[2];
    }

    mmul_4x4(&composite,&t,&composite);

    /* blast all the points thru the single composite matrix. */
    nnormal_xfrm(in,&composite,out,n);
}

/**
  * need to move this routine to the geometry library when finished.
  * 8/11/94  wes
**/
int
lcompute_normals(float *verts,
		 float **normals,
		 int usize,
		 int vsize)
{
    int uv,vv;
    int i,j,k;
    vertex_3d *n;
    vertex_3d s1,s2;
    vertex_3d *v;
    double mag;

    /**
      * here, we compute normals on a per-vertex basis.  if per-facet
      * normals are desired, they could be derived from those
      * computed here, or from the raw vertices directly
    **/

    uv = usize;
    vv = vsize;

    /* alloc space for one normal per vertex */
    n = (vertex_3d *)kmalloc(sizeof(vertex_3d)*uv*vv);
    
    v = (vertex_3d *)verts;

    /* do general case first. */
    for (j=0;j<vv;j++)
    {
	for (i=0;i<uv;i++)
	{
	    if ((i != 0) && (i != uv-1) && (j != 0) && (j != vv-1))
	    {
		for (k=0;k<3;k++)
		{
		    s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
		    s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
		}
		vertex_unit(&s1,&mag);
		vertex_unit(&s2,&mag);
		vertex_cross(&s1,&s2,(n+i+j*uv));
	    }
	}
    }
    
    /* do the j=0 row. */
    j=0;
    for (i=0;i<uv;i++)
    {
	if ((i != 0) && (i != uv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+1)->v[k] - (v+i-1)->v[k];
		s2.v[k] = (v+i+uv)->v[k] - (v+i)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i));
	}
    }
    
    /* do the j=vv-1 row. */
    j=vv-1;
    for (i=0;i<uv;i++)
    {
	if ((i != 0) && (i != uv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
		s2.v[k] = (v+i+j*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i+j*uv));
	}
    }
    
    /* do the i=0 column */
    i=0;
    for (j=0;j<vv;j++)
    {
	if ((j != 0) && (j != vv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i+j*uv)->v[k];
		s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i+j*uv));
	}
    }
    
    /* do the i=uv column */
    i=uv-1;
    for (j=0;j<vv;j++)
    {
	if ((j != 0) && (j != vv-1))
	{
	    for (k=0;k<3;k++)
	    {
		s1.v[k] = (v+i+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
		s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
	    }
	    vertex_unit(&s1,&mag);
	    vertex_unit(&s2,&mag);
	    vertex_cross(&s1,&s2,(n+i+j*uv));
	}
    }
    /* now, do the four corners. these normals are computed as the
      cross product of the segments adjoining the 2 nearest points. */
    
    /* i==0,j==0 */
    i=0;
    j=0;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i+j*uv)->v[k];
	s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+j*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    

    /* i=uv-1,j=0 */
    i=uv-1;
    j=0;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
	s2.v[k] = (v+i+(j+1)*uv)->v[k] - (v+i+j*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    
    /* i=0,j=vv-1 */
    i=0;
    j=vv-1;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+1+j*uv)->v[k] - (v+i+j*uv)->v[k];
	s2.v[k] = (v+i+j*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));
    
    /* i=uv-1,j=vv-1 */
    i=uv-1;
    j=vv-1;
    for (k=0;k<3;k++)
    {
	s1.v[k] = (v+i+j*uv)->v[k] - (v+i-1+j*uv)->v[k];
	s2.v[k] = (v+i+j*uv)->v[k] - (v+i+(j-1)*uv)->v[k];
    }
    vertex_unit(&s1,&mag);
    vertex_unit(&s2,&mag);
    vertex_cross(&s1,&s2,(n+i+j*uv));

    *normals = (float *)n;
    return(CHILL);
}

/**
  *
  * stuff to address object input.
  *
  * the notion is that there are up to N input for objects (N currently
  * #define'd to 4 (august 94).
  *
  * we keep static pointers around that point to the render objs,
  * thus establishing logical links between the N inputs and the
  * renderobjs in the object tree.  this will facilitate changing
  * the objects if either the input file changes in content or name.
  *
  * note that each of the input objects is a first-generation child
  * of the root object.
  *
**/
#if 0
static renderobj *input_objects[NUMBER_RENDER_INPUT_OBJECT_PORTS]=
{
    NULL,
    NULL,
    NULL,
    NULL
};
#endif
void replace_input_object(int n,
			  renderobj *r)
/**
  * n - index into input_objects[] array, ie, which input port is
  *     being tweaked.
  * r - the new renderobj
  *
  * 15 nov 1996 - add code so that surface attributes will not be
  * overwritten by the newly read object.  this is so that the user
  * can change the color, etc of an object, and when a new one is
  * read in, all of their changes won't be lost.
**/
{
    vertex_3d union_box_min,union_box_max;
    int i;
    
    /**
      * free up the memory associated with the old object
    **/
    if (input_objects[n] != NULL)
    {
	surface_attribs save_sa;
	surface_colors_rgba save_colors;
	int save_shader;
	/* 21 april 97 - add saving of certain pieces of the object
	 transformation matrix. */
	vertex_3d save_center;
	vertex_3d save_translate_vect;
	matrix4x4 save_matrix;

	kmemcpy(&save_sa,&(input_objects[n]->attribs),sizeof(surface_attribs));
	kmemcpy(&save_colors,&(input_objects[n]->colors),sizeof(surface_colors_rgba));
        save_shader = input_objects[n]->shader;

	/* save the center point */
	kmemcpy(&save_center,&(input_objects[n]->center),sizeof(vertex_3d));
	/* save the matrix. */
	kmemcpy(&save_matrix,&(input_objects[n]->xfrm),sizeof(matrix4x4));

	
	/* free the old object */
	free_renderobj(input_objects[n]);
	
	/**
	 * do a *copy, which avoids having to traverse through the
	 * object tree fixing up pointers to preserve the parent child
	 * relationships.
	 **/

	*(input_objects[n]) = *r;

	kmemcpy(&(input_objects[n]->attribs),&save_sa,sizeof(surface_attribs));
	kmemcpy(&(input_objects[n]->colors),&save_colors,sizeof(surface_colors_rgba));
        input_objects[n]->shader = save_shader;

	/* restore the center point and the matrix */
	kmemcpy(&(input_objects[n]->center),&save_center,sizeof(vertex_3d ));
	kmemcpy(&(input_objects[n]->xfrm),&save_matrix,sizeof(matrix4x4));

	determine_transparency(input_objects[n]);
	
	/**
	  * now, tweak the UI.
	**/
	/* free the storage used by "r".   */
	kfree(r);
    }
    else /* virgin object pointer. */
    {
	input_objects[n] = r;
	add_child(&root,r);
	num_objects_in_tree++;
    }

    /**
      * update parent's bounding box to reflect extents of the
      * child's bounding box.  this is union type operation.
      * the root object is the parent, the child object is "r".
    **/

#if 0
    union_bounding_boxes(&root,r,&union_box_min,&union_box_max);
#endif
    sync_root_box();


    /**
      * compute the object's center point based upon bounding box.
    **/
#if 0
    for (i=0;i<3;i++)
	root.center.v[i] = root.bmin.v[i] + 0.5 * (root.bmax.v[i] - root.bmin.v[i]);
#endif
    /**
      * rebuild the object list UI
    **/
      
    init_object_list_GUI(NULL);
    
    /** kfree(r) memory leak potential if this is commented out **/
}

static void
free_renderobj(renderobj *r)
{
    int i,n;
    render_prim *p;
   
    if (r == NULL)
       return;

    n = r->nprims;
    p = r->prims;
    
    for (i=0;i<n;i++,p++)
	free_render_prim(p);

    kfree(r->prims);
}

static void
free_render_prim(render_prim *p)
{
    if (p->verts)
	kfree(p->verts);

    if (p->prim_specifics.normals)
	kfree(p->prim_specifics.normals);
    
    if (p->rgb_color_list)
	kfree(p->rgb_color_list);
    
    if (p->rgba_color_list)
	kfree(p->rgba_color_list);

}

/**
  * action handler repsonsible for rereading an object when it`s
  * file contents changes.
**/
int
reread_input_object(xvobject dummy,
		    char *fname,
		    kaddr data)
{
    int *n,status;
    renderobj *t;
    char work[256];

    n = (int *)data;		/* what the heck is this? */

    kstrcpy(work,fname);
    status = whats_in_this_object(fname);

    if (status == RM_GEOMETRY)
    {
        t = create_renderobj();
	status = convert_kobj_to_renderobj(work,t);
        replace_input_object(*n,t);
	renderer_invoke_func();
    }
    else if (status == RM_VRSTUFF)
    {
        status = vr_process_object(fname);
	if (status == CHILL)
	    renderer_invoke_func();
    }
	/* else we'll just continue.. */
    return(TRUE);
}

int
get_object_bbox(renderobj *r,
		vertex_3d *bmin,
		vertex_3d *bmax)
{
    VCOPY(bmin,&(r->bmin));
    VCOPY(bmax,&(r->bmax));
    return(TRUE);
}


int
set_object_bbox(renderobj *r,
		vertex_3d *bmin,
		vertex_3d *bmax)
{
    /* range, etc., checking? */
    if (r == NULL)
    {
         /* issue insult */
         kerror("RENDER","render"," set_object_bbox: the input object is WHACKED, joser!");
    }
    VCOPY(&(r->bmin),bmin);
    VCOPY(&(r->bmax),bmax);

    /* auto-recompute center point?? */
    return(TRUE);
}

int
identity_current_object_matrix()
{
    renderobj *tr;

    tr = get_current_object();
    identity_4x4(&(tr->xfrm));

    /* also, set translate vect to 0. */
    {
        int i;
	for (i=0;i<3;i++)
	    tr->translate_vect.v[i] = 0.;
    }
    return(TRUE);
}

static int
build_disjoint_from_connected_perface(render_prim *p,
				      int nverts,
				      float *verts,
				      float *colors,
				      float *normals,
				      float *tcoords,
				      int layout,
				      int has_alpha)
{
    vertex_3d *work_verts,*work_normals,*v;
    surface_colors_rgba *wc_a;
    surface_colors_rgb  *wc;
    int work_nverts,do_tc,do_colors,offset;
    int i,j,index;
    vertex_3d p1,p2;
    vertex_3d *vverts;
    double dmag;

    work_nverts = (nverts-2)*3;

    do_colors = (colors == NULL) ? 0 : 1;
    do_tc = (tcoords == NULL) ? 0 : 1;
    work_verts = (vertex_3d *)kmalloc(sizeof(vertex_3d)*work_nverts);
    work_normals = (vertex_3d *)kmalloc(sizeof(vertex_3d)*work_nverts);

    v = (vertex_3d *)verts;

    /**
     * copy over the vertices from connected format to disconnected
     * format.
     **/
    offset=0;
    for (i=0;i<(nverts-2);i++)  /* loop on triangles */
    {
	if (i & 1)
	{
	    VCOPY(work_verts+offset,v+i+1);
	    offset++;
	    VCOPY(work_verts+offset,v+i);
	    offset++;
	    VCOPY(work_verts+offset,v+i+2);
	    offset++;
	}
	else
	{
	    VCOPY(work_verts+offset,v+i);
	    offset++;
	    VCOPY(work_verts+offset,v+i+1);
	    offset++;
	    VCOPY(work_verts+offset,v+i+2);
	    offset++;
	}
    }
    p->nverts = work_nverts;
    p->verts = (vertex_3d *)work_verts;
    kfree(verts);

    /**
     * compute normals, if necessary, using the new vertex list.
     **/
    if (normals == NULL)
    {
	vverts = work_verts;
	offset=0;
	for (i=0;i<(nverts-2);i++)  /* loop on triangles */
	{
	    index = i*3;
	    for (j=0;j<3;j++)
	    {
		p1.v[j] = vverts[index+1].v[j] - vverts[index].v[j];
		p2.v[j] = vverts[index+2].v[j] - vverts[index+1].v[j];
	    }
	    
	    vertex_cross(&p1,&p2,work_normals+offset);
	    vertex_unit(work_normals+i,&dmag);

	    /**
	     * flip normal.
	     **/
	    for (j=0;j<3;j++)
		work_normals[i].v[j] *= -1.;

	}
	p->nthings = nverts-2; /* per face normals */
    }
    else  /* have normals already, need to redistribute */
    {
	if (layout == KPER_VERTEX)
	{
	    v = (vertex_3d *)normals;
	    offset=0;
	    for (i=0;i<(nverts-2);i++)  /* loop on triangles */
	    {
		if (i & 1)
		{
		    VCOPY(work_normals+offset,v+i+1);
		    offset++;
		    VCOPY(work_normals+offset,v+i);
		    offset++;
		    VCOPY(work_normals+offset,v+i+2);
		    offset++;
		}
		else
	        {
		    VCOPY(work_normals+offset,v+i);
		    offset++;
		    VCOPY(work_normals+offset,v+i+1);
		    offset++;
		    VCOPY(work_normals+offset,v+i+2);
		    offset++;
		}
	    }
	}
	else /* layout == KPER_FACE */
	{
	    v = (vertex_3d *)normals;
	    offset=0;
	    for (i=0;i<(nverts-2);i++)  /* loop on triangles */
	    {
		VCOPY(work_normals+offset,v+i);
		offset++;
		VCOPY(work_normals+offset,v+i);
		offset++;
		VCOPY(work_normals+offset,v+i);
		offset++;
	    }
	}
    } /* end dealing with normals */
    p->nthings = work_nverts;
    p->prim_specifics.normals = (vertex_3d *)work_normals;
	
    /**
     * deal with color information, if any
     **/

    if (colors)
    {
	float *f;
	int stride;
	
	f = colors;

	if (has_alpha)
	{
	    stride = 4;
	    wc_a = (surface_colors_rgba *)
		kmalloc(sizeof(surface_colors_rgba)*work_nverts);
	}
	else
	{
	    stride = 3;
	    wc = (surface_colors_rgb *)
		kmalloc(sizeof(surface_colors_rgb)*work_nverts);
	}

	/*
	 *   previously did not consider per-face with no alpha
	 *   (alpha values of 1. were inserted)  this case can
	 *   be tested with a color legend generated from gcmaplegend.
	 *
	 *   sort out the colors differently depending on 
	 *   whether or not there is alpha information.
	 *
	 *   I suspect I did this in the most difficult way
	 *   possible :(  // SK
	 */
	if (has_alpha)
	{
	    if (layout == KPER_VERTEX)
	    {
		offset=0;
		for (i=0;i<(nverts-2);i++)  /* loop on triangles */
		{
		    if (i & 1)
		    {
			VCOPY((vertex_3d *)(wc_a+offset),
			      (vertex_3d *)(f+(i+1)*stride));
			wc_a[offset].a = *(f+(i+1)*stride+3);
			offset++;
			
			VCOPY((vertex_3d *)(wc_a+offset),
			      (vertex_3d *)(f+(i)*stride));
			wc_a[offset].a = *(f+(i)*stride+3);
			offset++;
			
			VCOPY((vertex_3d *)(wc_a+offset),
			      (vertex_3d *)(f+(i+2)*stride));
			wc_a[offset].a = *(f+(i+2)*stride+3);
			offset++;
		    }
		    else
		    {
			VCOPY((vertex_3d *)(wc_a+offset),
			      (vertex_3d *)(f+(i)*stride));
			wc_a[offset].a = *(f+(i)*stride+3);
			offset++;
			
			VCOPY((vertex_3d *)(wc_a+offset),
			      (vertex_3d *)(f+(i+1)*stride));
			wc_a[offset].a = *(f+(i+1)*stride+3);
			offset++;
			
			VCOPY((vertex_3d *)(wc_a+offset),
			      (vertex_3d *)(f+(i+2)*stride));
			wc_a[offset].a = *(f+(i+2)*stride+3);
			offset++;
		    }
		}
	    }
	    else /* layout == KPER_FACE */
	    {
		f = colors;
		offset=0;
		for (i=0;i<(nverts-2);i++)  /* loop on triangles */
		{
		    VCOPY((vertex_3d *)(wc_a+offset),
			  (vertex_3d *)(f+(i)*stride));
		    wc_a[offset].a = *(f+(i)*stride+3);
		    offset++;
		    VCOPY((vertex_3d *)(wc_a+offset),
			  (vertex_3d *)(f+(i)*stride));
		    wc_a[offset].a = *(f+(i)*stride+3);
		    offset++;
		    VCOPY((vertex_3d *)(wc_a+offset),
			  (vertex_3d *)(f+(i)*stride));
		    wc_a[offset].a = *(f+(i)*stride+3);
		    offset++;
		}
	    } /* end dealing with colors */
	    p->rgb_color_list = NULL;
	    p->rgba_color_list = wc_a;
	    p->ncolors = work_nverts;
	    kfree(colors);
	}
	else /* has no alpha */
	{
	    if (layout == KPER_VERTEX)
	    {
		offset=0;
		for (i=0;i<(nverts-2);i++)  /* loop on triangles */
		{
		    if (i & 1)
		    {
			VCOPY((vertex_3d *)(wc+offset),
			      (vertex_3d *)(f+(i+1)*stride));
			offset++;
			
			VCOPY((vertex_3d *)(wc+offset),
			      (vertex_3d *)(f+(i)*stride));
			offset++;
			
			VCOPY((vertex_3d *)(wc+offset),
			      (vertex_3d *)(f+(i+2)*stride));
			offset++;
		    }
		    else
		    {
			VCOPY((vertex_3d *)(wc+offset),
			      (vertex_3d *)(f+(i)*stride));
			offset++;
			
			VCOPY((vertex_3d *)(wc+offset),
			      (vertex_3d *)(f+(i+1)*stride));
			offset++;
			
			VCOPY((vertex_3d *)(wc+offset),
			      (vertex_3d *)(f+(i+2)*stride));
			offset++;
		    }
		}
	    }
	    else /* layout == KPER_FACE */
	    {
		f = colors;
		offset=0;
		for (i=0;i<(nverts-2);i++)  /* loop on triangles */
		{
		    VCOPY((vertex_3d *)(wc+offset),
			  (vertex_3d *)(f+(i)*stride));
		    offset++;
		    VCOPY((vertex_3d *)(wc+offset),
			  (vertex_3d *)(f+(i)*stride));
		    offset++;
		    VCOPY((vertex_3d *)(wc+offset),
			  (vertex_3d *)(f+(i)*stride));
		    offset++;
		}
	    } /* end dealing with colors */
	    p->rgb_color_list = wc;
	    p->rgba_color_list = NULL;
	    p->ncolors = work_nverts;
	    kfree(colors);
	}
    }
    else
    {
	p->ncolors = 0;
	p->rgba_color_list = NULL;
	p->rgb_color_list = NULL;
    }
    
    p->type = KGEOM_TRIANGLES_DISJOINT;
    return(CHILL);
}

static int
build_disjoint_from_connected_nonorms(render_prim *p,
				      int nverts,
				      float *verts,
				      float *colors,
				      int layout,
				      int has_alpha)
{
    /* if we have to compute normals, they will be computed
       per-face, and we'll rebuild an object which is a disjoint
       polytriangle primitive.  this simplifies a lot of downstream
       rendering code. */
	    
    vertex_3d *w_norms;
    vertex_3d *work,*lverts;
    int n,i,j,ntriangles;
    int color_size;

    vertex_3d *new_verts;
    vertex_3d *new_norms;
    float *new_colors;
    float *new_tcoords;

    ntriangles = n = nverts-2;

    /* first, compute all the normals from the source vertices and
     copy them over into the primitive. */
    
    w_norms = (vertex_3d *)kmalloc(sizeof(vertex_3d)*n);
    work = (vertex_3d *)kmalloc(sizeof(vertex_3d)*(n+1));

    /* compute c[i] = a[i]-b[i] */
    lverts = (vertex_3d *)verts;
    for (i=0;i<n+1;i++)
    {
	for (j=0;j<3;j++)
	    work[i].v[j] = lverts[i+1].v[j] - lverts[i].v[j];
    }

    for (i=0;i<n;i++)
	vertex_cross(work+i,work+i+1,w_norms+i);
    
    nnormal_unit(w_norms,w_norms,n);

    for (i=0;i<n;i++)
    {
	if (!(i&1)) 
	    /*		if ((i&1)) */
        { 
	    for (j=0;j<3;j++)
		w_norms[i].v[j] *= -1.;
	}

    }

    new_norms = (vertex_3d *)kmalloc(sizeof(vertex_3d)*(ntriangles*3));
    tstrip_to_disjoint_f((float *)w_norms,(float *)new_norms,
			 ntriangles*3,3);
    p->nthings = ntriangles *3;
    p->prim_specifics.normals = new_norms;

    kfree(work);
    kfree(w_norms);

    /* ok, now copy over the vertices */
    new_verts = (vertex_3d *)kmalloc(sizeof(vertex_3d)*(ntriangles*3));
    tstrip_to_disjoint_v(verts,(float *)new_verts,
			 ntriangles*3,3);
    p->verts = new_verts;
    p->nverts = ntriangles*3;
    
    /* can we kfree the verts???? */
    

    /* ok, now do color. */
    if (has_alpha)
	color_size = 4;
    else
	color_size = 3;
    if (colors)
    {
	new_colors = (float *)kmalloc(sizeof(float)*color_size*(ntriangles*3));
	tstrip_to_disjoint_v(colors,new_colors,ntriangles*3,color_size);
	if (has_alpha)
	    p->rgba_color_list = (surface_colors_rgba *)new_colors;
	else
	    p->rgb_color_list = (surface_colors_rgb *)new_colors;
	p->ncolors = ntriangles*3;
    }
    else
	p->ncolors = 0;

    /* can we kfree colors()?? */

    p->type = KGEOM_TRIANGLES_DISJOINT;
    return(CHILL);
}



static int
tstrip_to_disjoint_v(float *s,
		     float *d,
		     int n,
		     int size_s)
{
    /* copies from array S into array D using the assumption that
     the S entries are arranged 0,1,2,3,... and that they will end
     up in D as if 0,1,2,1,2,3,2,3,4....  */
    
    register float *rs,*rd;
    register int i,nbytes,index;

    rs = s;
    rd = d;
    nbytes = size_s * sizeof(float);
    
    for (i=0;i<n;i++,rd+=size_s)
    {
	index = size_s * (i/3 + i%3);
	kmemcpy(rd,rs+index,nbytes);
    }
    return(CHILL);
}
static int
tstrip_to_disjoint_f(float *s,
		     float *d,
		     int n,
		     int size_s)
    /* copies from array S into array D using the assumption that
     the S entries are arranged 0,1,2,3,... and that they will end
     up in D as if 0,0,0,1,1,1,2,2,2,... */
{    
    register float *rs,*rd;
    register int i,nbytes,index;

    rs = s;
    rd = d;
    nbytes = size_s * sizeof(float);
    
    for (i=0;i<n;i++,rd+= size_s)
    {
	index = size_s * (i/3);
	kmemcpy(rd,rs+index,nbytes);
    }
    return(CHILL);
}

static int
compute_octmesh_radii(int *mesh_dims,
		      float *r,
		      float *verts)
{
    int nvoxels;
    int i,j,k,l;
    vertex_3d *v1,*base;
    vertex_3d distances[3];
    double mags[3];
    int isize,ijsize;
    float *rwork;

    /* the voxel radius is computed as the average of the distance from
     the grid point (the voxel center) to each of the grid points in the
     +i, +j and +k directions.  this is cheesy, but will work.  voxels
     that lie at grid boundaries will have radius values of 0.0 */
    
    nvoxels = mesh_dims[0] * mesh_dims[1] * mesh_dims[2];
    isize = mesh_dims[0];
    ijsize = mesh_dims[0] * mesh_dims[1];
    
    base = (vertex_3d *)verts;

    /* set all radius value */
    for (i=0;i<nvoxels;i++)
	r[i] = 0.;

    /* compute the radius values for all non-boundary grid points */
    
    for (k=0;k<mesh_dims[2];k++)
    {
	for (j=0;j<mesh_dims[1];j++)
	{
	    v1 = base + k*ijsize + j*isize;
	    rwork = r + k*ijsize + j*isize;
	    for (i=0;i<mesh_dims[0];i++,v1++,rwork++)
	    {
		npoints_diff(v1+1,v1,distances,1); /* +i direction */
		npoints_diff(v1+isize,v1,distances+1,1); /* +j direction */
		npoints_diff(v1+ijsize,v1,distances+2,1); /* +k direction */

		for (l=0;l<3;l++)
		    vertex_mag(distances+l,mags+l);

		*rwork = mags[0];
		*rwork = (*rwork > mags[1]) ? *rwork : mags[1];
		*rwork = (*rwork > mags[2]) ? *rwork : mags[2];
	    }
	}
    }
    return(CHILL);
}

static int
whats_in_this_object(char *obj_name)
{
    kgeom_object *o;
    int i,num,*list;
    int rstat = RM_BOGOSITY;

    /* the purpose of this routine is to look inside the incoming
     * object and determine if it consists of
     * 1. geometry,
     * 2. a transformation event (including camera control operations)
     * 3. unrecognizable stuff.
     *
     * the following rules of thumb applies:
     * 1. if there's no prim list, the object is bogus.
     * 2. the species of the first item in the prim list determines
     *    what this routine reports.  in other words, if you have an
     *    object that has both geometry and vr stuff, you're hosed.
     */

   o = kgeom_read_object(obj_name);

   if (o == NULL || (num = kgeom_number_primitives(o)) == 0)
   {
       kerror("GEOMETRY","render","the input object <%s> is invalid or contains no primitives.",obj_name);
   }
   else
   {
     kgeom_primitive *p = kgeom_get_primitive(o, 0);
     switch (p->type)
       {
       case KGEOM_OBJECT_XFORM:
       case KGEOM_CAMERA_ORIENTATION:
	 rstat = RM_VRSTUFF;
	 break;
       
       case KGEOM_VIEW_XFORM:	/* can't deal with this... */
	 kerror("GEOMETRY","render","render can't deal with applying a transformation matrix in lieu of specific camera parameters such as eye point, look at point,etc.  Please specify those values explicitly if you want to control the camera position. i'm going to ignore this object and continue. ");
	 rstat = RM_BOGOSITY;
	 break;

       case KGEOM_POLYLINE_DISJOINT:
       case KGEOM_POLYLINE_CONNECTED:
       case KGEOM_TRIANGLES_DISJOINT:
       case KGEOM_TRIANGLES_CONNECTED:
       case KGEOM_SPHERES:
       case KGEOM_CYLINDERS:
       case KGEOM_DIRECTED_POINTS:
       case KGEOM_QUADMESH:
       case KGEOM_OCTMESH:
       case KGEOM_TEXT:
       case KGEOM_TEXTURE2D:
	 rstat = RM_GEOMETRY;
	 break;

       default:
	 rstat = RM_BOGOSITY;
	 kerror("GEOMETRY","render","the input object contains a primitive list, but i can't recognize the primitive.  please file a bug report. this object will be skipped, and i'll continue. ");
	 break;
	 
       }
   }
   kgeom_blast_object(o);

   return(rstat);
}

void
free_objects_memory()
{
    /* this routine frees all memory associated with objects at
     the end of the sessions. */

    int i;

    for (i=0;i<num_objects_in_tree-1;i++) /* "root" gets counted as one */
    {
	free_renderobj(input_objects[i]);
	kfree(input_objects[i]);
    }

    /* can't free the "root" object because it's static.  but, we
     can free any associated stuff.*/

    free_renderobj(get_root_object());
}
