
/*
 * Copyright (c) 1996,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 *
 */
 


#ifndef __render_objs_h_
#define __render_objs_h_

#include "geometry/geometry.h"

#define RM_GEOMETRY 0
#define RM_VRSTUFF 1
#define RM_BOGOSITY 2

#define TRANSPARENT_OBJ 0
#define OPAQUE_OBJ 1

#define MAXOFFSPRING 12

static char ROOT_OBJECT_NAME[] = {"root"};

typedef struct surface_attribs
{
    KGEOM_VERTEX_TYPE amb,diff,spec,spec_exp,spec_metal;
} surface_attribs;

typedef struct surface_colors_rgba
{
    KGEOM_VERTEX_TYPE r,g,b,a;
} surface_colors_rgba;


typedef struct surface_colors_rgb
{
    KGEOM_VERTEX_TYPE r,g,b;
} surface_colors_rgb;

typedef struct surface_colors_i
{
    KGEOM_VERTEX_TYPE intensity;
} surface_colors_i;

typedef struct surface_colors_ia
{
    KGEOM_VERTEX_TYPE intensity,alpha;
} surface_colors_ia;

typedef struct byte_rgba_colors
{
    unsigned char r,g,b,a;
} byte_rgba_colors;


typedef struct render_texture
{
    union
    {
	/* 2d texture images */
	surface_colors_i **i_image;
	surface_colors_ia **ia_image;
	surface_colors_rgb **rgb_image;
	surface_colors_rgba **rgba_image;

	/* 3d texture volumes */
	surface_colors_i ***i_volume;
	surface_colors_ia ***ia_volume;
	surface_colors_rgb ***rgb_volume;
	surface_colors_rgba ***rgba_volume;
    } texture_union;
    
    int ncomponents;		/* use 1 for intensity maps, 2 for
				   intensity+alpha maps,
				   3 for RGB textures, 4 for RGBA textures. */
    
    int dims;			/* use 2 for images, 3 for volumes */
    int dimensions[3];		/* size of each dimension */
    int any_transparency;	/* used to indicate if any alpha values,
				 if present, are not equal to 1.0 */
    
} render_texture;

typedef struct render_text_struct
{
    char *source_string;
    char *source_font;

    int r_width,r_height;
    unsigned char **mask;	/* bitmap for X-rendered character */
    unsigned char **alpha;	/* coverage information  */
    surface_colors_rgb *fg_color;
#if 0
    surface_colors_rgb bg_color; /* what's this for? */
#endif
    vertex_3d up,offset,flow;   /* not yet doing anything with these */
    /* the bitmap is a bitmap representation of the string.  where there's
     a 1, the pixel is turned on.  where 0, there is no coverage. the
     implication is that the bitmap mask is padded, rowwise, to the next
     higher byte boundary. */
    unsigned char **bitmap_mask;
} render_text_struct;


typedef struct render_prim
{
    int type;
    
    int nverts;
    vertex_3d *verts;
    
    int nthings;
    union
    {
	vertex_3d *normals;
	float *radii;
	render_text_struct *text;
    } prim_specifics;

    int mesh_dims[3]; /* used for holding sizing info for qmesh, omesh */
    float *oct_mesh_radii;	/* used for splat-footprint computation */
    int ncolors;

    /* NOTE: the following two should probably be merged into a single
     unionized structure or something.  jan 1996, wes. */

    surface_colors_rgb *rgb_color_list;
    surface_colors_rgba *rgba_color_list;
    byte_rgba_colors *byte_rgba_color_list;
    
    
    int pshader;

/*-- These are the possible fully prototyped clipfunc values.  To make this
     work under C++, I had to do some casting and with the actual function
     calls to make it compile.  -- Steve Jorgensen,  May 1, 1997

    int (*clipfunc)(Poly *, int, float *, int, int);
    int (*clipfunc)(render_prim *, Poly *, Poly *, int, float *, int, int);
*/
    int (*clipfunc)(void);

    int has_texture;
    char texture_transport_name[KGEOM_MAX_NAME_LENGTH];
    int tcoords_dimensions;
    float *tcoords;
    render_texture *the_texture;
} render_prim;


#define IGNOREBOX (1<<0)	/* implies no transforming */

typedef struct renderobj
{
    int nparents,nchildren;
    struct renderobj *parents[MAXOFFSPRING];
    struct renderobj *children[MAXOFFSPRING];
    char name[KGEOM_MAX_NAME_LENGTH];
    unsigned int flags;

    matrix4x4 xfrm;  
    
    vertex_3d translate_vect;
    vertex_3d center;
    vertex_3d bmin,bmax;  /* for bounding box. */
    
    surface_attribs attribs;
    surface_colors_rgba colors;

    int do_display;
    int is_transparent; /* this is different from HAS_ALPHA */
    int has_alpha;

    int nprims;
    render_prim *prims;

    int layout;

    /**
      * since a geom object is comprised of several primitives,
      * the notion of an object level shader applying to heterogeneous
      * primitives is slippery.  for now, we are assuming that
      * all the primitive types are homogeneous (ie, all volumetric,
      * all surface based, or all wireframe) and that all of the
      * primitives will use the object-level shader.
    **/
    int shader;

    /* i wonder if this should really be per-prim..but since data
     services has this as an object-level attribute, i'll leave it
     here for now.  6 mar 96, wes */
    
    int verts_nspace;		/* dimensionality of coords in object -
				 we expect either 2 or 3 coords.  we will
				 always store 3d coords, as data services will
				 always give us 3d coords */


} renderobj;


void init_render_obj PROTO((renderobj *, char *));
renderobj *create_renderobj PROTO((void));
int convert_kobj_to_renderobj PROTO((char *, renderobj *));
int add_child PROTO((renderobj *,renderobj *));
int find_renderobj PROTO((char *, renderobj **));
renderobj *get_current_object PROTO((void));
renderobj *get_root_object PROTO((void));

int set_object_bbox PROTO((renderobj *, vertex_3d *,vertex_3d *));
int get_object_bbox PROTO((renderobj *, vertex_3d *,vertex_3d *));
int set_current_object PROTO((renderobj *));
int set_object_shader PROTO((renderobj *, int));
int get_object_shader PROTO((renderobj *, int *));
int set_object_display_status PROTO((renderobj *, int));
int get_object_display_status PROTO((renderobj *, int *));
int get_object_rgba PROTO((renderobj *, float *, float *, float *, float *));
int set_object_rgba PROTO((renderobj *, float, float, float, float));

int set_object_amb PROTO((renderobj *, float));
int get_object_amb PROTO((renderobj *, float *));

int set_object_diff PROTO((renderobj *, float));
int get_object_diff PROTO((renderobj *, float *));

int set_object_spec PROTO((renderobj *, float));
int get_object_spec PROTO((renderobj *, float *));

int set_object_spec_exp PROTO((renderobj *, float));
int get_object_spec_exp PROTO((renderobj *, float *));

int set_object_metal PROTO((renderobj *, float));
int get_object_metal PROTO((renderobj *, float *));

void geom_xform_points PROTO((matrix4x4 *,vertex_3d*, vertex_3d *,vertex_3d *,vertex_3d *,matrix4x4 *,vertex_3d *,int));

void  replace_input_object PROTO((int,renderobj *));
int identity_current_object_matrix PROTO((void));

int have_any_transparent_objs PROTO((renderobj *r));
int is_opaque PROTO((renderobj *));
#endif
