 /*
  * Khoros: $Id: lifiltdesign.c,v 1.14.2.3 1997/05/31 19:15:04 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lifiltdesign.c,v 1.14.2.3 1997/05/31 19:15:04 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for ifiltdesign
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lifiltdesign
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
static double pix_dist PROTO((double, double, int, int, int));
static double ideal_lpf PROTO((double));
static double but_lpf   PROTO((double, int));
static double gauss_lpf PROTO((double, int));
static double ch1_lpf   PROTO((double, double, int));
static double ch2_lpf   PROTO((double, double, int));
static double lpf       PROTO((double, double, double, int, int));
static double hpf       PROTO((double, double, double, int, int));
static double bpf       PROTO((double, double, double, double, int, int));
static double bsf       PROTO((double, double, double, double, int, int));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lifiltdesign - generate 2-dimensional frequency filter
* 
*       Purpose: This routine generates a spatial frequency image of real values
*		 (zero phase) that represent the magnitude response of the 
*	         requested filter type and brand.  The filter types
*		 are are: lowpass, highpass, bandpass, and bandstop. The
*		 function brands are: ideal, Butterworth, Gaussian, Chebyshev I,
*		 and Chebyshev II. 
*
*                The output is symmetric across a diameter through the origin 
*                of the image; the origin can be either the center of the image 
*		 or the top left corner of the image.  Cutoff frequencies are 
*		 normalized to the Nyquist frequency, thus implying that 0 
*                corresponds to DC and 1.0 corresponds to the Nyquist frequency.
*
*                Note that the number of pixels between DC and Nyquist is
*                given by the image_width/2 in the width direction and
*                image_height/2 in the height direction. This means that if
*                you ask for a non-square image you will get a non-circularly
*                symmetric image passband in the output. This occurs because
*                it is assumed that the frequency sampling interval is the same
*                in both directions.
*
*         Input: type          - type of filter.
*		 brand         - brand of filter.
*		 order         - filter order.
*		 image_height  - height of the image in pixels.
*		 image_width   - width of the image in pixels.
*		 lower_cutoff  - lower cutoff for band pass or band stop,
*				 and cutoff for high pass filters, expressed in
*                                fraction of Nyquist.
*		 upper_cutoff  - upper cutoff for band pass or band stop,
*				 and cutoff for low pass filters, expressed in
*                                fraction of Nyquist.
*		 epsilon       - Chebyshev parameter to control the ripple in
*				 the passband or stopband.
*		 centered      - if TRUE, it puts low frequencies in the center
*				 of the image.  Otherwise it puts them in the
*				 corners.
*
*        Output: freq_image    - An open kobject in which to store the filter
*                                image
*
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: 
*    Written By: Steven Jorgensen, Scott Wilson
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int
lifiltdesign(
   int	    type,
   int	    brand,
   int	    order,
   int	    image_height,
   int	    image_width,
   double   lower_cutoff,
   double   upper_cutoff,
   double   epsilon,
   int	    centered,
   kobject  freq_image)
/* -library_def_end */

/* -library_code */
{
   kobject  fi_ref = NULL;
   double  *line = NULL;
   int      i, j;
   double   freq;  /* Actual frequency */
   char *lib="kimage_proc", *rtn="lifiltdesign";
   klist *objlist=NULL;

   /* Error check the parameters passed in by the user.
      type is checked by the switch statement below.  */
   if (brand < KFILT_IDEAL || brand > KFILT_CHEBYSHEV_II)
   {
      kerror(lib, rtn,
	     "Invalid Filter Brand, KFILT_IDEAL, KFILT_BUTTERWORTH, "
	     "KFILT_GAUSSIAN, KFILT_CHEBYSHEV_I, and KFILT_CHEBYSHEV_II "
	     "are the only valid brands.");
      return FALSE;
   }

   if (image_height <= 0)
   {
      kerror(lib, rtn, "Image height must be a positive value.");
      return FALSE;
   }
   if (image_width <= 0)
   {
      kerror(lib, rtn, "Image width must be a positive value.");
      return FALSE;
   }
   if (!freq_image)
   {
      kerror(lib, rtn, "Frequency Image object is not a valid data object.");
      return FALSE;
   }

   if (type < KFILT_LP || type > KFILT_BS)
   {
      kerror(lib,rtn,"Invalid filter type, valid types are KFILT_LP, KFILT_HP, "
	     "KFILT_BP, or KFILT_BS");
      return FALSE;
   }
   if ((type == KFILT_BP || type == KFILT_BS) &&
       ((upper_cutoff - lower_cutoff) <= 0))
   {
      kerror(lib,rtn,"Upper cutoff value must be greater than the Lower cutoff "
	     "value for this type of filter.");
      return FALSE;
   }

   KCALL((fi_ref = kpds_reference_object(freq_image)) != KOBJECT_INVALID);
   objlist = klist_add(objlist,fi_ref,"KOBJECT");

   KCALL(kpds_set_attributes(fi_ref, KPDS_VALUE_DATA_TYPE, KDOUBLE,
			     KPDS_VALUE_SIZE, image_width, image_height,
			     1, 1, 1, NULL));

   /* Alloc space for a line of image data */
   KCALL(!((line=(double *)
                kcalloc((size_t)image_width,sizeof(double)))== NULL));
   objlist = klist_add(objlist,line,"KMALLOC");

   /* For each line in image do */
   for (i = 0; i < image_height; i++)
   {
      /* For each element in the line do */
      for (j = 0; j < image_width; j++)
      {
	 /* compute frequency relative to DC */
	 freq = pix_dist((double)j, (double)i,image_width,
                                              image_height,centered);

         switch(type)
           {
             case KFILT_LP:  line[j]=lpf(freq,upper_cutoff,epsilon,order,brand);
                             break;
             case KFILT_HP:  line[j]=hpf(freq,lower_cutoff,epsilon,order,brand);
                             break;
             case KFILT_BP:  line[j]=bpf(freq,lower_cutoff,upper_cutoff,epsilon,
                                         order,brand);
                             break;
             case KFILT_BS:  line[j]=bsf(freq,lower_cutoff,upper_cutoff,epsilon,
                                         order,brand);
                             break;
             default: kerror(lib,rtn,"Bogus filter type.");
                      (void)klist_free(objlist,(void (*)(klist*))lkcall_free);
                      return FALSE;
           }
      }
      KCALL(kpds_put_data(fi_ref, KPDS_VALUE_LINE, line));
   }

   (void)klist_free(objlist,(void (*)(klist*))lkcall_free);
   return TRUE;
}

static double
pix_dist(
   double  x,
   double  y,
   int     ix,
   int     iy,
   int     centered)
{
   double dx, dy, d;

  /* Centered implies that the low freqencies are at the center of the image. */
   if (centered)
   {
      if (x < ((double) ix)/2.0)
	 dx = ((double) ix)/2.0 - x;
      else
	 dx = x - ((double) ix)/2.0;
      if (y < ((double) iy)/2.0)
	 dy = ((double) iy)/2.0 - y;
      else
	 dy = y - ((double) iy)/2.0;
   }

  /* Implies that the low freqencies are at the corners of the image.  */
   else
   {
      if (x < ((double) ix)/2.0)
	 dx = x;
      else
	 dx = ((double) ix) - x;
      if (y < ((double) iy)/2.0)
	 dy =  y;
      else
	 dy = ((double) iy) - y;
   }

  /* Normalize to the Nyquist frequency in each direction.  */ 
   dx /= (ix/2.0);
   dy /= (iy/2.0);

  /* Form aggregate frequency */
   d = ksqrt(dx*dx + dy*dy);

   return d;
}

static double
cheb_ord(
   double  x,
   int     order)
{
   /* This implements a simple recursive formula to compute the value of the
      order N Chebyshev polynomial at x, as described in Oppenheim and
      Schafer, p 847. */
   if (order < 0)
      return 1.0;

   if (order == 1)
      return x;

   return(2.0 * x * cheb_ord(x, order -1) - cheb_ord(x, order -2));
}

/* Perform a frequency transformation to obtain the correct
   type of filter. The transformed frequency is plugged into the
   appropriate unity-cutoff (normalized frequency) filter to obtain
   the actual magnitude response. The transformations are:

   Lowpass with cutoff at wc:
     w' = w/wc		
   Highpass with cutoff at wc:
     w'=wc/w
   Bandpass with lower cutoff at wl and upper cutoff at wu:
     w'=(w^2+wu*wl)/(w*(wu-wl))
   Bandstop with lower cutoff at wl and upper cutoff at wu:
     w'=(w*(wu-wl))/(w^2+wu*wl)

   These transformations can be found in Edward P. Cunningham,
   "Digital Filtering: An Introduction", ISBN 0-395-53989-7,
   p.89-91. Look out for the fact that s is substituted as
   j(omega), not just (omega).
*/

static double
lpf(
   double  freq,
   double  cutoff,
   double  epsilon,
   int     order,
   int     brand)
{
   double tfreq,val=0.0;;

   /* Transform the frequency into a normalized lowpass */
   tfreq = freq/cutoff;

   switch(brand)
     {
       case KFILT_IDEAL:        val=ideal_lpf(tfreq);
                                break;
       case KFILT_BUTTERWORTH:  val=but_lpf(tfreq,order);
                                break;
       case KFILT_GAUSSIAN:     val=gauss_lpf(tfreq,order);
                                break;
       case KFILT_CHEBYSHEV_I:  val=ch1_lpf(tfreq,epsilon,order);
                                break;
       case KFILT_CHEBYSHEV_II: val=ch2_lpf(tfreq,epsilon,order);
                                break;
     }
   return(val);
}

static double
hpf(
   double  freq,
   double  cutoff,
   double  epsilon,
   int     order,
   int     brand)
{
   double tfreq,val=0.0;;
 
   /* Transform the frequency into a normalized highpass */
   if (freq == 0.0)
     {
       /* Highpass should be zero at DC */
       return(0.0);
     }
   tfreq = cutoff/freq;
 
   switch(brand)
     {
       case KFILT_IDEAL:        val=ideal_lpf(tfreq);
                                break;
       case KFILT_BUTTERWORTH:  val=but_lpf(tfreq,order);
                                break;
       case KFILT_GAUSSIAN:     val=gauss_lpf(tfreq,order);
                                break;
       case KFILT_CHEBYSHEV_I:  val=ch1_lpf(tfreq,epsilon,order);
                                break;
       case KFILT_CHEBYSHEV_II: val=ch2_lpf(tfreq,epsilon,order);
                                break;
     }
   return(val);
}

static double
bpf(
   double  freq,
   double  lower_cutoff,
   double  upper_cutoff,
   double  epsilon,
   int     order,
   int     brand)
{
   double tfreq,val=0.0;;
 
   /* Transform the frequency into a normalized bandpass */
   if (freq == 0.0)
     {
       /* Bandpass should be zero at DC */
       return(0.0);
     }
   tfreq = (lower_cutoff*upper_cutoff-freq*freq)/
           ((upper_cutoff-lower_cutoff)*freq);
   tfreq = kabs(tfreq); 
 
   switch(brand)
     {
       case KFILT_IDEAL:        val=ideal_lpf(tfreq);
                                break;
       case KFILT_BUTTERWORTH:  val=but_lpf(tfreq,order);
                                break;
       case KFILT_GAUSSIAN:     val=gauss_lpf(tfreq,order);
                                break;
       case KFILT_CHEBYSHEV_I:  val=ch1_lpf(tfreq,epsilon,order);
                                break;
       case KFILT_CHEBYSHEV_II: val=ch2_lpf(tfreq,epsilon,order);
                                break;
     }
   return(val);
}

static double
bsf(
   double  freq,
   double  lower_cutoff,
   double  upper_cutoff,
   double  epsilon,
   int     order,
   int     brand)
{
   double tfreq,val=0.0;;
 
   /* Transform the frequency into a normalized bandstop */
   tfreq = (freq*(upper_cutoff-lower_cutoff))/
           (lower_cutoff*upper_cutoff-freq*freq);
   tfreq = kabs(tfreq); 
 
   switch(brand)
     {
       case KFILT_IDEAL:        val=ideal_lpf(tfreq);
                                break;
       case KFILT_BUTTERWORTH:  val=but_lpf(tfreq,order);
                                break;
       case KFILT_GAUSSIAN:     val=gauss_lpf(tfreq,order);
                                break;
       case KFILT_CHEBYSHEV_I:  val=ch1_lpf(tfreq,epsilon,order);
                                break;
       case KFILT_CHEBYSHEV_II: val=ch2_lpf(tfreq,epsilon,order);
                                break;
     }
   return(val);
}

/* Following are the basic building blocks: lowpass filters set up with
   cutoff at a normalized frequency of 1.0 */

static double
ideal_lpf(
   double  freq)
{
   return((freq <= 1.0) ? 1.0 : 0.0);
}

static double
but_lpf(
   double  freq,
   int     order)
{
   /* Oppenhiem and Schafer p.845, Eq. B.1. Note that the equation in the book
      is for the *squared* magnitude response, so we must take the square
      root here in order to get the magnitude response. The magnitude response
      is 1/sqrt(2) at the cutoff frequency. */
   double val;

   val = 1.0 / (1.0 + kpow(freq, 2.0*(double) order));
   val = sqrt(val);
   return (val);
}

static double
gauss_lpf(
   double  freq,
   int     order)
{
   /* This is an ad-hoc filter, a super-Gaussian.  The funky constant
      is obtained by looking by setting the magnitude response to 1/sqrt(2) at
      the cutoff frequency; you end up needing a scaling coefficient of
      2ln(sqrt(2)). So, the reponse function ends up being
          H(f) = exp(-0.5*(0.69314718)*[f^(2*order)])
      for positive f.
   */
   double val;

   val = kpow( freq, (double)(2.0*order));
   val = kexp(-0.5*(0.69314718)*val);

   return val;
}

static double
ch1_lpf(
   double  freq,
   double  epsilon,
   int     order)
{
   /* This is the Type I Chebyshev filter, which places the ripple in the
      passband. This filter is described in Oppenheim and Schafer, p. 847,
      eq. B.4. The expression in the book is for the *squared response*, so
      we must take the square root.  Note that epsilon controls the
      amount of passband ripple. In the passband, the magnitude response
      oscillates between 1 and 1/sqrt(1+epsilon^2). At the cutoff frequency, the
      reponse is exactly 1/sqrt(1+epsilon^2). The way to identify the required
      values of epsilon and order is to set epsilon to control the maximum
      ripple and then set the order to obtain the desired attenuation in
      the stopband. */
   double val;

   val = cheb_ord(freq, order);
   val = 1.0 / (1.0 + (epsilon*epsilon)*(val * val));
   val = sqrt(val);

   return val;
}

static double
ch2_lpf(
   double  freq,
   double  epsilon,
   int     order)
{
   /* This is the Type II Chebyshev filter, which places the ripple in the
      stopband. This filter is described in Oppenheim and Schafer, p. 848,
      eq. B.10. The expression in the book is for the *squared response*, so
      we must take the square root.  Epsilon controls the
      amount of stopband ripple. The reponse at the cutoff frequency is
      1/sqrt(1+(1/epsilon^2)). The reponse varies between 0 and
      1/sqrt(1+(1/epsilon^2)) in the stopband. */
   double val;

   if (freq == 0.0)
      return 1.0;

   val = cheb_ord(1.0/freq, order);
   val = 1.0 / (1.0 + (1.0 / ((epsilon*epsilon)*(val * val))));
   val = sqrt(val);

   return val;
}
/* -library_code_end */
