 /*
  * Khoros: $Id: lirotate.c,v 1.22.2.3 1997/05/31 19:15:17 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lirotate.c,v 1.22.2.3 1997/05/31 19:15:17 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for irotate
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lirotate
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lirotate - rotate image plane by arbitrary angle about an arbitrary point
* 
*       Purpose: rotate image plane about an arbitrary point and an
*                arbitrary angle. Pixel values are obtained by bilinear
*                interpolation.
*
*         Input: in_obj - input object to be processed
*                wctr - W coordinate of center of rotation
*                hctr - H coordinate of center of rotation
*                resize - if non-zero, resize output to fit data
*                planes - if non-zero, process date by full planes
*                             instead of prisms. Can be MUCH faster by planes
*                             than prisms, but is less well adapted for large 
*                             data sets
*                angle - rotation angle in degrees. Positive is CW
*                padding - how to handle areas of the output object that do not map to the input object; 1 : pad with zeros, 2 : pad with the value given by (padr,padi), or 3 : pad using the values stored in the KPDS_PAD_VALUE in the input object
*		 padr - real part of padding value
*		 padi - imaginary part of padding value
*		 padvalid - if a mask is present on the input data, then those areas on the output object that are given the pad value will be marked as valid if padvalid=1 or invalid otherwise
*
*        Output: out_obj - output object
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: 
*    Written By: Scott Wilson
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: *
*   Declaration: int lirotate(
*		 !   kobject in_obj,
*		 !   double wctr,
*		 !   double hctr,
*		 !   int resize,
*		 !   int planes,
*		 !   double angle,
*		 !   int padding,
*		 !   double padr,
*		 !   double padi,
*		 !   int padvalid,
*		 !   kobject out_obj)
****************************************************************/
/* -library_def */
int lirotate(
  kobject in_obj,
  double wctr,
  double hctr,
  int resize,
  int planes,
  double angle,
  int padding,
  double padr,
  double padi,
  int padvalid,
  kobject out_obj
)
/* -library_def_end */

/* -library_code */
{
        char *lib = "kimage_proc", *rtn = "lirotate";
        klist *objlist=NULL;
        unsigned int i,j,k,l,m;
        int map_exists,mask_exists;
	unsigned int wc,hc,dc,tc,ec; /* Dimensions of input data set */
	unsigned int wout,hout,dout,tout,eout; /* Dimensions of output data */
	unsigned int wro,hro,dro; /* Dimensions of output region */
	unsigned int wri,hri,dri; /* Dimensions of input region */
        double *bd=NULL,*bm=NULL,*bmask=NULL,*bdata=NULL;
        kdcomplex *bd_dc=NULL,*bdata_dc=NULL;
        double dang,sine,csine,in,kn,z1,z2,z3,z4,za,zb,zp;
        int jl,jh,il,ih;
        int beg[5],end[5];
        float wctri,hctri,wctro,hctro,xhi,xlo,yhi,ylo;
        float x1out,x2out,x3out,x4out;
        float y1out,y2out,y3out,y4out;
        double pad_real,pad_imag;
	int grid;
        int type;

        /* Make sure we have valid objects */
        if (in_obj == KOBJECT_INVALID || out_obj == KOBJECT_INVALID)
          {
            kerror(lib, rtn, "Bogus input or output object");
            return(FALSE);
          }
	if (kpds_query_value(in_obj) == 0)
          {
            kerror(lib, rtn, "No value data in input object.");
            return(FALSE);
          }

	/* We can't handle anything having to do with location data */
        if (kpds_query_location(in_obj))
        {
           if (!kpds_get_attribute(in_obj, KPDS_LOCATION_GRID, &grid))
           {
              kerror(lib, rtn, "Unable to retrieve location grid attribute.");
              return(FALSE);
           }
           if (grid != KUNIFORM)
           {
              kerror(lib, rtn,
                     "Can't handle anything except KUNIFORM location. Use"
                     " klocxform to rotate location data.");
              return(FALSE);
           }
        }

	/* Obtain the proper pad values */
	switch(padding)
          {
            case 1:
		pad_real = 0.0;
		pad_imag = 0.0;
		break;
            case 2:
		pad_real = padr;
		pad_imag = padi;
		break;
            case 3:
                KCALL(kpds_set_attribute(in_obj,KPDS_VALUE_PAD_VALUE,
                                         &pad_real,&pad_imag));
		break;
            default:
                kerror(lib, rtn, "Unknown padding behavior request: %d",padding);
                return(FALSE);
          }

        /* Reference the input object to avoid side effects, then add the
           reference to the list of goodies to be autmatically free'd on
           error. */
        KCALL((in_obj = kpds_reference_object(in_obj)) != KOBJECT_INVALID);
        objlist = klist_add(objlist,in_obj,"KOBJECT");

        mask_exists=kpds_query_mask(in_obj);
        map_exists =kpds_query_map(in_obj);

        /* If map exists, push all the data thru the map */
        if (map_exists)
          {
            KCALL(kpds_set_attribute(in_obj,KPDS_MAPPING_MODE,KMAPPED));
            KCALL(kpds_destroy_map(out_obj));
          }

        /* Create the output value segment if it's not there already and
           copy it's size and data type attributes from the input obj.  */
        KCALL(kpds_copy_object(in_obj,out_obj));

        /* Constrain the input data to KDOUBLE or KDCOMPLEX for reading */
        KCALL(kpds_get_attribute(in_obj,KPDS_VALUE_DATA_TYPE,&type));
        if (type != KCOMPLEX && type != KDCOMPLEX)
          {
            KCALL(kpds_set_attributes(in_obj,
                                      KPDS_VALUE_DATA_TYPE, KDOUBLE,
                                      KPDS_VALUE_INTERPOLATE,KPAD,NULL));
          }
        else
          {
            KCALL(kpds_set_attributes(in_obj,
                                      KPDS_VALUE_DATA_TYPE, KDCOMPLEX,
                                      KPDS_VALUE_INTERPOLATE,KPAD,NULL));
          }

        if (mask_exists) KCALL(kpds_set_attributes(in_obj,
                                 KPDS_MASK_DATA_TYPE,KDOUBLE,
                                 KPDS_VALUE_INTERPOLATE,KPAD,NULL));
 
        /* See how big the data is */
        KCALL(kpds_get_attributes(in_obj,
                                 KPDS_VALUE_SIZE,&wc,&hc,&dc,&tc,&ec,NULL));

        dang = angle*KPI/180.0; /* Convert to radians */
        csine = cos(dang);
        sine = sin(dang);
        if (angle == 90.0) { csine = 0.0; sine = 1.0; }
        else if (angle == -90.0) { csine = 0.0; sine = -1.0; }
        else if (angle == 180.0 ||
                 angle == -180.0) { csine = -1.0; sine = 0.0; }
        else if (angle == 360.0 ||
                 angle == -360.0) { csine = 1.0; sine = 0.0; }

        if (resize)
          {
            /* Look at all four corners of the input object to see where they
               will map to. Take the largest set of X and Y coordinates for
               the size of the output value region. We don't have to compute
               the absolute coords since we're only interested in the
               span here. Note that the rotation transformation is in the
               reverse direction to those used in the actual rotation code
               below.  */
            x1out = csine*(0-wctr) - sine*(0-hctr);
            y1out = sine*(0-wctr)  + csine*(0-hctr);

            x2out = csine*(wc-wctr) - sine*(0-hctr);
            y2out = sine*(wc-wctr)  + csine*(0-hctr);

            x3out = csine*(wc-wctr) - sine*(hc-hctr);
            y3out = sine*(wc-wctr)  + csine*(hc-hctr);

            x4out = csine*(0-wctr) - sine*(hc-hctr);
            y4out = sine*(0-wctr)  + csine*(hc-hctr);

            xlo = x1out;
            if (x2out < xlo) xlo = x2out;
            if (x3out < xlo) xlo = x3out;
            if (x4out < xlo) xlo = x4out;
            ylo = y1out;
            if (y2out < ylo) ylo = y2out;
            if (y3out < ylo) ylo = y3out;
            if (y4out < ylo) ylo = y4out;
            xhi = x1out;
            if (x2out > xhi) xhi = x2out;
            if (x3out > xhi) xhi = x3out;
            if (x4out > xhi) xhi = x4out;
            yhi = y1out;
            if (y2out > yhi) yhi = y2out;
            if (y3out > yhi) yhi = y3out;
            if (y4out > yhi) yhi = y4out;

            /* Figure out size for output image value seg */
            wout = kceil(xhi)-kfloor(xlo);
            hout = kceil(yhi)-kfloor(ylo);
            dout = dc;
            tout = tc;
            eout = ec;

            wctri = wctr;
            wctro = 0.5*(wout-1);
            hctri = hctr;
            hctro = 0.5*(hout-1);
          }
        else
          {
            wout = wc;
            hout = hc;
            dout = dc;
            tout = tc;
            eout = ec;
            wctri = wctr;
            wctro = 0.5*(wout-1);
            hctri = hctr;
            hctro = 0.5*(hout-1);
          }

        if (!planes)
          {
            /* Set the region sizes */
            wri = 2;  hri = 2;  dri = dc;
            wro = wout; hro = 1;  dro = dout;
          }
        else
          {
            /* Set the region sizes */
            wri = wc; hri = hc; dri = 1;
            wro = wout; hro = hout; dro = 1;
          }

        /* Set up the output object, reference it, and set presentation vals */
        KCALL(kpds_set_attributes(out_obj,
                                KPDS_VALUE_SIZE,wout,hout,dout,tout,eout,NULL));
        if (mask_exists)
          KCALL(kpds_set_attributes(out_obj,
                                 KPDS_MASK_SIZE,wout,hout,dout,tout,eout,NULL));

        KCALL((out_obj = kpds_reference_object(out_obj)) != KOBJECT_INVALID);
        objlist = klist_add(objlist,in_obj,"KOBJECT");
        if (type != KCOMPLEX && type != KDCOMPLEX)
          {
            KCALL(kpds_set_attributes(out_obj,
                                 KPDS_VALUE_DATA_TYPE, KDOUBLE,NULL));
          }
        else
          {
            KCALL(kpds_set_attributes(out_obj,
                                 KPDS_VALUE_DATA_TYPE, KDCOMPLEX,NULL));
          }

        if (mask_exists)
          {
            KCALL(kpds_set_attributes(out_obj,
                                 KPDS_MASK_DATA_TYPE,KDOUBLE,NULL));
            KCALL(kpds_set_attributes(out_obj,
                                 KPDS_MASK_SIZE,wout,hout,dout,tout,eout,NULL));
          }

        /* Set the region sizes on the input and output objects */
        KCALL(kpds_set_attribute(in_obj,
                                  KPDS_VALUE_REGION_SIZE,wri,hri,dri,1,1));
        KCALL(kpds_set_attribute(out_obj,
                                  KPDS_VALUE_REGION_SIZE,wro,hro,dro,1,1));
        if (mask_exists)
          { 
            KCALL(kpds_set_attribute(in_obj,
                                   KPDS_MASK_REGION_SIZE,wri,hri,dri,1,1));
            KCALL(kpds_set_attribute(out_obj,
                                   KPDS_MASK_REGION_SIZE,wro,hro,dro,1,1));
          }

        /* Set up output buffer */
        if (type != KCOMPLEX && type != KDCOMPLEX)
          {
            bd = (double *)kmalloc(wro*hro*dro*sizeof(double));
            objlist = klist_add(objlist,bd,"KMALLOC");
          }
        else
          {
            bd_dc = (kdcomplex *)kmalloc(wro*hro*dro*sizeof(kdcomplex));
            objlist = klist_add(objlist,bd_dc,"KMALLOC");
          }
        if (mask_exists)
          {
            bm = (double *)kmalloc(wro*hro*dro*sizeof(double));
            objlist = klist_add(objlist,bm,"KMALLOC");
          }

        if (!planes)
          {
            for (l=0; l<eout; l++)
              {
                for (m=0; m<tout; m++)
                  {
                    for (i=0; i<hout; i++)
                      {
                        if (type != KCOMPLEX && type != KDCOMPLEX)
                          kdata_fill(bd,(int)(wro*hro*dro),KDOUBLE,pad_real,0.0);
                        else
                          kdata_fill(bd_dc,(int)(wro*hro*dro),KDCOMPLEX,pad_real,pad_imag);
                        if (mask_exists)
                          {
                            if (padvalid == 1)
                              kdata_fill(bm,(int)(wro*hro*dro),KDOUBLE,1.0,0.0);
                            else
                              kdata_fill(bm,(int)(wro*hro*dro),KDOUBLE,0.0,0.0);
                          }
                        for (j=0; j<wout; j++)
                          {
                            kn = wctri + csine*(j-wctro) + sine*(i-hctro);
                            in = hctri - sine*(j-wctro)  + csine*(i-hctro);
                            jl = kfloor(kn); jh = jl+1;
                            il = kfloor(in); ih = il+1;
                            if (in >= 0 && in < hc &&
                                kn >= 0 && kn < wc &&
                                jl >= 0 && jl < wc &&
                                jh >= 0 && jh < wc &&
                                il >= 0 && il < hc &&
                                ih >= 0 && ih < hc)
                              {
/*
                                          z1(il,jl) +---------------+ z2(il,jh)
                                                    |               |
                                                    |               |
                                                    |               |
                                                 za + +zp(in,kn)    + zb
                                                    |               |
                                          z4(ih,jl) +---------------+ z3(ih,jh)
 
 
                                              Note: z1 is at 0+4k
                                                    z2 is at 1+4k
                                                    z4 is at 2+4k
                                                    z3 is at 3+4k
                                 */

                                /* Get a chunk of mask data if it's available */
                                if (mask_exists)
                                  {
                                    bdata = (double *)kdms_get_data(in_obj,KDMS_SEGMENT_MASK,beg,end,(kaddr)bmask);
                                  }

                                /* Go under KPDS to KDMS to speed the large data set process
                                ing case. This is NOT the way to do this in general, but is
                                a last grasp for speed and works by avoiding the overhead
                                imposed by KPDS. This will only work if the index order is
                                whdte, a condition that can be imposed by a kpds_open_object
                                call. This is normally the case for use by a kroutine
                                driver, but could be a problem for folks using this routine
                                in thier own processing code. */
                                beg[0] = jl; beg[1]=il; beg[2]=0;    beg[3]=0; beg[4]=0;
                                end[0] = jh; end[1]=ih; end[2]=dc-1; end[3]=0; end[4]=0;
                                if (type != KCOMPLEX && type != KDCOMPLEX)
                                  {
                                    bdata=(double *)kdms_get_data(in_obj,KDMS_SEGMENT_VALUE,beg,end,(kaddr)bdata);
                                    for (k=0; k<dro; k++)
                                      {
                                        z1 = bdata[4*k+0]; z2 = bdata[4*k+1];
                                        z4 = bdata[4*k+2]; z3 = bdata[4*k+3];
                                        za = z1*(ih-in)+z4*(in-il);
                                        zb = z2*(ih-in)+z3*(in-il);
                                        zp = za*(jh-kn)+zb*(kn-jl);
                                        bd[k*wro+j] = zp;
                                        if (mask_exists)
                                          {
                                            if (bmask[4*k+0] != 0 &&
                                                bmask[4*k+1] != 0 &&
                                                bmask[4*k+3] != 0 &&
                                                bmask[4*k+2] != 0) bm[k*wro+j] = 1;
                                            else bm[k*wro+j] = 0;
                                          }
                                      }
                                  }
                                else /* Complex data case */
                                  {
                                    bdata_dc=(kdcomplex *)kdms_get_data(in_obj,KDMS_SEGMENT_VALUE,beg,end,(kaddr)bdata_dc);
                                    for (k=0; k<dro; k++)
                                      {
                                        z1 = bdata_dc[4*k+0].r; z2 = bdata_dc[4*k+1].r;
                                        z4 = bdata_dc[4*k+2].r; z3 = bdata_dc[4*k+3].r;
                                        za = z1*(ih-in)+z4*(in-il);
                                        zb = z2*(ih-in)+z3*(in-il);
                                        zp = za*(jh-kn)+zb*(kn-jl);
                                        bd_dc[k*wro+j].r = zp;
                                        z1 = bdata_dc[4*k+0].i; z2 = bdata_dc[4*k+1].i;
                                        z4 = bdata_dc[4*k+2].i; z3 = bdata_dc[4*k+3].i;
                                        za = z1*(ih-in)+z4*(in-il);
                                        zb = z2*(ih-in)+z3*(in-il);
                                        zp = za*(jh-kn)+zb*(kn-jl);
                                        bd_dc[k*wro+j].i = zp;
                                        if (mask_exists)
                                          {
                                            if (bmask[4*k+0] != 0 &&
                                                bmask[4*k+1] != 0 &&
                                                bmask[4*k+3] != 0 &&
                                                bmask[4*k+2] != 0) bm[k*wro+j] = 1;
                                            else bm[k*wro+j] = 0;
                                          }
                                      }
                                  }
                              } /* Have completed processing the prism */
                          } /* Have completed process the set of prisms */
                        if (type != KCOMPLEX && type != KDCOMPLEX)
                          {
                            KCALL(kpds_put_data(out_obj,KPDS_VALUE_REGION,(kaddr)bd));
                          }
                        else
                          {
                            KCALL(kpds_put_data(out_obj,KPDS_VALUE_REGION,(kaddr)bd_dc));
                          }
                        if (mask_exists)
                          {
                            KCALL(kpds_put_data(out_obj,KPDS_MASK_REGION,(kaddr)bm));
                          }
                      }
                  }
              }
          }
        else /* Process data by full planes instead of prisms */
          {
/*
		      z1(il,jl)	+---------------+ z2(il,jh) 
                                |               |
                                |               |
                                |               |
                             za + +zp(in,kn)    + zb
                                |               |
		      z4(ih,jl) +---------------+ z3(ih,jh)

             */ 
            for (k=0; k<dout*tout*eout; k++)
              {
                if (type != KCOMPLEX && type != KDCOMPLEX)
                  {
                    bdata=(double *)kpds_get_data(in_obj,KPDS_VALUE_REGION,(kaddr)bdata);
                    kdata_fill(bd,(int)(wro*hro*dro),KDOUBLE,pad_real,(double)0.0);
                  }
                else
                  {
                    bdata_dc=(kdcomplex *)kpds_get_data(in_obj,KPDS_VALUE_REGION,(kaddr)bdata_dc);
                    kdata_fill(bd_dc,(int)(wro*hro*dro),KDCOMPLEX,pad_real,pad_imag);
                  }
                if (mask_exists)
                  {
                    bmask=(double *)kpds_get_data(in_obj,KPDS_MASK_REGION,(kaddr)bmask);
                    if (padvalid == 1)
                      kdata_fill(bm,(int)(wro*hro*dro),KDOUBLE,1.0,0.0);
                    else
                      kdata_fill(bm,(int)(wro*hro*dro),KDOUBLE,0.0,0.0);
                  }
                if (type != KCOMPLEX && type != KDCOMPLEX)
                  {
                    for (i=0; i<hout; i++)
                      {
                        for (j=0; j<wout; j++)
                          {
                            kn = wctri + csine*(j-wctro) + sine*(i-hctro);
                            in = hctri - sine*(j-wctro)  + csine*(i-hctro);
                            jl = kfloor(kn); jh = jl+1;
                            il = kfloor(in); ih = il+1;
                            if (in >= 0 && in < hc &&
                                kn >= 0 && kn < wc &&
                                jl >= 0 && jl < wc &&
                                jh >= 0 && jh < wc &&
                                il >= 0 && il < hc &&
                                ih >= 0 && ih < hc)
                              {
                                z1 = bdata[il*wc+jl]; z2 = bdata[il*wc+jh];
                                z4 = bdata[ih*wc+jl]; z3 = bdata[ih*wc+jh];
                                za = z1*(ih-in)+z4*(in-il);
                                zb = z2*(ih-in)+z3*(in-il);
                                zp = za*(jh-kn)+zb*(kn-jl);
                                bd[i*wout+j] = zp;
                                if (mask_exists)
                                  {
                                    if (bmask[il*wc+jl] != 0 && 
                                        bmask[il*wc+jh] != 0 &&
                                        bmask[ih*wc+jl] != 0 &&
                                        bmask[ih*wc+jh] != 0) bm[i*wout+j] = 1;
                                    else bm[i*wout+j] = 0;
                                  }
                              }
                          }
                      }
                    KCALL(kpds_put_data(out_obj,KPDS_VALUE_REGION,(kaddr)bd));
                    if (mask_exists)
                      KCALL(kpds_put_data(out_obj,KPDS_MASK_REGION,(kaddr)bm));
                  }
                else /* Complex data case */
                  {
                    for (i=0; i<hout; i++)
                      {
                        for (j=0; j<wout; j++)
                          {
                            kn = wctri + csine*(j-wctro) + sine*(i-hctro);
                            in = hctri - sine*(j-wctro)  + csine*(i-hctro);
                            jl = kfloor(kn); jh = jl+1;
                            il = kfloor(in); ih = il+1;
                            if (in >= 0 && in < hc &&
                                kn >= 0 && kn < wc &&
                                jl >= 0 && jl < wc &&
                                jh >= 0 && jh < wc &&
                                il >= 0 && il < hc &&
                                ih >= 0 && ih < hc)
                              {
                                z1 = bdata_dc[il*wc+jl].r; z2 = bdata_dc[il*wc+jh].r;
                                z4 = bdata_dc[ih*wc+jl].r; z3 = bdata_dc[ih*wc+jh].r;
                                za = z1*(ih-in)+z4*(in-il);
                                zb = z2*(ih-in)+z3*(in-il);
                                zp = za*(jh-kn)+zb*(kn-jl);
                                bd_dc[i*wout+j].r = zp;
                                z1 = bdata_dc[il*wc+jl].i; z2 = bdata_dc[il*wc+jh].i;
                                z4 = bdata_dc[ih*wc+jl].i; z3 = bdata_dc[ih*wc+jh].i;
                                za = z1*(ih-in)+z4*(in-il);
                                zb = z2*(ih-in)+z3*(in-il);
                                zp = za*(jh-kn)+zb*(kn-jl);
                                bd_dc[i*wout+j].i = zp;
                                if (mask_exists)
                                  {
                                    if (bmask[il*wc+jl] != 0 &&
                                        bmask[il*wc+jh] != 0 &&
                                        bmask[ih*wc+jl] != 0 &&
                                        bmask[ih*wc+jh] != 0) bm[i*wout+j] = 1;
                                    else bm[i*wout+j] = 0;
                                  }
                              }
                          }
                      }
                    KCALL(kpds_put_data(out_obj,KPDS_VALUE_REGION,(kaddr)bd_dc));
                    if (mask_exists)
                      KCALL(kpds_put_data(out_obj,KPDS_MASK_REGION,(kaddr)bm));
                  } /* End of complex case */
              }
          }

        /* Free up everything that has been allocated */
        kfree_and_NULL(bdata);
        if (mask_exists) kfree_and_NULL(bmask);
        (void)klist_free(objlist, (void (*)(klist *))lkcall_free);
 
        return TRUE;
   
}
/* -library_code_end */
