 /*
  * Khoros: $Id: lmtransform.c,v 1.9.2.4 1997/05/31 19:41:57 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lmtransform.c,v 1.9.2.4 1997/05/31 19:41:57 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for mtransform
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lmtransform
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* Around 10000 vectors seems to be a reasonable comprimise for most
architectures and operating systems. Your timings may vary and might
be improved by playing with this number. Look out for interactions with
the DMS buffer size though. */
#define MAX_VECTORS 10000
 
/* Special mapping macro for 5D access to the data region where
   (wi,hi,di,ti,ei) is the element we want access and
   (w,h,d,t,e) is the dimension of the region. Note that
   the use of this gimmick does not make for the most
   efficient code, but it is at least correct, which is
   the main goal. */
#define PIXEL(wi,hi,di,ti,ei) \
          ((ei)*w*h*d*t+(ti)*w*h*d+(di)*w*h+(hi)*w+(wi))
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lmtransform - transform all E vectors in a data object by a single matrix
* 
*       Purpose: lmtransform multiplies each of the w*h*d*t data vectors in
*                the input vector object by a single WxH matrix and stores the
*                resulting vectors in the output vector object.
*
*         Input: mat_obj - matrix object containing a single WxH matrix
*                in_obj - input vector object, E dimension must be equal to the
*                         W dimension of mat_obj
*
*                The data type of the output object is specified by
*                the data_type_string parameter, and should be one of
*                the following values:
*
*                .TS H
*                center tab(:) ;
*                lfB  lfB  lfB
*                l    l    lf(CW) .
*                Data Type:Abbreviation:Data Type
*                =
*                .TH
*                bit              :bi     : KBIT
*                byte             :by     : KBYTE
*                unsigned byte    :un by  : KUBYTE
*                ubyte            :uby    : KUBYTE
*                short            :sh     : KSHORT
*                unsigned short   :un sh  : KUSHORT
*                ushort           :ush    : KUSHORT
*                integer          :in     : KINT
*                unsigned integer :un in  : KUINT
*                uint             :ui     : KUINT
*                long             :lon    : KLONG
*                unsigned long    :un lon : KULONG
*                ulong            :ul     : KULONG
*                float            :fl     : KFLOAT
*                double           :do     : KDOUBLE
*                complex          :co     : KCOMPLEX
*                double complex   :do co  : KDCOMPLEX
*                dcomplex         :dc     : KDCOMPLEX
*                .T&
*                l s s.
*                Propagate Input Type - will not change the data type
*                .TE
*
*        Output: out_obj - output vector object, E dimension will be equal to
*                          the H dimension of mat_obj
*
*       Returns: TRUE (1) on success, FALSE (0) otherwise
*
*  Restrictions: Restrictions on data or input as applicable
*    Written By: 
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lmtransform(
kobject mat_obj,
kobject in_obj,
char *data_type_string,
kobject out_obj)
/* -library_def_end */

/* -library_code */
{
        char *lib = "kmatrix", *rtn = "lmtransform";
        klist *objlist=NULL;
        int nvecs,nrgns;
        int n,r,c;
        int type_m,type_i,ttype,otype;
        int w,h,d,t,e;
        int wi,hi,di,ti;
        int wc,hc,dc,tc,ec;             /* Size of vector input object */
        int wc_m,hc_m,dc_m,tc_m,ec_m;   /* Size of matrix input object */
        double *A=NULL,*ivec=NULL,*ovec=NULL;
        kdcomplex *AZ=NULL,*ivecZ=NULL,*ovecZ=NULL;
        kdcomplex sumZ;
        double sum;

        /* Make sure we have a valid input and output objects */
        if (mat_obj == KOBJECT_INVALID )
          {
            kerror(lib, rtn, "Bogus input matrix object");
            return(FALSE);
          }
        if (in_obj == KOBJECT_INVALID )
          {
            kerror(lib, rtn, "Bogus input vector object");
            return(FALSE);
          }
        if (out_obj == KOBJECT_INVALID)
          {
            kerror(lib, rtn, "Bogus output vector object");
            return(FALSE);
          }

        /* Reference the input objects to avoid side effects, then add the
           references to the list of goodies to be automatically free'd on
           error. */
        KCALL((mat_obj = kpds_reference_object(mat_obj)));
        objlist = klist_add(objlist,mat_obj,"KOBJECT");
        KCALL((in_obj = kpds_reference_object(in_obj)));
        objlist = klist_add(objlist,in_obj,"KOBJECT");
 
        /* See if the data type of the input objects are real or complex */
        KCALL(kpds_get_attribute(mat_obj,KPDS_VALUE_DATA_TYPE,&type_m));
        KCALL(kpds_get_attribute(in_obj, KPDS_VALUE_DATA_TYPE,&type_i));
        if (type_i == KCOMPLEX || type_m == KDCOMPLEX ||
            type_m == KCOMPLEX || type_m == KDCOMPLEX) ttype = KDCOMPLEX;
        else ttype = KDOUBLE;

        /* See what type of output data is desired */
        if (kstrcasecmp(data_type_string, "Propagate input type") == 0)
           otype = type_i;
        else if ((otype = kdatatype_to_define(data_type_string)) == 0)
        {
           kerror(lib, rtn, "Not a proper data type '%s'.", data_type_string);
           (void)klist_free(objlist, (void(*)(klist *))lkcall_free);
           return (FALSE);
        }

        /* Constrain the input data to double or double complex for read */
        KCALL(kpds_set_attribute(in_obj, KPDS_VALUE_DATA_TYPE, ttype));
        KCALL(kpds_set_attribute(mat_obj,KPDS_VALUE_DATA_TYPE, ttype));
 
        /* See how big the data is */
        KCALL(kpds_get_attribute(in_obj,KPDS_VALUE_SIZE,&wc,&hc,&dc,&tc,&ec));

        /* See how big the matrix is */
        KCALL(kpds_get_attribute(mat_obj,KPDS_VALUE_SIZE,
                                         &wc_m,&hc_m,&dc_m,&tc_m,&ec_m));
        if (dc_m != 1 || tc_m != 1 || ec_m != 1)
          {
            kerror(lib, rtn, "Matrix must be WxH only");
            (void)klist_free(objlist, (void(*)(klist *))lkcall_free);
            return(FALSE);
          }

        /* Make sure the matrix and vectors are conformal */
        if (ec != wc_m)
          {
            kerror(lib, rtn, "Vector size (E dimension of input vector object)"
                             "must match the number of colums in the matrix");
            (void)klist_free(objlist, (void(*)(klist *))lkcall_free);
            return(FALSE);
          }

        /* Come up with some kind of chunk size that makes sense */
        w=wc; h=1; d=1; t=1; e=ec;
        if (wc*hc < MAX_VECTORS) h=hc;
        if (wc*hc*dc < MAX_VECTORS) d=dc;
        if (wc*hc*dc*tc < MAX_VECTORS) t=tc;
        nrgns = wc*hc*dc*tc/(w*h*d*t);
        nvecs = wc*hc*dc*tc;
 
        /* Set up to process input by vectors */
        KCALL(kpds_set_attribute(in_obj,KPDS_VALUE_REGION_SIZE,w,h,d,t,e));
 
        /* Set the output object sizes and data type */
        if (!kpds_query_value(out_obj)) KCALL(kpds_create_value(out_obj));
        KCALL(kpds_set_attributes(out_obj,
                             KPDS_VALUE_SIZE,wc,hc,dc,tc,hc_m,
                             KPDS_VALUE_DATA_TYPE,otype,NULL));
        KCALL((out_obj = kpds_reference_object(out_obj)));
        objlist = klist_add(objlist,out_obj,"KOBJECT");
 
        /* Set the presentation data type to match that used internally */
        KCALL(kpds_set_attributes(out_obj,KPDS_VALUE_DATA_TYPE,ttype,NULL));
 
        /* Set up to write the output as the transformed vectors */
        KCALL(kpds_set_attribute(out_obj,KPDS_VALUE_REGION_SIZE,w,h,d,t,hc_m));

        /* Allocate buffers */
        if (ttype == KDCOMPLEX)
          {
            /* Transform matrix */
            KCALL(!((AZ=(kdcomplex *)kmalloc(wc_m*hc_m*sizeof(kdcomplex)))
                       == NULL));
            objlist = klist_add(objlist,AZ,"KMALLOC");

            /* Input vector buffer */
            KCALL(!((ivecZ=(kdcomplex *)kmalloc(w*h*d*t*e*sizeof(kdcomplex)))
                         == NULL));
            objlist = klist_add(objlist,ivecZ,"KMALLOC");

            /* Output vector buffer */
            KCALL(!((ovecZ=(kdcomplex *)kmalloc(w*h*d*t*hc_m*sizeof(kdcomplex)))
                         == NULL));
            objlist = klist_add(objlist,ovecZ,"KMALLOC");

          }
        else
          {
            /* Transform matrix */
            KCALL(!((A=(double *)kmalloc(wc_m*hc_m*sizeof(double)))== NULL));
            objlist = klist_add(objlist,A,"KMALLOC");

            /* Input vector buffer */
            KCALL(!((ivec=(double *)kmalloc(w*h*d*t*e*sizeof(double)))== NULL));
            objlist = klist_add(objlist,ivec,"KMALLOC");

            /* Output vector buffer */
            KCALL(!((ovec=(double *)kmalloc(w*h*d*t*hc_m*sizeof(double)))== NULL));
            objlist = klist_add(objlist,ovec,"KMALLOC");
          }
 
        for (n=0; n<nrgns; n++)
          {
            if (ttype == KDCOMPLEX)
              {
                KCALL(kpds_get_data(mat_obj,KPDS_VALUE_ALL,(kaddr)AZ));
                KCALL(kpds_get_data(in_obj,KPDS_VALUE_REGION,(kaddr)ivecZ));
                for (wi=0; wi<w; wi++)
                  {
                    for (hi=0; hi<h; hi++)
                      {
                        for (di=0; di<d; di++)
                          {
                            for (ti=0; ti<t; ti++)
                              {
                                /* Multiply the matrix onto the vector
                                   originating at (wi,hi,ti,di) */
                                for (r=0; r<hc_m; r++)
                                  {
                                    sumZ.r = 0;
                                    sumZ.i = 0;
                                    for (c=0; c<wc_m; c++)
                                      {
                                        sumZ = kdcadd(sumZ,
                                               kdcmult(AZ[r*wc_m+c],
                                               ivecZ[PIXEL(wi,hi,di,ti,c)]));
                                      }
                                    ovecZ[PIXEL(wi,hi,di,ti,r)] = sumZ;
                                  }
                              }
                          }
                      }
                  }
                KCALL(kpds_put_data(out_obj,KPDS_VALUE_REGION,(kaddr)ovecZ));
              }
            else
              {
                KCALL(kpds_get_data(mat_obj,KPDS_VALUE_ALL,(kaddr)A));
                KCALL(kpds_get_data(in_obj,KPDS_VALUE_REGION,(kaddr)ivec));
                for (wi=0; wi<w; wi++)
                  {
                    for (hi=0; hi<h; hi++)
                      {
                        for (di=0; di<d; di++)
                          {
                            for (ti=0; ti<t; ti++)
                              {
                                /* Multiply the matrix onto the vector
                                   originating at (wi,hi,ti,di) */
                                for (r=0; r<hc_m; r++)
                                  {
                                    sum = 0;
                                    for (c=0; c<wc_m; c++)
                                      {
                                        sum += A[r*wc_m+c]*
                                               ivec[PIXEL(wi,hi,di,ti,c)];
                                      }
                                    ovec[PIXEL(wi,hi,di,ti,r)] = sum;
                                  }
                              }
                          }
                      }
                  }
                KCALL(kpds_put_data(out_obj,KPDS_VALUE_REGION,(kaddr)ovec));
              }
          }
 
        (void)klist_free(objlist, (void(*)(klist *))lkcall_free);
 
	return TRUE;
}
/* -library_code_end */
