 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * Copyright (C) 1993-1997 Junior Barrera, Roberto Lotufo.
 * All rights reserved.  See $MMACH/repos/license/License.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vcast1
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvcast1
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvcast1 - Upconvert all images to the highest input image data type.
* 
*       Purpose: .I vcast1
This is a modified version of Khoros' vcast. Scaling and Normalization are turned off, so that BIT to BYTE_1 conversion yields pixel values of 0 or 255 on the BYTE image.
will upconvert or upcast each input image to the highest input image
data type.  Optionally, the user may elect to upcast all images to 
a specified data type, as long as it is a data type that is equal to
or greater than the highest data type of the input images.  If the
user tries to select a data type that is lower than the highest 
data type of the input images, an error message will be produced
indicating that down casting is not allowed.
.LP
There are no restrictions as to the input image data types.  The data
types can range from BIT to DOUBLE.  The input images may contain any
combination of sizes or number of data bands.  All map information from 
each input image is transferred to the corresponding output image. 
.LP
Note that when converting from type BIT to BYTE, the output values
will be 0 or 1.  This is different from the results one would obtain
using the defaults with vconvert.  In trying to preserve the 
integrity of the data, vcast1 scales the data by 1.  This routine
calls lvconvert to do the actual up convert/cast of the image data.
.LP
The inputs to lvcast1 include:
.LP
xvimage *images[]
.IP
This is an array containing pointers to all of the images that are to 
be upcast.  There is no limit on the number of images to convert.
.LP
int image_num
.IP
This is the number of pointers to images.  That is, the size of *images[].
.LP
int up_type
.IP
This is the optional target data type.  The target data type is specified
by passing in the integer representing the desired VIFF data type.  If
auto upconvert is to be used, then passing in the value "-1" will default
to finding the highest data type of the input images and upcasting the
remaining images to this data type.
.I lvcast1
will return a 1 upon success and a 0 upon failure.
*         Input: .IP "images" 15  
array of pointers to xvimage structures
.IP "image_num" 15
integer indicating the number of input images
.IP "up_type" 15
optional target data type.  
Pass in "-1" to select auto upconvert.
*        Output: .IP "images" 15
array of pointers to resulting xvimage structures.
.IP "Return Value:" 15
1 on success, 0 on failure.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: 
*    Written By: Charlie Gage
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Adriano N. Rodrigues: turned off scaling and normalization, so that bit to byte conversion yields pixel values of 0 or 255.
****************************************************************/
/* -library_def */
int
lvcast1(xvimage **images, int image_num, int up_type)
/* -library_def_end */

/* -library_code */
{
 int chg_flag,
     prev_type,
     max_type,
     type,
     i;

 chg_flag = 0;
 max_type = 0;
 prev_type = (int) images[0]->data_storage_type;

   /*-------------------------------------------------------
    * Determine highest data type of input images 
    *------------------------------------------------------*/
 for (i=0; i < image_num; i++)
 {
   type = (int) images[i]->data_storage_type;
   if (type > max_type)
   {
     max_type = type;
   }
          /* Set flag if any data type differs from any other */
   if (prev_type != type)
   {
     chg_flag = 1;
   }
   prev_type = type;
 }

   /*------------------------------------------------------
    * Determine if requested data type is equal to
    *     or greater highest data type of input images.
    *-----------------------------------------------------*/
 if (up_type != -1)
 {
   if (up_type >= max_type)
   {
     max_type = up_type;
     chg_flag = 1;
   }
   else
   {
     (void)kfprintf(kstderr, "lvcast1: You have requested a data type that\n");
     (void)kfprintf(kstderr, "is lower than the highest data type of the\n");
     (void)kfprintf(kstderr, "input images.  Unable to down cast data.\n");
     return(0);
   }
 }
   
   /*------------------------------------------------------
    * If images are of different data types, 
    *  then upconvert/cast them.
    *-----------------------------------------------------*/
 if (chg_flag == 1)
 {
   for (i=0; i < image_num; i++)
   {
                   /* Upconvert only those images that are not of max_type */
     if ((int) images[i]->data_storage_type != max_type)
     {
                       /* Call lvconvert to upconvert to max data type */
       if (lvconvert(images[i], max_type, 0, 0, 0, 0, 0) == 0)
       {
         (void)kfprintf(kstderr,"lvcast1: ERROR, lvconvert failed\n");
         return(0);
       }
     }
   }
 }
 else           /*** Do nothing if all images are of the same data type ***/
   return(1);

 return(1);
}  /*** End of lvcast1 ***/
/* -library_code_end */
