 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * Copyright (C) 1993-1997 Junior Barrera, Roberto Lotufo.
 * All rights reserved.  See $MMACH/repos/license/License.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vdil
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvdil
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvdil - Do the dilation of an image by a structuring element, with an operation gating mask.
* 
*       Purpose: .I lvdil
will do the dilation of an image by a structuring element.
img1 and b are, respectively, the input image and the structuring element;
img2 is the gating image, and mflg should be set to 1 if the gating image 
is to be used.
A non-zero mask pixel in img2 enables the output pixel to contain 
the dilation. A zero mask pixel in img2 just transfers the value 
of the corresponding pixel in img1 to the output pixel.
.LP
img3 is the output image. Note that img1 and img3 must point to different image structures.
The input, output and mask images must be of the same size, data type, and have 
the same number of data bands.
*         Input: .IP "img1" 15  
input image 
.IP "ker" 15
structuring element
.IP "img2" 15
gating mask image
.IP "mflg" 15
a flag set (equal 1) if gating image mask available
.IP "border" 15
mask to the last pixel of a row (used in formatted BIT images)
.IP
Note that the input and mask images must be of the same data type and size.
*        Output: .IP "img3" 15
holds the result of the dilation operation. The output data
type is the same as the input data type.
.IP "Return Value:" 15
1 on success, 0 on failure.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: .I lvdil
can be defined for all data types supported by Khoros, but at the moment it has been implemented just for the bit and unsigned char types.
The structuring elements are subsets of the 3x3 matrix and the origin is always at the center of this matrix.
.LP
Due to the simultaneous processing of 32 pixels in BIT images, input images of the BIT data type must be converted to a special data structure before the library is called. This can be accomplished by using the format routine. After the library call, the unformat routine can convert the image back to the BIT data type.
*    Written By: Junior Barrera & Roberto Lotufo & Ulisses Braga
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Bug fixes and MAT code optimization by Adriano Nagelschmidt Rodrigues & Luiz Fernando Vieira
*		 Memory allocated by kmalloc() is now properly unallocated by kfree() - E. Trettel <trettel@fee.unicamp.br>
****************************************************************/
/* -library_def */
int
lvdil(xvimage *img1, mat_str *ker, xvimage *img2, int mflg, unsigned long border, xvimage *img3)
/* -library_def_end */

/* -library_code */
{
    /*
     * Common variables
     */
    register int i,j,k;
    char     *program = "lvdil";

    /*
     *
     * Switch between element types
     *
     */
    switch(ker->tag) {

    case (MM_IMG): 
    {

    /*
     *
     * Variables for the IMG structuring element case
     *
     */
    int  type,        /* type of input image                 */
         cols,        /* number of columns of image          */ 
         rows,        /*   "    "  rows    "    "            */
         bands,       /*   "    "  bands   "    "            */
         sizeband,    /* size of each image band             */
         kerncols,    /* number of columns of kernel         */
         kernrows,    /* number of rows  of kernel           */
         kernbands,   /* number of bands of kernel           */
         sizekerband, /* size of each kernel band            */
         sizekernel,  /* total size of kernel                */
         sizebandout, /* size of each output buffer band     */
         sizeout,     /* size of output buffer               */
         kerdata,     /* a single value from the kernel      */
         colsbytes,   /* number of columns in bytes          */ 
         colslongs,   /* number of columns in longs          */ 
         cols_pad,    /* number of columns of line buffer    */
         rows_pad,    /* number of rows  of output buffer    */
         bands_pad;   /* number of bands of output buffer    */

    int  centerx,     /* x front upper left corner of kernel */
         centery,     /* y   "     "    "     "     "   "    */
         centerz,     /* z   "     "    "     "     "   "    */
         rightx,      /* x back lower right corner of kernel */
         downy,       /* y  "     "     "     "    "    "    */
         backz,       /* z  "     "     "     "    "    "    */
         offsetx,     /* displacement in the x direction     */
         offsety,     /*      "       "   "  y     "         */
         offsetz,     /*      "       "   "  z     "         */
         first_el,    /* flag to read new line of image      */
         first_col,   /* flag to do new shift of line        */
         pcx,         /* used in the shift for the bit case  */
         cxoff,       /*  "   "   "    "    "   "   "   "    */
         offoff,      /*  "   "   "    "    "   "   "   "    */
         offbits,     /*  "   "   "    "    "   "   "   "    */
         shift,       /*  "   "   "    "    "   "   "   "    */
         sum,         /* used in the max operation           */
         bits,        /* number of bits  in a long           */
         bytes,       /*   "    "  bytes "  "  "             */
         sizelong;    /* size of a long in bytes             */

    register int  x,y,z;
    unsigned char c1,c2;
    unsigned long l1,l2;
    unsigned char *pi;       /* pointer to input image  */
    short         *pk;       /* pointer to kernel image */

    /* buffer usage:
      1. line   buffer: shift processing
      2. output buffer: where the output image is built */

    /* BYTE case: */
    unsigned char *line,      /* line buffer */
                  *pl,        /* point to line buffer */    
                  *outimage,  /* output buffer */
                  *po;        /* point to output buffer */    

    /* BIT case:  */
    unsigned long *lline,      /* line buffer */
                  *ppl,        /* point to line buffer */    
                  *loutimage,  /* output buffer */
                  *ppo,        /* point to output buffer */    
                  *linesh;     /* Copy of line buffer to do the shifting */

    /*
     *
     * Code for IMG structuring element type 
     *
     */

    /* Check input image */
    type  = img1->data_storage_type;
    if ((type != VFF_TYP_BIT) && (type != VFF_TYP_1_BYTE)) { 
      kfprintf(kstderr,"%s: input image must be of type byte or bit\n",program); 
      return (0);
      } 
    if (!(proper_num_images(program, img1, 1, FALSE))) {
       (void) kfprintf (kstderr,"\n\n%s:   ", program);
       (void) kfprintf (kstderr,"Can only work on files with one image\n\n");
       return (0);
    }

    /* input image dimensions: */
    cols  = img1->row_size;       /* number of columns */
    rows  = img1->col_size;       /* number of rows    */
    bands = img1->num_data_bands; /* number of bands   */

    /* kernel dimensions: */
    kernrows    = ker->m;       /* number of rows    */
    kerncols    = ker->n;       /* number of columns */
    kernbands   = ker->p;       /* number of bands   */   
    sizekerband = kernrows*kerncols;      /* size of each band    */
    sizekernel  = kernbands*sizekerband;  /* total size of kernel */

/*  pi: the input image is read just once, sequentially, during the 
        processing of the program. This makes the algorithm faster */

    pi = (unsigned char *) img1->imagedata; 
    pk = (short *) ker->dat;                 

/*  The origin of the kernel is assumed to be on its center 
    In case of even length, the center is offset to the right */

    centerx = kerncols/2;
    centery = kernrows/2;
    centerz = kernbands/2;
    rightx  = (kerncols-1)/2;
    downy   = (kernrows-1)/2;
    backz   = (kernbands-1)/2;

/*
    Below is the outline of the output buffer (with up and down borders):

     +-------------------+ ^
     |                   | .
     |      centery      | .
     |                   | .
     +-------------------+ .   output buffer
     |                   | .
     |      original     | rows_pad
     |       image       | .
     |                   | .
     +-------------------+ .
     |                   | .
     |       downy       | .
     |                   | .
     +-------------------+ v
     <-------cols-------->

for the 3D case, there are front and back borders as well (not displayed
above), of size centerz and backz.
 
    Below is the line buffer (with left and right borders):

     +-----------+-------------------+----------+
     |  centerx  | one line of image |  rightx  |   line buffer
     +-----------+-------------------+----------+
     <-----------------cols_pad----------------->
*/

    if (type==VFF_TYP_BIT) {  /* BIT case */
    
    /* length of long for current architecture: */
    bytes = sizeof(long);
    bits  = 8*bytes;

    /* size of the output and line buffers in packed format: */
    rows_pad    = rows + kernrows-1;            /* x lenght */
    cols_pad    = (cols + kerncols-1 + bits-1)/bits + 1; /* y lenght */
    bands_pad   = bands + kernbands-1;          /* z lenght */
    colsbytes   = (cols +  7)/8;                /* number of columns in bytes     */
    colslongs   = (cols + bits-1)/bits;         /* number of columns in longs     */
    sizeband    = rows*colsbytes;               /* size in bytes of image band    */
    sizebandout = rows_pad*colslongs;           /* size in longs of output band   */
    sizeout     = rows_pad*colslongs*bands_pad; /* size in longs of output buffer */

    /* Indexes into the line buffer: */
    pcx   = (centerx + bits-1)/bits;
    if (centerx)
      cxoff = bits - centerx%bits;
    else
      cxoff = 0;

    /* dynamic memory allocation for buffers: */
    lline = (unsigned long *) kmalloc(cols_pad*sizeof(unsigned long));
    if (lline == NULL) {
        (void) kfprintf (kstderr, "\n%s bit, lline %d:", program, cols_pad);
        (void) kfprintf (kstderr, "Insufficient Space avaliable\n");
        return(0);
    }
    linesh = (unsigned long *) kmalloc(cols_pad*sizeof(unsigned long));
    if (linesh == NULL) {
        (void) kfprintf (kstderr, "\n%s bit,linesh %d:   ", program, cols_pad);
        (void) kfprintf (kstderr, "Insufficient Space avaliable\n");
        return(0);
    }
    loutimage =(unsigned long *) kmalloc(sizeout*sizeof(unsigned long));
    if (loutimage == NULL) {
        (void) kfprintf(kstderr, "\n%s bit, loutimage %d: ", program, sizeout);
        (void) kfprintf(kstderr, "Insufficient Space avaliable\n");
        return(0);
    }

    /* Initialize the buffers for the dilation case: */
    for(i=0; i<sizeout;  i++) loutimage[i] = 0;
    for(i=0; i<cols_pad; i++) lline[i] = 0;

/* ---------------------------------------------------------------------
   Order of loops:
   1. For each row
   2. For each kernel column
      : shift just once right or left  -> this makes the algor. faster
   3. For each kernel row
   4. For each column

   Another trick to make the algor. faster is to do the shift only when
   is necessary, this is implemented with the help of flags:
   flag first_el : indicate the reading of another input line
   flag first_col: indicate to do another shift
   ---------------------------------------------------------------------
*/
    for(k=0; k<bands; k++) 
    {
      for(j=0; j<rows; j++)
      {
        first_el = 1;                 /* flag to indicate to read another line */
        for(x=-centerx; x<=rightx; x++)
        {
          first_col = 1;        /* flag to indicate to make another shift */
          offsetx = centerx - x ;
          for(y=-centery; y<=downy; y++)
          {
            for(z=-centerz; z<=backz; z++)
            {
              /* Test kernel: */          
              if (*pk>=0)
              /* kernel differ from background value, so has comp. to do */
              {
                offsety = centery + y;
                offsetz = centerz + z; 
                if (first_el) /* read one line of input image */
                {
                  first_el = 0;
                  ppl = lline + pcx;
                  l2 = 0;
                  shift = bytes;
                  for (i=0; i<colsbytes; i++) {
                    l1 = *pi++ << 8*--shift;
                    l2 |= l1;
                    if (shift==0) {
                    /*bitreverse(&l2,1,1,bits);  reverse bits from khoros format*/
                      *ppl++ = l2;
                      l2 = 0;
                      shift = bytes;
                      }
                    }
                  if (shift!=bytes) { /* process last long */
                  /*bitreverse(&l2,1,1,bits);*/
                    *ppl = l2;
                    }
                }
                if (first_col)     /* shift right or left direction */
                {                  /* shift just once; */
                  first_col = 0;   /* for the BIT case, this is good */
                  /* Copy lline into linesh: */
                  (void) kbcopy((char *)lline,(char *)linesh,
                                 cols_pad*sizeof(unsigned long));
                  offoff  = offsetx + cxoff;
                  offbits = offoff%bits; 
                  if (offbits) { /* if offoff%bits=0, there is no work to do */
                    for (i=offoff/bits; i<cols_pad-1; i++) {
                      l1 = linesh[i] << offbits;
                      l2 = linesh[i+1] >> (bits-offbits);
                      linesh[i] = l1|l2;
                      }
                    linesh[cols_pad-1] <<= offbits; /* last long must be treated separately */
                    }
                }
                ppl = linesh + offoff/bits;
                ppo = loutimage + (offsety+j)*colslongs + (offsetz+k)*sizebandout;
                for(i=0; i<colslongs; i++)
                {
                  *ppo++ |= *ppl++;
                }
              }
            pk += sizekerband;
            }
          pk -= sizekernel - kerncols;
          }
        pk -= sizekerband - 1;
        }
      pk -= kerncols;
      }
    }

    /* put back the inner image, discarding the borders */
    pi = (unsigned char *) img3->imagedata; /* point to output image's pixels */
    ppo = loutimage + centerz*sizebandout;
    for(k=0; k<bands; k++) {
      ppo+=centery*colslongs;
    /*bitreverse(ppo,rows,colslongs,bits); reverse bits to khoros format*/ 
      for(j=0; j<rows; j++) {
        shift = bytes;
        for(i=0; i<colsbytes; i++) {
          *pi++ = (unsigned char) (*ppo >> 8*(--shift)) & 0xFF;
          if (shift==0) { 
            shift = bytes;
            ppo++;
            }
          }
        if (shift!=bytes) ppo++;
        }
      ppo+=downy*colslongs;
      }
     (void)kfree((kaddr)lline);
     (void)kfree((kaddr)linesh);
     (void)kfree((kaddr)loutimage);
    }

    else {  /* BYTE case */

    /* sizes for the output and line buffers: */
    rows_pad    = rows + kernrows-1;       /* x lenght */
    cols_pad    = cols + kerncols-1;       /* y lenght */
    bands_pad   = bands + kernbands-1;     /* z lenght */
    sizeband    = rows*cols;               /* size in bytes of band          */
    sizebandout = rows_pad*cols;           /* size in bytes of output band   */
    sizeout     = rows_pad*cols*bands_pad; /* size in bytes of output buffer */

    /* dynamic memory allocation for buffers */
    line = (unsigned char *) kmalloc(cols_pad*sizeof(unsigned char));
    if ( line == NULL) {
        (void) kfprintf (kstderr, "\n%s byte, line %d:   ", program, cols_pad);
        (void) kfprintf (kstderr, "Insufficient Space avaliable\n");
        return(0);
    }
    outimage = (unsigned char *) kmalloc(sizeout*sizeof(unsigned char));
    if (outimage == NULL) {
        (void) kfprintf (kstderr, "\n%s byte, outimage %d:   ", program, sizeout);
        (void) kfprintf (kstderr, "Insufficient Space avaliable\n");
        return(0);
    }

    /* Initialize buffers for the dilation case: */
    for(i=0; i<sizeout;  i++) outimage[i] = 0;
    for(i=0; i<cols_pad; i++) line[i] = 0;

/* ---------------------------------------------------------------------
   Order of loops:
   1. For each row
   2. For each kernel column
      : shift just once right or left  -> this makes the algor. faster
   3. For each kernel row
   4. For each column

   Another trick to make the algor. faster is to do the shift only when
   is necessary, this is implemented with the help of flags:
   flag first_el : indicate the reading of another input line
   flag first_col: indicate to do another shift
   ---------------------------------------------------------------------
*/
    for(k=0; k<bands; k++) 
    {
      for(j=0; j<rows; j++)
      {
        first_el = 1;            /* flag to indicate to read another line */
        for(x=-centerx; x<=rightx; x++)
        {
          offsetx = centerx - x ;
          for(y=-centery; y<=downy; y++) 
          {
            for(z=-centerz; z<=backz; z++)
            {
              /* Test kernel: */          
              kerdata = *pk; 
              if (kerdata>=0)
              /* kernel differ from background value, so has comp. to do */
              {
                offsety = centery + y;
                offsetz = centerz + z;
                if (first_el) /* read one line of input image */
                {
                  pl = line + centerx;
                  first_el = 0;
                  /* read one line of input image */
                  (void) kbcopy((char *)pi,(char *)pl,cols); 
                  pi+=cols;
                }
                pl = line + offsetx;   /* shift right or left direction */

                po = outimage + (offsety+j)*cols + (offsetz+k)*sizebandout;
                if (kerdata) {
                  /* Non-planar case: */  
                  for(i=0; i<cols;i++)
                  {
                    if (*pl) { 
                      sum = ((255-*pl)>kerdata) ? *pl+kerdata : 255;
                      if (sum > *po) *po = sum;
		      }
                    pl++;
                    po++;
                  }
                }
                else {
                  /* Planar case: */
                  for(i=0; i<cols;i++)
                  {
                    if (*pl > *po) *po = *pl;
                    pl++;
                    po++;
                  }
		}
              }
            pk += sizekerband;
            }
          pk -= sizekernel - kerncols;
          }
        pk -= sizekerband - 1;
        }
      pk -= kerncols;
      }
    }

    /* put back the inner image, discarding the borders */
    pi = (unsigned char *) img3->imagedata; /* point to output image's pixels */
    for (k=0; k<bands; k++) {
      po = outimage + centery*cols + (centerz+k)*sizebandout;
      (void) kbcopy((char *)po,(char *)pi,sizeband);
      pi += sizeband;
      }
    (void)kfree((kaddr)line);
    (void)kfree((kaddr)outimage);
    }
    /*
     * End of IMG code
     */

    break;
    }

    case (MM_MAT):
    {

    /*
     *
     * Variables for the MAT structuring element case
     *
     */ 
    unsigned char *c1,  /* output image */
                  *c2,  /* input image  */
                  *c3;  /* mask image   */

    unsigned long *l1,  /* output image */
                  *l2,  /* input image  */
                  *l3,  /* mask image   */
                  *l1r, *l2r; /* auxiliar vars. */        

    register int nc;    /* number of columns of image */
    int nb,             /* number of bands */
        nr,             /* number of rows */
        npb,            /* number of pixels in a band of an image */
        tnp;            /* total number of pixels */

    int col32;          /* number of columns of the compacted image */
    int jj;             /* evaluated pixel */  
    int row;            /* number of rows computed */
 
    register int  y0, y1, y2, y3, y4, y5, y6, y7, y8, n;
    int      d[9];
    unsigned long xx;      /* auxiliar word */
    register unsigned char x, y;
    int l, h, nr1, nc1, nc2, nr2;           /* auxiliar variables */
    vec_str b_vec;
    int knc, kncj, jinf, jsup;
    unsigned char *ref, *ref1, *ref2, *ref3, *ref4, *ref5, *ref6;

    /*
     *
     * Code for MAT structuring element type 
     *
     */

    if (img1 == img3) {      
      kfprintf(kstderr,"lvdil: Input and Output must be different image structures.\n"); 
      return(0);
    }       
    
    switch (img1->data_storage_type) {
        case VFF_TYP_BIT :                          /* bit image */

          border = ~ border;
          nr = img1->col_size;
          nb = img1->num_data_bands; 
          if ((img1->row_size < 65) || (img1->col_size < 3)) {
            kfprintf(kstderr,"lvdil: BIT images should be at least 65 pixels wide and 3 pixels high.\n"); 
            return(0);
          }            
          col32 = (img1->row_size + 31) >> 5;
          l2r = (unsigned long *)(img1->imagedata);
          l1r = (unsigned long *)(img3->imagedata);
          npb = nr * col32;                       
          k = 0;
          for (i=0; i<3; i++) {
            for (j=0; j<3; j++) {
              if (ker->x[j][i]) k++;
            }
          }


            y0 = ker->x[2][2];
            y1 = ker->x[1][2];
            y2 = ker->x[0][2];
            y3 = ker->x[2][1];
            y4 = ker->x[1][1];   /* flags to the pos. of the struct element */
            y5 = ker->x[0][1]; 
            y6 = ker->x[2][0];
            y7 = ker->x[1][0];
            y8 = ker->x[0][0];

            for (l=0; l<nb; l++) {
              l2 = &(l2r[l * npb]);
              l1 = &(l1r[l * npb]);              
   
              xx = 0;                                 /* REGION 1 */       
              if (y0 || y1 || y2) xx = 0;
              if (y3) xx |= (l2[0] >> 1);
              if (y4) xx |= l2[0];
              if (y5) xx |= (l2[0] << 1) | ((l2[1] & 0x80000000L) ? 1: 0);
              if (y6) xx |= (l2[col32] >> 1);
              if (y7) xx |= l2[col32];
              if (y8) xx |= (l2[col32] << 1) | ((l2[col32 + 1] & 0x80000000L) ? 1: 0);
              l1[0] = xx;
               
                                        
            for (j=1; j<col32 - 1; j++) {              /* REGION 2 */
              xx = 0;                          
              if (y0 || y1 || y2) xx = 0;
              if (y3) xx |= (l2[j] >> 1) | ((l2[j - 1] & 1) ? 0x80000000L: 0);
              if (y4) xx |= l2[j];
              if (y5) xx |= (l2[j] << 1) | ((l2[j + 1] & 0x80000000L) ? 1: 0);
              if (y6) xx |= (l2[j + col32] >> 1) | ((l2[j + col32 - 1] & 1) ? 0x80000000L: 0);
              if (y7) xx |= l2[j + col32];
              if (y8) xx |= (l2[j + col32] << 1) | ((l2[j + col32 + 1] & 0x80000000L) ? 1: 0);
              l1[j] = xx;
            }


                                                                                                                     
              xx = 0;                                 /* REGION 3 */
              if (y0 || y1 || y2) xx = 0;
              if (y3) xx |= (l2[col32 - 1] >> 1) | ((l2[col32 - 2] & 1) ? 0x80000000L: 0);
              if (y4) xx |= l2[col32 - 1];
              if (y5) xx |= (l2[col32 - 1] << 1) & border;
              if (y6) xx |= (l2[col32 - 1 + col32] >> 1) | ((l2[col32 + col32 - 2] & 1) ? 0x80000000L: 0);
              if (y7) xx |= l2[col32 - 1 + col32];
              if (y8) xx |= (l2[col32 - 1 + col32] << 1) & border;
              l1[col32 - 1] = xx;



            row = col32;                               /* REGION 4 */          
            for (i=1; i < nr - 1; i++) {                            
              xx = 0;
              if (y0) xx |= (l2[row - col32] >> 1);
              if (y1) xx |= l2[row - col32];
              if (y2) xx |= (l2[row - col32] << 1) | ((l2[row - col32 + 1] & 0x80000000L) ? 1: 0);
              if (y3) xx |= (l2[row] >> 1);
              if (y4) xx |= l2[row];
              if (y5) xx |= (l2[row] << 1) | ((l2[row + 1] & 0x80000000L) ? 1: 0);
              if (y6) xx |= (l2[row + col32] >> 1);
              if (y7) xx |= l2[row + col32];
              if (y8) xx |= (l2[row + col32] << 1) | ((l2[row + col32 + 1] & 0x80000000L) ? 1: 0);
              l1[row] = xx;
              row += col32;
            }


           row = col32;                                /* REGION 5 */   
           switch(k) { 
              case 0: 
                for (i=1;i < nr -1; i++) {
                  for (j=1; j<col32 -1 ; j++) {
                    l1[row+j] = 0;
                  }
                  row += col32;
                 }
                 break;
                                                                                                  
               case 9: 
                 for (i=1; i < nr - 1; i++) {
                   for (j=1; j < col32 - 1; j++) {
                     jj = row + j;       
                     xx = (l2[jj - col32] >> 1) | ((l2[jj - col32 - 1] & 1) ? 0x80000000L: 0);
                     xx |= l2[jj - col32];
                     xx |= (l2[jj - col32] << 1) | ((l2[jj - col32 + 1] & 0x80000000L) ? 1: 0);
                     xx |= (l2[jj] >> 1) | ((l2[jj - 1] & 1) ? 0x80000000L: 0);
                     xx |= l2[jj];
                     xx |= (l2[jj] << 1) | ((l2[jj + 1] & 0x80000000L) ? 1: 0);
                     xx |= (l2[jj + col32] >> 1) | ((l2[jj + col32 - 1] & 1) ? 0x80000000L: 0);
                     xx |= l2[jj + col32];
                     xx |= (l2[jj + col32] << 1) | ((l2[jj + col32 + 1] & 0x80000000L) ? 1: 0);
                     l1[jj] = xx;
                   }
                   row += col32;
                 }
                 break;

               default:
                 for (i=1; i < nr - 1; i++) {
                   for (j=1; j < col32 - 1; j++) {
                     xx = 0;
                     jj = row + j;   
                     if (y0) xx =  (l2[jj - col32] >> 1) | ((l2[jj - col32 - 1] & 1) ? 0x80000000L: 0);
                     if (y1) xx |= l2[jj - col32];
                     if (y2) xx |= (l2[jj - col32] << 1) | ((l2[jj - col32 + 1] & 0x80000000L) ? 1: 0);
                     if (y3) xx |=  (l2[jj] >> 1) | ((l2[jj - 1] & 1) ? 0x80000000L: 0);
                     if (y4) xx |= l2[jj];
                     if (y5) xx |= (l2[jj] << 1) | ((l2[jj + 1] & 0x80000000L) ? 1: 0);
                     if (y6) xx |= (l2[jj + col32] >> 1) | ((l2[jj + col32 - 1] & 1) ? 0x80000000L: 0);
                     if (y7) xx |= l2[jj + col32];
                     if (y8) xx |= (l2[jj + col32] << 1) | ((l2[jj + col32 + 1] & 0x80000000L) ? 1: 0);
                     l1[jj] = xx;
                   }
                   row += col32;
                 }
           }                                   

                              
            row = col32 + col32 -1;                    /* REGION 6 */          
            for (i=1; i < nr - 1; i++) {                            
              xx = 0;       
              if (y0) xx |= (l2[row - col32] >> 1) | ((l2[row - col32 - 1] & 1) ? 0x80000000L: 0);
              if (y1) xx |= l2[row - col32];
              if (y2) xx |= (l2[row - col32] << 1) & border;
              if (y3) xx |= (l2[row] >> 1) | ((l2[row - 1] & 1) ? 0x80000000L: 0);
              if (y4) xx |= l2[row];
              if (y5) xx |= (l2[row] << 1) & border;
              if (y6) xx |= (l2[row + col32] >> 1) | ((l2[row + col32 - 1] & 1) ? 0x80000000L: 0);
              if (y7) xx |= l2[row + col32];
              if (y8) xx |= (l2[row + col32] << 1) & border;
              l1[row] = xx;
              row += col32;
            }




              xx = 0;                                 /* REGION 7 */
              j = col32 * (nr - 1);
              if (y0) xx |= (l2[j - col32] >> 1);
              if (y1) xx |= l2[j - col32];
              if (y2) xx |= (l2[j - col32] << 1) | ((l2[j - col32 + 1] & 0x80000000L) ? 1: 0);
              if (y3) xx |= (l2[j] >> 1);
              if (y4) xx |= l2[j];
              if (y5) xx |= (l2[j] << 1) | ((l2[j + 1] & 0x80000000L) ? 1: 0);
              if (y6 || y7 || y8) xx |= 0;
              l1[j] = xx;
                            

             row = col32 * (nr - 1);                   /* REGION 8 */
             for (j=1; j < col32 - 1; j++) {                                  
               xx = 0;
               if (y0) xx |= (l2[row + j - col32] >> 1) | ((l2[row + j - col32 - 1] & 1) ? 0x80000000L: 0);
               if (y1) xx |= l2[row + j - col32];
               if (y2) xx |= (l2[row + j - col32] << 1) | ((l2[row + j - col32 + 1] & 0x80000000L) ? 1: 0);
               if (y3) xx |= (l2[row + j] >> 1) | ((l2[row + j - 1] & 1) ? 0x80000000L: 0);
               if (y4) xx |= l2[row + j];
               if (y5) xx |= (l2[row + j] << 1) | ((l2[row + j + 1] & 0x80000000L) ? 1: 0);
               if (y6 || y7 || y8) xx |= 0;
               l1[row + j] = xx;
            }
   
   
              xx = 0;                                 /* REGION 9 */
              j = col32 * nr - 1;
              if (y0) xx |= (l2[j - col32] >> 1) | ((l2[j - col32 - 1] & 1) ? 0x80000000L: 0);
              if (y1) xx |= l2[j - col32];
              if (y2) xx |= (l2[j - col32] << 1) & border;
              if (y3) xx |= (l2[j] >> 1) | ((l2[j - 1] & 1) ? 0x80000000L: 0);
              if (y4) xx |= l2[j];
              if (y5) xx |= (l2[j] << 1) & border;
              if (y6 || y7 || y8) xx |= 0;
              l1[j] = xx;
            
           }

              if (mflg) {
                tnp = nb*nr*col32;
                l3 = (unsigned long *)(img2->imagedata);
                l2 = l2r;
                l1 = l1r;              
                for (i = 0; i < tnp; i++) {
                  *l1 = (*l1 & *l3) | (*l2 & (~ *l3));
                  l1++;
                  l2++;
                  l3++;
                }
              }
          
          return(1);                                       
        break;

        case VFF_TYP_1_BYTE :                        /* byte image */



          nc = img1->row_size;
          nr = img1->col_size;
          nb = img1->num_data_bands;       
          c2 = (unsigned char *)(img1->imagedata);
          c1 = (unsigned char *)(img3->imagedata); 
          if (mflg) c3 = (unsigned char *)(img2->imagedata);
          if ((nc < 3) || (nr < 3))      /* 1D image (signal) */        
             if ((nc == 1) || (nr == 1)) {
               if (nc == 1) {
                 i = nr;
                 y3 = ker->x[1][0];
                 y2 = ker->x[1][1];
                 y1 = ker->x[1][2];
               }
               else {
                 i = nc;
                 y3 = ker->x[0][1];
                 y2 = ker->x[1][1];
                 y1 = ker->x[2][1];
               }
               if(!mflg) {      
                 if (y2) x = c2[0];          /* REGION 1 */
                 else x = 0;      
                 if ((y3) && (c2[1] > x)) x = c2[1];
                 c1[0] = x;
                 for (j = 1; j < i-1; j++) {            /* REGION 2 */
                   if (y2) x = c2[j];
                   else x = 0;
                   if ((y1) && (c2[j-1] > x)) x = c2[j-1];
                   if ((y3) && (c2[j+1] > x)) x = c2[j+1];         
                   c1[j] = x; 
                 }
                 if (y2) x = c2[i-1];         /* REGION 3 */
                 else x = 0;
                 if ((y1) && (c2[i-2] > x)) x = c2[i-2];
                 c1[i-1] = x;
               }
               else {                   /* Conditional dilation */
                 if (c3[0]) {               /* REGION 1 */
                   if (y2) x = c2[0];                   
                   else x = 0;      
                   if ((y3) && (c2[1] > x)) x = c2[1];
                   c1[0] = x;
                 }
                 for (j = 1; j < i-1; j++) {            /* REGION 2 */
                   if (c3[j]) {
                     if (y2) x = c2[j];
                     else x = 0;
                     if ((y1) && (c2[j-1] > x)) x = c2[j-1];
                     if ((y3) && (c2[j+1] > x)) x = c2[j+1];         
                     c1[j] = x; 
                   }
                 }
                 if (c3[i-1]) {                         /* REGION 3 */
                   if (y2) x = c2[i-1];        
                   else x = 0;
                   if ((y1) && (c2[i-2] > x)) x = c2[i-2];
                   c1[i-1] = x;
                 }
               }
             }
             else { 
               kfprintf(kstderr,"lvdil: BYTE images should be at least 3x3 or signal.\n");
               return(0);
             }
        else {

          npb = nc * nr;
          tnp = npb * nb;
          nr1 = nr - 1;
          nc1 = nc - 1;
          nc2 = nc - 2;
          nr2 = nr - 2;


                                                  /* REGION 1 */

           n = 0;
           for (j=0; j<2; j++) 
             for (i=0; i<2; i++) 
               if (ker->x[i][j])            
                 d[n++] = (1 - j) * nc + 1 - i;
              
                                             
          if (!mflg) { 
            if (n == 0)
              for (i = 0; i < nb; i++)
                *(c1 + i*npb) = 0;
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref = c2 + knc;
                x = *(ref + d[0]); 
                for (l = 1; l < n; l++) {
                      y = *(ref + d[l]);
                      if (y > x) x = y; 
                } 
                *(c1 + knc) = x;
              }
          }
          else {
            
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                if (*(c3 + knc)) 
                  *(c1 + knc) = 0;
                else
                  *(c1 + knc) = *(c2 + knc);
              }
            }              
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref = c2 + knc;
                if (*(c3 + knc)) {
                  x = *(ref + d[0]);
                  for (l = 1; l < n; l++) {
                     y = *(ref + d[l]);
                     if (y > x) x = y;
                  } 
                  *(c1 + knc) = x;
                }
                else 
                  *(c1 + knc) = *ref; 
              }
          }     
   
                                               /* REGION 2 */

           n = 0;
           for (j=0; j<2; j++) 
             for (i=0; i<3; i++) 
               if (ker->x[i][j])            
                 d[n++] = (1 - j) * nc + 1 - i;
  
          if (!mflg) {
            if (n == 0)
              for (i = 0; i < nb; i++) {
                ref = c1 + i*npb;                
                for (j = 1; j <= nc2; j++)
                  *(ref + j) = 0;
              }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                for (j = 1; j <= nc2; j++) {
                  kncj = knc + j;
                  ref = c2 + kncj;
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                     y = *(ref + d[l]);
                     if (y > x)  x = y; 
                  } 
                  *(c1 + kncj) = x;
                }
              } 
          }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                for (j = 1; j <= nc2; j++)
                  kncj = knc + j; 
                  if (*(c3 + kncj))
                    *(c1 + kncj) = 0;
                  else
                    *(c1 + kncj) = *(c2 + kncj);
              }
            }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                for (j = 1; j <= nc2; j++) {
                  kncj = knc + j;
                  if (*(c3 + kncj)) {
                    ref = c2 + kncj;
                    x = *(ref + d[0]); 
                    for (l = 1; l < n; l++) {
                        y = *(ref + d[l]);
                        if (y > x) x = y;
                    } 
                    *(c1 + kncj) = x;
                  }
                  else
                    *(c1 + kncj) = *(c2 + kncj);
                }
              }
           }    

                                        /* REGION 3 */ 
          n = 0;
          for (j=0; j<2; j++) 
            for (i=1; i<3; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i;

          if (!mflg) { 
            if (n == 0) {
              ref = c1 + nc1;
              for (i = 0; i < nb; i++)
                *(ref + i*npb) = 0;
            }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb + nc1;
                ref = c2 + knc; 
                x = *(ref + d[0]); 
                for (l = 1; l < n; l++) {
                   y = *(ref + d[l]);
                   if (y > x)  x = y; 
                } 
                *(c1 + knc) = x;
              }
          }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb + nc1;
                if (*(c3 + knc))
                  *(c1 + knc) = 0;
              }
            }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb + nc1;
                ref = c2 + knc;
                if (*(c3 + knc)) {
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                    y = *(ref + d[l]);
                    if (y > x)     x = y; 
                  } 
                  *(c1 + knc) = x;
                }
                else
                  *(c1 + knc) = *ref;
              }
           }   
  

                                         /* REGION 4 */ 
          n = 0;
          for (j=0; j<3; j++) 
            for (i=0; i<2; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i;
 
          if (!mflg) { 
            if (n == 0) 
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref = c1 + knc;
                for (j = 1; j <= nr2; j++)
                  *(ref + j*nc) = 0;
              }    
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref1 = c1 + knc;
                for (j = 1; j <= nr2; j++) {
                  ref = c2 + knc + j*nc;
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                    y = *(ref + d[l]);
                    if (y > x)     x = y; 
                  } 
                  *(ref1 + j*nc) = x;
                }
              } 
           }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref = c1 + knc;
                ref1 = c2 + knc;
                ref2 = c3 + knc;
                for (j = 1; j <= nr2; j++) 
                  if (*(ref2 + j*nc))
                    *(ref + j*nc) = 0;
                  else
                    *(ref + j*nc) = *(ref1 + j*nc);
              }
            }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref1 = c1 + knc;
                ref2 = c2 + knc;
                ref3 = c3 + knc;
                for (j = 1; j <= nr2; j++) {
                  ref = ref2 + j*nc;
                  if (*(ref3 + j*nc)) {
                    x = *(ref + d[0]); 
                    for (l = 1; l < n; l++) {
                      y = *(ref + d[l]);
                      if (y > x)     x = y; 
                    } 
                    *(ref1 + j*nc) = x;
                  }
                  else
                    *(ref1 + j*nc) = *(ref2 + j*nc);
                }
              } 
           }      


                                         /* REGION 5 */ 
          n = 0;
          for (j=0; j<3; j++) 
            for (i=0; i<3; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i;
 
          if (!mflg) {                 
            switch(n) {
              case 0:                                  /* 0 point struc */
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) 
                            *(c1 + j) = 0;
                                
                        }
                      }       
              break;

              case 1:                                  /* 1 point struc */
                      y1 = d[0];
                 
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          ref = c2 + y1;
                          for (j = jinf; j < jsup; j++) 
                            *(c1 + j) = *(ref + j);                                
                        }
                      }                
              break;        
  
              case 2:                                   /* 2 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
 
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);
                            if (y > x) *(c1 + j) = y;
                            else *(c1 + j) = x; 
                          }                                
                        }
                      }                
              break;

              case 3:                                    /* 3 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2]; 

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);
                            if (y > x) x = y;
                            y = *(ref + y3); 
                            if (y > x) *(c1 + j) = y;
                            else *(c1 + j) = x;
                          }
                        }
                      }     
              break;
         
              case 4:                                    /* 4 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
          
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);
                            if (y > x) x = y;
                            y = *(ref + y3);
                            if (y > x) x = y;
                            y = *(ref + y4);
                            if (y > x)  *(c1 + j) = y;
                            else *(c1 + j) = x;
                          }                              
                        }
                      }     
              break;
  
              case 5:                                    /* 5 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);
                            if (y > x) x = y;
                            y = *(ref + y3);
                            if (y > x) x = y;
                            y = *(ref + y4);                        
                            if (y > x) x = y;
                            y = *(ref + y5);                        
                            if (y > x)  *(c1 + j) = y;
                            else *(c1 + j) = x;
 
                          }                              
                        }
                      }     
              break;

              case 6:                                    /* 6 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
                      y6 = d[5]; 

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) { 
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);                        
                            if (y > x) x = y;
                            y = *(ref + y3);
                            if (y > x) x = y;
                            y = *(ref + y4);                        
                            if (y > x) x = y;
                            y = *(ref + y5);                        
                            if (y > x) x = y;
                            y = *(ref + y6);                        
                            if (y > x)  *(c1 + j) = y;
                            else *(c1 + j) = x;

                          }   
                        }
                      }   
              break;                          

              case 7:                                    /* 7 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
                      y6 = d[5];
                      y7 = d[6]; 

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) { 
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);                        
                            if (y > x) x = y;
                            y = *(ref + y3);
                            if (y > x) x = y;
                            y = *(ref + y4);                        
                            if (y > x) x = y;
                            y = *(ref + y5);                        
                            if (y > x) x = y;
                            y = *(ref + y6);                        
                            if (y > x) x = y;
                            y = *(ref + y7);                        
                            if (y > x)  *(c1 + j) = y;
                            else *(c1 + j) = x;
                          }   
                        }
                      }   
        break;

        case 8:                                    /* 8 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
                      y6 = d[5];
                      y7 = d[6]; 
                      y8 = d[7];

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) { 
                            ref = c2 + j;
                            y = *(ref + y2);
                            x = *(ref + y1);                        
                            if (y > x) x = y;
                            y = *(ref + y3);
                            if (y > x) x = y;
                            y = *(ref + y4);                        
                            if (y > x) x = y;
                            y = *(ref + y5);                       
                            if (y > x) x = y;
                            y = *(ref + y6);                        
                            if (y > x) x = y;
                            y = *(ref + y7);                        
                            if (y > x) x = y;
                            y = *(ref + y8);                        
                            if (y > x)  *(c1 + j) = y;
                            else *(c1 + j) = x;
                          }   
                        }
                      }   
              break;

              case 9:                                    /* 9 point struc */

                      for (i = 0; i < nb; i++) {       
                        knc = i*npb;
                        ref1 = c2 - nc;
                        ref2 = c2;
                        ref3 = c2 + nc;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;               
                          x = *(ref1 + knc);
                          y = *(ref2 + knc);
                          if (x > y)                    /* first column maximum */
                            y = *(ref3 + knc);
                          else
                            x = *(ref3 + knc);
                          if (x > y)
                            y1 = x;
                          else
                            y1 = y;
                          x = *(ref1 + jinf);         /* second column maximum */
                          y = *(ref2 + jinf);
                          if (x > y) 
                            y = *(ref3 + jinf);
                          else
                            x = *(ref3 + jinf);
                          if (x > y)
                            y2 = x;
                          else
                            y2 = y;                      
                          for (j = jinf; j < jsup; j++) {
                            ref = c2 + j + 1;   
                            x = *(ref - nc);
                            y = *(ref);
                            if (x > y)                 /* third column maximum */
                              y = *(ref + nc);
                            else
                              x = *(ref + nc);
                            if (x > y)
                              y3 = x;
                            else
                              y3 = y;  
                            if (y1 > y2)               /* maximum of maximum */
                              y = y1;
                            else 
                              y = y2;
                            if (y > y3)
                              *(c1 + j) = y;
                            else
                              *(c1 + j) = y3;
                            y1 = y2;
                            y2 = y3;
                          }                              
                        }
                      }
              break;         
           }
         }   
         else {
           switch(n) {
              case 0:                                  /* 0 point struc */
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) *(c1 + j) = 0;
                            else *(c1 + j) = *(c2 + j);
                          }                                
                        }
                      }       
              break;

              case 1:                                  /* 1 point struc */
                      y1 = d[0];
                 
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        ref = c2 + y1;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                             if (*(c3 + j)) *(c1 + j) = *(ref + j);
                             else *(c1 + j) = *(c2 + j);
                          }                                
                        }
                      }                
              break;        
  
              case 2:                                   /* 2 point struc */
                      y1 = d[0];
                      y2 = d[1];
 
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) {
                              ref = c2 + j;
                              y = *(ref + y2);
                              x = *(ref + y1);
                              if (y > x) *(c1 + j) = y;
                              else *(c1 + j) = x;
                            }
                            else
                              *(c1 + j) = *(c2 + j);
                          }                                
                        }
                      }                
              break;

              case 3:                                    /* 3 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) {
                              ref = c2 + j;
                              y = *(ref + y2);
                              x = *(ref + y1);                        
                              if (y >  x) x = y;
                              y = *(ref + y3); 
                              if (y < x) *(c1 + j) = x;
                              else *(c1 + j) = y;
                            }
                            else
                              *(c1 + j) = *(c2 + j);   
                          }                              
                        }
                      }     
              break;
         
              case 4:                                    /* 4 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
          
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) {
                              ref = c2 + j;
                              y = *(ref + y2);
                              x = *(ref + y1);                        
                              if (y > x) x = y;
                              y = *(ref + y3);
                              if (y > x) x = y;
                              y = *(ref + y4);                        
                              if (y > x)  *(c1 + j) = y;
                              else *(c1 + j) = x;
                            }
                            else
                              *(c1 + j) = *(c2 + j); 
                          }                              
                        }
                      }     
              break;
  
              case 5:                                    /* 5 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
 
                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) {
                             ref = c2 + j;
                             y = *(ref + y2);
                             x = *(ref + y1);                        
                             if (y > x) x = y;
                             y = *(ref + y3);
                             if (y > x) x = y;
                             y = *(ref + y4);                        
                             if (y > x) x = y;
                             y = *(ref + y5);                        
                             if (y > x)  *(c1 + j) = y;
                             else *(c1 + j) = x;
                            }
                            else
                              *(c1 + j) = *(c2 + j); 
                          }                              
                        }
                      }     
              break;

              case 6:                                    /* 6 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
                      y6 = d[5];

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) { 
                             ref = c2 + j;
                             y = *(ref + y2);
                             x = *(ref + y1);                        
                             if (y > x) x = y;
                             y = *(ref + y3);
                             if (y > x) x = y;
                             y = *(ref + y4);                        
                             if (y > x) x = y;
                             y = *(ref + y5);                        
                             if (y > x) x = y;
                             y = *(ref + y6);                        
                             if (y > x)  *(c1 + j) = y;
                             else *(c1 + j) = x;
                            }
                            else
                              *(c1 + j) = *(c2 + j);
                          }   
                        }
                      }   
              break;                          

              case 7:                                    /* 7 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
                      y6 = d[5];
                      y7 = d[6];

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) { 
                             ref = c2 + j;
                             y = *(ref + y2);
                             x = *(ref + y1);                        
                             if (y > x) x = y;
                             y = *(ref + y3);
                             if (y > x) x = y;
                             y = *(ref + y4);                        
                             if (y > x) x = y;
                             y = *(ref + y5);                        
                             if (y > x) x = y;
                             y = *(ref + y6);                        
                             if (y > x) x = y;
                             y = *(ref + y7);                        
                             if (y > x)  *(c1 + j) = y;
                             else *(c1 + j) = x;
                            }
                            else
                              *(c1 + j) = *(c2 + j);
                          }   
                        }
                      }   
        break;

        case 8:                                    /* 8 point struc */
                      y1 = d[0]; 
                      y2 = d[1];
                      y3 = d[2];
                      y4 = d[3];
                      y5 = d[4];
                      y6 = d[5];
                      y7 = d[6]; 
                      y8 = d[7];

                      for (i = 0; i < nb; i++) {
                        knc = i*npb;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) { 
                             ref = c2 + j;
                             y = *(ref + y2);
                             x = *(ref + y1);                        
                             if (y > x) x = y;
                             y = *(ref + y3);
                             if (y > x) x = y;
                             y = *(ref + y4);                        
                             if (y > x) x = y;
                             y = *(ref + y5);                        
                             if (y > x) x = y;
                             y = *(ref + y6);                        
                             if (y > x) x = y;
                             y = *(ref + y7);                        
                             if (y > x) x = y;
                             y = *(ref + y8);                        
                             if (y > x)  *(c1 + j) = y;
                             else *(c1 + j) = x;
                            }
                            else
                              *(c1 + j) = *(c2 + j);
                          }   
                        }
                      }   
              break;

              case 9:                                    /* 9 point struc */

                       for (i = 0; i < nb; i++) {       
                        knc = i*npb;
                        ref1 = c2 - nc - 1;
                        ref2 = c2 - 1;
                        ref3 = c2 + nc - 1;
                        ref4 = ref1 + 1;
                        ref5 = ref2 + 1;
                        ref6 = ref3 + 1;
                        for (k = 1; k < nr1; k++) {
                          knc += nc;
                          jinf = knc + 1;
                          jsup = knc + nc1;
                          y8 = 0;                /* new sequence of true value mask pixels */ 
                          for (j = jinf; j < jsup; j++) {
                            if (*(c3 + j)) {
                              if (y8 == 0) { 
                                x = *(ref1 + j);
                                y = *(ref2 + j);
                                if (x > y)               /* first column maximum */
                                  y = *(ref3 + j);
                                else
                                  x = *(ref3 + j);                              
                                if (x > y)
                                  y1 = x;
                                else
                                  y1 = y;                                    
                                x = *(ref4 + j);       /* second column maximum */
                                y = *(ref5 + j);
                                if (x > y) 
                                  y = *(ref6 + j);
                                else
                                  x = *(ref6 + j);                              
                                if (x > y)
                                  y2 = x;
                                else
                                  y2 = y;
                              }
                              ref = ref5 + j + 1;        /* third column maximum */
                              x = *(ref - nc);           
                              y = *(ref);
                              if (x > y)                 
                                y = *(ref + nc);
                              else
                                x = *(ref + nc);                
                              if (x > y)    
                                y3 = x; 
                              else
                                y3 = y;                      
                              if (y1 > y2)              /* maximum of maximum */
                                y = y1;
                              else
                                y = y2;
                              if (y > y3)
                                *(c1 + j) = y;
                              else
                                *(c1 + j) = y3;
                              y1 = y2;
                              y2 = y3;
                            }
                            else {
                              *(c1 + j) = *(c2 + j); 
                              y8 = 0;                      /* false value mask pixel */
                            }        
                        }                              
                      }
                    }
                     
              break;         
           }
         }
 
                                /* REGION 6 */
          n = 0;
          for (j=0; j<3; j++) 
            for (i=1; i<3; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i;
      
          if (!mflg) { 
            if (n == 0)
              for (i = 0; i < nb; i++) {
                ref = c1 + i*npb + nc1;
                for (j = 1; j <= nr2; j++)
                  *(ref + j*nc) = 0;
              }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref2 = c2 + knc;
                ref1 = c1 + knc;
                for (j = 1; j <= nr2; j++) {
                  ref = ref2 + j*nc + nc1;
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                    y = *(ref + d[l]);
                    if (y > x)     x = y; 
                  } 
                  *(ref1 + j*nc + nc1) = x;
                }
              } 
           }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref2 = c2 + knc + nc1;
                ref1 = c1 + knc + nc1;
                ref3 = c3 + knc + nc1;
                for (j = 1; j <= nr2; j++) 
                  if (*(ref3 + j*nc))
                    *(ref1 + j*nc) = 0;
                  else
                    *(ref1 + j*nc) = *(ref2 + j*nc);
              }
            }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref2 = c2 + knc + nc1;
                ref1 = c1 + knc + nc1;
                ref3 = c3 + knc + nc1;
                for (j = 1; j <= nr -2; j++) {
                  ref = ref2 + j*nc;
                  if (*(ref3 + j*nc)) {
                    x = *(ref + d[0]); 
                    for (l = 1; l < n; l++) {
                      y = *(ref + d[l]);
                      if (y > x)     x = y; 
                    } 
                    *(ref1 + j*nc) = x;
                  }
                  else
                    *(ref1 + j*nc) = *(ref2 + j*nc);
                }
              } 
           }    


                                         /* REGION 7 */
          n = 0;
          for (j=1; j<3; j++) 
            for (i=0; i<2; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i;
 
          if (!mflg) { 
            if (n == 0)
              for (i = 0; i < nb; i++)
                *(c1 + i*npb + nc*nr1) = 0;
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref = c2 + knc + nc*nr1;
                x = *(ref + d[0]); 
                for (l = 1; l < n; l++) {
                   y = *(ref + d[l]);
                   if (y > x)     x = y; 
                } 
                *(c1 + knc + nc*nr1) = x;
              }
          }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb + nc*nr1;
                if (*(c3 + knc))
                  *(c1 + knc) = 0;
                else
                  *(c1 + knc) = *(c2 + knc);
              }
            }  
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb + nc*nr1;
                ref = c2 + knc;    
                if (*(c3 + knc)) {
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                    y = *(ref + d[l]);
                    if (y > x)     x = y; 
                  } 
                  *(c1 + knc) = x;
                }
                else
                  *(c1 + knc) = *(c2 + knc);
              }
           }    

 
                         /* REGION 8 */ 

          n = 0;
          for (j=1; j<3; j++) 
            for (i=0; i<3; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i;
  
          if (!mflg) { 
            if (n == 0)
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref1 = c1 + knc + nc*nr1;
                for (j = 1; j <= nc2; j++)
                  *(ref1 + j) = 0;
              }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref1 = c1 + knc + nc*nr1;
                ref2 = c2 + knc + nc*nr1;
                for (j = 1; j <= nc2; j++) {
                  ref = ref2 + j;
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                    y = *(ref + d[l]);
                    if (y > x)     x = y; 
                  } 
                  *(ref1 + j) = x;
                }
              } 
          }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb + nc*nr1;
                ref1 = c1 + knc;
                ref2 = c2 + knc;
                ref3 = c3 + knc;
                for (j = 1; j <= nc2; j++) 
                  if (*(ref3 + j))
                    *(ref1 + j) = 0;
                  else
                    *(ref1 + j) = *(ref2 + j);
              }
            } 
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref1 = c1 + knc + nc*nr1;
                ref2 = c2 + knc + nc*nr1;
                ref3 = c3 + knc + nc*nr1;    
                for (j = 1; j <= nc2; j++) {
                  ref = ref2 + j;
                  if (*(ref3 + j)) {
                    x = *(ref + d[0]); 
                    for (l = 1; l < n; l++) {
                      y = *(ref + d[l]);
                      if (y > x)     x = y; 
                    } 
                    *(ref1 + j) = x;
                  }
                  else
                    *(ref1 + j) = *(ref2 + j);
                }
              } 
           }    

                                        /* REGION 9 */
          n = 0;
          for (j=1; j<3; j++) 
            for (i=1; i<3; i++) 
              if (ker->x[i][j])            
                d[n++] = (1 - j) * nc + 1 - i; 

          if (!mflg) { 
            if (n == 0)
              for (i = 0; i < nb; i++)
                *(c1 + i*npb + npb - 1) = 0;
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb;
                ref = c2 + knc + npb - 1;
                x = *(ref + d[0]); 
                for (l = 1; l < n; l++) {
                   y = *(ref + d[l]);
                   if (y > x)     x = y; 
                } 
                *(c1 + knc + npb - 1) = x;
              }
          }
          else {
            if (n == 0) {
              for (i = 0; i < nb; i++) {
                knc = i*npb + npb - 1;
                if (*(c3 + knc))
                  *(c1 + knc) = 0;
                else
                  *(c1 + knc) = *(c2 + knc);
              }
            }
            else                         
              for (i = 0; i < nb; i++) {
                knc = i*npb + npb - 1;
                ref = c2 + knc;
                if (*(c3 + knc)) {
                  x = *(ref + d[0]); 
                  for (l = 1; l < n; l++) {
                    y = *(ref + d[l]);
                    if (y > x)     x = y; 
                  } 
                  *(c1 + knc) = x;
                }
                else
                  *(c1 + knc) = *(c2 + knc);
              }
            }              
          }
          return(1);          
          break;
 
        case VFF_TYP_2_BYTE :
          kfprintf(kstderr,"lvdil: BYTE2 not implemented\n");
          return(0);
          break;

        case VFF_TYP_4_BYTE :
          kfprintf(kstderr,"lvdil: BYTE4 not implemented\n");
          return(0);
          break;

        case VFF_TYP_FLOAT :
          kfprintf(kstderr,"lvdil: FLOAT not implemented\n");
          return(0);
          break;

        case VFF_TYP_COMPLEX :
          kfprintf(kstderr,"lvdil: COMPLEX not implemented\n");
          return(0);
          break;

        case VFF_TYP_DOUBLE :
          kfprintf(kstderr,"lvdil: DOUBLE not implemented\n");
          return(0);
          break;

        case VFF_TYP_DCOMPLEX :
          kfprintf(kstderr,"lvdil: COMPLEXD not implemented\n");
          return(0);
          break;

        default:
          kfprintf(kstderr,"lvdil: Unknown data storage type, %d\n",img1->data_storage_type);
          return(0);
          break;
    }
    /*
     * End of MAT code
     */

    break;
    }

    /*
     *
     * Default action
     *
     */
    default:
          (void) kfprintf(kstderr,"unrecognizable element type\n");
          return(0);         
    } 

/*   (void)kfree((kaddr)lline);
   (void)kfree((kaddr)linesh);
   (void)kfree((kaddr)loutimage);
   (void)kfree((kaddr)line);
   (void)kfree((kaddr)outimage); */
    return(1); /* success */
}
/* -library_code_end */
