 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * Copyright (C) 1993-1997 Junior Barrera, Roberto Lotufo.
 * All rights reserved.  See $MMACH/repos/license/License.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vdilexpan
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvdilexpan
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvdilexpan - Expansion as a conditionally translation invariant dilation.
* 
*       Purpose: .I vdilexpan
expand an input image in both directions.
.LP
Let the domain of the input image be
           E = [0, 1, ..., m - 1] x [0, 1, ..., n - 1].
Let vs and hs be two positive integer values representing, respectively, the vertical and horizontal scale factors.
Let the domain of the output image be
           vhE = [0, 2, ..., vs(m - 1)] x [0, 2, ..., hs(n - 1)].
Let vo and ho be two non negative integer values less or equal than, respectively, vs and hs. vo and ho represent, respectively, the vertical and horizontal offsets.
Let a be the (conditionally) structuring function from E to P(vhE)
(the subsets of vhE) given by
           a(y) = (A + (y + (vo, ho))) n vhE (y member of E)
where n means "intersection" and where A is a subset of the 3 x 3
square centered at the origim, called structuring element.
The expansion of the input image Y is the dilation of Y by a, that is:
expand of Y = U{a(y): y member Y}.
.LP
If A is the 2 x 2 square then the dilation expands the input image in
the usual manner, that is by duplicating each original pixel four times
(twice in both directions).
If A is the 3 x 3 square then the expansion of an 8-connected subset
appears without ambiguity on the screen as a connected subset (in the
sense of the usual Euclidean topology). In this later case, we say that
the expansion preserve the connectivity 8.
.LP
img1 and b_mat are, respectively, the input image and the structuring element, and str_flag should be set to 1 if the structuring element 
is to be used;
img2 is the gating image, and mflg should be set to 1 if the gating image 
is to be used;
v_scale, h_scale, v_offset, h_offset are, respectively, the vertical scale factor, the horizontal scale factor, the vertical offset and the horizontal offset;
img3 is the output image.
.LP
The optional input mask image is used to gate the operation. 
.LP
A non-zero mask pixel enables the output pixel to contain the expansion. 
A zero mask pixel just transfers the value of the corresponding 
pixel in the first input image to the output pixel.
.LP
The input, output and mask images must be of the same data type, and have 
the same number of data bands. The output image column size is vs times
the input image column size and the output image row size is hs times
the input image row size.
*         Input: .IP "img1" 15  
input image 
.IP "b_mat" 15
structuring element
.IP "img2" 15
gating mask image
.IP "v_scale" 15
vertical scale factor by which to expand the image height, must be a positive integer
.IP "h_scale" 15
horizontal scale factor by which to expand the image width, must be a positive integer
.IP "v_offset" 15
vertical offset must be a non negative integer smaller or equal than v_scale
.IP "h_offset" 15
horizontal offset must be a non negative integer smaller or equal than h_scale
.IP "str_flag" 15
a flag set (equal 1) if structuring element available
.IP "mflg" 15
a flag set (equal 1) if gating image mask available
.IP "border" 15
mask to the last pixel of a row (used in formatted BIT images)
.IP
Note that the input and mask images must be of the same data type and size.
*        Output: .IP "img3" 15
holds the result of the expansion. The output data type is the same as
the input data type. The output image column size is v_scale times the
input image column size and the output image row size is h_scale times
the input image row size.
.IP "Return Value:" 15
1 on success, 0 on failure.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: .I lvdilexpan
can be defined for all data types supported by Khoros, but at the moment it has been implemented just for unsigned char types.
The structuring elements are subsets of the 3 x 3 square centered at the origim.
*    Written By: Marco Antonio Carvalho Pessoa
Gerald Jean Francis Banon (for last revision)
Sao Jose dos Campos, SP, Brazil, November, 9th, 94
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int
lvdilexpan(xvimage *img1, mat_str *b_mat, xvimage *img2, int v_scale, int h_scale, int v_offset, int h_offset , int str_flag, int mflg, unsigned long border, xvimage *img3)
/*    v_scale, vertical scale factor */
/*    h_scale, horizontal scale factor */
/*    v_offset, vertical offset */
/*    h_offset; horizontal offset */
/* -library_def_end */

/* -library_code */
{
        xvimage *img4; /* auxiliary image */
        int j, k, bands, 
            rows1, /* number of rows of the input image */
            cols1, /* number of columns of the input image */
            rows4, /* number of rows of the auxiliary and output images */
            cols4, /* number of columns of the auxiliary and output images */
            b1, /* cardinality of the input image domain */
            b4, /* cardinality of the auxiliary and output images domain */
            k4, /* number of pixels of the auxiliary and output images */
            nb; /* number of bands */
	unsigned char *c1, *c4, value4;
               
        value4 = 0;
        rows1 = img1->col_size;
        cols1 = img1->row_size;
        rows4 = img3->col_size;  
        cols4 = img3->row_size;
           
        nb = img1->num_data_bands;
        b1 = rows1*cols1;
        b4 = rows4*cols4;
        
        img4 = createimage( img3->col_size, img3->row_size, img3->data_storage_type, img3->num_of_images, img3->num_data_bands, img3->comment, img3->map_row_size , img3->map_col_size, img3->map_scheme,
img3->map_storage_type, img3->location_type, img3->location_dim);
    
        
	if (img4 == NULL) {
           (void) kfprintf(kstderr, "lvdilexpan: createimage failed\n");
           return(0);
        }
	
	if (v_scale <= 0) {
           (void) kfprintf(kstderr, "Vertical Scale Factor (v_scale) must be greater than zero\n\n");
           return(0);
	}

	if (h_scale <= 0) {
           (void) kfprintf(kstderr, "Horizontal Scale Factor (h_scale) must be greater than zero\n\n");
           return(0);
	}
	if (v_offset < 0) {
           (void) kfprintf(kstderr, " Vertical Offset (v_offset) must be greater or equal than zero\n\n");
           return(0);
	}
	if (h_offset < 0) {
           (void) kfprintf(kstderr, " Horizontal Offset (h_offset) must be greater or equal than zero\n\n");
           return(0);
	}

	if (v_offset >= v_scale) {
           (void) kfprintf(kstderr, "Vertical Offset (v_offset) must be smaller than Vertical Scale Factor (v_scale)\n\n");
           return(0);
        }
	if (h_offset >=  h_scale) {
           (void) kfprintf(kstderr, "Horizontal Offset (h_offset) must be smaller than Horizontal Scale Factor (h_scale)\n\n");
           return(0);
        }
		
        switch (img1->data_storage_type) {
               case VFF_TYP_BIT :
                    kfprintf(kstderr,"lvdilexpan:Don't know to expand a BIT image!\n");
                    return(0);
                    break;
        
               case VFF_TYP_1_BYTE : 
                    k4 = b4*nb;
                    c1 = (unsigned char *)(img1->imagedata);
 
                    c4 = (unsigned char *)(img4->imagedata);
                    if (!mflg) 
                       for (j = 0; j < k4 ; j++) 
                           *c4++ = (unsigned char) value4;    

                    c4 = (unsigned char *)(img4->imagedata);
                    for (j = 0; j < rows1; j++)
                        for (k = 0; k < cols1; k++)
                            for (bands = 0; bands < nb; bands++)
                                *(c4 + bands*b4 + (j*v_scale + v_offset) * cols4 + (k*h_scale + h_offset)) = *(c1 + bands*b1 + j*cols1 + k);
                          
     		    if (str_flag) {
                       if (!lvdil(img4, b_mat, img2, mflg, border, img3))
		          return(0);
		    }                               							
                
                    else
		         copyimagedata1(img4, img3);
		         break;

               case VFF_TYP_2_BYTE :
                    kfprintf(kstderr,"lvsup: BYTE2 not implemented\n");
                    return(0);
                    break;

               case VFF_TYP_4_BYTE :
                    kfprintf(kstderr,"lvsup: BYTE4 not implemented\n");
                    return(0);
                    break;

               case VFF_TYP_FLOAT :
                    kfprintf(kstderr,"lvsup: FLOAT not implemented\n");
                    return(0);
                    break;

               case VFF_TYP_COMPLEX :
                    kfprintf(kstderr,"lvsup: COMPLEX not implemented\n");
                    return(0);
                    break;

               case VFF_TYP_DOUBLE :
                    kfprintf(kstderr,"lvsup: DOUBLE not implemented\n");
                    return(0);
                    break;

               case VFF_TYP_DCOMPLEX :
                    kfprintf(kstderr,"lvsup: COMPLEXD not implemented\n");
                    return(0);
                    break;

               default:
                        kfprintf(kstderr,"lvsup: Unknown data storage type, %d\n",img1->data_storage_type);
                        return(0);
                        break;
        } 
 
        freeimage(img4);
        return(1);
}
/* -library_code_end */
