 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * Copyright (C) 1993-1997 Junior Barrera, Roberto Lotufo.
 * All rights reserved.  See $MMACH/repos/license/License.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vdilseg
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvdilseg
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
#include <values.h>
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvdilseg - Dilation by a straight line segment.
* 
*       Purpose: .I vdilseg
will do a dilation of the input image by a straight line segment.
.LP
the input image is "img_in"
the straight line segment is defined by :
   - its length : "length"
   - its angle from the horizontal direction "alfa"
the output image will be "img_out".
.LP
"im_trans" and "img_trav" are template images.
*         Input: .IP "img_in" 15  
input image
.IP "alfa" 15
angle of the straight line segment from the horizontal
.IP "length" 15
length of the straight line segment
.IP "border" 15
mask to the last pixel of a row (used in formatted BIT images)
*        Output: .IP "img_out" 15
holds the result of the dilations. The output data
type is the same as the input data type.
.IP "Return Value:" 15
1 on success, 0 on failure.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: .I lvdilseg
can be defined for all data types supported by Khoros, but at the moment it has been implemented just for the bit and unsigned char types.
.LP
Due to the simultaneous processing of 32 pixels in BIT images, input images of the BIT data type must be converted to a special data structure before the library is called. This can be accomplished by using the format routine. After the library call, the unformat routine can convert the image back to the BIT data type.
*    Written By: Guy Pilchen
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Adriano Nagelschmidt Rodrigues: port to new toolbox version.
****************************************************************/
/* -library_def */
int
lvdilseg(xvimage *img_in, int alfa, int length, unsigned long border, xvimage *img_out)
/* -library_def_end */

/* -library_code */
#define INT(x) (((x)-(int)(x)>0.5) ? (int)(x)+1 : (int) (x))
{
   /* local variables */
   xvimage       *img_trans,
                        *img_trav;
   int                  vectx[50],      /* length max is 50 */
                        vecty[50],   
                        lproj,          /* length projected on an axis */
                        nbpt,
                        dx,
                        dy,
                        dx_old,
                        dy_old,
                        i,
                        i_proj,
                        j,
                        sgn,
                        sgn_tan;
   mat_str              elt_str_mat;    /* structuring set */
   double               alfaF;          

   /* create a template image img_trans */
   img_trans = createimage(     img_in->col_size, 
                                img_in->row_size, 
                                img_in->data_storage_type, 
                                img_in->num_of_images,
                                img_in->num_data_bands, 
                                img_in->comment, 
                                img_in->map_row_size, 
                                img_in->map_col_size, 
                                img_in->map_scheme,
                                img_in->map_storage_type, 
                                img_in->location_type, 
                                img_in->location_dim);
   if (img_trans == NULL) {
         (void) kfprintf(kstderr, "createimage Failed\n");
          kexit(1);
   }

   /* create another template image img_trav */
   img_trav = createimage(      img_in->col_size, 
                                img_in->row_size, 
                                img_in->data_storage_type, 
                                img_in->num_of_images,
                                img_in->num_data_bands, 
                                img_in->comment, 
                                img_in->map_row_size, 
                                img_in->map_col_size, 
                                img_in->map_scheme,
                                img_in->map_storage_type, 
                                img_in->location_type, 
                                img_in->location_dim);
   if (img_trav == NULL) {
         (void) kfprintf(kstderr, "createimage Failed\n");
          kexit(1);
   }

   /* decompose the line segment in a list of adjacent points */
   alfaF = (alfa+180) * KPI / 180.;
   if (((alfaF >= -KPI/4)  && (alfaF <= KPI/4)) ||
       ((alfaF >= 3*KPI/4) && (alfaF <= KPI)) ||
       ((alfaF >= -KPI) && (alfaF <= -3*KPI/4))) {
        lproj = INT(length * cos(alfaF)*1.);
        if (lproj < 0) {     
           sgn = -1;     
        } else {     
           sgn = 1;
        }
        lproj = abs(lproj);
        if ((alfaF >= KPI/2) || (alfaF <= -KPI/2)) {
           sgn_tan = -1;
        } else {
           sgn_tan = 1;
        }
        for (i=0; i< lproj; i++) {  
                vectx[i] = sgn * i;
                vecty[i] = INT(i * sgn_tan * tan(alfaF));
        }
   } else {
        lproj = INT(length * sin(alfaF));
        if (lproj < 0) {
           sgn = -1;
        } else {
           sgn = 1;
        }
        lproj = abs(lproj);
        if (alfaF <= 0) {
           sgn_tan = -1;
        } else {
           sgn_tan = 1;
        }
        for (i=0; i< lproj; i++) {
                vectx[i] = INT(sgn_tan*i/tan(alfaF));
                vecty[i] = sgn * i;
        }
   }
           
   /* inicialize img_trav and img_out with img_in */
   if (!copyimagedata1(img_in, img_trav))
     return(0);
   if (!copyimagedata1(img_in, img_out))
     return(0);

   /* inicialize dx_old and dy_old to force to calculate the first structuring set */
   dx_old = dy_old = -2;

   /* bedin of the loop */
   for (i_proj=1; i_proj<lproj; i_proj++) {

        /* make ith structuring set if diferent from the (i-1)th */
        dx = vectx[i_proj] - vectx[i_proj-1];
        dy = vecty[i_proj] - vecty[i_proj-1];
        if ((dx!=dx_old)||(dy!=dy_old)) { /* it means the structuring set has changed */
           /* initialize the structure set elt_str_mat with 0 */
           for (i=0; i<3; i++)
              for (j=0; j<3; j++)
                 (&elt_str_mat)->x[i][j] = 0;
	   /* initialize the tag of elt_str_mat RAL 17/oct/95 */
	   elt_str_mat.tag = MM_MAT;
           dx_old = dx;
           dy_old = dy;
           switch (dx) {
              case (-1):  
                switch (dy) {
                  case (-1): (&elt_str_mat)->x[2][0] = 1; break;
                  case 0:    (&elt_str_mat)->x[2][1] = 1; break;
                  case 1:    (&elt_str_mat)->x[2][2] = 1; break;
                }
                break;
              case 0:
                 switch (dy) {
                    case (-1): (&elt_str_mat)->x[1][0] = 1; break;
                    case 0:    (&elt_str_mat)->x[1][1] = 1; break;
                    case 1:    (&elt_str_mat)->x[1][2] = 1; break;
                 } 
                 break;
              case 1:
                 switch (dy) {
                    case (-1): (&elt_str_mat)->x[0][0] = 1; break;
                    case 0:    (&elt_str_mat)->x[0][1] = 1; break;
                    case 1:    (&elt_str_mat)->x[0][2] = 1; break;
                 }
                 break;
           }
 
        }

        /* store the translated image in the working image */
        if (!copyimagedata1(img_trav, img_trans))
          return(0);

        /* make the translation of the working image */
        if (!lvdil(img_trans, &elt_str_mat, NULL, 0, border, img_trav))
          return(0);

        /* the output image is the union of the output image and the */
        /* working one                                               */
        if (!lvsup(img_out, img_trav, 1, 0, NULL, 0, img_out))
          return(0);
   }
   return(1);
}
/* -library_code_end */
