 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * Copyright (C) 1993-1997 Junior Barrera, Roberto Lotufo.
 * All rights reserved.  See $MMACH/repos/license/License.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vdiskstr
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvdiskstr
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvdiskstr - Creates disk-like structuring elements for use in  morphological operations
* 
*       Purpose: .I vgdisk
generates special structuring elements which are disks in one of three well-known
metrics:
.IP _
Euclidean : D(p,q) = sqrt( (px-qx)^2 + (py-qy)^2 )
.IP _
City-Block: D(p,q) = |px-qx| + |py-qy|
.IP _
Chessboard: D(p,q) = max( |px-qx| , |py-qy| )
.LP
Disks in each one of the above metrics are respectively circles,
diamonds and squares. For a concise review of this subject, see Gonzales &
Wintz, "Digital Image Processing", chapter 2.
.LP
The disks generated may be planar or nonplanar, 2-dimensional or 3-dimensional.
The default is euclidean planar 2-D.
.LP
The offset makes sense only for the nonplanar case, in which case the value
specified is added to the element. If no offset or 0 is specified, the element
values will be adjusted so as to have zero in the borders. For example, a
5x5 chessboard nonplanar 2-D element will look like this:
                          0 0 0 0 0
                          0 1 1 1 0
                          0 1 2 1 0
                          0 1 1 1 0
                          0 0 0 0 0
.LP
The disk-like structuring elements  are very useful in many morphological operations
such as  distance transforms, granulometry, skeleton algorithms etc.
*         Input: .IP
.br
ker    -  pointer to structuring element
.br
d      -  disk diameter
.br
type   -  disk type (planar or nonplanar)
.br
offset -  value to be added to nonplanar elements 
.br
metric -  string which specifies metric
*        Output: .IP
.br
ker  - desired structuring element
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: The disks generated are "planar", i.e., they define only a 2-D or 3-D domain,
assuming no values.
*    Written By: Ulisses Braga
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvgdisk (
      mat_str  *ker,
      int      d,
      int      type,
      int      offset,
      int      dim,
      char     *met)
/* -library_def_end */

/* -library_code */
{
    register int   i,j,k;
    int            r,side,side2,mside,min;
    xvimage *img1,*img2;
    unsigned char  *pi;
    short          *psi;
    char           *program = "lvgdisk";

    /* Error checking: */
    if (d<=0) {
      kfprintf(kstderr,"%s: disk diameter must be a positive number.\n",program); 
      return(0); 
      }
 
    if (d % 2) 
      side = d;
    else
      side = d + 1; 
    r = d / 2;         

    /* dim = 1; (3D) */
    /* kprintf("offset = %d\n",offset); */

    switch(type) {
    case 0: /* PLANAR */
   
    if (!dim) {  /* 2-D case */
      /* Switch among metrics: */
      if (!kstrcmp(met,"euclidean")) { 
        if (!lvgcirc(&img1, side, side, VFF_TYP_1_BYTE, 0.0, 1.0,
                       (float) r, (float) r, (float) d)) {
          kfprintf(kstderr,"%s: lvgcirc Failed.\n",program); 
          return(0); 
          }               
        }
      else if (!kstrcmp(met,"city-block")) {
        if (!gen_const(&img1, side, side, VFF_TYP_1_BYTE, 1, 0.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                        
        pi = (unsigned char *) img1->imagedata;
        for (i=0; i<=r; i++)
          for (j=r-i; j<=r+i; j++) 
            pi[i*side+j] = pi[(side-i-1)*side+j] = 1;
        }
      else if (!kstrcmp(met,"chessboard")) {
        if (!gen_const(&img1, side, side, VFF_TYP_1_BYTE, 1, (float) 1, (float) 0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                  
        }
      else {
        kfprintf(kstderr,"%s: unknown metric.\n",program); 
        return(0); 
        }   
      }

    else { /* 3-D case */
      /* Switch among metrics: */
      if (!kstrcmp(met,"euclidean")) { 
        if (!gen_const(&img1, side, side, VFF_TYP_1_BYTE, side, 0.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }
        pi = (unsigned char *) img1->imagedata;
        for (k=-r; k<=r; k++)
          for (i=-r; i<=r; i++)
            for (j=-r; j<=r; j++) {
              if (i*i+j*j+k*k <= r*r) *pi = 1;
              pi++;
              }
        }
      else if (!kstrcmp(met,"city-block")) {
        if (!gen_const(&img1, side, side, VFF_TYP_1_BYTE, side, 0.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                        
        side2 = side*side;
        pi = (unsigned char *) img1->imagedata;
        for (k=0; k<=r; k++)
          for (i=0; i<=k; i++)
            for (j=r-i; j<=r+i; j++) 
              pi[k*side2+(i+r-k)*side+j] =
              pi[k*side2+(side-i-r+k-1)*side+j] = 
              pi[(side-k-1)*side2+(i+r-k)*side+j] =
              pi[(side-k-1)*side2+(side-i-r+k-1)*side+j] = 1;
        }
      else if (!kstrcmp(met,"chessboard")) {
        if (!gen_const(&img1, side, side, VFF_TYP_1_BYTE, side, 1.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                  
        }
      else {
        kfprintf(kstderr,"%s: unknown metric.\n",program); 
        return(0); 
        }
      }
    break;
     
    case 1: /* NONPLANAR */
    if (!dim) {  /* 2-D case */
      /* Switch among metrics: */
      if (!kstrcmp(met,"euclidean")) { 
        if (!gen_const(&img1, side, side, VFF_TYP_2_BYTE, 1, -1.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                        
        psi = (short *) img1->imagedata;
        for (i=-r; i<=r; i++)
          for (j=-r; j<=r; j++) {
            if (i*i+j*j <= r*r) *psi = (int)(sqrt((float)r*r-i*i-j*j)+0.5) + offset; 
            psi++;
            }
        }
      else if (!kstrcmp(met,"city-block")) {
        if (!gen_const(&img1, side, side, VFF_TYP_2_BYTE, 1, -1.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                        
        psi = (short *) img1->imagedata;
        for (i=0; i<=r; i++)
          for (j=r-i; j<=r+i; j++) 
            psi[i*side+j] = psi[(side-i-1)*side+j] = r-(r-i+abs(j-r)) + offset;
        }
      else if (!kstrcmp(met,"chessboard")) {
        if (!gen_const(&img1, side, side, VFF_TYP_2_BYTE, 1, 0.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                  
        psi = (short *) img1->imagedata;
        for (i=0; i<=r; i++)
          for (j=0; j<=r; j++) 
            psi[i*side+j] = psi[i*side+(side-j-1)] = psi[(side-i-1)*side+j] =
            psi[(side-i-1)*side+(side-j-1)] = ((i<j)? i : j) + offset;
        }
      else {
        kfprintf(kstderr,"%s: unknown metric.\n",program); 
        return(0); 
        }   
      }

    else { /* 3-D case */
      /* Switch among metrics: */
      if (!kstrcmp(met,"euclidean")) { 
        if (!gen_const(&img1, side, side, VFF_TYP_2_BYTE, side, -1.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }
        psi = (short *) img1->imagedata;
        for (k=-r; k<=r; k++)
          for (i=-r; i<=r; i++)
            for (j=-r; j<=r; j++) {
              if (i*i+j*j+k*k <= r*r) *psi = (int)(sqrt((float)r*r-i*i-j*j-k*k)+0.5) + offset; 
              psi++;
              }
        }
      else if (!kstrcmp(met,"city-block")) {
        if (!gen_const(&img1, side, side, VFF_TYP_2_BYTE, side, -1.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }                        
        side2 = side*side;
        psi = (short *) img1->imagedata;
        for (k=0; k<=r; k++)
          for (i=0; i<=k; i++)
            for (j=r-i; j<=r+i; j++) 
              psi[k*side2+(i+r-k)*side+j] =
              psi[k*side2+(side-i-r+k-1)*side+j] = 
              psi[(side-k-1)*side2+(i+r-k)*side+j] =
              psi[(side-k-1)*side2+(side-i-r+k-1)*side+j] = k-(k-i+abs(j-r));
        }
      else if (!kstrcmp(met,"chessboard")) {
        if (!gen_const(&img1, side, side, VFF_TYP_2_BYTE, side, 0.0, 0.0)) {
          kfprintf(kstderr,"%s: gen_const Failed.\n",program); 
          return(0); 
          }
        side2 = side*side;                  
        psi = (short *) img1->imagedata;
        for (k=0; k<=r; k++)
          for (i=0; i<=r; i++)
            for (j=0; j<=r; j++) {
              min = i;
              if (j<min) min = j;
              if (k<min) min = k;
              psi[k*side2+i*side+j] =
              psi[k*side2+i*side+(side-j-1)] =
              psi[k*side2+(side-i-1)*side+j] =
              psi[k*side2+(side-i-1)*side+(side-j-1)] =
              psi[(side-k-1)*side2+i*side+j] =
              psi[(side-k-1)*side2+i*side+(side-j-1)] =
              psi[(side-k-1)*side2+(side-i-1)*side+j] =
              psi[(side-k-1)*side2+(side-i-1)*side+(side-j-1)] =
              min + offset;
	      }
        }
      else {
        kfprintf(kstderr,"%s: unknown metric.\n",program); 
        return(0); 
        }
      }  
    break;

    default:
      kfprintf(kstderr,"%s: type must be planar(0) or nonplanar(1).\n",program); 
      return(0); 
    }

    /* Convert to struc: */
    if (!lviff2str(img1, ker, type)) {
      kfprintf(kstderr,"%s: lviff2str Failed.\n",program); 
      return(0); 
      }                  
  
    return(1); /* Success */
}
/* -library_code_end */
