 /*
  * Khoros: $Id$
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id$";
#endif

/*
 * Copyright (C) 1993-1997 Junior Barrera, Roberto Lotufo.
 * All rights reserved.  See $MMACH/repos/license/License.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vwatersh
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvwatersh
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
#define MASK -2
#define WSHED 0
#define INIT -1
#define INQUEUE -3
#define BORDER -4
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvwatersh - Do the watershed of a gray-scale image.
* 
*       Purpose: .I lvwatersh
Will do the watershed of a gray-scale image.
img1 is the input image (of the VIFF_TYP_1_BYTE data type), img2 is the output image (of the VFF_TYP_2_BYTE data type). cv is an integer parameter that indicates wich connectivity is desired (4 for 4-connectivity or 8 for 8-connectivity). lw is a flag: true if border initialization with 0 is desired. If you want to use lwatersh in your program, note that if lw is true, the input image's inner boundary will be modified.
*         Input: .IP "img1" 15  
input xvimage structure of the VIFF_TYP_1_BYTE data type.
.IP "cv" 15
chosen connectivity: 4 or 8.
.IP "lw" 15
flag value, true if boundary initialization (with 0) is desired.
.IP
Note that img1 should be a binary image of the BYTE1 data type.
*        Output: .IP "img2" 15
holds the watershed operation. The output image should be of the VIFF_TYP_2_BYTE data type.
.IP "Return Value:" 15
1 on success, 0 on failure.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: .I lvwatersh
The input and output images must be of the same size. The input and output image data types should be, respectively, VFF_TYP_1_BYTE and VFF_TYP_2_BYTE.
*    Written By: Adriano Nagelschmidt Rodrigues
Based on an algorithm proposed by Pierre Soille and Luc Vincent.
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int
lvwatersh(xvimage *img1, int cv, int lw, xvimage *img2)
/* -library_def_end */

/* -library_code */
{
  int nc, nr, k;
  int d[8], cr_label, flag;
  unsigned long H[256], HC[257], *ims;
  int hmin, hmax, lower_b;
  register int i, j, l, r, t, pos;
  item_str p_num;
  queue_str q;
  unsigned char *c1;
  short *s1;

  if ((cv != 4) && (cv != 8)) {
    kfprintf(kstderr,"lvwatersh: illegal value on parameter cv\n");
    return(0);
  }

  switch (img1->data_storage_type) {
    case VFF_TYP_BIT :
      kfprintf(kstderr,"lvwatersh: BIT not implemented\n");
      return(0);
      break;

    case VFF_TYP_1_BYTE :
      nc = img1->row_size;
      nr = img1->col_size;
      k = nc * nr;
      c1 = (unsigned char *)(img1->imagedata); 
      s1 = (short *)(img2->imagedata);

      d[0] = -nc;          /* neighborhood pixels */
      d[1] = 1;
      d[2] = nc;
      d[3] = -1;
      d[4] = 1 - nc;
      d[5] = 1 + nc;
      d[6] = nc - 1;
      d[7] = -1 - nc; 

      if (lw) {              /* write 0 to image inner boundary */
        r = (nr - 2) * nc;
        for (i = 1; i < nc - 1; i++) { 
          c1[i + nc] = 0;
          c1[i + r] = 0;
        }
        for (i = nc; i < k - nc; i += nc) {
          c1[i + 1] = 0;
          c1[i + nc - 2] = 0;
        }
      }

      ims = (unsigned long *) kmalloc(k * sizeof(long));
      if (ims == NULL) {
        kfprintf(kstderr, "lvwatersh: memory allocation failed\n");
        return(0);
      }

      for (i = 0; i < k; i++)    /* output image initialization */
        s1[i] = INIT;

      for (i = 0; i < nc; i++) {      /* Mark image frame */
        s1[i] = BORDER;
        s1[i + k - nc] = BORDER;
      }
      for (i = 0; i < k; i += nc) {
        s1[i] = BORDER;
        s1[i + nc - 1] = BORDER;
      }

      queue_init(&q);            /* initialize pixels' queue */

      flag = 0;
      cr_label = 0;

      /* beginning of the tri-distributive function */

      for (i = 0; i < 256; i++) {   /* initialize histogram */
        H[i] = 0;
        HC[i] = 0;
      }

      hmax = hmin = c1[0];

      for (i = 1; i < nc - 1; i++)
        for (j = 1; j < nr - 1; j++) {
          l = i + j * nc;
          H[c1[l]] += 1;
          if (c1[l] < hmin)
            hmin = c1[l];
          if (c1[l] > hmax)
            hmax = c1[l];
        }

      HC[hmin] = 0;

      for (i = hmin + 1; i <= hmax + 1; i++)
        HC[i] = HC[i - 1] + H[i - 1];

      for (i = 1; i < nc - 1; i++)
        for (j = 1; j < nr - 1; j++) {
          l = i + j * nc;
          ims[HC[c1[l]]] = l;
          HC[c1[l]] += 1;
        }

      /* end of the tri-distributive function */


      for (i = hmin; i <= hmax; i++) {
        if (i != 0)
          lower_b = HC[i - 1];
        else
          lower_b = 0;

        for (j = lower_b; j < HC[i]; j++) {
          l = ims[j];
          s1[l] = MASK;
          for (r = 0; r < cv; r++) {
            pos = l + d[r];
            if ((s1[pos] > 0) || (s1[pos] == WSHED)) {
              p_num.address = l;
              queue_add(&q, &p_num);
              s1[l] = INQUEUE;
            }
          }
        }

        while (!queue_empty(&q)) {
          queue_first(&q, &p_num);
          l = p_num.address;
          for (r = 0; r < cv; r++) {
            pos = l + d[r];
            if (s1[pos] > 0) {
              if ((s1[l] == INQUEUE) || ((s1[l] == WSHED) && flag))
                s1[l] = s1[pos];
              else
                if ((s1[l] > 0) && (s1[l] != s1[pos])) {
                  s1[l] = WSHED;
                  flag = 0;
                }
            }
            else
              if (s1[pos] == WSHED) { 
                if (s1[l] == INQUEUE) {
                  s1[l] = WSHED;
                  flag = 1;
                }
              }
              else
                if (s1[pos] == MASK) {
                  s1[pos] = INQUEUE;
                  p_num.address = pos;
                  queue_add(&q, &p_num);
                }
          }
        }

        for (j = lower_b; j < HC[i]; j++) {
          l = ims[j];
          if (s1[l] == MASK) {
            s1[l] = ++cr_label;
            p_num.address = l;
            queue_add(&q, &p_num);
            while (!queue_empty(&q)) {
              queue_first(&q, &p_num);
              pos = p_num.address;
              for (r = 0; r < cv; r++) {
                t = pos + d[r];
                if (s1[t] == MASK) {
                  p_num.address = t;
                  queue_add(&q, &p_num);
                  s1[t] = cr_label;
                }
              }
            }
          }
        }

      }

           
      for (i = 0; i < nc; i++) {      /* Write WSHED to image frame */
        s1[i] = WSHED;
        s1[i + k - nc] = WSHED;
      }
      for (i = 0; i < k; i += nc) {
        s1[i] = WSHED;
        s1[i + nc - 1] = WSHED;
      }
 
      kfree_and_NULL(ims);
      return(1);
      break;
 
    case VFF_TYP_2_BYTE :
      kfprintf(kstderr,"lvwatersh: BYTE2 not implemented\n");
      return(0);
      break;

    case VFF_TYP_4_BYTE :
      kfprintf(kstderr,"lvwatersh: BYTE4 not implemented\n");
      return(0);
      break;

    case VFF_TYP_FLOAT :
      kfprintf(kstderr,"lvwatersh: FLOAT not implemented\n");
      return(0);
      break;

    case VFF_TYP_COMPLEX :
      kfprintf(kstderr,"lvwatersh: COMPLEX not implemented\n");
      return(0);
      break;

    case VFF_TYP_DOUBLE :
      kfprintf(kstderr,"lvwatersh: DOUBLE not implemented\n");
      return(0);
      break;

    case VFF_TYP_DCOMPLEX :
      kfprintf(kstderr,"lvwatersh: COMPLEXD not implemented\n");
      return(0);
      break;

    default:
      kfprintf(kstderr,"lvwatersh: Unknown data storage type, %d\n",img1->data_storage_type);
      return(0);
      break;
  }
}
/* -library_code_end */
