/*
 *--------------------------------------------------------------------------
 * COPYRIGHT
 *
 * A Mathematical Morphology Toolbox for the Khoros System 
 *
 * Copyright 1993 - University of Sao Paulo - USP
 *                  University of Campinas - UNICAMP
 *                  National Institute for Space Research - INPE
 *
 * All rights reserved.
 *
 * Due to potential patent problems, this program may only be used
 * for research in morphological image processing and practical classes
 * in University courses. The authors do not take any resposibilities
 * for use beyond this scope, nor do we claim any suitability and  
 * operability of this software for any purpose. It is provided "as is"
 * without express or implied warranty.
 *
 * I DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
 * NESS.  IN NO EVENT SHALL I BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
 * SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
 * FORMANCE OF THIS SOFTWARE.
 * 
 * No other rights, including, for example, the right to redis-
 * tribute this software and its documentation or the right to
 * prepare derivative works, are granted unless specifically
 * provided in a separate license agreement.
 *--------------------------------------------------------------------------
 */

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 >>>>
 >>>>         File Name: tools.c
 >>>>
 >>>> Date Last Updated: January 20 1994 
 >>>>
 >>>>          Routines: breverse, bitreverse, format, unformat,
 >>>>                    read_str, write_str, mat2vec_str,
 >>>>                    copyimagedata1, setimage gen_const;
 >>>>
 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>   <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

#include "internals.h"

/****************************************************************
* 
*  Routine Name: lvdiskstr - Creates disk-like structuring elements for use in  morphological operations
* 
*       Purpose: *    
*    Creates an image with constant pixel values.
*    
*
*         Input: *    
*    Input to gen_const :
*    
*    image          pointer to VIFF structure containing image data to
*                   be processed.
*    
*    rows           number of rows in output image.
*    
*    cols           number of columns in output image.
*    
*    type      Uses the data storage type convension from the  defines
*              in   the   viff   header.    VFF_TYP_1_BYTE   =   char,
*              VFF_TYP_2_BYTE   =   short,   VFF_TYP_4_BYTE   =   int,
*              VFF_TYP_FLOAT   =  float,  VFF_TYP_COMPLEX  =  complex,
*              VFF_TYP_DOUBLE = double, and VFF_TYP_DCOMPLEX  =  dcom-
*              plex.
*    
*    bnds      defines the number of  data  bands  for  the  resulting
*              image.
*
*    real      specifices the constant value (level) for the real part
*    
*    imag      specifices the constant value for the imaginary part.
*    
*
*        Output: *    
*    Output from gen_const :
*    
*    image          pointer to VIFF structure  containing  image  data
*                   after processing.
*    
*    Return Value:  1 on success, 0 on failure.
*    
*    
*
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: The disks generated are "planar", i.e., they define only a 2-D or 3-D domain,
assuming no values.
*    Written By: Ulisses Braga
*          Date: May 12, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
int gen_const (
    xvimage **image,
    int rows,
    int cols,
    int type,
    int bnds,
    float real,
    float imag)
{
    int     i, bsize;
    xvimage *kern;

    int      *iptr;
    short    *sptr;
    float    *fptr;
    double   *dptr;
    unsigned char *cptr, bit;


    /* kprintf("rows = %d \n cols = %d \n type = %d \n real = %f \n imag = %f \n bands = %d \n",rows,cols,type,real,imag,bnds); */

    kern = createimage((unsigned long) rows,    /* number of rows */
                        (unsigned long) cols,   /* number of columns */
                        (unsigned long) type,   /* data storage type */
                        (unsigned long) 1,      /* num_of_images */ 
                        (unsigned long) bnds,  /* num_data_bands */ 
                        "created by vgdisk",     /* comment */
                        (unsigned long) 0,      /* map_row_size */
                        (unsigned long) 0,      /* map_col_size */
                        (unsigned long)
                        VFF_MS_NONE,             /* map_scheme */
                        (unsigned long)
                        VFF_MAPTYP_NONE,        /* map_storage_type */
                        (unsigned long)
                        VFF_LOC_IMPLICIT,       /* location_type */
                        (unsigned long) 0);     /* location_dim */

    if (kern == NULL)
    {
       (void)kfprintf(kstderr,"gen_const: Unable to allocate new image!\n");
       *image = NULL;
       return(0);
     }

    if ( kern->data_storage_type == VFF_TYP_1_BYTE )
    {
       if ( real > 255 || real < 0 ) 
       {
          (void) kfprintf(kstderr,"gen_const: byte type image 0 < real < 255 !\n");
          return(0);
       }
    }

    switch (kern->data_storage_type)
    {
      case VFF_TYP_BIT:
           if (real != 0.0) bit = 0xff;
           else bit = 0;

           cptr = (unsigned char *) kern->imagedata;
           bsize = ((cols+7)/8)*rows*bnds;
           for (i = 0; i < bsize; i++) *cptr++ = bit;
           break;
      case VFF_TYP_1_BYTE:
           cptr = (unsigned char *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) *cptr++ = (unsigned char) real;
           break;
      case VFF_TYP_2_BYTE:
           sptr = (short *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) *sptr++ = (short) real;
           break;
      case VFF_TYP_4_BYTE:
           iptr = (int *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) *iptr++ = (int) real;
           break;
      case VFF_TYP_FLOAT:
           fptr = (float *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) *fptr++ = real;
           break;
      case VFF_TYP_DOUBLE:
           dptr = (double *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) *dptr++ = real;
           break;
      case VFF_TYP_COMPLEX:
           fptr = (float *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) {
               *fptr++ = real;
               *fptr++ = imag;
           }
           break;
      case VFF_TYP_DCOMPLEX:
           dptr = (double *) kern->imagedata;
           for (i = 0; i < rows*cols*bnds; i++) {
               *dptr++ = real;
               *dptr++ = imag;
           }
           break;
      default:
           (void) kfprintf(kstderr,"gen_const: Unsupported data storage type %d\n",
                       kern->data_storage_type);
        return(0);
    }

    *image = kern;
    return(1);
}

/****************************************************************
*
* Routine Name: breverse
*               This routine is used in the format and unformat routines.
*
*         Date: September 24 1993
*
*      Purpose: reverse bits in 8 bit clusters (32 bit words) 
*
*   Input data: bit image data formatted as following:
*               7..0 15..8 23..16 31..24  (32 bit word)
*          
*  Output data: bit image data formatted as following:    
*               0..7 8..15 16..23 24..31  (32 bit word)
*
*        Input: l2 - pointer to BIT imagedata
*               nr - total number of rows (of all image bands)
*               col32 - number columns of 32 bits
*
*   Written By: Adriano Nagelschmidt Rodrigues & Luiz Fernando Vieira
*        
****************************************************************/

void breverse(unsigned long *l2, int nr, int col32)
{
  int i, j, jj, count, row;
  unsigned long xx, mask, aux;

  row = 0;         
  for (i=0; i<nr; i++) {
    for (j=0; j<col32; j++) {
      jj = row + j;
      xx = 0;
      mask = 0x1010101;
      for (count=7; count>=-7; count-=2) {
        aux = mask & l2[jj];
        if (count>0) aux <<= count;
        else aux >>= (-count);
        xx |= aux;
        mask <<= 1;             
      }
      l2[jj] = xx; 
    }
    row += col32;
  }
}


/***********************************************************************
*
*  Routine Name: format
*
*          Date: October 20 1993
*        
*       Purpose: transforms image width to 32-multiple ready to be
*	  	 processed in 32 bit words  
*
*         Input: viff image
*
*        Output: 32-multiple number of columns viff image
*                right border mask (to the rightmost pixel) 
*
*    Written By: Luiz Fernando Vieira & Adriano Nagelschmidt Rodrigues  
*
* Modifications:
*
***********************************************************************/

xvimage *format(xvimage *img1, unsigned long *border)
{
  xvimage *imgout;
  int col32,		/* number of columns of 32 pixels */
      rwtot;            /* total number of rows */
  int nbl,              /* number of bytes per line */
      tsize,            /* total size of output image */
      new_row,          /* number of columns of output image */
      tofill,           /* number of bytes to fill */
      i, j;             /* auxiliar */
          
  rwtot = img1->col_size*img1->num_data_bands;
  col32 = (img1->row_size + 31) >> 5; 
         
  if (!(img1->row_size % 32))
    *border = 1;
  else
    *border = ((unsigned long) 0x80000000L) >> ((img1->row_size % 32) - 1);

  if (((img1->row_size % 32)<25) && (img1->row_size % 32)) {

    nbl = img1->row_size/8 + ((img1->row_size % 8) ? 1 : 0);
    tsize = nbl * rwtot;
    tofill = 4 - (nbl % 4);
    new_row = col32 * 32;

    imgout = createimage(img1->col_size, new_row, img1->data_storage_type, img1->num_of_images, img1->num_data_bands, img1->comment, img1->map_row_size, img1->map_col_size, img1->map_scheme, img1->map_storage_type, img1->location_type, img1->location_dim);

    if (imgout == NULL) {
      (void)  kfprintf(kstderr, "createimage Failed\n");
      exit(1);
    }

    j = 1;
    imgout->imagedata[0] = img1->imagedata[0];
    for (i=1; i<tsize; i++) {
      if (!(i % nbl))
        j += tofill;
      imgout->imagedata[j++] = img1->imagedata[i];
    }
    (void) breverse((unsigned long *)imgout->imagedata, rwtot, col32);
    return(imgout); 
  }
  else {
    (void) breverse((unsigned long *)img1->imagedata, rwtot, col32);
    return(img1);
  }
}

/***********************************************************************
*
*  Routine Name: unformat
*
*          Date: October 20 1993
*        
*       Purpose: undo modifications made by the format routine 
*
*         Input: formatted size image (output of the above routine)
*
*        Output: original size image
*
*    Written By: Luiz Fernando Vieira & Adriano Nagelschmidt Rodrigues 
*
* Modifications:
*
***********************************************************************/

void unformat(xvimage *img1, xvimage *img2)
{
  int col32,		/* number of columns of 32 pixels */
      rwtot;            /* total number of rows */
  int nbl,              /* number of bytes per line */
      tsize,            /* total size of output image */
      tofill,           /* number of bytes to fill */
      i, j;             /* auxiliar */
  
  rwtot = img1->col_size*img1->num_data_bands;
  col32 = (img2->row_size + 31) >> 5;

  (void) breverse((unsigned long *)img1->imagedata, rwtot, col32);

   if (img1 != img2) {
     nbl = img2->row_size/8 + ((img2->row_size % 8) ? 1 : 0);
     tsize = nbl * rwtot;
     tofill = 4 - (nbl % 4);
     j = 1;
     img2->imagedata[0] = img1->imagedata[0];
     for (i=1; i<tsize; i++) {
       if (!(i % nbl))
         j += tofill;
       img2->imagedata[i] = img1->imagedata[j++];
     }
   }
}


/********************************************************
*
* Routine Name: read_str
*
*      Purpose: read a structuring element file. 
*
*        Input: file name
*
*       Output: structuring element in the generic form.
*      
*       Return: 1 on success, 0 on failure,  
* 
*   Written by: Ulisses de Mendonca Braga Neto
*
****************************************************************/

int
read_str(char name[50], mat_str *ker)
{
  int      i,j,size,in,flag_mat=1;
  char     sign[6],buf[512];
  short    *pi;
  kfile    *f;

  /* Find complete path: */
  if (!kfullpath(name,NULL,buf)) return(0);

  /* Try to open file: */
  if ((f= kfopen(buf, "rb")) == NULL) {
    (void) kfprintf(kstderr, "read_str: fopen failed\n");
    return(0);
    } 
  if (kfread(sign, 1, 5, f) < 5) {
    (void) kfprintf(kstderr, "read_str: fread Failed\n");
    kfclose(f);
    return(0);
    }  
  sign[5] = '\0';

  /* See if we have an STRUC file */
  if (strcmp(sign, "STRUC") == 0) {
    kfscanf(f," %d",&ker->m);  /* Read dimensions */
    kfscanf(f," %d",&ker->n);
    kfscanf(f," %d\n",&ker->p);
    size = ker->m*ker->n*ker->p;

    ker->dat = (short *) malloc(size*sizeof(short));
    for (i=0; i<size; i++) {  /* Read data */
      kfscanf(f, "%d ", &in);
      if (in>0) flag_mat=0;
      ker->dat[i] = in;
      }

    if (ker->n==3 && ker->m==3 && ker->p==1 && flag_mat) {
      /* This should become a MAT structuring element: */
      pi = (short *) ker->dat;
      for (i=0; i<3; i++)
        for (j=0; j<3; j++) {
          ker->x[j][i] = *pi++ + 1;
          /* OBS: MAT format is the transposed of Khoros format */        
          }
      ker->tag = MM_MAT;
      }
    
    else {
      /* Let's stick with the IMG structuring element: */
      ker->tag = MM_IMG;
      }

    kfclose(f);
    return(1); /* Success */
    }

  /* Unknown file format: */
  (void) kfprintf(kstderr,"read_str: unrecognizable file format\n");
  kfclose(f);
  return(0); 
}

/********************************************************
*
* Routine Name: write_str
*
*      Purpose: write a structuring element file.
*
*        Input: file name & structuring element in the
*                 generic form.
*
*       Output: file written.
*
*       Return: 1 on success, 0 on failure,  
* 
*   Written by: Ulisses de Mendonca Braga Neto 
*
****************************************************************/

int
write_str(char name[50], mat_str *ker)
{
  int            i,j,out,size;
  char           buf[512];
  kfile           *f;

  /* Find complete path: */
  if (!kfullpath(name,NULL,buf)) return(0);

  /* Try to open file: */
  if ((f= kfopen(buf, "w")) == NULL) {
       (void) kfprintf(kstderr, "write_str: fopen Failed\n");
       return(0);
       } 
 
  kfprintf(f, "STRUC"); /* Output signature */

  switch (ker->tag) {
  case MM_MAT:
    kfprintf(f, " 3 3 1\n");  /* Output dimensions */
    for (i=0; i<3; i++) 
      for (j=0; j<3; j++) {
        out = (ker->x[j][i]) ? 0 : -1;  
        /* OBS: MAT format is the transposed of Khoros format */ 
        kfprintf(f, "%d ", out);  /* Output data */
      }
    break;

  case MM_IMG:
    kfprintf(f, " %d",ker->m);  /* Output dimensions */
    kfprintf(f, " %d",ker->n);
    kfprintf(f, " %d\n",ker->p);
    size = (ker->m)*(ker->n)*(ker->p);
    for (i=0; i<size; i++) {
      kfprintf(f, "%d ", (ker->dat)[i]);
      if (!((i+1)%20)) kfprintf(f, "\n");   /* Output data */
      }
    break;
    }
   
  kfclose(f);
  return(1); /* Success */
}

/********************************************************
*
* Routine Name: mat2vec_str
*
*      Purpose: Converts a structuring element from the matrix form to the
*               vector form.
*
*        Input: structuring element in the mat form.
*
*               number of the image region (1 to 9)
*
*                   1 !       2         ! 3
*                 -------------------------
*                     !                 ! 
*
*                   4 !       5         ! 6 
*
*                     !                 !
*                 -------------------------
*                   7 !       8         ! 9        
*
*      Output: structuring element in the vectorial form. 
*
*      Return: 1 on success, 0 on failure.
*
*  Written By: Junior Barrera
*
****************************************************************/

mat2vec_str(mat_str *b_mat, int region, vec_str *b_vec)

{
 int i, j;
 int imin, imax;
 int jmin, jmax;

 switch(region) {

    case 1: imin = 1;
            imax = 2;
            jmin = 1;
            jmax = 2;
            break;

    case 2: imin = 0;
            imax = 2;
            jmin = 1;
            jmax = 2;
            break;
      
   case 3:  imin = 0;
            imax = 1;
            jmin = 1;
            jmax = 2;
            break;

     
   case 4:  imin = 1;
            imax = 2;
            jmin = 0;
            jmax = 2;
            break;

     
   case 5:  imin = 0;
            imax = 2;
            jmin = 0;
            jmax = 2;
            break;

     
   case 6:  imin = 0;
            imax = 1;
            jmin = 0;
            jmax = 2;
            break;

     
   case 7:  imin = 1;
            imax = 2;
            jmin = 0;
            jmax = 1;
            break;

     
   case 8:  imin = 0;
            imax = 2;
            jmin = 0;
            jmax = 1;
            break;

   case 9:  imin = 0;
            imax = 1;
            jmin = 0;
            jmax = 1;
            break;
  }


b_vec->n = 0;

for (i = imin; i <= imax; i++)
  for (j = jmin; j <= jmax; j++) { 

      if (b_mat->x[i][j] == 1) {
          switch(i) {

             case 0:  switch(j) {

                        case 0:  b_vec->x[b_vec->n] = -1;
                                 b_vec->y[b_vec->n] = 1;
                                 break;

                        case 1:  b_vec->x[b_vec->n] = -1;
                                 b_vec->y[b_vec->n] = 0;
                                 break;

                        case 2:  b_vec->x[b_vec->n] = -1;
                                 b_vec->y[b_vec->n] = -1;
                                 break;
                       }
                       break; 



             case 1:  switch(j) {

                        case 0:  b_vec->x[b_vec->n] = 0;
                                 b_vec->y[b_vec->n] = 1;
                                 break;

                        case 1:  b_vec->x[b_vec->n] = 0;
                                 b_vec->y[b_vec->n] = 0;
                                 break;

                        case 2:  b_vec->x[b_vec->n] =  0;
                                 b_vec->y[b_vec->n] = -1;
                                 break;
                       }
                       break; 


             case 2:  switch(j) {

                        case 0:  b_vec->x[b_vec->n] = 1;
                                 b_vec->y[b_vec->n] = 1;
                                 break;

                        case 1:  b_vec->x[b_vec->n] = 1;
                                 b_vec->y[b_vec->n] = 0;
                                 break;

                        case 2:  b_vec->x[b_vec->n] = 1;
                                 b_vec->y[b_vec->n] = -1;
                                 break;
                       }
                       break; 

             }

       b_vec->n = (b_vec->n) + 1;
    }
  }
  return (1);
}  




/**************************************************************
*
* Routine Name: copyimagedata1
*
*      Purpose: Copy image data
*
*        Input: img1 -- original image
*
*       Output: img2 -- new image
*
*       Return: 1 upon success and 0 upon failure.
* 
*   Written by: Junior Barrera
*
**************************************************************/

copyimagedata1(xvimage *img1, xvimage *img2)
{
long	lsize;

/* room for the image data */

    switch (img1->data_storage_type)
    {
	case VFF_TYP_BIT:
	    lsize = ((img1->row_size+7)/8)*img1->col_size;
	    break;
	case VFF_TYP_1_BYTE:
	    lsize = img1->col_size*img1->row_size;
	    break;
	case VFF_TYP_2_BYTE:
	    lsize = img1->col_size*img1->row_size*sizeof(short);
	    break;
	case VFF_TYP_4_BYTE:
	    lsize = img1->col_size*img1->row_size*sizeof(int);
	    break;
	case VFF_TYP_FLOAT:
	    lsize = img1->col_size*img1->row_size*sizeof(float);
	    break;
	case VFF_TYP_DOUBLE:
	    lsize = img1->col_size*img1->row_size*sizeof(double);
	    break;
	case VFF_TYP_COMPLEX:
	    lsize = img1->col_size*img1->row_size*sizeof(float)*2;
	    break;
	case VFF_TYP_DCOMPLEX:
	    lsize = img1->col_size*img1->row_size*sizeof(double)*2;
	    break;
	default:
	    (void) kfprintf(kstderr,"copyimagedata: Not valid data \
storage type\n");
	    return(0);
	    break;
    }
    lsize = lsize*img1->num_of_images*img1->num_data_bands;

    kbcopy(img1->imagedata, img2->imagedata, (int) lsize);

    img2->data_storage_type = img1->data_storage_type;
    img2->row_size = img1->row_size;
    img2->col_size = img1->col_size;
    img2->num_of_images = img1->num_of_images;
    img2->num_data_bands = img1->num_data_bands;

    return(1);
}

/**************************************************************
*
* Routine Name: setimage
*
*      Purpose: write maxval to all image pixels
*
*        Input: img -- xv image
*
*       Output: img -- set xv image
*
* 
*   Written by: Adriano Nagelschmidt Rodrigues
*
**************************************************************/

int setimage(xvimage *img)
{
  int    j, col32, k;
  unsigned char *c;
  unsigned long *l;

  switch (img->data_storage_type) {
    case VFF_TYP_BIT :
      col32 = (img->row_size + 31) >> 5;
      k = img->col_size * col32;
      l = (unsigned long *)(img->imagedata);
      for(j=0; j<k; j++) {
        *l = 0xffffffffL;
        l++;
      }
      return(1);
      break;

    case VFF_TYP_1_BYTE : 
      k = img->row_size * img->col_size * img->num_data_bands;
      c = (unsigned char *)(img->imagedata);     
      for (j=0; j<k; j++) {
        *c = 255;
        c++;
      }     
      return(1);
      break;
 
    case VFF_TYP_2_BYTE :
      kfprintf(kstderr,"setimage: BYTE2 not implemented\n");
      return(0);
      break;

    case VFF_TYP_4_BYTE :
      kfprintf(kstderr,"setimage: BYTE4 not implemented\n");
      return(0);
      break;

    case VFF_TYP_FLOAT :
      kfprintf(kstderr,"setimage: FLOAT not implemented\n");
      return(0);
      break;

    case VFF_TYP_COMPLEX :
      kfprintf(kstderr,"setimage: COMPLEX not implemented\n");
      return(0);
      break;

    case VFF_TYP_DOUBLE :
      kfprintf(kstderr,"setimage: DOUBLE not implemented\n");
      return(0);
      break;

    case VFF_TYP_DCOMPLEX :
      kfprintf(kstderr,"setimage: COMPLEXD not implemented\n");
      return(0);
      break;

    default:
      kfprintf(kstderr,"setimage: Unknown data storage type, %d\n",img->data_storage_type);
      return(0);
      break;
  }
}

/**************************************************************
*
* Routine Name: clearimage
*
*      Purpose: write 0 to all image pixels
*
*        Input: img -- xv image
*
*       Output: img -- cleared xv image
*
* 
*   Written by: Adriano Nagelschmidt Rodrigues
*
**************************************************************/

int clearimage(xvimage *img)
{
  int    j, col32, k;
  short *s;
  unsigned char *c;
  unsigned long *l;

  switch (img->data_storage_type) {
    case VFF_TYP_BIT :
      col32 = (img->row_size + 31) >> 5;
      k = img->col_size * col32;
      l = (unsigned long *)(img->imagedata);
      for(j=0; j<k; j++)
        l[j] = 0;
      return(1);
      break;

    case VFF_TYP_1_BYTE : 
      k = img->row_size * img->col_size * img->num_data_bands;
      c = (unsigned char *)(img->imagedata);     
      for (j=0; j<k; j++)
        c[j] = 0;
      return(1);
      break;
 
    case VFF_TYP_2_BYTE :
      k = img->row_size * img->col_size * img->num_data_bands;
      s = (short *)(img->imagedata);     
      for (j=0; j<k; j++)
        s[j] = 0;
      return(1);
      break;

    case VFF_TYP_4_BYTE :
      kfprintf(kstderr,"setimage: BYTE4 not implemented\n");
      return(0);
      break;

    case VFF_TYP_FLOAT :
      kfprintf(kstderr,"setimage: FLOAT not implemented\n");
      return(0);
      break;

    case VFF_TYP_COMPLEX :
      kfprintf(kstderr,"setimage: COMPLEX not implemented\n");
      return(0);
      break;

    case VFF_TYP_DOUBLE :
      kfprintf(kstderr,"setimage: DOUBLE not implemented\n");
      return(0);
      break;

    case VFF_TYP_DCOMPLEX :
      kfprintf(kstderr,"setimage: COMPLEXD not implemented\n");
      return(0);
      break;

    default:
      kfprintf(kstderr,"setimage: Unknown data storage type, %d\n",img->data_storage_type);
      return(0);
      break;
  }
}
