 /*
  * Khoros: $Id: lvconvert.c,v 1.16.2.3 1997/05/31 20:16:15 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvconvert.c,v 1.16.2.3 1997/05/31 20:16:15 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vconvert
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvconvert
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
static int lvcvt PROTO((xvimage *, int, int, int, double, double));
static int lvcset_sizes PROTO((xvimage *, int, int, int *, int *));
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvconvert - Convert an image data type
* 
*       Purpose: lvconvert - performs data storage type conversions from any
*		 VIFF type to any other, with the exception of the transform
*		 types.  Scaling or (mutually exclusive) normalization can be
*		 performed if requested.  The scaling and normalization options
*		 are important for controlling the loss of precision when
*		 converting from a higher data type to a lower data type.
*
*		 Output data storage type BYTE is a special case. If scaling is
*		 requested, then the input image is scaled, and then clipped to
*		 the range [0..255] before output. If normalization is
*		 requested, the data is offset adjusted and scaled to map onto
*		 the range [0..255] before output.
*
*		 Normalization for other output types simply scales the data so
*		 that the most positive data value is equal to the normalization
*		 target. No offset adjustment is performed.  If the most
*		 positive value in the image is zero then a zero valued image
*		 of the desired data type will be produced.
*
*		 When converting from type BIT to type BYTE, the output values
*		 will be 0 or 255 depending on whether the bit corresponding to
*		 the input pixel is 0 or 1.  For conversions from BIT to any
*		 type other than BYTE, the output value receives either 0 or 1
*		 depending on the input bit value.  This was done so that an
*		 BYTE image would be displayable immediately after conversion
*		 from type BIT.
*
*		 The nflg and sflg are used to indicate whether or not the
*		 image should be scaled or normalized.  If the image is to be
*		 scaled or normalized, the norm_factor and scale_factor give the*		 appropriate values.
*
*		 Note that img1 is used for both the input xvimage structure and
*		 the resulting output xvimage structure.  This is done to save
*		 space, but you must be careful not to overwrite important data.
*
*		 Use of the bflg option is important when trying to normalize a
*		 multiband or vector image. If bflg is false then the largest
*		 value of any vector component in the image is used to compute
*		 the normalization multiplier. If bflg is true then a different
*		 normalization multiplier is used for each band, as dictated by
*		 the data in that band.
*         Input: image - pointer to xvimage structure to be processed
*		 k - VIFF data storage type to convert input image into
*		 sflg - if non-zero then perform scaling using the scale_factor
*		 nflg - if non-zero then perform normalization to norm_factor
*		 norm_factor - value to normalize output image data to
*		 scale_factor - value to scale image data by
*		 bflg - logic switch to request each band be processed
*			independently
*        Output: image - receives the output image after processing
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: There needs to be a way to specify the conversion from a
*		 COMPLEX type to a non-COMPLEX type to be either the real part
*		 or the magnitude (the current version does the real part only).
*    Written By: Scott Wilson
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Scott Wilson and Tom Sauer 11/23/88 - Added cmap type
*		 checking and support for color models other than VFF_CM_GREY.
*
*		 Tom Sauer -- Tue Dec 27 15:15:26 MST 1988 - Updated
*		 lvcset_sizes to handle generic color models
*
*		 Scott Wilson -- updated to XV III 18-Mar-89
*
*		 Mark Young and Tait Cyrus 11/9/89 - lvconvert was not
*		 converting BIT images correctly by not incrementing past the
*		 padding at the end of each scanline.
*
*		 Scott Wilson -- 8-Feb-90 - Fixed bug that caused BIT to * to
*		 ignore scale and normalization factors.
*
*		 Scott Wilson -- 6-Jun-90 - Fixed missing kfree_and_NULL() calls
*
*		 Scott Wilson -- 15-Aug-90 - Renamed lvconvert as lvcvt, for
*		 use as a toolbox routine for the new lvconvert, which is new
*		 code that can do multiband processing with each band being
*		 normalized or scaled independently.  If the band processing
*		 is chosen to be global, then it just falls through to the old
*		 lvconvert code.
*
*		 Scott Wilson -- 8-Oct-90 - Modified the normalization
*		 procedure so that if the max is zero then you get back a zero
*		 image. Before, you would get an error message and no image.
*		 This really screws up editimage.
*
*		 Jeremy Worley Tue Feb 12 14:32:01 MST 1991 - Added support
*		 for double and double complex data types.
*
*		 Jeremy Worley Fri Feb 15 16:24:36 MST 1991 - complex to
*		 complex case indexing bug fixed.  It was indexing (i,2*i+1)
*		 when it should have been indexing (i,i+1)
*
*		 Jeremy Worley Sun Feb 23 11:50:03 MST 1992 - fixed scale_factor
*		 usage so that is is set to 1.0 is sflg is 0. also fixed
*		 messages that indicate invalid source or destination data 
*		 types
****************************************************************/
/* -library_def */
int lvconvert(
	xvimage *image,
	int k,
	int nflg,
	int sflg,
	float norm_factor,
	float scale_factor,
	int bflg)
/* -library_def_end */

/* -library_code */
{
    xvimage *it;  /* Work image area for multiband */
    int i;
    int insize,outsize,trash;
    char *data;

    if (bflg)
      {
        /* User requests independent band processing */
        /* Fix up a 1-band image for a work area */
        it = (xvimage *)kmalloc(sizeof(xvimage));
        if (it == NULL)
          {
            kerror("vipl","lvconvert", "Unable to allocate work area\n");
            return(FALSE);
          }
        copyheader(image,it);
        it->num_data_bands = 1;

        /* Attach space for one band only */
        if (!imagesize(image,&insize,&trash,&trash,&trash,&trash,&trash))
          {
            kerror("vipl","lvconvert", "Unknown image configuration\n");
            freeimage(it);
            return(FALSE);
          }
        insize /= image->num_data_bands;
        it->data_storage_type = k;
        if (!imagesize(it,&outsize,&trash,&trash,&trash,&trash,&trash))
          {
            kerror("vipl","lvconvert", "Unknown image configuration\n");
            freeimage(it);
            return(FALSE);
          }
        it->data_storage_type = image->data_storage_type;
        data = (char *)kmalloc(outsize*image->num_data_bands);
        if (data == NULL)
          {
            kerror("vipl","lvconvert", "Not enough space for new image\n");
            freeimage(it);
            return(FALSE);
          }
        
        /* Process each band separately */
        for (i=0; i<image->num_data_bands; i++)
          {
            /* Copy the i'th band into the work image */
            it->imagedata = (char *)kmalloc((size_t) insize);
            if (it->imagedata == NULL)
              {
                kerror("vipl","lvconvert",
		       "Unable to allocate work data area\n");
                freeimage(it);
                return(FALSE);
              }
            (void) kmemcpy((kaddr) it->imagedata, (kaddr) (image->imagedata+i*insize), (size_t) insize);
            lvcvt(it,k,nflg,sflg,(double) norm_factor,(double) scale_factor);
            (void) kmemcpy((kaddr) (data+i*outsize), (kaddr) it->imagedata, (size_t) outsize);
            it->data_storage_type = image->data_storage_type;
          }
        kfree_and_NULL(image->imagedata);
        image->imagedata = data;
        image->data_storage_type = k;
        freeimage(it);
        return(TRUE);
      }
    else
      {
        /* User requests processing all bands at once */
        if (!lvcvt(image,k,nflg,sflg,(double) norm_factor,(double) scale_factor))
          {
            kerror("vipl","lvconvert", "Unable to convert image\n");
            return(FALSE);
          }
        return(TRUE);
      }
  }
            
static int lvcvt(
	xvimage *image,
	int k,
	int nflg,
	int sflg,
	double norm_factor,
	double scale_factor)
{
    register int i, j, it;
    register short s;
    register float f;
    register unsigned char c;
    int    bmsize,rows,cols,r;
    int    *iptr,*iptr1,imax,imin;
    short  *sptr,*sptr1,smax,smin;
    float  *fptr,*fptr1,fmax,fmin;
    double *dptr; /* new for double conversions */
    unsigned char *bptr,*b,bit;
    unsigned char *cptr,*cptr1,cmax;
    int  size,sz;
    float  multiplier=1.0; /* initial value; jw */
    int really_to_double; /* new for double conversions */
    rows = image->col_size;
    cols = image->row_size;
    size = rows*cols;

/*
** this mod to allow for double and double complex...
*/

    if(image->data_storage_type==VFF_TYP_DOUBLE){
       if((fptr = (float *)kmalloc(size*image->num_data_bands
           *sizeof(float)))==NULL){
          kerror("vipl","vconvert", "Can't malloc memory..\n");
          return(FALSE);
       }
       dptr = (double*)(image->imagedata);
       for(i=0;i<size*image->num_data_bands;i++)fptr[i] = (float)dptr[i];  
       kfree_and_NULL(dptr);
       image->imagedata = (char *)fptr;
       image->data_storage_type = VFF_TYP_FLOAT;
    }else if(image->data_storage_type==VFF_TYP_DCOMPLEX){
       if((fptr = (float *)kmalloc(2*size*image->num_data_bands
           *sizeof(float)))==NULL){
          kerror("vipl","vconvert", "Can't malloc memory..\n");
          return(FALSE);
       }
       dptr = (double*)(image->imagedata);
       for(i=0;i<2*size*image->num_data_bands;i++)fptr[i] = (float)dptr[i];
       image->imagedata = (char *)fptr;
       image->data_storage_type = VFF_TYP_COMPLEX;
    }

    if(k==VFF_TYP_DOUBLE){
       k = VFF_TYP_FLOAT;
       really_to_double = TRUE;
    }else if(k==VFF_TYP_DCOMPLEX){
       k = VFF_TYP_COMPLEX;
       really_to_double = TRUE;
    }else{
       really_to_double = FALSE;
    }
 
/*
** the original lvcvt code ...
*/

    switch(k)
      {
        case VFF_TYP_BIT : /* Target is bit */
          {
            switch(image->data_storage_type)
              {
                case VFF_TYP_BIT: /* Bit to Bit */
                  {
                    /* NOTHING TO CONVERT! */
                    break;
                  }
                case VFF_TYP_1_BYTE: /* CHAR to bit */
                  {
                    if (image->map_scheme != VFF_MS_NONE)
                      {
                        kerror("vipl","lvconvert",
		"Source must be unmapped for conversion to BIT target\n");
                        return(FALSE);
                      }
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    bmsize = ((cols + 7)/8)*r;
                    bptr = (unsigned char *)kmalloc((size_t)bmsize);
                    b = bptr;
                    cptr = (unsigned char *)(image->imagedata);
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to set or clear */
                            if (*cptr++ > 127) *b = *b | bit;
                            else *b = *b & ~bit;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                          {
                            for (j=k; j<8; j++) *b = *b & (~(1 << j));
                            b++;
                          }
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)bptr;
                    image->data_storage_type =  VFF_TYP_BIT;
                    break;
                  }
                case VFF_TYP_2_BYTE: /* SHORT to bit */
                  {
                    if (image->map_scheme != VFF_MS_NONE)
                      {
                        kerror("vipl","lvconvert",
		"Source must be unmapped for conversion to BIT target\n");
                        return(FALSE);
                      }
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    bmsize = ((cols + 7)/8)*r;
                    bptr = (unsigned char *)kmalloc((size_t)bmsize);
                    b = bptr;
                    sptr = (short *)(image->imagedata);
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to set or clear */
                            if (*sptr++) *b = *b | bit;
                            else *b = *b & ~bit;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                          {
                            for (j=k; j<8; j++) *b = *b & (~(1 << j));
                            b++;
                          }
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)bptr;
                    image->data_storage_type =  VFF_TYP_BIT;
                    break;
                  }
                case VFF_TYP_4_BYTE: /* INT to bit */
                  {
                    if (image->map_scheme != VFF_MS_NONE)
                      {
                        kerror("vipl","lvconvert",
		"Source must be unmapped for conversion to BIT target\n");
                        return(FALSE);
                      }
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    bmsize = ((cols + 7)/8)*r;
                    bptr = (unsigned char *)kmalloc((size_t)bmsize);
                    b = bptr;
                    iptr = (int *)(image->imagedata);
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to set or clear */
                            if (*iptr++) *b = *b | bit;
                            else *b = *b & ~bit;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                          {
                            for (j=k; j<8; j++) *b = *b & (~(1 << j));
                            b++;
                          }
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)bptr;
                    image->data_storage_type =  VFF_TYP_BIT;
                    break;
                  }
                case VFF_TYP_FLOAT: /* FLOAT to bit */
                  {
                    if (image->map_scheme != VFF_MS_NONE)
                      {
                        kerror("vipl","lvconvert",
		"Source must be unmapped for conversion to BIT target\n");
                        return(FALSE);
                      }
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    bmsize = ((cols + 7)/8)*r;
                    bptr = (unsigned char *)kmalloc((size_t)bmsize);
                    b = bptr;
                    fptr = (float *)(image->imagedata);
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to set or clear */
                            if (*fptr++) *b = *b | bit;
                            else *b = *b & ~bit;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                          {
                            for (j=k; j<8; j++) *b = *b & (~(1 << j));
                            b++;
                          }
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)bptr;
                    image->data_storage_type =  VFF_TYP_BIT;
                    break;
                  }
                case VFF_TYP_COMPLEX: /* COMPLEX to bit */
                  {
                    if (image->map_scheme != VFF_MS_NONE)
                      {
                        kerror("vipl","lvconvert",
		"Source must be unmapped for conversion to BIT target\n");
                        return(FALSE);
                      }
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    bmsize = ((cols + 7)/8)*r;
                    bptr = (unsigned char *)kmalloc((size_t)bmsize);
                    b = bptr;
                    fptr = (float *)(image->imagedata);
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to set or clear */
                            if (*fptr) *b = *b | bit; /* REAL PART ONLY! */
                            else *b = *b & ~bit;
                            fptr++; fptr++;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                          {
                            for (j=k; j<8; j++) *b = *b & (~(1 << j));
                            b++;
                          }
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)bptr;
                    image->data_storage_type =  VFF_TYP_BIT;
                    break;
                  }
                default :
                  {
                    kerror("vipl","lvconvert",
			   "Unknown source conversion type\n");
                    return(FALSE);
                    /* break; */
                  }
              } /* switch */
            break;
          }
        case VFF_TYP_1_BYTE : /* Target is char */
          {
            switch(image->data_storage_type)
              {
                case VFF_TYP_BIT : /* Bit to char */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    cptr = (unsigned char *)kmalloc((size_t)sz);
                    cptr1 = cptr;
                    bptr = (unsigned char *)image->imagedata;
                    b = bptr;
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to look at */
                            if (*b & bit) 
                              {
                                if (sflg) *cptr1++ = scale_factor;
                                else if (nflg) *cptr1++ = norm_factor;
                                else *cptr1++ = 255;
                              }
                            else *cptr1++ = 0;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                            b++;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)cptr;
                    image->data_storage_type =  VFF_TYP_1_BYTE;
                    image->map_scheme =  VFF_MS_NONE;
                    break;
                  }
                case VFF_TYP_1_BYTE : /* Char to char */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    cptr = (unsigned char *)kmalloc((size_t)sz);
                    if (nflg)              /* Get normalization factor */
                      {
                        if (norm_factor > 255)
                          {  
                            kerror("vipl","lvconvert",
				   "Normalization target too large!\n");
                            return(FALSE);
                          }
                        cmax = 0;
                        for (i=0; i<sz; i++)
                          {
                            c = ((unsigned char *)(image->imagedata))[i];
                            if (c > cmax) cmax = c;
                          }
                        if (cmax == 0) multiplier = 0.0;
                        else multiplier = norm_factor / (double) cmax; 
                      }
                    else
                      {
                           multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0; i<sz; i++)
                      {
                            c = ((unsigned char *)(image->imagedata))[i];
                            cptr[i] = c*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)cptr;
                    image->data_storage_type =  VFF_TYP_1_BYTE;
                    break;
                  }
                case VFF_TYP_2_BYTE : /* short to char */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    cptr = (unsigned char *)kmalloc((size_t)sz);
                    smin = 0;
                    if (nflg)              /* Get normalization factor */
                      {
                        if (norm_factor > 255)
                          {  
                            kerror("vipl","lvconvert",
				   "Normalization target too large!\n");
                            return(FALSE);
                          }
                        smax = ((short *)(image->imagedata))[0];
                        smin = smax;
                        for (i=0; i<sz; i++)
                          {
                            s = ((short *)(image->imagedata))[i];
                            if (s > smax) smax = s;
                            else if (s < smin) smin = s;
                          }
                        if (smin == 0 && smax == 0) multiplier = 0.0;
                        else if (smin == smax)
                          {
                            multiplier = norm_factor/smin;
                            smin = 0;
                          }
                        else multiplier = norm_factor/(smax-smin); 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0; i<sz; i++)
                      {
                        s = ((short *)(image->imagedata))[i];
                        s = (s-smin)*multiplier;
                        if (s > 255) s = 255;
                        else if (s < 0) s = 0;
                        cptr[i] = s;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)cptr;
                    image->data_storage_type =  VFF_TYP_1_BYTE;
                    break;
                  }
                case VFF_TYP_4_BYTE : /* int to char */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    cptr = (unsigned char *)kmalloc((size_t)sz);
                    imin = 0;
                    if (nflg)              /* Get normalization factor */
                      {
                        if (norm_factor > 255)
                          {  
                            kerror("vipl","lvconvert",
				   "Normalization target too large!\n");
                            return(FALSE);
                          }
                        imax = ((int *)(image->imagedata))[0];
                        imin = imax;
                        for (i=0; i<sz; i++)
                          {
                            it = ((int *)(image->imagedata))[i];
                            if (it > imax) imax = it;
                            else if (it < imin) imin = it;
                          }
                        if (imin == 0 && imax == 0) multiplier = 0.0;
                        else if (imin == imax)
                          {
                            multiplier = norm_factor/imin;
                            imin = 0;
                          }
                        else multiplier = norm_factor/(imax-imin); 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        it = ((int *)(image->imagedata))[i];
                        it = (it-imin)*multiplier;
                        if (it > 255) it = 255;
                        else if (it < 0) it = 0;
                        cptr[i] = it;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)cptr;
                    image->data_storage_type =  VFF_TYP_1_BYTE;
                    break;
                  }
                case VFF_TYP_FLOAT : /* float to char */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    cptr = (unsigned char *)kmalloc((size_t)sz);
                    fmin = 0.0;
                    if (nflg)              /* Get normalization factor */
                      {
                        if (norm_factor > 255)
                          {  
                            kerror("vipl","lvconvert",
				   "Normalization target too large!\n");
                            return(FALSE);
                          }
                        fmax = ((float *)(image->imagedata))[0];
                        fmin = fmax;
                        for (i=0; i<sz; i++)
                          {
                            f = ((float *)(image->imagedata))[i];
                            if (f > fmax) fmax = f;
                            else if (f < fmin) fmin = f;
                          }
                        if (fmin == 0.0 && fmax == 0.0) multiplier = 0.0;
                        else if (fmin == fmax)
                          {
                            multiplier = norm_factor/fmin;
                            fmin = 0;
                          }
                        else multiplier = norm_factor/(fmax-fmin); 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        f = ((float *)(image->imagedata))[i];
                        f = (f-fmin)*multiplier;
                        if (f > 255.0) f=255.0;
                        else if (f < 0.0) f=0.0;
                        cptr[i] = f;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)cptr;
                    image->data_storage_type =  VFF_TYP_1_BYTE;
                    break;
                  }
                case VFF_TYP_COMPLEX : /* complex to char */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    cptr = (unsigned char *)kmalloc((size_t)sz);
                    fmin = 0.0;
                    if (nflg)              /* Get normalization factor */
                      {
                        if (norm_factor > 255)
                          {  
                            kerror("vipl","lvconvert",
				   "Normalization target too large!\n");
                            return(FALSE);
                          }
                        fmax = ((float *)(image->imagedata))[0];
                        fmin = fmax;
                        for (i=0; i<2*sz; i += 2)
                          {
                            f = ((float *)(image->imagedata))[i];
                            if (f > fmax) fmax = f;
                            else if (f < fmin) fmin = f;
                          }
                        if (fmin == 0.0 && fmax == 0.0) multiplier = 0.0;
                        else if (fmin == fmax)
                          {
                            multiplier = norm_factor/fmin;
                            fmin = 0;
                          }
                        else multiplier = norm_factor/(fmax-fmin); 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }

                    for (i=0,j=0; i<2*sz; i += 2)
                      {
                            f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY */
                            f = (f-fmin)*multiplier;
                            if (f > 255.0) f=255.0;
                            else if (f < 0.0) f=0.0;
                            cptr[j++] = (unsigned char)f;
                          }
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)cptr;
                    image->data_storage_type =  VFF_TYP_1_BYTE;
                    break;
                default :
                  {
                    kerror("vipl","lvconvert",
			   "Unknown source conversion type\n");
                    return(FALSE);
                    /* break; */
                  }
              } /* switch */
            break;
          } /* case */
        case VFF_TYP_2_BYTE: /* Target is short */
          {
            switch(image->data_storage_type)
              {
                case VFF_TYP_BIT : /* Bit to short */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    sptr = (short *)kmalloc(sz*sizeof(short));
                    sptr1 = sptr;
                    bptr = (unsigned char *)image->imagedata;
                    b = bptr;
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to look at */
                            if (*b & bit) 
                              {
                                if (sflg) *sptr1++ = scale_factor;
                                else if (nflg) *sptr1++ = norm_factor;
                                else *sptr1++ = 1;
                              }
                            else *sptr1++ = 0;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                            b++;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)sptr;
                    image->data_storage_type =  VFF_TYP_2_BYTE;
                    image->map_scheme =  VFF_MS_NONE;
                    break;
                  }
                case VFF_TYP_1_BYTE : /* Char to short */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    sptr = (short *)kmalloc(sz*sizeof(short));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        cmax = 0;
                        for (i=0; i<sz; i++)
                          {
                            c = ((unsigned char *)(image->imagedata))[i];
                            if (c > cmax) cmax = c;
                          }
                        if (cmax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/ (double) cmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        c = ((unsigned char *)(image->imagedata))[i];
                        sptr[i] = multiplier*c;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)sptr;
                    image->data_storage_type =  VFF_TYP_2_BYTE;
                    break;
                  }
                case VFF_TYP_2_BYTE : /* Short to short */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    sptr = (short *)kmalloc(sz*sizeof(short));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        smax = ((short *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            s = ((short *)(image->imagedata))[i];
                            if (s > smax) smax = s;
                          }
                        if (smax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/smax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        s = ((short *)(image->imagedata))[i];
                        sptr[i] = s*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)sptr;
                    image->data_storage_type =  VFF_TYP_2_BYTE;
                    break;
                  }
                case VFF_TYP_4_BYTE : /* int to short */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    sptr = (short *)kmalloc(sz*sizeof(short));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        imax = ((int *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            it = ((int *)(image->imagedata))[i];
                            if (it > imax) imax = it;
                          }
                        if (imax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/imax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        it = ((int *)(image->imagedata))[i];
                        sptr[i] = it*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)sptr;
                    image->data_storage_type =  VFF_TYP_2_BYTE;
                    break;
                  }
                case VFF_TYP_FLOAT : /* float to short */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    sptr = (short *)kmalloc(sz*sizeof(short));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            f = ((float *)(image->imagedata))[i];
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        f = ((float *)(image->imagedata))[i];
                        sptr[i] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)sptr;
                    image->data_storage_type =  VFF_TYP_2_BYTE;
                    break;
                  }
                case VFF_TYP_COMPLEX : /* complex to short */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    sptr = (short *)kmalloc(sz*sizeof(short));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz*2; i += 2)
                          {
                            f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY! */
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0,j=0; i<sz*2; i+= 2)
                      {
                        f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY */
                        sptr[j++] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)sptr;
                    image->data_storage_type =  VFF_TYP_2_BYTE;
                    break;
                  }
                default :
                  {
                    kerror("vipl","lvconvert",
			   "Unknown source conversion type\n");
                    return(FALSE);
                    /* break; */
                  }
              } /* switch */
            break;
          } /* case */
        case VFF_TYP_4_BYTE: /* Target is int */
          {
            switch(image->data_storage_type)
              {
                case VFF_TYP_BIT : /* Bit to int */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    iptr = (int *)kmalloc(sz*sizeof(int));
                    iptr1 = iptr;
                    bptr = (unsigned char *)image->imagedata;
                    b = bptr;
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to look at */
                            if (*b & bit) 
                              {
                                if (sflg) *iptr1++ = scale_factor;
                                else if (nflg) *iptr1++ = norm_factor;
                                else *iptr1++ = 1;
                              }
                            else *iptr1++ = 0;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                            b++;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)iptr;
                    image->data_storage_type =  VFF_TYP_4_BYTE;
                    image->map_scheme =  VFF_MS_NONE;
                    break;
                  }
                case VFF_TYP_1_BYTE : /* Char to int */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    iptr = (int *)kmalloc(sz*sizeof(int));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        cmax = 0;
                        for (i=0; i<sz; i++)
                          {
                            c = ((unsigned char *)(image->imagedata))[i];
                            if (c > cmax) cmax = c;
                          }
                        if (cmax == 0) multiplier = 0.0;
/* changed: else multiplier = norm_factor/ (double) cmax; to float cast 4-15-95 DK*/
                        else multiplier = norm_factor/ (float) cmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        c = ((unsigned char *)(image->imagedata))[i];
                        iptr[i] = (int) (multiplier*(float)c);
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)iptr;
                    image->data_storage_type =  VFF_TYP_4_BYTE;
                    break;
                  }
                case VFF_TYP_2_BYTE : /* Short to int */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    iptr = (int *)kmalloc(sz*sizeof(int));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        smax = ((short *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            s = ((short *)(image->imagedata))[i];
                            if (s > smax) smax = s;
                          }
                        if (smax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/smax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        s = ((short *)(image->imagedata))[i];
                        iptr[i] = s*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)iptr;
                    image->data_storage_type =  VFF_TYP_4_BYTE;
                    break;
                  }
                case VFF_TYP_4_BYTE : /* int to int */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    iptr = (int *)kmalloc(sz*sizeof(int));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        imax = ((int *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            it = ((int *)(image->imagedata))[i];
                            if (it > imax) imax = it;
                          }
                        if (imax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/imax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        it = ((int *)(image->imagedata))[i];
                        iptr[i] = it*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)iptr;
                    image->data_storage_type =  VFF_TYP_4_BYTE;
                    break;
                  }
                case VFF_TYP_FLOAT : /* float to int */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    iptr = (int *)kmalloc(sz*sizeof(int));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            f = ((float *)(image->imagedata))[i];
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        f = ((float *)(image->imagedata))[i];
                        iptr[i] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)iptr;
                    image->data_storage_type =  VFF_TYP_4_BYTE;
                    break;
                  }
                case VFF_TYP_COMPLEX : /* complex to int */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    iptr = (int *)kmalloc(sz*sizeof(int));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz*2; i += 2)
                          {
                            f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY! */
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0,j=0; i<sz*2; i += 2)
                      {
                        f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY */
                        iptr[j++] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)iptr;
                    image->data_storage_type =  VFF_TYP_4_BYTE;
                    break;
                  }
                default :
                  {
                    kerror("vipl","lvconvert",
			   "Unknown source conversion type\n");
                    return(FALSE);
                    /* break; */
                  }
              } /* switch */
            break;
          } /* case */
        case VFF_TYP_FLOAT: /* Target is float */
          {
            switch(image->data_storage_type)
              {
                case VFF_TYP_BIT : /* Bit to FLOAT */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float));
                    fptr1 = fptr;
                    bptr = (unsigned char *)image->imagedata;
                    b = bptr;
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to look at */
                            if (*b & bit) 
                              {
                                if (sflg) *fptr1++ = scale_factor;
                                else if (nflg) *fptr1++ = norm_factor;
                                else *fptr1++ = 1.0;
                              }
                            else *fptr1++ = 0.0;
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                            b++;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_FLOAT;
                    image->map_scheme =  VFF_MS_NONE;
                    break;
                  }
                case VFF_TYP_1_BYTE : /* Char to float */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        cmax = 0;
                        for (i=0; i<sz; i++)
                          {
                            c = ((unsigned char *)(image->imagedata))[i];
                            if (c > cmax) cmax = c;
                          }
                        if (cmax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/ (double) cmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        c = ((unsigned char *)(image->imagedata))[i];
                        fptr[i] = (float)c*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_FLOAT;
                    break;
                  }
                case VFF_TYP_2_BYTE : /* Short to float */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        smax = ((short *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            s = ((short *)(image->imagedata))[i];
                            if (s > smax) smax = s;
                          }
                        if (smax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/smax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        s = ((short *)(image->imagedata))[i];
                        fptr[i] = s*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_FLOAT;
                    break;
                  }
                case VFF_TYP_4_BYTE : /* int to float */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        imax = ((int *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            it = ((int *)(image->imagedata))[i];
                            if (it > imax) imax = it;
                          }
                        if (imax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/imax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        it = ((int *)(image->imagedata))[i];
                        fptr[i] = it*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_FLOAT;
                    break;
                  }
                case VFF_TYP_FLOAT : /* float to float */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            f = ((float *)(image->imagedata))[i];
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0 ;
                      }
                    for (i=0; i<sz; i++)
                      {
                        f = ((float *)(image->imagedata))[i];
                        fptr[i] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_FLOAT;
                    break;
                  }
                case VFF_TYP_COMPLEX : /* complex to float */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float));
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz*2; i += 2)
                          {
                            f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY! */
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0,j=0; i<sz*2; i += 2)
                      {
                        f = ((float *)(image->imagedata))[i]; /* REAL PART ONLY */
                        fptr[j++] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_FLOAT;
                    break;
                  }
                default :
                  {
                    kerror("vipl","lvconvert",
			   "Unknown source conversion type\n");
                    return(FALSE);
                    /* break; */
                  }
              } /* switch */
            break;
          } /* case */
        case VFF_TYP_COMPLEX: /* Target is complex */
          {
            switch(image->data_storage_type)
              {
                case VFF_TYP_BIT : /* Bit to complex */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float)*2);
                    fptr1 = fptr;
                    bptr = (unsigned char *)image->imagedata;
                    b = bptr;
                    for (i=0; i<r; i++)
                      {
                        k = 0;
                        for (j=0; j<cols; j++)
                          {
                            bit = 1 << k;      /* Bit to look at */
                            if (*b & bit) 
                              { /* NOTE: This only sets the REAL PART */
                                if (sflg) *fptr1++ = scale_factor;
                                else if (nflg) *fptr1++ = norm_factor;
                                else *fptr1++ = 1.0;
                              }
                            else *fptr1++ = 0.0;
                            fptr++;
                            *fptr++ = 0.0; /* IMAGINARY PART */
                            if (k == 7)
                              {
                                k = 0;
                                b++;
                              }
                            else k++;
                          }
                        if (k != 0)
                            b++;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_COMPLEX;
                    image->map_scheme = VFF_MS_NONE;
                    break;
                  }
                case VFF_TYP_1_BYTE : /* Char to complex */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float)*2);
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        cmax = 0;
                        for (i=0; i<sz; i++)
                          {
                            c = ((unsigned char *)(image->imagedata))[i];
                            if (c > cmax) cmax = c;
                          }
                        if (cmax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/ (double) cmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0; i<sz; i++)
                      {
                        c = ((unsigned char *)(image->imagedata))[i];
                        fptr[2*i] = (float)c*multiplier;
                        fptr[2*i+1] = 0.0;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_COMPLEX;
                    break;
                  }
                case VFF_TYP_2_BYTE : /* short to complex */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float)*2);
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        smax = ((short *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            s = ((short *)(image->imagedata))[i];
                            if (s > smax) smax = s;
                          }
                        if (smax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/smax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0; i<sz; i++)
                      {
                        s = ((short *)(image->imagedata))[i];
                        fptr[2*i] = s*multiplier;
                        fptr[2*i+1] = 0.0;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_COMPLEX;
                    break;
                  }
                case VFF_TYP_4_BYTE : /* int to complex */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float)*2);
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        imax = ((int *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            it = ((int *)(image->imagedata))[i];
                            if (it > imax) imax = it;
                          }
                        if (imax == 0) multiplier = 0.0;
                        else multiplier = norm_factor/imax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0; i<sz; i++)
                      {
                        it = ((int *)(image->imagedata))[i];
                        fptr[2*i] = it*multiplier;
                        fptr[2*i+1] = 0.0;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_COMPLEX;
                    break;
                  }
                case VFF_TYP_FLOAT : /* float to complex */
                  {
                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float)*2);
                    if (nflg != 0)              /* Get normalization factor */
                      {
                        fmax = ((float *)(image->imagedata))[0];
                        for (i=0; i<sz; i++)
                          {
                            f = ((float *)(image->imagedata))[i];
                            if (f > fmax) fmax = f;
                          }
                        if (fmax == 0.0) multiplier = 0.0;
                        else multiplier = norm_factor/fmax; 
                      }
                    else
                      {
                        multiplier = sflg ? scale_factor : 1.0;
                      }
                    for (i=0; i<sz; i++)
                      {
                        f = ((float *)(image->imagedata))[i];
                        fptr[2*i] = f*multiplier;
                        fptr[2*i+1] = 0.0;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_COMPLEX;
                    break;
                  }
                case VFF_TYP_COMPLEX : /* complex to complex */
                  {
                    if (nflg)
                      {
                        kerror("vipl","lvconvert",
			       "Cannot normalize COMPLEX.\n");
                        return(FALSE);
                      }
                    if (sflg) multiplier = scale_factor;

                    if (!lvcset_sizes(image,rows,size,&r,&sz)) return(FALSE);
                    fptr = (float *)kmalloc(sz*sizeof(float)*2);
                    for (i=0; i<sz*2; i += 2) 
                      {
                        f = ((float *)(image->imagedata))[i];
                        fptr[i] = f*multiplier;
                        f = ((float *)(image->imagedata))[i+1]; /*was 2*i+1-jw*/
                        fptr[i+1] = f*multiplier;
                      }
                    kfree_and_NULL(image->imagedata);
                    image->imagedata = (char *)fptr;
                    image->data_storage_type =  VFF_TYP_COMPLEX;
                    break;
                  }
                default :
                  {
                    kerror("vipl","lvconvert",
			   "Unknown source conversion type\n");
                    return(FALSE);
                    /* break; */
                  }
              } /* switch */
            break;
          } /* case */
        default :
          {
            kerror("vipl","lvconvert", "Unknown target data type: %d\n",k);
            return(FALSE);
            /* break; */
          }
      } /* switch */

/*
** now for a little more kludge code for double and double complex...
*/

   if(really_to_double){
      if(k==VFF_TYP_FLOAT){
         if((dptr = (double *)kmalloc(size*image->num_data_bands*
            sizeof(double)))==NULL){
            kerror("vipl","lvconvert", "failure to malloc stuff.\n");
            return(FALSE);
         }
         for(i=0;i<size*image->num_data_bands;i++){
             dptr[i] = (double)fptr[i];
         }
         kfree_and_NULL(image->imagedata);
         image->imagedata = (char *)dptr;
         image->data_storage_type =  VFF_TYP_DOUBLE;
         image->map_scheme =  VFF_MS_NONE;
      }else{
         if((dptr = (double *)kmalloc(2*size*image->num_data_bands
            *sizeof(double)))==NULL){
            kerror("vipl","lvconvert", "failure to malloc stuff.\n");
            return(FALSE);
         }
         for(i=0;i<2*size*image->num_data_bands;i++){
             dptr[i] = (double)fptr[i];
         }
         kfree_and_NULL(image->imagedata);
         image->imagedata = (char *)dptr;
         image->data_storage_type =  VFF_TYP_DCOMPLEX;
         image->map_scheme =  VFF_MS_NONE;
      }
   }

/*
** the end .... whew!
*/

    return(TRUE);
}

static int lvcset_sizes(
	xvimage *image,
	int rows,
	int size,
	int *r,
	int *s)
{
    *r = rows*image->num_data_bands;
    *s = size*image->num_data_bands;
    return(TRUE);
}
/* -library_code_end */
