 /*
  * Khoros: $Id: lvenhance.c,v 1.16.2.2 1997/05/31 20:16:24 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvenhance.c,v 1.16.2.2 1997/05/31 20:16:24 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for venhance
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvenhance
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvenhance - Enhances an image using the local standard deviation and mean.
* 
*       Purpose: venhance uses a local transformation based on the mean and
*		 variance of the intensity maps and the intensity of the
*		 input image f(x,y) to create a new image g(x,y) by
*		 performing the following transformation at each pixel
*		 location (x,y):
*
*		 !  g(x,y) = A(x,y).[f(x,y) -m(x,y)] + m(x,y),
*
*		 where
*
*		 !  A(x,y) = k . M / sigma(x,y)     0 < k < 1
*
*		 where m(x,y) and sigma(x,y) are the gray-level mean and
*		 standard deviation computed in a neighborhood centered at
*		 (x,y), M is the global mean of f(x,y), and k is a
*		 constant in the range indicated above.
*
*         Input: img    - pointer to the input image
*		 xsize  - horizontal dimension of the window being used
*		 ysize  - vertical dimension of the window being used
*		 factor - tuning factor (parameter k, see kman of venhance)
*		 minstd - minimum allowed standard deviation
*        Output: img    - pointer to the output image
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: lvenhance only operates on byte type images.
*    Written By: Marcelo Teran
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvenhance(
	xvimage *img,
	int xsize,
	int ysize,
	float factor,
	float minstd)
/* -library_def_end */

/* -library_code */
{

    int row,                            /* points image row */ 
        col,                            /* points image col */ 
        xend,                           /* step in img up to here */
        yend,                           /* step in img up to here */
        nc,                             /* column size of image */
        nr,                             /* row size of image */
        xoff, yoff,
        z, h,
        points,
        ind, adex, bdex,
        i, j, k, m;                     /* general counters */

    float *hist;                        /* holds histogram of window */

    unsigned char *ptr;                 /* pointer to the image data char */
    float      *result;                 /* pointer to resulting image char */
    float       value, sum, avg, rms;
    float       MEAN, gain, var;
    float       std;

    char       *program;                /* contains the library name */

    program = "lvenhance";

   /* check for odd windows */
    if((xsize & 1) == 0 || (ysize & 1) == 0){
        (void) kerror("vipl","lvenhance","Error window must be odd");
        return(FALSE);
    }

    propertype(program,img,VFF_TYP_1_BYTE,1);      /* check img. type */        

    nr = img->col_size;
    nc = img->row_size;
    points = nc * nr;

    xend = nc - xsize;
    yend = nr - ysize + 1;
    m = 0; h = 0;

    xoff = xsize / 2;
    yoff = ysize / 2;

   /* Allocate space */ 
   hist = (float *) kmalloc((unsigned int) xsize * ysize* sizeof(float));
    if(hist == NULL) {
        (void) kerror("vipl","lvhmed", "Insufficient memory available");
        return(FALSE);
    }

   /* Allocate space for resulting image */ 
    result = (float *)kmalloc((unsigned int)nc*nr*sizeof(float ));
    if(result == NULL) {
        (void) kerror("vipl","lvhmed", "Insufficient memory available");
        return(FALSE);
    }
   /* Assign image data address to ptr */
    ptr = (unsigned char *)img->imagedata;

    sum = 0; 
   /* compute the overall mean */
    for(i = 0; i < points; i++){
        sum += ptr[i];
    }

    MEAN = sum / (float)points;

   /* perform enhancement to the image */
    for (row = 0 ; row < yend; row++){

       /* setup the first window for the current row */
        for(i = 0; i < ysize; i++){
            for(j = 0; j < xsize; j++){
                hist[i*xsize + j] = (float)ptr[nc*(row + i) + j];
            }
        }

        sum = 0; rms = 0;
       /* get the standard deviation value */
        for(ind = 0; ind < xsize * ysize; ind++){
            sum += hist[ind];
            rms += hist[ind] * hist[ind];
        }

        avg = sum /(xsize * ysize);
        var = rms /(xsize * ysize) - avg * avg;
        std = sqrt((double)var);
        if(std < minstd){
            gain = factor * MEAN;
        } else {
            gain = factor * MEAN / std;
        }
        value = gain *((float)ptr[nc*(row + yoff) + xoff + m++ ] - avg ) + avg;

        result[nc * (row + yoff) + xoff + h++] = value;

       /* step through the current row */
        for (col = 0 ; col < xend ; col++){

            for(k = 0; k < ysize; k++){
                for(z = 0; z < xsize; z++){
                   /* delete the lefmost column of the previous window*/
                    adex = z + k*xsize;
                    bdex = z + k*nc + col + row*nc;
                    hist[adex] = (float)ptr[bdex + 1];
                }
            }

            sum = 0; rms = 0;
            /* get the variance value */
            for(ind = 0; ind < xsize * ysize; ind++){
                sum += hist[ind];
                rms += hist[ind] * hist[ind];
            }
            avg = sum /(xsize * ysize);
            var = rms /(xsize * ysize) - avg * avg;
            std = sqrt((double)var);
            if(std < minstd){
                gain = factor * MEAN;
            } else {
                gain = factor * MEAN / std;
            }
            value = gain *((float)ptr[nc*(row + yoff) + xoff + m++ ] - avg ) + avg;

            result[nc * (row + yoff) + xoff + h++] = value;

        } m = 0; h = 0;


    }

    (void)kfree_and_NULL(img->imagedata);

    img->data_storage_type = VFF_TYP_FLOAT;
    img->imagedata = (char *) result;

    if( !lvconvert(img,1,0,1,255.0,1.0,0)){
        (void) kerror("vipl","lvconvert", "lvconvert Failed");
        return(FALSE);
    }

    return(TRUE);
}
/* -library_code_end */
