 /*
  * Khoros: $Id: lvgpwl.c,v 1.15.2.2 1997/05/31 20:16:43 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvgpwl.c,v 1.15.2.2 1997/05/31 20:16:43 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vgpwl
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvgpwl
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvgpwl - Create a two dimensional piecewise linear image
* 
*       Purpose: creates a two dimensional piecewise linear image.  The image
*		 pixel values follow a square waveform. The user can specify
*		 the period, phase, minima and maxima, sampling frequency, rise
*		 and fall times, and image dimensions.
*
*		 NOTE: According to the Sampling Theorem, the sampling frequency
*		 should be at least twice that of the signal frequency to
*		 generate a signal properly.
*
*         Input: image - resulting image.
*		 rows - number of rows in image
*		 cols - number of columns in image
*		 fs - sampling frequency
*		 minimum - minimum value of function
*		 maximum - maximum value of function
*		 xperiod - X period of the function in seconds
*		 yperiod - Y period of the function in seconds
*		 xrise - X rise time in seconds
*		 yrise - Y rise time in seconds
*		 xfall - X fall time in seconds
*		 yfall - Y fall time in seconds
*		 xwidth - X width of pulse
*		 ywidth - Y width of pulse
*		 type - data type of output image.  This must be either
*		        VFF_TYP_FLOAT or VFF_TYP_COMPLEX.
*        Output: image - holds the result of the data generation.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: can only generate VFF_TYP_FLOAT and VFF_TYP_COMPLEX type images
*    Written By: Jeremy Worley
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Jeremy Worley 05 Mar 1992 08:26 MST -- Fixed up an fprintf
*		 function that had too few arguments based on the format
*		 field.  Added rows,cols, and bands to the list of explicitly
*		 declared arguments to lvgpwl().
****************************************************************/
/* -library_def */
int lvgpwl(
    xvimage **image, 
    int rows, 
    int cols, 
    int bands, 
    float fs, 
    float minimum, 
    float maximum, 
    float xperiod, 
    float yperiod, 
    float xrise, 
    float yrise, 
    float xfall, 
    float yfall, 
    float xwidth, 
    float ywidth, 
    int type)
/* -library_def_end */

/* -library_code */
{
  char  *comment="signal ensemble courtesy of lvgpwl.";
  int   j,b;
  float *rpart,*ipart=NULL, *data;
/*
** Immediately check bounds on arguments. 
*/

  if(rows<1 || cols<1 || bands<1){
     kerror("vipl","lvgpwl","Image dimension out of bounds.\n");
     return(FALSE);
  }else if(fs<=0.0){
     kerror("vipl","lvgpwl", "Sampling frequency out of bounds.\n");
     return(FALSE);
  }else if(xperiod<=0.0 || yperiod<=0.0){
     kerror("vipl","lvgpwl", "Period out of bounds.\n");
     return(FALSE);
  }else if(type!=VFF_TYP_FLOAT && type!=VFF_TYP_COMPLEX){
     kerror("vipl","lvgpwl", "Illegal image data type specified.\n");
     return(FALSE);
  }else if(xrise < 0.0 || xfall < 0.0 || yrise < 0.0 || yfall < 0.0){
     kerror("vipl","lvgpwl", "Transition time must be non negative.\n");
     return(FALSE);
  }

/*
** Generate image
*/
  
  *image = createimage((unsigned long) rows,
		       (unsigned long) cols,
		       (unsigned long) type,
		       (unsigned long) 1,
		       (unsigned long) bands,comment,
                       (unsigned long)0, /* map_row_size */
                       (unsigned long)0, /* map_col_size */
                       (unsigned long)VFF_MS_NONE,
                       (unsigned long)VFF_MAPTYP_NONE,
                       (unsigned long)VFF_LOC_IMPLICIT,
                       (unsigned long)0);/* location dimension */

  if(*image==NULL){
     kerror("vipl","lvgpwl","Unable to allocate memory for image.\n");
     return(FALSE);
  }

  data = (float *)(*image)->imagedata;

/*
** malloc appropriate amount of memory for real part of signal.
** if there are any errors along the way, free up whatever memory was 
** allocated before exiting.
*/

  rpart = (float *)kmalloc(rows*cols*sizeof(float));
  if(rpart==NULL){
     kerror("vipl","lvgpwl","Unable to malloc temporary workspace.\n");
     return(FALSE);
  }

/*
** if this is a complex image, malloc appropriate amount of 
** memory for the imaginary part.  if there are any errors along the way, 
** free up whatever memory was allocated before exiting.
*/
 
  if(type==VFF_TYP_COMPLEX){
     ipart = (float *)kmalloc(rows*cols*sizeof(float));
     if(ipart==NULL){
        kerror("vipl","lvgpwl","Unable to malloc temporary workspace.\n");
        kfree_and_NULL(rpart);
        return(FALSE);
     }
  }

/*
** loop through all bands
*/

  for(b=0;b<bands;b++){

/*
** call low level sine wave generator for real part.
*/

     j=gpwl(rpart,rows,cols,(float) fs,(float) minimum,(float) maximum,(float) yperiod,(float) xperiod, (float) yrise,(float) xrise,(float) yfall,(float) xfall,(float) ywidth,(float) xwidth);
     if(j==0){
        kerror("vipl","lvgpwl","Error during signal generation\n");
        kfree_and_NULL(rpart);
        if(type==VFF_TYP_COMPLEX)kfree_and_NULL(ipart);
        return(FALSE);
     }

/*
** call low level sine wave generator for imaginary part.
*/

     if(type==VFF_TYP_COMPLEX){
        j=gpwl(ipart,rows,cols,(float) fs,(float) minimum,(float) maximum,(float) yperiod,(float) xperiod,
               (float) yrise,(float) xrise,(float) yfall,(float) xfall,(float) ywidth,(float) xwidth);
        if(j==0){
           kerror("vipl","lvgpwl","Error during signal generation\n");
           kfree_and_NULL(rpart);
           kfree_and_NULL(ipart);
           return(FALSE);
        }
     }

/*
** interlace real and imaginary parts if necessary
** reorganize data and place it in ensemble
*/

     if(type==VFF_TYP_COMPLEX){
        for(j=0;j<rows*cols;j++){
            *data++ = rpart[j];
            *data++ = ipart[j];
        }
     }else{
        kmemcpy((kaddr) data, (kaddr) rpart,(unsigned) rows*cols*sizeof(float));
        data += rows*cols;
     }
  }

/*
** clean up mess
*/ 

  kfree_and_NULL(rpart);
  if(type==VFF_TYP_COMPLEX)kfree_and_NULL(ipart);

/*
** normal return
*/

  return(TRUE);

}
/* -library_code_end */
