 /*
  * Khoros: $Id: lvhmed.c,v 1.16.2.2 1997/05/31 20:16:49 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvhmed.c,v 1.16.2.2 1997/05/31 20:16:49 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vhmed
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvhmed
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvhmed - Perform median filtering on an image using a histogram to find the median.
* 
*       Purpose: lvhmed computes a two dimensional median filter of width w
*		 and height h over the given image, img.  The median is
*		 computed by moving a w by h pixel mask over the image
*		 and updating the masked histogram.
*
*		 In rectangular odd windows (masks) the median value is
*		 replaced in the center pixel location of the window. The
*		 default window format is (3x3).
*
*		 The resultant image will have a border of pixels of
*		 value 0 since the masking operation works in much the same
*		 way as a convolution.
*
*         Input: img   - input image
*		 xsize - window width
*		 ysize - window height
*
*        Output: img - holds the result of the median filtered image
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: lvhmed does not support even window dimensions and only works
*		 on BYTE images.
*    Written By: Marcelo Teran, Ramiro Jordan
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvhmed(
	xvimage *img,
	int      xsize,
	int      ysize)
/* -library_def_end */

/* -library_code */
{

    int row,                            /* points image row */ 
        col,                            /* points image col */ 
        th,                             /* gives the mediun of the wind. */
        tmdn,                           /* numb of pixels <= median */ 
        g1,                             /* auxiliar variable */
        xoff,                           /* horiz. offset in the img */
        yoff,                           /* vertic. offset in the img */
        xend,                           /* step in img up to here */
        yend,                           /* step in img up to here */
        mdn,                            /* this is the median value */
        nc,                             /* column size of image */
        nr,                             /* row size of image */
        i, j, k, m;                     /* general counters */

    unsigned int hist[256];             /* holds histogram of window */

    unsigned char *ptr,                 /* pointer to the image data char */
                  *result;              /* pointer to resulting image char */

    char  *program;                     /* contains the library name */

    program = "lvhmed";

   /* check for odd windows */
    if((xsize & 1) == 0 || (ysize & 1) == 0){
        (void) kerror("vipl", "lvhmed","Error window must be odd\n");
        return(FALSE);
    }

    propertype(program,img,(unsigned long) VFF_TYP_1_BYTE,1);      /* check img. type */        

    th = xsize * ysize / 2;

    nr = img->col_size;
    nc = img->row_size;

    xoff = xsize/2;
    yoff = ysize/2;

    xend = nc - xsize;
    yend = nr - ysize + 1;
    m = 0; tmdn = 0; mdn = 0;

   /* Zero histogram */
    kmemset((kaddr) hist, 0, 256 * sizeof(unsigned int));

   /* Allocate space for resulting image */ 
    result = (unsigned char *)kmalloc((unsigned int)nc*nr*sizeof(unsigned char));
    if(result == NULL) {
        (void) kerror("vipl","lvhmed", "Insufficient memory available\n");
        return(FALSE);
    }
   /* Assign image data address to ptr */
    ptr = (unsigned char *)img->imagedata;

   /* Zero the allocated array */
    kmemset((kaddr) result, 0, nc*nr * sizeof(unsigned char));

   /* perform median filtering of an image */
    for (row = 0 ; row < yend; row++){

       /* setup the histogram for the current row */
        for(i = 0; i < ysize; i++){
            for(j = 0; j < xsize; j++){
                g1 = ptr[nc*(row + i) + j];
                hist[g1]++;
            }
        }

       /* get the median value */
        if(tmdn > th){
            do {
                mdn = mdn - 1;
                tmdn = tmdn - hist[mdn];
            } while(tmdn > th);

        } else {
            while(tmdn + hist[mdn] <= th){
                tmdn = tmdn + hist[mdn];
                mdn = mdn + 1;
            }
        }
        result[nc *(row + yoff) + xoff + m++] = mdn;

       /* step through the current row */
        for (col = 0 ; col < xend ; col++){

            for(k = 0; k < ysize; k++){

               /* delete the lefmost column of the previous window*/
                g1 =  ptr[col + nc*(row + k)];
                hist[g1] = hist[g1] -1;

                if(g1 < mdn) tmdn = tmdn -1;

               /* add the rightmost column of the current window*/
                g1 = ptr[xsize+ nc*(row + k) +col];
                hist[g1] = hist[g1] + 1;

                if(g1 < mdn) tmdn = tmdn + 1;

            }

       /* get the median value */
            if(tmdn > th){
                do {
                    mdn = mdn - 1;
                    tmdn = tmdn - hist[mdn];
                } while(tmdn > th);
            } else {
                while(tmdn + hist[mdn] <= th){
                    tmdn = tmdn + hist[mdn];
                    mdn = mdn + 1;
                }
            }
            result[nc *(row + yoff) + xoff + m++] = mdn;
        }
       /* Reset histogram and variables */
        kmemset((kaddr) hist, 0, 256 * sizeof(unsigned int));
        m = 0; tmdn = 0; mdn = 0;
    }

    kfree_and_NULL (img->imagedata);
    img->imagedata = (char *) result;

    return(TRUE);
}
/* -library_code_end */
