 /*
  * Khoros: $Id: lvmediaxis.c,v 1.15.2.2 1997/05/31 20:17:00 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvmediaxis.c,v 1.15.2.2 1997/05/31 20:17:00 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vmediaxis
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvmediaxis
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvmediaxis - Computes the medial axis transform
* 
*       Purpose: vmediaxis computes the medial axis transform on
*		 a byte binary image.
*
*		 The median axis of a region in a binary image is
*		 the set of pixels whose minimum distance from
*		 the edge is a local maximum. It is a thinning
*		 algorithm that is comparable to the morphological
*		 skeletonization.  The median axis is generally
*		 not connected.
*
*		 An optional flag allows the user to obtain the
*		 medial axis pixel values set either to their
*		 distance from the border, or to 255. If the flag
*		 is set to one, it leads to a better visualization
*		 of the median axis but the information that
*		 permits the user to reconstruct the original
*		 binary image is lost.
*
*		 If the flag is not set to 1, it is possible to
*		 reconstruct the original image from the medial
*		 axis.
*
*		 Presentation of the MEDIAL AXIS TRANSFORM algorithm :
*
* !              U(m,n) : original image. This image must be a two-valued
* !                       byte image, 0 for the background, and any other
* !                       value for the objects.
* !
* !              Uk(i,j) : Image after k iterations.
* !
* !              U0(m,n)= U(m,n),  and for k= 1, 2, 3, ...
* !
* !              Uk(m,n)=U(m,n)+min{U0(i,j);((i,j):dist(m,n;i,j)<= 1)},
*
*		 where:
*
* !              dist(m,n;i,j) is the mediaxis between (m,n) and (i,j).
* !
* !              dist(i,j-1;i,j) = dist(i,j+1;i,j) =
* !              dist(i-1,j;i,j) = dist(i+1,j;i,j) = 1.
* !
* !              dist(i-1,j-1;i,j) = dist(i+1,j-1;i,j) = dist(i-1,j+1;i,j) = 
* !              dist(i+1,j+1;i,j) = 2.
*
*		 The transform is completed when k equals the
*		 maximum thickness of the region. That is when Uk+1
*		 equals Uk.
*
*		 The last part consists of using this image and
*		 of keeping only the pixels that are locally
*		 maximum in the 4 neighborhod.
*
*		 The MEDIAL AXIS IMAGE is a form of SKELETONIZATION. 
*
*		 SET OF FUNCTIONS that can be useful when using
*		 vmediaxis:
*
*		 vthresh: generates a binary image
*
*		 For this routine, the image must be a byte
*		 binary image, with the gray level value 0
*		 for the background, so it is advisable to
*		 use vthresh on the image before using vmediaxis.
*
*         Input: image - an xvimage structure.
*		 wflag - logical integer.
*        Output: image - holds the result median axis image.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: vmediaxis works only on Byte binary images where
*		 the background is set to zero.
*    Written By: Pascal ADAM
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: Ghosted by Pascal ADAM - Mon Jun  4 09:22:26 MDT 1990
*		 Modification of description in lvmediaxis(3) and vmediaxis(1)
****************************************************************/
/* -library_def */
int lvmediaxis(
	xvimage *image,
	int wflag)
/* -library_def_end */

/* -library_code */
{
    unsigned char *im0;
    unsigned char *cptr, *cptr0;

    int i, j, nc, nr, val;

    int max;
 
    /* Check type,... of image */
    if (!(propertype("lvmediaxis", image, (unsigned long)VFF_TYP_1_BYTE, FALSE))) {
        kerror ("vipl", "lvmediaxis", "Image must be of type byte\n");
        return (FALSE);
    }
    if (!(proper_num_images ("lvmediaxis", image, (unsigned long) 1, FALSE))) {
        kerror ("vipl", "lvmediaxis",
		"Can only work on files with one image\n\n");
        return (FALSE);
    }
    if (!(proper_num_bands ("lvmediaxis", image, (unsigned long) 1, FALSE))) {
        kerror ("vipl", "lvmediaxis",
		"Can only work on images with 1 data band\n\n");
        return (FALSE);
    }

    nr = image->col_size;
    nc = image->row_size;


    /* dynamic memory allocation for temporary im0 stockage */
    im0 = (unsigned char *) kmalloc( nc * nr *sizeof(unsigned char));
    if (im0 == NULL) {
        (void) kerror ("vipl", "lvmediaxis", "Insufficient Space avaliable\n");
        return(FALSE);
    }

 
    if(! lvdistance(image))
    {
        return(FALSE);
    }

    cptr   = (unsigned char *) image->imagedata;
    cptr0  = im0;
 
    for (i=1;i<nr-1;i++)
    {
        for (j=1;j<nc-1;j++)
        { 
            val = *(cptr+i*nc+j);
            if (val == 0) *(cptr+i*nc+j) = 0;
            else
            {
               max = *(cptr+(i-1)*nc+j);
               if ((unsigned char) max < *(cptr+(i+1)*nc+j)) max = *(cptr+(i+1)*nc+j);
               if ((unsigned char) max < *(cptr+i*nc+j+1))   max = *(cptr+i*nc+j+1);
               if ((unsigned char) max < *(cptr+i*nc+j-1))   max = *(cptr+i*nc+j-1);
            
               if  (max <= val)  
                   *(cptr0+i*nc+j)=(wflag==1)? 255: *(cptr+i*nc+j);

               else  *(cptr0+i*nc+j) = 0; 
            }

        }
    }
    
    for (i=0;i<nr;i++) *(cptr0+i*nc) = *(cptr0+i*nc+nc-1)   = 0;
    for (j=0;j<nc;j++) *(cptr0+j)    = *(cptr0+(nr-1)*nc+j) = 0;
        
    cptr   = (unsigned char *) image->imagedata;
    cptr0  = im0;

    for (i=0; i< nc*nr ;i++)  *cptr++ = *cptr0++;

    kfree_and_NULL(im0);

    return(TRUE);
        
}
/* -library_code_end */
