 /*
  * Khoros: $Id: lvmos2band.c,v 1.15.2.2 1997/05/31 20:17:05 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvmos2band.c,v 1.15.2.2 1997/05/31 20:17:05 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vmos2band
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvmos2band
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvmos2band - Convert a mosaic input image to a multiband output image
* 
*       Purpose: lvmos2band converts a mosaic input image to an output image in
*		 which each subimage of the input mosaic constitutes a band of
*		 the output image.
*
*		 The output image generated by lvmos2band, img1, will be
*		 (r-2*bs) rows by (c-2*bs) columns with n bands, where r and c
*		 are the number of rows and columns per subimage, bs is the
*		 subimage border size, and n is the number of subimages
*		 that are to be put into band format.  Lvmos2band assumes that
*		 all subimages are same size (r*c) and all subimages have the
*		 same border size (bs).  Subimages from the input mosaic image
*		 (image) are assigned to bands in the multiband output image
*		 (img1) in the following manner:
*
*		 The first subimage specified is assigned to the first band in
*		 img1.  Subsequent bands are assigned subsequent subimages
*		 indexed from left to right, top to bottom.  Lvmos2band accepts
*		 VIFF data_types (excluding VFF_TYP_BIT) and returns that same
*		 data_type.  If a map is associated with the input mosaic
*		 image, the same map will be assigned to the output multiband
*		 image.
*
*		 ! Number of columns and number of rows in subimages:
*
*		 The \fIsub_rows\fP and \fIsub_cols\fP parameters are used to
*		 specify the number of rows and the number of columns,
*		 respecively, per subimage.  Lvmos2band assumes that all
*		 subimages in the mosaic image are the same size.  The sub_rows
*		 and sub_cols values SHOULD include any borders associated with
*		 the subimages.
*
*		 ! Subimage border size:
*
*		 The sub_border parameter specifies the subimage border size
*		 in pixels.  Again, this parameter will be applied to all
*		 subimages in the mosaic.  When lvmos2band converts the mosaic
*		 image to a multiband image, the subimage borders will be
*		 removed from the subimages before storing them in bands.  The
*		 size of the output image, img1, will therefore be
*		 (sub_rows-2*sub_border) rows by (sub_cols-2*sub_border)
*		 columns by n bands.
*
*		 ! Overall mosaic image border size
*
*		 The img_border parameter specifies, in pixels, the size of an
*		 outer border that should be removed from the mosaic image
*		 before extracting the subimages.  The overall mosaic image
*		 border and the subimage borders are independent.  
*
*		 ! Horizontal and vertical offset of first subimage to be banded
*
*		 The the first subimage in the mosaic that will be put into
*		 the multiband output image is specified by sub_x_off
*		 (horizontal offset) and sub_y_off (vertical offset).  The
*		 specified subimage will then be the first band in the output
*		 image.  The indices, x and y, index through the subimages
*		 (they are NOT indices through the input mosaic image rows and
*		 columns).  Therefore, if a mosaic image contained 9 4-row by
*		 4-col subimages arranged as 3 subimages in the horizontal
*		 direction by 3 subimages in the vertical direction, the center
*		 subimage would be indexed by sub_x_offset=1, sub_y_offset=1 
*		 (not by sub_x_offset=4, sub_y_offset=4).
*
*		 ! Number of subimages to be put into band format
*
*		 Two parameters are passed to lvmos2band to indicate the number
*		 of subimages that should be put into band format, starting
*		 with the first subimage described by sub_x_offset and
*		 sub_y_offset.  The num_flag is set if the user specifies the
*		 number of subimages option.  In the case that num_flag is TRUE,
*		 the value passed in by num_subs is used.  If the specified
*		 number of subimages cannot be extracted, lvmos2band will return
*		 an error and the output image will not be created.  If
*		 num_flag is FALSE, lvmos2band calculates the number of
*		 subimages that will be banded, taking into account both the
*		 subimage skip value and the subimage offsets.
*
*		 ! Subimage skip value
*
*		 The subimage skip value parameter (sub_skip) specifies 
*		 the number of subimages to be skipped each time a subimage is
*		 stored as a band.  For example, a skip value of sub_skip=1 in
*		 a 3x3 subimage mosaic would band the subimages in a
*		 checkerboard pattern (every other subimage is a band in the
*		 output image).  When applying the subimage skip value,
*		 lvmos2band looks at the series of subimages as a linear array.
*		 Therefore, the row/column arrangement of subimages does not
*		 affect how the skipping occurs.
*         Input: image - pointer to input xvimage structure
*		 sub_rows - num rows per subimage (including borders) 
*		 sub_cols - num cols per subimage (including borders) 
*		 sub_border - subimage border (inner border) size in pixels
*		 sub_x_off - col index through subimages (1 img per index)
*		 sub_y_off - row index through subimages (1 img per index)
*		 sub_skip - skip value between subimages to be banded 
*		 num_flag - if set, user specifies num subimages banded
*		 num_subs - number of subimages that will be banded
*        Output: img1 - output banded image
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: Vmos2band does not work on explicit location data and will
*		 return an error when such an image is encountered.  No output
*		 image will be created.
*    Written By: Donna Koechner
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvmos2band(
	xvimage *image,
	xvimage **img1,
	int sub_rows,
	int sub_cols,
	int sub_border,
	int img_border,
	int sub_x_off,
	int sub_y_off,
	int sub_skip,
	int num_flag,
	int num_subs)
/* -library_def_end */

/* -library_code */
{

  int           sub_col_cnt,    /* num subimages in horizontal direction */
                sub_row_cnt,    /* num of subimages in vertical direction */
                total_num_subs, /* total number subimages in mosaic */
                first_sub_loc,  /* 1st subimage position - from -x, -y */
                num_subs_left,  /* total - first subimage */
                num_subs_avail, /* number of subimage available after x,y,s */
                skip,           /* sub_skip + 1  -- used as an index */
                skew,           /* skew at beginning of rows due to skipping */
                band_cnt,
                *i1, *i2,
                i,j,m,n;

  short         *s1, *s2;
  unsigned char *c1, *c2;
  float         *f1, *f2;

  xvimage *img2;

  unsigned long height,         /* subimage height (minus subimage border) */
                width,          /* subimage width (minus subimage border) */
                x_off,          /* column offset into mosaic image */
                y_off;          /* row offset into mosaic image */
/*
  * Verify that given sub border sizes are within bounds
  */
  if (2*sub_border >= sub_cols || 2*sub_border >= sub_rows)
  {
    kerror("vipl","lvmos2band", "sub image border size too large:\n\
sub image border size = %d\n",sub_border);
    return(FALSE);
  }


/*
  * Verify that given border sizes and subimage sizes agree with input 
  * mosaic image size.
  */
  if (((image->row_size - 2*img_border)%sub_cols) != 0 )
  {
    kerror("vipl","lvmos2band", "Incorrect horizontal dimensions:\n\
number of cols in image = %d\ncols per subimage = %d\n\
image border size = %d\n\n",image->row_size,sub_cols,img_border);
   return(FALSE);
  }
  sub_col_cnt = (image->row_size - 2*img_border)/sub_cols;

  if (((image->col_size - 2*img_border)%sub_rows) != 0 )
  {
    kerror("vipl","lvmos2band", "Incorrect vertical dimensions:\n\
number of rows in image = %d\n\rows per subimage = %d\n\
image border size = %d\n\n",image->col_size,sub_rows,img_border);
    return(FALSE);
  }
  sub_row_cnt = (image->col_size - 2*img_border)/sub_rows;


/*
  * Verify that horizontal and vertical subimage offset values are valid.
  */
  if ( (sub_x_off < 0) || (sub_x_off >= sub_col_cnt) )
  {
    kerror("vipl","lvmos2band", "Invalid horizontal subimage offset:\n\
offset = %d, number subimages in horizontal direction = %d\n\
Max offset is number subimages - 1\n\n", sub_x_off, sub_col_cnt);
    return(FALSE);
  }
 
  if ( (sub_y_off < 0) || (sub_y_off >= sub_row_cnt) )
  {
    kerror("vipl","lvmos2band", "Invalid vertical subimage offset:\n\
offset = %d, number subimages in vertical direction = %d\n\
Max offset is number subimages - 1\n\n", sub_y_off, sub_row_cnt);
    return(FALSE);
  }


/*
  * Calculate the number of subimages available to be put into bands.
  * If number of subimages flag is set, get the value passed in.  
  */
  total_num_subs = sub_row_cnt * sub_col_cnt;
  first_sub_loc = (sub_col_cnt * sub_y_off) + sub_x_off;
  num_subs_left = total_num_subs - first_sub_loc;
  num_subs_avail = ( (num_subs_left - 1)/(sub_skip +1) ) + 1;
  if (!num_flag)
  {
    num_subs = num_subs_avail;
  }
  

 /* 
  * If more subimages were specified to be banded than are available, 
  * return.
  */
  if (num_subs > num_subs_avail)
  {
    kerror("vipl", "lvmos2band",
	   "Number subimages to band option too large.\n");
    return(FALSE);
  }

  if ((width = sub_cols - 2*sub_border) < 1)
  {
    kerror("vipl", "lvmos2band",
	   "subimage borders larger than subimage cols\n");
    return(FALSE);
  }

  if ((height = sub_rows - 2*sub_border) < 1)
  {
    kerror("vipl", "lvmos2band",
	   "subimage borders larger than subimage rows\n");
    return(FALSE);
  }

/*
  * Create the resulting image
  */
  img2 = createimage(height,                    /* number rows per subimage */
                     width,                     /* number cols per subimage */
                     image->data_storage_type,
                     (unsigned long) 1,         /* number of images */
                     (unsigned long) num_subs,  /* number of data bands */
                     "created by vmos2band",    /* comment */
                     image->map_row_size,
                     image->map_col_size,
                     image->map_scheme,
                     image->map_storage_type,
                     (unsigned long) VFF_LOC_IMPLICIT,
                     (unsigned long) 0);        /* location dimension */

                                
  if (img2 == NULL)
  {
    kerror("vipl","lvmos2vect", "createimage failed\n");
    return(FALSE);
  }

  if (img2->imagedata == NULL)
  {
    kerror("vipl","lvmos2band", "Pointer to img2->imagedata invalid.\n");
    return(FALSE);
  }

 /* 
  * Change necessary fields in output image.
  * Startx and starty, which specify location of subimage in original
  * image, is set to the first subimage extracted.
  */
  img2->startx = img_border + (sub_x_off * sub_cols) + sub_border;
  img2->starty = img_border + (sub_y_off * sub_rows) + sub_border;
  img2->pixsizx = image->pixsizx;
  img2->pixsizy = image->pixsizy;
  img2->data_encode_scheme = image->data_encode_scheme;
  img2->map_enable = image->map_enable;
  img2->color_space_model = image->color_space_model;


 /* 
  * Copy map data from image to img2 
  */
  if ( ! copymapdata(image, &img2, FALSE) ) 
  {
    kerror("vipl","lvmos2band", "Could not copy map data!\n");
    return(FALSE);
  }


/*
  * Initialize common loop variable.
  */
  skew = 0;
  skip = sub_skip + 1;
  band_cnt = 0;
  

/*
  * Copy specified subimages from input image to bands in output image.
  */
  switch (image->data_storage_type)
  {
    case VFF_TYP_BIT:
         kerror("vipl","lvmos2band", "No BIT images allowed!\n");
         return(FALSE);

    case VFF_TYP_1_BYTE:
         
         c1 = (unsigned char *)(image->imagedata);
         c2 = (unsigned char *)(img2->imagedata);

/*
         * First extract subimages in the partial(?) row defined by 
         * sub_x_off and sub_y_off
         */
         for ( j=sub_x_off ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip )
         {
           x_off = img_border + ( j * sub_cols ) + sub_border;
           y_off = img_border + ( sub_y_off * sub_rows ) + sub_border;
  
           for (m=0; m<height; m++)
           {
             for (n=0; n<width; n++)
             {
               c2[(band_cnt * height * width ) + m * width + n] 
                        = c1[(y_off+m)*image->row_size+(x_off+n)];
             }
           }
  
           band_cnt ++;
    
           if ( (sub_col_cnt - j) <= skip )
           {
             skew = skip - (sub_col_cnt - j);
           }
         } 
  
        /* 
         * Extract the rest of the (complete) rows.
         */
  
         for ( i=(sub_y_off+1) ; (i<sub_row_cnt) && (band_cnt<num_subs) ; i++ )
         {
           for ( j=skew ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip ) 
           {
             x_off = img_border + ( j * sub_cols ) + sub_border;
             y_off = img_border + ( i * sub_rows ) + sub_border;
  
             for (m=0; m<height; m++)
             {
               for (n=0; n<width; n++)
               {
                 c2[(band_cnt * height * width ) + m * width + n] 
                          = c1[(y_off+m)*image->row_size+(x_off+n)];
               }
             }
  
             band_cnt ++;
      
             if ( (sub_col_cnt - j) <= skip )
             {
               skew = skip - (sub_col_cnt - j);
             }
           }  /* end of loop through subimage columns */
         }  /* end of loop through subimage rows */
         break; /* end case VFF_TYP_1_BYTE */


    case VFF_TYP_2_BYTE:
         
         s1 = (short *)(image->imagedata);
         s2 = (short *)(img2->imagedata);

/*
         * First extract subimages in the partial(?) row defined by 
         * sub_x_off and sub_y_off
         */
         for ( j=sub_x_off ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip )
         {
           x_off = img_border + ( j * sub_cols ) + sub_border;
           y_off = img_border + ( sub_y_off * sub_rows ) + sub_border;
  
           for (m=0; m<height; m++)
           {
             for (n=0; n<width; n++)
             {
               s2[(band_cnt * height * width ) + m * width + n] 
                        = s1[(y_off+m)*image->row_size+(x_off+n)];
             }
           }
  
           band_cnt ++;
    
           if ( (sub_col_cnt - j) <= skip )
           {
             skew = skip - (sub_col_cnt - j);
           }
         } 
  
        /* 
         * Extract the rest of the (complete) rows.
         */
  
         for ( i=(sub_y_off+1) ; (i<sub_row_cnt) && (band_cnt<num_subs) ; i++ )
         {
           for ( j=skew ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip ) 
           {
             x_off = img_border + ( j * sub_cols ) + sub_border;
             y_off = img_border + ( i * sub_rows ) + sub_border;
  
             for (m=0; m<height; m++)
             {
               for (n=0; n<width; n++)
               {
                 s2[(band_cnt * height * width ) + m * width + n] 
                          = s1[(y_off+m)*image->row_size+(x_off+n)];
               }
             }
  
             band_cnt ++;
      
             if ( (sub_col_cnt - j) <= skip )
             {
               skew = skip - (sub_col_cnt - j);
             }
           }  /* end of loop through subimage columns */
         }  /* end of loop through subimage rows */
         break; /* end case VFF_TYP_2_BYTE */



    case VFF_TYP_4_BYTE:

         i1 = (int *)(image->imagedata);
         i2 = (int *)(img2->imagedata);

/*
         * First extract subimages in the partial(?) row defined by 
         * sub_x_off and sub_y_off
         */
         for ( j=sub_x_off ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip )
         {
           x_off = img_border + ( j * sub_cols ) + sub_border;
           y_off = img_border + ( sub_y_off * sub_rows ) + sub_border;
  
           for (m=0; m<height; m++)
           {
             for (n=0; n<width; n++)
             {
               i2[(band_cnt * height * width ) + m * width + n] 
                        = i1[(y_off+m)*image->row_size+(x_off+n)];
             }
           }
  
           band_cnt ++;
    
           if ( (sub_col_cnt - j) <= skip )
           {
             skew = skip - (sub_col_cnt - j);
           }
         } 
  
        /* 
         * Extract the rest of the (complete) rows.
         */
  
         for ( i=(sub_y_off+1) ; (i<sub_row_cnt) && (band_cnt<num_subs) ; i++ )
         {
           for ( j=skew ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip ) 
           {
             x_off = img_border + ( j * sub_cols ) + sub_border;
             y_off = img_border + ( i * sub_rows ) + sub_border;
  
             for (m=0; m<height; m++)
             {
               for (n=0; n<width; n++)
               {
                 i2[(band_cnt * height * width ) + m * width + n] 
                          = i1[(y_off+m)*image->row_size+(x_off+n)];
               }
             }
  
             band_cnt ++;
      
             if ( (sub_col_cnt - j) <= skip )
             {
               skew = skip - (sub_col_cnt - j);
             }
           }  /* end of loop through subimage columns */
         }  /* end of loop through subimage rows */
         break; /* end case VFF_TYP_4_BYTE */
         

    case VFF_TYP_FLOAT:

         f1 = (float *)(image->imagedata);
         f2 = (float *)(img2->imagedata);

/*
         * First extract subimages in the partial(?) row defined by 
         * sub_x_off and sub_y_off
         */
         for ( j=sub_x_off ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip )
         {
           x_off = img_border + ( j * sub_cols ) + sub_border;
           y_off = img_border + ( sub_y_off * sub_rows ) + sub_border;
  
           for (m=0; m<height; m++)
           {
             for (n=0; n<width; n++)
             {
               f2[(band_cnt * height * width ) + m * width + n] 
                        = f1[(y_off+m)*image->row_size+(x_off+n)];
             }
           }
  
           band_cnt ++;
    
           if ( (sub_col_cnt - j) <= skip )
           {
             skew = skip - (sub_col_cnt - j);
           }
         } 
  
        /* 
         * Extract the rest of the (complete) rows.
         */
  
         for ( i=(sub_y_off+1) ; (i<sub_row_cnt) && (band_cnt<num_subs) ; i++ )
         {
           for ( j=skew ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip ) 
           {
             x_off = img_border + ( j * sub_cols ) + sub_border;
             y_off = img_border + ( i * sub_rows ) + sub_border;
  
             for (m=0; m<height; m++)
             {
               for (n=0; n<width; n++)
               {
                 f2[(band_cnt * height * width ) + m * width + n] 
                          = f1[(y_off+m)*image->row_size+(x_off+n)];
               }
             }
  
             band_cnt ++;
      
             if ( (sub_col_cnt - j) <= skip )
             {
               skew = skip - (sub_col_cnt - j);
             }
           }  /* end of loop through subimage columns */
         }  /* end of loop through subimage rows */
         break; /* end case VFF_TYP_FLOAT */


    case VFF_TYP_COMPLEX:

         f1 = (float *)(image->imagedata);
         f2 = (float *)(img2->imagedata);

/*
         * First extract subimages in the partial(?) row defined by 
         * sub_x_off and sub_y_off
         */
         for ( j=sub_x_off ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip )
         {
           x_off = img_border + ( j * sub_cols ) + sub_border;
           y_off = img_border + ( sub_y_off * sub_rows ) + sub_border;

           for (m=0; m<height; m++)
           {
             for (n=0; n<width*2; n++)
             {
               f2[(band_cnt * height * width * 2) + m * width * 2 + n] 
                        = f1[(y_off+m)*2*image->row_size+(x_off*2+n)];
             }
           }
  
           band_cnt ++;
    
           if ( (sub_col_cnt - j) <= skip )
           {
             skew = skip - (sub_col_cnt - j);
           }
         } 
  
        /* 
         * Extract the rest of the (complete) rows.
         */
  
         for ( i=(sub_y_off+1) ; (i<sub_row_cnt) && (band_cnt<num_subs) ; i++ )
         {
           for ( j=skew ; (j<sub_col_cnt) && (band_cnt<num_subs) ; j+=skip ) 
           {
             x_off = img_border + ( j * sub_cols ) + sub_border;
             y_off = img_border + ( i * sub_rows ) + sub_border;
  
             for (m=0; m<height; m++)
             {
               for (n=0; n<width*2; n++)
               {
                  f2[(band_cnt * height * width * 2) + m * width * 2 + n] 
                          = f1[(y_off+m)*2*image->row_size+(x_off*2+n)];
               }
             }
  
             band_cnt ++;
      
             if ( (sub_col_cnt - j) <= skip )
             {
               skew = skip - (sub_col_cnt - j);
             }
           }  /* end of loop through subimage columns */
         }  /* end of loop through subimage rows */
         break; /* end case VFF_TYP_COMPLEX */

    default:
         kerror("vipl","lvmos2band", "Unknown data storage type, %d\n",
                img2->data_storage_type);
          return(FALSE);
  } /* end type case statement */

  *img1 = img2;
  return(TRUE);

}
/* -library_code_end */
