 /*
  * Khoros: $Id: lvpml.c,v 1.15.2.3 1997/05/31 20:17:07 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvpml.c,v 1.15.2.3 1997/05/31 20:17:07 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vpml
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvpml
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvpml - Estimates the Fractal Dimension of an image using the P(m,L) method.
* 
*       Purpose: To estimate the fractal dimension of an image based
*		 on the probability that there are m pixels within a
*		 window of size L centered on a pixel from a particular
*		 class.  For a selected range of window sizes (L), the
*		 window is centered on the first occurrence of the
*		 pixel belonging to a particular class.  The number
*		 of pixels within a window of size L, belonging to a
*		 specific class are counted (including the center
*		 pixel of the window), and a "histogram" is formed as
*		 the window is moved over the image.  This "histogram"
*		 represents the total number of occurrences, m, of a
*		 class of pixels in a window of size L.  From this, a
*		 normalized histogram is formed, which yields an
*		 estimate of the probability density function, P(m,L),
*		 for each window size L.  
*
*		 All selected moments (q) of the P(m,L) distributions
*		 for each of the desired window sizes are determined,
*		 and a linear regression or "best fit line" is found
*		 for the moment generating function, log(M(L))^1/q,
*		 versus the log of the window size, log(L).  The slope
*		 of the "best fit line" provides an estimate for the
*		 fractal dimension, D.
*
*		 The center pixel of the largest window, Lmax, is
*		 replaced with the computed fractal dimension based
*		 on the probability density function, P(m,L).  The
*		 largest window size, Lmax, determines the resulting
*		 size of the fractal dimension image.  This results
*		 in a border of size (Lmax / 2) around the image,
*		 which is set to zero. 
*
*		 The input image MUST be of data storage type BYTE
*		 or INTEGER. 
*
*		 The input arguments are described as follows:
*
*		 img specifies the input image, which must be
*		 of data type BYTE or INTEGER.  The input image
*		 must be a single band image.
*
*		 img1 specifies the output image, which will be a
*		 multiband image representing the fractal dimension
*		 of the input image for each of the specified moments.
*		 The resulting output image will be of data storage
*		 type FLOAT, and will have a border of pixels of
*		 value zero.  The size of the border will be determined
*		 by the size of the largest window specified.  This
*		 can be determined from the following formulation,
*		 (Lmax -1) / 2 = border size.  Where Lmax is the size
*		 of the largest window used.  For example, if the
*		 largest window size is 9, then the border size will
*		 be (9 - 1) / 2 = 4 pixels.
*
*		 fd_img specifies fractal dimension image, representing
*		 the fractal dimension vectors for each class and moment.
*		 This image will always have a row size of 1 and a
*		 column size determined by the number of classes in the
*		 input image.  The number of data bands will be equal
*		 to the number of moments used in the estimation of
*		 the fractal dimension.
*
*		 printdev uses an ASCII file as output for the image
*		 information specifying the size of the input image,
*		 number of classes in the input image, and the range
*		 of moments used in the estimation of the fractal
*		 dimension.  Also included is a listing of the fractal
*		 dimensions for each class and moment.
*
*		 lower specifies the initial or lower size of the
*		 sliding window.  This MUST be an odd number resulting
*		 in a window with a center pixel.  This means that the
*		 minimum size of the sliding window is 3 x 3.  The
*		 default value is 3, resulting in an initial window
*		 size of 3 x 3.
*
*		 upper specifies the final or upper size of the sliding
*		 window.  This MUST also be an odd number resulting in
*		 a window with a center pixel.  The size of the upper
*		 window MUST be at least one step greater than the
*		 lower window size.  That is, if the lower window is
*		 3 x 3, then the minimum size of the upper window must
*		 be 5 x 5, or one step greater than the lower window.
*		 This is necessary, since there must be at least two
*		 points for a best fit line to be formed determining
*		 the slope and ultimately the fractal dimension.
*
*		 step specifies the step size or interval used when
*		 specifying a range of window sizes.  The default
*		 value is two, corresponding to the next odd window
*		 size with a center pixel.  This may be helpful when
*		 a wide range of window sizes is required, as it will
*		 reduce the number of points generated for the best
*		 fit line and hence the number of calculations.
*
*		 range_q specifies the range of moments to base the
*		 P(m,L) fractal dimension calculations on.  This
*		 must be an odd number, corresponding to the number
*		 of moments centered on zero.  For example, a value
*		 of 5 would result in the computation of the fractal
*		 dimension for the range of moments, -2, -1, 0, 1, 2.
*		 The default value is 3, resulting in the range of
*		 -1, 0, 1.
*
*         Input: img     - input image structure
*		 lower   - specifies the initial size of the sliding window.
*		 upper   - specifies the upper size of the sliding window.
*		 step    - specifies the step size or interval for a range
*		           of window sizes.
*		 range_q - specifies the range of moments to base the P(m,L)
*			   calculations on.
*
*        Output: img1     - output image structure, may be multiband.
*		 fd_img   - fractal dimension image, represents the fractal
*			    dimension vectors.
*		 printdev - ASCII output file for the image information.
*
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: The input image MUST be of data storage type BYTE or INTEGER. 
*		 The output images are of data storage type FLOAT.
*    Written By: Charlie Gage
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvpml(
	xvimage *img,
	xvimage *img1,
	xvimage *fd_img,
	kfile *printdev,
	int lower,
	int upper,
	int step,
	int range_q)
/* -library_def_end */

/* -library_code */
{
 int   i,             /* indices, counters and temporary variables */
       j,
       k,
       p,
       t,
       row,
       col,
       count,
       num_pts,
       num_L,         /* number of windows used */
       L,             /* window size index */
       l_sqr,         /* maximum size of histogram array */
       w_sqr,         /* max size of histogram array for current window */
       max_class,     /* maximum number of classes in image */
       nr,            /* number of rows in image */
       nc,            /* number of cols in image */
       nb,            /* number of bands in image */
       offset,        /* offset into image, dependent on max window size */
       xend,          /* step into image up to here (x-direction) */
       yend,          /* step into image up to here (y-direction) */
       w_offset,      /* offset into sliding window */
       class1,        /* current class of pixel */
       x;             /* current pixel value in image */

 int
       *imgptr,       /* pointer to input image data */
       **workptr;     /* pointer to 2-D input image data working array */

 float
       q,             /* moment index */
       m,             /* temp histogram variable */
       sum_m,         /* temp summing variable */
       prob,          /* current probability value */
       tmp_slope,     /* temp slope variable */
       sum_L,         /* sum of log(L), window size */
       sum_sqr,       /* sum of squares */
       tmp,           /* temp variable for calculating slope */
       avg;           /* average of log(L) */

 float *result,       /* pointer to resulting multiband fractal image */
       *fract_dim,    /* pointer to resulting fractal dim vector */
       ***moment,     /* pointer to 3-D array of M(L) for each moment(q) & L */
       **histo;       /* pointer to 2-D array of normalized histogram data */
 
  /* determine number of windows used */
  num_L = ((upper - lower) / step) +1;

  /* Initialize variables */
  nb = range_q;              /* number of bands equals number of moments */
  nr = img->col_size;        /* number of rows (y-direction) in image */
  nc = img->row_size;        /* number of cols (x-direction) in image */

 /*-------------------------------------------------------------------*
  *    ERROR Checking statements follow:
  *-------------------------------------------------------------------*/

 /* Ensure that Windows are odd */
 if ((lower % 2) == 0  ||  (upper % 2) == 0)
  {
   kerror("vipl", "lvpml", "Window size must be an odd number.");
   return(FALSE);
  }

 /* Ensure that Windows are greater than 1 */
 if ((lower <= 1)  ||  (upper <= 1))
  {
    (void) kerror("vipl", "lvpml", "Windows must be greater than 1.");
    return(FALSE);
  }

  /* Ensure that lower window size does not exceed image size */
  if ((lower > nc) || (lower > nr))
  {
   kerror("vipl","lvpml", "Lower window size exceeds image size.\n \
Requested lower window size of %d, but image size is %d by %d.", lower, nc, nr);
   return(FALSE);
  }

  /* Ensure that upper window size does not exceed image size */
  if ((upper > nc) || (upper > nr))
  {
   kerror("vipl","lvpml", "Upper window size exceeds image size.\n \
Requested upper window size of %d, but image size is %d by %d.", upper, nc, nr);
   return(FALSE);
  }

  /* Ensure that lower window size does not exceed upper window size */
  /* Also, need at least two points to define the slope */
  if (lower >= upper)
  {
   kerror("vipl","lvpml", "Lower window size equals or exceeds upper window size.\n \
Requested lower window size of %d, but upper window size is %d", lower, upper);
   return(FALSE);
  }

 /* Ensure that step size is an even number. Windows must be odd. */
 if ((step % 2) != 0)
  {
    kerror("vipl","lvpml", "Step size must be an even number.");
    return(FALSE);
  }

 /* Ensure that step size does not exceed difference between upper and lower */
 if (step > (upper - lower))
  {
   kerror("vipl","lvpml", "Step size exceeds range of window sizes.");
   return(FALSE);
  }

 /* Ensure that the range of moments is an odd number. */
 if ((range_q % 2) == 0)
  {
    kerror("vipl","lvpml", "Range of moments (q) must be an odd number.");
    return(FALSE);
  }

  /*---------------  End of Error Checking  ------------------------------*/

 /*---------------------------------------------------------------------*
  *  Determine offsets and bookkeeping variables for image pointer 
  *   and Initialize variables.
  *---------------------------------------------------------------------*/
  count = 0;
  p = 0;                     /* initialize counter for each window (L) */
  sum_m = 0.0;
  l_sqr = upper * upper + 1; /* max number of cols in histogram array */
  offset = upper/2;          /* offset into image */
  xend = nc - upper + 1;     /* step in image up to here (x-direction) */
  yend = nr - upper + 1;     /* step in image up to here (y-direction) */
  max_class = 0;             /* initialize max number of classes to zero */

                             /* Assign input image data address to ptr */
  imgptr = (int *)img->imagedata;

  /*-----------------------------------------------------------------------*
   *  Allocate memory for arrays and images 
   *-----------------------------------------------------------------------*/

  /* Allocate space for 2-D input image working array (workptr) */
  workptr = (int **) kmalloc((unsigned int) nr * sizeof(int *));
  if (workptr == NULL)
  {
    kerror("vipl","lvpml", "Insufficient memory available");
    return(FALSE);
  }
  for (i = 0; i < nr; i++)
  {
   workptr[i] = (int *) kmalloc((unsigned int) nc * sizeof(int));
   if (workptr[i] == NULL)
   {
     kerror("vipl","lvpml", "Insufficient memory available");
     return(FALSE);
   }
  }   

  /* Assign input image data to 2-D input image working array */
  for (i = 0; i < nr; i++)
  {
   for (j = 0; j < nc; j++)
       workptr[i][j] = imgptr[nc * i + j];
  }

  /* release memory from input image */
  (void) kfree_and_NULL(img->imagedata);

  /* Scan image to determine number of classes in classified image */
  for (i = 0; i < nr; i++)
  {
    for (j = 0; j < nc; j++)
    {
      x = workptr[i][j];
      if (x > max_class)
         max_class = x;
    }
  }
             /* max number of classes in image is one greater than the 
                maximum pixel (class) value found, since 0 is a class  */
  max_class++;

  /* Allocate space for resulting fractal dimension image (multiband) */
  result = (float *)kmalloc((unsigned int) nr * nc * nb * sizeof(float));
  if (result == NULL)
  {
    kerror("vipl","lvpml", "(2) insufficient memory available!");
    return(FALSE);
  }

  /* Allocate space for resulting fractal dimension vlaues (multiband) */
  fract_dim = (float *)kmalloc((unsigned int) max_class * nb * sizeof(float));
  if (fract_dim == NULL)
  {
    kerror("vipl","lvpml", "(3) insufficient memory available!");
    return(FALSE);
  }

  /* Allocate space for resulting 2-D histogram array */
  histo = (float **) kmalloc((unsigned int) max_class * sizeof(float *));
  if (histo == NULL)
  {
    kerror("vipl","lvpml", "Insufficient memory available");
    return(FALSE);
  }
  for (i = 0; i < max_class; i++)
  {
   histo[i] = (float *) kmalloc((unsigned int) l_sqr * sizeof(float));
   if (histo[i] == NULL)
   {
    kerror("vipl","lvpml", "Insufficient memory available");
    return(FALSE);
   }
  }   

  /* Allocate space for 3-D array of M(L) for each moment, class, and L. */
  moment = (float ***) kmalloc((unsigned int) num_L * sizeof(float **));
  if (moment == NULL)
  {
    kerror("vipl","lvpml", "Insufficient memory available");
    return(FALSE);
  }
     /* Set up 3 dimensional array */
  for (i = 0; i < num_L; i++)
  {
              /* Allocate space for each band */
   moment[i] = (float **) kmalloc((unsigned int) max_class * sizeof(float *));
   if (moment[i] == NULL)
   {
     kerror("vipl","lvpml", "Insufficient memory available");
     return(FALSE);
   }
     /* Allocate rows and set pointers to them */
   for (j = 0; j < max_class; j++)
   {
    moment[i][j] = (float *) kmalloc((unsigned int) range_q * sizeof(float));
    if (moment[i][j] == NULL)
    {
      kerror("vipl","lvpml", "Insufficient memory available");
      return(FALSE);
    }
   }
  }

  /*---------------------------------------------------------------------*
   *  Initialize data arrays to zero
   *---------------------------------------------------------------------*/

  for (i = 0; i < max_class; i++)
  {
   for (j = 0; j < l_sqr; j++)
       histo[i][j] = 0.0;
  }

  /* Zero out 3-D array of moments */

  for (i = 0; i < num_L; i++)
  {
   for (j = 0; j < max_class; j++)
   {
    for (k = 0; k < range_q; k++)
         moment[i][j][k] = 0.0;
   }
  }

  /* Zero out resulting fractal dimension image */
  for (i = 0; i < nr; i++)
  {
   for (j = 0; j < nc; j++)
   {
    for (k = 0; k < nb; k++)
         result[i*nc+j*nb+k] = 0.0;
   }
  }

  /* Zero out fractal dimension centers */
  for (i = 0; i < max_class; i++)
  {
   for (j = 0; j < nb; j++)
       fract_dim[i*nb+j] = 0.0;
  }

 /* Print input file and program information to ASCII file */
 kfprintf(printdev,"***  Input File Information  ***\n");
 kfprintf(printdev,"Input file must be a single band BYTE or INTEGER image.\n");
 kfprintf(printdev,"Num ROWS (nr) = %d   Num COLS (nc) = %d\n", nr, nc);
 kfprintf(printdev,"Maximum number of classes in input image: %d\n",max_class);
 kfprintf(printdev,"The number of classes corresponds to the number of\n");
 kfprintf(printdev,"different pixel values in the input image.\n");
 kfprintf(printdev,"\n***  Program Information  ***\n");
 kfprintf(printdev,"Lower Window Size: %d\n", lower);
 kfprintf(printdev,"Upper Window Size: %d\n", upper);
 kfprintf(printdev,"Window Step Size: %d\n", step);
 kfprintf(printdev,"Number of windows used: %d\n", num_L);
 kfprintf(printdev,"Range of moments selected: %d\n", range_q);
 kfprintf(printdev,"\n***  Output File Information  ***\n");
 kfprintf(printdev,"Num ROWS (nr) = %d   Num COLS (nc) = %d\n", nr, nc);
 kfprintf(printdev,"Num BANDS (nb) = %d\n", nb);
 kfprintf(printdev,"The resulting output file is a multiband float image,\n");
 kfprintf(printdev,"with a border (value = 0.0) width of %d pixels.\n",upper/2);
 kfprintf(printdev,"-----------------------------------------------------\n");

  /*---------------------------------------------------------------------*
   * Step 1 - Form histogram and calculate qth moment of P(m,L)
   *---------------------------------------------------------------------*/
                               /* for all specified window sizes (L) */
  for (L = lower; L <= upper; L += step)
  {                            /* over the entire image */
   w_sqr = L * L;              /* max size of histo array for current L */
   for (row = 0; row < yend; row++)
   {
    for (col = 0; col < xend; col++)
    {                    /* current class is center pixel of current window */
     class1 = workptr[row + offset][col + offset];
     for (i = 0; i < L; i++)    /* rows of window of size L */
     {
      w_offset = offset - (L / 2);
      for (j = 0; j < L; j++)   /* cols of window of size L */
      {
       x = workptr[row + w_offset + i][col + w_offset + j];
       if (x == class1)
           count++;        /* count number of pixels matching center pixel */
      }
     }
     histo[class1][count]++;           /* increment bin (m) of histogram */
     histo[class1][0]++;               /* sum number of elements in histogram 
                                         and store in 1st col of histo array. */
     count = 0;                       /* reset counter */
    }
   }

   /*----------------------------------------------------------------------*
    * Step 2 - Normalize histogram for each class
    * Need to preserve normalized histogram, histo[] for the current value
    *     of L, in order to calculate M(L) for each of the moments, q.
    *----------------------------------------------------------------------*/
   for (i = 0; i < max_class; i++)
   {
    if (histo[i][0] != 0.0)  /* skip class if no entries in histogram */
    {   
     for (j = 1; j <= w_sqr; j++) /* use only bins of current window size */
     {
      m = histo[i][j];
      if (m != 0.0)               /* skip bins with no entries */
         histo[i][j] = m / histo[i][0]; 
     }
    }
   }
   q = (float) -(range_q / 2);        /* determine the initial moment, q */

  /*------------------------------------------------------------------------*
   * Step 3 - Determine M(L) for all moments (q) using normalized 
   *          histogram data.
   *------------------------------------------------------------------------*/
   for (k = 0; k < range_q; k++)   /* do for all moments -10 <= q <= 10  */
   {
    for (i = 0; i < max_class; i++)
    {
     if (histo[i][0] != 0.0)  /* skip class if no entries in histogram */
     {
      for (j = 1; j <= w_sqr; j++) /* use only bins of current window size */
      {
       prob = histo[i][j];
       if (prob != 0.0)            /* skip bins with no entries */
        {
         if (q == 0.0)
            m = klog((double)j) * prob;
         else
            m = prob * kpow((double)j, (double)q);
         sum_m += m;
        }
      }       /* end of j loop */
      if (q == 0.0)
       {
        moment[p][i][k] = sum_m;
       }
      else
       {
        moment[p][i][k] = klog(kpow((double)sum_m,(double)(1.0/q)));
       }
      sum_m = 0.0;                /* reset sum_m for next iteration */
     }    /* end of if() */
    }     /* end of i loop */
    q++;                        /* increment moment counter, q */
   }       /* end of k loop */

   p++;                   /* increment window size (L) counter */
                          /* Zero out histogram for next window */
  for (i = 0; i < max_class; i++)
  {
    for (j = 0; j < l_sqr; j++)
        histo[i][j] = 0.0;
  }

 }   /* End of window size (L) loop  */
  
  /*---------------------------------------------------------------
  *  Step 4 - find slope and intercept of best fit line for the qth 
  *   moment of P(m,L) distributions for all selected window sizes,
  *   L.    ie.  find best fit line of log[M(L)] vs log(L)
  *---------------------------------------------------------------*/
                               /* initialize variables */
  tmp_slope = 0.0;
  sum_L = 0.0;                   /* defines sum of log(L) */
  sum_sqr = 0.0;
  num_pts = 0;

  for ( i = 0; i < max_class; i++)    /* for each class */
  {
   for (k = 0; k < range_q; k++)     /* for each moment, q */
   {
    for (L = lower; L <= upper; L += step)
    {
      sum_L += klog((double)L);
      num_pts++;
    }
    avg = sum_L / (float)num_pts;
    p = 0;
    for (L = lower; L <= upper; L += step)
    {
      tmp = klog((double)L) - avg;
      sum_sqr += tmp * tmp;
      tmp_slope += tmp * moment[p][i][k];
      p++;
    }
            /* slope is an estimate of the fractal dimension, D.  */
             /* find slope of best fit line */
    fract_dim[i + (k * max_class)] = (tmp_slope / sum_sqr);
                        /* reset counters and variables */
    sum_L = 0;
    sum_sqr = 0;
    num_pts = 0;
    tmp_slope = 0;
   }
  }
 
 /*------------------------------------------------------------------------*
  * Construct multi-band fractal dimension image from classified image
  *------------------------------------------------------------------------*/ 
  for (i = 0; i < nr; i++)
  {
   for (j = 0; j < nc; j++)
   {
            /* determine fractal dimension only for valid region of image */
    if((i >= offset) && (i < (nr-offset)) && (j >= offset) && (j < (nc-offset)))
    {
      class1 = workptr[i][j];       /* get current class of image */
      for (k = 0; k < range_q; k ++)            /* k is currnet moment */
      {                       /* get fractal dimension for current moment */
       result[nc * i + j + (k * nr * nc)] = fract_dim[class1 + (k * max_class)];
      }
    }
   }
  }

  /*  Print out the fractal dimension vectors for each band  */
  t = -(range_q / 2);
  (void)kfprintf(printdev,"\n\n**********  Fractal Dimension Vectors  ***********\n");
  (void)kfprintf(printdev,"\t\tclass\n");
  (void)kfprintf(printdev,"moments  ");
  for (i = 0; i < max_class; i++)
  {
   (void)kfprintf(printdev," %d  ", i);
  }
  (void)kfprintf(printdev,"\n-------  -------------------------\n");
  for (k = 0; k < range_q; k++)
  {
   (void)kfprintf(printdev,"q = %d\t", t);
   for (i = 0; i < max_class; i++)
   {
    (void)kfprintf(printdev," %.1f", fract_dim[i + (k * max_class)]);
   }
   (void)kfprintf(printdev,"\n----------------------------------\n");
   t++;
  }

  /* adjust header in the output image */
  img1->data_storage_type = VFF_TYP_FLOAT;
  img1->imagedata = (char *) result;
  img1->num_data_bands = range_q;
  fd_img->data_storage_type = VFF_TYP_FLOAT;
  fd_img->imagedata = (char *) fract_dim;
  fd_img->num_data_bands = range_q;

  return(1);

}  /*** End of lvpml ***/
/* -library_code_end */
