 /*
  * Khoros: $Id: lvsamp.c,v 1.15.2.2 1997/05/31 20:17:14 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvsamp.c,v 1.15.2.2 1997/05/31 20:17:14 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vsamp
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvsamp
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvsamp - Sample an image along an arbitrary line.
* 
*       Purpose: NOTE:  the lvsamp function will be removed in the future.
*                lvsamp functionality is replaced by the
*                lksampline() function in the DATAMANIP::kdatamanip
*                library.  Additional input parameters are required by
*                lksampline().
*
*       	 Given an image of arbitrary size and a line through
*		 that image, sample the image along the line.  The
*		 sampling interval is independent of image size and
*		 line orientation. The value at each sample point is
*		 obtained by bilinear interpolation.
*
*		 The first sample point is precisely at the starting
*		 point and the final sample is precisely at the ending
*		 point.
*         Input: image - pointer to image to be sampled
*		 line  - pointer to image pointer to receive the output image
*		 px1    - X coordinate of start point
*		 py1    - Y coordinate of start point
*		 px2    - X coordinate of end point
*		 py2    - Y coordinate of end point
*		 n     - number of samples to generate between the endpoints
*        Output: line  - xvimage pointer indicated by line is attached
*			 to the new image
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: lvsamp will only work on images of type FLOAT.
*    Written By: Richard Krukar
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 6-Feb-91 Scott Wilson  - Cleaned up for installation,
*					  completed documentation.
*		 8-Feb-91 Scott Wilson  - changed length to n, added
*					  pixel size stuff.
*		 10-Feb-91 Scott Wilson - re-wrote the entire inner
*				 	  loop - the original had bugs
*					  and did not handle the boundary
*					  conditions.
*		 20-Feb-91 Scott Wilson - corrected pixel size computation bug.
****************************************************************/
/* -library_def */
int lvsamp(
	xvimage *image,
	xvimage **line,
	int px1,
	int py1,
	int px2,
	int py2,
	int n)
/* -library_def_end */

/* -library_code */
{
    int i,rows,cols;
    float x_delta,y_delta;
    float *i_data,*line_dat;
    xvimage *l;
    int yl,yh,xl,xh;
    float pyn,pxn;
    float z1,z2,z3,z4,za,zb,zp;
    float a,b;

/* uncomment this as soon as lvcost is removed from retro 
    kerror("vipl", "lvsamp", "lvsamp function NOT executed!\n"
           "  lvsamp() functionality is replaced by lksampline()\n"
           "  in the DATAMANIP::kdatamanip library.");
    if (TRUE) return FALSE;
*/

    /* Check image characteristics */
    if (image->data_storage_type != VFF_TYP_FLOAT)
      {
        kerror("vipl", "lvsamp", "Input image must be of type FLOAT\n");
        return(FALSE);
      }
    if (image->num_data_bands != 1 || image->map_scheme != VFF_MS_NONE)
      {
        kerror("vipl", "lvsamp", "Input image must be single band, no map!\n");
        return(FALSE);
      }

    rows = image->col_size;
    cols = image->row_size;

    /* Check line endpoints to make sure they are in bounds */
    if (px1 < 0 || px1 > cols-1 ||
        py1 < 0 || py1 > rows-1 ||
        px2 < 0 || px2 > cols-1 ||
        py2 < 0 || py2 > rows-1)
      {
        kerror("vipl", "lvsamp",
	       "Line endpoints must lie within image boundaries!\n");
        return(FALSE);
      }
    if (px1 == px2 && py1 == py2)
      {
        kerror("vipl", "lvsamp",
	       "Line endpoints must not be the same point!\n");
        return(FALSE);
      }

    /* Make output image */
    *line = createimage((unsigned long) 1,(unsigned long) n,(unsigned long) VFF_TYP_FLOAT, (unsigned long) 1,(unsigned long) 1,"created by vsamp", (unsigned long) 0,(unsigned long) 0,(unsigned long) VFF_MS_NONE,(unsigned long) VFF_MAPTYP_NONE, (unsigned long) VFF_LOC_IMPLICIT,(unsigned long) 0);
    if (*line == NULL)
      {
        kerror("vipl", "lvsamp", "Unable to allocate memory for line image\n");
        return(FALSE);
      }

    l = *line;
    line_dat = (float *)(l->imagedata);
    for ( i = 0; i < n ; i++ )
      {
	line_dat[i] = 0.0;
      }

    /* Compute sampling intervals */
    x_delta = (float)(px2-px1)/(float)(l->row_size-1);
    y_delta = (float)(py2-py1)/(float)(l->row_size-1);

    /* Set new pixel sizes */
    a = (px2-px1)*image->pixsizx;
    b = (py2-py1)*image->pixsizy;
    l->pixsizx = sqrt((double)(a*a+b*b))/(float)(l->row_size-1);
    l->pixsizy = l->pixsizx;

    i_data = (float *)(image->imagedata);
    for(i=0; i<(l->row_size)-1; i++)
      {
        /* Compute coordinates of i'th sample point */
        pxn = px1 + i*x_delta;
        pyn = py1 + i*y_delta;
        /* Compute coordinates of pixels bounding the box the sample
           point lies in */
        xl = floor((double)pxn);
        xh = xl+1;
        yl = floor((double)pyn);
        yh = yl+1;
        if (xl == cols-1 && yl == rows-1)
          {
            /* Desired point is corner value - just return it */
            *line_dat = i_data[PIXEL(xl,yl,rows,cols)];
          }
        else if (xl == cols-1 && yl != rows-1)
          {
            /* If were at the right hand edge then interpolate only
               on the y values.(HERE) */ 
            z1 = i_data[PIXEL(xl,yl,rows,cols)];
            z2 = i_data[PIXEL(xl,yh,rows,cols)];
            zp = z1*(yh-pyn)+z2*(pyn-yl);
            *line_dat = zp;
          }
        else if (xl != cols-1 && yl == rows-1)
          {
            /* Were on the bottom edge, so interpolate only on the
               x values */
            z1 = i_data[PIXEL(xl,yl,rows,cols)];
            z2 = i_data[PIXEL(xh,yl,rows,cols)];
            zp = z1*(xh-pxn)+z2*(pxn-xl);
            *line_dat = zp;    
          }
        else
          {
            /* If the containing box is fully in bounds then interpolate the 
               four pixel values to find the intermediate value */
            z1 = i_data[PIXEL(xl,yl,rows,cols)];
            z2 = i_data[PIXEL(xh,yl,rows,cols)];
            z3 = i_data[PIXEL(xh,yh,rows,cols)];
            z4 = i_data[PIXEL(xl,yh,rows,cols)];
            za = z1*(yh-pyn)+z4*(pyn-yl);
            zb = z2*(yh-pyn)+z3*(pyn-yl);
            zp = za*(xh-pxn)+zb*(pxn-xl);
            *line_dat = zp;
          }
        line_dat++;
      }
    *line_dat = i_data[PIXEL(px2,py2,rows,cols)];
    return(TRUE);
}
/* -library_code_end */
