 /*
  * Khoros: $Id: lvwmdd.c,v 1.16.2.3 1997/05/31 20:17:32 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: lvwmdd.c,v 1.16.2.3 1997/05/31 20:17:32 mirage Exp $";
#endif

/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */

/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>> 
   >>>> 	Library Routine for vwmdd
   >>>> 
   >>>>  Private: 
   >>>> 
   >>>>   Static: 
   >>>>   Public: 
   >>>> 	lvwmdd
   >>>> 
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */


#include "internals.h"

/* -library_includes */
/* -library_includes_end */


/****************************************************************
* 
*  Routine Name: lvwmdd - Weighted minimum distance detector
* 
*       Purpose: lvwmdd is an Weighted minimum distance detector
*		 or Classifier.  The main idea behind this detector
*		 is that it will distinguish a single class form
*		 the rest of the data. However, it will also work
*		 to distinguish multiple classes from each other. 
*
*		 Theory
*
*		 To classify or detect spatial patterns in an image, we
*		 must have a template to compare against. Creating a
*		 template can be thought of as training on representive
*		 (ideal) data. This can be done by using a clustering
*		 algorithm on the representive data. Clustering routines
*		 such as vkmeans, vquant, isodata, etc can be used to
*		 create a template.  The idea here is to over cluster
*		 the representive data, and quantize that into k classes.
*		 Pseudo color in xvdisplay, and vcustom can be used to 
*		 create a class image. When the clustering is performed,
*		 two items need to be computed; (1) the cluster center
*		 values, and (2) the variances for each axis in the
*		 cluster. The variances can be obtained by looking at
*		 the diagonal of the covariance matrix. 
*
*		 Explained below is a method of creating a the class
*		 assignments for each cluster center. The class assignments
*		 may, for example, quantize the space from 12 clusters
*		 to 2 classes. Example
* !
* !
* !        Cluster Center               Class
* !              1 ---------------------  2
* !              2 ---------------------  2
* !              3 ---------------------  1
* !              4 ---------------------  2
* !                    .
* !                    .
* !                    .
* !             12 ---------------------  1
* !
* !   
*
*		 This routine expects an image that contains the diagonal
*		 terms of the covariance matrix, called the variance image,
*		 and an image that contains the cluster center values and
*		 class assignment for each cluster center. And of course
*		 the image to be classified.  This detector is to approximate
*		 the following equation:
* !
* !         (X - Mi)~ Inv(C) (X - Mi)                        (1)
* !
*		 where Mi is the mean, inv(C) is the inverse of the Covariance
*		 matrix, and X is the data points, and ~ denotes transpose.
*		 This can be written as
* !
* !     (d1 d2 ... dn)~ - Q1           - (d1 d2 ... dn)     (2)
* !                     |    Q2        |
* !                     |      .   0   |
* !                     |        .     |
* !                     |  0       .   |
* !                     |           Qn |
* !                     -              -
* !
*		 Which equals,
* !
* !          sq(d1)   sq(d2)           sq(dn)
* !          ------ + ------ + ..... + ------             (3)
* !            Q1       Q2               Qn 
* ! 
* !		 Notation:
* !		            sq => square
* !		            Qi => the ith variance
* !		            di => the ith (X - Mi) value
* !		            the matrix is actually the inverse of C
* !
*
*		 Since the inverse of the Covariance matrix can not be
*		 easily determined, we only consider the diagonal terms
*		 or simlpy the variance terms.
*
*		 There are two methods for detecting classes in the input
*		 image. Each data point is put through a test based on
*		 the ideas above. The detector works in two modes as follows:
*
*		 (1) uses the summed method -a option set to yes. This is
*		 an apprioximation to method (2). 
* !
* !		             sq( || X - Mi || )   <
* !		      Vi =   ------------------   >   1
* !		                 K x sq(Si)
* !
* !		      where:    sq( || X - Mi || ) =  sq(di) 
* !		                                   = Eucludian Distance Squared
* !		                sq(Si) = trace(C)
* !		                   the trace of C is the sum of the 
* !		                   diagonal terms (variances).
* !		                K is a Constant
* !
*		 (2) non summed method (default). 
* !
* !		              sq(d1)   sq(d2)           sq(dn)   <
* !		      Vi =    ------ + ------ + ..... + ------   >   1
* !		              Q1 x K   Q2 x K           Qn x K
* !
* !		      where:  sq(dj) = sq( || xj - Mi || ) 
* !		                     = Eucludian Distance Squared of the
* !		                       jth element of the vector X.
* !		              Qj = the jth variance element of the varance 
* !		                   vector for the ith cluster center.
* !		              K = Constant
* !
*
*		 In both cases the constant K is used. This is used to
*		 adjust the variance value(s). If K is large, Qi can
*		 increase such that Vi is always less than 1. There are
*		 no specific rules as to the optimal value for K.
*
*		 The data point X is put throught the test for each cluster,
*		 the data point X is assigned to a Class based on the
*		 following criteria: choose the smallest Vi value, where
*		 Vi is the result for cluster i.
*
*		 If Vi is greater than 1, assign X to the NULL class (NULL
*		 class is always 0).  Otherwise assign Vi to class that
*		 corresponds to the ith cluster.
*
*		 The scale factor (k_factor) must be choosen by trial and
*		 error. It has been found that if the summed method is
*		 being used a small (1.0 - 3.0) value for k is sufficent.
*		 If the average method is not being used a larger k (8.0 -
*		 10.0) value is sufficient. Note, this may change based on
*		 the data, so these ranges are only a starting point.
*
*		 All input images must be of data storage type FLOAT. 
*
*		 img the input image that is to be classified. This is
*		 also used as the output classified image, so must be
*		 careful not to overwrite important data.
*
*		 center the image containing the center values (or prototype
*		 vectors) with the last data band being the class image.
*		 This image must contain n + 1 data bands, where n = # of
*		 data bands in the img image, and 1 data band that specifies
*		 the class assignments.
*
*		 varimg the image containing the variance values. This
*		 image must have the same row and column size of the center
*		 image, and have the same number of data bands as the input
*		 image (img).
*
*		 border the border width on the image. The border corresponds
*		 to a no class assignment.  The border area is ignored during
*		 the classification.
*
*		 k_factor the fudge factor, explained in detail above.
*
*		 use_avg if use_avg = 1, use the summing method, if
*		 use_avg = 0 use the non-summing method, as explained above.
*
*		 This routine was written with the help of and ideas from
*		 Dr. Don Hush, University of New Mexico, Dept. of EECE.
*
*         Input: img      - the input image that is to be classified.
*		 center   - the image containing the center values (or
*			    prototype vectors) with the last data band
*			    being the class image. This image must have
*			    n + 1 data bands where n is the number of
*			    data bands in img.
*		 varimg   - the image containing the variance values for
*			    the clusters.  Should have the same number of
*			    data bands as the input image (img).
*		 border   - the border width on the image. The border
*			    corresponds to a no class assignment.  i.e. the
*			    border region is ignored during the classification
*			    process.
*		 k_factor - scalar, fudge factor.
*		 use_avg  - 0 or 1
* !                         0 - use non-summing method
* !                         1 - use summing method
*
*        Output: img      - a single band classified image.
*       Returns: TRUE (1) on success, FALSE (0) on failure
*  Restrictions: All input images must be of data storage type FLOAT.
*    Written By: Tom Sauer
*          Date: July 1, 1997
*      Verified: 
*  Side Effects: 
*      Examples: 
* Modifications: 
****************************************************************/
/* -library_def */
int lvwmdd (
	xvimage *img,
	xvimage *center,
	xvimage *varimg,
	int border,
	float k_factor,
	int use_avg )
/* -library_def_end */

/* -library_code */
{
    int     nc,                 /* column size of image */
            nr,                 /* row size of image */
            cc_nc,              /* column size of image */
            cc_nr,              /* row size of image */
            vect,               /* number of vectors in image */
            dimension,          /* size of vector - num_bands */
            n_centers,
            num_centers,
            i,                  /* loop indices */
            j, m, k,x , z, c;

    float  d, a, q, dmin;
    float  *ptr,                /* pointer to the image data char */
           *varptr,
          **input_vectors,      /* pointer to image data reorganized */
           *new_class,          /* pointer to result image */
           *class1,             /* pointer to class data that corresponds */
           *class_valid,
                                /* to the cluster centers */
          **c_center,           /* cluster centers */
          **var_vector;         /* square distance values */

    nr = img -> col_size;                /* number of rows in image */
    nc = img -> row_size;                /* number of columns in image */
    cc_nr = center->col_size;   /* number of rows in center image */
    cc_nc = center->row_size;   /* number of columns in center image */

    dimension = img -> num_data_bands;
    vect = (nr - (border * 2)) * (nc - (border * 2)); /* number of vectors */
    n_centers = cc_nr * cc_nc;

 /* Assign class to the last band of the center image */

    ptr = (float *) center -> imagedata;
    class1 = (float *) (&ptr[cc_nr * cc_nc * dimension]);
    num_centers = 0;
    for (i =0 ; i< n_centers; i++)
    {
        if (class1[i] != 0.0)
           num_centers++;
    }

    class_valid = (float *) kmalloc (sizeof (float *) * num_centers);
    if (class_valid == NULL) {
        kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
        return (FALSE);
    }

    ptr = class_valid;
    for (i =0 ; i< n_centers; i++)
    {
       if (class1[i] != 0.0)
       {
         *ptr = class1[i];
         ptr++;
       }
    }

    /* Allocate space for input_vectors image */
    /* reorganize the data because it is not organized in an efficient */
    /* manner for the computational complexity of the routine */

    input_vectors = (float **) kmalloc ((unsigned int) vect * sizeof (float));
    if (input_vectors == NULL) {
        kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
        return (FALSE);
    }

    for (i = 0; i < vect; i++) {
        input_vectors[i] = (float *) kmalloc (dimension * sizeof (float));
        if (input_vectors[i] == NULL) {
            kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
            return (FALSE);
        }
    }

/* Allocate space for cluster centers */

    c_center = (float **) kmalloc (sizeof (float *) * num_centers);
    if (c_center == NULL) {
        kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
        return (FALSE);
    }

    for (i = 0; i < num_centers; i++) {
        c_center[i] = (float *) kmalloc (sizeof (float) * dimension);
        if (c_center[i] == NULL) {
            kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
            return (FALSE);
        }
    }

/* Allocate space for variance values */

    var_vector = (float **) kmalloc (sizeof (float *) * num_centers);
    if (var_vector == NULL) {
        kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
        return (FALSE);
    }

    for (i = 0; i < num_centers; i++) {
        var_vector[i] = (float *) kmalloc (sizeof (float) * dimension);
        if (var_vector[i] == NULL) {
            kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
            return (FALSE);
        }
    }


    /* Allocate space for output image data */

    new_class = (float *) kmalloc ((unsigned int) nr * nc * sizeof (float));
    if (new_class == NULL) {
        kerror ("vipl", "lvwmdd", "Insufficient memory available\n");
        return (FALSE);
    }

    /* Zero out output imagedata */

    for (i =0 ; i< nr; i++) {
        for (j =0 ; j< nc; j++) {
	    new_class[i * nc + j] = 0.0;
        }
    }


    /* Assign image data address to ptr */

    ptr = (float *) img -> imagedata;

    /* Assign ptr to input vectors minus the border */
    m = 0;
    for (i = border; i < nr - border; i++) {
        for (j = border; j < nc - border; j++) {
            for (k = 0; k < dimension; k++) {
                input_vectors[m][k] = ptr[(i * nc + j) + (k * nc * nr)];
            }
            m++;
        }
    }

 /* Assign cluster center image data to c_center */
 /* Assign variance values to var_vector */


   ptr = (float *) center -> imagedata;
   varptr = (float *) varimg -> imagedata;
   k = 0;
   for (i = 0; i < cc_nr * cc_nc; i++) 
   {
      if (class1[i] != 0.0)
      {
         for (j = 0; j < dimension; j++) 
         {
            c_center[k][j] = ptr[i + (cc_nc * cc_nr * j)];
            var_vector[k][j] = varptr[i + (cc_nc * cc_nr * j)];
         }
         k++;
      }
   }



/*  run the classifier */

      /* two methods: (1) use_avg = FALSE => sum the sqaure distance
                                             divided by the variance along each
                                             axis for every point.
                      (2) use_avg = TRUE => 1. sum the variance along each
                                               axis for every cluster center.
                                            2. sum the sqaure distance 
                                               divided by the (sum of the 
                                               variance along each axis) 
                                               for every point.
     */
    if (use_avg == FALSE) 
    {
       i = 0;
       for (x = border; x < nr - border; x++) {
          for (z = border; z < nc - border; z++) {
        
              dmin = KMAXFLOAT;
   
              for (j = 0; j < num_centers; j++) {
   
                      /* compute the distance along each axis */
   
                  d = 0.0;
                  for (k = 0; k < dimension; k++) {
                      a = input_vectors[i][k] - c_center[j][k];
                      q = a * a;
                      d = d + (q / var_vector[j][k]);
                  }
                  d = d / (k_factor * dimension);
   
                  if (d < dmin) {
                     dmin = d;
                     c = j;
                  }
              }
                 /* assign the classified vector */
                 new_class[x * nc + z] = class_valid[c];
              i++;
          }
       }
    }

      /* use_avg == TRUE => use the average method */

    else
    {
       for (j = 0; j < num_centers; j++) 
       {
          a = 0;
          for (k = 0; k < dimension; k++) 
          {
             a += var_vector[j][k];
          }
          var_vector[j][0] = a; 
       }

       i = 0;
       for (x = border; x < nr - border; x++) {
          for (z = border; z < nc - border; z++) {
        
   
              dmin = KMAXFLOAT;
              for (j = 0; j < num_centers; j++) {
   
                      /* compute the distance along each axis */
   
                  d = 0.0;
                  for (k = 0; k < dimension; k++) 
                  {
                      a = input_vectors[i][k] - c_center[j][k];
                      d = d + (a * a);
                  }
                  d = d / (dimension * k_factor * var_vector[j][0]);

                  if (d < dmin) 
                  {
                     dmin = d;
                     c = j;
                  }
              }
                 /* assign the classified vector */
   
                 new_class[x * nc + z] = class_valid[c];
              i++;
          }
       }
    }

    /* release memory from input image */

    (void) kfree_and_NULL (img->imagedata);
    (void) kfree_and_NULL (img->maps);

       /* adjust header in the output image */
    img->imagedata = (char *) new_class;
    img->num_data_bands = 1;
    img->map_scheme = VFF_MS_NONE;
    img->map_row_size = 0;
    img->map_col_size = 0;
    img->map_subrow_size = 0;
    img->color_space_model = VFF_CM_NONE;

    return (TRUE);
}
/* -library_code_end */
