/*
  * Khoros: $Id: loadvector.c,v 1.6 1997/01/20 07:45:20 mirage Exp $
  */

#if !defined(__lint) && !defined(__CODECENTER__)
static char rcsid[] = "Khoros: $Id: loadvector.c,v 1.6 1997/01/20 07:45:20 mirage Exp $";
#endif


/*
 * Copyright (C) 1993 - 1997, Khoral Research, Inc., ("KRI").
 * All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.
 */


/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>		File Contents Title
   >>>>		
   >>>>   Static:
   >>>>                 load_char_vector()
   >>>>                 load_short_vector()
   >>>>                 load_int_vector()
   >>>>                 load_float_vector()
   >>>>                 load_complex_vector()
   >>>>                 load_double_vector()
   >>>>                 load_dcomplex_vector()
   >>>>   Public:
   >>>>                 load_vector()
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */

#include "internals.h"

static char **load_char_vector PROTO((xvimage *, int, int *, int *));
static short **load_short_vector PROTO((xvimage *, int, int *, int *));
static int **load_int_vector PROTO((xvimage *, int, int *, int *));
static float **load_float_vector PROTO((xvimage *, int, int *, int *));
static float **load_complex_vector PROTO((xvimage *, int, int *, int *));
static double **load_double_vector PROTO((xvimage *, int, int *, int *));
static double **load_dcomplex_vector PROTO((xvimage *, int, int *, int *));

/************************************************************
*
*  Routine Name: load_vector - convert image data into vector data
*
*       Purpose: takes a multiband image and rearanges the data so
*	         that the data is stored as a list of vectors.
*	         Each pixel in a multiband image can be thought of
*	         as a vector looking down all the bands. Thus, each 
*	         pixel is a vector with dimension = number of data bands.
*	      
*                This routine will scan the image row by row making
*	         a linear 2-dimensional array, where the first array index
*                points to a vector (which corresponds to some pixel),
*	         the second array index contains the values for that
*	         pixel looking down the bands on the image.
*
*                When a border is specified a row and coluwn on each
*	         side of the image is dropped.
*	       
*
*
*         Input: image - a pointer to the image to vectorize.
*		 border - the border width to skip.
*
*        Output: num_vects - the total number of vectors loaded into the array
*	         dimension - the dimension of each vector.
*
*       Returns: a (char **) pointer to a 2-d array containing the
*		 vectorized data.
*
*  Restrictions: Restrictions on data or input as applicable
*    Written By: Tom Sauer
*          Date: Apr 5, 1990
*      Verified:
*  Side Effects:
* Modifications:
*
*************************************************************/
char **load_vector(
	xvimage *image,
	int	border,
	int	*num_vects,
	int	*dimension)

{
    char **output_vectors;
    float **float_vectors;
    double **double_vectors;
    int **int_vectors;
    short **short_vectors;
    char *program = "load_vector";

    switch(image->data_storage_type) 
    {
       case VFF_TYP_BIT:
          kfprintf(kstderr, "%s: Can not work on BIT data storage type\n",program);
          kfprintf(kstderr, "Can only convert type Byte, Short, Integer, ");
          kfprintf(kstderr, "Float\nand Complex\n");
          return(NULL);

       case VFF_TYP_1_BYTE:
          output_vectors = load_char_vector(image,border,num_vects,dimension);
          if (output_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_char_vector failed\n", program);
             return(NULL);
          }
          break;

       case VFF_TYP_2_BYTE:
          short_vectors = load_short_vector(image,border,num_vects,dimension);
          if (short_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_short_vector failed\n", program);
             return(NULL);
          }
          output_vectors = (char **) short_vectors;
          break;

       case VFF_TYP_4_BYTE:
          int_vectors = load_int_vector(image,border,num_vects,dimension);
          if (int_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_int_vector failed\n", program);
             return(NULL);
          }
          output_vectors = (char **) int_vectors;
          break;

       case VFF_TYP_FLOAT:
          float_vectors = load_float_vector(image,border,num_vects,dimension);
          if (float_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_float_vector failed\n", program);
             return(NULL);
          }
          output_vectors = (char **) float_vectors;
          break;

       case VFF_TYP_COMPLEX:
          float_vectors = load_complex_vector(image,border,num_vects,dimension);
          if (float_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_complex_vector failed\n", program);
             return(NULL);
          }

          output_vectors = (char **) float_vectors;
          break;

       case VFF_TYP_DOUBLE:
          double_vectors = load_double_vector(image,border,num_vects,dimension);
          if (double_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_double_vector failed\n", program);
             return(NULL);
          }
          output_vectors = (char **) double_vectors;
          break;

       case VFF_TYP_DCOMPLEX:
          double_vectors = load_dcomplex_vector(image,border,num_vects,
			   dimension);
          if (double_vectors == NULL)
          {
             kfprintf(kstderr, "%s: load_dcomplex_vector failed\n", program);
             return(NULL);
          }

          output_vectors = (char **) double_vectors;
          break;

       default:
          kfprintf(kstderr, "%s: Invalid image Data Storage type\n", program);
          kfprintf(kstderr, "Can only convert type Byte, Short, Integer, ");
          kfprintf(kstderr, "Float\nand Complex\n");
          return(NULL);
      }

       return(output_vectors);
}


/*-----------------------------------------------------------
|
|  Routine Name:   load_char_vector()
|
|       Purpose:     a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|
|         Input: argument1 - explanation
|                argument2 - explanation
|                argument3 - explanation
|         Input: image - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: returns a char pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By: Tom Sauer
|          Date:    Thu Apr  5 11:22:17 MDT 1990
| Modifications:
|
------------------------------------------------------------*/
static char **load_char_vector(
	xvimage	*image,
	int	border,
	int	*num_vects,
	int	*dimension)
{


    char *program = "load_char_vector";
    char **output_vectors, *cptr;
    int m, i, j, k;
    int nr, nc;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;

    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));


              /* Allocate space for output_vectors image */

    output_vectors = (char **)kmalloc((unsigned int) *num_vects * sizeof (char *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available\n",program);
	return (NULL);
    }
      /* vectorize the data, such that the border width is ignored */

    for (i = 0; i < *num_vects; i++) 
    {
       output_vectors[i] = (char *) kmalloc (*dimension * sizeof (char));
       if (output_vectors[i] == NULL) 
       {
          (void) kfprintf (kstderr,
               "%s: insufficient memory available\n",program);
           return (NULL);
       }
    }

    /* Assign image data address to ptr */

    cptr = (char *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       for (j = border; j < nc - border; j++) 
       {
	  for (k = 0; k < *dimension; k++) 
          {
	       output_vectors[m][k] = cptr[(i * nc + j) + (k * nc * nr)];
	  }
	  m++;
       }
    }
    return(output_vectors);
}

/*-----------------------------------------------------------
|
|  Routine Name:   load_short_vector()
|
|       Purpose:     a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|         Input: image  - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: returns a short pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By:  Tom Sauer
|          Date:    Thu Apr  5 11:22:17 MDT 1990
| Modifications:
|
------------------------------------------------------------*/
static short **load_short_vector(
	xvimage	*image,
	int	border,
	int	*num_vects,
	int	*dimension)
{

    char *program = "load_short_vector";
    short **output_vectors, *sptr;
    int m, i, j, k;
    int nr, nc;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;

    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));


              /* Allocate space for output_vectors image */

    output_vectors = (short **)kmalloc((unsigned int) *num_vects * sizeof (short *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available\n",program);
	return (NULL);
    }
      /* vectorize the data, such that the border width is ignored */

    for (i = 0; i < *num_vects; i++) 
    {
       output_vectors[i] = (short *) kmalloc (*dimension * sizeof (short));
       if (output_vectors[i] == NULL) 
       {
          (void) kfprintf (kstderr,
               "%s: insufficient memory available\n",program);
           return (NULL);
       }
    }

    /* Assign image data address to ptr */

    sptr = (short *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       for (j = border; j < nc - border; j++) 
       {
	  for (k = 0; k < *dimension; k++) 
          {
	       output_vectors[m][k] = sptr[(i * nc + j) + (k * nc * nr)];
	  }
	  m++;
       }
    }
    return(output_vectors);
}

/*-----------------------------------------------------------
|
|  Routine Name:   load_int_vector()
|
|       Purpose:     a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|
|         Input: image - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: a int pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By:  Tom Sauer
|          Date:  Thu Apr  5 11:22:17 MDT 1990
| Modifications:
|
------------------------------------------------------------*/
static int **load_int_vector(
	xvimage	*image,
	int	border,
	int	*num_vects,
	int	*dimension)
{

    char *program = "load_int_vector";
    int **output_vectors, *iptr;
    int m, i, j, k;
    int nr, nc;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;

    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));


              /* Allocate space for output_vectors image */

    output_vectors = (int **)kmalloc((unsigned int) *num_vects * sizeof (int *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available\n",program);
	return (NULL);
    }
      /* vectorize the data, such that the border width is ignored */

    for (i = 0; i < *num_vects; i++) 
    {
       output_vectors[i] = (int *) kmalloc (*dimension * sizeof (int));
       if (output_vectors[i] == NULL) 
       {
          (void) kfprintf (kstderr,
               "%s: insufficient memory available\n",program);
           return (NULL);
       }
    }

    /* Assign image data address to ptr */

    iptr = (int *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       for (j = border; j < nc - border; j++) 
       {
	  for (k = 0; k < *dimension; k++) 
          {
	       output_vectors[m][k] = iptr[(i * nc + j) + (k * nc * nr)];
	  }
	  m++;
       }
    }
    return(output_vectors);
}

/*-----------------------------------------------------------
|
|  Routine Name:   load_float_vector()
|
|       Purpose:     a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|         Input: image - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: a float pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By:  Tom Sauer
|          Date:    Thu Apr  5 11:22:17 MDT 1990
| Modifications:
|
------------------------------------------------------------*/
static float **load_float_vector(
	xvimage *image,
	int	border,
	int	*num_vects,
	int	*dimension)
{


    char *program = "load_float_vector";
    float **output_vectors, *fptr;
    int m, i, j, k;
    int nr, nc;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;

    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));


              /* Allocate space for output_vectors image */

    output_vectors = (float **)kmalloc((unsigned int) *num_vects * sizeof (float *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available for pointer array\n",program);
	return (NULL);
    }
      /* vectorize the data, such that the border width is ignored */

    for (i = 0; i < *num_vects; i++) 
    {
       output_vectors[i] = (float *) kmalloc (*dimension * sizeof (float));
       if (output_vectors[i] == NULL) 
       {
          (void) kfprintf (kstderr,
               "%s: insufficient memory available for additional output vector\n",program);
           return (NULL);
       }
    }

    /* Assign image data address to ptr */

    fptr = (float *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       for (j = border; j < nc - border; j++) 
       {
	  for (k = 0; k < *dimension; k++) 
          {
	       output_vectors[m][k] = fptr[(i * nc + j) + (k * nc * nr)];
	  }
	  m++;
       }
    }
    return(output_vectors);
}

/*-----------------------------------------------------------
|
|  Routine Name:   load_complex_vector()
|
|       Purpose:     takes a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|         Input: image - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: a complex pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By:  Tom Sauer
|          Date:    Thu Apr  5 11:22:17 MDT 1990
| Modifications:
|
------------------------------------------------------------*/
static float **load_complex_vector(
	xvimage	*image,
	int	border,
	int	*num_vects,
	int	*dimension)
{

    char *program = "load_complex_vector";
    float **output_vectors, *fptr;
    int m, i, j, k, p;
    int nr, nc, q;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;


    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));

              /* Allocate space for output_vectors image */

    output_vectors = (float **)kmalloc((unsigned int) *num_vects * sizeof (float *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available\n",program);
	return (NULL);
    }
    for (i = 0; i < *num_vects; i++) 
    {
        output_vectors[i] = (float *)kmalloc(*dimension * 2 * sizeof (float));
        if (output_vectors[i] == NULL) 
        {
           (void) kfprintf (kstderr,
                  "%s: insufficient memory available\n", program);
	   return (NULL);
	}
    }

    /* Assign image data address to ptr */

    fptr = (float *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       q = border*2;
       for (j = border; j < nc - border; j++ ) 
       {
          p = 0;
          for (k = 0; k < *dimension; k++) 
          {
	    output_vectors[m][p] = fptr[(i * nc*2 + q)+(k * nc * nr*2)];
            p++;
            output_vectors[m][p] = fptr[(i * nc*2 + q+1)+(k * nc *nr*2)];
            p++;
          }
          q +=2;
          m++;
       }
    }
    return(output_vectors);
}


/*-----------------------------------------------------------
|
|  Routine Name:   load_double_vector()
|
|       Purpose:     a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|         Input: image - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: a double pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By:  Tom Sauer
|          Date:    Mon Feb 25 12:43:24 MST 1991
| Modifications:
|
------------------------------------------------------------*/
static double **load_double_vector(
	xvimage	*image,
	int	border,
	int	*num_vects,
	int	*dimension)
{


    char *program = "load_double_vector";
    double **output_vectors, *dptr;
    int m, i, j, k;
    int nr, nc;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;

    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));


              /* Allocate space for output_vectors image */

    output_vectors = (double **)kmalloc((unsigned int) *num_vects * sizeof (double *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available\n",program);
	return (NULL);
    }
      /* vectorize the data, such that the border width is ignored */

    for (i = 0; i < *num_vects; i++) 
    {
       output_vectors[i] = (double *) kmalloc (*dimension * sizeof (double));
       if (output_vectors[i] == NULL) 
       {
          (void) kfprintf (kstderr,
               "%s: insufficient memory available\n",program);
           return (NULL);
       }
    }

    /* Assign image data address to ptr */

    dptr = (double *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       for (j = border; j < nc - border; j++) 
       {
	  for (k = 0; k < *dimension; k++) 
          {
	       output_vectors[m][k] = dptr[(i * nc + j) + (k * nc * nr)];
	  }
	  m++;
       }
    }
    return(output_vectors);
}

/*-----------------------------------------------------------
|
|  Routine Name:   load_dcomplex_vector()
|
|       Purpose:     takes a multiband image and rearanges the data so
|              	     that the data is stored as a list of vectors.
|                    Each pixel in a multiband image can be thought of
|                    as a vector looking down all the bands. Thus, each
|                    pixel is a vector with dimension = number of data bands.
|    
|                    This routine will scan the image row by row making
|                    a linear 2-dimensional array, where the first array index
|                    points to a vector (which corresponds to some pixel),
|                    the second array index contains the values for that
|                    pixel looking down the bands on the image.
|
|                    When a border is specified a row and coluwn on each
|                    side of the image is dropped.
|
|         Input: image - a pointer to the image to vectorize.
|                border - the border width to skip.
|
|        Output: num_vects - the total number of vectors loaded into the array
|                dimension - the dimension of each vector.
|
|       Returns: a double complex pointer to a 2-d array containing 
|		 the vectorized data.
|
|    Written By:  Tom Sauer
|          Date:    Mon Feb 25 13:12:32 MST 1991
| Modifications:
|
------------------------------------------------------------*/
static double **load_dcomplex_vector(
	xvimage *image,
	int	border,
	int	*num_vects,
	int	*dimension)
{

    char *program = "load_dcomplex_vector";
    double **output_vectors, *dptr;
    int m, i, j, k, p;
    int nr, nc, q;

    *dimension = image->num_data_bands;
    nr = image->col_size;
    nc = image->row_size;


    if ((nr - (border*2) <= 0) || (nc  - (border*2) <= 0))
    {
       (void) kfprintf (kstderr,"%s: border size is too large\n",program);
       return (NULL);
    }

        /* calculate the total number of vectors */

    *num_vects = (image->row_size-(border*2)) * (image->col_size-(border*2));

              /* Allocate space for output_vectors image */

    output_vectors = (double **)kmalloc((unsigned int) *num_vects * sizeof (double *));
    if (output_vectors == NULL) 
    {
        (void) kfprintf (kstderr, "%s: insufficient memory available\n",program);
	return (NULL);
    }
    for (i = 0; i < *num_vects; i++) 
    {
        output_vectors[i] = (double *)kmalloc(*dimension * 2 * sizeof (double));
        if (output_vectors[i] == NULL) 
        {
           (void) kfprintf (kstderr,
                  "%s: insufficient memory available\n", program);
	   return (NULL);
	}
    }

    /* Assign image data address to ptr */

    dptr = (double *) image->imagedata;

          /* Assign ptr to output vectors minus the border */

    m = 0;
    for (i = border; i < nr - border; i++) 
    {
       q = border*2;
       for (j = border; j < nc - border; j++ ) 
       {
          p = 0;
          for (k = 0; k < *dimension; k++) 
          {
	    output_vectors[m][p] = dptr[(i * nc*2 + q)+(k * nc * nr*2)];
            p++;
            output_vectors[m][p] = dptr[(i * nc*2 + q+1)+(k * nc *nr*2)];
            p++;
          }
          q +=2;
          m++;
       }
    }
    return(output_vectors);
}
